/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file     cpsw_lld_if.c
 *
 *  \brief    CPSW interface private source file.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <hw/inout.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <hw/nicinfo.h>
#include <io-pkt/iopkt_driver.h>
#include <sys/io-pkt.h>

#include <ti/osal/osal.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/enet/enet.h>

#if defined (CPSW9G)
#include <ti/drv/ipc/ipc.h>
#include <ti/csl/cslr_gtc.h>
#endif /* defined (CPSW9G)*/

#include "j7_cpsw.h"
#include "enetlld_if_utils.h"
#include "enetlld_if_memutils.h"
#include "enetlld_if.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#if defined (CPSW9G)
  #define IPC_RPMESSAGE_OBJ_SIZE  256

  #define RPMSG_DATA_SIZE         (256*512 + IPC_RPMESSAGE_OBJ_SIZE)
  #define VQ_BUF_SIZE             2048
  #define REMOTE_DEVICE_ENDPT     (26)

  #define VRING_BASE_ADDRESS      0xAA000000

  #define CORE_IN_TEST            9

  #define CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL   (28U)
  #define CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM      (2U)
#endif /* defined (CPSW9G)*/

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* CPSW configuration */
EnetIf_Obj gEnetIfObj;

/* devnp driver object */
extern struct cpsw_dev *g_cpsw;

#if defined (CPSW9G)
uint8_t  g_taskStackBuf[IPC_TASK_STACKSIZE];

uint8_t  gCntrlBuf[RPMSG_DATA_SIZE] __attribute__ ((aligned (8)));
uint8_t  sysVqBuf[VQ_BUF_SIZE]  __attribute__ ((aligned (8)));
uint8_t *pCntrlBuf = gCntrlBuf;
uint8_t *pTaskBuf = g_taskStackBuf;
uint8_t *pSysVqBuf = sysVqBuf;

uint32_t selfProcId = IPC_MPU1_0;
uint32_t remoteProc[] =
{
    IPC_MCU1_0, IPC_MCU1_1, IPC_MCU2_0, IPC_MCU2_1, IPC_MCU3_0, IPC_MCU3_1, IPC_C66X_1, IPC_C66X_2, IPC_C7X_1
};

uint32_t *pRemoteProcArray = remoteProc;
uint32_t  gNumRemoteProc = sizeof(remoteProc)/sizeof(uint32_t);
uint32_t remoteProcNum[IPC_MAX_PROCS];
uint32_t rpmsgDataSize = RPMSG_DATA_SIZE;
#endif /* defined (CPSW9G) */

EnetDma_Pkt gPktInfoMem[CPSW_MEM_TX_PKTS_MAX + CPSW_MEM_RX_PKTS_MAX];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
#if defined (CPSW9G)
void appLogPrintf(const char *format, ...)
{
    char printBuffer[256];
    va_list arguments;

    if (256 < strlen(format))
    {
        assert(false);
    }

    /* Start the varargs processing */
    va_start(arguments, format);
    vsnprintf(printBuffer, sizeof(printBuffer), format, arguments);

    slogf(_SLOGC_NETWORK, _SLOG_INFO, printBuffer);

    /* End the varargs processing */
    va_end(arguments);
}

static void printDevInfo(struct rpmsg_kdrv_ethswitch_device_data *ethDevData)
{
    char *tf[] = {"false", "true"};

    EnetIf_print("ETHFW Version:%2d.%2d.%2d\n",
                  ethDevData->fw_ver.major,
                  ethDevData->fw_ver.minor,
                  ethDevData->fw_ver.rev);
    EnetIf_print("ETHFW Build Date (YYYY/MMM/DD):%c%c%c%c/%c%c%c/%c%c\n",
                  ethDevData->fw_ver.year[0], ethDevData->fw_ver.year[1], ethDevData->fw_ver.year[2], ethDevData->fw_ver.year[3],
                  ethDevData->fw_ver.month[0], ethDevData->fw_ver.month[1], ethDevData->fw_ver.month[2],
                  ethDevData->fw_ver.date[0], ethDevData->fw_ver.date[1]);
    EnetIf_print("ETHFW Commit SHA:%c%c%c%c%c%c%c%c\n",
                  ethDevData->fw_ver.commit_hash[0],
                  ethDevData->fw_ver.commit_hash[1],
                  ethDevData->fw_ver.commit_hash[2],
                  ethDevData->fw_ver.commit_hash[3],
                  ethDevData->fw_ver.commit_hash[4],
                  ethDevData->fw_ver.commit_hash[5],
                  ethDevData->fw_ver.commit_hash[6],
                  ethDevData->fw_ver.commit_hash[7]);
    EnetIf_print("ETHFW PermissionFlag:0x%x, UART Connected:%s,UART Id:%d",
                  ethDevData->permission_flags,
                  tf[ethDevData->uart_connected],
                  ethDevData->uart_id);
}
static CpswProxy_Handle CpswRemoteApp_initCpswProxy(void)
{
    CpswProxy_Config proxyConfig;
    CpswProxy_Handle hProxy;

    strncpy(proxyConfig.device_name, ETHREMOTEDEVICE_DEVICE_NAME_MPU_1_0, (sizeof(proxyConfig.device_name) - 1));
    proxyConfig.device_name[(sizeof(proxyConfig.device_name) - 1)] = 0;
    proxyConfig.deviceDataNotifyCb = &printDevInfo;
    proxyConfig.masterCoreId = IPC_MCU2_0;
    proxyConfig.rpmsgEndPointId = REMOTE_DEVICE_ENDPT;

    hProxy = CpswProxy_init(&proxyConfig);

    return hProxy;
}

static uint64_t EnetIf_getLocalTime(uint64_t baseAddress)
{
    uint64_t gtcTimeLo = 0UL, gtcTimeHi = 0UL;
    uint64_t gtcTime = 0UL;

    gtcTimeLo = (uint64_t) CSL_REG32_RD(baseAddress + CSL_GTC_CFG1_CNTCV_LO);
    gtcTimeHi = (uint64_t) CSL_REG32_RD(baseAddress + CSL_GTC_CFG1_CNTCV_HI);
    gtcTime = (gtcTimeHi << 32U) | gtcTimeLo;

    return gtcTime;
}

static uint64_t EnetIf_getSynchronizedTime(EnetIf_SyncTimerObj *hSyncTimerObj)
{
    uint64_t gtcTime = 0UL, synchronizedTime = 0UL;

    /* Get GTC time */
    gtcTime = EnetIf_getLocalTime((uint64_t) hSyncTimerObj->GTC0_GTC_CFG1_Base);

    /* Compute synchronized time from GTC time */
    synchronizedTime = (uint64_t)((hSyncTimerObj->rate * (double)gtcTime) + hSyncTimerObj->offset);

    return synchronizedTime;
}

static void EnetIf_calcSyncTimeParams(CpswCpts_HwPush hwPushNum, uint64_t syncTime)
{
    uint64_t gtcTime = 0U;
    EnetIf_SyncTimerObj *hSyncTimerObj = (EnetIf_SyncTimerObj*)&gEnetIfObj.syncTimerObj;

    if (hwPushNum == CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM)
    {
        if(hSyncTimerObj->prevLocalTime == 0U)
        {
            /* Set GTC time */
            gtcTime = EnetIf_getLocalTime((uint64_t) gEnetIfObj.syncTimerObj.GTC0_GTC_CFG1_Base) &
                                            (0xFFFFFFFFFFFFFFFFUL << CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);
        }
        else
        {
            /* Increment GTC time used for computation based on selected bit for event */
            gtcTime = hSyncTimerObj->prevLocalTime + (1UL << (CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL+1));
        }

        if ((hSyncTimerObj->prevLocalTime != 0U) &&
            (hSyncTimerObj->prevCptsTime != 0U))
        {
            /* Logic:
             *  T1, T2 - Previous & Current CPTS time
             *  t1, t2 - Previous & Current local GTC time
             *  rate = (T2-T1) / (t2-t1)
             *  offset = T2 - rate*t2
             */
            hSyncTimerObj->rate = (double)((double)(syncTime - hSyncTimerObj->prevCptsTime) /
                                          (double)(gtcTime - hSyncTimerObj->prevLocalTime));
            hSyncTimerObj->offset = (double) syncTime - (hSyncTimerObj->rate * (double) gtcTime);
        }

        hSyncTimerObj->prevLocalTime = gtcTime;
        hSyncTimerObj->prevCptsTime = syncTime;
    }
}

#endif /* defined (CPSW9G) */

void EnetIf_InitObj()
{
    /* Calling this so that the EnetSoc lib gets linked */
    EnetSoc_init();

#if defined (CPSW2G)

    memset(&gEnetIfObj, 0, sizeof(EnetIf_Obj));

    gEnetIfObj.enetType = ENET_CPSW_2G;
    gEnetIfObj.instId = 0;
    gEnetIfObj.macPort = ENET_MAC_PORT_1;
    gEnetIfObj.coreId = IPC_MPU1_0;
    gEnetIfObj.macMode = RGMII;
    gEnetIfObj.preferredUdmaChannel = UDMA_DMA_CH_ANY;
    gEnetIfObj.numTxPkts = CPSW_MEM_TX_PKTS_DEFAULT;
    gEnetIfObj.numRxPkts = CPSW_MEM_RX_PKTS_DEFAULT;

    EnetIfBoard_cpsw2gMacModeConfig(gEnetIfObj.macPort, gEnetIfObj.macMode);

    EnetIf_selectCptsClock(gEnetIfObj.enetType, ENETIF_CPTS_CLKSEL_MAIN_SYSCLK0);

#elif defined (CPSW9G)

    uint32_t numProc = gNumRemoteProc;
    Ipc_VirtIoParams vqParam;

    memset(&gEnetIfObj, 0, sizeof(EnetIf_Obj));

    gEnetIfObj.macPort = ENET_MAC_PORT_1;
    gEnetIfObj.coreId = IPC_MPU1_0;
    gEnetIfObj.enetType = ENET_CPSW_9G;
    gEnetIfObj.instId = 0;
    gEnetIfObj.preferredUdmaChannel = UDMA_DMA_CH_ANY;
    gEnetIfObj.numTxPkts = CPSW_MEM_TX_PKTS_DEFAULT;
    gEnetIfObj.numRxPkts = CPSW_MEM_RX_PKTS_DEFAULT;

    /* Step1 : Initialize the multiproc */
    Ipc_mpSetConfig(selfProcId, numProc, pRemoteProcArray);

    EnetIf_print("CPSW9G Client (core : %s) .....\r\n",
         Ipc_mpGetSelfName());

    Ipc_init(NULL);

    /* Step2 : Initialize Virtio */
    vqParam.vqObjBaseAddr = (void*)pSysVqBuf;
    vqParam.vqBufSize     = numProc * Ipc_getVqObjMemoryRequiredPerCore();
    vqParam.vringBaseAddr = (void*)VRING_BASE_ADDRESS;
    vqParam.vringBufSize  = IPC_VRING_BUFFER_SIZE;
    vqParam.timeoutCnt    = 100;  /* Wait for counts */
    Ipc_initVirtIO(&vqParam);

    /* Step 3: Initialize RPMessage */
    RPMessage_Params cntrlParam;

    /* Initialize the param */
    RPMessageParams_init(&cntrlParam);

    /* Set memory for HeapMemory for control task */
    cntrlParam.buf         = pCntrlBuf;
    cntrlParam.bufSize     = rpmsgDataSize;
    cntrlParam.stackBuffer = pTaskBuf;
    cntrlParam.stackSize   = IPC_TASK_STACKSIZE;
    RPMessage_init(&cntrlParam);

    /* Step 5: Start Cpsw Proxy */
    gEnetIfObj.hCpswProxy = CpswRemoteApp_initCpswProxy();
    gEnetIfObj.enetType = ENET_CPSW_9G;
    gEnetIfObj.macPort = ENET_MAC_PORT_1;
    gEnetIfObj.coreId = IPC_MPU1_0;

    EnetIf_print("CpswProxy_start .....\r\n");
    CpswProxy_start(gEnetIfObj.hCpswProxy);

    CpswProxy_attach(gEnetIfObj.hCpswProxy,
                   gEnetIfObj.enetType,
                   &gEnetIfObj.hEnet,
                   &gEnetIfObj.coreKey,
                   &gEnetIfObj.rxMtu,
                   gEnetIfObj.txMtu);
    CpswProxy_allocMac(gEnetIfObj.hCpswProxy,
                       gEnetIfObj.hEnet,
                       gEnetIfObj.coreKey,
                       gEnetIfObj.hostMacAddr);
#else
    #error "Unsupported!"
#endif

    return;
}

void EnetIf_SetPerferredDmaChNum(uint32_t preferred_udma_channel)
{
    gEnetIfObj.preferredUdmaChannel = preferred_udma_channel;
}

void EnetIf_SetDescriptorCount(uint32_t tx_descriptors_count,
                               uint32_t rx_descriptors_count)
{
    gEnetIfObj.numTxPkts = tx_descriptors_count;
    gEnetIfObj.numRxPkts = rx_descriptors_count;
}

int32_t EnetIf_GetMacAddr(uint8_t macAddr[][ENET_MAC_ADDR_LEN])
{
#if defined (CPSW2G)
    uint32_t num = 1;
    EnetSoc_getEFusedMacAddrs(macAddr, &num);
#elif defined (CPSW9G)
    memcpy(macAddr[0], gEnetIfObj.hostMacAddr, ETH_MAC_ADDR_LEN);
    EnetIf_print("MAC Address: %x:%x:%x:%x:%x:%x\n", macAddr[0][0],macAddr[0][1],
                                               macAddr[0][2],macAddr[0][3],
                                               macAddr[0][4],macAddr[0][5]);
#else
    #error "Unsupported!"
#endif
    return ENET_SOK;
}

static void CpswCptsEvent_setDefaultPortEventPrms(
    CpswMacPort_TsEventCfg *tsPortEventCfg)
{
    tsPortEventCfg->commonPortIpCfg.ttlNonzeroEn = true;
    tsPortEventCfg->commonPortIpCfg.tsIp107En    = true;
    tsPortEventCfg->commonPortIpCfg.tsIp129En    = true;
    tsPortEventCfg->commonPortIpCfg.tsIp130En    = true;
    tsPortEventCfg->commonPortIpCfg.tsIp131En    = true;
    tsPortEventCfg->commonPortIpCfg.tsIp132En    = true;
    tsPortEventCfg->commonPortIpCfg.tsIp132En    = true;
    tsPortEventCfg->commonPortIpCfg.tsPort320En  = true;
    tsPortEventCfg->commonPortIpCfg.unicastEn    = false;
    tsPortEventCfg->domainOffset                 = 4U;
    tsPortEventCfg->ltype2En                     = false;
    tsPortEventCfg->rxAnnexDEn                   = true;
    tsPortEventCfg->rxAnnexEEn                   = true;
    tsPortEventCfg->rxAnnexFEn                   = true;
    tsPortEventCfg->txAnnexDEn                   = true;
    tsPortEventCfg->txAnnexEEn                   = true;
    tsPortEventCfg->txAnnexFEn                   = true;
    tsPortEventCfg->txHostTsEn                   = true;
    tsPortEventCfg->mcastType                    = 0U;
    tsPortEventCfg->messageType                  = 0xFFFFU;
    tsPortEventCfg->rxVlanType                   = ENET_MACPORT_VLAN_TYPE_SINGLE_TAG;
    tsPortEventCfg->seqIdOffset                  = 30U;
    tsPortEventCfg->txVlanType                   = ENET_MACPORT_VLAN_TYPE_SINGLE_TAG;
    tsPortEventCfg->vlanLType1                   = 0x8100;
    tsPortEventCfg->vlanLType2                   = 0U;
}

int32_t EnetIf_PtpInit()
{
    int32_t status = ENET_SOK;
    Enet_IoctlPrms prms;
    CpswCpts_RegisterStackInArgs registerStackInArgs;
    CpswMacPort_EnableTsEventInArgs enableTsEventInArgs;

    EnetIf_print("%s:%d PTP", __FUNCTION__, __LINE__);

    /* CPTS */
    if (status == ENET_SOK)
    {
        registerStackInArgs.eventNotifyCb = EnetIf_EventNotifyCallback;
        registerStackInArgs.eventNotifyCbArg = &gEnetIfObj;

        ENET_IOCTL_SET_IN_ARGS(&prms, &registerStackInArgs);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                    gEnetIfObj.coreId,
                    CPSW_CPTS_IOCTL_REGISTER_STACK,
                    &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: failed CPSW_CPTS_IOCTL_REGISTER_STACK: %d",
                __FUNCTION__, __LINE__, status);
        }
        else {
            EnetIf_print("%s:%d CPSW_CPTS_IOCTL_REGISTER_STACK is sucessful",
                __FUNCTION__, __LINE__);
        }
    }

    if (status == ENET_SOK)
    {
        /* Enable PTP over Bare Ethernet - Annex F (IEEE802.3) */
        enableTsEventInArgs.macPort = ENET_MAC_PORT_FIRST;
        CpswCptsEvent_setDefaultPortEventPrms(&enableTsEventInArgs.tsEventCfg);

        ENET_IOCTL_SET_IN_ARGS(&prms, &enableTsEventInArgs);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT,
                            &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: failed CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    return status;
}

void EnetIf_EventNotifyCallback(void *hEventNotifyCallbackArg,
                                 CpswCpts_Event *eventInfo)
{
    /* Do Nothing */
    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        EnetIf_print("Entered cpts Callback eventtype=%d, ts=%lld, msgType=%d, seqId=%d ",
            eventInfo->eventType, eventInfo->tsVal, eventInfo->msgType, eventInfo->seqId);
    }
}

int32_t EnetIf_GetTime(uint64_t *ts)
{
    int32_t status = ENET_SOK;
#if defined (CPSW2G)
    Enet_IoctlPrms prms;

    if (gEnetIfObj.hEnet == NULL)
    {
        return ENET_ENOTFOUND;
    }

    if (ts == NULL)
    {
        return ENET_EBADARGS;
    }

    *ts = 0;
    ENET_IOCTL_SET_OUT_ARGS(&prms, ts);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_TIMESYNC_IOCTL_GET_CURRENT_TIMESTAMP,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed ENET_TIMESYNC_IOCTL_GET_CURRENT_TIMESTAMP: %d",
            __FUNCTION__, __LINE__, status);
    }
    else {
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            EnetIf_print("%s:%d The timestamp is: %lld", __FUNCTION__, __LINE__, *ts);
        }
    }
#elif defined (CPSW9G)
    if (ts == NULL)
    {
        return ENET_EBADARGS;
    }

    *ts = EnetIf_getSynchronizedTime(&gEnetIfObj.syncTimerObj);
#endif
    return status;
}

int32_t EnetIf_SetTime(uint64_t ts)
{
    int32_t status = ENET_SOK;
    Enet_IoctlPrms prms;

    if (gEnetIfObj.hEnet == NULL)
    {
        return ENET_ENOTFOUND;
    }

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        EnetIf_print("%s:%d set timestamp is: %lld", __FUNCTION__, __LINE__, ts);
    }

    ENET_IOCTL_SET_IN_ARGS(&prms, &ts);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_TIMESYNC_IOCTL_SET_TIMESTAMP,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed ENET_TIMESYNC_IOCTL_SET_TIMESTAMP: %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t EnetIf_GetTimestamp(uint8_t msgType,
                             uint16_t sequenceId,
                             uint8_t tx,
                             uint64_t *ts)
{
    int32_t status = ENET_SOK;
    CpswCpts_Event lookUpEventInArgs;
    CpswCpts_Event lookUpEventOutArgs;
    Enet_IoctlPrms prms;

    if (gEnetIfObj.hEnet == NULL)
    {
        return ENET_ENOTFOUND;
    }

    if (ts == NULL)
    {
        return ENET_EBADARGS;
    }

    *ts = 0;
    if (tx)
    {
        lookUpEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT;
    }
    else
    {
        lookUpEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_RECEIVE;
    }
    lookUpEventInArgs.msgType = msgType;
    lookUpEventInArgs.seqId = sequenceId;
    lookUpEventInArgs.portNum = gEnetIfObj.macPort;
    lookUpEventInArgs.domain = 0;

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        EnetIf_print("%s:%d PTP_GET_TX_TIMESTAMP/PTP_GET_RX_TIMESTAMP",
                __FUNCTION__, __LINE__);
        EnetIf_print("%s:%d msg_type:%d, sequence_id:%d, tx:%d",
                __FUNCTION__, __LINE__, msgType, sequenceId, tx);
    }
    ENET_IOCTL_SET_INOUT_ARGS(&prms, &lookUpEventInArgs, &lookUpEventOutArgs);

    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        CPSW_CPTS_IOCTL_LOOKUP_EVENT,
                        &prms);
    if (status != ENET_SOK)
    {
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            EnetIf_print("%s:%d Error: failed, no retries, CPSW_CPTS_IOCTL_LOOKUP_EVENT: %d",
                __FUNCTION__, __LINE__, status);
        }
    }
    else
    {
        *ts = lookUpEventOutArgs.tsVal;
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            EnetIf_print("%s:%d The CPSW_CPTS_IOCTL_LOOKUP_EVENT timestamp is: %lld",
                __FUNCTION__, __LINE__, *ts);
        }
    }

    return status;
}

int32_t EnetIf_SetCompensation(int32_t nudge)
{
    int32_t status = ENET_SOK;
    Enet_IoctlPrms prms;

    if (gEnetIfObj.hEnet == NULL)
    {
        return ENET_ENOTFOUND;
    }

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        EnetIf_print("%s:%d nudge is: %d", __FUNCTION__, __LINE__, nudge);
    }

    if (nudge > 127 || nudge < -128)
    {
        uint64_t ts = 0;
        EnetIf_GetTime(&ts);

        ts -= nudge;

        EnetIf_SetTime(ts);
        return ENET_SOK;
    }

    ENET_IOCTL_SET_IN_ARGS(&prms, &nudge);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        CPSW_CPTS_IOCTL_SET_TS_NUDGE,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed CPSW_CPTS_IOCTL_SET_TS_NUDGE: %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}


int32_t EnetIf_Init(int mac_to_mac, int speed, uint8_t *currentMacAddr)
{
    int32_t status = ENET_SOK;
#if defined (CPSW2G)
    Enet_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;

#if 0
    EnetAppUtils_enableClocks(gEnetIfObj.enetType,
                              MAC_CONN_TYPE_RGMII_FORCE_1000_FULL);
#endif
#endif

    /* Open the CPSW */
    status = EnetIf_openCpsw(mac_to_mac, speed, currentMacAddr);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: Failed to open CPSW: %d",
            __FUNCTION__, __LINE__, status);
    }

    /* Open CPSW DMA driver */
    if (status == ENET_SOK)
    {
        status = EnetIf_openDma();
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: Failed to open CPSW DMA: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

#if defined (CPSW2G)
    /* Enable host port */
    if (status == ENET_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        ENET_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
        prms.outArgs = NULL;
        status       = Enet_ioctl(gEnetIfObj.hEnet,
                                  gEnetIfObj.coreId,
                                  CPSW_ALE_IOCTL_SET_PORT_STATE,
                                  &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print( "%s:%d Error: Failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d",
                __FUNCTION__, __LINE__, status);
        }

        if (status == ENET_SOK)
        {
            ENET_IOCTL_SET_NO_ARGS(&prms);
            status = Enet_ioctl(gEnetIfObj.hEnet,
                                gEnetIfObj.coreId,
                                ENET_HOSTPORT_IOCTL_ENABLE,
                                &prms);
            if (status != ENET_SOK)
            {
                EnetIf_print("%s:%d Error: Failed to enable host port: %d",
                    __FUNCTION__, __LINE__, status);
            }
        }
    }
#endif
    return status;
}

int32_t EnetIf_InitPhy()
{
#if defined (CPSW2G)
    int32_t status = ENET_SOK;
    Enet_IoctlPrms prms;
    int8_t         i;
    bool           alive;

    /* Show alive PHYs */
    if (status == ENET_SOK)
    {
        for (i = 0U; i < ENET_MDIO_PHY_CNT_MAX; i++)
        {
            ENET_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Enet_ioctl(gEnetIfObj.hEnet,
                                gEnetIfObj.coreId,
                                ENET_MDIO_IOCTL_IS_ALIVE,
                                &prms);
            if (status == ENET_SOK)
            {
                if (alive == true)
                {
                    EnetIf_print("PHY %d is alive", i);
                }
            }
            else
            {
                EnetIf_print("%s:%d Error: Failed to get PHY %d alive status: %d",
                    __FUNCTION__, __LINE__, i, status);
            }
        }
    }
    return status;
#elif defined (CPSW9G)
    return ENET_SOK;
#endif
}


int32_t EnetIf_CheckPortLinkUp(uint32_t timeout_in_sec)
{
    bool linked = false;
    int32_t timeout_cnt = (timeout_in_sec * 1000 / 100);

    do {
#if defined (CPSW2G)
        linked = EnetIf_isPortLinkUp(gEnetIfObj.hEnet,
                                     gEnetIfObj.coreId,
                                     gEnetIfObj.macPort);
#elif defined (CPSW9G)
        linked = CpswProxy_isPhyLinked(gEnetIfObj.hCpswProxy,
                                      gEnetIfObj.hEnet,
                                      gEnetIfObj.coreKey,
                                      gEnetIfObj.macPort);
#endif
        if ((linked) || (!timeout_cnt)) {
            break;
        }
        EnetIf_wait(100);
        Enet_periodicTick(gEnetIfObj.hEnet);
        timeout_cnt--;
    } while (timeout_cnt);

    if (!linked) {
        return ENET_EFAIL;
    }
    return ENET_SOK;
}


int32_t EnetIf_Close()
{
#if defined (CPSW2G)
    int32_t status = ENET_SOK;
    Enet_IoctlPrms prms;

    /* Retreive any pending packets */
    EnetIf_retrieveFreeTxPkts();

    /* Print CPSW statistics of all ports */
    if (status == ENET_SOK)
    {
        EnetIf_showStats();

        EnetIf_print("rxReadyQ count = %d", EnetQueue_getQCount(&gEnetIfObj.rxReadyQ));
        EnetIf_print("freePktInfoQ count = %d", EnetQueue_getQCount(&gEnetIfObj.freePktInfoQ));
    }

    /* Disable host port */
    if (status == ENET_SOK)
    {
        ENET_IOCTL_SET_NO_ARGS(&prms);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            ENET_HOSTPORT_IOCTL_DISABLE,
                            &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: Failed to disable host port: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    /* De-register ptp */
    if (status == ENET_SOK)
    {
        ENET_IOCTL_SET_NO_ARGS(&prms);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            CPSW_CPTS_IOCTL_UNREGISTER_STACK,
                            &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: Failed to de-register ptp: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    /* Print DMA statistics */
    if (status == ENET_SOK)
    {
        /* Print RX flow DMA statistics */
        EnetIf_showRxFlowStats(gEnetIfObj.hRxFlow);
        /* Print TX channel DMA statistics */
        EnetIf_showTxChStats(gEnetIfObj.hTxCh);
    }

    /* Retreive any pending packets */
    EnetIf_retrieveFreeTxPkts();

    /* Close CPSW DMA driver */
    EnetIf_closeDma();

    /* Close CPSW */
    EnetIf_closeCpsw();

    /* Close UDMA driver */
    EnetIf_udmaclose(gEnetIfObj.hUdmaDrv);

    pthread_mutex_destroy(&gEnetIfObj.cpswMutex);

    gEnetIfObj.hEnet = NULL;
    EnetIf_print("cpsw deinit done");

#if 0
    /* Disable peripheral clocks */
    EnetAppUtils_disableClocks(gEnetIfObj.enetType);
#endif

    /* Deinit Enet driver */
    Enet_deinit();

    return status;
#elif defined (CPSW9G)
    return ENET_SOK;
#endif
}

uintptr_t EnetIf_disableAllIntr(void)
{
    pthread_mutex_lock(&gEnetIfObj.cpswMutex);
    return 0;
}

void EnetIf_restoreAllIntr(uintptr_t key)
{
    pthread_mutex_unlock(&gEnetIfObj.cpswMutex);
}

int32_t EnetIf_openCpsw(int mac_to_mac, int speed, uint8_t *currentMacAddr)
{
    Cpsw_Cfg cpswCfg;
    EnetOsal_Cfg osalCfg;
    EnetUtils_Cfg utilsCfg;
    EnetUdma_Cfg dmaCfg;
#if defined (CPSW2G)
    Enet_IoctlPrms prms;
    EnetPer_PortLinkCfg portLinkCfg;
    CpswMacPort_Cfg macCfg;
#endif
    int32_t status = ENET_SOK;

    // set the mutex with resursive attribute set
    pthread_mutexattr_t mattr;
    pthread_mutexattr_init(&mattr);
    pthread_mutexattr_setrecursive(&mattr, PTHREAD_RECURSIVE_ENABLE);
    pthread_mutex_init(&gEnetIfObj.cpswMutex, &mattr);
    pthread_mutexattr_destroy (&mattr);

    /* Initialize Enet driver (use default OSAL and utils) */
    Enet_initOsalCfg(&osalCfg);
    Enet_initUtilsCfg(&utilsCfg);

    /* Utils Overrides */
    utilsCfg.print     = (Enet_Print)EnetIf_print;
    utilsCfg.physToVirt = (Enet_PhysToVirt)EnetIf_phyToVirtFxn;
    utilsCfg.virtToPhys = (Enet_VirtToPhys)EnetIf_virtToPhyFxn;

    //Note: Override the osal api to disableAllIntr/restoreAllIntr
    osalCfg.disableAllIntr = &EnetIf_disableAllIntr;
    osalCfg.restoreAllIntr = &EnetIf_restoreAllIntr;

    Enet_init(&osalCfg, &utilsCfg);

    /********************/
    /* Open Enet driver */
    /********************/

    /* Set configuration parameters */

#if defined (CPSW9G)
    /* TODO: May be ignored for 9G */
#endif
    /* Set initial config */
    Enet_initCfg(gEnetIfObj.enetType, gEnetIfObj.instId, &cpswCfg, sizeof(cpswCfg));

    /* Peripheral config */
    cpswCfg.vlanCfg.vlanAware = false;

    /* Host port config */
    cpswCfg.hostPortCfg.removeCrc      = true;
    cpswCfg.hostPortCfg.padShortPacket = true;
    cpswCfg.hostPortCfg.passCrcErrors  = true;

    /* ALE config */
    cpswCfg.aleCfg.modeFlags                          = CPSW_ALE_CFG_MODULE_EN;
    cpswCfg.aleCfg.agingCfg.autoAgingEn               = true;
    cpswCfg.aleCfg.agingCfg.agingPeriodInMs           = 1000;
    cpswCfg.aleCfg.nwSecCfg.vid0ModeEn                = true;
    cpswCfg.aleCfg.vlanCfg.aleVlanAwareMode           = false;
    cpswCfg.aleCfg.vlanCfg.cpswVlanAwareMode          = false;
    cpswCfg.aleCfg.vlanCfg.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    cpswCfg.aleCfg.vlanCfg.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    cpswCfg.aleCfg.vlanCfg.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
    /* QNX: override based on user need */
    if (g_cpsw->promiscious)
    {
        cpswCfg.aleCfg.modeFlags = CPSW_ALE_CFG_BYPASS_EN;
    }

#if defined (CPSW2G)
    /* CPTS config */
    cpswCfg.cptsCfg.hostRxTsEn = true;
    cpswCfg.cptsCfg.cptsRftClkFreq = CPSW_CPTS_RFTCLK_FREQ_500MHZ; /* Setting cpts refclk frequency */

    EnetIf_initResourceConfig(gEnetIfObj.enetType,
                               gEnetIfObj.coreId,
                               &cpswCfg.resCfg,
                               currentMacAddr);
#else
    /* CPTS config */
    cpswCfg.cptsCfg.hostRxTsEn = false;
#endif

    if (gEnetIfObj.enetType == ENET_CPSW_9G)
    {
        EnetIf_print("ENET_CPSW_9G on MAIN NAVSS");
    }
    else if (gEnetIfObj.enetType == ENET_CPSW_2G)
    {
        EnetIf_print("ENET_CPSW_2G on MCU NAVSS");
    }

    dmaCfg.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;
    // Note: Using preferred channel
    dmaCfg.rxChInitPrms.preferredChNum = gEnetIfObj.preferredUdmaChannel;

    /* App should open UDMA first as UDMA handle is needed to initialize
     * CPSW RX channel */
    gEnetIfObj.hUdmaDrv = EnetIf_udmaOpen(gEnetIfObj.enetType, NULL);
    EnetIf_assert(NULL != gEnetIfObj.hUdmaDrv);

    dmaCfg.hUdmaDrv = gEnetIfObj.hUdmaDrv;
    cpswCfg.dmaCfg = &dmaCfg;

#if defined (CPSW2G)
    /* Set Enet global runtime log level */
    Enet_setTraceLevel(ENET_TRACE_DEBUG);

    /* Open the Enet driver */
    gEnetIfObj.hEnet = Enet_open(gEnetIfObj.enetType, gEnetIfObj.instId, &cpswCfg, sizeof(cpswCfg));
    if (gEnetIfObj.hEnet == NULL)
    {
        EnetIf_print("%s:%d Error: Failed to open: %d",
            __FUNCTION__, __LINE__, status);
        status = ENET_EFAIL;
    }

    if (status == ENET_SOK)
    {
        EnetMacPort_LinkCfg *linkCfg = &portLinkCfg.linkCfg;
        EnetMacPort_Interface *mii   = &portLinkCfg.mii;
        EnetPhy_Cfg *phyCfg          = &portLinkCfg.phyCfg;

        portLinkCfg.macPort = gEnetIfObj.macPort;
        portLinkCfg.macCfg = &macCfg;

        CpswMacPort_initCfg(&macCfg);

        /* Note: we currently support macmode=RGMII only */
        switch (gEnetIfObj.macMode)
        {
            case RMII:
                mii->layerType    = ENET_MAC_LAYER_MII;
                mii->sublayerType = ENET_MAC_SUBLAYER_REDUCED;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            case RGMII:
                mii->layerType    = ENET_MAC_LAYER_GMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_REDUCED;
                mii->variantType  = ENET_MAC_VARIANT_FORCED;
                break;
            case GMII:
                mii->layerType    = ENET_MAC_LAYER_GMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_STANDARD;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            case SGMII:
                mii->layerType    = ENET_MAC_LAYER_GMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_SERIAL;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            case QSGMII:
                mii->layerType    = ENET_MAC_LAYER_GMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_QUAD_SERIAL_MAIN;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            case QSGMII_SUB:
                mii->layerType    = ENET_MAC_LAYER_GMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_QUAD_SERIAL_SUB;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            case XFI:
                mii->layerType    = ENET_MAC_LAYER_XGMII;
                mii->sublayerType = ENET_MAC_SUBLAYER_STANDARD;
                mii->variantType  = ENET_MAC_VARIANT_NONE;
                break;

            default:
                EnetIf_print("%s:%d Error: Invalid MAC mode: %u\n",
                    __FUNCTION__, __LINE__, gEnetIfObj.macMode);
                EnetIf_assert(false);
                break;
        }

        if (mac_to_mac == 1) {
            /* mac-to-mac configuration -> no phy */
            phyCfg->phyAddr   = ENETPHY_INVALID_PHYADDR;

            if (speed == 100) {
                linkCfg->speed     = ENET_SPEED_100MBIT;
            }
            else if (speed == 1000) {
                linkCfg->speed     = ENET_SPEED_1GBIT;
            }
            linkCfg->duplexity = ENET_DUPLEX_FULL;
        }
        else
        {
            EnetIfBoard_setPhyConfig(gEnetIfObj.enetType,
                                       portLinkCfg.macPort,
                                       mii,
                                       phyCfg);

            linkCfg->speed     = ENET_SPEED_AUTO;
            linkCfg->duplexity = ENET_DUPLEX_AUTO;
        }

        ENET_IOCTL_SET_IN_ARGS(&prms, &portLinkCfg);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            ENET_PER_IOCTL_OPEN_PORT_LINK,
                            &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: failed to open MAC port: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    if (status == ENET_SOK)
    {
        uint32_t coreId;
        EnetPer_AttachCoreOutArgs attachCoreOutArgs;

        /* Attach the core with RM */
        coreId = gEnetIfObj.coreId;

        ENET_IOCTL_SET_INOUT_ARGS(&prms, &coreId, &attachCoreOutArgs);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            ENET_PER_IOCTL_ATTACH_CORE,
                            &prms);
        if (status != ENET_SOK)
        {
            EnetIf_print("%s:%d Error: failed ENET_PER_IOCTL_ATTACH_CORE: %d",
                __FUNCTION__, __LINE__, status);
        }
        else
        {
            gEnetIfObj.coreKey = attachCoreOutArgs.coreKey;
        }
    }
#elif defined (CPSW9G)
    /* CPSW9G proxy driver open & attach in cpsw_attach() */
    /* Open UDMA for CPSW NAVSS instance type */
    gEnetIfObj.dmaDataPathConfig.hUdmaDrv = gEnetIfObj.hUdmaDrv;
    gEnetIfObj.dma = EnetUdma_initDataPath(gEnetIfObj.enetType, gEnetIfObj.instId, &gEnetIfObj.dmaDataPathConfig);
    EnetIf_assert(NULL != gEnetIfObj.dma);
#endif

    if (status == ENET_SOK)
    {
        /* appmem open should happen after Cpsw is opened as it uses CpswUtils_Q
         * functions */
        status = EnetIfMem_init();
        EnetIf_assert(ENET_SOK == status);
    }

    return status;
}

void EnetIf_closeCpsw(void)
{
    Enet_IoctlPrms prms;
    int32_t status;

    /* Close port link */
    ENET_IOCTL_SET_IN_ARGS(&prms, &gEnetIfObj.macPort);

    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_PER_IOCTL_CLOSE_PORT_LINK,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed to close port link: %d",
            __FUNCTION__, __LINE__, status);
    }

    /* Detach core */
    ENET_IOCTL_SET_IN_ARGS(&prms, &gEnetIfObj.coreKey);

    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_PER_IOCTL_DETACH_CORE,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: Failed to detach core key: %d",
            __FUNCTION__, __LINE__, status);
    }

    /* Close Enet driver */
    Enet_close(gEnetIfObj.hEnet);

}

void EnetIf_showStats(void)
{
    Enet_IoctlPrms prms;
    CpswStats_PortStats portStats;
    int32_t status = ENET_SOK;

    /* Show host port statistics */
    ENET_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_STATS_IOCTL_GET_HOSTPORT_STATS,
                        &prms);
    if (status == ENET_SOK)
    {
        EnetIf_print(" Port 0 Statistics");
        EnetIf_print("-----------------------------------------");
        EnetIf_printHostPortStats2G((CpswStats_HostPort_2g *)&portStats);
        EnetIf_print("");
    }
    else
    {
        EnetIf_print("%s:%d Error: failed to get host stats: %d",
            __FUNCTION__, __LINE__, status);
    }

    /* Show MAC port statistics */
    if (status == ENET_SOK)
    {
        ENET_IOCTL_SET_INOUT_ARGS(&prms, &gEnetIfObj.macPort, &portStats);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            ENET_STATS_IOCTL_GET_MACPORT_STATS,
                            &prms);
        if (status == ENET_SOK)
        {
            EnetIf_print(" Port 1 Statistics");
            EnetIf_print("-----------------------------------------");
            EnetIf_printMacPortStats2G((CpswStats_MacPort_2g *)&portStats);
            EnetIf_print("");
        }
        else
        {
            EnetIf_print("%s:%d Error: failed to get MAC stats: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    /* Show ALE table entries */
    if (status == ENET_SOK)
    {
        ENET_IOCTL_SET_NO_ARGS(&prms);
        status = Enet_ioctl(gEnetIfObj.hEnet,
                            gEnetIfObj.coreId,
                            CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        EnetIf_assert(status == ENET_SOK);
    }
}

extern void SetRxPulse();
void EnetIf_rxIsrFxn(void *appData)
{
    SetRxPulse();
}

void EnetIf_txIsrFxn(void *appData)
{
    // nothing to do here
}

int32_t EnetIf_initRxReadyPktQ(void)
{
    EnetDma_PktQ rxFreeQ;
    EnetDma_Pkt *pPktInfo;
    int32_t status = ENET_SOK;
    int i;
    struct mbuf *m;
    struct nw_work_thread   *wtp = WTP;

    EnetQueue_initQ(&rxFreeQ);
    EnetQueue_initQ(&gEnetIfObj.rxReadyQ);

    for (i = 0U; i < gEnetIfObj.numRxPkts; i++)
    {
        m = m_getcl_wtp (M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
        if (m == NULL) {
            EnetIf_print("%s:%d Error: m_getcl_wtp failed ",
                __FUNCTION__, __LINE__);
            status = ENET_EALLOC;
            return status;
        }

        pPktInfo = (EnetDma_Pkt *)EnetQueue_deq(&gEnetIfObj.freePktInfoQ);
        EnetIf_assert(pPktInfo != NULL);
        pPktInfo->bufPtr     = (uint8_t *)m->m_data;
        pPktInfo->bufPtrPhys = (uint8_t *)pool_phys(m->m_data, m->m_ext.ext_page);
        pPktInfo->userBufLen = MAX_ETH_PACKET_SIZE;
        pPktInfo->orgBufLen  = MAX_ETH_PACKET_SIZE;
        pPktInfo->appPriv    = m;
        ENET_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, ENET_PKTSTATE_APP_WITH_FREEQ);
        EnetQueue_enq(&rxFreeQ, &pPktInfo->node);
    }

    /* Retrieve any CPSW packets which are ready */
    status = EnetDma_retrieveRxPktQ(gEnetIfObj.hRxFlow, &gEnetIfObj.rxReadyQ);
    EnetIf_assert(status == ENET_SOK);
    /* There should not be any packet with DMA during init */
    EnetIf_assert(EnetQueue_getQCount(&gEnetIfObj.rxReadyQ) == 0U);

#if (1U == ENET_CFG_DEV_ERROR)
    EnetIf_validatePacketState(&rxFreeQ,
                                 ENET_PKTSTATE_APP_WITH_FREEQ,
                                 ENET_PKTSTATE_APP_WITH_DRIVER);
#endif

    status = EnetDma_submitRxPktQ(gEnetIfObj.hRxFlow,
                            &rxFreeQ);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed to send packet.. ret-%d",
            __FUNCTION__, __LINE__, status);
    }

    /* Assert here as during init no. of DMA descriptors should be equal to
     * no. of free Ethernet buffers available with app */

    EnetIf_assert(0U == EnetQueue_getQCount(&rxFreeQ));

    return status;
}

void EnetIf_initPktInfoQ()
{
    uint32_t i;
    EnetDma_Pkt *pPktInfo;

    /* Initialize all queues */
    EnetQueue_initQ(&gEnetIfObj.freePktInfoQ);

    memset(&gPktInfoMem, 0U,
        sizeof(EnetDma_Pkt) * (CPSW_MEM_TX_PKTS_MAX + CPSW_MEM_RX_PKTS_MAX));

    /* Initialize TX & RX EthPkts and queue them to freePktInfoQ */
    for (i = 0U; i < (gEnetIfObj.numTxPkts + gEnetIfObj.numRxPkts); i++)
    {
        pPktInfo = &gPktInfoMem[i];
        ENET_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, ENET_PKTSTATE_APP_WITH_FREEQ);
        EnetQueue_enq(&gEnetIfObj.freePktInfoQ, &pPktInfo->node);
    }

    EnetIf_print("initQs() freePktInfoQ initialized with %d pkts",
                       EnetQueue_getQCount(&gEnetIfObj.freePktInfoQ));

}

int32_t EnetIf_openDma(void)
{
    int32_t status = ENET_SOK;
    EnetUdma_OpenRxFlowPrms cpswRxFlowCfg;
    EnetUdma_OpenTxChPrms cpswTxChCfg;

    EnetIf_initPktInfoQ();

    /* Open the CPSW TX channel  */
    if (status == ENET_SOK)
    {
        EnetDma_initTxChParams(&cpswTxChCfg);

        cpswTxChCfg.hUdmaDrv = gEnetIfObj.hUdmaDrv;
        cpswTxChCfg.cbArg   = &gEnetIfObj;
        // Note: we do not register callback for TX channel
        cpswTxChCfg.notifyCb = NULL;
        // Note: Using preferred channel
        cpswTxChCfg.preferredChNum = gEnetIfObj.preferredUdmaChannel;

        EnetIf_setCommonTxChPrms(&cpswTxChCfg, gEnetIfObj.numTxPkts);

#if defined (CPSW2G)
        EnetIf_openTxCh(gEnetIfObj.hEnet,
                              gEnetIfObj.coreKey,
                              gEnetIfObj.coreId,
                              &gEnetIfObj.txChNum,
                              &gEnetIfObj.hTxCh,
                              &cpswTxChCfg);
#elif defined (CPSW9G)
        EnetIf_openTxCh(&gEnetIfObj,
                         &cpswTxChCfg);
#endif
        // Note: We do not call EnetDma_enableTxEvent call for
        // registering the TX event

        if (NULL == gEnetIfObj.hTxCh)
        {
#if defined (CPSW2G)
            /* Free the Ch Num if open Tx Ch failed */
            EnetIf_freeTxCh(gEnetIfObj.hEnet,
                             gEnetIfObj.coreKey,
                             gEnetIfObj.coreId,
                             gEnetIfObj.txChNum);
#elif defined (CPSW9G)
            CpswProxy_freeTxCh(gEnetIfObj.hCpswProxy,
                               gEnetIfObj.hEnet,
                               gEnetIfObj.coreKey,
                               gEnetIfObj.txChNum);
#endif
            EnetIf_print("%s:%d Error: EnetDma_openTxCh failed to open: %d",
                __FUNCTION__, __LINE__, status);
            status = ENET_EFAIL;
        }
    }

    /* Open the CPSW RX flow  */
    if (status == ENET_SOK)
    {
#if defined (CPSW2G)
        EnetDma_initRxChParams(&cpswRxFlowCfg);
        cpswRxFlowCfg.notifyCb = EnetIf_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gEnetIfObj.hUdmaDrv;
        cpswRxFlowCfg.cbArg   = &gEnetIfObj;

        EnetIf_setCommonRxFlowPrms(&cpswRxFlowCfg, gEnetIfObj.numRxPkts);

        EnetIf_openRxFlow(gEnetIfObj.hEnet,
                           gEnetIfObj.coreKey,
                           gEnetIfObj.coreId,
                           true,
                           &gEnetIfObj.rxStartFlowIdx,
                           &gEnetIfObj.rxFlowIdx,
                           &gEnetIfObj.hostMacAddr[0U],
                           &gEnetIfObj.hRxFlow,
                           &cpswRxFlowCfg);
#elif defined (CPSW9G)
        EnetDma_initRxChParams(&cpswRxFlowCfg);
        EnetIf_setCommonRxFlowPrms(&cpswRxFlowCfg, gEnetIfObj.numRxPkts);
        cpswRxFlowCfg.notifyCb = EnetIf_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gEnetIfObj.hUdmaDrv;
        cpswRxFlowCfg.cbArg   = &gEnetIfObj;
        cpswRxFlowCfg.rxFlowMtu = gEnetIfObj.rxMtu;

        EnetIf_openRxFlow(&gEnetIfObj,
                           false,
                           &cpswRxFlowCfg);
#endif

        if (NULL == gEnetIfObj.hRxFlow)
        {
#if defined (CPSW2G)
            EnetIf_freeRxFlow(gEnetIfObj.hEnet,
                               gEnetIfObj.coreKey,
                               gEnetIfObj.coreId,
                               gEnetIfObj.rxFlowIdx);
#elif defined (CPSW9G)
            CpswProxy_freeRxFlow(gEnetIfObj.hCpswProxy,
                                 gEnetIfObj.hEnet,
                                 gEnetIfObj.coreKey,
                                 gEnetIfObj.rxFlowIdx);
#endif
            EnetIf_print("%s:%d Error: EnetDma_openRxCh failed to open: %d",
                __FUNCTION__, __LINE__, status);
            EnetIf_assert(NULL != gEnetIfObj.hRxFlow);
        }
        else
        {
            EnetIf_print("Host MAC address: ");
            EnetIf_printMacAddr(gEnetIfObj.hostMacAddr);
#if defined (CPSW2G)
            /* For MAC loopback disable secure flag for the host port entry */
            EnetIf_changeHostAleEntry(&gEnetIfObj.hostMacAddr[0U]);
#elif defined (CPSW9G)
            /* TODO: Is this needed for 9G */
#endif

            /* Submit all ready RX buffers to DMA.*/
            status = EnetIf_initRxReadyPktQ();
        }
    }

    return status;
}

void EnetIf_freePktInfoQ(EnetDma_PktQ *pPktInfoQ)
{
    EnetDma_Pkt *pktInfo;
    uint32_t pktCnt, i;
    struct mbuf *m = NULL;

    pktCnt = EnetQueue_getQCount(pPktInfoQ);
    /* Free all retrieved packets from DMA */
    for (i = 0U; i < pktCnt; i++)
    {
        pktInfo = (EnetDma_Pkt *)EnetQueue_deq(pPktInfoQ);
        if (pktInfo->appPriv != NULL) {
            m = pktInfo->appPriv;
            m_freem (m);
            pktInfo->appPriv = NULL;
        }
    }
}

void EnetIf_closeDma(void)
{
    EnetDma_PktQ fqPktInfoQ;
    EnetDma_PktQ cqPktInfoQ;

    EnetQueue_initQ(&fqPktInfoQ);
    EnetQueue_initQ(&cqPktInfoQ);

    /* There should not be any ready packet */
    EnetIf_assert(0U == EnetQueue_getQCount(&gEnetIfObj.rxReadyQ));

#if defined (CPSW2G)
    /* Close RX Flow */
    EnetIf_closeRxFlow(gEnetIfObj.hEnet,
                        gEnetIfObj.coreKey,
                        gEnetIfObj.coreId,
                        true,
                        &fqPktInfoQ,
                        &cqPktInfoQ,
                        gEnetIfObj.rxStartFlowIdx,
                        gEnetIfObj.rxFlowIdx,
                        gEnetIfObj.hostMacAddr,
                        gEnetIfObj.hRxFlow);
#elif defined (CPSW9G)
    /* Close RX Flow */
    EnetIf_closeRxFlow(&gEnetIfObj,
                        true,
                        &fqPktInfoQ,
                        &cqPktInfoQ);
#endif

    EnetIf_freePktInfoQ(&fqPktInfoQ);
    EnetIf_freePktInfoQ(&cqPktInfoQ);
    /* Close TX channel */
    EnetQueue_initQ(&fqPktInfoQ);
    EnetQueue_initQ(&cqPktInfoQ);

#if defined (CPSW2G)
    EnetIf_closeTxCh(gEnetIfObj.hEnet,
                      gEnetIfObj.coreKey,
                      gEnetIfObj.coreId,
                      &fqPktInfoQ,
                      &cqPktInfoQ,
                      gEnetIfObj.hTxCh,
                      gEnetIfObj.txChNum);
#elif defined (CPSW9G)
    EnetIf_closeTxCh(&gEnetIfObj,
                      &fqPktInfoQ,
                      &cqPktInfoQ);
#endif
    EnetIf_freePktInfoQ(&fqPktInfoQ);
    EnetIf_freePktInfoQ(&cqPktInfoQ);
    EnetIf_freePktInfoQ(&gEnetIfObj.rxReadyQ);
    EnetIf_freePktInfoQ(&gEnetIfObj.freePktInfoQ);

    EnetIfMem_deInit();
}

void EnetIf_changeHostAleEntry(uint8_t macAddr[])
{
    int32_t status;
    Enet_IoctlPrms prms;
    uint32_t setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.trunk   = false;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;

    ENET_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        CPSW_ALE_IOCTL_ADD_UCAST,
                        &prms);
    if (status != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d",
            __FUNCTION__, __LINE__, status);
    }
}

uint32_t EnetIf_retrieveFreeTxPkts(void)
{
    EnetDma_PktQ txFreeQ;
    EnetDma_Pkt *pktInfo;
    int32_t status;
    uint32_t txFreeQCnt = 0U;
    struct mbuf *m = NULL;

    EnetQueue_initQ(&txFreeQ);

    /* Retrieve any CPSW packets that may be free now */
    status = EnetDma_retrieveTxPktQ(gEnetIfObj.hTxCh, &txFreeQ);
    if (status == ENET_SOK)
    {
        txFreeQCnt = EnetQueue_getQCount(&txFreeQ);

        pktInfo = (EnetDma_Pkt *)EnetQueue_deq(&txFreeQ);
        while (NULL != pktInfo)
        {
            // start
            m = pktInfo->appPriv;
            EnetIf_assert(m != NULL);
            m_freem (m);
            pktInfo->appPriv = NULL;

#if (1U == ENET_CFG_DEV_ERROR)
            EnetDma_checkPktState(&pktInfo->pktState,
                                    ENET_PKTSTATE_MODULE_APP,
                                    ENET_PKTSTATE_APP_WITH_DRIVER,
                                    ENET_PKTSTATE_APP_WITH_FREEQ);
#endif

            EnetQueue_enq(&gEnetIfObj.freePktInfoQ, &pktInfo->node);
            pktInfo = (EnetDma_Pkt *)EnetQueue_deq(&txFreeQ);
        }
    }
    else
    {
        EnetIf_print("%s:%d Error: failed to retrieve pkts: %d",
            __FUNCTION__, __LINE__, status);
    }

    return txFreeQCnt;
}

uint32_t EnetIf_receivePkts(void)
{
    int32_t status;
    uint32_t rxReadyCnt = 0U;

    EnetQueue_initQ(&gEnetIfObj.rxReadyQ);

    /* Retrieve any CPSW packets which are ready */
    status = EnetDma_retrieveRxPktQ(gEnetIfObj.hRxFlow, &gEnetIfObj.rxReadyQ);
    if (status == ENET_SOK)
    {
        rxReadyCnt = EnetQueue_getQCount(&gEnetIfObj.rxReadyQ);
    }
    else
    {
        EnetIf_print("%s:%d Error: failed to retrieve pkts: %d",
            __FUNCTION__, __LINE__, status);
    }

    return rxReadyCnt;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void EnetIf_Tick()
{
    Enet_periodicTick(gEnetIfObj.hEnet);
}

void EnetIf_GetStats(nic_stats_t *stats)
{
#if defined (CPSW2G)
    Enet_IoctlPrms prms;
    CpswStats_PortStats portStats;
    int32_t status = ENET_SOK;
    nic_ethernet_stats_t    *estats = &stats->un.estats;
    CpswStats_MacPort_2g *st = (CpswStats_MacPort_2g *)&portStats;

    if (gEnetIfObj.hEnet == NULL)
    {
        return;
    }

    ENET_IOCTL_SET_INOUT_ARGS(&prms, &gEnetIfObj.macPort, &portStats);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_STATS_IOCTL_GET_MACPORT_STATS,
                        &prms);
    if (status == ENET_SOK)
    {
        stats->rxed_ok = st->rxGoodFrames;
        stats->rxed_broadcast = st->rxBcastFrames;
        stats->rxed_multicast = st->rxMcastFrames;
        stats->octets_rxed_ok = st->rxOctets;

        stats->txed_ok = st->txGoodFrames;
        stats->txed_broadcast = st->txBcastFrames;
        stats->txed_multicast = st->txMcastFrames;
        stats->octets_txed_ok = st->txOctets;

        estats->align_errors = st->rxAlignCodeErrors;
        estats->single_collisions = st->txSingleCollFrames;
        estats->multi_collisions = st->txMultipleCollFrames;
        estats->fcs_errors = st->rxCrcErrors;
        estats->tx_deferred = st->txDeferredFrames;
        estats->late_collisions = st->txLateCollFrames;
        estats->xcoll_aborted = st->txExcessiveCollFrames;
        estats->no_carrier = st->txCarrierSenseErrors;
        estats->oversized_packets = st->rxOversizedFrames;
        estats->jabber_detected = st->rxJabberFrames;
        estats->short_packets = st->rxUndersizedFrames;
        estats->total_collision_frames = st->txCollisionFrames;
    }
    else
    {
        EnetIf_print("%s:%d Error: failed to get MAC stats: %d",
            __FUNCTION__, __LINE__, status);
    }
    EnetIf_showStats();

    EnetIf_print("rxReadyQ count = %d", EnetQueue_getQCount(&gEnetIfObj.rxReadyQ));
    EnetIf_print("freePktInfoQ count = %d", EnetQueue_getQCount(&gEnetIfObj.freePktInfoQ));

#if 0
    /* Print CPSW regs */
    ENET_IOCTL_SET_NO_ARGS(&prms);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_PER_IOCTL_PRINT_REGS,
                        &prms);
#endif

    /* Print CPSW PHY regs */
    ENET_IOCTL_SET_IN_ARGS(&prms, &gEnetIfObj.macPort);
    status = Enet_ioctl(gEnetIfObj.hEnet,
                        gEnetIfObj.coreId,
                        ENET_PHY_IOCTL_PRINT_REGS,
                        &prms);

#elif defined (CPSW9G)
    /* TODO: Add CPSW9G Proxy support for statistics */
#endif
}


int32_t EnetIf_SendPkt(struct mbuf* m)
{
    int32_t retVal;
    EnetDma_Pkt *pktInfo;
    EnetDma_PktQ txSubmitQ;

    EnetIf_assert(m != NULL);

    /* Transmit a single packet */
    EnetQueue_initQ(&txSubmitQ);

    /* Dequeue one free TX Eth packet */
    pktInfo = (EnetDma_Pkt *) EnetQueue_deq(&gEnetIfObj.freePktInfoQ);
    if (pktInfo == NULL)
    {
        return ENET_EFAIL;
    }

    // start
    if (!ENET_UTILS_IS_ALIGNED(m->m_data, UDMA_CACHELINE_ALIGNMENT))
    {
        EnetIf_print("%s:%d: Error: Send m buffer addr=0x%lx is not aligned",
            __FUNCTION__, __LINE__, m->m_data);
    }

    memset(&pktInfo->node, 0, sizeof(pktInfo->node));
    pktInfo->bufPtr = (uint8_t *)m->m_data;
    pktInfo->bufPtrPhys = (uint8_t *)pool_phys(m->m_data, m->m_ext.ext_page);
    pktInfo->orgBufLen = m->m_ext.ext_size;
    pktInfo->userBufLen = m->m_len;
    pktInfo->txPortNum = ENET_MAC_PORT_INV;
    pktInfo->appPriv = m;
    // end

#if (1U == ENET_CFG_DEV_ERROR)
    EnetDma_checkPktState(&pktInfo->pktState,
                            ENET_PKTSTATE_MODULE_APP,
                            ENET_PKTSTATE_APP_WITH_FREEQ,
                            ENET_PKTSTATE_APP_WITH_DRIVER);
#endif

    /* Enqueue the packet for later transmission */
    EnetQueue_enq(&txSubmitQ, &pktInfo->node);

    retVal = EnetDma_submitTxPktQ(gEnetIfObj.hTxCh,
                                  &txSubmitQ);
    if (retVal != ENET_SOK)
    {
        EnetIf_print("%s:%d Error: failed to send packet.. ret-%d",
            __FUNCTION__, __LINE__, retVal);
    }

    return retVal;
}

extern struct mbuf * cpsw_process_rx(void *arg, struct mbuf *);

int32_t EnetIf_GetRx(void *arg)
{
    int32_t retVal = ENET_SOK;
    EnetDma_PktQ rxFreeQ;
    EnetDma_Pkt *pktInfo;
    uint32_t rxReadyCnt;
    struct mbuf *m, *mNew;

    /* Get the packets received so far */
    rxReadyCnt = EnetIf_receivePkts();
    if (rxReadyCnt > 0U)
    {
        EnetQueue_initQ(&rxFreeQ);

        /* Consume the received packets and release them */
        pktInfo = (EnetDma_Pkt *) EnetQueue_deq( &gEnetIfObj.rxReadyQ);
        while (NULL != pktInfo)
        {
#if (1U == ENET_CFG_DEV_ERROR)
            EnetDma_checkPktState(&pktInfo->pktState,
                                    ENET_PKTSTATE_MODULE_APP,
                                    ENET_PKTSTATE_APP_WITH_DRIVER,
                                    ENET_PKTSTATE_APP_WITH_READYQ);
#endif
            m = pktInfo->appPriv;
            m->m_pkthdr.len = m->m_len = pktInfo->userBufLen;
            /* Get a packet/buffer to replace the one that was filled */
            /* Consume the packet by sending to upper layer */
            mNew = cpsw_process_rx(arg, m);
            if (mNew == NULL)
            {
                EnetIf_print("%s:%d: Error: cpsw_process_rx returned NULL",
                    __FUNCTION__, __LINE__);
                EnetIf_assert(false);
                //TODO - instead of assert, push pktInfo to freePktInfoQ and populate it
                // later when m bufs gets available.
            }

            EnetIf_assert(ENET_UTILS_IS_ALIGNED(mNew->m_data, UDMA_CACHELINE_ALIGNMENT));

            pktInfo->bufPtr     = (uint8_t *)mNew->m_data;
            pktInfo->bufPtrPhys = (uint8_t *)pool_phys(mNew->m_data, mNew->m_ext.ext_page);
            pktInfo->orgBufLen  = MAX_ETH_PACKET_SIZE;
            pktInfo->userBufLen = MAX_ETH_PACKET_SIZE;
            pktInfo->appPriv    = mNew;

#if (1U == ENET_CFG_DEV_ERROR)
            EnetDma_checkPktState(&pktInfo->pktState,
                                    ENET_PKTSTATE_MODULE_APP,
                                    ENET_PKTSTATE_APP_WITH_READYQ,
                                    ENET_PKTSTATE_APP_WITH_FREEQ);
#endif

            EnetQueue_enq(&rxFreeQ, &pktInfo->node);
            pktInfo = (EnetDma_Pkt *) EnetQueue_deq( &gEnetIfObj.rxReadyQ);
        }

        /*Submit now processed buffers */
        {
#if (1U == ENET_CFG_DEV_ERROR)
            EnetIf_validatePacketState(&rxFreeQ,
                                             ENET_PKTSTATE_APP_WITH_FREEQ,
                                             ENET_PKTSTATE_APP_WITH_DRIVER);
#endif

            retVal = EnetDma_submitRxPktQ(gEnetIfObj.hRxFlow,
                                    &rxFreeQ);
            if (retVal != ENET_SOK)
            {
                EnetIf_print("%s:%d Error: failed to send packet.. ret-%d",
                    __FUNCTION__, __LINE__, retVal);
            }
        }
    }

    return retVal;
}


uint32_t EnetIf_GetTxFreeQCnt() {
    return EnetQueue_getQCount(&gEnetIfObj.freePktInfoQ);
}

#if defined (CPSW9G)
void EnetIf_RegisterIPv4Address(uint8_t *ipv4Addr)
{
    CpswProxy_registerIPV4Addr(gEnetIfObj.hCpswProxy,
                               gEnetIfObj.hEnet,
                               gEnetIfObj.coreKey,
                               gEnetIfObj.hostMacAddr,
                               ipv4Addr);
}

void EnetIf_InitSyncTimer(void)
{
    int32_t status;
    uint64_t baseaddr;

    memset(&gEnetIfObj.syncTimerObj, 0, sizeof(EnetIf_SyncTimerObj));

    gEnetIfObj.syncTimerObj.GTC0_GTC_CFG0_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG0_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG0_BASE);
    EnetIf_assert(MAP_FAILED != gEnetIfObj.syncTimerObj.GTC0_GTC_CFG0_Base);

    gEnetIfObj.syncTimerObj.GTC0_GTC_CFG1_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG1_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG1_BASE);
    EnetIf_assert(MAP_FAILED != gEnetIfObj.syncTimerObj.GTC0_GTC_CFG1_Base);

    gEnetIfObj.syncTimerObj.GTC0_GTC_CFG2_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG2_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG2_BASE);
    EnetIf_assert(MAP_FAILED != gEnetIfObj.syncTimerObj.GTC0_GTC_CFG2_Base);

    gEnetIfObj.syncTimerObj.GTC0_GTC_CFG3_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG3_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG3_BASE);
    EnetIf_assert(MAP_FAILED != gEnetIfObj.syncTimerObj.GTC0_GTC_CFG3_Base);

    EnetIf_print("EnetIf_InitSyncTimer(): Set GTC0 Push event from BIT %d\n", CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);
    baseaddr = (uint64_t) gEnetIfObj.syncTimerObj.GTC0_GTC_CFG0_Base;
    baseaddr += CSL_GTC_CFG0_PUSHEVT;
    /* Configure GTC push event */
    CSL_REG32_WR(baseaddr, CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);

    /* Register callback */
    status = CpswProxy_registerHwPushNotifyCb(gEnetIfObj.hCpswProxy, EnetIf_calcSyncTimeParams);
    if (status == ENET_EALREADYOPEN)
    {
        EnetIf_print("CpswProxy_registerHwPushNotifyCb(): Callback is registered already\n");
    }
    else
    {
        EnetIf_print("CpswProxy_registerRemoteTimer() for GTC0 Push Event\n");
        /* Send request to Ethfw to configure TSR */
        CpswProxy_registerRemoteTimer(gEnetIfObj.hCpswProxy,
                                      gEnetIfObj.hEnet,
                                      gEnetIfObj.coreKey,
                                      CSLR_TIMESYNC_INTRTR0_IN_GTC0_GTC_PUSH_EVENT_0,
                                      CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM);
    }

    /* Enable Timer */
    baseaddr = (uint64_t) gEnetIfObj.syncTimerObj.GTC0_GTC_CFG1_Base;
    baseaddr += CSL_GTC_CFG1_CNTCR;
    if ((CSL_REG32_RD(baseaddr) & 0x1U) == 0U)
    {
        EnetIf_print("EnetIf_InitSyncTimer(): GTC0 is not enabled. Enable it now.\n");
        CSL_REG32_WR(baseaddr, 0x1U);
    }

    return;
}

#endif
/* end of file */
