/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_app.h
 *
 *  \brief Common CPSW utility header file.
 */

#ifndef CPSW_APP_H_
#define CPSW_APP_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <hw/inout.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <hw/nicinfo.h>

#include <ti/osal/osal.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/dp83867.h>

#if defined (CPSW9G)
#include <ethremotecfg/client/include/cpsw_proxy.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
void CpswApp_print(const char *pcString, ...);

extern void error_log(const char *str, const char *fileName, int32_t lineNum);

static inline void CpswApp_assertLocal(bool condition,
                                            const char *str,
                                            const char *fileName,
                                            int32_t lineNum)
{
    if (!(condition))
    {
        error_log(str, fileName, lineNum);
    }

    return;
}


#define CpswApp_assert(cond)                                     \
    (CpswApp_assertLocal((bool) (cond), (const char *) # cond, \
                    (const char *) __FILE__, (int32_t) __LINE__))

#ifndef htons
/** \brief Host to network byte order conversion for short integer */
#define htons(a)                        ((((a) & 0x00FFU) << 8) | \
                                         (((a) & 0xFF00U) >> 8))
#endif

#ifndef ntohs
/** \brief Network to host byte order conversion for long integer */
#define ntohs(a)                        htons(a)
#endif

/** \brief VLAN tag's Tag Protocol Identifier (TPID) */
#define ETHERTYPE_VLAN_TAG              (0x8100U)

/** \brief MAC address length in bytes */
#define ETH_MAC_ADDR_LEN                 (6U)

/** \brief Total bytes in header */
#define ETH_HDR_LEN                     (14U)

/** \brief Max octets in payload */
#define ETH_PAYLOAD_LEN                 (1500U)

/** \brief VLAN tag length in bytes */
#define ETH_VLAN_TAG_LEN                (4U)

/** \brief Test frame's header length in bytes */
#define ETH_TEST_DATA_HDR_LEN           (4U)

/* \brief Octets of payload per console row */
#define OCTETS_PER_ROW                  (16U)
/** \brief Max frame length */
#define ETH_MAX_FRAME_LEN               (1522U)

#define RGMII_ID_DISABLE_MASK           (0x10)

#define MAX_ETH_PACKET_SIZE             (1536U)

/* \brief Support macros for MMR lock/unlock functions */
#define MMR_KICK0_UNLOCK_VAL            (0x68EF3490U)
#define MMR_KICK1_UNLOCK_VAL            (0xD172BC5AU)
#define MMR_KICK_LOCK_VAL               (0x00000000U)

#define CSL_MMR_KICK_UNLOCKED_MASK          (0x00000001U)
#define CSL_MMR_KICK_UNLOCKED_SHIFT         (0x0U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */
typedef enum
{
    GMII = 0,
    RMII,
    RGMII,
    SGMII,
    QSGMII
}emac_mode;

/**
 *  \brief  CPTS Clock Select Mux enum
 */
typedef enum
{
    CPSWAPP_CPTS_CLKSEL_MAIN_PLL3_HSDIV1_CLKOUT = 0x0U,
    /**< Main PLL3 HSDIV1 Clockout for CPTS*/
    CPSWAPP_CPTS_CLKSEL_MAIN_PLL0_HSDIV6_CLKOUT = 0x1U,
    /**< Main PLL0 HSDIV6 Clockout for CPTS*/
    CPSWAPP_CPTS_CLKSEL_MCU_CPTS_REF_CLK        = 0x2U,
    /**< MCU CPTS Reference Clockout*/
    CPSWAPP_CPTS_CLKSEL_CPTS_RFT_CLK            = 0x3U,
    /**< CPTS RFT Clock */
    CPSWAPP_CPTS_CLKSEL_MCU_EXT_REFCLK0         = 0x4U,
    /**< MCU External Reference Clock0*/
    CPSWAPP_CPTS_CLKSEL_EXT_REFCLK1             = 0x5U,
    /**< External Reference Clock 1 for CPTS*/
    CPSWAPP_CPTS_CLKSEL_SERDES0_IP2_LN0_TXMCLK  = 0x6U,
    /**< SERDES0 IP2 Lane0 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES0_IP2_LN1_TXMCLK  = 0x7U,
    /**< SERDES0 IP2 Lane1 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES1_IP2_LN0_TXMCLK  = 0x8U,
    /**< SERDES1 IP2 Lane0 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES1_IP2_LN1_TXMCLK  = 0x9U,
    /**< SERDES1 IP2 Lane1 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES2_IP2_LN0_TXMCLK  = 0xAU,
    /**< SERDES2 IP2 Lane0 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES2_IP2_LN1_TXMCLK  = 0xBU,
    /**< SERDES2 IP2 Lane1 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES3_IP2_LN0_TXMCLK  = 0xCU,
    /**< SERDES3 IP2 Lane0 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_SERDES3_IP2_LN1_TXMCLK  = 0xDU,
    /**< SERDES3 IP2 Lane1 Tx MCLK*/
    CPSWAPP_CPTS_CLKSEL_MCU_PLL2_HSDIV1_CLKOUT  = 0xEU,
    /**< MCU PLL2 HSDIV1 Clockout*/
    CPSWAPP_CPTS_CLKSEL_MAIN_SYSCLK0            = 0xFU,
    /**< Main System Clock*/
} CpswApp_CptsClkSelMux;

/**
 *  \brief  Lock/unlock enum
 */
typedef enum
{
    CPSWAPP_LOCK_MMR = 0U,
    /**< lock control MMR */
    CPSWAPP_UNLOCK_MMR,
    /**< unlock control MMR */
} CpswApp_MmrLockType;

/**
 *  \brief CTRL MMR enum for MCU and Main domain
 */
typedef enum
{
    CPSWAPP_MMR_LOCK0 = 0U,
    /**< control MMR lock 0*/
    CPSWAPP_MMR_LOCK1,
    /**< control MMR lock 1*/
    CPSWAPP_MMR_LOCK2,
    /**< control MMR lock 2*/
    CPSWAPP_MMR_LOCK3,
    /**< control MMR lock 3*/
    CPSWAPP_MMR_LOCK4,
    /**< control MMR lock 4*/
    CPSWAPP_MMR_LOCK5,
    /**< control MMR lock 5*/
    CPSWAPP_MMR_LOCK6,
    /**< control MMR lock 6*/
    CPSWAPP_MMR_LOCK7,
    /**< control MMR lock 7*/
} CpswApp_CtrlMmrType;

typedef struct
{
    uint8_t dstMac[ETH_MAC_ADDR_LEN];
    uint8_t srcMac[ETH_MAC_ADDR_LEN];
    uint16_t etherType;
} __attribute__ ((packed)) EthFrameHeader;

typedef struct
{
    EthFrameHeader hdr;
    uint8_t payload[ETH_PAYLOAD_LEN + ETH_VLAN_TAG_LEN];
} __attribute__ ((packed)) EthFrame;

typedef struct
{
    uint8_t dstMac[ETH_MAC_ADDR_LEN];
    uint8_t srcMac[ETH_MAC_ADDR_LEN];
    uint16_t tpid;
    uint16_t tci;
    uint16_t etherType;
} __attribute__ ((packed)) EthVlanFrameHeader;

typedef struct
{
    EthVlanFrameHeader hdr;
    uint8_t payload[ETH_PAYLOAD_LEN];
} __attribute__ ((packed)) EthVlanFrame;

#if defined (CPSW9G)
typedef struct CpswRemoteApp_SyncTimerObj_s
{
    uint32_t *GTC0_GTC_CFG0_Base;
    uint32_t *GTC0_GTC_CFG1_Base;
    uint32_t *GTC0_GTC_CFG2_Base;
    uint32_t *GTC0_GTC_CFG3_Base;
    uint64_t currLocalTime;
    uint64_t prevLocalTime;
    uint64_t currCptsTime;
    uint64_t prevCptsTime;
    double rate;
    double offset;
} CpswApp_SyncTimerObj;
#endif

typedef struct
{
    /* CPSW instance type */
    Cpsw_Type            cpswType;

    /* Core Id */
    uint32_t             coreId;
    /*Core Key */
    uint32_t             coreKey;

    /* MAC port number */
    Cpsw_MacPort         portNum;

#if defined (CPSW9G)
    /* CPSW Proxy Driver handle */
    CpswProxy_Handle     hCpswProxy;
    /* MTU of rx packet */
    uint32_t             rxMtu;
    /* Max Tx packet size per priority */
    uint32_t             txMtu[CPSW_PRIORITY_NUM];
    /*! DMA handle */
    CpswDma_Handle dma;
    /*! DMA Config */
    CpswDma_DataPathConfig dmaDataPathConfig;
    /*! Sync Timer Object */
    CpswApp_SyncTimerObj syncTimerObj;
#endif

    /* CPSW driver handle */
    Cpsw_Handle          hCpsw;
    Udma_DrvHandle       hUdmaDrv;

    CpswDma_PktInfoQ     freePktInfoQ;

    CpswDma_RxFlowHandle hRxFlow;
    CpswDma_PktInfoQ     rxReadyQ;

    CpswDma_TxChHandle   hTxCh;

    uint8_t              hostMacAddr[ETH_MAC_ADDR_LEN];

    uint32_t             rxStartFlowIdx;

    uint32_t             rxFlowIdx;

    uint32_t             txChNum;

    uint32_t             preferredUdmaChannel;

    uint32_t             numTxPkts;
    uint32_t             numRxPkts;

    pthread_mutex_t      cpswMutex;
} CpswApp_Obj;


/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */
#define CpswApp_wait               delay

/* cpswapp_rmcfg.c */
const CpswRm_ResourcePrms *CpswAppRm_getResourcePartitionInfo(Cpsw_Type cpswType);
const CpswRm_IoctlPermissionTable *CpswAppRm_getIoctlPermissionInfo(Cpsw_Type cpswType);

uint32_t CpswApp_getNavSSInstanceId(Cpsw_Type cpswType);
Udma_DrvHandle CpswApp_udmaOpen(Cpsw_Type cpswType,
                                Udma_InitPrms *pInitPrms);
void CpswApp_udmaclose(Udma_DrvHandle hUdmaDrv);
void CpswApp_printMacAddr(uint8_t macAddr[]);
void CpswApp_printFrame(EthFrame *frame,
                        uint32_t len);
void CpswApp_printHostPortStats2G(CpswStats_HostPort_2g *st);
void CpswApp_printMacPortStats2G(CpswStats_MacPort_2g *st);
void CpswApp_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                             CpswDma_PktInfoQ *pSubmitQ);
int32_t CpswApp_submitTxPackets(CpswDma_TxChHandle hTxCh,
                                CpswDma_PktInfoQ *pSubmitQ);
void CpswApp_validatePacketState(CpswDma_PktInfoQ *pQueue,
                                 uint32_t expectedState,
                                 uint32_t newState);
uint64_t CpswApp_virtToPhyFxn(const void *virtAddr,
                              void *appData);
void *CpswApp_phyToVirtFxn(uint64_t phyAddr,
                           void *appData);
uint64_t CpswApp_udmaVirtToPhyFxn(const void *virtAddr,
                                  uint32_t chNum,
                                  void *appData);
void CpswApp_mcuMmrCtrl(CpswApp_CtrlMmrType mmrNum,
                        CpswApp_MmrLockType lock);
void CpswApp_mainMmrCtrl(CpswApp_CtrlMmrType mmrNum,
                         CpswApp_MmrLockType lock);
void CpswApp_selectCptsClock(Cpsw_Type cpswType,
                             CpswApp_CptsClkSelMux clkSelMux);
void *CpswApp_udmaPhyToVirtFxn(uint64_t phyAddr,
                               uint32_t chNum,
                               void *appData);
int32_t CpswApp_allocRxFlow(Cpsw_Handle hCpsw,
                            uint32_t coreKey,
                            uint32_t coreId,
                            uint32_t *rxFlowStartIdx,
                            uint32_t *flowIdx);
int32_t CpswApp_allocMac(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         uint8_t *macAddress);
int32_t CpswApp_freeRxFlow(Cpsw_Handle hCpsw,
                           uint32_t coreKey,
                           uint32_t coreId,
                           uint32_t rxFlowIdx);
int32_t CpswApp_freeTxCh(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         uint32_t txChNum);
int32_t CpswApp_freeMac(Cpsw_Handle hCpsw,
                        uint32_t coreKey,
                        uint32_t coreId,
                        uint8_t *macAddress);
int32_t CpswApp_registerDefaultRxFlow(Cpsw_Handle hCpsw,
                                      uint32_t coreKey,
                                      uint32_t coreId,
                                      uint32_t rxFlowStartIdx,
                                      uint32_t rxFlowIdx);
int32_t CpswApp_unregisterDefaultRxFlow(Cpsw_Handle hCpsw,
                                        uint32_t coreKey,
                                        uint32_t coreId,
                                        uint32_t rxFlowStartIdx,
                                        uint32_t rxFlowIdx);
int32_t CpswApp_registerDstMacRxFlow(Cpsw_Handle hCpsw,
                                     uint32_t coreKey,
                                     uint32_t coreId,
                                     uint32_t rxFlowStartIdx,
                                     uint32_t rxFlowIdx,
                                     uint8_t macAddress[CPSW_MAC_ADDR_LEN]);
int32_t CpswApp_unregisterDstMacRxFlow(Cpsw_Handle hCpsw,
                                       uint32_t coreKey,
                                       uint32_t coreId,
                                       uint32_t rxFlowStartIdx,
                                       uint32_t rxFlowIdx,
                                       uint8_t macAddress[CPSW_MAC_ADDR_LEN]);
bool CpswApp_isPortLinkUp(Cpsw_Handle hCpsw,
                          uint32_t coreId,
                          Cpsw_MacPort portNum);
void CpswApp_initResourceConfig(Cpsw_Type cpswType,
                                uint32_t selfCoreId,
                                CpswRm_ResourceConfig *rscCfg,
                                uint8_t *currentMacAddr);
void CpswApp_setCommonRxFlowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms, uint32_t numRxPkts);
void CpswApp_addHostPortEntry(Cpsw_Handle hCpsw,
                              uint32_t coreId,
                              uint8_t *macAddr);
void CpswApp_setCommonTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms, uint32_t numTxPkts);
#if defined (CPSW2G)
void CpswApp_openTxCh(Cpsw_Handle hCpsw,
                      uint32_t coreKey,
                      uint32_t coreId,
                      uint32_t *pTxChNum,
                      CpswDma_TxChHandle *pTxChHandle,
                      CpswDma_OpenTxChPrms *pCpswTxChCfg);
void CpswApp_closeTxCh(Cpsw_Handle hCpsw,
                       uint32_t coreKey,
                       uint32_t coreId,
                       CpswDma_PktInfoQ *pFqPktInfoQ,
                       CpswDma_PktInfoQ *pCqPktInfoQ,
                       CpswDma_TxChHandle hTxChHandle,
                       uint32_t txChNum);
void CpswApp_openRxFlow(Cpsw_Handle hCpsw,
                        uint32_t coreKey,
                        uint32_t coreId,
                        bool useDefaultFlow,
                        uint32_t *pRxFlowStartIdx,
                        uint32_t *pRxFlowIdx,
                        uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                        CpswDma_RxFlowHandle *pRxFlowHandle,
                        CpswDma_OpenRxFlowPrms *pRxFlowPrms);
void CpswApp_closeRxFlow(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         bool useDefaultFlow,
                         CpswDma_PktInfoQ *pFqPktInfoQ,
                         CpswDma_PktInfoQ *pCqPktInfoQ,
                         uint32_t rxFlowStartIdx,
                         uint32_t rxFlowIdx,
                         uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                         CpswDma_RxFlowHandle hRxFlow);
#elif defined (CPSW9G)
void CpswApp_openTxCh(CpswApp_Obj *obj,
                      CpswDma_OpenTxChPrms *pCpswTxChCfg);
void CpswApp_closeTxCh(CpswApp_Obj *obj,
                       CpswDma_PktInfoQ *pFqPktInfoQ,
                       CpswDma_PktInfoQ *pCqPktInfoQ);
void CpswApp_openRxFlow(CpswApp_Obj *obj,
                        bool useDefaultFlow,
                        CpswDma_OpenRxFlowPrms *pRxFlowPrms);
void CpswApp_closeRxFlow(CpswApp_Obj *obj,
                         bool useDefaultFlow,
                         CpswDma_PktInfoQ *pFqPktInfoQ,
                         CpswDma_PktInfoQ *pCqPktInfoQ);
#endif
int32_t CpswApp_showRxFlowStats(CpswDma_RxFlowHandle hRxFlow);
int32_t CpswApp_showTxChStats(CpswDma_TxChHandle hTxCh);
void CpswApp_setNoPhyCfgRgmii(CpswMacPort_Interface *interface,
                              CpswPhy_Config *phyConfig);

int32_t CpswAppBoard_cpsw2gMacModeConfig(uint32_t portNum,
                                         uint8_t mode);
uint32_t CpswAppBoard_getPhyAddr(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum);
void CpswAppBoard_setPhyConfig(Cpsw_Type cpswType,
                               Cpsw_MacPort portNum,
                               CpswMacPort_Interface *interface,
                               CpswPhy_Config *phyConfig);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_APP_H_ */
