/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_appmem.c
 *
 *  \brief CPSW DMA memory allocation utility functions.
 *
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* This is needed for memset/memcpy */
#include <string.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>

#include "j7_cpsw.h"
#include "cpsw_app.h"
#include "cpsw_appmem.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define CPSW_APPMEM_RING_MAX_ELEM_CNT          (CPSW_MEM_TX_PKTS_MAX + \
                                                CPSW_MEM_RX_PKTS_MAX)

#define CPSW_APPMEM_NUM_DESCS ((CPSW_MEM_TX_PKTS_MAX * CPSW_CFG_TX_CHANNELS_NUM) + \
                               (CPSW_MEM_RX_PKTS_MAX * CPSW_CFG_RX_FLOWS_NUM))

/* TDCQ is allocated in driver, we allocate only FQ and CQ */
#define CPSW_APPMEM_NUM_RINGS_TYPES (2U)

#define CPSW_APPMEM_NUM_RINGS (CPSW_APPMEM_NUM_RINGS_TYPES * \
                               (CPSW_CFG_TX_CHANNELS_NUM + CPSW_CFG_RX_FLOWS_NUM))

#define CPSW_APPMEM_RING_MAX_SIZE \
    (CPSW_DMA_RING_MEM_SIZE * CPSW_APPMEM_RING_MAX_ELEM_CNT)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/**
 *  \brief
 */
typedef struct CpswAppMem_DmaDescMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! DMA descriptor element */
    CpswDma_DmaDesc dmaDesc
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /*! DMA descriptor state, refer to CpswUtils_DescStateMemMgr */
    uint32_t dmaDescState;
}CpswAppMem_DmaDescMem;

/**
 *  \brief
 */
typedef struct CpswAppMem_RingMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Ring memory element */
    uint8_t ringEle[CPSW_UTILS_ALIGN(CPSW_APPMEM_RING_MAX_SIZE, UDMA_CACHELINE_ALIGNMENT)]
    __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));
}CpswAppMem_RingMem;

typedef queue_t CpswAppMem_DmaDescMemQ;

typedef queue_t CpswAppMem_RingMemQ;


/**
 *  \brief
 */
typedef struct CpswAppMem_MemAllocObj_s
{
    /**< DMA packet Q */
    bool memInitFlag;

    /**< DMA packet Q */
    CpswAppMem_DmaDescMemQ dmaDescFreeQ;

    /**< Ring memory Q */
    CpswAppMem_RingMemQ ringMemQ;

} CpswAppMem_MemAllocObj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

typedef struct CpswAppMem_Mem_s
{
    CpswAppMem_DmaDescMem gDmaDescMemArray[CPSW_APPMEM_NUM_DESCS]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));
    CpswAppMem_RingMem gRingMemArray[CPSW_APPMEM_NUM_RINGS]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));
}CpswAppMem_Mem;

static void *gMemBuf        = NULL;
static uint64_t gMemBufPhys  = 0;
static CpswAppMem_Mem *gMem = NULL;

/* Cpsw mem utils driver object */
static CpswAppMem_MemAllocObj gCpswAppMemObj = {.memInitFlag = false};

/* Array storing the physical address of the DMA descriptor array */
uint64_t  gDmaDescMemArrayBase = 0;
uint64_t  gDmaDescMemArrayBasePhys = 0;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */


/*! Ring memory allocation function  */
uint8_t *CpswAppMem_allocRingMemFxn(void *appPriv,
                                    uint32_t numRingEle,
                                    uint32_t alignSize)
{
    uint8_t *ringMemPtr = NULL;
    CpswAppMem_RingMem *pRingMemEle;

    if (gCpswAppMemObj.memInitFlag == true)
    {
        CpswApp_assert(numRingEle <= CPSW_APPMEM_RING_MAX_ELEM_CNT);
        pRingMemEle = (CpswAppMem_RingMem *)CpswUtils_deQ(&gCpswAppMemObj.ringMemQ);
        if (pRingMemEle != NULL)
        {
            ringMemPtr = &pRingMemEle->ringEle[0U];

            CpswApp_assert(CPSW_UTILS_IS_ALIGNED(ringMemPtr, alignSize));
        }
        else
        {
            ringMemPtr = NULL;
        }
    }

    return ringMemPtr;
}

/*! Ring memory free function  */
void CpswAppMem_freeRingMemFxn(void *appPriv,
                               void *ringMemPtr,
                               uint32_t numRingEle)
{
    CpswAppMem_RingMem *pRingMemEle;

    if (gCpswAppMemObj.memInitFlag == true)
    {
        pRingMemEle = container_of((const uint8_t *)ringMemPtr, CpswAppMem_RingMem, ringEle[0U]);
        /* TODO - just to get it compiling - Need to fix this after discussion with Misa/Badri */
        CpswUtils_enQ(&gCpswAppMemObj.ringMemQ, &pRingMemEle->node);
    }

    return;
}

/*! DMA packet allocation function  */
CpswDma_DmaDesc *CpswAppMem_allocDmaDescFxn(void *appPriv,
                                                 uint32_t alignSize)
{
    CpswDma_DmaDesc *dmaDescPtr = NULL;
    CpswAppMem_DmaDescMem *pDmaDescMem;

    if (gCpswAppMemObj.memInitFlag == true)
    {
        pDmaDescMem = (CpswAppMem_DmaDescMem *)
                      CpswUtils_deQ(&gCpswAppMemObj.dmaDescFreeQ);
        if (NULL != pDmaDescMem)
        {
            dmaDescPtr = &pDmaDescMem->dmaDesc;
            if (!CPSW_UTILS_IS_ALIGNED(dmaDescPtr, alignSize))
            {
                CpswUtils_enQ(&gCpswAppMemObj.dmaDescFreeQ,
                              &pDmaDescMem->node);
                dmaDescPtr = NULL;
            }
            else
            {
                CpswUtils_checkDescState(&pDmaDescMem->dmaDescState,
                                         CPSW_DESCSTATE_MEMMGR_FREE,
                                         CPSW_DESCSTATE_MEMMGR_ALLOC);
            }
        }
    }

    return dmaDescPtr;
}

/*! DMA packet free function  */
void CpswAppMem_freeDmaDescFxn(void *appPriv,
                               CpswDma_DmaDesc *dmaDescPtr)
{
    CpswAppMem_DmaDescMem *pDmaDescMem;

    if (gCpswAppMemObj.memInitFlag == true)
    {
        pDmaDescMem = container_of(dmaDescPtr, CpswAppMem_DmaDescMem, dmaDesc);
        CpswUtils_checkDescState(&pDmaDescMem->dmaDescState,
                                 CPSW_DESCSTATE_MEMMGR_ALLOC,
                                 CPSW_DESCSTATE_MEMMGR_FREE);
        CpswUtils_enQ(&gCpswAppMemObj.dmaDescFreeQ, &pDmaDescMem->node);
    }

    return;
}


void *cpsw_alloc(size_t size, paddr64_t *paddr)
{
    off64_t offset;
    void  *buf;

    buf = alloc_typed_mem(size, paddr);
    if (buf != MAP_FAILED)
        return buf;

    /* Will come here if the typed memory is not specified */
    buf = mmap(0, size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PHYS | MAP_PRIVATE, NOFD, 0);
    if(buf != MAP_FAILED) {
        if (mem_offset64(buf, NOFD, 1, &offset, 0) == -1)
        {
            CpswApp_print("%s:%d: Error: Could not obtain buffer physical address",
                __FUNCTION__, __LINE__);
            munmap(buf, size);
            return MAP_FAILED;
        }
        if (paddr)
        {
            *paddr = (uintptr_t) offset;
        }
    }
    CpswApp_print("%s:%d: Alloc successfull; Virt: 0x%lx, Phys: 0x%lx",
                __FUNCTION__, __LINE__, buf, offset);
    return buf;
}

void cpsw_free(void *addr, size_t size)
{
    munmap(addr, size);
}

#ifdef DEBUG_MODE
#include <assert.h>
#endif

void error_log(const char *str, const char *fileName, int32_t lineNum)
{
    CpswApp_print("\nAssertion @ Line: %d in %s: %s : failed !!!\n",
                         lineNum, fileName, str);
#ifdef DEBUG_MODE
    assert(0);
#endif
}

int32_t CpswAppMem_init(void)
{
    uint32_t i;
    int32_t retVal = CPSW_SOK;
    uint32_t alignSize = UDMA_CACHELINE_ALIGNMENT;
    uint32_t alignment_offset = 0;

    gMemBuf = cpsw_alloc(sizeof(CpswAppMem_Mem) + UDMA_CACHELINE_ALIGNMENT, (paddr64_t *)&gMemBufPhys);
    if (gMemBuf == MAP_FAILED)
    {
        CpswApp_print("%%s:%d: Error: alloc fail ",__FUNCTION__, __LINE__);
        return CPSW_EFAIL;
    }

    if (!CPSW_UTILS_IS_ALIGNED(gMemBuf, UDMA_CACHELINE_ALIGNMENT))
    {
        CpswApp_print("%s:%d: Error: addr=0x%lx is not aligned",
            __FUNCTION__, __LINE__, (uintptr_t)gMemBuf);
        gMem = (CpswAppMem_Mem *)(((uintptr_t)gMemBuf + UDMA_CACHELINE_ALIGNMENT) & ~(UDMA_CACHELINE_ALIGNMENT-1));
        alignment_offset = (uintptr_t)gMem - (uintptr_t)gMemBuf;
        gMemBufPhys += alignment_offset;
    }
    else
    {
        CpswApp_print("%s: addr=0x%lx is  aligned", __FUNCTION__, (uintptr_t)gMemBuf);
        gMem = gMemBuf;
    }

    CpswApp_print("%s: addr=0x%lx, size=0x%x, gMem=0x%lx, gMemBufPhys=0x%lx ",
        __FUNCTION__, (uintptr_t)gMemBuf, (unsigned)sizeof(CpswAppMem_Mem), (uintptr_t)gMem, (uintptr_t)gMemBufPhys);
    CpswApp_print("%s: gDmaDescMemArray=0x%lx, size=0x%x ",
        __FUNCTION__, (uintptr_t)&gMem->gDmaDescMemArray,(unsigned)sizeof(gMem->gDmaDescMemArray));
    CpswApp_print("%s: gRingMemArray=0x%lx, size=0x%x ",
        __FUNCTION__, (uintptr_t)&gMem->gRingMemArray,(unsigned)sizeof(gMem->gRingMemArray));

    if (gCpswAppMemObj.memInitFlag == false)
    {
        memset(&gCpswAppMemObj, 0U, sizeof(CpswAppMem_MemAllocObj));
        memset(gMem, 0U, sizeof(CpswAppMem_Mem));
        gDmaDescMemArrayBase = (uint64_t)gMem->gDmaDescMemArray;
        gDmaDescMemArrayBasePhys = gMemBufPhys + ((uint64_t)gDmaDescMemArrayBase - (uint64_t)gMem);
        CpswApp_print("%s: gDmaDescMemArrayBasePhys=0x%lx, gDmaDescMemArrayBase=0x%lx",
            __FUNCTION__, gDmaDescMemArrayBasePhys,gDmaDescMemArrayBase);

        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemObj.dmaDescFreeQ);
            for (i = 0U; i < CPSW_APPMEM_NUM_DESCS; i++)
            {
                gMem->gDmaDescMemArray[i].dmaDescState = 0;
                if (!CPSW_UTILS_IS_ALIGNED(&gMem->gDmaDescMemArray[i].dmaDesc, alignSize))
                {
                    CpswApp_print("%s:%d: Error: alignment issue! ",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }
                CpswUtils_enQ(&gCpswAppMemObj.dmaDescFreeQ, &gMem->gDmaDescMemArray[i].node);
                CPSW_UTILS_SET_DESC_MEMUTILS_STATE(&gMem->gDmaDescMemArray[i].dmaDescState, CPSW_DESCSTATE_MEMMGR_FREE);
            }
        }

        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemObj.ringMemQ);
            for (i = 0U; i < CPSW_APPMEM_NUM_RINGS; i++)
            {
                if (!CPSW_UTILS_IS_ALIGNED(&gMem->gRingMemArray[i].ringEle, alignSize))
                {
                    CpswApp_print("%s:%d: Error: alignment error! ",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }
                CpswUtils_enQ(&gCpswAppMemObj.ringMemQ, &gMem->gRingMemArray[i].node);
            }
        }

        if (CPSW_SOK == retVal)
        {
            gCpswAppMemObj.memInitFlag = true;
        }
    }

    CpswApp_assert(retVal == CPSW_SOK);
    return retVal;
}

void CpswAppMem_deInit(void)
{
    uint32_t i;

    if (gCpswAppMemObj.memInitFlag)
    {
        CpswApp_assert(CpswUtils_getQCount(&gCpswAppMemObj.ringMemQ) ==
                            CPSW_APPMEM_NUM_RINGS);
        CpswApp_assert(CpswUtils_getQCount(&gCpswAppMemObj.dmaDescFreeQ) ==
                            CPSW_APPMEM_NUM_DESCS);

        memset(&gCpswAppMemObj, 0U, sizeof(CpswAppMem_MemAllocObj));

        for (i = 0U; i < CPSW_APPMEM_NUM_DESCS; i++)
        {
            CpswUtils_checkDescState(&gMem->gDmaDescMemArray[i].dmaDescState,
                                     CPSW_DESCSTATE_MEMMGR_FREE,
                                     CPSW_DESCSTATE_MEMMGR_FREE);
        }
    }

    cpsw_free(gMemBuf, (sizeof(CpswAppMem_Mem) + UDMA_CACHELINE_ALIGNMENT));

}

/* ========================================================================== */
/*                          Static Function Definitions                       */
/* ========================================================================== */

/* end of file */
