/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_appboardutils_j721e_gesi.c
 *
 * \brief    This file contains the board specific utilities for J721E GESI brd.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <ti/osal/osal.h>
#include <ti/drv/i2c/I2C.h>
#include <ti/drv/i2c/soc/I2C_soc.h>
#include <ti/csl/soc.h>
#include <ti/drv/cpsw/include/dp83867.h>
#include <ti/board/board.h>
#include <ti/board/src/j721e_evm/include/board_pinmux.h>
#include <ti/board/src/j721e_evm/include/board_utils.h>
#include <ti/board/src/j721e_evm/include/board_control.h>
#include <ti/drv/cpsw/cpsw.h>

#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

/* Hack to run EthFw with Linux
 * We do only limited pinmux configuration and also don't configure RMII and MDIO board muxes
 * as it causes conflict with Linux in using Main domain I2C0. If we need any of configuration on I2c
 * we should ask Linux to do that. Currently we wouldn't use RMII port in EthFw and we don't need MDIO
 * board mux as default is CPSW9G
 */
void CpswAppBoardUtils_initEthFw(void)
{
//TODO: commenting for QNX.. need to revisit
#ifndef QNX_OS
    Board_STATUS status;
    Board_initCfg boardCfg;
    Board_initParams_t initParams;
    Board_PinmuxConfig_t pinmuxCfg;
    bool gesiBoardDetected = false;

    /* Configure CPSW9G PADs for Ports on the GESI board */
    Board_pinmuxGetCfg(&pinmuxCfg);
    pinmuxCfg.gesiExp = BOARD_PINMUX_GESI_CPSW9G;
    Board_pinmuxSetCfg(&pinmuxCfg);

    Board_getInitParams(&initParams);
    /* On Jacinto 7, EthFw running on mcu2_0 uses UART instance 2 */
    initParams.uartInst      = CPSW_UTILS_MCU2_0_UART_INSTANCE;
    initParams.uartSocDomain = BOARD_SOC_DOMAIN_MAIN;

    /* Use non-exclusive mode in board lib when accessing the module for clock enable
     * to let Linux also access the module */
    initParams.pscMode = BOARD_PSC_DEVICE_MODE_NONEXCLUSIVE;

    Board_setInitParams(&initParams);

    boardCfg = BOARD_INIT_MODULE_CLOCK_MAIN|
               BOARD_INIT_ENETCTRL_CPSW9G|
               BOARD_INIT_UART_STDIO;

    status = Board_init(boardCfg);
    CpswAppUtils_assert(status == BOARD_SOK);

    status = CpswAppBoardUtils_configEthFwPinmux();
    CpswAppUtils_assert(status == BOARD_SOK);

    gesiBoardDetected = Board_detectBoard(BOARD_ID_GESI);
    if (gesiBoardDetected)
    {
        CpswAppUtils_print("GESI board detected\n");
    }
    else
    {
        CpswAppUtils_print("Warning!!! GESI board not detected, "
                           "CPSW9G test will not work on TI EVM\n");
    }
#endif
}

void CpswAppBoardUtils_init(void)
{
//TODO: commenting for QNX.. need to revisit
#ifndef QNX_OS
    Board_STATUS status;
    Board_initCfg boardCfg;
    Board_PinmuxConfig_t pinmuxCfg;

    /* Configure CPSW9G PADs for Ports on the GESI board */
    Board_pinmuxGetCfg(&pinmuxCfg);
    pinmuxCfg.gesiExp = BOARD_PINMUX_GESI_CPSW9G;
    Board_pinmuxSetCfg(&pinmuxCfg);

#if defined(BUILD_MCU2_0)
    Board_initParams_t initParams;

    Board_getInitParams(&initParams);
    /* On Jacinto 7, EthFw running on mcu2_0 uses UART instance 2 */
    initParams.uartInst      = CPSW_UTILS_MCU2_0_UART_INSTANCE;
    initParams.uartSocDomain = BOARD_SOC_DOMAIN_MAIN;

    /* Use non-exclusive mode in board lib when accessing the module for clock enable
     * to let Linux also access the module */
    initParams.pscMode = BOARD_PSC_DEVICE_MODE_NONEXCLUSIVE;

    Board_setInitParams(&initParams);

#endif

    boardCfg = BOARD_INIT_PINMUX_CONFIG|
               BOARD_INIT_MODULE_CLOCK|
               BOARD_INIT_ENETCTRL_CPSW2G|
               BOARD_INIT_ENETCTRL_CPSW9G|
               BOARD_INIT_UART_STDIO;

    status = Board_init(boardCfg);
    CpswAppUtils_assert(status == BOARD_SOK);

#if defined(BUILD_MCU2_0)
    bool gesiBoardDetected = false;

    /* Set IO Expander to use RMII on GESI board */
    status = Board_control(BOARD_CTRL_CMD_SET_RMII_DATA_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);

    /* Enable CPSW9G MDIO mux */
    status = Board_control(BOARD_CTRL_CMD_SET_CPSW9G_MDIO_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);

    gesiBoardDetected = Board_detectBoard(BOARD_ID_GESI);
    if (gesiBoardDetected)
    {
        CpswAppUtils_print("GESI board detected\n");
    }
    else
    {
        CpswAppUtils_print("Warning!!! GESI board not detected, "
                           "CPSW9G test will not work on TI EVM\n");
    }
#endif
#endif

}

void CpswAppBoardUtils_deInit(void)
{
//TODO: commenting for QNX.. need to revisit
#ifndef QNX_OS
    Board_STATUS boardInitStatus;
    Board_initCfg boardCfg;

    boardCfg = BOARD_INIT_PINMUX_CONFIG |
               BOARD_INIT_MODULE_CLOCK |
               BOARD_INIT_UART_STDIO;

    boardInitStatus = Board_deinit(boardCfg);
    CpswAppUtils_assert(boardInitStatus == BOARD_SOK);
#endif
}

uint32_t CpswAppBoardUtils_getPhyAddr(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum)
{
    uint32_t phyAddr = CPSW_PHY_INVALID_PHYADDR;

    switch (cpswType)
    {
        case CPSW_2G:
            phyAddr = 0x00U;
            break;

        case CPSW_9G:
            switch (portNum)
            {
                case CPSW_MAC_PORT_0:
                    phyAddr = 0x0CU;
                    break;

                case CPSW_MAC_PORT_1:
                    phyAddr = 0x0FU;
                    break;

                case CPSW_MAC_PORT_2:
                    phyAddr = 0x00U;
                    break;

                case CPSW_MAC_PORT_3:
                    phyAddr = 0x03U;
                    break;

                case CPSW_MAC_PORT_7:
                    phyAddr = 0x17U;
                    break;

                default:
                    CpswAppUtils_assert(false);
                    break;
            }

            break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

    return phyAddr;
}

void CpswAppBoardUtils_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    Dp83867_Config extendedConfig;

    Dp83867_initParams(&extendedConfig);
    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoardUtils_getPhyAddr(cpswType, portNum);

    switch (cpswType)
    {
        case CPSW_2G:
            CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);

            /* DP83867 specific configuration */
            extendedConfig.txFifoDepth = 4U;

            extendedConfig.txClkShiftEnable     = true;
            extendedConfig.rxClkShiftEnable     = true;
            extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
            extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
            extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
            extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
            extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

            /* RGMII interface type */
            interface->layerType    = CPSW_MAC_LAYER_GMII;
            interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
            interface->variantType  = CPSW_MAC_VARIANT_NONE;
            break;

        case CPSW_9G:
            /* DP83867 specific configuration */
            extendedConfig.txDelayInPs          = 2750U; /* 2.75 ns */
            extendedConfig.rxDelayInPs          = 2500U; /* 2.50 ns */
            extendedConfig.txFifoDepth          = 4U;
            extendedConfig.txClkShiftEnable     = true;
            extendedConfig.rxClkShiftEnable     = true;
            extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
            extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
            extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
            extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
            extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

            /* RGMII interface type */
            interface->layerType    = CPSW_MAC_LAYER_GMII;
            interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
            interface->variantType  = CPSW_MAC_VARIANT_NONE;
            break;
    }

    Cpsw_setPhyExtendedParams(phyConfig,
                              &extendedConfig,
                              sizeof(extendedConfig));
}

void CpswAppBoardUtils_getMacAddrList(uint8_t macAddr[][CPSW_MAC_ADDR_LEN],
                                      uint32_t maxMacEntries,
                                      uint32_t *pAvailMacEntries)
{
//TODO: commenting for QNX.. need to revisit
#ifndef QNX_OS
    uint8_t macAddrBuf[CPSW_RM_NUM_MACADDRESS_MAX * BOARD_MAC_ADDR_BYTES];
    Board_STATUS boardStatus;
    uint32_t macEntries;
    uint32_t i;

    CpswAppUtils_assert(pAvailMacEntries != NULL);
    boardStatus = Board_readMacAddrCount(BOARD_ID_GESI, pAvailMacEntries);
    CpswAppUtils_assert(boardStatus == BOARD_SOK);
    CpswAppUtils_assert(*pAvailMacEntries <= CPSW_RM_NUM_MACADDRESS_MAX);
    boardStatus = Board_readMacAddr(BOARD_ID_GESI,
                                    macAddrBuf,
                                    sizeof(macAddrBuf),
                                    &macEntries);
    CpswAppUtils_assert(boardStatus == BOARD_SOK);
    CpswAppUtils_assert(*pAvailMacEntries == macEntries);
    macEntries = CpswUtils_min(*pAvailMacEntries, maxMacEntries);
    for (i = 0; i < macEntries; i++)
    {
        CPSW_UTILS_COMPILETIME_ASSERT(CPSW_MAC_ADDR_LEN == BOARD_MAC_ADDR_BYTES);
        memcpy(macAddr[i], &macAddrBuf[i * BOARD_MAC_ADDR_BYTES], CPSW_MAC_ADDR_LEN);
    }
#endif
}
