/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_phy.c
 *
 *  \brief File containing the ENETPHY functions.
 *
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Event.h>
#include <ti/sysbios/utils/Load.h>

#include <xdc/runtime/Assert.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Types.h>
#include <xdc/runtime/Timestamp.h>
#include <ti/sysbios/hal/Seconds.h>
#include <ti/sysbios/hal/Hwi.h>
#include <ti/sysbios/hal/Cache.h>

/* OSAL Header files */
#include <ti/osal/osal.h>

/* UART Header files */
#include <ti/drv/uart/UART.h>
#include <ti/drv/uart/UART_stdio.h>

/* This module's Header files */
#include "app_profile.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define UTILS_ARRAYSIZE(x) sizeof(x) / sizeof(x[0U])
#define UTILS_ALIGN(x, align)  ((((x) + ((align) - 1)) / (align)) * (align))

#define THRLOAD_MAX_NUM_TASKS             16
#define THRLOAD_LOAD_TABLE_LENGTH         100

#define APP_PROFILE_START_EVENT_ID       (Event_Id_00)
#define APP_PROFILE_STOP_EVENT_ID        (Event_Id_01)

#define HOST_PORT_NUM                    (0)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */
typedef struct loadEntry
{
    char *tskName;
    Load_Stat load_stat;
}loadEntry;

typedef struct loadEntryTbl
{
    Event_Struct profileControlEvt;
    Event_Handle hProfileControlEvt;
    uint32_t loadEntryIdx;
    struct loadEntryRecord_s
    {
        loadEntry tsk[THRLOAD_MAX_NUM_TASKS];
        Load_Stat isr;
        Load_Stat swi;
        uint32_t cpuLoad;
        uint32_t activeTaskCount;
        uint32_t totalTaskCount;
        Types_Timestamp64 windowTs64;
    } record[THRLOAD_LOAD_TABLE_LENGTH];
} loadEntryTbl;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
#pragma DATA_SECTION(apploadEntryTbl,".benchmark_buffer")
loadEntryTbl apploadEntryTbl =
{.hProfileControlEvt = NULL};
volatile uint32_t appLoadDebugFlag = 1;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
static void app_loadInit(void)
{
    apploadEntryTbl.loadEntryIdx = 0;
    // memset (&apploadEntryTbl.record[0], 0, sizeof (apploadEntryTbl.record));
}

static void app_GetTaskLoad(Task_Handle curTask,
                            loadEntry *pLocalLoadEntry,
                            uint32_t *pActiveTaskCnt)
{
    Bool retVal;

    if (NULL != curTask)
    {
        retVal = Load_getTaskLoad(curTask, &pLocalLoadEntry->load_stat);
        // Assert_isTrue((TRUE == retVal), Assert_E_assertFailed);
        if ((retVal == TRUE) && (pLocalLoadEntry->load_stat.threadTime != 0))
        {
            if (*pActiveTaskCnt < THRLOAD_MAX_NUM_TASKS)
            {
                // Assert_isTrue((TRUE == retVal), Assert_E_assertFailed);
                pLocalLoadEntry->tskName = Task_Handle_name(curTask),
                *pActiveTaskCnt          = *pActiveTaskCnt + 1;
            }
        }
    }
}

void app_loadUpdateCb(void)
{
    Task_Handle curTask = NULL, nextTask = NULL;
    uint32_t nTaskCnt, activeTaskCnt;
    uint32_t staticTaskIndex;
    loadEntry *pLocalLoadEntry;
    struct loadEntryRecord_s *curRecord;

    nTaskCnt      = 0;
    activeTaskCnt = 0;

    if (apploadEntryTbl.loadEntryIdx >= THRLOAD_LOAD_TABLE_LENGTH)
    {
        apploadEntryTbl.loadEntryIdx = 0;
    }

    curRecord = &apploadEntryTbl.record[apploadEntryTbl.loadEntryIdx];
    Timestamp_get64(&curRecord->windowTs64);
    for (staticTaskIndex = 0; staticTaskIndex < Task_Object_count(); staticTaskIndex++)
    {
        pLocalLoadEntry = &(curRecord->tsk[activeTaskCnt]);
        curTask         = Task_Object_get(NULL, staticTaskIndex);
        if (NULL != curTask)
        {
            app_GetTaskLoad(curTask, pLocalLoadEntry, &activeTaskCnt);
            nTaskCnt++;
        }
    }

    nextTask = Task_Object_first();
    do
    {
        pLocalLoadEntry = &(curRecord->tsk[activeTaskCnt]);
        curTask         = nextTask;
        if (NULL != curTask)
        {
            app_GetTaskLoad(curTask, pLocalLoadEntry, &activeTaskCnt);
            nTaskCnt++;
            nextTask = Task_Object_next(curTask);
        }
    }
    while (nextTask != NULL);

    Load_getGlobalHwiLoad(&curRecord->isr);
    Load_getGlobalSwiLoad(&curRecord->swi);
    curRecord->cpuLoad         = Load_getCPULoad();
    curRecord->activeTaskCount = activeTaskCnt;
    curRecord->totalTaskCount  = nTaskCnt;
    apploadEntryTbl.loadEntryIdx++;
}

static void loadPrintLoadStats(Load_Stat *stats)
{
    UART_printf("%u,%u,%u,%u,",
                stats->threadTime,
                stats->totalTime,
                Load_calculateLoad(stats));
}

/* Function to print Load info */
static void loadPrint(void)
{
    uint32_t windowIdx, taskIdx;
    struct loadEntryRecord_s *curRecord;

    /* Profile information */
    UART_printf("Num profile Windows, %d \n",
                apploadEntryTbl.loadEntryIdx);
    UART_printf("#,TS64Hi,TS64Lo,");
    UART_printf("NUMTSK,ACTIVETSK,"
                "CPU,"
                "HWI Thread,HWI Total,HWI Load,"
                "SWI Thread,SWI Total,SWI Load,"
                "TSK name,TSK Thread,TSK Total,TSK Load,");
    UART_printf("\n");

    for (windowIdx = 0; windowIdx < apploadEntryTbl.loadEntryIdx; windowIdx++)
    {
        curRecord = &apploadEntryTbl.record[windowIdx];
        UART_printf("%u,%u,%u,",
                    windowIdx,
                    curRecord->windowTs64.hi,
                    curRecord->windowTs64.lo);
        UART_printf("%u,",
                    curRecord->totalTaskCount);
        UART_printf("%u,",
                    curRecord->activeTaskCount);
        UART_printf("%u,",
                    curRecord->cpuLoad);
        loadPrintLoadStats(&curRecord->isr);
        loadPrintLoadStats(&curRecord->swi);
        for (taskIdx = 0; taskIdx < curRecord->activeTaskCount; taskIdx++)
        {
            UART_printf("%s,", curRecord->tsk[taskIdx].tskName);
            loadPrintLoadStats(&curRecord->tsk[taskIdx].load_stat);
        }

        UART_printf("\n");
    }

    return;
}

void app_profileStart(void)
{
    if (apploadEntryTbl.hProfileControlEvt)
    {
        Types_Timestamp64 windowTs64;

        Timestamp_get64(&windowTs64);
        UART_printf("\n\nStartTime: Hi:Lo::%u:%u\n\n", windowTs64.hi, windowTs64.lo);
        Event_post(apploadEntryTbl.hProfileControlEvt, APP_PROFILE_START_EVENT_ID);
    }
    else
    {
        while (appLoadDebugFlag)
        {
            ;
        }
    }
}

void app_profileStop(void)
{
    if (apploadEntryTbl.hProfileControlEvt)
    {
        Types_Timestamp64 windowTs64;

        Timestamp_get64(&windowTs64);
        UART_printf("\n\nStopTime: Hi:Lo::%u:%u\n\n", windowTs64.hi, windowTs64.lo);

        Event_post(apploadEntryTbl.hProfileControlEvt, APP_PROFILE_STOP_EVENT_ID);
    }
    else
    {
        while (appLoadDebugFlag)
        {
            ;
        }
    }
}

void app_profile(UArg a0,
                 UArg a1)
{
    uint32_t profilePrintInterval = (uint32_t)a0;
    uint32_t profileWindowMs      = (uint32_t)a1;
    Event_Params evtParams;
    UInt postedEvent;
    bool enableProfiling;
    uint32_t profileElapsedTime;

    Event_Params_init(&evtParams);  // Initialize this config-params structure with supplier-specified defaults before instance creation
    Event_construct(&apploadEntryTbl.profileControlEvt, &evtParams);
    apploadEntryTbl.hProfileControlEvt = Event_handle(&apploadEntryTbl.profileControlEvt);

    enableProfiling    = FALSE;
    profileElapsedTime = 0;

    while (1)
    {
        postedEvent =
            Event_pend(apploadEntryTbl.hProfileControlEvt,
                       0,
                       (APP_PROFILE_START_EVENT_ID | APP_PROFILE_STOP_EVENT_ID),
                       profileWindowMs);
        if (postedEvent)
        {
            switch (postedEvent)
            {
                case APP_PROFILE_START_EVENT_ID:
                    Load_reset();
                    app_loadInit();
                    enableProfiling    = TRUE;
                    profileElapsedTime = 0;
                    break;

                case APP_PROFILE_STOP_EVENT_ID:
                    Load_reset();
                    loadPrint();
                    enableProfiling    = FALSE;
                    profileElapsedTime = 0;
                    break;

                default:
                    /* No other event combo should be posted */
                    while (appLoadDebugFlag)
                    {
                        ;
                    }
            }
        }

        if (enableProfiling)
        {
            Load_update();
        }

        if (enableProfiling)
        {
            profileElapsedTime += profileWindowMs;
            if (profileElapsedTime >= profilePrintInterval)
            {
                Load_reset();
                loadPrint();
                app_loadInit();
                profileElapsedTime = 0;
            }
        }
    }
}

/* ========================================================================== */
/*                          Static Function Definitions                       */
/* ========================================================================== */
