/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_appmemutils.c
 *
 *  \brief CPSW DMA memory allocation utility functions.
 *
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* This is needed for memset/memcpy */
#include <string.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>

#include "inc/cpsw_appmemutils.h"
#include "inc/cpsw_appmemutils_cfg.h"
#include "inc/cpsw_apputils.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_APPMEMUTILS_NUM_DESCS ((CPSW_APPMEMUTILS_NUM_TX_PKTS * CPSW_CFG_TX_CHANNELS_NUM) + \
                                    (CPSW_APPMEMUTILS_NUM_RX_PKTS * CPSW_CFG_RX_FLOWS_NUM))

/* TDCQ is allocated in driver, we allocate only FQ and CQ */
#define CPSW_APPMEMUTILS_NUM_RINGS_TYPES (2U)

#define CPSW_APPMEMUTILS_NUM_RINGS (CPSW_APPMEMUTILS_NUM_RINGS_TYPES * \
                                    (CPSW_CFG_TX_CHANNELS_NUM + CPSW_CFG_RX_FLOWS_NUM))

#define CPSW_APPMEMUTILS_RING_MAX_SIZE \
    (CPSW_DMA_RING_MEM_SIZE * CPSW_APPMEMUTILS_RING_MAX_ELEM_CNT)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/**
 *  \brief
 */
typedef struct CpswAppMemUtils_DmaDescMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! DMA descriptor element */
    CpswDma_DmaDesc dmaDesc
    __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /*! DMA descriptor state, refer to CpswUtils_DescStateMemMgr */
    uint32_t dmaDescState;
}CpswAppMemUtils_DmaDescMem;

/**
 *  \brief
 */
typedef struct CpswAppMemUtils_RingMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Ring memory element */
    uint8_t ringEle[CPSW_UTILS_ALIGN(CPSW_APPMEMUTILS_RING_MAX_SIZE, UDMA_CACHELINE_ALIGNMENT)]
    __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));
}CpswAppMemUtils_RingMem;

/**
 *  \brief
 */
typedef struct CpswAppMemUtils_EthPktMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Eth packet info structure - shared with driver */
    CpswDma_PktInfo appPktInfo;

    /*! Original packet size - we can't use this info from appPktInfo as app can change it */
    uint32_t orgBufSize;
}CpswAppMemUtils_AppPktInfoMem;

typedef queue_t CpswAppMemUtils_DmaDescMemQ;

typedef queue_t CpswAppMemUtils_RingMemQ;

typedef queue_t CpswAppMemUtils_EthPktMemQ;

/**
 *  \brief
 */
typedef struct CpswAppMemUtils_MemAllocObj_s
{
    /**< DMA packet Q */
    bool memUtilsInitFlag;

    /**< DMA packet Q */
    CpswAppMemUtils_DmaDescMemQ dmaDescFreeQ;

    /**< Ring memory Q */
    CpswAppMemUtils_RingMemQ ringMemQ;

    /**< Pool1 ethernet packet memory Q */
    CpswAppMemUtils_EthPktMemQ ethPktMem_LargePoolQ;

    /**< Pool3 ethernet packet memory Q */
    CpswAppMemUtils_EthPktMemQ ethPktMem_MediumPoolQ;

    /**< Pool3 ethernet packet memory Q */
    CpswAppMemUtils_EthPktMemQ ethPktMem_SmallPoolQ;
} CpswAppMemUtils_MemAllocObj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

#ifdef QNX_OS

typedef struct CpswAppMemUtils_Mem_s
{

    CpswAppMemUtils_DmaDescMem gDmaDescMemArray[CPSW_APPMEMUTILS_NUM_DESCS]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));

    CpswAppMemUtils_RingMem gRingMemArray[CPSW_APPMEMUTILS_NUM_RINGS]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));

    CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_LargePool[CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM];

    uint8_t gEthPktMem_LargePool[CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM][CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));

    CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_MediumPool[CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM];

    uint8_t gEthPktMem_MediumPool[CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM][CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));

    CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_SmallPool[CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM];

    uint8_t gEthPktMem_SmallPool[CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM][CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE]
                    __attribute__(( aligned(UDMA_CACHELINE_ALIGNMENT)));
}CpswAppMemUtils_Mem;

static void             *gMemBuf = NULL;
static CpswAppMemUtils_Mem *gMem = NULL;

#else
/* CPSW DMA DESC memories */
static CpswAppMemUtils_DmaDescMem gDmaDescMemArray[CPSW_APPMEMUTILS_NUM_DESCS]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT),
                section(".bss:CPSW_DMA_DESC_MEMPOOL")));

/* RX & TX RingAcc memories */
static CpswAppMemUtils_RingMem gRingMemArray[CPSW_APPMEMUTILS_NUM_RINGS]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT),
                section(".bss:CPSW_DMA_RING_MEMPOOL")));

/* Eth packet info memory Q - large pool */
static CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_LargePool[CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM];

/* Eth packet large pool memories */
static uint8_t gEthPktMem_LargePool[CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM][CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT),
                section(".bss:CPSW_DMA_PKT_MEMPOOL")));

/* Eth packet info memory Q - medium pool */
static CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_MediumPool[CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM];

/* Eth packet medium pool memories */
static uint8_t gEthPktMem_MediumPool[CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM][CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT),
                section(".bss:CPSW_DMA_PKT_MEMPOOL")));

/* Eth packet info memory Q - small pool */
static CpswAppMemUtils_AppPktInfoMem gAppPktInfoMem_SmallPool[CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM];

/* Eth packet medium pool memories */
static uint8_t gEthPktMem_SmallPool[CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM][CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT),
                section(".bss:CPSW_DMA_PKT_MEMPOOL")));
#endif
/* Cpsw mem utils driver object */
static CpswAppMemUtils_MemAllocObj gCpswAppMemUtilsObj = {.memUtilsInitFlag = false};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

/*! Ethernet packet allocation function  */
CpswDma_PktInfo *CpswAppMemUtils_allocEthPktFxn(void *appPriv,
                                                uint32_t pktSize,
                                                uint32_t alignSize)
{
    CpswDma_PktInfo *pPktInfo                     = NULL;
    CpswAppMemUtils_AppPktInfoMem *pAppPktInfoMem = NULL;
    CpswAppMemUtils_EthPktMemQ *selectedQ         = NULL;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        if (CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE >= pktSize)
        {
            /* Allocate packet from  smallest pool*/
            selectedQ = &gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ;
        }
        else if (CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE >= pktSize)
        {
            selectedQ = &gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ;
        }
        else if (CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE >= pktSize)
        {
            selectedQ = &gCpswAppMemUtilsObj.ethPktMem_LargePoolQ;
        }
        else
        {
            selectedQ = NULL;
        }

        if (NULL != selectedQ)
        {
            pAppPktInfoMem = (CpswAppMemUtils_AppPktInfoMem *)CpswUtils_deQ(selectedQ);
        }

        if (NULL != pAppPktInfoMem)
        {
            pPktInfo = &pAppPktInfoMem->appPktInfo;

            if (!CPSW_UTILS_IS_ALIGNED(pPktInfo->bufPtr, alignSize))
            {
                CpswUtils_enQ(selectedQ, &pAppPktInfoMem->node);
                pPktInfo = NULL;
            }
            else
            {
                pPktInfo->appPriv = (void *)appPriv;
                CpswUtils_checkPktState(&pPktInfo->pktState,
                                        CPSW_PKTSTATE_MODULE_MEMUTILS,
                                        CPSW_PKTSTATE_MEMMGR_FREE,
                                        CPSW_PKTSTATE_MEMMGR_ALLOC);
                CPSW_UTILS_SET_PKT_DRIVER_STATE(&pPktInfo->pktState,
                                                CPSW_PKTSTATE_DMA_NOT_WITH_HW);
            }
        }
        else
        {
            pPktInfo = NULL;
        }
    }

    return pPktInfo;
}

/*! Ethernet packet free function  */
void CpswAppMemUtils_freeEthPktFxn(CpswDma_PktInfo *pPktInfo)
{
    CpswAppMemUtils_AppPktInfoMem *pAppPktInfoMem;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        CpswUtils_checkPktState(&pPktInfo->pktState,
                                CPSW_PKTSTATE_MODULE_MEMUTILS,
                                CPSW_PKTSTATE_MEMMGR_ALLOC,
                                CPSW_PKTSTATE_MEMMGR_FREE);
        pPktInfo->appPriv = NULL;
        pAppPktInfoMem    = container_of(pPktInfo, CpswAppMemUtils_AppPktInfoMem, appPktInfo);

        CPSW_UTILS_COMPILETIME_ASSERT(offsetof(CpswAppMemUtils_AppPktInfoMem, node) == 0);
        if (CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE == pAppPktInfoMem->orgBufSize)
        {
            CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ,
                          &pAppPktInfoMem->node);
        }
        else if (CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE == pAppPktInfoMem->orgBufSize)
        {
            CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ,
                          &pAppPktInfoMem->node);
        }
        else if (CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE == pAppPktInfoMem->orgBufSize)
        {
            CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ,
                          &pAppPktInfoMem->node);
        }
        else
        {
            /* packet is not allocated by me */
            CpswAppUtils_assert(false);
        }
    }

    return;
}

/*! Ring memory allocation function  */
uint8_t *CpswAppMemUtils_allocRingMemFxn(void *appPriv,
                                         uint32_t numRingEle,
                                         uint32_t alignSize)
{
    uint8_t *ringMemPtr = NULL;
    CpswAppMemUtils_RingMem *pRingMemEle;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        CpswAppUtils_assert(numRingEle <= CPSW_APPMEMUTILS_RING_MAX_ELEM_CNT);
        pRingMemEle = (CpswAppMemUtils_RingMem *)CpswUtils_deQ(&gCpswAppMemUtilsObj.ringMemQ);
        if (pRingMemEle != NULL)
        {
            ringMemPtr = &pRingMemEle->ringEle[0U];

            CpswAppUtils_assert(CPSW_UTILS_IS_ALIGNED(ringMemPtr, alignSize));
        }
        else
        {
            ringMemPtr = NULL;
        }
    }

    return ringMemPtr;
}

/*! Ring memory free function  */
void CpswAppMemUtils_freeRingMemFxn(void *appPriv,
                                    void *ringMemPtr,
                                    uint32_t numRingEle)
{
    CpswAppMemUtils_RingMem *pRingMemEle;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        pRingMemEle = container_of((const uint8_t *)ringMemPtr, CpswAppMemUtils_RingMem, ringEle[0U]);
        /* TODO - just to get it compiling - Need to fix this after discussion with Misa/Badri */
        CpswUtils_enQ(&gCpswAppMemUtilsObj.ringMemQ, &pRingMemEle->node);
    }

    return;
}

/*! DMA packet allocation function  */
CpswDma_DmaDesc *CpswAppMemUtils_allocDmaDescFxn(void *appPriv,
                                                 uint32_t alignSize)
{
    CpswDma_DmaDesc *dmaDescPtr = NULL;
    CpswAppMemUtils_DmaDescMem *pDmaDescMem;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        pDmaDescMem = (CpswAppMemUtils_DmaDescMem *)
                      CpswUtils_deQ(&gCpswAppMemUtilsObj.dmaDescFreeQ);
        if (NULL != pDmaDescMem)
        {
            dmaDescPtr = &pDmaDescMem->dmaDesc;
            if (!CPSW_UTILS_IS_ALIGNED(dmaDescPtr, alignSize))
            {
                CpswUtils_enQ(&gCpswAppMemUtilsObj.dmaDescFreeQ,
                              &pDmaDescMem->node);
                dmaDescPtr = NULL;
            }
            else
            {
                CpswUtils_checkDescState(&pDmaDescMem->dmaDescState,
                                         CPSW_DESCSTATE_MEMMGR_FREE,
                                         CPSW_DESCSTATE_MEMMGR_ALLOC);
            }
        }
    }

    return dmaDescPtr;
}

/*! DMA packet free function  */
void CpswAppMemUtils_freeDmaDescFxn(void *appPriv,
                                    CpswDma_DmaDesc *dmaDescPtr)
{
    CpswAppMemUtils_DmaDescMem *pDmaDescMem;

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == true)
    {
        pDmaDescMem = container_of(dmaDescPtr, CpswAppMemUtils_DmaDescMem, dmaDesc);
        CpswUtils_checkDescState(&pDmaDescMem->dmaDescState,
                                 CPSW_DESCSTATE_MEMMGR_ALLOC,
                                 CPSW_DESCSTATE_MEMMGR_FREE);
        CpswUtils_enQ(&gCpswAppMemUtilsObj.dmaDescFreeQ, &pDmaDescMem->node);
    }

    return;
}

#ifdef QNX_OS

void *cpsw_alloc(size_t size, paddr64_t *paddr)
{
    off64_t offset;
    void  *buf;
    
    buf = mmap(0, size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PHYS | MAP_PRIVATE, NOFD, 0);

    if(buf != MAP_FAILED) {
        if (mem_offset64(buf, NOFD, 1, &offset, 0) == -1)
        {
            CpswAppUtils_print("Error: Could not obtain buffer physical address\n");
            munmap(buf, size);
            return MAP_FAILED;
        }
        if (paddr)
        {
            *paddr = (uint32_t) offset;
        }
    }
    return buf;
}

void cpsw_free(void *addr, size_t size)
{
    munmap(addr, size);
}

#endif

int32_t CpswAppMemUtils_init(void)
{
    uint32_t i;
    int32_t retVal = CPSW_SOK;
    CpswDma_PktInfo *pAppPktInfo;
    CpswAppMemUtils_AppPktInfoMem *pAppPktInfoMem;
    uint32_t alignSize = UDMA_CACHELINE_ALIGNMENT;

#ifdef QNX_OS
    //gMem = (CpswAppMemUtils_Mem *)mmap_device_memory(0, sizeof(CpswAppMemUtils_Mem), PROT_READ|PROT_WRITE, 0, CPSW_RESERVE_ADDR);
    gMemBuf = cpsw_alloc(sizeof(CpswAppMemUtils_Mem) + UDMA_CACHELINE_ALIGNMENT, NULL);
    if (gMemBuf == MAP_FAILED)
    {
        CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
        return CPSW_EFAIL;
    }
    gMem = (CpswAppMemUtils_Mem *)(((long)gMemBuf + UDMA_CACHELINE_ALIGNMENT) & ~(UDMA_CACHELINE_ALIGNMENT-1));

    CpswAppUtils_print("%s: addr=0x%x, size=0x%x, gMem=0x%x\n",
        __FUNCTION__, (long)gMemBuf, (unsigned)sizeof(CpswAppMemUtils_Mem), (unsigned)(uintptr_t)gMem);
    CpswAppUtils_print("%s: gDmaDescMemArray=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gDmaDescMemArray,(unsigned)sizeof(gMem->gDmaDescMemArray));
    CpswAppUtils_print("%s: gRingMemArray=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gRingMemArray,(unsigned)sizeof(gMem->gRingMemArray));
    CpswAppUtils_print("%s: gAppPktInfoMem_LargePool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gAppPktInfoMem_LargePool,(unsigned)sizeof(gMem->gAppPktInfoMem_LargePool));
    CpswAppUtils_print("%s: gEthPktMem_LargePool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gEthPktMem_LargePool,(unsigned)sizeof(gMem->gEthPktMem_LargePool));
    CpswAppUtils_print("%s: gAppPktInfoMem_MediumPool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gAppPktInfoMem_MediumPool,(unsigned)sizeof(gMem->gAppPktInfoMem_MediumPool));
    CpswAppUtils_print("%s: gEthPktMem_MediumPool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gEthPktMem_MediumPool,(unsigned)sizeof(gMem->gEthPktMem_MediumPool));
    CpswAppUtils_print("%s: gAppPktInfoMem_SmallPool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gAppPktInfoMem_SmallPool,(unsigned)sizeof(gMem->gAppPktInfoMem_SmallPool));
    CpswAppUtils_print("%s: gEthPktMem_SmallPool=0x%x, size=0x%x \n",
        __FUNCTION__, (unsigned)(uintptr_t)&gMem->gEthPktMem_SmallPool,(unsigned)sizeof(gMem->gEthPktMem_SmallPool));

    if (gCpswAppMemUtilsObj.memUtilsInitFlag == false)
    {
        memset(&gCpswAppMemUtilsObj, 0U, sizeof(CpswAppMemUtils_MemAllocObj));
        memset(gMem, 0U, sizeof(CpswAppMemUtils_Mem));

        /*********************** DMA packet Q ************************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.dmaDescFreeQ);
            /* Initialize the DMA packet and enqueue into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_NUM_DESCS; i++)
            {
                gMem->gDmaDescMemArray[i].dmaDescState = 0;
                if (!CPSW_UTILS_IS_ALIGNED(&gMem->gDmaDescMemArray[i].dmaDesc, alignSize))
                {
                    CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }

                CpswUtils_enQ(&gCpswAppMemUtilsObj.dmaDescFreeQ, &gMem->gDmaDescMemArray[i].node);
                CPSW_UTILS_SET_DESC_MEMUTILS_STATE(&gMem->gDmaDescMemArray[i].dmaDescState, CPSW_DESCSTATE_MEMMGR_FREE);
            }
        }

        /*********************** Ring Mem Q ************************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ringMemQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_NUM_RINGS; i++)
            {
                if (!CPSW_UTILS_IS_ALIGNED(&gMem->gRingMemArray[i].ringEle, alignSize))
                {
                    CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }

                CpswUtils_enQ(&gCpswAppMemUtilsObj.ringMemQ, &gMem->gRingMemArray[i].node);
            }
        }

        /****************** memory Q for Large pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gMem->gAppPktInfoMem_LargePool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gMem->gEthPktMem_LargePool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;

                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState,
                                                  CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ, &pAppPktInfoMem->node);
            }
        }

        /****************** memory Q for Medium pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gMem->gAppPktInfoMem_MediumPool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gMem->gEthPktMem_MediumPool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState, CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ, &pAppPktInfoMem->node);
            }
        }

        /****************** memory Q for Small pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gMem->gAppPktInfoMem_SmallPool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gMem->gEthPktMem_SmallPool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;

                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    CpswAppUtils_print("%s:%d: error! \n",__FUNCTION__, __LINE__);
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState, CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ, &pAppPktInfoMem->node);
            }
        }

        if (CPSW_SOK == retVal)
        {
            gCpswAppMemUtilsObj.memUtilsInitFlag = true;
        }
    }
#else
    if (gCpswAppMemUtilsObj.memUtilsInitFlag == false)
    {
        memset(&gCpswAppMemUtilsObj, 0U, sizeof(CpswAppMemUtils_MemAllocObj));
        memset(&gRingMemArray, 0U, sizeof(gRingMemArray));

        /*********************** DMA packet Q ************************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.dmaDescFreeQ);
            /* Initialize the DMA packet and enqueue into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_NUM_DESCS; i++)
            {
                gDmaDescMemArray[i].dmaDescState = 0;
                if (!CPSW_UTILS_IS_ALIGNED(&gDmaDescMemArray[i].dmaDesc, alignSize))
                {
                    retVal = CPSW_EFAIL;
                    break;
                }

                CpswUtils_enQ(&gCpswAppMemUtilsObj.dmaDescFreeQ, &gDmaDescMemArray[i].node);
                CPSW_UTILS_SET_DESC_MEMUTILS_STATE(&gDmaDescMemArray[i].dmaDescState, CPSW_DESCSTATE_MEMMGR_FREE);
            }
        }

        /*********************** Ring Mem Q ************************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ringMemQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_NUM_RINGS; i++)
            {
                if (!CPSW_UTILS_IS_ALIGNED(&gRingMemArray[i].ringEle, alignSize))
                {
                    retVal = CPSW_EFAIL;
                    break;
                }

                CpswUtils_enQ(&gCpswAppMemUtilsObj.ringMemQ, &gRingMemArray[i].node);
            }
        }

        /****************** memory Q for Large pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gAppPktInfoMem_LargePool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gEthPktMem_LargePool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;

                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState,
                                                  CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ, &pAppPktInfoMem->node);
            }
        }

        /****************** memory Q for Medium pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gAppPktInfoMem_MediumPool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gEthPktMem_MediumPool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_MEDIUM_POOL_PKT_SIZE;
                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState, CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ, &pAppPktInfoMem->node);
            }
        }

        /****************** memory Q for Small pool ethernet packets *****************/
        if (CPSW_SOK == retVal)
        {
            CpswUtils_initQ(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ);
            /* Enqueue ring memory element into free packet Q */
            for (i = 0U; i < CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM; i++)
            {
                pAppPktInfoMem = &gAppPktInfoMem_SmallPool[i];
                /* Keep record of allocated size - we use this in free */
                pAppPktInfoMem->orgBufSize = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;
                pAppPktInfo                = &pAppPktInfoMem->appPktInfo;
                memset(&pAppPktInfo->node, 0U, sizeof(pAppPktInfo->node));
                pAppPktInfo->bufPtr     = &gEthPktMem_SmallPool[i][0U];
                pAppPktInfo->userBufLen = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;
                pAppPktInfo->orgBufLen  = CPSW_APPMEMUTILS_SMALL_POOL_PKT_SIZE;

                if (!CPSW_UTILS_IS_ALIGNED(pAppPktInfo->bufPtr, alignSize))
                {
                    retVal = CPSW_EFAIL;
                    break;
                }

                CPSW_UTILS_SET_PKT_MEMUTILS_STATE(&pAppPktInfo->pktState, CPSW_PKTSTATE_MEMMGR_FREE);
                CpswUtils_enQ(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ, &pAppPktInfoMem->node);
            }
        }

        if (CPSW_SOK == retVal)
        {
            gCpswAppMemUtilsObj.memUtilsInitFlag = true;
        }
    }
#endif

    CpswAppUtils_assert(retVal == CPSW_SOK);
    return retVal;
}

void CpswAppMemUtils_deInit(void)
{
    uint32_t i;
    CpswAppMemUtils_AppPktInfoMem *pAppPktInfoMem;
#ifdef QNX_OS
    if (gCpswAppMemUtilsObj.memUtilsInitFlag)
    {
        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ringMemQ) ==
                            CPSW_APPMEMUTILS_NUM_RINGS);
        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.dmaDescFreeQ) ==
                            CPSW_APPMEMUTILS_NUM_DESCS);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ) ==
                            CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ) ==
                            CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ) ==
                            CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM);

        memset(&gCpswAppMemUtilsObj, 0U, sizeof(CpswAppMemUtils_MemAllocObj));

        for (i = 0U; i < CPSW_APPMEMUTILS_NUM_DESCS; i++)
        {
            CpswUtils_checkDescState(&gMem->gDmaDescMemArray[i].dmaDescState,
                                     CPSW_DESCSTATE_MEMMGR_FREE,
                                     CPSW_DESCSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gMem->gAppPktInfoMem_SmallPool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gMem->gAppPktInfoMem_MediumPool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gMem->gAppPktInfoMem_LargePool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }
    }

    cpsw_free(gMemBuf, (sizeof(CpswAppMemUtils_Mem) + UDMA_CACHELINE_ALIGNMENT));

#else
    if (gCpswAppMemUtilsObj.memUtilsInitFlag)
    {
        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ringMemQ) ==
                            CPSW_APPMEMUTILS_NUM_RINGS);
        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.dmaDescFreeQ) ==
                            CPSW_APPMEMUTILS_NUM_DESCS);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_LargePoolQ) ==
                            CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_MediumPoolQ) ==
                            CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM);

        CpswAppUtils_assert(
                            CpswUtils_getQCount(&gCpswAppMemUtilsObj.ethPktMem_SmallPoolQ) ==
                            CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM);

        memset(&gCpswAppMemUtilsObj, 0U, sizeof(CpswAppMemUtils_MemAllocObj));

        for (i = 0U; i < CPSW_APPMEMUTILS_NUM_DESCS; i++)
        {
            CpswUtils_checkDescState(&gDmaDescMemArray[i].dmaDescState,
                                     CPSW_DESCSTATE_MEMMGR_FREE,
                                     CPSW_DESCSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_SMALL_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gAppPktInfoMem_SmallPool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_MEDIUM_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gAppPktInfoMem_MediumPool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }

        for (i = 0U; i < CPSW_APPMEMUTILS_LARGE_POOL_PACKET_NUM; i++)
        {
            pAppPktInfoMem = &gAppPktInfoMem_LargePool[i];
            CpswDma_PktInfo *pAppPktInfo = &pAppPktInfoMem->appPktInfo;

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_MEMUTILS,
                                    CPSW_PKTSTATE_MEMMGR_FREE,
                                    CPSW_PKTSTATE_MEMMGR_FREE);
        }
    }
#endif
}

/* ========================================================================== */
/*                          Static Function Definitions                       */
/* ========================================================================== */

/* end of file */
