/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_apputils.h
 *
 * \brief    This file contains the function prototypes of the CPSW
 *           application utility functions used in the CPSW examples.
 *
 * NOTE: This library is meant only for CPSW examples. Customers are not
 * encouraged to use this layer as these are very specific to the examples
 * written and the API behaviour and signature can change at any time to
 * suit the examples.
 */

#ifndef CPSW_APPUTILS_H_
#define CPSW_APPUTILS_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
#ifdef QNX_OS
#include <hw/inout.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#endif

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>

#include <ti/drv/cpsw/nimucpsw/ndk2cpsw_appif.h>
#include "cpswapp_ethutils.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/** \brief Size of an array */
#define CPSWAPPUTILS_ROUND_UP(x, y)     ((((x) + ((y) - 1U)) / (y)) * (y))
#define CPSWAPPUTILS_ALIGN(x)           (CPSWAPPUTILS_ROUND_UP((x), 128U))

/* \brief Octets of payload per console row */
#define OCTETS_PER_ROW                  (16U)
/** \brief Max frame length */
#define ETH_MAX_FRAME_LEN               (1522U)

/* \brief Support macros for MMR lock/unlock functions */
#define MMR_KICK0_UNLOCK_VAL            (0x68EF3490U)
#define MMR_KICK1_UNLOCK_VAL            (0xD172BC5AU)
#define MMR_KICK_LOCK_VAL               (0x00000000U)

#define CSL_MMR_KICK_UNLOCKED_MASK          (0x00000001U)
#define CSL_MMR_KICK_UNLOCKED_SHIFT         (0x0U)

#define ENET_CTRL_RGMII_ID_SHIFT            (4U)
#define ENET_CTRL_RGMII_ID_INTTXDLY         (0U)
#define ENET_CTRL_RGMII_ID_NODELAY          (1U)

#define CPSW_UTILS_MCU2_0_UART_INSTANCE     (2U)

#define CpswAppUtils_assert(cond)                                     \
    (CpswAppUtils_assertLocal((bool) (cond), (const char *) # cond, \
                    (const char *) __FILE__, (int32_t) __LINE__))

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/**
 *  \brief  Type/Speed/Duplex connection type
 *
 *  Ethernet connections based on the type (MII, RMII, RGMII), speed
 *  (10Mbps, 100Mbps, 1Gbps) and duplexity (half, full).
 */
typedef enum
{
    MAC_CONN_TYPE_RMII_10               = 0x01U,
    /**< MAC connection type for 10Mbps RMII mode */
    MAC_CONN_TYPE_RMII_100              = 0x02U,
    /**< MAC connection type for 100Mbps RMII mode */
    MAC_CONN_TYPE_RGMII_FORCE_100_HALF  = 0x03U,
    /**< MAC connection type for forced half-duplex 100Mbps RGMII mode */
    MAC_CONN_TYPE_RGMII_FORCE_100_FULL  = 0x04U,
    /**< MAC connection type for forced full-duplex 100Mbps RGMII mode */
    MAC_CONN_TYPE_RGMII_FORCE_1000_FULL = 0x05U,
    /**< MAC connection type for forced full-duplex 1000Mbps RGMII mode */
    MAC_CONN_TYPE_RGMII_DETECT_INBAND   = 0x06U,

    /**< MAC connection type for RGMII inband detection mode
     *   (speed determined based on received RGMII Rx clock) */
} CpswAppUtils_MacConType;

/**
 *  \brief  Clkout frequency select enum
 */
typedef enum
{
    CPSWAPPUTILS_CLKOUT_FREQ_50MHZ = 0U,
    /**< Select 50MHz output on clkout pin */
    CPSWAPPUTILS_CLKOUT_FREQ_25MHZ,
    /**< Select 25MHz output on clkout pin */
} CpswAppUtils_ClkOutFreqType;

/**
 *  \brief  Lock/unlock enum
 */
typedef enum
{
    CPSWAPPUTILS_LOCK_MMR = 0U,
    /**< lock control MMR */
    CPSWAPPUTILS_UNLOCK_MMR,
    /**< unlock control MMR */
} CpswAppUtils_MmrLockType;

/**
 *  \brief CTRL MMR enum for MCU and Main domain
 */
typedef enum
{
    CPSWAPPUTILS_MMR_LOCK0 = 0U,
    /**< control MMR lock 0*/
    CPSWAPPUTILS_MMR_LOCK1,
    /**< control MMR lock 1*/
    CPSWAPPUTILS_MMR_LOCK2,
    /**< control MMR lock 2*/
    CPSWAPPUTILS_MMR_LOCK3,
    /**< control MMR lock 3*/
    CPSWAPPUTILS_MMR_LOCK4,
    /**< control MMR lock 4*/
    CPSWAPPUTILS_MMR_LOCK5,
    /**< control MMR lock 5*/
    CPSWAPPUTILS_MMR_LOCK6,
    /**< control MMR lock 6*/
    CPSWAPPUTILS_MMR_LOCK7,
    /**< control MMR lock 7*/
} CpswAppUtils_CtrlMmrType;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Perform cache writeback operation
 *
 * Performs cache write-back operation based on whether cache is coherent
 * or not. This internally uses the OSAL API.
 *
 * \param addr  Start address of the cache line/s
 * \param size  Size (in bytes) of the memory to be written back
 *
 */
void CpswAppUtils_cacheWb(const void *addr,
                          int32_t size);

/*!
 * \brief Perform cache invalidate operation
 *
 * Perform cache invalidate operation based on whether cache is
 * coherent or not. This internally uses the OSAL API.
 *
 * \param addr  Start address of the cache line/s
 * \param size  Size (in bytes) of the memory to invalidate
 *
 */
void CpswAppUtils_cacheInv(const void *addr,
                           int32_t size);

/*!
 * \brief Perform cache writeback and invalidate operation
 *
 * Performs cache write-back and invalidation operation based on whether
 * cache is coherent or not. This internally uses the OSAL API.
 *
 * \param addr  Start address of the cache line/s
 * \param size  Size (in bytes) of the memory to writeback and invalidate
 */
void CpswAppUtils_cacheWbInv(const void *addr,
                             int32_t size);

/*!
 * \brief Prints the string on UART or console
 */
void CpswAppUtils_print(const char *pcString,
                        ...);

/*!
 * \brief Input utility used to accept a character from terminal
 */
char CpswAppUtils_getChar(void);

/*!
 * \brief Input utility used to accept a number from terminal
 */
int32_t CpswAppUtils_getNum(void);

/*!
 * \brief Wait function utility. In TI RTOS case, we sleep for waitTime whereas
 *        in baremetal, we busy loop for approx. no. of cycles equivalent of
 *        waitTime
 */
void CpswAppUtils_wait(uint32_t waitTime);

/*!
 * \brief Rand number function
 */
uint32_t CpswAppUtils_randFxn(uint32_t min,
                              uint32_t max);

/*!
 * \brief Wait for emulator connection
 */
void CpswAppUtils_waitEmuConnect(void);

/**
 *  \brief Returns if print can be supported for a platform/build
 *
 *  \return TRUE if print can be supported for a platform. Else FALSE
 */
uint32_t CpswAppUtils_isPrintSupported(void);

/**
 *  \brief Allocates memory for UDMA drv object and open UDMA driver
 *
 *  \return Non-NULL handle to UDMA driver
 */
Udma_DrvHandle CpswAppUtils_udmaOpen(Cpsw_Type cpswType,
                                     Udma_InitPrms *pInitPrms);

/**
 *  \brief Closes UDMA driver earlier opened via CpswAppUtils_udmaOpen
 */
void CpswAppUtils_udmaclose(Udma_DrvHandle hUdmaDrv);

/**
 *  \brief Utils function for printing "non-zero" MAC port statistics for CPSW2G
 */
void CpswAppUtils_printMacPortStats2G(CpswStats_MacPort_2g *st);

/**
 *  \brief Utils function for printing "non-zero" Host port statistics for CPSW2G
 */
void CpswAppUtils_printHostPortStats2G(CpswStats_HostPort_2g *st);

/**
 *  \brief Utils function for printing "non-zero" MAC port statistics for CPSW9G
 */
void CpswAppUtils_printMacPortStats9G(CpswStats_MacPort_9g *st);

/**
 *  \brief Utils function for printing "non-zero" Host port statistics for CPSW9G
 */
void CpswAppUtils_printHostPortStats9G(CpswStats_HostPort_9g *st);

/**
 *  \brief Utils function for printing Ethernet frame with src MAC, dest MAC,
 *         VLAN type
 */
void CpswAppUtils_printFrame(EthFrame *frame,
                             uint32_t len);

/**
 *  \brief Utils function for printing MAC address
 */
void CpswAppUtils_printMacAddr(uint8_t macAddr[]);

/**
 *  \brief Submits TX ready packet for transmission and copies unused(returned by
 *         HW) packets back to ready Q so it can be submitted later.
 *
 *  \return Success or failure
 */
int32_t CpswAppUtils_submitTxPackets(CpswDma_TxChHandle hTxChannel,
                                     CpswDma_PktInfoQ *pSubmitQ);

/**
 *  \brief Submits Rx free packets to HW and copies unused(returned by HW)
 *         packets back to free Q so it can be submitted later.
 */
void CpswAppUtils_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                  CpswDma_PktInfoQ *pSubmitQ);

/**
 *  \brief Validates the state of all packets in the given
 *         queue with the expectedState and changes it to
 *         newState.
 */
void CpswAppUtils_validatePacketState(CpswDma_PktInfoQ *pQueue,
                                      uint32_t expectedState,
                                      uint32_t newState);

/**
 *  \brief Retrieves packets from Completion queue and Free
 *         queue and frees the packets.
 */
void CpswAppUtils_freePktInfoQ(CpswDma_PktInfoQ *pPktInfoQ);

/**
 *  \brief Utility function to enable CPSW module clocks via SCI_CLIENT
 */
void CpswAppUtils_enableClocks(Cpsw_Type cpswType,
                               CpswAppUtils_MacConType macType);

/**
 *  \brief Utility function to disable CPSW module clocks via SCI_CLIENT
 */
void CpswAppUtils_disableClocks(Cpsw_Type cpswType);

/**
 *  \brief Utility function to convert CPU virtual address to SoC 64-bit address
 *         required by UDMA
 */
uint64_t CpswAppUtils_virtToPhyFxn(const void *virtAddr,
                                   void *appData);

/**
 *  \brief Utility function to convert SoC 64-bit address to CPU virtual address
 */
void *CpswAppUtils_phyToVirtFxn(uint64_t phyAddr,
                                void *appData);

/**
 *  \brief lock/unlock the MCU MMR CNTR register
 */
void CpswAppUtils_mcuMmrCtrl(CpswAppUtils_CtrlMmrType mmrNum,
                             CpswAppUtils_MmrLockType lock);

/**
 *  \brief Lock/unlock the Main MMR CNTR register
 */
void CpswAppUtils_mainMmrCtrl(CpswAppUtils_CtrlMmrType mmrNum,
                              CpswAppUtils_MmrLockType lock);

/**
 *  \brief Configure CLK_OUT for CPSW2G and 9G.
 */
void CpswAppUtils_enableClkOut(Cpsw_Type cpswType,
                               CpswAppUtils_ClkOutFreqType clkOut);

/**
 *  \brief Utils API to get NAVSS instance id abstracting all usage anomalies
 */
uint32_t CpswAppUtils_getNavSSInstanceId(Cpsw_Type cpswType);

/**
 *  \brief Set loopback configuration for RGMII mode
 */
void CpswAppUtils_setNoPhyCfgRgmii(CpswMacPort_Interface *interface,
                                   CpswPhy_Config *phyConfig);

/**
 *  \brief Utility function to get cpsw type and mac port configuration from user
 */
void CpswAppUtils_getTestConfig(Cpsw_Type *cpswType,
                                Cpsw_MacPort *macPort);

/**
 *  \brief Sets common TX channel parameters when used apputils and memutils
 */
void CpswAppUtils_setCommonTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms);

/**
 *  \brief Sets common RX Flow parameters when used apputils and memutils
 */
void CpswAppUtils_setCommonRxFlowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms);

/**
 *  \brief Utils assert function
 */
static inline void CpswAppUtils_assertLocal(bool condition,
                                            const char *str,
                                            const char *fileName,
                                            int32_t lineNum);

int32_t CpswAppUtils_allocRxFlow(Cpsw_Handle hCpsw,
                                 uint32_t coreKey,
                                 uint32_t coreId,
                                 uint32_t *rxFlowStartIdx,
                                 uint32_t *rxFlowIdx);

int32_t CpswAppUtils_allocTxCh(Cpsw_Handle hCpsw,
                               uint32_t coreKey,
                               uint32_t coreId,
                               uint32_t *txPSILThreadId);

int32_t CpswAppUtils_freeRxFlow(Cpsw_Handle hCpsw,
                                uint32_t coreKey,
                                uint32_t coreId,
                                uint32_t rxFlowIdx);

int32_t CpswAppUtils_freeTxCh(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              uint32_t txChNum);

int32_t CpswAppUtils_showRxFlowStats(CpswDma_RxFlowHandle hRxFlow);

int32_t CpswAppUtils_showTxChStats(CpswDma_TxChHandle hTxCh);

int32_t CpswAppUtils_unregisterDstMacRxFlow(Cpsw_Handle hCpsw,
                                            uint32_t coreKey,
                                            uint32_t coreId,
                                            uint32_t rxFlowStartIdx,
                                            uint32_t rxFlowIdx,
                                            uint8_t *macAddress);

int32_t CpswAppUtils_registerDstMacRxFlow(Cpsw_Handle hCpsw,
                                          uint32_t coreKey,
                                          uint32_t coreId,
                                          uint32_t rxFlowStartIdx,
                                          uint32_t rxFlowIdx,
                                          uint8_t *macAddress);

int32_t CpswAppUtils_unregisterDefaultRxFlow(Cpsw_Handle hCpsw,
                                             uint32_t coreKey,
                                             uint32_t coreId,
                                             uint32_t rxFlowStartIdx,
                                             uint32_t rxFlowIdx);

int32_t CpswAppUtils_registerDefaultRxFlow(Cpsw_Handle hCpsw,
                                           uint32_t coreKey,
                                           uint32_t coreId,
                                           uint32_t rxFlowStartIdx,
                                           uint32_t rxFlowIdx);

int32_t CpswAppUtils_allocMac(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              uint8_t *macAddress);

int32_t CpswAppUtils_freeMac(Cpsw_Handle hCpsw,
                             uint32_t coreKey,
                             uint32_t coreId,
                             uint8_t *macAddress);

bool CpswAppUtils_isPortLinkUp(Cpsw_Handle hCpsw,
                               uint32_t coreId,
                               Cpsw_MacPort portNum);

/**
 *  \brief Utility function to init resource config structure part of Cpsw_Config
 */
void CpswAppUtils_initResourceConfig(Cpsw_Type cpswType,
                                     uint32_t selfCoreId,
                                     CpswRm_ResourceConfig *rscCfg);

uint32_t CpswAppUtils_getStartFlowIdx(Cpsw_Handle hCpsw,
                                      uint32_t coreId);

void CpswAppUtils_absFlowIdx2FlowIdxOffset(Cpsw_Handle hCpsw,
                                           uint32_t coreId,
                                           uint32_t absRxFlowId,
                                           uint32_t *pStartFlowIdx,
                                           uint32_t *pFlowIdxOffset);

void CpswAppUtils_addHostPortEntry(Cpsw_Handle hCpsw,
                                   uint32_t coreId,
                                   uint8_t *macAddr);

void CpswAppUtils_openRxFlow(Cpsw_Handle hCpsw,
                             uint32_t coreKey,
                             uint32_t coreId,
                             bool useDefaultFlow,
                             uint32_t *pRxFlowStartIdx,
                             uint32_t *pRxFlowIdx,
                             uint8_t *macAddr,
                             CpswDma_RxFlowHandle *pRxFlowHandle,
                             CpswDma_OpenRxFlowPrms *pRxFlowPrms);

void CpswAppUtils_closeRxFlow(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              bool useDefaultFlow,
                              CpswDma_PktInfoQ *pFqPktInfoQ,
                              CpswDma_PktInfoQ *pCqPktInfoQ,
                              uint32_t rxFlowStartIdx,
                              uint32_t rxFlowIdx,
                              uint8_t *macAddr,
                              CpswDma_RxFlowHandle hRxFlow);

void CpswAppUtils_openTxCh(Cpsw_Handle hCpsw,
                           uint32_t coreKey,
                           uint32_t coreId,
                           uint32_t *pTxChNum,
                           CpswDma_TxChHandle *pTxChHandle,
                           CpswDma_OpenTxChPrms *pCpswTxChCfg);

void CpswAppUtils_closeTxCh(Cpsw_Handle hCpsw,
                            uint32_t coreKey,
                            uint32_t coreId,
                            CpswDma_PktInfoQ *pFqPktInfoQ,
                            CpswDma_PktInfoQ *pCqPktInfoQ,
                            CpswDma_TxChHandle hTxChHandle,
                            uint32_t txChNum);


/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

static inline void CpswAppUtils_assertLocal(bool condition,
                                            const char *str,
                                            const char *fileName,
                                            int32_t lineNum)
{
    volatile static bool gCpswAssertWaitInLoop = TRUE;

    if (!(condition))
    {
        CpswAppUtils_print("Assertion @ Line: %d in %s: %s : failed !!!\n",
                           lineNum, fileName, str);
        while (gCpswAssertWaitInLoop)
        {
        }
    }

    return;
}
#ifdef __cplusplus
}
#endif

#endif  /* CPSW_APPUTILS_H_ */
