/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_loopback.c
 *
 * \brief    This file contains the loopback example implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#ifdef QNX_OS
#include <hw/inout.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#endif

#include <ti/osal/osal.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apprm.h>

#include "cpsw_loopback_baremetal.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define TEST_LEN                    (500U)

#define TEST_NUM_LOOP               (10U)
#define TEST_PTK_NUM                (1000U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

typedef struct
{
    /* CPSW instance type */
    Cpsw_Type cpswType;

    /* Core Id */
    uint32_t coreId;
    /*Core Key */
    uint32_t coreKey;

    /* MAC port number */
    Cpsw_MacPort portNum;

    /* CPSW driver handle */
    Cpsw_Handle hCpsw;
    Udma_DrvHandle hUdmaDrv;

    CpswDma_RxFlowHandle hRxFlow;
    CpswDma_PktInfoQ rxFreeQ;
    CpswDma_PktInfoQ rxReadyQ;

    CpswDma_TxChHandle hTxCh;
    CpswDma_PktInfoQ txFreePktInfoQ;

    uint8_t hostMacAddr[ETH_MAC_ADDR_LEN];

    /* Whether to test PHY loopback or MAC loopback */
    bool testPhyLoopback;

    /* Whether to print received Ethernet frames */
    bool printFrame;

    uint32_t rxStartFlowIdx;

    uint32_t rxFlowIdx;

    uint32_t txChNum;
} CpswApp_Obj;

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

int32_t  CpswApp_openCpsw(void);

void     CpswApp_closeCpsw(void);

void     CpswApp_showStats(void);

int32_t  CpswApp_openDma(void);

void     CpswApp_closeDma(void);

int32_t  CpswApp_pktRxTx(void);

uint32_t CpswApp_retrieveFreeTxPkts(void);

uint32_t CpswApp_receivePkts(void);

void     CpswApp_changeHostAleEntry(uint8_t macAddr[]);

#ifdef QNX_OS
/* For linker errors */
uint64_t TTBR3_BASE_ADDR, TTBR2_BASE_ADDR, TTBR1_BASE_ADDR;
#endif

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/* Broadcast address */
static uint8_t bcastAddr[ETH_MAC_ADDR_LEN] =
{0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU};

/* CPSW configuration */
static CpswApp_Obj gCpswLpbkAppObj =
{
    .testPhyLoopback = false,
    .printFrame      = false,
};

static volatile bool txSem = false, rxSem = false;

#define QNX_OS_MEM_TRACKING
#ifdef QNX_OS_MEM_TRACKING
typedef struct
{
    uint64_t paddr;
    uint64_t vaddr;
} virt_phys_table;

#define MAX_TABLE_ENTRY         4096
static virt_phys_table g_virt_phys_table[MAX_TABLE_ENTRY];
int g_virt_phys_table_index = 0;
#endif

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswApp_loopbackTestInitCfg(void)
{
    CpswAppUtils_getTestConfig(&gCpswLpbkAppObj.cpswType, &gCpswLpbkAppObj.portNum);
}

int32_t CpswApp_loopbackTest(CpswApp_ClkHandle clkhandle,
                             uint32_t iteration)
{
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    bool alive;
    int8_t i;
    int32_t status;

    CpswAppUtils_enableClocks(gCpswLpbkAppObj.cpswType,
                              MAC_CONN_TYPE_RGMII_FORCE_1000_FULL);

    CpswAppUtils_print("=================================\n");
    CpswAppUtils_print("Cpsw Loopback app: Iteration %d\n", iteration);
    CpswAppUtils_print("=================================\n");

    gCpswLpbkAppObj.coreId = CpswAppSoc_getCoreId();

    /* Open the CPSW */
    status = CpswApp_openCpsw();
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to open CPSW: %d\n", status);
    }

    /* Open CPSW DMA driver */
    if (status == CPSW_SOK)
    {
        status = CpswApp_openDma();
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to open CPSW DMA: %d\n", status);
        }
    }

    /* Enable host port */
    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
        prms.outArgs = NULL;
        status       = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                  gCpswLpbkAppObj.coreId,
                                  CPSW_ALE_IOCTL_SET_PORT_STATE,
                                  &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print(
                               "CpswApp_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
        }

        if (status == CPSW_SOK)
        {
            CPSW_IOCTL_SET_NO_ARGS(&prms);
            status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreId,
                                CPSW_HOSTPORT_IOCTL_ENABLE,
                                &prms);
            if (status != CPSW_SOK)
            {
                CpswAppUtils_print("Failed to enable host port: %d\n", status);
            }
        }
    }

    /* Show alive PHYs */
    if (status == CPSW_SOK)
    {
        for (i = 0U; i < CPSW_MDIO_MAX_PHY_CNT; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreId,
                                CPSW_MDIO_IOCTL_GET_ALIVE_STATUS,
                                &prms);
            if (status == CPSW_SOK)
            {
                if (alive == true)
                {
                    CpswAppUtils_print("PHY %d is alive\n", i);
                }
            }
            else
            {
                CpswAppUtils_print("Failed to get PHY %d alive status: %d\n", i,
                                   status);
            }
        }
    }

    /* Start timer */
    CpswApp_startClock(clkhandle);

    if (status == CPSW_SOK)
    {
        if (gCpswLpbkAppObj.testPhyLoopback)
        {
            while (!CpswAppUtils_isPortLinkUp(gCpswLpbkAppObj.hCpsw,
                                              gCpswLpbkAppObj.coreId,
                                              gCpswLpbkAppObj.portNum))
            {
                CpswAppUtils_wait(10);
            }
        }

        status = CpswApp_pktRxTx();
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to enable host port: %d\n", status);
        }
    }

    /* Print CPSW statistics of all ports */
    if (status == CPSW_SOK)
    {
        CpswApp_showStats();
    }

    /* Disable host port */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_HOSTPORT_IOCTL_DISABLE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to disable host port: %d\n", status);
        }
    }

    CpswApp_stopClock(clkhandle);

    if (status == CPSW_SOK)
    {
        /* Print RX flow DMA statistics */
        CpswAppUtils_showRxFlowStats(gCpswLpbkAppObj.hRxFlow);
        /* Print TX channel DMA statistics */
        CpswAppUtils_showTxChStats(gCpswLpbkAppObj.hTxCh);
        /* Close CPSW DMA driver */
        CpswApp_closeDma();
        /* Close CPSW */
        CpswApp_closeCpsw();
        /* Close UDMA driver */
        CpswAppUtils_udmaclose(gCpswLpbkAppObj.hUdmaDrv);
    }

    CpswAppUtils_disableClocks(gCpswLpbkAppObj.cpswType);

    return status;
}

static uint64_t CpswApp_virtToPhyFxn(const void *virtAddr,
                                     void       *appData)
{
#ifdef QNX_OS
    int ret;
    off64_t    phyAddr;
    uint32_t   length;

    CpswAppUtils_assert(virtAddr != NULL);

    if(appData != NULL_PTR) {
        length = (uint32_t) *((uint32_t *) appData);
    }
    else {
#ifdef QNX_OS_MEM_TRACKING
        length = 4;
#else
        length = 8192; // Default to 8K
#endif
    }

    /* Get destination physical address */
    ret = mem_offset64(virtAddr, NOFD, length, &phyAddr, NULL);
    if (ret) {
        CpswAppUtils_print("%s:Error from mem_offset\n",__FUNCTION__);
    }
    //QNX_DEBUG_PRINT("%s: 0x%x -> 0x%x \n",__FUNCTION__, virtAddr, phyAddr);
#ifdef QNX_OS_MEM_TRACKING
    int i, found = 0;

    for(i = 0; i < g_virt_phys_table_index; i++)
    {
        if(g_virt_phys_table[i].paddr == phyAddr)
        {
            found = 1;
            break;
        }
    }
    if (!found)
    {
        g_virt_phys_table[g_virt_phys_table_index].paddr = phyAddr;
        g_virt_phys_table[g_virt_phys_table_index].vaddr = (uint64_t)virtAddr;
        g_virt_phys_table_index++;
        if (g_virt_phys_table_index >= MAX_TABLE_ENTRY) {
            CpswAppUtils_print("%s: Error... virt_phys_tablee overflow \n",__FUNCTION__);
            delay(100);
        }
    }
#endif
    return (uint64_t ) phyAddr;
#else
    return ((uint64_t) virtAddr);
#endif
}

static void *CpswApp_phyToVirtFxn(uint64_t phyAddr,
                                  void    *appData)
{
#ifdef QNX_OS
    uint64_t *temp = 0;
    uint32_t length = 0;

    CpswAppUtils_assert(phyAddr != 0);

#ifdef QNX_OS_MEM_TRACKING
    uint32_t i = 0;

    for(i = 0; i < g_virt_phys_table_index; i++)
    {
        if(g_virt_phys_table[i].paddr == phyAddr)
        {
            return (void *) g_virt_phys_table[i].vaddr;
        }
    }
#endif

    if(appData != NULL_PTR) {
        length = (uint32_t) *((uint32_t *) appData);
    }
    else {
        length = 8192; // Default to 8K
    }

    temp  = mmap_device_memory(0, length, PROT_READ|PROT_WRITE, 0, phyAddr);
    if((temp == MAP_FAILED))
    {
        CpswAppUtils_print("%s: mmmap_device_memory failed\n",__FUNCTION__);
    }
    //QNX_DEBUG_PRINT("%s: 0x%x -> 0x%x size -> %d\n",__FUNCTION__, phyAddr, temp, length);
#else
#if defined (__aarch64__)
    uint64_t temp = phyAddr;
#else
    /* R5 is 32-bit machine, need to truncate to avoid void * typecast error */
    uint32_t temp = (uint32_t) phyAddr;
#endif
#endif
    return ((void *) temp);
}

void CpswApp_initAleConfig(CpswAle_Config *aleConfig)
{
    aleConfig->modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging           = TRUE;
    aleConfig->agingConfig.agingPeriodInMs           = 1000;
    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
}

int32_t CpswApp_openCpsw(void)
{
    Cpsw_Config cpswCfg;
    CpswOsal_Prms osalPrms;
    CpswUtils_Prms utilsPrms;
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    uint32_t coreId;
    Cpsw_AttachCoreOutArgs attachCoreOutArgs;

    /* Set configuration parameters */
    Cpsw_initParams(&cpswCfg);
    cpswCfg.vlanConfig.vlanAware          = false;
    cpswCfg.hostPortConfig.removeCrc      = true;
    cpswCfg.hostPortConfig.padShortPacket = true;
    cpswCfg.hostPortConfig.passCrcErrors  = true;
    CpswAppUtils_initResourceConfig(gCpswLpbkAppObj.cpswType,
                                    gCpswLpbkAppObj.coreId,
                                    &cpswCfg.resourceConfig);
    CpswApp_initAleConfig(&cpswCfg.aleConfig);

    if (gCpswLpbkAppObj.cpswType == CPSW_9G)
    {
        CpswAppUtils_print("CPSW_9G Test on MAIN NAVSS\n");
    }
    else if (gCpswLpbkAppObj.cpswType == CPSW_2G)
    {
        CpswAppUtils_print("CPSW_2G Test on MCU NAVSS\n");
    }

    cpswCfg.dmaConfig.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;

    /* App should open UDMA first as UDMA handle is needed to initialize
     * CPSW RX channel */
    gCpswLpbkAppObj.hUdmaDrv = CpswAppUtils_udmaOpen(gCpswLpbkAppObj.cpswType, NULL);
    CpswAppUtils_assert(NULL != gCpswLpbkAppObj.hUdmaDrv);

    cpswCfg.dmaConfig.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;

    /* Initialize CPSW driver with default OSAL, utils */
#ifdef QNX_OS
    utilsPrms.printFxn     = printf;
    utilsPrms.traceFxn     = printf;
    utilsPrms.phyToVirtFxn = &CpswApp_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswApp_virtToPhyFxn;

#else
    utilsPrms.printFxn     = UART_printf;
    utilsPrms.traceFxn     = UART_printf;
    utilsPrms.phyToVirtFxn = &CpswAppUtils_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswAppUtils_virtToPhyFxn;
#endif

    Cpsw_initOsalPrms(&osalPrms);

    Cpsw_init(gCpswLpbkAppObj.cpswType, &osalPrms, &utilsPrms);

    /* Open the CPSW driver */
    gCpswLpbkAppObj.hCpsw = Cpsw_open(gCpswLpbkAppObj.cpswType, &cpswCfg);
    if (gCpswLpbkAppObj.hCpsw == NULL)
    {
        CpswAppUtils_print("CpswApp_openCpsw() failed to open: %d\n", status);
        status = CPSW_EFAIL;
    }

    if (status == CPSW_SOK)
    {
        /* Attach the core with RM */
        coreId = gCpswLpbkAppObj.coreId;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreId, &attachCoreOutArgs);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_IOCTL_ATTACH_CORE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_loopbackTest failed CPSW_IOCTL_ATTACH_CORE: %d\n", status);
        }
        else
        {
            gCpswLpbkAppObj.coreKey = attachCoreOutArgs.coreKey;
        }
    }

    if (status == CPSW_SOK)
    {
        /* memutils open should happen after Cpsw is opened as it uses CpswUtils_Q
         * functions */
        status = CpswAppMemUtils_init();
        CpswAppUtils_assert(CPSW_SOK == status);
    }

    if (status == CPSW_SOK)
    {
        Cpsw_OpenPortLinkInArgs linkArgs =
        {
            .portNum = gCpswLpbkAppObj.portNum,
        };
        CpswMacPort_Config *macConfig      = &linkArgs.macConfig;
        CpswMacPort_LinkConfig *linkConfig = &linkArgs.linkConfig;
        CpswPhy_Config *phyConfig          = &linkArgs.phyConfig;
        CpswMacPort_Interface *interface   = &linkArgs.interface;

        Cpsw_initMacPortParams(macConfig);

        if (gCpswLpbkAppObj.testPhyLoopback)
        {
            CpswAppBoardUtils_setPhyConfig(gCpswLpbkAppObj.cpswType,
                                           linkArgs.portNum,
                                           interface,
                                           phyConfig);
            macConfig->enableLoopback = false;
            linkConfig->speed         = CPSW_SPEED_AUTO;
            linkConfig->duplexity     = CPSW_DUPLEX_AUTO;
        }
        else
        {
            CpswAppUtils_setNoPhyCfgRgmii(interface, phyConfig);
            macConfig->enableLoopback = true;
            linkConfig->speed         = CPSW_SPEED_1GBIT;
            linkConfig->duplexity     = CPSW_DUPLEX_FULL;
        }

        CPSW_IOCTL_SET_IN_ARGS(&prms, &linkArgs);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_IOCTL_OPEN_PORT_LINK,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_openCpsw() failed to open MAC port: %d\n",
                               status);
        }
    }

    return status;
}

void CpswApp_closeCpsw(void)
{
    Cpsw_IoctlPrms prms;
    uint32_t coreKey;
    Cpsw_GenericPortLinkInArgs args =
    {
        .portNum = gCpswLpbkAppObj.portNum,
    };
    int32_t status;

    coreKey = gCpswLpbkAppObj.coreKey;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &coreKey);

    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_IOCTL_DETACH_CORE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_closeCpsw() failed CPSW_IOCTL_DETACH_CORE: %d\n", status);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &args);
    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_IOCTL_CLOSE_PORT_LINK,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("close() failed to close MAC port: %d\n", status);
    }

    Cpsw_close(gCpswLpbkAppObj.hCpsw);

    Cpsw_deinit(gCpswLpbkAppObj.cpswType);
}

void CpswApp_showStats(void)
{
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status = CPSW_SOK;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_STATS_IOCTL_GET_HOSTPORT_STATS,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("\n Port 0 Statistics\n");
        CpswAppUtils_print("-----------------------------------------\n");
        CpswAppUtils_printHostPortStats2G((CpswStats_HostPort_2g *)&portStats);
        CpswAppUtils_print("\n");
    }
    else
    {
        CpswAppUtils_print("CpswApp_showStats() failed to get host stats: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        inArgs.portNum = gCpswLpbkAppObj.portNum;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswAppUtils_print("\n Port 1 Statistics\n");
            CpswAppUtils_print("-----------------------------------------\n");
            CpswAppUtils_printMacPortStats2G((CpswStats_MacPort_2g *)&portStats);
            CpswAppUtils_print("\n");
        }
        else
        {
            CpswAppUtils_print("CpswApp_showStats() failed to get MAC stats: %d\n",
                               status);
        }
    }
}

void CpswApp_rxIsrFxn(void *appData)
{
    rxSem = true;
}

void CpswApp_txIsrFxn(void *appData)
{
    txSem = true;
}

void CpswApp_initTxFreePktQ(void)
{
    CpswDma_PktInfo *pPktInfo;
    uint32_t i;

    /* Initialize all queues */
    CpswUtils_initQ(&gCpswLpbkAppObj.txFreePktInfoQ);

    /* Initialize TX EthPkts and queue them to txFreePktInfoQ */
    for (i = 0U; i < CPSW_APPMEMUTILS_NUM_TX_PKTS; i++)
    {
        pPktInfo = CpswAppMemUtils_allocEthPktFxn(&gCpswLpbkAppObj,
                                                  CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE,
                                                  UDMA_CACHELINE_ALIGNMENT);
        CpswAppUtils_assert(pPktInfo != NULL);
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);

        CpswUtils_enQ(&gCpswLpbkAppObj.txFreePktInfoQ, &pPktInfo->node);
    }

    CpswAppUtils_print("initQs() txFreePktInfoQ initialized with %d pkts\n",
                       CpswUtils_getQCount(&gCpswLpbkAppObj.txFreePktInfoQ));
}

void CpswApp_initRxReadyPktQ(void)
{
    CpswDma_PktInfoQ rxReadyQ;
    CpswDma_PktInfo *pPktInfo;
    int32_t status;
    uint32_t i;

    CpswUtils_initQ(&gCpswLpbkAppObj.rxFreeQ);
    CpswUtils_initQ(&gCpswLpbkAppObj.rxReadyQ);
    CpswUtils_initQ(&rxReadyQ);

    for (i = 0U; i < CPSW_APPMEMUTILS_NUM_RX_PKTS; i++)
    {
        pPktInfo = CpswAppMemUtils_allocEthPktFxn(&gCpswLpbkAppObj,
                                                  CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE,
                                                  UDMA_CACHELINE_ALIGNMENT);
        CpswAppUtils_assert(pPktInfo != NULL);
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);
        CpswUtils_enQ(&gCpswLpbkAppObj.rxFreeQ, &pPktInfo->node);
    }

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswLpbkAppObj.hRxFlow, &rxReadyQ);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* There should not be any packet with DMA during init */
    CpswAppUtils_assert(CpswUtils_getQCount(&rxReadyQ) == 0U);

    CpswAppUtils_validatePacketState(&gCpswLpbkAppObj.rxFreeQ,
                                     CPSW_PKTSTATE_APP_WITH_FREEQ,
                                     CPSW_PKTSTATE_APP_WITH_DRIVER);

    CpswAppUtils_submitRxPackets(gCpswLpbkAppObj.hRxFlow,
                                 &gCpswLpbkAppObj.rxFreeQ);

    /* Assert here as during init no. of DMA descriptors should be equal to
     * no. of free Ethernet buffers available with app */

    CpswAppUtils_assert(0U == CpswUtils_getQCount(&gCpswLpbkAppObj.rxFreeQ));
}

int32_t CpswApp_openDma(void)
{
    int32_t status = CPSW_SOK;
    CpswDma_OpenRxFlowPrms cpswRxFlowCfg;
    CpswDma_OpenTxChPrms cpswTxChCfg;

    /* Open the CPSW TX channel  */
    if (status == CPSW_SOK)
    {
        CpswDma_initTxChParams(&cpswTxChCfg);

        cpswTxChCfg.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;
        cpswTxChCfg.hCbArg   = &gCpswLpbkAppObj;
        cpswTxChCfg.notifyCb = CpswApp_txIsrFxn;

        CpswAppUtils_setCommonTxChPrms(&cpswTxChCfg);

        CpswAppUtils_openTxCh(gCpswLpbkAppObj.hCpsw,
                              gCpswLpbkAppObj.coreKey,
                              gCpswLpbkAppObj.coreId,
                              &gCpswLpbkAppObj.txChNum,
                              &gCpswLpbkAppObj.hTxCh,
                              &cpswTxChCfg);

        CpswApp_initTxFreePktQ();

        if (NULL != gCpswLpbkAppObj.hTxCh)
        {
            status = CpswDma_enableTxEvent(gCpswLpbkAppObj.hTxCh);
            if (CPSW_SOK != status)
            {
                /* Free the Ch Num if enable event failed */
                CpswAppUtils_freeTxCh(gCpswLpbkAppObj.hCpsw,
                                      gCpswLpbkAppObj.coreKey,
                                      gCpswLpbkAppObj.coreId,
                                      gCpswLpbkAppObj.txChNum);
                CpswAppUtils_print("CpswDma_startTxCh() failed: %d\n", status);
                status = CPSW_EFAIL;
            }
        }
        else
        {
            /* Free the Ch Num if open Tx Ch failed */
            CpswAppUtils_freeTxCh(gCpswLpbkAppObj.hCpsw,
                                  gCpswLpbkAppObj.coreKey,
                                  gCpswLpbkAppObj.coreId,
                                  gCpswLpbkAppObj.txChNum);
            CpswAppUtils_print("CpswDma_openTxCh() failed to open: %d\n",
                               status);
            status = CPSW_EFAIL;
        }
    }

    /* Open the CPSW RX flow  */
    if (status == CPSW_SOK)
    {
        CpswDma_initRxFlowParams(&cpswRxFlowCfg);
        cpswRxFlowCfg.notifyCb = CpswApp_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;
        cpswRxFlowCfg.hCbArg   = &gCpswLpbkAppObj;

        CpswAppUtils_setCommonRxFlowPrms(&cpswRxFlowCfg);

        CpswAppUtils_openRxFlow(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreKey,
                                gCpswLpbkAppObj.coreId,
                                true,
                                &gCpswLpbkAppObj.rxStartFlowIdx,
                                &gCpswLpbkAppObj.rxFlowIdx,
                                &gCpswLpbkAppObj.hostMacAddr[0U],
                                &gCpswLpbkAppObj.hRxFlow,
                                &cpswRxFlowCfg);

        if (NULL == gCpswLpbkAppObj.hRxFlow)
        {
            CpswAppUtils_freeRxFlow(gCpswLpbkAppObj.hCpsw,
                                    gCpswLpbkAppObj.coreKey,
                                    gCpswLpbkAppObj.coreId,
                                    gCpswLpbkAppObj.rxFlowIdx);
            CpswAppUtils_print("CpswDma_openRxFlow() failed to open: %d\n",
                               status);
            CpswAppUtils_assert(NULL != gCpswLpbkAppObj.hRxFlow);
        }
        else
        {
            CpswAppUtils_print("Host MAC address: ");
            CpswAppUtils_printMacAddr(gCpswLpbkAppObj.hostMacAddr);
            /* For MAC loopback disable secure flag for the host port entry */
            CpswApp_changeHostAleEntry(&gCpswLpbkAppObj.hostMacAddr[0U]);

            /* Submit all ready RX buffers to DMA.*/
            CpswApp_initRxReadyPktQ();
        }
    }

    return status;
}

void CpswApp_closeDma(void)
{
    CpswDma_PktInfoQ fqPktInfoQ;
    CpswDma_PktInfoQ cqPktInfoQ;

    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

    /* There should not be any ready packet */
    CpswAppUtils_assert(0U == CpswUtils_getQCount(&gCpswLpbkAppObj.rxReadyQ));

    /* Close RX Flow */
    CpswAppUtils_closeRxFlow(gCpswLpbkAppObj.hCpsw,
                             gCpswLpbkAppObj.coreKey,
                             gCpswLpbkAppObj.coreId,
                             true,
                             &fqPktInfoQ,
                             &cqPktInfoQ,
                             gCpswLpbkAppObj.rxStartFlowIdx,
                             gCpswLpbkAppObj.rxFlowIdx,
                             gCpswLpbkAppObj.hostMacAddr,
                             gCpswLpbkAppObj.hRxFlow);
    CpswAppUtils_freePktInfoQ(&fqPktInfoQ);
    CpswAppUtils_freePktInfoQ(&cqPktInfoQ);

    /* Close TX channel */
    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

    CpswAppUtils_closeTxCh(gCpswLpbkAppObj.hCpsw,
                           gCpswLpbkAppObj.coreKey,
                           gCpswLpbkAppObj.coreId,
                           &fqPktInfoQ,
                           &cqPktInfoQ,
                           gCpswLpbkAppObj.hTxCh,
                           gCpswLpbkAppObj.txChNum);
    CpswAppUtils_freePktInfoQ(&fqPktInfoQ);
    CpswAppUtils_freePktInfoQ(&cqPktInfoQ);

    CpswAppUtils_freePktInfoQ(&gCpswLpbkAppObj.rxFreeQ);
    CpswAppUtils_freePktInfoQ(&gCpswLpbkAppObj.txFreePktInfoQ);

    CpswAppMemUtils_deInit();
}

void CpswApp_changeHostAleEntry(uint8_t macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.trunk   = false;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print(
                           "CpswApp_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
    }
}

int32_t CpswApp_pktRxTx(void)
{
    CpswDma_PktInfoQ txSubmitQ;
    CpswDma_PktInfo *pktInfo;
    EthFrame *frame;
    uint32_t txRetrievePktCnt, loopTxPktCnt, totalTxCnt;
    uint32_t rxReadyCnt;
    uint32_t loopCnt, pktCnt, loopRxPktCnt;
    int32_t status = CPSW_SOK;

    totalTxCnt = 0U;
    for (loopCnt = 0U; loopCnt < TEST_NUM_LOOP; loopCnt++)
    {
        pktCnt = 0U;
        while (pktCnt < TEST_PTK_NUM)
        {
            loopRxPktCnt = loopTxPktCnt = 0U;
            /* Transmit a single packet */
            CpswUtils_initQ(&txSubmitQ);

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.txFreePktInfoQ);

            while (NULL != pktInfo)
            {
                pktCnt++;
                /* Fill the TX Eth frame with test content */
                frame = (EthFrame *)pktInfo->bufPtr;
                memcpy(frame->hdr.dstMac, bcastAddr, ETH_MAC_ADDR_LEN);
                memcpy(frame->hdr.srcMac, &gCpswLpbkAppObj.hostMacAddr[0U], ETH_MAC_ADDR_LEN);
                frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);
                memset(&frame->payload[0U], (uint8_t)(0xA5 + pktCnt), TEST_LEN);
                pktInfo->userBufLen = TEST_LEN + sizeof(EthFrameHeader);
                pktInfo->appPriv    = &gCpswLpbkAppObj;
                CpswUtils_checkPktState(&pktInfo->pktState,
                                        CPSW_PKTSTATE_MODULE_APP,
                                        CPSW_PKTSTATE_APP_WITH_FREEQ,
                                        CPSW_PKTSTATE_APP_WITH_DRIVER);

                /* Enqueue the packet for later transmission */
                CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

                if (pktCnt >= TEST_PTK_NUM)
                {
                    break;
                }

                /* Dequeue one free TX Eth packet */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.txFreePktInfoQ);
            }

            loopTxPktCnt = CpswUtils_getQCount(&txSubmitQ);
            while (0U != CpswUtils_getQCount(&txSubmitQ))
            {
                uint32_t txCnt = CpswUtils_getQCount(&txSubmitQ);
                status = CpswAppUtils_submitTxPackets(gCpswLpbkAppObj.hTxCh,
                                                      &txSubmitQ);
                while (txSem != true)
                {
                    ;
                }

                txSem = false;

                /* Retrieve TX free packets */
                if (status == CPSW_SOK)
                {
                    txCnt            = txCnt - CpswUtils_getQCount(&txSubmitQ);
                    txRetrievePktCnt = 0U;
                    while (txRetrievePktCnt != txCnt)
                    {
                        /* TODO this is not failure as HW is busy sending packets, we
                         * need to wait and again call retrieve packets */
                        CpswAppUtils_wait(1);
                        txRetrievePktCnt += CpswApp_retrieveFreeTxPkts();
                    }
                }
                else
                {
                    break;
                }
            }

            /* Wait for packet reception */
            do
            {
                while (rxSem != true)
                {
                    ;
                }

                rxSem = false;

                /* Get the packets received so far */
                rxReadyCnt = CpswApp_receivePkts();
                if (rxReadyCnt > 0U)
                {
                    /* Consume the received packets and release them */
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(
                                                               &gCpswLpbkAppObj.rxReadyQ);
                    while (NULL != pktInfo)
                    {
                        CpswUtils_checkPktState(&pktInfo->pktState,
                                                CPSW_PKTSTATE_MODULE_APP,
                                                CPSW_PKTSTATE_APP_WITH_READYQ,
                                                CPSW_PKTSTATE_APP_WITH_FREEQ);

                        frame = (EthFrame *)pktInfo->bufPtr;

                        /* Consume the packet by just printing its content */
                        if (gCpswLpbkAppObj.printFrame)
                        {
                            CpswAppUtils_printFrame(frame,
                                                    pktInfo->userBufLen - sizeof(EthFrameHeader));
                        }

                        /* Release the received packet */
                        CpswUtils_enQ(&gCpswLpbkAppObj.rxFreeQ, &pktInfo->node);
                        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(
                                                                   &gCpswLpbkAppObj.rxReadyQ);
                    }

                    /*Submit now processed buffers */
                    if (status == CPSW_SOK)
                    {
                        CpswAppUtils_validatePacketState(&gCpswLpbkAppObj.rxFreeQ,
                                                         CPSW_PKTSTATE_APP_WITH_FREEQ,
                                                         CPSW_PKTSTATE_APP_WITH_DRIVER);

                        CpswAppUtils_submitRxPackets(gCpswLpbkAppObj.hRxFlow,
                                                     &gCpswLpbkAppObj.rxFreeQ);
                    }
                }

                loopRxPktCnt += rxReadyCnt;
            }
            while (loopRxPktCnt < loopTxPktCnt);

            totalTxCnt += loopTxPktCnt;

            CpswAppUtils_print("Transmitted & Received %d packets\n", totalTxCnt);
        }
    }

    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("Transmitted & Received %d packets\n", totalTxCnt);
    }
    else
    {
        CpswAppUtils_print("Failed to transmit/receive packets: %d, transmitted: %d \n", TEST_PTK_NUM, totalTxCnt);
    }

    return status;
}

uint32_t CpswApp_retrieveFreeTxPkts(void)
{
    CpswDma_PktInfoQ txFreeQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t txFreeQCnt = 0U;

    CpswUtils_initQ(&txFreeQ);

    /* Retrieve any CPSW packets that may be free now */
    status = CpswDma_retrieveTxDonePackets(gCpswLpbkAppObj.hTxCh, &txFreeQ);
    if (status == CPSW_SOK)
    {
        txFreeQCnt = CpswUtils_getQCount(&txFreeQ);

        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        while (NULL != pktInfo)
        {
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_FREEQ);

            CpswUtils_enQ(&gCpswLpbkAppObj.txFreePktInfoQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        }
    }
    else
    {
        CpswAppUtils_print("retrieveFreeTxPkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return txFreeQCnt;
}

uint32_t CpswApp_receivePkts(void)
{
    CpswDma_PktInfoQ rxReadyQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t rxReadyCnt = 0U;

    CpswUtils_initQ(&rxReadyQ);

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswLpbkAppObj.hRxFlow, &rxReadyQ);
    if (status == CPSW_SOK)
    {
        rxReadyCnt = CpswUtils_getQCount(&rxReadyQ);

        /* Queue the received packet to rxReadyQ and pass new ones from rxFreeQ
        **/
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxReadyQ);
        while (pktInfo != NULL)
        {
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_READYQ);

            CpswUtils_enQ(&gCpswLpbkAppObj.rxReadyQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxReadyQ);
        }
    }
    else
    {
        CpswAppUtils_print("receivePkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return rxReadyCnt;
}

Cpsw_Handle CpswApp_getCpswHandle(void)
{
    return gCpswLpbkAppObj.hCpsw;
}

/* end of file */
