/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     main_baremetal.c
 *
 * \brief    Main file for baremetal build.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#ifdef QNX_OS
#include <hw/inout.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#endif

#include <ti/drv/sciclient/sciclient.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/osal/TimerP.h>

#include "cpsw_loopback_baremetal.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#ifdef QNX_OS
#define NUM_LOOPBACK_ITERATION                (1)
#else
#define NUM_LOOPBACK_ITERATION                (5)
#endif


/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static void CpswApp_timerISR(uintptr_t arg);

static CpswApp_ClkHandle CpswApp_createClock(void);

extern Cpsw_Handle CpswApp_getCpswHandle(void);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int main(void)
{
    uint8_t i;
    int32_t status;
    CpswApp_ClkHandle clkhandle = CpswApp_createClock();
#ifndef QNX_OS
    if (clkhandle != NULL)
#endif
    {
        /* This should not be removed, as this is fix for linker issue in release mode.
         * If removed linker will complain "undefined reference to `_sbrk'".
         * This arises whenever library invokes printf() but the app doesn't.
         */
#ifndef UART_ENABLED
        printf("CPSW Bare metal Loopback App \n");
#endif

#ifdef QNX_OS
        /* Get IO priveleges */
        if (ThreadCtl(_NTO_TCTL_IO, NULL) == -1) {
            perror("ThreadCtl(_NTO_TCTL_IO");
            return 1;
        }
#else
        CpswAppBoardUtils_init();
#endif

        CpswApp_loopbackTestInitCfg();

        for (i = 0; i < NUM_LOOPBACK_ITERATION; i++)
        {
            /* Run the loopback test */
            status = CpswApp_loopbackTest(clkhandle, i);
        }

        if (status == CPSW_SOK)
        {
            CpswAppUtils_print("Loopback application completed\n");
        }
        else
        {
            CpswAppUtils_print("Loopback application failed to complete\n");
        }

#ifdef QNX_OS
#else
        CpswAppBoardUtils_deInit();
#endif
        TimerP_delete(clkhandle);
    }
#ifndef QNX_OS
    else
    {
        CpswAppUtils_assert(false);
    }
#endif

    return 0;
}

static CpswApp_ClkHandle CpswApp_createClock(void)
{
    TimerP_Handle timerHandle;
    TimerP_Params timerParams;
    uint32_t period = 1000 * 100; /* usecs */

    TimerP_Params_init(&timerParams);

    timerParams.startMode  = TimerP_StartMode_USER;
    timerParams.periodType = TimerP_PeriodType_MICROSECS;
    timerParams.period     = period;

    timerHandle = TimerP_create(TimerP_ANY,
                                (TimerP_Fxn) & CpswApp_timerISR,
                                &timerParams);

    return (CpswApp_ClkHandle)timerHandle;
}

void CpswApp_startClock(CpswApp_ClkHandle handle)
{
    TimerP_Handle timerHandle = (TimerP_Handle)handle;

    /* start the timer */
    TimerP_start(timerHandle);
}

void CpswApp_stopClock(CpswApp_ClkHandle handle)
{
    TimerP_Handle timerHandle = (TimerP_Handle)handle;

    TimerP_stop(timerHandle);
}

static void CpswApp_timerISR(uintptr_t arg)
{
    Cpsw_Handle hCpsw = CpswApp_getCpswHandle();

    Cpsw_periodicTick(hCpsw);
}

void CpswAppUtils_wait(uint32_t waitTime)
{
#ifdef QNX_OS
        delay(waitTime); // in ms
#else
    uint32_t i;

    /* we multiply waitTime by 10000 as 400MHz R5 takes 2.5ns for single cycle
     * and we assume for loop takes 4 cycles */
    for (i = 0; i < (waitTime * 10000); i++)
    {
        ;
    }
#endif
}
