/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_config_server.c
 *
 *  \brief CPSW configuration server
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <inttypes.h>

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/utils/Load.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Mailbox.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_networkutils.h>

#include <ti/drv/cpsw/nimucpsw/ndk2cpsw_appif.h>

/* API Headers */
#include "priv/inc/jsmn.h"
#include "priv/inc/cpsw_serial.h"

#include "cpsw_cfgserver.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* Configuration server application stack size */
#define CPSWSERVERAPP_TSK_STACK_MAIN              (16U * 1024U)
#define CPSWSERVERAPP_SOCKETPORTNUM               (5555U)
/*Buffer size for config server*/
#define CPSW_CONFIG_SERVER_BUFFER_SIZE            (1024)
#define CPSW_CONFIG_SERVER_UART_CFG_BUFFER_SIZE   (1024U * 2U)
#define CPSW_CONFIG_SERVER_XMODEM_ACK             (0x06)
#define CPSWDISPATCHER_MBOX_MSG_COUNT             (10U)
#define CPSW_CONFIG_SERVER_JSMN_TOKEN_COUNT       (1000U)
#define CPSW_CONFIG_SERVER_CPU_LOAD_MAX_LEN       (16U)
#define UART_TRANSPORT_ENABLE                     (0U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct CpswDispatcher_MailboxObj_s
{
    char *recv;

    char *send;
} CpswDispatcher_MailboxObj;

typedef struct CpswCfgServer_AppObj_s
{
    /* CPSW Handle */
    Cpsw_Handle hCpsw;

    /* CPSW Type */
    Cpsw_Type cpswType;

    /* Core Id */
    uint32_t coreId;

    /* Dispatch Request Mailbox */
    Mailbox_Handle hMboxIn;

    /* Dispatch Response Mailbox */
    Mailbox_Handle hMboxOut;

    /* Task Handles  */
    Task_Handle taskNetwork;

    Task_Handle taskDispatcher;

    Task_Handle taskUart;

    jsmn_parser parser;
} CpswCfgServer_AppObj;
/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* Task functions */
static void    CpswCfgServer_networkTask(UArg arg0,
                                         UArg arg1);

static void    CpswCfgServer_dispatcherTask(UArg arg0,
                                            UArg arg1);

#if UART_TRANSPORT_ENABLE
static void    CpswCfgServer_uartTask(UArg arg0,
                                      UArg arg1);

#endif

/* IOCTL Functions */
static void    CpswCfgServer_enabledisableMcast(char *recvBuff,
                                                char *sendBuff);

static void    CpswCfgServer_addVlanEntry(char *recvBuff,
                                          char *sendBuff);

static void    CpswCfgServer_getMacPortStats(char *recvBuff,
                                             char *sendBuff);

static void    CpswCfgServer_getHostPortStats(char *recvBuff,
                                              char *sendBuff);

static void    CpswCfgServer_lookupUnicast(char *recvBuff,
                                           char *sendBuff);

static void    CpswCfgServer_getBcastMcastLimits(char *recvBuff,
                                                 char *sendBuff);

static void    CpswCfgServer_getVersion(char *recvBuff,
                                        char *sendBuff);

static void    CpswCfgServer_initFxnTable(void);

static void    CpswCfgServer_parseInterVlanRoutingCfg(char *recvBuff,
                                                      CpswCfgServer_InterVlanConfig *pinterVlanCfg);

static void    CpswCfgServer_addIpNxtHdrFilter(char *recvBuff,
                                               char *sendBuff);

static void    CpswCfgServer_enableRateLimiting(char *recvBuff,
                                                char *sendBuff);

static int32_t CpswCfgServer_setPolicerPriorityRateLimiting(CpswCfgServer_RateLimitingCfg *pRateLimitingCfg);

static int32_t CpswCfgServer_addAleEntryRateLimiting(CpswCfgServer_RateLimitingCfg *pRateLimitingCfg);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
CpswCfgServer_AppObj gCpswCfgServerObj;

/* IOCTL Function Table */
void(*cpswCfgServer_fxn_table[CPSW_CONFIG_SERVER_FXN_TBL_SIZE]) (char *recvBuff, char *sendBuff);

/* Objects required for init using UART*/
CpswSerial_ContextObj *gSerialContextObj;

/* Required for communication with clients*/
jsmntok_t token[CPSW_CONFIG_SERVER_JSMN_TOKEN_COUNT];
uint8_t gNwRecvBuff[CPSW_CONFIG_SERVER_BUFFER_SIZE];
uint8_t gNwSendBuff[CPSW_CONFIG_SERVER_BUFFER_SIZE];
uint8_t gUartRecvBuff[CPSW_CONFIG_SERVER_BUFFER_SIZE];
uint8_t gUartSendBuff[CPSW_CONFIG_SERVER_BUFFER_SIZE];

static uint8_t gCpswNetworkTskStackMain[CPSWSERVERAPP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
static uint8_t gCpswDispatcherTskStackMain[CPSWSERVERAPP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
#if UART_TRANSPORT_ENABLE

static uint8_t gCpswUartTskStackMain[CPSWSERVERAPP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
#endif

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
int32_t CpswCfgServer_init(Cpsw_Type cpswType)
{
    int32_t status = CPSW_SOK;
    Task_Params params;
    Mailbox_Params mboxParams;
    Error_Block eb;

    gCpswCfgServerObj.hCpsw    = Cpsw_getHandle(cpswType);
    gCpswCfgServerObj.coreId   = CpswAppSoc_getCoreId();
    gCpswCfgServerObj.cpswType = cpswType;

    if (gCpswCfgServerObj.hCpsw == NULL)
    {
        status = CPSW_EFAIL;
    }

    /* Create Dispatch Request Mailbox */
    if (status == CPSW_SOK)
    {
        Error_init(&eb);
        Mailbox_Params_init(&mboxParams);
        mboxParams.instance->name = "Dispatch Request MailBox";
        gCpswCfgServerObj.hMboxIn = Mailbox_create(sizeof(CpswDispatcher_MailboxObj),
                                                   CPSWDISPATCHER_MBOX_MSG_COUNT,
                                                   &mboxParams,
                                                   &eb);

        if (gCpswCfgServerObj.hMboxIn == NULL)
        {
            status = CPSW_EFAIL;
        }
    }

    if (status == CPSW_SOK)
    {
        /* Create Dispatch Response Mailbox */
        Error_init(&eb);
        Mailbox_Params_init(&mboxParams);
        mboxParams.instance->name  = "Dispatch Response MailBox";
        gCpswCfgServerObj.hMboxOut = Mailbox_create(sizeof(CpswDispatcher_MailboxObj),
                                                    CPSWDISPATCHER_MBOX_MSG_COUNT,
                                                    &mboxParams,
                                                    &eb);
        if (gCpswCfgServerObj.hMboxOut == NULL)
        {
            status = CPSW_EFAIL;
        }
    }

    if (status == CPSW_SOK)
    {
        /* Initialize dispatch function table */
        CpswCfgServer_initFxnTable();
    }

    if (status == CPSW_SOK)
    {
        /* Create Network Listener Task.  */
        Task_Params_init(&params);
        Error_init(&eb);
        params.priority               = 7U;
        params.stack                  = gCpswNetworkTskStackMain;
        params.stackSize              = sizeof(gCpswNetworkTskStackMain);
        params.instance->name         = "Network Task";
        gCpswCfgServerObj.taskNetwork = Task_create(CpswCfgServer_networkTask, &params, &eb);
        if (gCpswCfgServerObj.taskNetwork == NULL)
        {
            status = CPSW_EFAIL;
        }
    }

    if (status == CPSW_SOK)
    {
        /* Create Dispatcher Task.  */
        Task_Params_init(&params);
        Error_init(&eb);
        params.priority                  = 7U;
        params.stack                     = gCpswDispatcherTskStackMain;
        params.stackSize                 = sizeof(gCpswDispatcherTskStackMain);
        params.instance->name            = "Dispatcher Task";
        gCpswCfgServerObj.taskDispatcher = Task_create(CpswCfgServer_dispatcherTask, &params, &eb);
        if (gCpswCfgServerObj.taskDispatcher == NULL)
        {
            status = CPSW_EFAIL;
        }
    }

#if UART_TRANSPORT_ENABLE
    if (status == CPSW_SOK)
    {
        /* Create UART Listener task.  */
        Task_Params_init(&params);
        Error_init(&eb);
        params.priority            = 14U;
        params.stack               = gCpswUartTskStackMain;
        params.stackSize           = sizeof(gCpswUartTskStackMain);
        params.instance->name      = "UART Task";
        gCpswCfgServerObj.taskUart = Task_create(CpswCfgServer_uartTask, &params, &eb);
        if (gCpswCfgServerObj.taskUart == NULL)
        {
            status = CPSW_EFAIL;
        }
    }
#endif

    return status;
}

static void CpswCfgServer_getCpuLoad(char *recvBuff,
                                     char *sendBuff)
{
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];
    uint8_t cpu_load = 0U;

    cpu_load = Load_getCPULoad();
    snprintf(outString, (CPSW_CONFIG_SERVER_CPU_LOAD_MAX_LEN - 1U), "{\"load\":%d}", cpu_load);
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));
    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_initFxnTable(void)
{
    cpswCfgServer_fxn_table[0] = NULL;  /* Reserved for Dump ALE Table */
    cpswCfgServer_fxn_table[1] = &CpswCfgServer_getVersion;
    cpswCfgServer_fxn_table[2] = &CpswCfgServer_getBcastMcastLimits;
    cpswCfgServer_fxn_table[3] = &CpswCfgServer_lookupUnicast;
    cpswCfgServer_fxn_table[4] = &CpswCfgServer_getHostPortStats;
    cpswCfgServer_fxn_table[5] = &CpswCfgServer_getMacPortStats;
    cpswCfgServer_fxn_table[6] = &CpswCfgServer_addVlanEntry;
    cpswCfgServer_fxn_table[7] = &CpswCfgServer_getCpuLoad;
    cpswCfgServer_fxn_table[8] = &CpswCfgServer_enabledisableMcast;
    /* Cmd 9 and 10 are reserved for SW & HW Inter VLAN Routing */
    cpswCfgServer_fxn_table[11] = &CpswCfgServer_addIpNxtHdrFilter;
    cpswCfgServer_fxn_table[12] = &CpswCfgServer_enableRateLimiting;
}

void CpswCfgServer_getCpswCfg(Cpsw_Config *cpswCfg)
{
    CpswSerial_ConfigObj cfg;
    int32_t maxLen;
    int tokenCount, val, i = 0;
    char recvBuff[CPSW_CONFIG_SERVER_UART_CFG_BUFFER_SIZE];
    int32_t status = CpswSerial_invalid_read_len;

    cfg.magic                               = CPSW_SERIAL_MODULE_MAGIC;
    cfg.protocol                            = CPSW_SERIAL_PROTOCOL_XMODEM;
    cfg.protocolConfig.xmodem.maxErrorCount = 10U;
    cfg.protocolConfig.xmodem.ackTimeoutSec = 3U;
    cfg.protocolConfig.xmodem.charTimeoutMs = 2U;

    gSerialContextObj = (CpswSerial_ContextObj *)CpswSerial_init(&cfg);
    CpswSerial_open(gSerialContextObj);
    CpswSerial_start(gSerialContextObj);

    status = CpswSerial_read(gSerialContextObj, (uint8_t *)recvBuff, CPSW_UTILS_ARRAYSIZE(recvBuff));

    if (status == CpswSerial_success)
    {
        jsmn_init(&gCpswCfgServerObj.parser);
        tokenCount = jsmn_parse(&gCpswCfgServerObj.parser,
                                (char *)recvBuff,
                                CPSW_CONFIG_SERVER_UART_CFG_BUFFER_SIZE,
                                token,
                                CPSW_UTILS_ARRAYSIZE(token));

        for (i = 1U; i < tokenCount; i++)
        {
            char choice[CPSW_CONFIG_SERVER_TOKEN_SIZE];
            if (jsoneq((char *)recvBuff, &token[i], "vlanAware") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(choice, maxLen, "%s", recvBuff + token[i + 1].start);
                val                           = atoi(choice);
                cpswCfg->vlanConfig.vlanAware = val;
            }

            if (jsoneq((char *)recvBuff, &token[i], "removeCrc") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(choice, maxLen, "%s", recvBuff + token[i + 1].start);
                val                               = atoi(choice);
                cpswCfg->hostPortConfig.removeCrc = val;
            }

            if (jsoneq((char *)recvBuff, &token[i], "padShortPacket") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(choice, maxLen, "%s", recvBuff + token[i + 1].start);
                val                                    = atoi(choice);
                cpswCfg->hostPortConfig.padShortPacket = val;
            }

            if (jsoneq((char *)recvBuff, &token[i], "passCrcErrors") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(choice, maxLen, "%s", recvBuff + token[i + 1].start);
                val                                   = atoi(choice);
                cpswCfg->hostPortConfig.passCrcErrors = val;
            }
        }
    }
}

static void CpswCfgServer_networkTask(UArg arg0,
                                      UArg arg1)
{
    CpswDispatcher_MailboxObj msg;

    /* Network socket object for RX */
    static CpswNetworkUtils_SockObj netRxSockObj;

    netRxSockObj.port = CPSWSERVERAPP_SOCKETPORTNUM;
    CpswNetworkUtils_createSocket(&netRxSockObj);

    msg.recv = (char *)gNwRecvBuff;
    msg.send = (char *)gNwSendBuff;

    while (true)
    {
        if (CpswNetworkUtils_readPayload(&netRxSockObj, &gNwRecvBuff[0U], 1024) == 0U)
        {
            Mailbox_post(gCpswCfgServerObj.hMboxIn, &msg, BIOS_WAIT_FOREVER);
            Mailbox_pend(gCpswCfgServerObj.hMboxOut, &msg, BIOS_WAIT_FOREVER);
            CpswNetworkUtils_writePayload(&netRxSockObj, &gNwSendBuff[0U], 1024);
        }
        else
        {
            Task_sleep(10);
        }
    }
}

#if UART_TRANSPORT_ENABLE
static void CpswCfgServer_uartTask(UArg arg0,
                                   UArg arg1)
{
    CpswSerial_ConfigObj cfg;
    CpswDispatcher_MailboxObj msg;
    int32_t status = Serial_invalid_read_len;

    msg.recv = (char *)gUartRecvBuff;
    msg.send = (char *)gUartSendBuff;

    cfg.magic                               = CPSW_SERIAL_MODULE_MAGIC;
    cfg.protocol                            = CPSW_SERIAL_PROTOCOL_XMODEM;
    cfg.protocolConfig.xmodem.maxErrorCount = 10U;
    cfg.protocolConfig.xmodem.ackTimeoutSec = 3U;
    cfg.protocolConfig.xmodem.charTimeoutMs = 2U;

    gSerialContextObj = (Serial_Context_Obj *)Serial_init((void *)&cfg);
    Serial_open(gSerialContextObj);
    Serial_start((void *)gSerialContextObj);

    status = Serial_read((void *)gSerialContextObj, gUartRecvBuff, CPSW_UTILS_ARRAYSIZE(gUartRecvBuff));

    if (status == Serial_success)
    {
        Mailbox_post(gCpswCfgServerObj.hMboxIn, &msg, BIOS_WAIT_FOREVER);
        Mailbox_pend(gCpswCfgServerObj.hMboxOut, &msg, BIOS_WAIT_FOREVER);
    }
}

#endif

static void CpswCfgServer_dispatcherTask(UArg arg0,
                                         UArg arg1)
{
    int i, tokenCount, cmd;
    char choice[5U];
    int32_t maxLen;

    while (true)
    {
        CpswDispatcher_MailboxObj msg;
        Mailbox_pend(gCpswCfgServerObj.hMboxIn, &msg, BIOS_WAIT_FOREVER);

        jsmn_init(&gCpswCfgServerObj.parser);
        tokenCount = jsmn_parse(&gCpswCfgServerObj.parser,
                                msg.recv,
                                CPSW_CONFIG_SERVER_BUFFER_SIZE,
                                token,
                                CPSW_UTILS_ARRAYSIZE(token));
        cmd = -1;
        for (i = 0; i < tokenCount; i++)
        {
            if (jsoneq(msg.recv, &token[i], "cmd") == 0)
            {
                memset(choice, 0U, sizeof(choice));
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       sizeof(choice));
                snprintf(choice, maxLen, "%s", msg.recv + token[i + 1].start);
                cmd = atoi(choice);
                i++;
                break;
            }
        }

        if (cmd != -1 && cmd < CPSW_CONFIG_SERVER_FXN_TBL_SIZE && NULL != cpswCfgServer_fxn_table[cmd])
        {
            if (cmd == 9U || cmd == 10U)
            {
                CpswCfgServer_InterVlanConfig interVlanCfg;

                CpswCfgServer_parseInterVlanRoutingCfg(msg.recv,
                                                       &interVlanCfg);
                cpswCfgServer_fxn_table[cmd]((char *)&interVlanCfg, msg.send);
                Mailbox_post(gCpswCfgServerObj.hMboxOut, &msg, BIOS_WAIT_FOREVER);
            }
            else
            {
                cpswCfgServer_fxn_table[cmd](msg.recv, msg.send);
                Mailbox_post(gCpswCfgServerObj.hMboxOut, &msg, BIOS_WAIT_FOREVER);
            }
        }
        else
        {
            char errorString[CPSW_CONFIG_SERVER_BUFFER_SIZE] = {"{\"Error\":\"No cmd\"}"};
            memcpy(msg.send, &errorString[0], sizeof(errorString));
            Mailbox_post(gCpswCfgServerObj.hMboxOut, &msg, BIOS_WAIT_FOREVER);
        }
    }
}

static void CpswCfgServer_getVersion(char *recvBuff,
                                     char *sendBuff)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    Cpsw_GetVersionsOutArgs outArgs;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    memset(&outString[0], 0, sizeof(outString));
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &outArgs);
    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_IOCTL_GET_VERSION, &prms);
    if (status != CPSW_SOK)
    {
        sprintf(outString, "\"Error\":%d", status);
        memcpy(&sendBuff[0U], outString, sizeof(outString));
        return;
    }

    sprintf(outString, "{\"ss\":{\"major\":%u,\"minor\":%u,\"rtl\":%u,\"id\":%u},"
            "\"cpsw\":{\"major\":%u,\"minor\":%u,\"rtl\":%u,\"id\":%u},"
            "\"ale\":{\"major\":%u,\"minor\":%u,\"rtl\":%u,\"id\":%u},"
            "\"cpts\":{\"major\":%u,\"minor\":%u,\"rtl\":%u,\"id\":%u},"
            "\"mdio\":{\"major\":%u,\"minor\":%u,\"rtl\":%u,\"id\":%u,\"scheme\":%u,\"bu\":%u}}",
            outArgs.ss.major, outArgs.ss.minor, outArgs.ss.rtl, outArgs.ss.id,
            outArgs.cpsw.major, outArgs.cpsw.minor, outArgs.cpsw.rtl, outArgs.cpsw.id,
            outArgs.ale.major, outArgs.ale.minor, outArgs.ale.rtl, outArgs.ale.id,
            outArgs.cpts.major, outArgs.cpts.minor, outArgs.cpts.rtl, outArgs.cpts.id,
            outArgs.mdio.major, outArgs.mdio.minor, outArgs.mdio.rtl, outArgs.mdio.id,
            outArgs.mdio.scheme, outArgs.mdio.bu);

    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_getBcastMcastLimits(char *recvBuff,
                                              char *sendBuff)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_GetBcastMcastRateLimitOutArgs outArgs;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    memset(&outString[0], 0, sizeof(outString));
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &outArgs);
    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                        gCpswCfgServerObj.coreId,
                        CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT,
                        &prms);
    if (status != CPSW_SOK)
    {
        sprintf(outString, "\"Error\":%d", status);
        memcpy(&sendBuff[0U], outString, sizeof(outString));
        return;
    }

    sprintf(outString, "{\"rateLimitEnabled\":%d,\"rateLimitAtTxPort\":%d,\"numPorts\":%d,"
            "\"portPrms\":[{\"portNum\":%d,\"enableBRateLimitForPort\":%d,"
            "\"enableMRateLimitForPort\":%d,\"bcastLimitNumPktsPerSec\":%d,"
            "\"mcastLimitNumPktsPerSec\":%d},{\"portNum\":%d,"
            "\"enableBRateLimitForPort\":%d,\"enableMRateLimitForPort\":%d,"
            "\"bcastLimitNumPktsPerSec\":%d,\"mcastLimitNumPktsPerSec\":%d}]}",
            outArgs.rateLimitEnabled, outArgs.rateLimitAtTxPort, outArgs.numPorts,
            outArgs.portPrms[0].portNum, outArgs.portPrms[0].enableBRateLimitForPort,
            outArgs.portPrms[0].enableMRateLimitForPort, outArgs.portPrms[0].bcastLimitNumPktsPerSec,
            outArgs.portPrms[0].mcastLimitNumPktsPerSec, outArgs.portPrms[1].portNum,
            outArgs.portPrms[1].enableBRateLimitForPort, outArgs.portPrms[1].enableMRateLimitForPort,
            outArgs.portPrms[1].bcastLimitNumPktsPerSec, outArgs.portPrms[1].mcastLimitNumPktsPerSec);

    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_lookupUnicast(char *recvBuff,
                                        char *sendBuff)
{
    int32_t status, i, tokenCount, maxLen;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];
    Cpsw_IoctlPrms prms;
    CpswAle_MacAddrInfo inArgs;
    CpswAle_GetUcastEntryOutArgs outArgs;

    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));
    memset(&outString[0], 0, CPSW_UTILS_ARRAYSIZE(outString));
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));

    for (i = 1U; i < tokenCount; i++)
    {
        char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
        if (jsoneq(recvBuff, &token[i], "vlanId") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            inArgs.vlanId = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "addr") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            status = CpswAppUtils_macAddrAtoI(value, &inArgs.addr[0U]);
            CpswAppUtils_assert(status == CPSW_SOK);
            i++;
        }
    }

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);
    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_ALE_IOCTL_LOOKUP_UNICAST, &prms);
    if (status != CPSW_SOK)
    {
        sprintf(outString, "\"Error\":%d", status);
        memcpy(&sendBuff[0U], outString, sizeof(outString));
        return;
    }

    sprintf(outString, "{\"touched\":%d,\"aleEntryIndex\":%d,"
            "\"unicastInfo\":{\"portNum\":%d,\"blocked\":%d,\"secure\":%d,"
            "\"super\":%d,\"ageable\":%d,\"trunk\":%d}}",
            outArgs.touched, outArgs.aleEntryIndex,
            outArgs.unicastInfo.portNum, outArgs.unicastInfo.blocked, outArgs.unicastInfo.secure,
            outArgs.unicastInfo.super, outArgs.unicastInfo.ageable, outArgs.unicastInfo.trunk);

    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_getHostPortStats(char *recvBuff,
                                           char *sendBuff)
{
    int32_t status;
    uint8_t i;
    Cpsw_IoctlPrms prms;
    CpswStats_PortStats outArgs;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &outArgs);
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];
    memset(&outString[0], 0, sizeof(outString));
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));

    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_STATS_IOCTL_GET_HOSTPORT_STATS, &prms);
    if (status != CPSW_SOK)
    {
        sprintf(outString, "\"Error\":%d", status);
        memcpy(&sendBuff[0U], outString, sizeof(outString));
        return;
    }

    sprintf(outString, "{\"val\":[");

    for (i = 0U; i < CPSW_STATS_BLOCK_ELEM_NUM; i++)
    {
#if defined(BUILD_MPU1_0)
        if (i != (CPSW_STATS_BLOCK_ELEM_NUM - 1U))
        {
            sprintf(outString + strlen(outString), "%lu,", outArgs.val[i]);
        }
        else
        {
            sprintf(outString + strlen(outString), "%lu", outArgs.val[i]);
        }

#else
        if (i != (CPSW_STATS_BLOCK_ELEM_NUM - 1U))
        {
            sprintf(outString + strlen(outString), "%llu,", outArgs.val[i]);
        }
        else
        {
            sprintf(outString + strlen(outString), "%llu", outArgs.val[i]);
        }
#endif
    }

    sprintf(outString + strlen(outString), "]}");
    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_getMacPortStats(char *recvBuff,
                                          char *sendBuff)
{
    int32_t status, maxLen;
    Cpsw_IoctlPrms prms;
    CpswStats_PortStats outArgs;
    CpswStats_GenericMacPortInArgs inArgs;
    uint32_t i, tokenCount;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));

    for (i = 1U; i < tokenCount; i++)
    {
        if (jsoneq(recvBuff, &token[i], "portNum") == 0)
        {
            char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            int8_t val = atoi(value);
            switch (val)
            {
                case 0:
                    inArgs.portNum = CPSW_MAC_PORT_0;
                    break;

                case 1:
                    inArgs.portNum = CPSW_MAC_PORT_1;
                    break;

                case 2:
                    inArgs.portNum = CPSW_MAC_PORT_2;
                    break;

                case 3:
                    inArgs.portNum = CPSW_MAC_PORT_3;
                    break;

                case 4:
                    inArgs.portNum = CPSW_MAC_PORT_4;
                    break;

                case 5:
                    inArgs.portNum = CPSW_MAC_PORT_5;
                    break;

                case 6:
                    inArgs.portNum = CPSW_MAC_PORT_6;
                    break;

                case 7:
                    inArgs.portNum = CPSW_MAC_PORT_7;
                    break;
            }

            i++;
        }
    }

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);
    memset(&outString[0], 0, sizeof(outString));
    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));

    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_STATS_IOCTL_GET_MACPORT_STATS, &prms);
    if (status != CPSW_SOK)
    {
        sprintf(outString, "\"Error\":%d", status);
        memcpy(&sendBuff[0U], outString, sizeof(outString));
        return;
    }

    sprintf(outString, "{\"val\":[");

    for (i = 0U; i < CPSW_STATS_BLOCK_ELEM_NUM; i++)
    {
#if defined(BUILD_MPU1_0)
        if (i != (CPSW_STATS_BLOCK_ELEM_NUM - 1U))
        {
            sprintf(outString + strlen(outString), "%lu,", outArgs.val[i]);
        }
        else
        {
            sprintf(outString + strlen(outString), "%lu", outArgs.val[i]);
        }

#else
        if (i != (CPSW_STATS_BLOCK_ELEM_NUM - 1U))
        {
            sprintf(outString + strlen(outString), "%llu,", outArgs.val[i]);
        }
        else
        {
            sprintf(outString + strlen(outString), "%llu", outArgs.val[i]);
        }
#endif
    }

    sprintf(outString + strlen(outString), "]}");
    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_addVlanEntry(char *recvBuff,
                                       char *sendBuff)
{
    uint32_t vlanId = 0U, vlanMemberMask = 0U, isEnable = 0U;
    int32_t status, i, tokenCount, maxLen;
    Cpsw_IoctlPrms prms;
    CpswAle_VlanEntryInfo inArgs;
    CpswAle_AddEntryOutArgs outArgs;
    CpswAle_VlanIdInfo getVlanInArgs;
    CpswAle_GetVlanEntryOutArgs getVlanOutArgs;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));

    for (i = 1; i < tokenCount; i++)
    {
        char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
        if (jsoneq(recvBuff, &token[i], "vlanId") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            vlanId = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "vlanMemberMask") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            vlanMemberMask = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "isEnable") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            isEnable = atoi(value);
            i++;
        }
    }

    if (isEnable)
    {
        inArgs.disallowIPFragmentation  = FALSE;
        inArgs.forceUntaggedEgressMask  = 0U;
        inArgs.limitIPNxtHdr            = FALSE;
        inArgs.noLearnMask              = 0U;
        inArgs.regMcastFloodMask        = vlanMemberMask;
        inArgs.unregMcastFloodMask      = 0U;
        inArgs.vidIngressCheck          = FALSE;
        inArgs.vlanMemberList           = vlanMemberMask;
        inArgs.vlanIdInfo.vlanId        = vlanId;
        inArgs.vlanIdInfo.outerVlanFlag = FALSE;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                            gCpswCfgServerObj.coreId,
                            CPSW_ALE_IOCTL_ADD_VLAN,
                            &prms);
        if (status != CPSW_SOK)
        {
            getVlanInArgs.vlanId        = vlanId;
            getVlanInArgs.outerVlanFlag = FALSE;
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getVlanInArgs, &getVlanOutArgs);

            status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                                gCpswCfgServerObj.coreId,
                                CPSW_ALE_IOCTL_LOOKUP_VLAN,
                                &prms);
            if (status != CPSW_SOK)
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"NA\"}"); // Not Added
            }
            else
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":%d}", getVlanOutArgs.aleEntryIndex); // Already Existed
            }
        }
        else
        {
            memset(&outString[0], 0, sizeof(outString));
            sprintf(outString, "{\"aleEntryIndex\":%d}", outArgs.aleEntryIndex); // Added normally
        }
    }
    else
    {
        getVlanInArgs.vlanId        = vlanId;
        getVlanInArgs.outerVlanFlag = FALSE;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getVlanInArgs, &getVlanOutArgs);

        status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                            gCpswCfgServerObj.coreId,
                            CPSW_ALE_IOCTL_LOOKUP_VLAN,
                            &prms);
        if (status != CPSW_SOK)
        {
            memset(&outString[0], 0, sizeof(outString));
            sprintf(outString, "{\"aleEntryIndex\":\"NF\"}"); // Not Found
        }
        else
        {
            CPSW_IOCTL_SET_IN_ARGS(&prms, &getVlanInArgs);
            status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                                gCpswCfgServerObj.coreId,
                                CPSW_ALE_IOCTL_REMOVE_VLAN,
                                &prms);
            if (status != CPSW_SOK)
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"NR\"}"); // Not Removed
            }
            else
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"R\"}"); // Removed
            }
        }
    }

    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));
    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_enabledisableMcast(char *recvBuff,
                                             char *sendBuff)
{
    uint8_t macAddr[6] = {0};
    uint32_t vlanId = 0U, vlanMemberMask = 0U, isEnable = 0U;
    int32_t status, i, tokenCount, maxLen;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setMcastOutArgs;
    CpswAle_SetMcastEntryInArgs setMcastInArgs;
    CpswAle_GetMcastEntryInArgs getMcastInArgs;
    CpswAle_GetMcastEntryOutArgs getMcastOutArgs;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));

    for (i = 1; i < tokenCount; i++)
    {
        char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
        if (jsoneq(recvBuff, &token[i], "addr") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            status = CpswAppUtils_macAddrAtoI(value, &macAddr[0U]);
            CpswAppUtils_assert(status == CPSW_SOK);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "vlanId") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            vlanId = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "vlanMemberMask") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            vlanMemberMask = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "isEnable") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
            isEnable = atoi(value);
            i++;
        }
    }

    memcpy(&setMcastInArgs.addr.addr[0], macAddr, sizeof(setMcastInArgs.addr.addr));
    setMcastInArgs.addr.vlanId = vlanId;

    if (isEnable)
    {
        setMcastInArgs.info.superFlag  = false;
        setMcastInArgs.info.fwdState   = CPSW_ALE_FWDSTLVL_FWD_LRN;
        setMcastInArgs.info.portMask   = vlanMemberMask;
        setMcastInArgs.info.numIgnBits = 0;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);

        status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                            gCpswCfgServerObj.coreId,
                            CPSW_ALE_IOCTL_ADD_MULTICAST,
                            &prms);
        if (status != CPSW_SOK)
        {
            memset(&outString[0], 0, sizeof(outString));
            sprintf(outString, "{\"aleEntryIndex\":\"NA\"}"); // Not Added
        }
        else
        {
            memcpy(&getMcastInArgs.addr.addr[0U], macAddr, sizeof(macAddr));
            getMcastInArgs.addr.vlanId = vlanId;
            getMcastInArgs.numIgnBits  = 0;
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getMcastInArgs, &getMcastOutArgs);

            status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                                gCpswCfgServerObj.coreId,
                                CPSW_ALE_IOCTL_LOOKUP_MULTICAST,
                                &prms);
            if (status == CPSW_SOK)
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":%d}", getMcastOutArgs.aleIndex); // Added
            }
            else
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"NF\"}"); // Not Found
            }
        }
    }
    else
    {
        memcpy(&getMcastInArgs.addr.addr[0U], macAddr, sizeof(macAddr));
        getMcastInArgs.addr.vlanId = vlanId;
        getMcastInArgs.numIgnBits  = 0;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getMcastInArgs, &getMcastOutArgs);

        status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                            gCpswCfgServerObj.coreId,
                            CPSW_ALE_IOCTL_LOOKUP_MULTICAST,
                            &prms);
        if (status != CPSW_SOK)
        {
            memset(&outString[0], 0, sizeof(outString));
            sprintf(outString, "{\"aleEntryIndex\":\"NF\"}");  // Not Found
        }
        else
        {
            CpswAle_MacAddrInfo removeInArgs;
            memcpy(&removeInArgs.addr[0U], &getMcastInArgs.addr.addr[0U], sizeof(macAddr));
            removeInArgs.vlanId = getMcastInArgs.addr.vlanId;

            CPSW_IOCTL_SET_IN_ARGS(&prms, &removeInArgs);
            status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                                gCpswCfgServerObj.coreId,
                                CPSW_ALE_IOCTL_REMOVE_ADDR,
                                &prms);
            if (status == CPSW_SOK)
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"R\"}"); // Removed
            }
            else
            {
                memset(&outString[0], 0, sizeof(outString));
                sprintf(outString, "{\"aleEntryIndex\":\"NR\"}"); // Not Removed
            }
        }
    }

    memset(&sendBuff[0U], 0, CPSW_CONFIG_SERVER_BUFFER_SIZE * sizeof(sendBuff[0]));
    memcpy(&sendBuff[0U], outString, sizeof(outString));
}

static void CpswCfgServer_parseInterVlanRoutingCfg(char *recvBuff,
                                                   CpswCfgServer_InterVlanConfig *pinterVlanCfg)
{
    int32_t i, tokenCount, status, maxLen;

    if (pinterVlanCfg != NULL && recvBuff != NULL)
    {
        jsmn_init(&gCpswCfgServerObj.parser);
        tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));
        for (i = 1; i < tokenCount; i++)
        {
            char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
            if (jsoneq(recvBuff, &token[i], "src_mac_addr") == 0)

            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s", recvBuff + token[i + 1].start);
                status = CpswAppUtils_macAddrAtoI(value, &pinterVlanCfg->srcMacAddr[0U]);
                CpswAppUtils_assert(status == CPSW_SOK);
                i++;
            }
            if (jsoneq(recvBuff, &token[i], "src_IpAddr") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                status = CpswAppUtils_ipAddrAtoI(value, &pinterVlanCfg->srcIpv4Addr[0]);
                CpswAppUtils_assert(status == CPSW_SOK);
                i++;
            }
            if (jsoneq(recvBuff, &token[i], "ing_vlanId") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                pinterVlanCfg->ingVlanId = atoi(value);
                i++;
            }

            if (jsoneq(recvBuff, &token[i], "ing_portNum") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                pinterVlanCfg->ingPortNum = atoi(value);
                i++;
            }

            if (jsoneq(recvBuff, &token[i], "dst_mac_addr") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                status = CpswAppUtils_macAddrAtoI(value, &pinterVlanCfg->dstMacAddr[0U]);
                CpswAppUtils_assert(status == CPSW_SOK);
                i++;
            }
            if (jsoneq(recvBuff, &token[i], "dst_IpAddr") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                status = CpswAppUtils_ipAddrAtoI(value, &pinterVlanCfg->dstIpv4Addr[0]);
                CpswAppUtils_assert(status == CPSW_SOK);
                i++;
            }

            if (jsoneq(recvBuff, &token[i], "egr_vlanId") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                pinterVlanCfg->egrVlanId = atoi(value);
                i++;
            }

            if (jsoneq(recvBuff, &token[i], "egr_portNum") == 0)
            {
                maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                       CPSW_CONFIG_SERVER_TOKEN_SIZE);
                snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
                pinterVlanCfg->egrPortNum = atoi(value);
                i++;
            }
        }
    }
}

static void CpswCfgServer_addIpNxtHdrFilter(char *recvBuff,
                                            char *sendBuff)
{
    uint32_t vlanId = 0U, vlanMemberMask = 0U;
    int32_t status, i, tokenCount, maxLen;
    Cpsw_IoctlPrms prms;
    CpswAle_VlanEntryInfo inArgs;
    CpswAle_AddEntryOutArgs outArgs;
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));

    for (i = 1; i < tokenCount; i++)
    {
        char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
        if (jsoneq(recvBuff, &token[i], "vlanId") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            vlanId = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "vlanMemberMask") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            vlanMemberMask = atoi(value);
            i++;
        }
    }

    inArgs.disallowIPFragmentation  = false;
    inArgs.forceUntaggedEgressMask  = 0U;
    inArgs.limitIPNxtHdr            = true;
    inArgs.noLearnMask              = 0U;
    inArgs.regMcastFloodMask        = vlanMemberMask;
    inArgs.unregMcastFloodMask      = 0U;
    inArgs.vidIngressCheck          = false;
    inArgs.vlanMemberList           = vlanMemberMask;
    inArgs.vlanIdInfo.vlanId        = vlanId;
    inArgs.vlanIdInfo.outerVlanFlag = false;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw,
                        gCpswCfgServerObj.coreId,
                        CPSW_ALE_IOCTL_ADD_VLAN,
                        &prms);
    if (status != CPSW_SOK)
    {
        /* Entry not added */
        memset(&outString[0], 0, sizeof(outString));
        sprintf(outString, "{\"aleEntryIndex\":\"NA\"}");
    }
}

static void CpswCfgServer_enableRateLimiting(char *recvBuff,
                                             char *sendBuff)
{
    int32_t status, i, tokenCount, maxLen;
    CpswCfgServer_RateLimitingCfg rateLimitingCfg = {0};
    char outString[CPSW_CONFIG_SERVER_BUFFER_SIZE];

    memset(&rateLimitingCfg, 0, sizeof(rateLimitingCfg));
    jsmn_init(&gCpswCfgServerObj.parser);
    tokenCount = jsmn_parse(&gCpswCfgServerObj.parser, recvBuff, CPSW_CONFIG_SERVER_BUFFER_SIZE, token, CPSW_UTILS_ARRAYSIZE(token));

    for (i = 1; i < tokenCount; i++)
    {
        char value[CPSW_CONFIG_SERVER_TOKEN_SIZE];
        if (jsoneq(recvBuff, &token[i], "src_mac_addr") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            status = CpswAppUtils_macAddrAtoI(value, &rateLimitingCfg.srcMacAddr[0]);
            CpswAppUtils_assert(status == CPSW_SOK);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "ing_portNum") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            rateLimitingCfg.ingPortNum = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "dst_mac_addr") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            status = CpswAppUtils_macAddrAtoI(value, &rateLimitingCfg.dstMacAddr[0]);
            CpswAppUtils_assert(status == CPSW_SOK);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "egr_portNum") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            rateLimitingCfg.egrPortNum = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "PIR") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            rateLimitingCfg.pir = atoi(value);
            i++;
        }

        if (jsoneq(recvBuff, &token[i], "CIR") == 0)
        {
            maxLen = CpswUtils_min(token[i + 1].end - token[i + 1].start + 1U,
                                   CPSW_CONFIG_SERVER_TOKEN_SIZE);
            snprintf(value, maxLen, "%s",recvBuff + token[i + 1].start);
            rateLimitingCfg.cir = atoi(value);
            i++;
        }
    }

    status = CpswCfgServer_addAleEntryRateLimiting(&rateLimitingCfg);

    if (status == CPSW_SOK)
    {
        status = CpswCfgServer_setPolicerPriorityRateLimiting(&rateLimitingCfg);
    }

    if (status != CPSW_SOK)
    {
        /* Entry not added */
        memset(&outString[0], 0, sizeof(outString));
        sprintf(outString, "{\"RateLimitingStatus\":\"Failed\"}");
    }

}

static int32_t CpswCfgServer_addAleEntryRateLimiting(CpswCfgServer_RateLimitingCfg *pRateLimitingCfg)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], &pRateLimitingCfg->dstMacAddr[0U],
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0;
    setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(pRateLimitingCfg->egrPortNum);
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);

    CpswAppUtils_assert(status == CPSW_SOK);

    return status;
}

static int32_t CpswCfgServer_setPolicerPriorityRateLimiting(CpswCfgServer_RateLimitingCfg *pRateLimitingCfg)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    uint32_t i;

    for (i = 0; i < CPSW_PRIORITY_NUM; i++)
    {
        setPolicerInArgs.policerMatch.policerMatchEnableMask = CPSW_ALE_POLICER_MATCH_PRIORITY;
        setPolicerInArgs.policerMatch.priority               = i;

        setPolicerInArgs.policerMatch.policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_MACSRC;
        memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], &pRateLimitingCfg->srcMacAddr[0U],
               sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
        setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(pRateLimitingCfg->ingPortNum);

        setPolicerInArgs.threadIdEnable         = false;
        setPolicerInArgs.threadId               = 0;
        setPolicerInArgs.peakRateInBitsPerSec   = pRateLimitingCfg->pir;
        setPolicerInArgs.commitRateInBitsPerSec = pRateLimitingCfg->cir;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

        status = Cpsw_ioctl(gCpswCfgServerObj.hCpsw, gCpswCfgServerObj.coreId, CPSW_ALE_IOCTL_SET_POLICER,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswCfgServer_setPolicerPriorityRateLimiting() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                               status);
            break;
        }
    }

    return status;
}
