/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_serial.c
 *
 *  \brief Top level serial protocol layer
 *
 */
/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "inc/cpsw_serial.h"

#include <stdio.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/*   A global context is used. Only one transfer can be active at a time */
CpswSerial_ContextObj Serial_Context __attribute__ ((section(".serialContext")));

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

/*   Validates the configuration, then passes through to the specific protocol
 *   initialization
 */
void *CpswSerial_init(CpswSerial_ConfigObj *cfg)
{
    /* Sanity check on the input structure */
    if (cfg->magic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return(NULL);
    }

    if (cfg->protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_init((void *)cfg));
    }
    else
    {
        return(NULL);
    }
}

/* Dispatches to the correct protocol handler */
int32_t CpswSerial_open(CpswSerial_ContextObj *cxt)
{
    /* Sanity check on the context */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_open((void *)cxt));
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}

/* The reception protocol is initiated */
int32_t CpswSerial_start(CpswSerial_ContextObj *cxt)
{
    /* Sanity check on the context */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_start((void *)cxt));
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}

/* Debug info is passed through to the protocol handler for output */
int32_t CpswSerial_sendDebugInfo(CpswSerial_ContextObj *cxt,
                                 uint8_t c)
{
    /* Sanity check on the context */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_sendDebugInfo((void *)cxt, c));
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}

/* Dispatches to the correct protocol handler */
int32_t CpswSerial_read(CpswSerial_ContextObj *cxt,
                        uint8_t *addr,
                        int32_t len)
{
    /* Context sanity check */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    /* Fall through to the protocol specific handler */
    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        int32_t length = CpswSerial_invalid_read_len;
        char buff[CPSW_SERIAL_XMODEM_DATA_MAX_LEN] = {0};
        char *pRecvBuff = (char *)addr;

        CpswSerialXmodem_read((void *)cxt,
                              (uint8_t *)buff,
                              CPSW_SERIAL_XMODEM_DATA_MAX_LEN);
        while (length != CpswSerial_read_past_end)
        {
            length = CpswSerial_readStatus(cxt, (uint8_t **)&buff);

            if (length > 0U)
            {
                /* concatenate  buff to recvBuff */
                sprintf(pRecvBuff, "%.*s", length, buff);
                pRecvBuff += length;

                /* memset tempBuff and re-submit for next packet*/
                memset(&buff[0U], 0U, CPSW_SERIAL_XMODEM_DATA_MAX_LEN);
                CpswSerialXmodem_read((void *)cxt,
                                      (uint8_t *)buff,
                                      CPSW_SERIAL_XMODEM_DATA_MAX_LEN);
            }
        }

        return(CpswSerial_success);
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}

/* Dispatches to the correct protocol handler */
int32_t CpswSerial_readStatus(CpswSerial_ContextObj *cxt,
                              uint8_t **dataAddr)
{
    /* Context sanity check */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    /* Fall through to the protocol specific handler */
    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_readStatus((void *)cxt, dataAddr));
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}

/* Dispatches to the correct protocol handler */
int32_t CpswSerial_close(CpswSerial_ContextObj *cxt,
                         uint32_t timeoutMsec)
{
    /* Context sanity check */
    if (cxt->serialMagic != CPSW_SERIAL_MODULE_MAGIC)
    {
        return((int32_t)CpswSerial_bad_magic);
    }

    /* Fall through to the protocol specific handler */
    if (cxt->cfg.protocol == CPSW_SERIAL_PROTOCOL_XMODEM)
    {
        return(CpswSerialXmodem_close((void *)cxt, timeoutMsec));
    }
    else
    {
        return(CpswSerial_bad_protocol);
    }
}
