/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file main_tirtos.c
 *
 *  \brief Template file for Integrating NDK
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
#include <ti/drv/sciclient/sciclient.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/utils/Load.h>
#include <ti/sysbios/knl/Semaphore.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_mcm.h>

#include <ti/drv/cpsw/nimucpsw/nimu_ndk.h>
#include <ti/drv/cpsw/nimucpsw/ndk2cpsw_appif.h>

/* NDK headers */
#include <ti/ndk/inc/netmain.h>
#include <ti/ndk/inc/stkmain.h>
#include <ti/ndk/inc/socket.h>
#include <ti/ndk/inc/_stack.h>
#include <ti/ndk/inc/tools/servers.h>
#include <ti/ndk/inc/tools/console.h>

/* OSAL Header files */
#include <ti/osal/osal.h>

#include "webpage.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct
{
    /* CPSW instance type */
    Cpsw_Type cpswType;

    /* MAC port number */
    Cpsw_MacPort portNum;

    CpswMcm_Handle hMcm[CPSW_COUNT];

    /* Handle to UDMA Driver */
    Udma_DrvHandle hUdmaDrv;

    /* Host Port MAC address to be used by the application */
    uint8_t hostMacAddr[CPSW_MAC_ADDR_LEN];
} CpswNimu_AppObj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static int32_t CpswApp_init(void);

void           CpswApp_deInit(void);

/* ========================================================================== */
/*                          Extern variables                                  */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*!
 * \brief
 *   NIMUDeviceTable
 *
 * \details
 *  This is the NIMU Device Table for the Platform.
 *  This should be defined for each platform. Since the current platform
 *  has a single network Interface; this has been defined here. If the
 *  platform supports more than one network interface this should be
 *  defined to have a list of "initialization" functions for each of the
 *  interfaces.
 */
NIMU_DEVICE_TABLE_ENTRY NIMUDeviceTable[2U] =
{
    /*!
     * \brief  NIMU_NDK_Init for this network device
     */
    {&NIMU_NDK_init},
    {NULL}
};

static void *hEcho    = 0;
static void *hEchoUdp = 0;
static void *hData    = 0;
static void *hNull    = 0;
static void *hOob     = 0;

/* App string used by Telnet server */
char *VerStr = "NIMU CPSW Example";

static CpswNimu_AppObj gCpswNimuAppObj =
{
#if defined(SOC_AM65XX)
    .cpswType = CPSW_2G,
    .portNum  = CPSW_MAC_PORT_0,
#elif defined(SOC_J721E)
#if defined(BUILD_MCU2_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
    .cpswType = CPSW_9G,
    .portNum  = CPSW_MAC_PORT_1,
#else
    .cpswType = CPSW_2G,
    .portNum  = CPSW_MAC_PORT_0,
#endif
#endif
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int main(void)
{
    CpswAppBoardUtils_init();

    CpswAppUtils_print("Board_init success\n");

    CpswAppUtils_enableClocks(gCpswNimuAppObj.cpswType);

    /* User can include their application code here */

    BIOS_start();    /* does not return */

    return(0);
}

static void CpswApp_initAleConfig(CpswAle_Config *aleConfig)
{
    int32_t status = CPSW_SOK;

    aleConfig->modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging = TRUE;
    aleConfig->agingConfig.agingPeriodInMs = 1000;

    CpswAppUtils_assert(status == CPSW_SOK);

    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->policerGlobalConfig.policingEnable    = true;
    aleConfig->policerGlobalConfig.yellowDropEnable  = false;
    /*! Enables the ALE to drop the red colored packets. */
    aleConfig->policerGlobalConfig.redDropEnable = false;
    /*! Policing match mode */
    aleConfig->policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_GREEN;
}

void CpswApp_initLinkArgs(Cpsw_OpenPortLinkInArgs *linkArgs,
                          Cpsw_MacPort macPort)
{
    CpswMacPort_Config *macConfig      = &linkArgs->macConfig;
    CpswMacPort_LinkConfig *linkConfig = &linkArgs->linkConfig;
    CpswMacPort_Interface *interface   = &linkArgs->interface;
    CpswPhy_Config *phyConfig          = &linkArgs->phyConfig;

    linkArgs->portNum = macPort;

    Cpsw_initMacPortParams(macConfig);

    CpswAppBoardUtils_setPhyConfig(gCpswNimuAppObj.cpswType,
                                   macConfig,
                                   macPort,
                                   interface,
                                   phyConfig);

    if (phyConfig->phyAddr == CPSW_PHY_INVALID_PHYADDR)
    {
        linkConfig->speed     = CPSW_SPEED_1GBIT;
        linkConfig->duplexity = CPSW_DUPLEX_FULL;
    }
    else
    {
        linkConfig->speed     = CPSW_SPEED_AUTO;
        linkConfig->duplexity = CPSW_DUPLEX_AUTO;
    }
}

static int32_t CpswApp_init(void)
{
    int32_t status = CPSW_SOK;
    CpswMcm_InitConfig cpswMcmCfg;
    Cpsw_Config cpswCfg;
    Cpsw_MacPort macPortList[] = {gCpswNimuAppObj.portNum};
    uint8_t numMacPorts        = (sizeof(macPortList) / sizeof(macPortList[0U]));

    CpswAppUtils_assert(numMacPorts <= Cpsw_getMacPortMax(gCpswNimuAppObj.cpswType));

    /* Set configuration parameters */
    Cpsw_initParams(&cpswCfg);
    cpswCfg.vlanConfig.vlanAware                         = false;
    cpswCfg.hostPortConfig.removeCrc                     = true;
    cpswCfg.hostPortConfig.padShortPacket                = true;
    cpswCfg.hostPortConfig.passCrcErrors                 = true;
    cpswCfg.resourcePartitionConfig.isDefaultRmPartition = true;
    cpswCfg.resourcePartitionConfig.rmPartitionPrms      = NULL;

    CpswApp_initAleConfig(&cpswCfg.aleConfig);

    cpswCfg.dmaConfig.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;

    /* Open UDMA */
    gCpswNimuAppObj.hUdmaDrv = CpswAppUtils_udmaOpen(gCpswNimuAppObj.cpswType, NULL);
    CpswAppUtils_assert(NULL != gCpswNimuAppObj.hUdmaDrv);

    cpswCfg.dmaConfig.hUdmaDrv = gCpswNimuAppObj.hUdmaDrv;

    cpswMcmCfg.cpswType           = gCpswNimuAppObj.cpswType;
    cpswMcmCfg.pCpswCfg           = &cpswCfg;
    cpswMcmCfg.setPortLinkCfg     = CpswApp_initLinkArgs;
    cpswMcmCfg.numMacPorts        = numMacPorts;
    cpswMcmCfg.periodicTaskPeriod = CPSW_PHY_FSM_TICK_PERIOD_MS; /* msecs */

    memcpy(&cpswMcmCfg.macPortList[0U], &macPortList[0U], sizeof(macPortList));

    status = CpswSoc_getMacAddr(gCpswNimuAppObj.cpswType,
                                macPortList[0],
                                &cpswCfg.aleConfig.macAddr[0]);

    memcpy(&gCpswNimuAppObj.hostMacAddr[0U],
           &cpswCfg.aleConfig.macAddr[0],
           CPSW_MAC_ADDR_LEN);

    CpswAppUtils_print("Host MAC address: ");
    CpswAppUtils_printMacAddr(&gCpswNimuAppObj.hostMacAddr[0U]);

    gCpswNimuAppObj.hMcm[gCpswNimuAppObj.cpswType] = CpswMcm_init(&cpswMcmCfg);
    CpswAppUtils_assert(NULL != gCpswNimuAppObj.hMcm[gCpswNimuAppObj.cpswType]);

    return status;
}

/* The Below hook functions are to be added to the nimu Configuration file
 * similar to the way done in nimu_example.cfg in cpsw_nimu_example
 */

/* Stack Init Hook is the first function in application
 * called by NDK when it comes up
 */
void CpswNimuApp_stackInitHook(void *hCfg)
{
    uint32_t rc;

    /* increase stack size */
    rc = 16384;
    CfgAddEntry(hCfg, CFGTAG_OS, CFGITEM_OS_TASKSTKBOOT,
                CFG_ADDMODE_UNIQUE, sizeof(uint32_t), (uint8_t *)&rc, 0);

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
    /* We do not want to see debug messages less than WARNINGS */
    rc = DBG_INFO;
    CfgAddEntry(hCfg, CFGTAG_OS, CFGITEM_OS_DBGPRINTLEVEL,
                CFG_ADDMODE_UNIQUE, sizeof(uint32_t), (uint8_t *)&rc, 0);
#endif

    AddWebFiles();
}

void CpswNimuApp_stackDeleteHook(void *hCfg)
{
    RemoveWebFiles();
}

/* CpswNimuApp_ipAddrHookFxn is the function called by NDK,
 * once it gets an IP Address.
 */
void CpswNimuApp_ipAddrHookFxn(uint32_t IPAddr,
                               uint32_t IfIdx,
                               uint32_t fAdd)
{
    char ipAddrStr[20];

    NtIPN2Str(IPAddr, ipAddrStr);

    if (1U == fAdd)
    {
        CpswAppUtils_print("CPSW NIMU App: Added Network IP address I/F 1: %s", ipAddrStr);
    }
    else
    {
        CpswAppUtils_print("CPSW NIMU App: Removed Network IP address I/F 1: %s", ipAddrStr);
    }
}

void CpswNimuApp_netOpenHook(void)
{
    /* Create our local servers */
    hEcho = DaemonNew(SOCK_STREAMNC, 0, 7, dtask_tcp_echo,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hEchoUdp = DaemonNew(SOCK_DGRAM, 0, 7, dtask_udp_echo,
                         OS_TASKPRINORM, OS_TASKSTKNORM, 0, 1);
    hData = DaemonNew(SOCK_STREAM, 0, 1000, dtask_tcp_datasrv,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hNull = DaemonNew(SOCK_STREAMNC, 0, 1001, dtask_tcp_nullsrv,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hOob = DaemonNew(SOCK_STREAMNC, 0, 999, dtask_tcp_oobsrv,
                     OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
}

void CpswNimuApp_netCloseHook(void)
{
    DaemonFree(hOob);
    DaemonFree(hNull);
    DaemonFree(hData);
    DaemonFree(hEchoUdp);
    DaemonFree(hEcho);
}

/* NIMU/any application using CPSW has to call the CpswAppIf_getHandles
 * function to get the handle to CPSW and UDMA driver,Core Id, Core Key,
 * Host Port Rx MTU, Tx MTU of all priorities and flag indicating whether to use
 * Default flow or not.
 *
 * With the current implementation the first app to call the CpswAppIf_getHandles
 * will get the Default flow. This can be changed according to the user's application.
 */
void CpswAppIf_getHandles(CpswAppIf_HandleInfo *pAppIfHandleInfo)
{
    int32_t status;
    CpswMcm_HandleInfo handleInfo;

    if (gCpswNimuAppObj.hMcm[gCpswNimuAppObj.cpswType] == NULL)
    {
        status                          = CpswApp_init();
        pAppIfHandleInfo->isDefaultFlow = true;

        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to open CPSW: %d\n", status);
            CpswAppUtils_assert(status == CPSW_SOK);
        }
    }
    else
    {
        pAppIfHandleInfo->isDefaultFlow = false;
    }

    CpswMcm_getHandle(gCpswNimuAppObj.hMcm[gCpswNimuAppObj.cpswType], &handleInfo);

    pAppIfHandleInfo->hCpsw         = handleInfo.hCpsw;
    pAppIfHandleInfo->hUdmaDrv      = handleInfo.hUdmaDrv;
    pAppIfHandleInfo->coreId        = handleInfo.coreId;
    pAppIfHandleInfo->coreKey       = handleInfo.coreKey;
    pAppIfHandleInfo->hostPortRxMtu = handleInfo.hostPortRxMtu;
    memcpy(&pAppIfHandleInfo->txMtu[0U], &handleInfo.txMtu[0U],
           CPSW_UTILS_ARRAYSIZE(pAppIfHandleInfo->txMtu));

    pAppIfHandleInfo->printFxnCb = &CpswAppUtils_print;
    pAppIfHandleInfo->macPort    = gCpswNimuAppObj.portNum;
}

void CpswAppIf_releaseHandles(void)
{
    CpswMcm_releaseHandle(gCpswNimuAppObj.hMcm[gCpswNimuAppObj.cpswType]);
}

void CpswApp_deInit(void)
{
    while (Task_getMode(task) != Task_Mode_TERMINATED)
    {
        Task_sleep(10);
    }

    if (Task_deleteTerminatedTasks == FALSE)
    {
        Task_delete(&task);
    }

    CpswAppUtils_udmaclose(gCpswNimuAppObj.hUdmaDrv);

    memset(&gCpswNimuAppObj, 0U, sizeof(CpswNimu_AppObj));
}
