/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_appboardutils_j721e_gesi.c
 *
 * \brief    This file contains the board specific utilities for J721E GESI brd.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <ti/osal/osal.h>
#include <ti/drv/i2c/I2C.h>
#include <ti/drv/i2c/soc/I2C_soc.h>
#include <ti/csl/soc.h>
#include <ti/drv/cpsw/include/dp83867.h>
#include <ti/board/board.h>
#if defined (SOC_J721E)
#include <ti/board/src/j721e_evm/include/board_pinmux.h>
#include <ti/board/src/j721e_evm/include/board_utils.h>
#include <ti/board/src/j721e_evm/include/board_control.h>
#include <ti/board/src/j721e_evm/include/board_ethernet_config.h>
#include <ti/board/src/j721e_evm/include/board_serdes_cfg.h>
#elif defined (SOC_J7200)
#include <ti/board/src/j7200_evm/include/board_pinmux.h>
#include <ti/board/src/j7200_evm/include/board_utils.h>
#include <ti/board/src/j7200_evm/include/board_control.h>
#include <ti/board/src/j7200_evm/include/board_ethernet_config.h>
#include <ti/board/src/j7200_evm/include/board_serdes_cfg.h>
#else
#error "Invalid SOC"
#endif
#include <ti/drv/cpsw/cpsw.h>

#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

#if defined(BUILD_MCU2_0)
static void CpswAppBoardUtils_enetExpCfg(void)
{
    Board_STATUS boardStatus;

    /* Release PHY reset. Note this is needed for both SGMII and QSGMII boards */
    boardStatus = Board_cpswEnetExpPhyReset(0U);
    CpswAppUtils_assert(BOARD_SOK == boardStatus);

    if (Board_detectBoard(BOARD_ID_ENET))
    {
        /* Release the COMA_MODE pin */
        boardStatus = Board_cpswEnetExpComaModeCfg(0U);
        CpswAppUtils_assert(BOARD_SOK == boardStatus);
    }
}
#endif

#if defined(BUILD_MCU2_0)

#if defined (SOC_J721E)
static void CpswAppBoardUtils_configSierra0Clks(void)
{
    uint32_t moduleId, clkId, clkRateHz;

    /* Configure the required PLLs for SERDES0 */
    moduleId  = TISCI_DEV_SERDES_16G0;
    clkId     = TISCI_DEV_SERDES_16G0_CORE_REF1_CLK;
    clkRateHz = 100000000U;
    CpswAppUtils_clkRateSet(moduleId, clkId, clkRateHz);

    clkId     = TISCI_DEV_SERDES_16G0_CORE_REF_CLK;
    clkRateHz = 100000000U;
    CpswAppUtils_clkRateSet(moduleId, clkId, clkRateHz);
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_ON, 0U);
}
#endif

#if defined (SOC_J721E)
static void CpswAppBoardUtils_configSierra1Clks(void)
{
    uint32_t moduleId, clkId, clkRateHz;

    /* Configure the required PLLs for SERDES1 */
    moduleId  = TISCI_DEV_SERDES_16G1;
    clkId     = TISCI_DEV_SERDES_16G1_CORE_REF1_CLK;
    clkRateHz = 100000000U;
    CpswAppUtils_clkRateSet(moduleId, clkId, clkRateHz);

    clkId     = TISCI_DEV_SERDES_16G1_CORE_REF_CLK;
    clkRateHz = 100000000U;
    CpswAppUtils_clkRateSet(moduleId, clkId, clkRateHz);
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_ON, 0U);
}
#endif

static void CpswAppBoardUtils_configTorrentClks(void)
{
    uint32_t moduleId, clkId, clkRateHz;

     /* Configure the required PLLs for SERDES1 */
#if defined (SOC_J721E)
    moduleId  = TISCI_DEV_SERDES_10G0;
    clkId     = TISCI_DEV_SERDES_10G0_CORE_REF_CLK;
#elif defined (SOC_J7200)
    moduleId  = TISCI_DEV_SERDES_10G1;
    clkId     = TISCI_DEV_SERDES_10G1_CORE_REF_CLK;
#endif
    clkRateHz = 100000000U;
    CpswAppUtils_clkRateSet(moduleId, clkId, clkRateHz);
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_ON, 0U);
}

#if defined (SOC_J721E)
static void CpswAppBoardUtils_disableSierra0Clks(void)
{
    uint32_t moduleId;

    /* Set module set to HW AUTO */
    moduleId = TISCI_DEV_SERDES_16G0;
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_AUTO_OFF, 0U);
}
#endif

#if defined (SOC_J721E)
static void CpswAppBoardUtils_disableSierra1Clks(void)
{
    uint32_t moduleId;

    /* Set module set to HW AUTO */
    moduleId  = TISCI_DEV_SERDES_16G1;
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_AUTO_OFF, 0U);
}
#endif

static void CpswAppBoardUtils_disableTorrentClks(void)
{
    uint32_t moduleId;

    /* Set module set to HW AUTO */
#if defined (SOC_J721E)
    moduleId  = TISCI_DEV_SERDES_10G0;
#elif defined (SOC_J7200)
    moduleId  = TISCI_DEV_SERDES_10G1;
#endif
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_AUTO_OFF, 0U);
}
#endif

static void CpswAppBoardUtils_boardInit(Board_initCfg *boardCfg)
{
    Board_STATUS status;
    Board_PinmuxConfig_t pinmuxCfg;

    /* Configure CPSW9G PADs for Ports on the GESI board */
    Board_pinmuxGetCfg(&pinmuxCfg);
    pinmuxCfg.gesiExp = BOARD_PINMUX_GESI_CPSW;
    Board_pinmuxSetCfg(&pinmuxCfg);

#if defined(BUILD_MCU2_0)
    Board_initParams_t initParams;

    Board_getInitParams(&initParams);
    /* mcu2_0 UART instances:
     *  - J721E EVM: UART2
     *  - J7200 EVM: UART3 */
    initParams.uartInst      = CPSW_UTILS_MCU2_0_UART_INSTANCE;
    initParams.uartSocDomain = BOARD_SOC_DOMAIN_MAIN;
    /* Use non-exclusive mode in board lib when accessing the module for clock enable
     * to let Linux also access the module */
    initParams.pscMode = BOARD_PSC_DEVICE_MODE_NONEXCLUSIVE;
    Board_setInitParams(&initParams);
#endif

    status = Board_init(*boardCfg);
    CpswAppUtils_assert(status == BOARD_SOK);

    /* Avoid I2C conflicts between EthFw and u-boot */
#if !defined(CPSWAPPUTILS_BYPASS_I2C)
    if (Board_detectBoard(BOARD_ID_GESI))
    {
        CpswAppUtils_print("GESI board detected\n");
    }

    if (Board_detectBoard(BOARD_ID_ENET))
    {
        CpswAppUtils_print("QSGMII board detected\n");
    }
#endif
}

/* Hack to run EthFw with Linux
 * We do only limited pinmux configuration and also don't configure RMII and MDIO board muxes
 * as it causes conflict with Linux in using Main domain I2C0. If we need any of configuration on I2c
 * we should ask Linux to do that. Currently we wouldn't use RMII port in EthFw and we don't need MDIO
 * board mux as default is CPSW9G
 */
void CpswAppBoardUtils_initEthFw(void)
{
    Board_initCfg boardCfg = 0U;
    Board_STATUS status;
    uint32_t i, moduleCnt, appFlags;
    uint32_t clkModuleId[] = { TISCI_DEV_DDR0,
                               TISCI_DEV_TIMER12,
                               TISCI_DEV_TIMER13,
                               TISCI_DEV_UART2, };

    appFlags = 0U;
    moduleCnt = CPSW_UTILS_ARRAYSIZE(clkModuleId);
    for(i = 0; i < moduleCnt; i++)
    {
        CpswAppUtils_setDeviceState(clkModuleId[i],
                                    TISCI_MSG_VALUE_DEVICE_SW_STATE_ON,
                                    appFlags);
    }

    status = CpswAppBoardUtils_configEthFwPinmux();
    CpswAppUtils_assert(status == BOARD_SOK);

    /* Setting up UART3 in J7200 requires I2C accesses, so need to be bypassed for mcu2_0 */
#if defined(CPSWAPPUTILS_UART_ALLOWED)
    boardCfg |= BOARD_INIT_UART_STDIO;
#endif

#if defined (SOC_J721E)
    boardCfg |= BOARD_INIT_ENETCTRL_CPSW9G;
#elif defined (SOC_J7200)
    boardCfg |= BOARD_INIT_ENETCTRL_CPSW5G;
#endif

    CpswAppBoardUtils_boardInit(&boardCfg);

#if defined(CPSWAPPUTILS_UART_ALLOWED) && defined (SOC_J7200) && defined(BUILD_MCU2_0)
    /* Set SOM's MUX2 for UART3 route */
    status = Board_control(BOARD_CTRL_CMD_SET_SOM_UART_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);
#endif

#if !defined(CPSWAPPUTILS_BYPASS_I2C) && defined(BUILD_MCU2_0)
    /* Configure SGMII/QSGMII expansion board
     * Note - We call below without any check as these functions can be called
     * without daughter board also. They just change the output state of IO
     * expander pins on CP board which goes to daughter board.
     * No negative impact when they are configured without DB. */
    CpswAppBoardUtils_enetExpCfg();
#endif

#if !defined(CPSWAPPUTILS_BYPASS_QSGMII) && defined(BUILD_MCU2_0)
#if defined (SOC_J721E)
    /* On J721E EVM, port 1 (SERDES0, lane 1) is brought for SGMII */
    CpswAppBoardUtils_configSierra0Clks();
#elif defined (SOC_J7200)
    /* On J721E EVM, port 1 (SERDES0, lane 1) is brought for SGMII */
    CpswAppBoardUtils_configTorrentClks();
#endif

    if (Board_detectBoard(BOARD_ID_ENET))
    {
        status = Board_serdesCfgQsgmii();
        CpswAppUtils_assert(status == BOARD_SOK);
    }
#endif
}

void CpswAppBoardUtils_init(void)
{
    Board_initCfg boardCfg;
#if defined(BUILD_MCU2_0)
    Board_STATUS status;
#endif
#if defined(BUILD_MCU2_0)
    int32_t enetCard;
#endif

    boardCfg = BOARD_INIT_PINMUX_CONFIG|
               BOARD_INIT_MODULE_CLOCK|
               BOARD_INIT_ENETCTRL_CPSW2G|
               BOARD_INIT_UART_STDIO;

#if defined (SOC_J721E)
    boardCfg |= BOARD_INIT_ENETCTRL_CPSW9G;
#elif defined (SOC_J7200)
    boardCfg |= BOARD_INIT_ENETCTRL_CPSW5G;
#endif

    CpswAppBoardUtils_boardInit(&boardCfg);

#if defined(BUILD_MCU2_0)
#if defined (SOC_J7200)
    /* Set SOM's MUX2 for UART3 route */
    status = Board_control(BOARD_CTRL_CMD_SET_SOM_UART_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);
#endif

    /* Enable CPSW9G MDIO mux */
    status = Board_control(BOARD_CTRL_CMD_SET_GESI_CPSW_MDIO_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);

#if defined (SOC_J721E)
    /* Set IO Expander to use RMII on GESI board */
    status = Board_control(BOARD_CTRL_CMD_SET_RMII_DATA_MUX, NULL);
    CpswAppUtils_assert(status == BOARD_SOK);

    /* On J721E EVM, port 1 (SERDES0, lane 1) is brought for SGMII */
    CpswAppBoardUtils_configSierra0Clks();

    /* We don't need to enable SERDES1/2 clocks for CPSW tests on EVM. Below is test code
     * so customers have ready to use functions.
     */
    CpswAppBoardUtils_configSierra1Clks();
#endif

    CpswAppBoardUtils_configTorrentClks();

    /* Configure SGMII/QSGMII expansion board
     * Note - We call below without any check as these functions can be called
     * without daughter board also. They just change the output state of IO
     * expander pins on CP board which goes to daughter board.
     * No negative impact when they are configured without DB. */
    CpswAppBoardUtils_enetExpCfg();

    enetCard = Board_detectEnetCard();
    if (enetCard == BOARD_ENET_QSGMII)
    {
        status = Board_serdesCfgQsgmii();
        CpswAppUtils_assert(status == BOARD_SOK);
    }
    else if (enetCard == BOARD_ENET_SGMII)
    {
        status = Board_serdesCfgSgmii();
        CpswAppUtils_assert(status == BOARD_SOK);
    }
#endif
}

void CpswAppBoardUtils_deInit(void)
{
    Board_STATUS boardInitStatus;
    Board_initCfg boardCfg;

#if defined(BUILD_MCU2_0)
#if defined (SOC_J721E)
    CpswAppBoardUtils_disableSierra0Clks();
    CpswAppBoardUtils_disableSierra1Clks();
#endif
    CpswAppBoardUtils_disableTorrentClks();
#endif

    boardCfg = BOARD_INIT_PINMUX_CONFIG|
               BOARD_INIT_MODULE_CLOCK|
               BOARD_INIT_ENETCTRL_CPSW2G|
               BOARD_INIT_ENETCTRL_CPSW9G|
               BOARD_INIT_UART_STDIO;

    boardInitStatus = Board_deinit(boardCfg);
    CpswAppUtils_assert(boardInitStatus == BOARD_SOK);
}

uint32_t CpswAppBoardUtils_getPhyAddr(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum)
{
    uint32_t phyAddr = CPSW_PHY_INVALID_PHYADDR;

    switch (cpswType)
    {
        case CPSW_2G:
            CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);
            phyAddr = 0x00U;
            break;

        case CPSW_5G:
            switch (portNum)
            {
                case CPSW_MAC_PORT_0:
                    /* QSGMII port */
                    phyAddr = 16U;
                    break;

                case CPSW_MAC_PORT_1:
                    /* Use RGMII port in GESI board when QSGMII is disabled */
#if defined(CPSWAPPUTILS_BYPASS_QSGMII)
                    /* RGMII port */
                    phyAddr = 0U;
#else
                    /* QSGMII port */
                    phyAddr = 17U;
#endif
                    break;

                case CPSW_MAC_PORT_2:
                    /* QSGMII port */
                    phyAddr = 18U;
                    break;

                case CPSW_MAC_PORT_3:
                    /* QSGMII port */
                    phyAddr = 19U;
                    break;

                default:
                    CpswAppUtils_assert(false);
                    break;
            }

            break;

        case CPSW_9G:
            switch (portNum)
            {
                case CPSW_MAC_PORT_0:
                    /* RGMII port */
                    phyAddr = 12U;
                    break;

                case CPSW_MAC_PORT_2:
                    /* RGMII port */
                    phyAddr = 0U;
                    break;

                case CPSW_MAC_PORT_3:
                    /* RGMII port */
                    phyAddr = 3U;
                    break;

                case CPSW_MAC_PORT_7:
                    /* RGMII port */
                    phyAddr = 15U;
                    break;

                case CPSW_MAC_PORT_1:
                    /* QSGMII port */
                    phyAddr = 16U;
                    break;

                case CPSW_MAC_PORT_4:
                    /* QSGMII port */
                    phyAddr = 17U;
                    break;

                case CPSW_MAC_PORT_5:
                    /* QSGMII port */
                    phyAddr = 18U;
                    break;

                case CPSW_MAC_PORT_6:
                    /* QSGMII port */
                    phyAddr = 19U;
                    break;

                default:
                    CpswAppUtils_assert(false);
                    break;
            }

            break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

    return phyAddr;
}

uint32_t CpswAppBoardUtils_getSgmiiPhyAddr(void)
{
    /* Both Q/SGMII DB and GESI board use MAC port 2, so we can't use generic CpswAppBoardUtils_getPhyAddr
     * function */
    return 10U;
}

uint32_t CpswAppBoardUtils_getRmiiPhyAddr(void)
{
    /* Both QSGMII DB and GESI board use MAC port 7, so we can't use generic CpswAppBoardUtils_getPhyAddr
     * function */
    return 23U;
}

void CpswAppBoardUtils_setPhyConfigSgmii(Cpsw_MacPort portNum,
                                         CpswMacPort_Config *macConfig,
                                         CpswMacPort_Interface *interface,
                                         CpswPhy_Config *phyConfig)
{
    Board_STATUS boardStatus;
    Cpsw_Type cpswType;

    /* Override the ENET control set by board lib */
#if defined (SOC_J721E)
    CpswAppUtils_assert(portNum == CPSW_MAC_PORT_1);
    cpswType = CPSW_9G;
#elif defined (SOC_J7200)
    CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);
    cpswType = CPSW_5G;
#endif

    /* Override the ENET control set by board lib */
    CpswAppBoardUtils_setEnetControl(cpswType, portNum, SGMII);

    boardStatus = Board_serdesCfgSgmii();
    CpswAppUtils_assert(boardStatus == BOARD_SOK);

    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoardUtils_getSgmiiPhyAddr();
    phyConfig->skipExtendedConfig = true;

    /* RGMII interface type */
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_SERIAL;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;

    macConfig->sgmiiConfig.sgmiiMode = CPSW_MAC_SGMIIMODE_SGMII_WITH_PHY;
}

void CpswAppBoardUtils_setPhyConfigQsgmii(Cpsw_Type cpswType,
                                          Cpsw_MacPort portNum,
                                          CpswMacPort_Config *macConfig,
                                          CpswMacPort_Interface *interface,
                                          CpswPhy_Config *phyConfig)
{
    bool isAlpha = false;
    Cpsw_MacPort qsgmiiMain;

    /*
     * Disable QSGMII on alpha revision EVMs due to ADASVISION-3869 cpsw - mdio
     * instability for QSGMII PHYs on alpha EVM. Issue is resolved in the Beta
     * HW revision.
     */
    isAlpha = Board_isAlpha(BOARD_ID_CP);

    if (isAlpha)
    {
        CpswAppUtils_print("QSGMII DB not supported on alpha revision of EVM\n");
        CpswAppUtils_assert(false);
    }

#if defined (SOC_J721E)
    /* QSGMII PHYs are routed as follows in QUAD PORT ETH EXP board:
     *  MAC port 1 = QSGMII MAIN
     *  MAC port 4 = QSGMII SUB
     *  MAC port 5 = QSGMII SUB
     *  MAC port 6 = QSGMII SUB
     */
    CpswAppUtils_assert((portNum == CPSW_MAC_PORT_1) ||
                        (portNum == CPSW_MAC_PORT_4) ||
                        (portNum == CPSW_MAC_PORT_5) ||
                        (portNum == CPSW_MAC_PORT_6));

    qsgmiiMain = CPSW_MAC_PORT_1;
#elif defined (SOC_J7200)
    CpswAppUtils_assert((portNum == CPSW_MAC_PORT_0) ||
                        (portNum == CPSW_MAC_PORT_1) ||
                        (portNum == CPSW_MAC_PORT_2) ||
                        (portNum == CPSW_MAC_PORT_3));

    qsgmiiMain = CPSW_MAC_PORT_0;
#endif

    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoardUtils_getPhyAddr(cpswType, portNum);

    /* QSGMII ports */
    /* In QSGMII config each port is independent port just going over one link so
     * we need set up each port just like SGMII ports - autoneg, mac control etc.*/
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;
    macConfig->sgmiiConfig.sgmiiMode = CPSW_MAC_SGMIIMODE_SGMII_WITH_PHY;

    if (portNum == qsgmiiMain)
    {
        interface->sublayerType = CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN;
    }
    else
    {
        interface->sublayerType = CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB;
    }
}

void CpswAppBoardUtils_setPhyConfigRmii(Cpsw_MacPort portNum,
                                        CpswMacPort_Config *macConfig,
                                        CpswMacPort_Interface *interface,
                                        CpswPhy_Config *phyConfig)
{
    CpswAppUtils_assert(portNum == CPSW_MAC_PORT_7);

    /* Override the ENET control set by board lib */
    CpswAppBoardUtils_setEnetControl(CPSW_9G, portNum, RMII);

    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoardUtils_getRmiiPhyAddr();
    phyConfig->skipExtendedConfig = false;

    /* Mac port 7 on GESI board is RMII port */
    interface->layerType    = CPSW_MAC_LAYER_MII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;
}

void CpswAppBoardUtils_setPhyConfigRgmii(Cpsw_Type cpswType,
                                         Cpsw_MacPort portNum,
                                         CpswMacPort_Config *macConfig,
                                         CpswMacPort_Interface *interface,
                                         CpswPhy_Config *phyConfig)
{
    Dp83867_Config extendedConfig;

#if defined (SOC_J721E)
    /* RGMII PHYs can be used on the following MAC ports as per board library pinmux:
     *   GESI board: 0, 1, 2, 3
     *  SGMII board: 2, 3, 6, 7
     * QSGMII board: 0, 2, 3, 7
     */
    CpswAppUtils_assert((portNum == CPSW_MAC_PORT_0) ||
                        (portNum == CPSW_MAC_PORT_1) ||
                        (portNum == CPSW_MAC_PORT_2) ||
                        (portNum == CPSW_MAC_PORT_3) ||
                        (portNum == CPSW_MAC_PORT_6) ||
                        (portNum == CPSW_MAC_PORT_7));
#elif defined (SOC_J7200)
    if (cpswType == CPSW_2G)
    {
        CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);
    }
    else if (cpswType == CPSW_5G)
    {
        CpswAppUtils_assert(portNum == CPSW_MAC_PORT_1);
        /* Override the ENET control set by board lib */
        CpswAppBoardUtils_setEnetControl(cpswType, portNum, RGMII);
    }
#endif

    Cpsw_initPhyParams(phyConfig);

#if defined(SOC_J7200)
    phyConfig->phyAddr = 0U;
#else
    phyConfig->phyAddr = CpswAppBoardUtils_getPhyAddr(cpswType, portNum);
#endif

    /* DP83867 specific configuration */
    Dp83867_initParams(&extendedConfig);
    extendedConfig.txFifoDepth          = 4U;
    extendedConfig.txClkShiftEnable     = true;
    extendedConfig.rxClkShiftEnable     = true;
    extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
    extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
    extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
    extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
    extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

    if ((cpswType == CPSW_9G) || (cpswType == CPSW_5G))
    {
        extendedConfig.txDelayInPs      = 2750U; /* 2.75 ns */
        extendedConfig.rxDelayInPs      = 2500U; /* 2.50 ns */
    }

    Cpsw_setPhyExtendedParams(phyConfig,
                              &extendedConfig,
                              sizeof(extendedConfig));

    /* RGMII interface type */
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;
}

void CpswAppBoardUtils_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Config *macConfig,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    switch (cpswType)
    {
        case CPSW_2G:
            CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);

            CpswAppBoardUtils_setPhyConfigRgmii(cpswType,
                                                portNum,
                                                macConfig,
                                                interface,
                                                phyConfig);
            break;

        case CPSW_5G:
            /* Use RGMII port in GESI board when QSGMII is disabled */
#if defined(CPSWAPPUTILS_BYPASS_QSGMII)
            if (CPSW_MAC_PORT_1 == portNum)
            {
                CpswAppBoardUtils_setPhyConfigRgmii(cpswType,
                                                    portNum,
                                                    macConfig,
                                                    interface,
                                                    phyConfig);
            }
#else
            CpswAppBoardUtils_setPhyConfigQsgmii(cpswType,
                                                 portNum,
                                                 macConfig,
                                                 interface,
                                                 phyConfig);
#endif
            break;

        case CPSW_9G:

            if ( CPSW_MAC_PORT_0 == portNum ||
                 CPSW_MAC_PORT_2 == portNum ||
                 CPSW_MAC_PORT_3 == portNum ||
                 CPSW_MAC_PORT_7 == portNum )
            {
                CpswAppBoardUtils_setPhyConfigRgmii(cpswType,
                                                    portNum,
                                                    macConfig,
                                                    interface,
                                                    phyConfig);
            }
            else
            {
                CpswAppBoardUtils_setPhyConfigQsgmii(cpswType,
                                                     portNum,
                                                     macConfig,
                                                     interface,
                                                     phyConfig);
            }

            break;
    }

}

void CpswAppBoardUtils_setLpbkCfg(bool phyLpbk,
                                  Cpsw_MacPort portNum,
                                  const CpswMacPort_Interface *interface,
                                  CpswMacPort_Config *macConfig,
                                  CpswPhy_Config *phyConfig,
                                  CpswMacPort_LinkConfig *linkConfig)
{
    /* Set speed according to the interface type */
    if (interface->layerType == CPSW_MAC_LAYER_MII)
    {
        linkConfig->speed = CPSW_SPEED_100MBIT;
    }
    else
    {
        linkConfig->speed = CPSW_SPEED_1GBIT;
    }

    linkConfig->duplexity = CPSW_DUPLEX_FULL;

    /* MAC and PHY loopbacks are mutually exclusive */
    phyConfig->enableLoopback = phyLpbk;
    macConfig->enableLoopback = !phyLpbk;
}

void CpswAppBoardUtils_getMacAddrList(uint8_t macAddr[][CPSW_MAC_ADDR_LEN],
                                      uint32_t maxMacEntries,
                                      uint32_t *pAvailMacEntries)
{
#if !defined(CPSWAPPUTILS_BYPASS_I2C)
    uint8_t macAddrBuf[CPSW_RM_NUM_MACADDRESS_MAX * BOARD_MAC_ADDR_BYTES];
    Board_STATUS boardStatus;
    uint32_t macAddrCnt, tempCnt;
    uint32_t allocMacEntries = 0;
    uint32_t i, j;

    CpswAppUtils_assert(pAvailMacEntries != NULL);

    if (Board_detectBoard(BOARD_ID_GESI))
    {
        /* Read number of MAC addresses in GESI board */
        boardStatus = Board_readMacAddrCount(BOARD_ID_GESI, &macAddrCnt);
        CpswAppUtils_assert(boardStatus == BOARD_SOK);
        CpswAppUtils_assert(macAddrCnt <= CPSW_RM_NUM_MACADDRESS_MAX);

        /* Read MAC addresses */
        boardStatus = Board_readMacAddr(BOARD_ID_GESI,
                                        macAddrBuf,
                                        sizeof(macAddrBuf),
                                        &tempCnt);
        CpswAppUtils_assert(boardStatus == BOARD_SOK);
        CpswAppUtils_assert(tempCnt == macAddrCnt);

        /* Save only those required to meet the max number of MAC entries */
        macAddrCnt = CpswUtils_min(macAddrCnt, maxMacEntries);
        for (i = 0U; i < macAddrCnt; i++)
        {
            CPSW_UTILS_COMPILETIME_ASSERT(CPSW_MAC_ADDR_LEN == BOARD_MAC_ADDR_BYTES);
            memcpy(macAddr[i], &macAddrBuf[i * BOARD_MAC_ADDR_BYTES], CPSW_MAC_ADDR_LEN);
        }

        allocMacEntries = macAddrCnt;
    }

    if (Board_detectBoard(BOARD_ID_ENET))
    {
        /* Read number of MAC addresses in QUAD Eth board */
        boardStatus = Board_readMacAddrCount(BOARD_ID_ENET, &macAddrCnt);
        CpswAppUtils_assert(boardStatus == BOARD_SOK);
        CpswAppUtils_assert(macAddrCnt <= CPSW_RM_NUM_MACADDRESS_MAX);

        /* Read MAC addresses */
        boardStatus = Board_readMacAddr(BOARD_ID_ENET,
                                        macAddrBuf,
                                        sizeof(macAddrBuf),
                                        &tempCnt);
        CpswAppUtils_assert(boardStatus == BOARD_SOK);
        CpswAppUtils_assert(tempCnt == macAddrCnt);

        /* Save only those required to meet the max number of MAC entries */
        macAddrCnt = CpswUtils_min(macAddrCnt, maxMacEntries - allocMacEntries);
        for (i = 0U, j = allocMacEntries; i < macAddrCnt; i++, j++)
        {
            CPSW_UTILS_COMPILETIME_ASSERT(CPSW_MAC_ADDR_LEN == BOARD_MAC_ADDR_BYTES);
            memcpy(macAddr[j], &macAddrBuf[i * BOARD_MAC_ADDR_BYTES], CPSW_MAC_ADDR_LEN);
        }

        allocMacEntries += macAddrCnt;
    }

    *pAvailMacEntries = allocMacEntries;

    if (allocMacEntries == 0U)
    {
        CpswAppUtils_print("CpswAppBoardUtils_getMacAddrList Failed - GESI/ENET board not present\n");
        CpswAppUtils_assert(false);
    }
#else
    /*
     * Workaround for EthFw/u-boot I2C conflicts:
     * EthFw reads MAC addresses from GESI and QUAD_ETH boards during EthFw
     * initialization which are stored in EEPROM memories and are read via
     * I2C.  These I2C accesses tend to occur around the same u-boot is also
     * performing I2C accesses, causing transactions to timeout or other
     * similar symptoms.
     *
     * I2C sharing is a known limitation but no current solution exists at
     * this time.  As a temporary workaround, EthFw will use fixed MAC
     * addresses in Linux builds. For RTOS build, MAC addresses will still
     * be read from EEPROM as such I2C contention isn't an problem.
     */
    uint8_t macAddrBuf[][CPSW_MAC_ADDR_LEN] =
    {
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0x92U, 0xC1U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0x92U, 0xC2U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0x92U, 0xC3U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0x92U, 0xC4U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0x92U, 0xC5U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0X8BU, 0xC4U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0X8BU, 0xC5U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0X8BU, 0xC6U },
        { 0x70U, 0xFFU, 0x76U, 0x1DU, 0X8BU, 0xC7U },
    };
    uint32_t macAddrCnt = CPSWAPPUTILS_ARRAY_SIZE(macAddrBuf);

    /* Save only those required to meet the max number of MAC entries */
    *pAvailMacEntries = CpswUtils_min(macAddrCnt, maxMacEntries);
    memcpy(&macAddr[0U][0U], &macAddrBuf[0U][0U], *pAvailMacEntries * CPSW_MAC_ADDR_LEN);
#endif
}

void CpswAppBoardUtils_setEnetControl(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum,
                                      uint32_t modeSel)
{
    int32_t status = BOARD_FAIL;

    switch (cpswType)
    {
        case CPSW_2G:
            status = Board_cpsw2gMacModeConfig(modeSel);
            break;
#if defined (SOC_J7200)
        case CPSW_5G:
            status = Board_cpsw5gEthConfig(portNum, modeSel);
            break;
#endif
#if defined (SOC_J721E)
        case CPSW_9G:
            status = Board_cpsw9gEthConfig(portNum, modeSel);
            break;
#endif
        default:
            break;
    }

    CpswAppUtils_assert(status == BOARD_SOK);
}
