/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpswapp_rmcfg.c
 *
 * \brief    This file contains the CPSW driver resource management
 *           configuration used to initialize the RM init parameters passed
 *           during driver init.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/ipc/ipc.h>
/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*!
 *  \brief CPSW5G default configuration
 *
 *   Note: If user wishes to change the Resource Partition the following
 *   things must be considered:
 *   1. Sum of numTxCh allocated to each core should not exceed 8.
 *   2. Sum of numRxFlows allocated to each core should not exceed 63 (not 64),
 *      as one Rx flow is reserved to the master core.
 *
 */
static CpswRm_ResourcePrms gCpswAppRmDefCfg_5G =
{
    .coreDmaResourceInfo =
    {
        [0] =
        {
            .coreId        = IPC_MPU1_0,
            .numTxCh       = 1U,
            .numRxFlows    = 1U,
            .numMacAddress = 1U,
        },
        [1] =
        {
            .coreId        = IPC_MCU2_0,
            .numTxCh       = 4U,
            .numRxFlows    = 5U,
            .numMacAddress = 1U,
        },
        [2] =
        {
            .coreId        = IPC_MCU2_1,
            .numTxCh       = 1U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
        [3] =
        {
            .coreId        = IPC_MCU1_0,
            .numTxCh       = 2U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
    },
    .numCores = 4U,
};

/*!
 *  \brief CPSW9G default configuration
 *
 *   Note: If user wishes to change the Resource Partition the following
 *   things must be considered:
 *   1. Sum of numTxCh allocated to each core should not exceed 8.
 *   2. Sum of numRxFlows allocated to each core should not exceed 63 (not 64),
 *      as one Rx flow is reserved to the master core.
 *
 */
static CpswRm_ResourcePrms gCpswAppRmDefCfg_9G =
{
    .coreDmaResourceInfo =
    {
        [0] =
        {
            .coreId        = IPC_MPU1_0,
            .numTxCh       = 1U,
            .numRxFlows    = 1U,
            .numMacAddress = 1U,
        },
        [1] =
        {
            .coreId        = IPC_MCU2_0,
            .numTxCh       = 4U,
            .numRxFlows    = 5U,
            .numMacAddress = 1U,
        },
        [2] =
        {
            .coreId        = IPC_MCU2_1,
            .numTxCh       = 1U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
        [3] =
        {
            .coreId        = IPC_MCU1_0,
            .numTxCh       = 2U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
    },
    .numCores = 4U,
};

/*!
 *  \brief CPSW2G default configuration
 *         Note: If user wishes to change the Resource Partition the following
 *         things must be considered:
 *         1. Sum of numTxCh allocated to each core should not exceed 8.
 *         2. Sum of numRxFlows allocated to each core should not exceed 63 (not 64),
 *            as one Rx flow is reserved to the master core.
 *
 */
static CpswRm_ResourcePrms gCpswAppRmDefCfg_2G =
{
    .coreDmaResourceInfo =
    {
        [0] =
        {
            .coreId        = IPC_MPU1_0,
            .numTxCh       = 2U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
        [1] =
        {
            .coreId        = IPC_MCU2_1,
            .numTxCh       = 2U,
            .numRxFlows    = 2U,
            .numMacAddress = 1U,
        },
        [2] =
        {
            .coreId        = IPC_MCU1_0,
            .numTxCh       = 4U,
            .numRxFlows    = 4U,
            .numMacAddress = 1U,
        },
    },
    .numCores = 3,
};

/* Cores IOCTL Privileges */
static const CpswRm_IoctlPermissionTable gCpswAppIoctlPermission_2G =
{
    .defaultPermittedCoreMask = (CPSW_SET_BIT(IPC_MPU1_0) |
                                 CPSW_SET_BIT(IPC_MCU2_0) |
                                 CPSW_SET_BIT(IPC_MCU2_1) |
#if defined (SOC_J721E)
                                 CPSW_SET_BIT(IPC_MCU3_0) |
#endif
                                 CPSW_SET_BIT(IPC_MCU1_0)),
    .numEntries = 0,
};

#if defined (SOC_J7200)
static const CpswRm_IoctlPermissionTable gCpswAppIoctlPermission_5G =
{
    .defaultPermittedCoreMask = (CPSW_SET_BIT(IPC_MPU1_0) |
                                 CPSW_SET_BIT(IPC_MCU2_0) |
                                 CPSW_SET_BIT(IPC_MCU2_1) |
                                 CPSW_SET_BIT(IPC_MCU1_0)),
    .numEntries = 0,
};
#endif

#if defined (SOC_J721E)
static const CpswRm_IoctlPermissionTable gCpswAppIoctlPermission_9G =
{
    .defaultPermittedCoreMask = (CPSW_SET_BIT(IPC_MPU1_0) |
                                 CPSW_SET_BIT(IPC_MCU2_0) |
                                 CPSW_SET_BIT(IPC_MCU2_1) |
                                 CPSW_SET_BIT(IPC_MCU3_0) |
                                 CPSW_SET_BIT(IPC_MCU1_0)),
    .numEntries = 0,
};
#endif
/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

const CpswRm_ResourcePrms *CpswAppRm_getResourcePartitionInfo(Cpsw_Type cpswType)
{
    const CpswRm_ResourcePrms *rmInitPrms = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        {
            rmInitPrms = &gCpswAppRmDefCfg_2G;
            break;
        }

        case CPSW_5G:
        {
            rmInitPrms = &gCpswAppRmDefCfg_5G;
            break;
        }

        case CPSW_9G:
        {
            rmInitPrms = &gCpswAppRmDefCfg_9G;
            break;
        }

        default:
        {
            rmInitPrms = NULL;
            break;
        }
    }

    return(rmInitPrms);
}

const CpswRm_IoctlPermissionTable *CpswAppRm_getIoctlPermissionInfo(Cpsw_Type cpswType)
{
    const CpswRm_IoctlPermissionTable *ioctlPerm = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        {
            ioctlPerm = &gCpswAppIoctlPermission_2G;
            break;
        }

#if defined (SOC_J7200)
        case CPSW_5G:
        {
            ioctlPerm = &gCpswAppIoctlPermission_5G;
            break;
        }
#endif

#if defined (SOC_J721E)
        case CPSW_9G:
        {
            ioctlPerm = &gCpswAppIoctlPermission_9G;
            break;
        }
#endif

        default:
        {
            ioctlPerm = NULL;
            break;
        }
    }

    return(ioctlPerm);
}
