/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_apputils.c
 *
 * \brief    Common CPSW application utility used in all CPSW example.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <ti/csl/soc.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/osal/osal.h>
#include <ti/board/board.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/uart/UART.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/pm/pmlib.h>

#include "inc/cpsw_apputils.h"
#include "inc/cpsw_appmemutils.h"
#include "inc/cpsw_appmemutils_cfg.h"
#include "inc/cpsw_appboardutils.h"

#include "inc/cpsw_appsoc.h"
#include "inc/cpsw_apprm.h"

#include <ti/drv/cpsw/nimucpsw/ndk2cpsw_appif.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* UDMA driver objects */
static struct Udma_DrvObj udmaDrvObj[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

uint32_t CpswAppUtils_getNavSSInstanceId(Cpsw_Type cpswType)
{
    uint32_t instId;

#if defined(SOC_AM65XX)
    CpswAppUtils_assert(cpswType == CPSW_2G);

    /* Due to errata "i2006 - UDMA-P: UDMA-P Real-Time Remote Peer Registers
     * not Functional Across UDMA-P Domains" in AM65xx we access CPSW_2G via MCU
     * navss only */
    instId = UDMA_INST_ID_MCU_0;
#else
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU1_1)
    instId = UDMA_INST_ID_MCU_0;
#else
    instId = UDMA_INST_ID_MAIN_0;
#endif
#endif

    return instId;
}

Udma_DrvHandle CpswAppUtils_udmaOpen(Cpsw_Type cpswType,
                                     Udma_InitPrms *pInitPrms)
{
    Udma_InitPrms initPrms;
    Udma_DrvHandle hUdmaDrv;
    int32_t retVal;
    uint32_t instId;

    hUdmaDrv = &udmaDrvObj[cpswType];
    memset(hUdmaDrv, 0U, sizeof(*hUdmaDrv));

    /* Use default params if not passed by caller */
    if (NULL == pInitPrms)
    {
        instId = CpswAppUtils_getNavSSInstanceId(cpswType);
        /* Initialize the UDMA driver based on NAVSS instance */
        UdmaInitPrms_init(instId, &initPrms);
        initPrms.printFxn = (Udma_PrintFxn) & CpswAppUtils_print;
        retVal            = Udma_init(hUdmaDrv, &initPrms);
    }
    else
    {
        retVal = Udma_init(hUdmaDrv, pInitPrms);
    }

    /* assert if UDMA failed to open */
    CpswAppUtils_assert(UDMA_SOK == retVal);

    return hUdmaDrv;
}

void CpswAppUtils_udmaclose(Udma_DrvHandle hUdmaDrv)
{
    CpswAppUtils_assert(NULL != hUdmaDrv);
    /* App should make sure that all Rx & Tx channels & are closed */
    /* Deinitialize the UDMA driver */
    Udma_deinit(hUdmaDrv);
}

void CpswAppUtils_cacheWb(const void *addr,
                          int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

    if (isCacheCoherent == false)
    {
        CacheP_wb(addr, size);
    }
}

void CpswAppUtils_cacheInv(const void *addr,
                           int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

    if (isCacheCoherent == FALSE)
    {
        CacheP_Inv(addr, size);
    }
}

void CpswAppUtils_cacheWbInv(const void *addr,
                             int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

    if (isCacheCoherent == false)
    {
        CacheP_wbInv(addr, size);
    }
}

void CpswAppUtils_vprint(const char *pcString,
                         va_list args)
{
    char printBuffer[CPSW_CFG_PRINT_BUF_LEN];

    vsnprintf(printBuffer, sizeof(printBuffer), pcString, args);

#if defined(UART_ENABLED) && defined(CPSWAPPUTILS_UART_ALLOWED)
    UART_printf("%s", printBuffer);
#else
    if (TRUE == CpswAppUtils_isPrintSupported())
    {
        printf(printBuffer);
    }
#endif
}

void CpswAppUtils_print(const char *pcString,
                        ...)
{
    char printBuffer[CPSW_CFG_PRINT_BUF_LEN];
    va_list arguments;

#if (1U == CPSW_CFG_DEV_ERROR)
    if (CPSW_CFG_PRINT_BUF_LEN < strlen(pcString))
    {
        /* Don't use CpswAppUtils_assert as it uses CpswAppUtils_print function  */
        assert(false);
    }
#endif

    /* Start the varargs processing */
    va_start(arguments, pcString);
    vsnprintf(printBuffer, sizeof(printBuffer), pcString, arguments);

    {
#if defined(UART_ENABLED) && defined(CPSWAPPUTILS_UART_ALLOWED)
        UART_printf("%s", printBuffer);
#else
        if (TRUE == CpswAppUtils_isPrintSupported())
        {
            printf(printBuffer);
        }
#endif
    }

    /* End the varargs processing */
    va_end(arguments);
}

char CpswAppUtils_getChar(void)
{
    char ch;

#if defined(UART_ENABLED) && defined(CPSWAPPUTILS_UART_ALLOWED)
    ch = UART_getc();
#else
    scanf("%c", &ch);
#endif

    return ch;
}

int32_t CpswAppUtils_getNum(void)
{
    int32_t num;

#if defined(UART_ENABLED) && defined(CPSWAPPUTILS_UART_ALLOWED)
    UART_scanFmt("%d", &num);
#else
    scanf("%d", &num);
#endif

    return num;
}

uint32_t CpswAppUtils_randFxn(uint32_t min,
                              uint32_t max)
{
    return (rand() % (max - min + 1)) + min;
}

void CpswAppUtils_waitEmuConnect(void)
{
    volatile uint32_t emuWaitFlag = 0x1U;

    while (emuWaitFlag)
    {
        ;
    }
}

uint32_t CpswAppUtils_isPrintSupported(void)
{
    uint32_t retVal = TRUE;

#if defined(BUILD_MPU)
    retVal = FALSE;
#endif

    return(retVal);
}

static void CpswAppUtils_printStatsNonZero(const char *pcString,
                                           uint64_t statVal)
{
    if (0U != statVal)
    {
        CpswAppUtils_print(pcString, statVal);
    }
}

static void CpswAppUtils_printStatsWithIdxNonZero(const char *pcString,
                                                  uint32_t idx,
                                                  uint64_t statVal)
{
    if (0U != statVal)
    {
        CpswAppUtils_print(pcString, idx, statVal);
    }
}

void CpswAppUtils_printMacAddr(uint8_t macAddr[])
{
    CpswAppUtils_print("%02x:%02x:%02x:%02x:%02x:%02x\n",
                       macAddr[0] & 0xFF,
                       macAddr[1] & 0xFF,
                       macAddr[2] & 0xFF,
                       macAddr[3] & 0xFF,
                       macAddr[4] & 0xFF,
                       macAddr[5] & 0xFF);
}

void CpswAppUtils_printFrame(EthFrame *frame,
                             uint32_t len)
{
    uint8_t *payload;
    uint32_t i;

    CpswAppUtils_print("Dst addr : ");
    CpswAppUtils_printMacAddr(&frame->hdr.dstMac[0]);

    CpswAppUtils_print("Src addr : ");
    CpswAppUtils_printMacAddr(&frame->hdr.srcMac[0]);

    if (frame->hdr.etherType == htons(ETHERTYPE_VLAN_TAG))
    {
        EthVlanFrame *vlanFrame = (EthVlanFrame *)frame;

        CpswAppUtils_print("TPID     : 0x%04x\n", ntohs(vlanFrame->hdr.tpid) & 0xFFFFU);
        CpswAppUtils_print("Priority : %d\n", (ntohs(vlanFrame->hdr.tci) & 0xFFFFU) >> 13);
        CpswAppUtils_print("VLAN Id  : %d\n", ntohs(vlanFrame->hdr.tci) & 0xFFFU);
        CpswAppUtils_print("EtherType: 0x%04x\n", ntohs(vlanFrame->hdr.etherType) & 0xFFFFU);
        payload = vlanFrame->payload;
        len    -= ETH_VLAN_TAG_LEN;
    }
    else
    {
        CpswAppUtils_print("EtherType: 0x%04x\n", ntohs(frame->hdr.etherType) & 0xFFFFU);
        payload = frame->payload;
    }

    CpswAppUtils_print("Payload  : ");
    for (i = 0; i < len; i++)
    {
        CpswAppUtils_print("0x%02x ", payload[i]);
        if (i && (((i + 1) % OCTETS_PER_ROW) == 0))
        {
            CpswAppUtils_print("\n           ");
        }
    }

    if (len && ((len % OCTETS_PER_ROW) != 0))
    {
        CpswAppUtils_print("\n");
    }

    CpswAppUtils_print("\n");
}

void CpswAppUtils_printHostPortStats2G(CpswStats_HostPort_2g *st)
{
    CpswAppUtils_printStatsNonZero("  rxGoodFrames            = %llu\n", st->rxGoodFrames);
    CpswAppUtils_printStatsNonZero("  rxBcastFrames           = %llu\n", st->rxBcastFrames);
    CpswAppUtils_printStatsNonZero("  rxMcastFrames           = %llu\n", st->rxMcastFrames);
    CpswAppUtils_printStatsNonZero("  rxCrcErrors             = %llu\n", st->rxCrcErrors);
    CpswAppUtils_printStatsNonZero("  rxOversizedFrames       = %llu\n", st->rxOversizedFrames);
    CpswAppUtils_printStatsNonZero("  rxUndersizedFrames      = %llu\n", st->rxUndersizedFrames);
    CpswAppUtils_printStatsNonZero("  rxFragments             = %llu\n", st->rxFragments);
    CpswAppUtils_printStatsNonZero("  aleDrop                 = %llu\n", st->aleDrop);
    CpswAppUtils_printStatsNonZero("  aleOverrunDrop          = %llu\n", st->aleOverrunDrop);
    CpswAppUtils_printStatsNonZero("  rxOctets                = %llu\n", st->rxOctets);
    CpswAppUtils_printStatsNonZero("  txGoodFrames            = %llu\n", st->txGoodFrames);
    CpswAppUtils_printStatsNonZero("  txBcastFrames           = %llu\n", st->txBcastFrames);
    CpswAppUtils_printStatsNonZero("  txMcastFrames           = %llu\n", st->txMcastFrames);
    CpswAppUtils_printStatsNonZero("  txOctets                = %llu\n", st->txOctets);
    CpswAppUtils_printStatsNonZero("  octetsFrames64          = %llu\n", st->octetsFrames64);
    CpswAppUtils_printStatsNonZero("  octetsFrames65to127     = %llu\n", st->octetsFrames65to127);
    CpswAppUtils_printStatsNonZero("  octetsFrames128to255    = %llu\n", st->octetsFrames128to255);
    CpswAppUtils_printStatsNonZero("  octetsFrames256to511    = %llu\n", st->octetsFrames256to511);
    CpswAppUtils_printStatsNonZero("  octetsFrames512to1023   = %llu\n", st->octetsFrames512to1023);
    CpswAppUtils_printStatsNonZero("  octetsFrames1024        = %llu\n", st->octetsFrames1024);
    CpswAppUtils_printStatsNonZero("  netOctets               = %llu\n", st->netOctets);
    CpswAppUtils_printStatsNonZero("  rxBottomOfFifoDrop      = %llu\n", st->rxBottomOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  portMaskDrop            = %llu\n", st->portMaskDrop);
    CpswAppUtils_printStatsNonZero("  rxTopOfFifoDrop         = %llu\n", st->rxTopOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  aleRateLimitDrop        = %llu\n", st->aleRateLimitDrop);
    CpswAppUtils_printStatsNonZero("  aleVidIngressDrop       = %llu\n", st->aleVidIngressDrop);
    CpswAppUtils_printStatsNonZero("  aleDAEqSADrop           = %llu\n", st->aleDAEqSADrop);
    CpswAppUtils_printStatsNonZero("  aleBlockDrop            = %llu\n", st->aleBlockDrop);
    CpswAppUtils_printStatsNonZero("  aleSecureDrop           = %llu\n", st->aleSecureDrop);
    CpswAppUtils_printStatsNonZero("  aleAuthDrop             = %llu\n", st->aleAuthDrop);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcast         = %llu\n", st->aleUnknownUcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcastBcnt     = %llu\n", st->aleUnknownUcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcast         = %llu\n", st->aleUnknownMcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcastBcnt     = %llu\n", st->aleUnknownMcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcast         = %llu\n", st->aleUnknownBcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcastBcnt     = %llu\n", st->aleUnknownBcastBcnt);
    CpswAppUtils_printStatsNonZero("  alePolicyMatch          = %llu\n", st->alePolicyMatch);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchRed       = %llu\n", st->alePolicyMatchRed);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchYellow    = %llu\n", st->alePolicyMatchYellow);
    CpswAppUtils_printStatsNonZero("  txMemProtectError       = %llu\n", st->txMemProtectError);
}

void CpswAppUtils_printMacPortStats2G(CpswStats_MacPort_2g *st)
{
    uint32_t i;

    CpswAppUtils_printStatsNonZero("  rxGoodFrames            = %llu\n", st->rxGoodFrames);
    CpswAppUtils_printStatsNonZero("  rxBcastFrames           = %llu\n", st->rxBcastFrames);
    CpswAppUtils_printStatsNonZero("  rxMcastFrames           = %llu\n", st->rxMcastFrames);
    CpswAppUtils_printStatsNonZero("  rxPauseFrames           = %llu\n", st->rxPauseFrames);
    CpswAppUtils_printStatsNonZero("  rxCrcErrors             = %llu\n", st->rxCrcErrors);
    CpswAppUtils_printStatsNonZero("  rxAlignCodeErrors       = %llu\n", st->rxAlignCodeErrors);
    CpswAppUtils_printStatsNonZero("  rxOversizedFrames       = %llu\n", st->rxOversizedFrames);
    CpswAppUtils_printStatsNonZero("  rxJabberFrames          = %llu\n", st->rxJabberFrames);
    CpswAppUtils_printStatsNonZero("  rxUndersizedFrames      = %llu\n", st->rxUndersizedFrames);
    CpswAppUtils_printStatsNonZero("  rxFragments             = %llu\n", st->rxFragments);
    CpswAppUtils_printStatsNonZero("  aleDrop                 = %llu\n", st->aleDrop);
    CpswAppUtils_printStatsNonZero("  aleOverrunDrop          = %llu\n", st->aleOverrunDrop);
    CpswAppUtils_printStatsNonZero("  rxOctets                = %llu\n", st->rxOctets);
    CpswAppUtils_printStatsNonZero("  txGoodFrames            = %llu\n", st->txGoodFrames);
    CpswAppUtils_printStatsNonZero("  txBcastFrames           = %llu\n", st->txBcastFrames);
    CpswAppUtils_printStatsNonZero("  txMcastFrames           = %llu\n", st->txMcastFrames);
    CpswAppUtils_printStatsNonZero("  txPauseFrames           = %llu\n", st->txPauseFrames);
    CpswAppUtils_printStatsNonZero("  txDeferredFrames        = %llu\n", st->txDeferredFrames);
    CpswAppUtils_printStatsNonZero("  txCollisionFrames       = %llu\n", st->txCollisionFrames);
    CpswAppUtils_printStatsNonZero("  txSingleCollFrames      = %llu\n", st->txSingleCollFrames);
    CpswAppUtils_printStatsNonZero("  txMultipleCollFrames    = %llu\n", st->txMultipleCollFrames);
    CpswAppUtils_printStatsNonZero("  txExcessiveCollFrames   = %llu\n", st->txExcessiveCollFrames);
    CpswAppUtils_printStatsNonZero("  txLateCollFrames        = %llu\n", st->txLateCollFrames);
    CpswAppUtils_printStatsNonZero("  rxIPGError              = %llu\n", st->rxIPGError);
    CpswAppUtils_printStatsNonZero("  txCarrierSenseErrors    = %llu\n", st->txCarrierSenseErrors);
    CpswAppUtils_printStatsNonZero("  txOctets                = %llu\n", st->txOctets);
    CpswAppUtils_printStatsNonZero("  octetsFrames64          = %llu\n", st->octetsFrames64);
    CpswAppUtils_printStatsNonZero("  octetsFrames65to127     = %llu\n", st->octetsFrames65to127);
    CpswAppUtils_printStatsNonZero("  octetsFrames128to255    = %llu\n", st->octetsFrames128to255);
    CpswAppUtils_printStatsNonZero("  octetsFrames256to511    = %llu\n", st->octetsFrames256to511);
    CpswAppUtils_printStatsNonZero("  octetsFrames512to1023   = %llu\n", st->octetsFrames512to1023);
    CpswAppUtils_printStatsNonZero("  octetsFrames1024        = %llu\n", st->octetsFrames1024);
    CpswAppUtils_printStatsNonZero("  netOctets               = %llu\n", st->netOctets);
    CpswAppUtils_printStatsNonZero("  rxBottomOfFifoDrop      = %llu\n", st->rxBottomOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  portMaskDrop            = %llu\n", st->portMaskDrop);
    CpswAppUtils_printStatsNonZero("  rxTopOfFifoDrop         = %llu\n", st->rxTopOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  aleRateLimitDrop        = %llu\n", st->aleRateLimitDrop);
    CpswAppUtils_printStatsNonZero("  aleVidIngressDrop       = %llu\n", st->aleVidIngressDrop);
    CpswAppUtils_printStatsNonZero("  aleDAEqSADrop           = %llu\n", st->aleDAEqSADrop);
    CpswAppUtils_printStatsNonZero("  aleBlockDrop            = %llu\n", st->aleBlockDrop);
    CpswAppUtils_printStatsNonZero("  aleSecureDrop           = %llu\n", st->aleSecureDrop);
    CpswAppUtils_printStatsNonZero("  aleAuthDrop             = %llu\n", st->aleAuthDrop);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcast         = %llu\n", st->aleUnknownUcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcastBcnt     = %llu\n", st->aleUnknownUcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcast         = %llu\n", st->aleUnknownMcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcastBcnt     = %llu\n", st->aleUnknownMcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcast         = %llu\n", st->aleUnknownBcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcastBcnt     = %llu\n", st->aleUnknownBcastBcnt);
    CpswAppUtils_printStatsNonZero("  alePolicyMatch          = %llu\n", st->alePolicyMatch);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchRed       = %llu\n", st->alePolicyMatchRed);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchYellow    = %llu\n", st->alePolicyMatchYellow);
    CpswAppUtils_printStatsNonZero("  txMemProtectError       = %llu\n", st->txMemProtectError);

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPri); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPri[%u]                = %llu\n", i, st->txPri[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriBcnt[%u]            = %llu\n", i, st->txPriBcnt[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDrop); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDrop[%u]            = %llu\n", i, st->txPriDrop[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDropBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDropBcnt[%u]        = %llu\n", i, st->txPriDropBcnt[i]);
    }
}

void CpswAppUtils_printHostPortStats9G(CpswStats_HostPort_9g *st)
{
    uint_fast32_t i;

    CpswAppUtils_printStatsNonZero("  rxGoodFrames            = %llu\n", st->rxGoodFrames);
    CpswAppUtils_printStatsNonZero("  rxBcastFrames           = %llu\n", st->rxBcastFrames);
    CpswAppUtils_printStatsNonZero("  rxMcastFrames           = %llu\n", st->rxMcastFrames);
    CpswAppUtils_printStatsNonZero("  rxCrcErrors             = %llu\n", st->rxCrcErrors);
    CpswAppUtils_printStatsNonZero("  rxOversizedFrames       = %llu\n", st->rxOversizedFrames);
    CpswAppUtils_printStatsNonZero("  rxUndersizedFrames      = %llu\n", st->rxUndersizedFrames);
    CpswAppUtils_printStatsNonZero("  rxFragments             = %llu\n", st->rxFragments);
    CpswAppUtils_printStatsNonZero("  aleDrop                 = %llu\n", st->aleDrop);
    CpswAppUtils_printStatsNonZero("  aleOverrunDrop          = %llu\n", st->aleOverrunDrop);
    CpswAppUtils_printStatsNonZero("  rxOctets                = %llu\n", st->rxOctets);
    CpswAppUtils_printStatsNonZero("  txGoodFrames            = %llu\n", st->txGoodFrames);
    CpswAppUtils_printStatsNonZero("  txBcastFrames           = %llu\n", st->txBcastFrames);
    CpswAppUtils_printStatsNonZero("  txMcastFrames           = %llu\n", st->txMcastFrames);
    CpswAppUtils_printStatsNonZero("  txOctets                = %llu\n", st->txOctets);
    CpswAppUtils_printStatsNonZero("  octetsFrames64          = %llu\n", st->octetsFrames64);
    CpswAppUtils_printStatsNonZero("  octetsFrames65to127     = %llu\n", st->octetsFrames65to127);
    CpswAppUtils_printStatsNonZero("  octetsFrames128to255    = %llu\n", st->octetsFrames128to255);
    CpswAppUtils_printStatsNonZero("  octetsFrames256to511    = %llu\n", st->octetsFrames256to511);
    CpswAppUtils_printStatsNonZero("  octetsFrames512to1023   = %llu\n", st->octetsFrames512to1023);
    CpswAppUtils_printStatsNonZero("  octetsFrames1024        = %llu\n", st->octetsFrames1024);
    CpswAppUtils_printStatsNonZero("  netOctets               = %llu\n", st->netOctets);
    CpswAppUtils_printStatsNonZero("  rxBottomOfFifoDrop      = %llu\n", st->rxBottomOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  portMaskDrop            = %llu\n", st->portMaskDrop);
    CpswAppUtils_printStatsNonZero("  rxTopOfFifoDrop         = %llu\n", st->rxTopOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  aleRateLimitDrop        = %llu\n", st->aleRateLimitDrop);
    CpswAppUtils_printStatsNonZero("  aleVidIngressDrop       = %llu\n", st->aleVidIngressDrop);
    CpswAppUtils_printStatsNonZero("  aleDAEqSADrop           = %llu\n", st->aleDAEqSADrop);
    CpswAppUtils_printStatsNonZero("  aleBlockDrop            = %llu\n", st->aleBlockDrop);
    CpswAppUtils_printStatsNonZero("  aleSecureDrop           = %llu\n", st->aleSecureDrop);
    CpswAppUtils_printStatsNonZero("  aleAuthDrop             = %llu\n", st->aleAuthDrop);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcast         = %llu\n", st->aleUnknownUcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcastBcnt     = %llu\n", st->aleUnknownUcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcast         = %llu\n", st->aleUnknownMcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcastBcnt     = %llu\n", st->aleUnknownMcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcast         = %llu\n", st->aleUnknownBcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcastBcnt     = %llu\n", st->aleUnknownBcastBcnt);
    CpswAppUtils_printStatsNonZero("  alePolicyMatch          = %llu\n", st->alePolicyMatch);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchRed       = %llu\n", st->alePolicyMatchRed);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchYellow    = %llu\n", st->alePolicyMatchYellow);
    CpswAppUtils_printStatsNonZero("  aleMultSADrop           = %llu\n", st->aleMultSADrop);
    CpswAppUtils_printStatsNonZero("  aleDualVlanDrop         = %llu\n", st->aleDualVlanDrop);
    CpswAppUtils_printStatsNonZero("  aleLenErrorDrop         = %llu\n", st->aleLenErrorDrop);
    CpswAppUtils_printStatsNonZero("  aleIpNextHdrDrop        = %llu\n", st->aleIpNextHdrDrop);
    CpswAppUtils_printStatsNonZero("  aleIPv4FragDrop         = %llu\n", st->aleIPv4FragDrop);
    CpswAppUtils_printStatsNonZero("  ietRxAssemblyErr        = %llu\n", st->ietRxAssemblyErr);
    CpswAppUtils_printStatsNonZero("  ietRxAssemblyOk         = %llu\n", st->ietRxAssemblyOk);
    CpswAppUtils_printStatsNonZero("  ietRxSmdError           = %llu\n", st->ietRxSmdError);
    CpswAppUtils_printStatsNonZero("  ietRxFrag               = %llu\n", st->ietRxFrag);
    CpswAppUtils_printStatsNonZero("  ietTxHold               = %llu\n", st->ietTxHold);
    CpswAppUtils_printStatsNonZero("  ietTxFrag               = %llu\n", st->ietTxFrag);
    CpswAppUtils_printStatsNonZero("  txMemProtectError       = %llu\n", st->txMemProtectError);

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPri); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPri[%u]                = %llu\n", i, st->txPri[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriBcnt[%u]            = %llu\n", i, st->txPriBcnt[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDrop); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDrop[%u]            = %llu\n", i, st->txPriDrop[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDropBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDropBcnt[%u]        = %llu\n", i, st->txPriDropBcnt[i]);
    }
}

void CpswAppUtils_printMacPortStats9G(CpswStats_MacPort_9g *st)
{
    uint_fast32_t i;

    CpswAppUtils_printStatsNonZero("  rxGoodFrames            = %llu\n", st->rxGoodFrames);
    CpswAppUtils_printStatsNonZero("  rxBcastFrames           = %llu\n", st->rxBcastFrames);
    CpswAppUtils_printStatsNonZero("  rxMcastFrames           = %llu\n", st->rxMcastFrames);
    CpswAppUtils_printStatsNonZero("  rxPauseFrames           = %llu\n", st->rxPauseFrames);
    CpswAppUtils_printStatsNonZero("  rxCrcErrors             = %llu\n", st->rxCrcErrors);
    CpswAppUtils_printStatsNonZero("  rxAlignCodeErrors       = %llu\n", st->rxAlignCodeErrors);
    CpswAppUtils_printStatsNonZero("  rxOversizedFrames       = %llu\n", st->rxOversizedFrames);
    CpswAppUtils_printStatsNonZero("  rxJabberFrames          = %llu\n", st->rxJabberFrames);
    CpswAppUtils_printStatsNonZero("  rxUndersizedFrames      = %llu\n", st->rxUndersizedFrames);
    CpswAppUtils_printStatsNonZero("  rxFragments             = %llu\n", st->rxFragments);
    CpswAppUtils_printStatsNonZero("  aleDrop                 = %llu\n", st->aleDrop);
    CpswAppUtils_printStatsNonZero("  aleOverrunDrop          = %llu\n", st->aleOverrunDrop);
    CpswAppUtils_printStatsNonZero("  rxOctets                = %llu\n", st->rxOctets);
    CpswAppUtils_printStatsNonZero("  txGoodFrames            = %llu\n", st->txGoodFrames);
    CpswAppUtils_printStatsNonZero("  txBcastFrames           = %llu\n", st->txBcastFrames);
    CpswAppUtils_printStatsNonZero("  txMcastFrames           = %llu\n", st->txMcastFrames);
    CpswAppUtils_printStatsNonZero("  txPauseFrames           = %llu\n", st->txPauseFrames);
    CpswAppUtils_printStatsNonZero("  txDeferredFrames        = %llu\n", st->txDeferredFrames);
    CpswAppUtils_printStatsNonZero("  txCollisionFrames       = %llu\n", st->txCollisionFrames);
    CpswAppUtils_printStatsNonZero("  txSingleCollFrames      = %llu\n", st->txSingleCollFrames);
    CpswAppUtils_printStatsNonZero("  txMultipleCollFrames    = %llu\n", st->txMultipleCollFrames);
    CpswAppUtils_printStatsNonZero("  txExcessiveCollFrames   = %llu\n", st->txExcessiveCollFrames);
    CpswAppUtils_printStatsNonZero("  txLateCollFrames        = %llu\n", st->txLateCollFrames);
    CpswAppUtils_printStatsNonZero("  rxIPGError              = %llu\n", st->rxIPGError);
    CpswAppUtils_printStatsNonZero("  txCarrierSenseErrors    = %llu\n", st->txCarrierSenseErrors);
    CpswAppUtils_printStatsNonZero("  txOctets                = %llu\n", st->txOctets);
    CpswAppUtils_printStatsNonZero("  octetsFrames64          = %llu\n", st->octetsFrames64);
    CpswAppUtils_printStatsNonZero("  octetsFrames65to127     = %llu\n", st->octetsFrames65to127);
    CpswAppUtils_printStatsNonZero("  octetsFrames128to255    = %llu\n", st->octetsFrames128to255);
    CpswAppUtils_printStatsNonZero("  octetsFrames256to511    = %llu\n", st->octetsFrames256to511);
    CpswAppUtils_printStatsNonZero("  octetsFrames512to1023   = %llu\n", st->octetsFrames512to1023);
    CpswAppUtils_printStatsNonZero("  octetsFrames1024        = %llu\n", st->octetsFrames1024);
    CpswAppUtils_printStatsNonZero("  netOctets               = %llu\n", st->netOctets);
    CpswAppUtils_printStatsNonZero("  rxBottomOfFifoDrop      = %llu\n", st->rxBottomOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  portMaskDrop            = %llu\n", st->portMaskDrop);
    CpswAppUtils_printStatsNonZero("  rxTopOfFifoDrop         = %llu\n", st->rxTopOfFifoDrop);
    CpswAppUtils_printStatsNonZero("  aleRateLimitDrop        = %llu\n", st->aleRateLimitDrop);
    CpswAppUtils_printStatsNonZero("  aleVidIngressDrop       = %llu\n", st->aleVidIngressDrop);
    CpswAppUtils_printStatsNonZero("  aleDAEqSADrop           = %llu\n", st->aleDAEqSADrop);
    CpswAppUtils_printStatsNonZero("  aleBlockDrop            = %llu\n", st->aleBlockDrop);
    CpswAppUtils_printStatsNonZero("  aleSecureDrop           = %llu\n", st->aleSecureDrop);
    CpswAppUtils_printStatsNonZero("  aleAuthDrop             = %llu\n", st->aleAuthDrop);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcast         = %llu\n", st->aleUnknownUcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownUcastBcnt     = %llu\n", st->aleUnknownUcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcast         = %llu\n", st->aleUnknownMcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownMcastBcnt     = %llu\n", st->aleUnknownMcastBcnt);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcast         = %llu\n", st->aleUnknownBcast);
    CpswAppUtils_printStatsNonZero("  aleUnknownBcastBcnt     = %llu\n", st->aleUnknownBcastBcnt);
    CpswAppUtils_printStatsNonZero("  alePolicyMatch          = %llu\n", st->alePolicyMatch);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchRed       = %llu\n", st->alePolicyMatchRed);
    CpswAppUtils_printStatsNonZero("  alePolicyMatchYellow    = %llu\n", st->alePolicyMatchYellow);
    CpswAppUtils_printStatsNonZero("  aleMultSADrop           = %llu\n", st->aleMultSADrop);
    CpswAppUtils_printStatsNonZero("  aleDualVlanDrop         = %llu\n", st->aleDualVlanDrop);
    CpswAppUtils_printStatsNonZero("  aleLenErrorDrop         = %llu\n", st->aleLenErrorDrop);
    CpswAppUtils_printStatsNonZero("  aleIpNextHdrDrop        = %llu\n", st->aleIpNextHdrDrop);
    CpswAppUtils_printStatsNonZero("  aleIPv4FragDrop         = %llu\n", st->aleIPv4FragDrop);
    CpswAppUtils_printStatsNonZero("  ietRxAssemblyErr        = %llu\n", st->ietRxAssemblyErr);
    CpswAppUtils_printStatsNonZero("  ietRxAssemblyOk         = %llu\n", st->ietRxAssemblyOk);
    CpswAppUtils_printStatsNonZero("  ietRxSmdError           = %llu\n", st->ietRxSmdError);
    CpswAppUtils_printStatsNonZero("  ietRxFrag               = %llu\n", st->ietRxFrag);
    CpswAppUtils_printStatsNonZero("  ietTxHold               = %llu\n", st->ietTxHold);
    CpswAppUtils_printStatsNonZero("  ietTxFrag               = %llu\n", st->ietTxFrag);
    CpswAppUtils_printStatsNonZero("  txMemProtectError       = %llu\n", st->txMemProtectError);

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPri); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPri[%u]                = %llu\n", i, st->txPri[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriBcnt[%u]            = %llu\n", i, st->txPriBcnt[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDrop); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDrop[%u]            = %llu\n", i, st->txPriDrop[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDropBcnt); i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("  txPriDropBcnt[%u]        = %llu\n", i, st->txPriDropBcnt[i]);
    }
}

void CpswAppUtils_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                  CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitRxPackets(hRxFlow, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswAppUtils_print("CpswDma_submitRxPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        /* Copy unused packets to back to readyQ */
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }
}

int32_t CpswAppUtils_submitTxPackets(CpswDma_TxChHandle hTxCh,
                                     CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitTxReadyPackets(hTxCh, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswAppUtils_print("CpswDma_submitTxReadyPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }

    return retVal;
}

void CpswAppUtils_validatePacketState(CpswDma_PktInfoQ *pQueue,
                                      uint32_t expectedState,
                                      uint32_t newState)
{
    uint32_t i;
    CpswDma_PktInfo *pktInfo = (CpswDma_PktInfo *)pQueue->pHead;

    for (i = 0; i < CpswUtils_getQCount(pQueue); i++)
    {
        CpswUtils_checkPktState(&pktInfo->pktState,
                                CPSW_PKTSTATE_MODULE_APP,
                                expectedState,
                                newState);
        pktInfo = (CpswDma_PktInfo *)pktInfo->node.pNext;
    }
}

void CpswAppUtils_freePktInfoQ(CpswDma_PktInfoQ *pPktInfoQ)
{
    CpswDma_PktInfo *pktInfo;
    uint32_t pktCnt, i;

    pktCnt = CpswUtils_getQCount(pPktInfoQ);
    /* Free all retrieved packets from DMA */
    for (i = 0U; i < pktCnt; i++)
    {
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(pPktInfoQ);
        CpswAppMemUtils_freeEthPktFxn(pktInfo);
    }
}

uint64_t CpswAppUtils_virtToPhyFxn(const void *virtAddr,
                                   void *appData)
{
    return((uint64_t)virtAddr);
}

void *CpswAppUtils_phyToVirtFxn(uint64_t phyAddr,
                                void *appData)
{
#if defined(__aarch64__)
    uint64_t temp = phyAddr;
#else
    /* R5 is 32-bit machine, need to truncate to avoid void * typecast error */
    uint32_t temp = (uint32_t)phyAddr;
#endif
    return((void *)temp);
}

/* Don't use CpswAppUtils_assert or print functions as UART might not be
 * initialized when this function is called.  */
void CpswAppUtils_setDeviceState(uint32_t moduleId,
                                 uint32_t requiredState,
                                 uint32_t appFlags)
{
    int32_t status;
    uint32_t moduleState      = 0U;
    uint32_t resetState       = 0U;
    uint32_t contextLossState = 0U;
    bool turnOn;
    bool turnOff;

    /* Get the module state */
    status = Sciclient_pmGetModuleState(moduleId,
                                        &moduleState,
                                        &resetState,
                                        &contextLossState,
                                        SCICLIENT_SERVICE_WAIT_FOREVER);
    assert(status == CSL_PASS);

    /* No need to change if already in the requested state. Note that
     * getter (HW_STATE_*) and setter (SW_STATE_*) can't be directly compared */
    turnOn = (moduleState == TISCI_MSG_VALUE_DEVICE_HW_STATE_OFF) &&
             (requiredState == TISCI_MSG_VALUE_DEVICE_SW_STATE_ON);
    turnOff = (moduleState == TISCI_MSG_VALUE_DEVICE_HW_STATE_ON) &&
              (requiredState == TISCI_MSG_VALUE_DEVICE_SW_STATE_AUTO_OFF);

    /* We assert in this function upon failure as this is first function called
     * from the app to enable clocks and app can't recover from this failure */
    if (turnOn || turnOff)
    {
        status = Sciclient_pmSetModuleState(moduleId,
                                            requiredState,
                                            (appFlags |
                                             TISCI_MSG_FLAG_AOP |
                                             TISCI_MSG_FLAG_DEVICE_RESET_ISO),
                                            SCICLIENT_SERVICE_WAIT_FOREVER);
        assert(status == CSL_PASS);

        if (requiredState == TISCI_MSG_VALUE_DEVICE_SW_STATE_ON)
        {
            /* Reset if changed state to enabled */
            status = Sciclient_pmSetModuleRst(moduleId,
                                              0x0U /*resetBit*/,
                                              SCICLIENT_SERVICE_WAIT_FOREVER);
            assert(status == CSL_PASS);
        }
    }

    /* Dummy assignment to avoid -Werror=unused-but-set-variable error. The GCC compiler
     * doesn't consider usage of variable in assert */
    status = status;
}

void CpswAppUtils_clkRateSetState(uint32_t moduleId,
                                     uint32_t clockId,
                                     uint32_t additionalFlag,
                                     uint32_t state)
{
    int32_t status;

    status = Sciclient_pmModuleClkRequest
                 (moduleId,
                 clockId,
                 state,
                 additionalFlag,
                 SCICLIENT_SERVICE_WAIT_FOREVER);
    if (status != CSL_PASS)
    {
        CpswAppUtils_print("Sciclient_pmModuleClkRequest failed for module=%d, clockId=%d\n",
                           moduleId, clockId);
        CpswAppUtils_assert(false);
    }
}

void CpswAppUtils_clkRateSet(uint32_t moduleId,
                             uint32_t clkId,
                             uint64_t clkRateHz)
{
    int32_t status;
    uint64_t currClkFreqHz;

    status = PMLIBClkRateGet(moduleId, clkId, &currClkFreqHz);
    if ((status == CSL_PASS) &&
        (currClkFreqHz != clkRateHz))
    {
        status = PMLIBClkRateSet(moduleId, clkId, clkRateHz);
        if (status != CSL_PASS)
        {
            CpswAppUtils_print("PMLIBClkRateSet failed for clock Id = %d\n", clkId);
            CpswAppUtils_assert(false);
        }
    }
}

static void CpswAppUtils_unlockMmr(volatile uint32_t *kick0,
                                   volatile uint32_t *kick1)
{
    uint32_t kick0Val;

    /* if kick lock registers are locked*/
    if ((CSL_REG32_RD(kick0) & 0x1U) == 0U)
    {
        /* unlock the partition by writing the unlock values to the kick lock registers*/
        CSL_REG32_WR(kick0, MMR_KICK0_UNLOCK_VAL);
        CSL_REG32_WR(kick1, MMR_KICK1_UNLOCK_VAL);
    }

    kick0Val = CSL_REG32_FEXT(kick0, MMR_KICK_UNLOCKED);

    /* check to see if unlock bit in lock0 register is set */
    if (1U != kick0Val)
    {
        CpswAppUtils_assert(false);
    }
}

static void CpswAppUtils_lockMmr(volatile uint32_t *kick0,
                                 volatile uint32_t *kick1)
{
    uint32_t kick0Val;

    /* check to see if either of the kick registers are unlocked.*/
    if ((CSL_REG32_RD(kick0) & 0x1) == 1U)
    {
        /* write the kick lock value to the kick lock registers to lock the partition*/
        CSL_REG32_WR(kick0, MMR_KICK_LOCK_VAL);
        CSL_REG32_WR(kick1, MMR_KICK_LOCK_VAL);
    }

    /* check to see if unlock bit in lock0 register is set */
    kick0Val = CSL_REG32_FEXT(kick0, MMR_KICK_UNLOCKED);

    if (0U != kick0Val)
    {
        CpswAppUtils_assert(false);
    }
}

void CpswAppUtils_mcuMmrCtrl(CpswAppUtils_CtrlMmrType mmrNum,
                             CpswAppUtils_MmrLockType lock)
{
    CSL_mcu_ctrl_mmr_cfg0Regs *regs =
        (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;
    volatile uint32_t *kick0 = NULL, *kick1 = NULL;

    switch (mmrNum)
    {
        case CPSWAPPUTILS_MMR_LOCK0:
            kick0 = &regs->LOCK0_KICK0;
            kick1 = &regs->LOCK0_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK1:
            kick0 = &regs->LOCK1_KICK0;
            kick1 = &regs->LOCK1_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK2:
            kick0 = &regs->LOCK2_KICK0;
            kick1 = &regs->LOCK2_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK3:
            kick0 = &regs->LOCK3_KICK0;
            kick1 = &regs->LOCK3_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK4:
            kick0 = &regs->LOCK4_KICK0;
            kick1 = &regs->LOCK4_KICK1;
            break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

    if ((NULL != kick0) && (NULL != kick1))
    {
        if (CPSWAPPUTILS_UNLOCK_MMR == lock)
        {
            CpswAppUtils_unlockMmr(kick0, kick1);
        }
        else
        {
            CpswAppUtils_lockMmr(kick0, kick1);
        }
    }
}

void CpswAppUtils_mainMmrCtrl(CpswAppUtils_CtrlMmrType mmrNum,
                              CpswAppUtils_MmrLockType lock)
{
    CSL_main_ctrl_mmr_cfg0Regs *regs =
        (CSL_main_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_CTRL_MMR0_CFG0_BASE;
    volatile uint32_t *kick0 = NULL, *kick1 = NULL;

    switch (mmrNum)
    {
        case CPSWAPPUTILS_MMR_LOCK0:
            kick0 = &regs->LOCK0_KICK0;
            kick1 = &regs->LOCK0_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK1:
            kick0 = &regs->LOCK1_KICK0;
            kick1 = &regs->LOCK1_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK2:
            kick0 = &regs->LOCK2_KICK0;
            kick1 = &regs->LOCK2_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK3:
            kick0 = &regs->LOCK3_KICK0;
            kick1 = &regs->LOCK3_KICK1;
            break;

        case CPSWAPPUTILS_MMR_LOCK4:
            kick0 = &regs->LOCK4_KICK0;
            kick1 = &regs->LOCK4_KICK1;
            break;

#if defined (SOC_J721E) || defined (SOC_J7200)
        case CPSWAPPUTILS_MMR_LOCK5:
            kick0 = &regs->LOCK5_KICK0;
            kick1 = &regs->LOCK5_KICK1;
            break;
#endif

#if defined (SOC_J721E) || defined (SOC_AM65XX)
        case CPSWAPPUTILS_MMR_LOCK6:
            kick0 = &regs->LOCK6_KICK0;
            kick1 = &regs->LOCK6_KICK1;
            break;
#endif

        case CPSWAPPUTILS_MMR_LOCK7:
            kick0 = &regs->LOCK7_KICK0;
            kick1 = &regs->LOCK7_KICK1;
            break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

    if ((NULL != kick0) && (NULL != kick1))
    {
        if (CPSWAPPUTILS_UNLOCK_MMR == lock)
        {
            CpswAppUtils_unlockMmr(kick0, kick1);
        }
        else
        {
            CpswAppUtils_lockMmr(kick0, kick1);
        }
    }

}

void CpswAppUtils_enableClkOut(Cpsw_Type cpswType,
                               CpswAppUtils_ClkOutFreqType clkOut)
{
    uint32_t clkVal;

    if (CPSWAPPUTILS_CLKOUT_FREQ_50MHZ == clkOut)
    {
        clkVal = 0U;
    }
    else
    {
        /* 25 MHz */
        clkVal = 1U;
    }

    switch (cpswType)
    {
        case CPSW_2G:
        {
            CSL_mcu_ctrl_mmr_cfg0Regs *mcuRegs;
            mcuRegs = (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;

            CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);
            CSL_REG32_FINS(&mcuRegs->MCU_CLKOUT0_CTRL,
                           MCU_CTRL_MMR_CFG0_MCU_CLKOUT0_CTRL_CLK_SEL,
                           clkVal);
            CSL_REG32_FINS(&mcuRegs->MCU_CLKOUT0_CTRL,
                           MCU_CTRL_MMR_CFG0_MCU_CLKOUT0_CTRL_CLK_EN,
                           1U);
            CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
        }
        break;

#if defined (SOC_J721E) || defined (SOC_J7200)
    case CPSW_5G:
    case CPSW_9G:
        {
            CSL_main_ctrl_mmr_cfg0Regs *mainRegs;
            mainRegs = (CSL_main_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_CTRL_MMR0_CFG0_BASE;

            CpswAppUtils_mainMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);
            CSL_REG32_FINS(&mainRegs->CLKOUT_CTRL,
                           MAIN_CTRL_MMR_CFG0_CLKOUT_CTRL_CLK_SEL,
                           clkVal);
            CSL_REG32_FINS(&mainRegs->CLKOUT_CTRL,
                           MAIN_CTRL_MMR_CFG0_CLKOUT_CTRL_CLK_EN,
                           1U);
            CpswAppUtils_mainMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
        }
        break;
#endif
        default:
            /* Assert if case new mac mode gets added - ex. for SGMII clock config */
            CpswAppUtils_assert(false);
            break;
    }

}

static void CpswAppUtils_selectCptsClock(Cpsw_Type cpswType,
                                         CpswAppUtils_CptsClkSelMux clkSelMux)
{
    uint32_t muxVal;

    muxVal = (uint32_t) clkSelMux;

#if defined(SOC_AM65XX)
    switch (cpswType)
    {
        case CPSW_2G:
        {
            CSL_cptsRegs *regs = NULL;

            regs = (CSL_cptsRegs *)(uintptr_t)(CSL_MCU_CPSW0_NUSS_BASE +
                                               CPSW_CPTS_OFFSET);
            CSL_CPTS_setRFTCLKSelectReg(regs, muxVal);
        }
        break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

#elif defined (SOC_J721E) || defined (SOC_J7200)
    switch (cpswType)
    {
        case CPSW_2G:
        {
            CSL_mcu_ctrl_mmr_cfg0Regs *mcuRegs;
            mcuRegs = (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;

            CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);
            CSL_REG32_FINS(&mcuRegs->MCU_ENET_CLKSEL,
                           MCU_CTRL_MMR_CFG0_MCU_ENET_CLKSEL_CPTS_CLKSEL,
                           muxVal);
            CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
        }
        break;

#if defined (SOC_J721E) || defined (SOC_J7200)
    case CPSW_5G:
    case CPSW_9G:
        {
            CSL_main_ctrl_mmr_cfg0Regs *mainRegs;
            mainRegs = (CSL_main_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_CTRL_MMR0_CFG0_BASE;

            CpswAppUtils_mainMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);
            CSL_REG32_FINS(&mainRegs->CPSW_CLKSEL,
                           MAIN_CTRL_MMR_CFG0_CPSW_CLKSEL_CPTS_CLKSEL,
                           muxVal);
            CpswAppUtils_mainMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
        }
        break;
#endif
        default:
            CpswAppUtils_assert(false);
            break;
    }
#else
#error "Unsupported platform"
#endif
}

void CpswAppUtils_enableClocks(Cpsw_Type cpswType)
{
    const CpswSoc_Config *pSocCfg;
    uint32_t moduleId;
    uint32_t appFlags = 0U;

    /* Verify CPSW main clock frequency in SoC config is valid */
    pSocCfg = CpswSoc_getConfig(cpswType);
    CpswAppUtils_assert(pSocCfg->cppiClkFreqHz != 0U);
    moduleId = pSocCfg->dmscDevIdCpsw;

    CpswAppUtils_print("Enabling clocks for CPSW!\n");

    switch (cpswType)
    {
        case CPSW_2G:
            break;

#if defined (SOC_J721E) || defined (SOC_J7200)
        case CPSW_5G:
        case CPSW_9G:
            /* mcu2_0 as a owner of CPSW9G/CPSW5G needs exclusive access, preventing clock
             * changes by any other cores */
            appFlags = TISCI_MSG_FLAG_DEVICE_EXCLUSIVE;
            break;
#endif
        default:
            CpswAppUtils_assert(false);
    }

    CpswAppUtils_clkRateSet(moduleId, pSocCfg->clocks.cppiClkId, pSocCfg->cppiClkFreqHz);
    CpswAppUtils_clkRateSet(moduleId, pSocCfg->clocks.rgmii250MHzClkId, CPSW_SOC_RGMII_MHZ_250_CLK_VAL);
    CpswAppUtils_clkRateSet(moduleId, pSocCfg->clocks.rgmii50MHzClkId, CPSW_SOC_RGMII_MHZ_50_CLK_VAL);
    CpswAppUtils_clkRateSet(moduleId, pSocCfg->clocks.rgmii5MHzClkId, CPSW_SOC_RGMII_MHZ_5_CLK_VAL);

    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_ON, appFlags);

#if defined(SOC_AM65XX)
    CpswAppUtils_selectCptsClock(cpswType, CPSWAPPUTILS_CPTS_CLKSEL_CPSWHSDIV_CLKOUT2);
#elif defined (SOC_J721E) || defined (SOC_J7200)
    CpswAppUtils_selectCptsClock(cpswType, CPSWAPPUTILS_CPTS_CLKSEL_MAIN_SYSCLK0);
#endif
}

void CpswAppUtils_disableClocks(Cpsw_Type cpswType)
{
    uint32_t moduleId;
    const CpswSoc_Config *pSocCfg;
    uint32_t appFlags = 0U;

    /* Verify CPSW main clock frequency in SoC config is valid */
    pSocCfg = CpswSoc_getConfig(cpswType);
    CpswAppUtils_assert(pSocCfg->cppiClkFreqHz != 0U);
    moduleId = pSocCfg->dmscDevIdCpsw;

    CpswAppUtils_print("Disabling clocks for CPSW!\n");

    switch (cpswType)
    {
    case CPSW_2G:
        break;

#if defined (SOC_J721E) || defined (SOC_J7200)
    case CPSW_5G:
    case CPSW_9G:
            /* mcu2_0 as a owner of CPSW9G/CPSW5G needs exclusive access, preventing clock
             * changes by any other cores */
            appFlags = TISCI_MSG_FLAG_DEVICE_EXCLUSIVE;
            break;
#endif
        default:
            CpswAppUtils_assert(false);
    }

    /* set clock set to auto so when module is disabled, clocks shuts off */
    CpswAppUtils_clkRateSetState(moduleId, pSocCfg->clocks.cppiClkId, 0, TISCI_MSG_VALUE_CLOCK_SW_STATE_AUTO);
    CpswAppUtils_clkRateSetState(moduleId,
                                 pSocCfg->clocks.rgmii250MHzClkId,
                                 0,
                                 TISCI_MSG_VALUE_CLOCK_SW_STATE_AUTO);
    CpswAppUtils_clkRateSetState(moduleId,
                                 pSocCfg->clocks.rgmii50MHzClkId,
                                 0,
                                 TISCI_MSG_VALUE_CLOCK_SW_STATE_AUTO);
    CpswAppUtils_clkRateSetState(moduleId,
                                 pSocCfg->clocks.rgmii5MHzClkId,
                                 0,
                                 TISCI_MSG_VALUE_CLOCK_SW_STATE_AUTO);
    /* Set module set to HW AUTO */
    CpswAppUtils_setDeviceState(moduleId, TISCI_MSG_VALUE_DEVICE_SW_STATE_AUTO_OFF, appFlags);
}

int32_t CpswAppUtils_allocRxFlow(Cpsw_Handle hCpsw,
                                 uint32_t coreKey,
                                 uint32_t coreId,
                                 uint32_t *rxFlowStartIdx,
                                 uint32_t *flowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_AllocRxFlowOutArgs allocRxFlowOutArgs;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, &allocRxFlowOutArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_RX_FLOW,
                        &prms);

    if (status == CPSW_SOK)
    {
        *rxFlowStartIdx = allocRxFlowOutArgs.startIdx;
        *flowIdx        = allocRxFlowOutArgs.flowIdx;
    }
    else
    {
        CpswAppUtils_print("CpswAppUtils_allocRxFlow() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_allocTxCh(Cpsw_Handle hCpsw,
                               uint32_t coreKey,
                               uint32_t coreId,
                               uint32_t *txPSILThreadId)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    /* Allocate Tx Ch */
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, txPSILThreadId);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_TX_CH_PEERID,
                        &prms);
    if (status != CPSW_SOK)
    {
        *txPSILThreadId = CPSW_RM_TXCHNUM_INVALID;
        CpswAppUtils_print("CpswAppUtils_allocTxCh() failed: %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_allocMac(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              uint8_t *macAddress)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_AllocMacAddrOutArgs allocMacOutArgs;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, &allocMacOutArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_MAC_ADDR,
                        &prms);

    if (status == CPSW_SOK)
    {
        memcpy(macAddress, allocMacOutArgs.macAddr, sizeof(allocMacOutArgs.macAddr));
    }
    else
    {
        CpswAppUtils_print("CpswAppUtils_allocMac() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_freeRxFlow(Cpsw_Handle hCpsw,
                                uint32_t coreKey,
                                uint32_t coreId,
                                uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeRxFlowInArgs freeRxFlowInArgs;

    /*Free Rx Flow*/
    freeRxFlowInArgs.coreKey = coreKey;
    freeRxFlowInArgs.flowIdx = rxFlowIdx;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeRxFlowInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_RX_FLOW,
                        &prms);

    return status;
}

int32_t CpswAppUtils_freeTxCh(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              uint32_t txChNum)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeTxChInArgs freeTxChInArgs;

    /* Release Tx Ch */
    freeTxChInArgs.coreKey = coreKey;
    freeTxChInArgs.txChNum = txChNum;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeTxChInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_TX_CH_PEERID,
                        &prms);

    return status;
}

int32_t CpswAppUtils_freeMac(Cpsw_Handle hCpsw,
                             uint32_t coreKey,
                             uint32_t coreId,
                             uint8_t *macAddress)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeMacAddrInArgs freeMacInArgs;

    freeMacInArgs.coreKey = coreKey;
    CPSW_UTILS_ARRAY_COPY(freeMacInArgs.macAddr, macAddress);

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeMacInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_MAC_ADDR,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_freeMac() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_registerDefaultRxFlow(Cpsw_Handle hCpsw,
                                           uint32_t coreKey,
                                           uint32_t coreId,
                                           uint32_t rxFlowStartIdx,
                                           uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_DefaultFlowInfo defaultFlow;

    defaultFlow.coreKey  = coreKey;
    defaultFlow.startIdx = rxFlowStartIdx;
    defaultFlow.flowIdx  = rxFlowIdx;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &defaultFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_REGISTER_RX_DEFAULT_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_registerDefaultRxFlow() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_unregisterDefaultRxFlow(Cpsw_Handle hCpsw,
                                             uint32_t coreKey,
                                             uint32_t coreId,
                                             uint32_t rxFlowStartIdx,
                                             uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_DefaultFlowInfo defaultFlow;

    defaultFlow.coreKey  = coreKey;
    defaultFlow.startIdx = rxFlowStartIdx;
    defaultFlow.flowIdx  = rxFlowIdx;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &defaultFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_UNREGISTER_RX_DEFAULT_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_unregisterDefaultRxFlow() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_registerDstMacRxFlow(Cpsw_Handle hCpsw,
                                          uint32_t coreKey,
                                          uint32_t coreId,
                                          uint32_t rxFlowStartIdx,
                                          uint32_t rxFlowIdx,
                                          uint8_t macAddress[CPSW_MAC_ADDR_LEN])
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_MacDstFlowInfo macDstFlow;

    macDstFlow.coreKey  = coreKey;
    macDstFlow.startIdx = rxFlowStartIdx;
    macDstFlow.flowIdx  = rxFlowIdx;
    CPSW_UTILS_ARRAY_COPY(macDstFlow.macAddress, macAddress);
    CPSW_IOCTL_SET_IN_ARGS(&prms, &macDstFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_REGISTER_DSTMAC_RX_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_registerDstMacRxFlow() failed : %d\n", status);
    }

    return status;
}

int32_t CpswAppUtils_unregisterDstMacRxFlow(Cpsw_Handle hCpsw,
                                            uint32_t coreKey,
                                            uint32_t coreId,
                                            uint32_t rxFlowStartIdx,
                                            uint32_t rxFlowIdx,
                                            uint8_t macAddress[CPSW_MAC_ADDR_LEN])
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_MacDstFlowInfo macDstFlow;

    macDstFlow.coreKey  = coreKey;
    macDstFlow.startIdx = rxFlowStartIdx;
    macDstFlow.flowIdx  = rxFlowIdx;
    CPSW_UTILS_ARRAY_COPY(macDstFlow.macAddress, macAddress);
    CPSW_IOCTL_SET_IN_ARGS(&prms, &macDstFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_UNREGISTER_DSTMAC_RX_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_registerDstMacRxFlow() failed : %d\n", status);
    }

    return status;
}

bool CpswAppUtils_isPortLinkUp(Cpsw_Handle hCpsw,
                               uint32_t coreId,
                               Cpsw_MacPort portNum)
{
    Cpsw_IoctlPrms prms;
    Cpsw_GenericPortLinkInArgs portLinkInArgs;
    bool linked;
    int32_t status;

    portLinkInArgs.portNum = portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &portLinkInArgs, &linked);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_IOCTL_IS_PORT_LINK_UP, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_isPortLinkUp() failed to get port %u's link status: %d\n",
                           portNum, status);
        linked = false;
    }

    return linked;
}

static void CpswAppUtils_reduceCoreMacAllocation(CpswRm_ResourcePrms *rscPrms,
                                                 uint32_t *pReduceCount,
                                                 uint32_t coreMinCount,
                                                 bool skipCore,
                                                 uint32_t skipCoreId)
{
    uint32_t i;

    for (i = 0; (i < rscPrms->numCores) && (*pReduceCount > 0); i++)
    {
        if ((rscPrms->coreDmaResourceInfo[i].numMacAddress > coreMinCount)
            &&
            ((skipCore == false) || (skipCoreId != rscPrms->coreDmaResourceInfo[i].coreId)))
        {
            uint32_t coreMacAddrReducedCount = (rscPrms->coreDmaResourceInfo[i].numMacAddress - coreMinCount);

            if (*pReduceCount >= coreMacAddrReducedCount)
            {
                *pReduceCount -= coreMacAddrReducedCount;
            }
            else
            {
                coreMacAddrReducedCount -= *pReduceCount;
                *pReduceCount            = 0;
            }

            CpswAppUtils_print("CpswAppUtils_reduceCoreMacAllocation: "
                               "Reduced Mac Address Allocation for CoreId:%u From %u To %u \n",
                               rscPrms->coreDmaResourceInfo[i].coreId,
                               rscPrms->coreDmaResourceInfo[i].numMacAddress,
                               (rscPrms->coreDmaResourceInfo[i].numMacAddress - coreMacAddrReducedCount));
            rscPrms->coreDmaResourceInfo[i].numMacAddress -= coreMacAddrReducedCount;
        }
    }
}

static void CpswAppUtils_updateMacResourcePartition(CpswRm_ResourcePrms *rscPrms,
                                                    uint32_t availMacCount,
                                                    uint32_t selfCoreId)
{
    uint32_t totalRscPatitionMacCount;
    uint32_t i;

    totalRscPatitionMacCount = 0;
    for (i = 0; i < rscPrms->numCores; i++)
    {
        totalRscPatitionMacCount += rscPrms->coreDmaResourceInfo[i].numMacAddress;
    }

    if (totalRscPatitionMacCount > availMacCount)
    {
        uint32_t reduceCount = totalRscPatitionMacCount - availMacCount;

        /* First reduce mac count for cores with more than one mac address allocation */
        CpswAppUtils_reduceCoreMacAllocation(rscPrms, &reduceCount, 1, false, selfCoreId);
        if (reduceCount)
        {
            /* Next reduce mac address for core other than self core to 0 */
            CpswAppUtils_reduceCoreMacAllocation(rscPrms, &reduceCount, 0, true, selfCoreId);
        }

        /* Finally reduce self core also to 0 */
        if (reduceCount)
        {
            /* Next reduce mac address for core other than self core to 0 */
            CpswAppUtils_reduceCoreMacAllocation(rscPrms, &reduceCount, 0, false, selfCoreId);
        }

        CpswAppUtils_assert(reduceCount == 0);
    }
}

void CpswAppUtils_initResourceConfig(Cpsw_Type cpswType,
                                     uint32_t selfCoreId,
                                     CpswRm_ResourceConfig *rscCfg)
{
    int32_t status;
    const CpswRm_ResourcePrms *rscPrms         = CpswAppRm_getResourcePartitionInfo(cpswType);
    const CpswRm_IoctlPermissionTable *ioPerms = CpswAppRm_getIoctlPermissionInfo(cpswType);

    CpswAppUtils_assert(rscPrms != NULL);
    rscCfg->resourcePartitionInfo = *rscPrms;

    CpswAppUtils_assert(ioPerms != NULL);
    rscCfg->ioctlPermissionInfo = *ioPerms;

    status =
        CpswAppSoc_getMacAddrList(cpswType,
                                  rscCfg->macList.macAddress,
                                  CPSW_UTILS_ARRAYSIZE(rscCfg->macList.macAddress),
                                  &rscCfg->macList.numMacAddress);
    CpswAppUtils_assert(status == CPSW_SOK);
    if (rscCfg->macList.numMacAddress > CPSW_UTILS_ARRAYSIZE(rscCfg->macList.macAddress))
    {
        CpswAppUtils_print("CpswAppUtils_initResourceConfig: "
                           "Limiting number of mac address entries to rscCfg->macList.macAddress size"
                           "Available:%u, LimitedTo: %u",
                           rscCfg->macList.numMacAddress,
                           CPSW_UTILS_ARRAYSIZE(rscCfg->macList.macAddress));
        rscCfg->macList.numMacAddress = CPSW_UTILS_ARRAYSIZE(rscCfg->macList.macAddress);
    }

    CpswAppUtils_updateMacResourcePartition(&rscCfg->resourcePartitionInfo,
                                            rscCfg->macList.numMacAddress,
                                            selfCoreId);
    rscCfg->selfCoreId = selfCoreId;
}

void CpswAppUtils_setCommonRxFlowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    pRxFlowPrms->numRxPkts           = CPSW_APPMEMUTILS_NUM_RX_PKTS;
    pRxFlowPrms->disableCacheOpsFlag = false;
    pRxFlowPrms->rxFlowMtu           = CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE;

    pRxFlowPrms->ringMemAllocFxn = &CpswAppMemUtils_allocRingMemFxn;
    pRxFlowPrms->ringMemFreeFxn  = &CpswAppMemUtils_freeRingMemFxn;

    pRxFlowPrms->dmaDescAllocFxn = &CpswAppMemUtils_allocDmaDescFxn;
    pRxFlowPrms->dmaDescFreeFxn  = &CpswAppMemUtils_freeDmaDescFxn;
}

void CpswAppUtils_addHostPortEntry(Cpsw_Handle hCpsw,
                                   uint32_t coreId,
                                   uint8_t *macAddr)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = CPSW_ALE_HOST_PORT_NUM;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = true;
    setUcastInArgs.info.super   = false;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_addHostPortEntry() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    CpswAppUtils_assert(status == CPSW_SOK);
}

static void CpswAppUtils_delAddrEntry(Cpsw_Handle hCpsw,
                                      uint32_t coreId,
                                      uint8_t *macAddr)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_MacAddrInfo addrInfo;

    memset(&addrInfo, 0, sizeof(addrInfo));
    memcpy(&addrInfo.addr[0U], macAddr, sizeof(addrInfo.addr));
    addrInfo.vlanId = 0U;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &addrInfo);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_ALE_IOCTL_REMOVE_ADDR, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswAppUtils_delAddrEntry() failed CPSW_ALE_IOCTL_REMOVE_ADDR: %d\n",
                           status);
    }

    CpswAppUtils_assert(status == CPSW_SOK);
}

void CpswAppUtils_setCommonTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms)
{
    pTxChPrms->numTxPkts           = CPSW_APPMEMUTILS_NUM_TX_PKTS;
    pTxChPrms->disableCacheOpsFlag = false;

    pTxChPrms->ringMemAllocFxn = &CpswAppMemUtils_allocRingMemFxn;
    pTxChPrms->ringMemFreeFxn  = &CpswAppMemUtils_freeRingMemFxn;

    pTxChPrms->dmaDescAllocFxn = &CpswAppMemUtils_allocDmaDescFxn;
    pTxChPrms->dmaDescFreeFxn  = &CpswAppMemUtils_freeDmaDescFxn;
}

uint32_t CpswAppUtils_getStartFlowIdx(Cpsw_Handle hCpsw,
                                      uint32_t coreId)
{
    Cpsw_IoctlPrms prms;
    uint32_t p0FlowIdOffset;
    int32_t status;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &p0FlowIdOffset);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_HOSTPORT_GET_FLOW_ID_OFFSET,
                        &prms);

    CpswAppUtils_assert(status == CPSW_SOK);
    return p0FlowIdOffset;
}

void CpswAppUtils_absFlowIdx2FlowIdxOffset(Cpsw_Handle hCpsw,
                                           uint32_t coreId,
                                           uint32_t absRxFlowId,
                                           uint32_t *pStartFlowIdx,
                                           uint32_t *pFlowIdxOffset)
{
    uint32_t p0FlowIdOffset;

    p0FlowIdOffset = CpswAppUtils_getStartFlowIdx(hCpsw, coreId);
    CpswAppUtils_assert(absRxFlowId >= p0FlowIdOffset);
    *pStartFlowIdx  = p0FlowIdOffset;
    *pFlowIdxOffset = (absRxFlowId - p0FlowIdOffset);
}

void CpswAppUtils_openTxCh(Cpsw_Handle hCpsw,
                           uint32_t coreKey,
                           uint32_t coreId,
                           uint32_t *pTxChNum,
                           CpswDma_TxChHandle *pTxChHandle,
                           CpswDma_OpenTxChPrms *pCpswTxChCfg)
{
    int32_t status;

    status = CpswAppUtils_allocTxCh(hCpsw,
                                    coreKey,
                                    coreId,
                                    pTxChNum);
    CpswAppUtils_assert(CPSW_SOK == status);

    pCpswTxChCfg->chNum = *pTxChNum;

    *pTxChHandle = CpswDma_openTxCh(pCpswTxChCfg);
    CpswAppUtils_assert(NULL != *pTxChHandle);
}

void CpswAppUtils_closeTxCh(Cpsw_Handle hCpsw,
                            uint32_t coreKey,
                            uint32_t coreId,
                            CpswDma_PktInfoQ *pFqPktInfoQ,
                            CpswDma_PktInfoQ *pCqPktInfoQ,
                            CpswDma_TxChHandle hTxChHandle,
                            uint32_t txChNum)
{
    int32_t status;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    CpswDma_disableTxEvent(hTxChHandle);
    status = CpswDma_closeTxCh(hTxChHandle, pFqPktInfoQ, pCqPktInfoQ);
    CpswAppUtils_assert(CPSW_SOK == status);

    status = CpswAppUtils_freeTxCh(hCpsw,
                                   coreKey,
                                   coreId,
                                   txChNum);
    CpswAppUtils_assert(CPSW_SOK == status);
}

void CpswAppUtils_openRxFlow(Cpsw_Handle hCpsw,
                             uint32_t coreKey,
                             uint32_t coreId,
                             bool useDefaultFlow,
                             uint32_t *pRxFlowStartIdx,
                             uint32_t *pRxFlowIdx,
                             uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                             CpswDma_RxFlowHandle *pRxFlowHandle,
                             CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    int32_t status = CPSW_SOK;

    status = CpswAppUtils_allocRxFlow(hCpsw,
                                      coreKey,
                                      coreId,
                                      pRxFlowStartIdx,
                                      pRxFlowIdx);
    CpswAppUtils_assert(status == CPSW_SOK);

    pRxFlowPrms->startIdx = *pRxFlowStartIdx;
    pRxFlowPrms->flowIdx  = *pRxFlowIdx;

    *pRxFlowHandle = CpswDma_openRxFlow(pRxFlowPrms);
    CpswAppUtils_assert(*pRxFlowHandle != NULL);

    status = CpswAppUtils_allocMac(hCpsw, coreKey, coreId, macAddr);
    CpswAppUtils_assert(status == CPSW_SOK);

    CpswAppUtils_addHostPortEntry(hCpsw, coreId, macAddr);
    if (useDefaultFlow)
    {
        status = CpswAppUtils_registerDefaultRxFlow(hCpsw,
                                                    coreKey,
                                                    coreId,
                                                    *pRxFlowStartIdx,
                                                    *pRxFlowIdx);
    }
    else
    {
        status = CpswAppUtils_registerDstMacRxFlow(hCpsw,
                                                   coreKey,
                                                   coreId,
                                                   *pRxFlowStartIdx,
                                                   *pRxFlowIdx,
                                                   macAddr);
    }

    CpswAppUtils_assert(status == CPSW_SOK);
}

void CpswAppUtils_closeRxFlow(Cpsw_Handle hCpsw,
                              uint32_t coreKey,
                              uint32_t coreId,
                              bool useDefaultFlow,
                              CpswDma_PktInfoQ *pFqPktInfoQ,
                              CpswDma_PktInfoQ *pCqPktInfoQ,
                              uint32_t rxFlowStartIdx,
                              uint32_t rxFlowIdx,
                              uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                              CpswDma_RxFlowHandle hRxFlow)
{
    int32_t status = CPSW_SOK;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    CpswDma_disableRxEvent(hRxFlow);

    if (useDefaultFlow)
    {
        status = CpswAppUtils_unregisterDefaultRxFlow(hCpsw,
                                                      coreKey,
                                                      coreId,
                                                      rxFlowStartIdx,
                                                      rxFlowIdx);
    }
    else
    {
        status = CpswAppUtils_unregisterDstMacRxFlow(hCpsw,
                                                     coreKey,
                                                     coreId,
                                                     rxFlowStartIdx,
                                                     rxFlowIdx,
                                                     macAddr);
    }

    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_delAddrEntry(hCpsw, coreId, macAddr);
    status = CpswDma_closeRxFlow(hRxFlow,
                                 pFqPktInfoQ,
                                 pCqPktInfoQ);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_freeMac(hCpsw,
                                  coreKey,
                                  coreId,
                                  macAddr);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_freeRxFlow(hCpsw,
                                     coreKey,
                                     coreId,
                                     rxFlowIdx);
    CpswAppUtils_assert(status == CPSW_SOK);
}

static void CpswAppUtils_printDmaDescStats(CpswDma_DmaDescStats *descstats)
{
    // TODO when DMA stats are enabled.
}

static void CpswAppUtils_printCbStats(CpswDma_CbFxnStats *cbStats)
{
    uint32_t i;

    CpswAppUtils_printStatsNonZero("Data Notify Count          = %llu\n", cbStats->dataNotifyCnt);
    CpswAppUtils_printStatsNonZero("Zero Notify Count          = %llu\n", cbStats->zeroNotifyCnt);
    CpswAppUtils_printStatsNonZero("Total Packets Count        = %llu\n", cbStats->totalPktCnt);
    CpswAppUtils_printStatsNonZero("Total Cycles Count         = %llu\n", cbStats->totalCycleCnt);
    CpswAppUtils_printStatsNonZero("Packets per Notify Max     = %llu\n", cbStats->pktsPerNotifyMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("Packets per Notify[%d] = %llu\n", i, cbStats->pktsPerNotify[i]);
    }

    CpswAppUtils_printStatsNonZero("Cycles per Notify Max      = %llu\n", cbStats->cycleCntPerNotifyMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("Cycles per Notify[%d]  = %llu\n", i, cbStats->cycleCntPerNotify[i]);
    }

    CpswAppUtils_printStatsNonZero("Cycles per Packet Max      = %llu\n", cbStats->cycleCntPerPktMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("Cycles per Packet[%d]  = %llu\n", i, cbStats->cycleCntPerPkt[i]);
    }

    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswAppUtils_printStatsWithIdxNonZero("Ready Desc Q Count[%d] = %llu\n", i, cbStats->readyDmaDescQCnt[i]);
    }
}

static void CpswAppUtils_printRxFlowStats(CpswDma_RxFlowStats *stats)
{
    CpswAppUtils_print(" RX Flow Statistics\n");
    CpswAppUtils_print("-----------------------------------------\n");
    CpswAppUtils_printCbStats(&stats->submitPktStats);
    CpswAppUtils_printCbStats(&stats->retrievePktStats);
    CpswAppUtils_printDmaDescStats(&stats->dmaDescStats);

    CpswAppUtils_printStatsNonZero("RX Submit Packet EnQ count           = %llu\n", stats->rxSubmitPktEnq);
    CpswAppUtils_printStatsNonZero("RX Submit Packet Underflow           = %llu\n", stats->rxSubmitPktUnderFlowCnt);
    CpswAppUtils_printStatsNonZero("RX Submit Packet DeQ count           = %llu\n", stats->rxRetrievePktDq);
}

static void CpswAppUtils_printTxChStats(CpswDma_TxChStats *stats)
{
    CpswAppUtils_print(" RX Flow Statistics\n");
    CpswAppUtils_print("-----------------------------------------\n");
    CpswAppUtils_printCbStats(&stats->submitPktStats);
    CpswAppUtils_printCbStats(&stats->retrievePktStats);
    CpswAppUtils_printDmaDescStats(&stats->dmaDescStats);

    CpswAppUtils_printStatsNonZero("TX Submit Packet EnQ count           = %llu\n", stats->txSubmitPktEnq);
    CpswAppUtils_printStatsNonZero("TX Submit Packet Underflow           = %llu\n", stats->txSubmitPktOverFlowCnt);
    CpswAppUtils_printStatsNonZero("TX Submit Packet DeQ count           = %llu\n", stats->txRetrievePktDq);
}

int32_t CpswAppUtils_showRxFlowStats(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal = CPSW_SOK;

    CpswDma_RxFlowStats rxFlowStats;

    retVal = CpswDma_getRxFlowStats(hRxFlow, &rxFlowStats);
    if (CPSW_ENOTSUPPORTED != retVal)
    {
        CpswAppUtils_printRxFlowStats(&rxFlowStats);
        CpswDma_resetRxFlowStats(hRxFlow);
    }

    return retVal;
}

int32_t CpswAppUtils_showTxChStats(CpswDma_TxChHandle hTxCh)
{
    int32_t retVal = CPSW_SOK;

    CpswDma_TxChStats txChStats;

    retVal = CpswDma_getTxChStats(hTxCh, &txChStats);
    if (CPSW_ENOTSUPPORTED != retVal)
    {
        CpswAppUtils_printTxChStats(&txChStats);
        CpswDma_resetTxChStats(hTxCh);
    }

    return retVal;
}

void CpswAppUtils_setNoPhyCfgRgmii(CpswMacPort_Interface *interface,
                                   CpswPhy_Config *phyConfig)
{
    phyConfig->phyAddr      = CPSW_PHY_INVALID_PHYADDR;
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_FORCED;
}

void CpswAppUtils_setNoPhyCfgRmii(CpswMacPort_Interface *interface,
                                   CpswPhy_Config *phyConfig)
{
    phyConfig->phyAddr      = CPSW_PHY_INVALID_PHYADDR;
    interface->layerType    = CPSW_MAC_LAYER_MII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;
}

void CpswAppUtils_setNoPhyCfgSgmii(CpswMacPort_Interface *interface,
                                   CpswMacPort_Config *macConfig,
                                   CpswPhy_Config *phyConfig)
{
    phyConfig->phyAddr      = CPSW_PHY_INVALID_PHYADDR;
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_SERIAL;
    interface->variantType  = CPSW_MAC_VARIANT_NONE;

    macConfig->sgmiiConfig.sgmiiMode = CPSW_MAC_SGMIIMODE_SGMII_FORCEDLINK;
}

int8_t CpswAppUtils_hex2Num(char hex)
{
    int8_t num = -1;

    if ((hex >= '0') && (hex <= '9'))
    {
        num =  hex - '0';
    }
    else if (hex >= 'a' && hex <= 'f')
    {
        num = hex - 'a' + 10;
    }
    else if (hex >= 'A' && hex <= 'F')
    {
        num = hex - 'A' + 10;
    }

    return num;
}

int32_t CpswAppUtils_macAddrAtoI(const char *txt, uint8_t *addr)
{
    int32_t status = CPSW_SOK;
    int8_t a, b, i;

    for (i = 0; i < 6; i++)
    {
        a = CpswAppUtils_hex2Num(*txt++);
        if (a < 0)
        {
            status = CPSW_EFAIL;
        }

        b = CpswAppUtils_hex2Num(*txt++);
        if (b < 0)
        {
            status = CPSW_EFAIL;
        }

        *addr++ = (a << 4) | b;

        if ((i < 5) && (*txt++ != ':'))
        {
            status = CPSW_EFAIL;
            break;
        }
    }

    return status;
}

int32_t CpswAppUtils_ipAddrAtoI(const char* txt, uint8_t *addr)
{
    int32_t status = CPSW_SOK;
    uint8_t i;

    for (i = 0U; i < 4U; i++)
    {
        addr[i] = strtoul(txt, NULL, 10U);
        txt = strchr(txt, '.');
        if (((txt == NULL) || (*txt == '\0')) && (i != 3U))
        {
            status = CPSW_EFAIL;
            break;
        }
        txt++;
    }
    return status;

}

int32_t CpswAppUtils_setTimeSyncRouter(Cpsw_Type cpswType, uint32_t input, uint32_t output)
{
    int32_t  status = CPSW_SOK;

#if defined(SOC_AM65XX)
    CpswAppUtils_assert(cpswType == CPSW_2G);
#endif

    /** TODO: Remove the below ifdef based on SOC_J721E once TSR
     * macros are added for AM65xx */
#if defined (SOC_J721E) || defined (SOC_J7200)
    CSL_IntrRouterCfg   irRegs;

    irRegs.pIntrRouterRegs = (CSL_intr_router_cfgRegs *) CSL_TIMESYNC_INTRTR0_INTR_ROUTER_CFG_BASE;
    irRegs.pIntdRegs       = (CSL_intr_router_intd_cfgRegs *) NULL;
    irRegs.numInputIntrs   = CPSW_APP_SOC_TSYNC_ROUTER_NUM_INPUT;
    irRegs.numOutputIntrs  = CPSW_APP_SOC_TSYNC_ROUTER_NUM_OUTPUT;

    switch (cpswType)
    {
        case CPSW_2G:
        {
            if ((output != CSLR_TIMESYNC_INTRTR0_OUTL_MCU_CPSW0_CPTS_HW3_PUSH_0) &&
                (output != CSLR_TIMESYNC_INTRTR0_OUTL_MCU_CPSW0_CPTS_HW4_PUSH_0))
            {
                status = CPSW_EINVALID_PARAMS;
            }
        }
        break;

#if defined (SOC_J721E) || defined (SOC_J7200)
    case CPSW_5G:
    case CPSW_9G:
        {
            if ((output < CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW1_PUSH_0) ||
                (output > CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW8_PUSH_0))
            {
                status = CPSW_EINVALID_PARAMS;
            }
        }
        break;
#endif
        default:
        {
            CpswAppUtils_assert(false);
            status = CPSW_EINVALID_PARAMS;
        }
        break;
    }

    if (status == CPSW_SOK)
    {
        CSL_intrRouterCfgMux(&irRegs, input, output);
    }
    else
    {
        CpswAppUtils_print("Wrong Input/Output params for TSync Router Mux configuration: %d \n", status);
    }
#endif
    return status;
}

/* end of file */
