/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_apputils_rtos.c
 *
 * \brief    CPSW application utility to get Tasks Stack Status.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

#include <ti/drv/cpsw/cpsw.h>
#include "inc/cpsw_apputils.h"
#include "inc/cpsw_apputils_rtos.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static void CpswAppUtils_printTaskStackStats(Task_Handle hTask)
{
    Task_Stat statBuff;
    uint32_t stackPercent;
    char *taskName = NULL;
    char nameLessTask[20] = "empty-task-name";

    CpswAppUtils_assert(NULL != hTask);

    /* Function get Task status */
    Task_stat(hTask, &statBuff);
    taskName = Task_Handle_name(hTask);
    if (NULL == taskName)
    {
        taskName = nameLessTask;
    }

    if (statBuff.stackSize != 0U)
    {
        stackPercent = CPSW_UTILS_DIV_ROUNDUP(100U * statBuff.used, statBuff.stackSize);
    }
    else
    {
        stackPercent = 0U;
    }

    CpswAppUtils_print("%-30s\t %6d\t\t%6d\t\t%6d\t\t%6d%%\n",
                       taskName,
                       statBuff.priority,
                       statBuff.stackSize,
                       statBuff.used,
                       stackPercent);

    if (stackPercent > 90U)
    {
        CpswAppUtils_print("Task stack (%u%%) is about to overflow\n", stackPercent);
    }
}

void CpswAppUtils_printTaskStackUsage(void)
{
    Task_Object *task;
    uint32_t i;

    CpswAppUtils_print("==========================================================================="
                       "====================\n");
    CpswAppUtils_print("Task Name\t\t\tPriority\tStack Size\tStack Peak\tPercentage Used \n");
    CpswAppUtils_print("==========================================================================="
                       "====================\n");

    /* Below for loop is to traverse through the list of statically created tasks
     * Task_Object_count() returns the size of the array */
    for (i = 0; i < Task_Object_count(); i++)
    {
        task = Task_Object_get(NULL, i);
        CpswAppUtils_printTaskStackStats(task);
    }

    /* Below while loop is used to traverse the linked list maintaining
     * dynamically created tasks */
    task = Task_Object_first();
    while (NULL != task)
    {
        CpswAppUtils_printTaskStackStats(task);
        task = Task_Object_next(task);
    }
}

void CpswAppUtils_wait(uint32_t waitTime)
{
    Task_sleep(waitTime);
}

