/*
 * Copyright (c) 2014-2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_networkutils.c
 *
 *  \brief File containing the Network functions.
 *
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdint.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>

/* OSAL Header files */
#include <ti/osal/osal.h>

/* CSL Header files */
#include <ti/csl/csl_types.h>

/* NDK Dependencies */
#include <ti/ndk/inc/socket.h>
#include <ti/ndk/inc/netmain.h>
#include <ti/ndk/inc/stkmain.h>

#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>

/* Module Header files */
#include "inc/cpsw_networkutils.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_NETWORKUTILS_INVALID_SOCKET              (0xFFFFFFFFU)
#define CPSW_NETWORKUTILS_TX_SERVER_CLOSED            (0U)
#define CPSW_NETWORKUTILS_TX_SERVER_LISTEN            (1U)
#define CPSW_NETWORKUTILS_TX_SERVER_CONNECTED         (2U)
#define CPSW_NETWORKUTILS_TX_SERVER_POLL_TIMEOUT      (10U)
/* Maximum number of connects to queue */
#define CPSW_NETWORKUTILS_MAX_CONNECTS                (5)
/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static int32_t CpswNetworkUtils_open(CpswNetworkUtils_SockObj *pObj,
                                     uint32_t port);

static int32_t CpswNetworkUtils_waitConnect(CpswNetworkUtils_SockObj *pObj,
                                            uint32_t timeout);

static int32_t CpswNetworkUtils_read(CpswNetworkUtils_SockObj *pObj,
                                     uint8_t *dataBuf,
                                     uint32_t *dataSize);

static int32_t CpswNetworkUtils_write(CpswNetworkUtils_SockObj *pObj,
                                      uint8_t *dataBuf,
                                      uint32_t dataSize);

static int32_t CpswNetworkUtils_openSession(void *handle);

static int32_t CpswNetworkUtils_sessionClose(void *handle);

static int32_t CpswNetworkUtils_waitConnection(CpswNetworkUtils_SockObj *pObj);

static int32_t CpswNetworkUtils_close(CpswNetworkUtils_SockObj *pObj,
                                      bool closeServerSock);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswNetworkUtils_createSocket(CpswNetworkUtils_SockObj *pObj)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    CpswNetworkUtils_openSession(NULL);

    status = CpswNetworkUtils_open(pObj, pObj->port);

    if (status != CPSW_NETWORKUTILS_SUCCESS)
    {
        CpswAppUtils_print("Network_open failed !!!\n");
    }
    else
    {
        pObj->state = CPSW_NETWORKUTILS_TX_SERVER_LISTEN;
    }

    return status;
}

int32_t CpswNetworkUtils_deleteSocket(CpswNetworkUtils_SockObj *pObj)
{
    CpswNetworkUtils_close(pObj, true);

    CpswNetworkUtils_sessionClose(NULL);

    pObj->state = CPSW_NETWORKUTILS_TX_SERVER_CLOSED;

    CpswAppUtils_print("Network Utils: Server Closed (port=%d) !!!\n",
                       pObj->port);

    return 0U;
}

int32_t CpswNetworkUtils_writePayload(CpswNetworkUtils_SockObj *pObj,
                                      uint8_t *bufAddr,
                                      uint32_t bufSize)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    status = CpswNetworkUtils_waitConnection(pObj);

    if (pObj->state == CPSW_NETWORKUTILS_TX_SERVER_CONNECTED)
    {
        /* write header to client, if write failed, then client
         * disconnected, so go back to listening
         */
        status = CpswNetworkUtils_write(pObj,
                                        bufAddr,
                                        bufSize);

        if (status != CPSW_NETWORKUTILS_SUCCESS)
        {
            CpswNetworkUtils_close(pObj, false);

            pObj->state = CPSW_NETWORKUTILS_TX_SERVER_LISTEN;
        }
    }

    return status;
}

int32_t CpswNetworkUtils_readPayload(CpswNetworkUtils_SockObj *pObj,
                                     uint8_t *bufAddr,
                                     uint32_t bufSize)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    status = CpswNetworkUtils_waitConnection(pObj);

    if (pObj->state == CPSW_NETWORKUTILS_TX_SERVER_CONNECTED)
    {
        /* write header to client, if write failed, then client
         * disconnected, so go back to listening
         */
        status = CpswNetworkUtils_read(pObj,
                                       bufAddr,
                                       &bufSize);

        if (status != 0)
        {
            CpswNetworkUtils_close(pObj, false);

            pObj->state = CPSW_NETWORKUTILS_TX_SERVER_LISTEN;
        }
    }

#if (defined(DEBUG))
    CpswAppUtils_print("NETWORK_RX: NETWORK_RX: BUF%d: %d bytes recevied (port=%d)!!!\n",
                       i,
                       bufSize[i],
                       pObj->createArgs.networkServerPort
                       );
#endif

    return status;
}

/* ========================================================================== */
/*                          Static Function Definitions                       */
/* ========================================================================== */

static int32_t CpswNetworkUtils_waitConnection(CpswNetworkUtils_SockObj *pObj)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;
    int32_t isConnected;

    if (pObj->state == CPSW_NETWORKUTILS_TX_SERVER_LISTEN)
    {
        isConnected = CpswNetworkUtils_waitConnect(pObj,
                                                   CPSW_NETWORKUTILS_TX_SERVER_POLL_TIMEOUT);

        if (CPSW_NETWORKUTILS_SUCCESS == isConnected)
        {
            /* connected to client */
            pObj->state = CPSW_NETWORKUTILS_TX_SERVER_CONNECTED;
        }
        else
        {
            status = CPSW_NETWORKUTILS_ERROR;
        }
    }

    return status;
}

static int32_t CpswNetworkUtils_read(CpswNetworkUtils_SockObj *pObj,
                                     uint8_t *dataBuf,
                                     uint32_t *dataSize)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    uint32_t tmpDataSize;
    int32_t actDataSize = 0;

    tmpDataSize = *dataSize;

    while (tmpDataSize > 0U)
    {
        actDataSize = NDK_recv(pObj->connectedSockFd, (void *)dataBuf, tmpDataSize, 0);
        if (actDataSize <= 0)
        {
            *dataSize = 0U;
            status    = CPSW_NETWORKUTILS_ERROR;
            break;
        }
        else
        {
            dataBuf     += actDataSize;
            tmpDataSize -= (uint32_t)actDataSize;
        }
    }

    return status;
}

static int32_t CpswNetworkUtils_write(CpswNetworkUtils_SockObj *pObj,
                                      uint8_t *dataBuf,
                                      uint32_t dataSize)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    int32_t actDataSize = 0;

    while (dataSize > 0)
    {
        actDataSize = (int32_t)NDK_send(pObj->connectedSockFd, dataBuf, (size_t)dataSize, 0);

        if (actDataSize <= 0)
        {
            break;
        }
        else
        {
            dataBuf  += actDataSize;
            dataSize -= (uint32_t)actDataSize;
        }
    }

    if (dataSize > 0)
    {
        status = CPSW_NETWORKUTILS_ERROR;
    }

    return status;
}

static int32_t CpswNetworkUtils_open(CpswNetworkUtils_SockObj *pObj,
                                     uint32_t port)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    struct sockaddr_in sin1;
    int32_t option = 1;

    pObj->connectedSockFd = (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET;
    pObj->port            = port;
    pObj->sockFd          = (SOCKET)NDK_socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (pObj->sockFd == (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET)
    {
        CpswAppUtils_print("Network Utils: Unable to open socket (port=%d)!!!\n", port);
        status = CPSW_NETWORKUTILS_ERROR;
    }
    else
    {
        /* Bind to the specified Server port */
        memset(&sin1, 0U, sizeof(struct sockaddr_in));
        sin1.sin_family      = AF_INET;
        sin1.sin_addr.s_addr = INADDR_ANY;
        sin1.sin_port        = (uint16_t)(htons((uint16_t)(pObj->port)));

        NDK_setsockopt(pObj->sockFd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option));

        if (NDK_bind(pObj->sockFd, (struct sockaddr *)&sin1, sizeof(sin1)) < 0)
        {
            CpswAppUtils_print("Network Utils: Unable to bind() (port=%d) !!!\n", port);
            fdClose(pObj->sockFd);
            pObj->sockFd = (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET;
            status       = CPSW_NETWORKUTILS_ERROR;
        }
        else
        {
            if (NDK_listen(pObj->sockFd, CPSW_NETWORKUTILS_MAX_CONNECTS) < 0)
            {
                fdClose(pObj->sockFd);
                pObj->sockFd = (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET;
                status       = CPSW_NETWORKUTILS_ERROR;
            }
        }
    }

    return status;
}

static int32_t CpswNetworkUtils_close(CpswNetworkUtils_SockObj *pObj,
                                      bool closeServerSock)
{
    if (pObj->connectedSockFd != (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET)
    {
        fdClose(pObj->connectedSockFd);
        pObj->connectedSockFd = (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET;
    }

    if (closeServerSock)
    {
        if (pObj->sockFd != (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET)
        {
            fdClose(pObj->sockFd);
            pObj->sockFd = (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET;
        }
    }

    return 0;
}

static int32_t CpswNetworkUtils_waitConnect(CpswNetworkUtils_SockObj *pObj,
                                            uint32_t timeout)
{
    int32_t status = CPSW_NETWORKUTILS_SUCCESS;

    pObj->pollitem[0].fd              = pObj->sockFd;
    pObj->pollitem[0].eventsRequested = 0x1 /* POLLIN */;

    if (fdPoll(pObj->pollitem, 1U, timeout) == SOCKET_ERROR)
    {
        CpswAppUtils_print("Network Utils: fdPoll() failed with SOCKET_ERROR (port=%d) !!!\n", pObj->port);
        status = CPSW_NETWORKUTILS_ERROR;
    }
    else
    {
        if (pObj->pollitem[0].eventsDetected == FALSE)
        {
            /* NO connection, retry */
            status = CPSW_NETWORKUTILS_SUCCESS;
        }
        else
        {
            if ((uint32_t)(pObj->pollitem[0].eventsDetected) & (uint32_t)0x08 /* POLLNVAL */)
            {
                CpswAppUtils_print("Network Utils: fdPoll() failed with POLLNVAL (port=%d) !!!\n", pObj->port);
                status = CPSW_NETWORKUTILS_ERROR;
            }
            else
            {
                if ((uint32_t)(pObj->pollitem[0].eventsDetected) & (uint32_t)0x1)
                {
                    pObj->connectedSockFd = (SOCKET)NDK_accept(pObj->sockFd, 0, 0);

                    if (pObj->connectedSockFd != (SOCKET)CPSW_NETWORKUTILS_INVALID_SOCKET)
                    {
                        status = 1;
                    }
                    else
                    {
                        CpswAppUtils_print("Not able to connect to PC client !!!\n");
                    }
                }
            }
        }
    }

    /* NO connection, retry */
    return status;
}

static int32_t CpswNetworkUtils_openSession(void *handle)
{
    if (handle == NULL)
    {
        handle = TaskSelf();
    }

    /* Allocate the file environment for this task */
    fdOpenSession(handle);

    return 0;
}

static int32_t CpswNetworkUtils_sessionClose(void *handle)
{
    if (handle == NULL)
    {
        handle = TaskSelf();
    }

    fdCloseSession(handle);

    return 0;
}

/* end of file */
