/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_ale.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the Address Lookup Engine (ALE).
 */

/*!
 *  \ingroup DRV_CPSW_MODULE
 *  \defgroup CPSW_ALE_API ALE Handler and Driver API
 *
 *  The ALE Driver  provides low level access to switch configuration
 *
 *  Following features are supported:
 *      - VLAN aware or unaware switching including Unknown VLAN and null VLAN
 *        controls
 *      - Auto Learn on a per VLAN and per port bases
 *      - Manual or fully Automatic aging of agable addresses
 *      - STP required forwarding state controls and status
 *      - MACSEC forwarding controls and status
 *      - Port and VLAN, ONU, or Address traffic mirroring
 *      - Port Trunking for up to four trunks across any port combination
 *      - Multicast Ranging for up to 1024 addresses per ALE table entry
 *                  - Overlapping Multicast Ranges
 *      - Packet Classification based on Port, Priority, ONU, DA, SA, S-VLAN,
 *        C-VLAN,
 *        EtherType, IPDA and/or IPSA combination with full CIDR masking support
 *        for IPv4 and
 *        IPv6 addresses
 *      - Policing of ingress data flows based on any packet classification
 *      - Host Thread mapping based on any packet classification
 */

/*!
 * \defgroup ALE_IOCTLS  ALE IOCTLS
 * \ingroup CPSW_ALE_API
 */

/*!
 * \defgroup CPSW_ALE_OPENPARAMS  ALE OPENPARAMS
 * \ingroup CPSW_ALE_API
 */

/*!
 * \defgroup CPSW_ALE_ENUMS  ALE ENUMS
 * \ingroup CPSW_ALE_API
 */

/*!
 * \addtogroup CPSW_ALE_API
 * @{
 */

#ifndef CPSW_ALE_H_
#define CPSW_ALE_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Number of switch ports in the subsystem */
#define CPSW_ALE_NUM_PORTS                           ((uint32_t)9)

/*! \brief Number of external ports in the subsystem */
#define CPSW_ALE_NUM_MAC_PORTS                       (CPSW_ALE_NUM_PORTS - 1U)

/*! \brief Maximum IP next header whilelist */
#define CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR            (4U)

/*! \brief Mask value for the all the PORTS */
#define CPSW_ALE_ALL_PORTS_MASK                      ((1U << CPSW_ALE_NUM_PORTS) - 1U)

/*! \brief Mask value for the all MAC PORTS */
#define CPSW_ALE_ALL_MACPORTS_MASK                   (((1U << CPSW_ALE_NUM_PORTS) - 1U) - 1U)

/*! \brief Mask value for the host PORT */
#define CPSW_ALE_HOST_PORT_MASK                      (0x1U)

/*! \brief Port number for the host PORT */
#define CPSW_ALE_HOST_PORT_NUM                       (0x0U)

/*! \brief Maximum number of ignore bits in multicast address */
#define CPSW_ALE_MCAST_IGN_BITS_MAX                  (10U)

/*! \brief Num of octets in MAC address */
#define CPSW_ALE_MACADDR_NUM_OCTETS                  (CPSW_MAC_ADDR_LEN)

/*! \brief Number of octets in OUI address */
#define CPSW_ALE_OUIADDR_NUM_OCTETS                  (3U)

/*! \brief Number of octets in IPV4 address */
#define CPSW_ALE_IPV4ADDR_NUM_OCTETS                 (4U)

/*! \brief Number of octets in IPV6 address */
#define CPSW_ALE_IPV6ADDR_NUM_OCTETS                 (16U)

/*! \brief Disable peak bit rate */
#define CPSW_ALE_PEAKBITRATE_DISABLE                 (0U)

/*! \brief Disable commit bit rate */
#define CPSW_ALE_COMMITBITRATE_DISABLE               (0U)

/*! \brief ALE invalid thread id */
#define CPSW_ALE_THREADID_INVALID                    (~0U)

/*!
 * \name ALE classifier match types
 *
 * Classifier match type supported by ALE. Each type is represented by a bit
 * To enable multiple classifiers in a single classifier entry, create bitmask
 * ORing required classifier match type defines
 *  @{
 */
/*! Enable classifier match with PORT number */
#define CPSW_ALE_POLICER_MATCH_PORT                  (CPSW_SET_BIT(0U))

/*! Enable classifier match with received packet  priority */
#define CPSW_ALE_POLICER_MATCH_PRIORITY              (CPSW_SET_BIT(1U))

/*! Enable classifier match with OUI portion of source MAC address */
#define CPSW_ALE_POLICER_MATCH_OUI                   (CPSW_SET_BIT(2U))

/*! Enable classifier match with MAC destination address entry with/without VLAN **/
#define CPSW_ALE_POLICER_MATCH_MACDST                (CPSW_SET_BIT(3U))

/*! Enable classifier match with MAC source address entry with/without VLAN */
#define CPSW_ALE_POLICER_MATCH_MACSRC                (CPSW_SET_BIT(4U))

/*! Enable classifier match with outer VLAN entry */
#define CPSW_ALE_POLICER_MATCH_OVLAN                 (CPSW_SET_BIT(5U))

/*! Enable classifier match with inner VLAN entry */
#define CPSW_ALE_POLICER_MATCH_IVLAN                 (CPSW_SET_BIT(6U))

/*! Enable classifier match with Ethertype */
#define CPSW_ALE_POLICER_MATCH_ETHERTYPE             (CPSW_SET_BIT(7U))

/*! Enable classifier match with IPv4/IPv6 source address */
#define CPSW_ALE_POLICER_MATCH_IPSRC                 (CPSW_SET_BIT(8U))

/*! Enable classifier match with IPv4/IPv6 destination address */
#define CPSW_ALE_POLICER_MATCH_IPDST                 (CPSW_SET_BIT(9U))
/*  @} */

/*!
 * \name ALE classifier aleEntry delete bitmask
 *
 * Associated with some classifier types are ALE table entries.
 * When deleting the classifier entry application needs to specify by means
 * of below bitmask if the ALE entry associated with the classifier should also
 * be deleted. If not set the ALE entry will remain.
 *  @{
 */

/*! \brief Delete OUI ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_OUI       (CPSW_ALE_POLICER_MATCH_PRIORITY)

/*! \brief Delete MAC source address ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACSRC    (CPSW_ALE_POLICER_MATCH_MACSRC)

/*! \brief Delete MAC destination address ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST    (CPSW_ALE_POLICER_MATCH_MACDST)

/*! \brief Delete InnerVLAN classifier associated ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_IVLAN     (CPSW_ALE_POLICER_MATCH_IVLAN)

/*! \brief Delete OuterVLAN classifier associated ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_OVLAN     (CPSW_ALE_POLICER_MATCH_OVLAN)

/*! \brief Delete EtherType classifier associated ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_ETHERTYPE (CPSW_ALE_POLICER_MATCH_ETHERTYPE)

/*! \brief Delete IPv4/IPv6 source address classifier associated ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPSRC     (CPSW_ALE_POLICER_MATCH_IPSRC)

/*! \brief Delete IPv4/IPv6 destination address classifier associated ALE table entry */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPDST     (CPSW_ALE_POLICER_MATCH_IPDST)

/*! \brief Delete all ale entries associated with classifier entries */
#define CPSW_ALE_POLICER_TABLEENTRY_DELETE_ALL       (CPSW_ALE_POLICER_TABLEENTRY_DELETE_OUI |       \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACSRC |    \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST |    \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_IVLAN |     \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_OVLAN |     \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_ETHERTYPE | \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPSRC |     \
                                                      CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPDST)

/*  @} */

/*! \brief Base index for MAC ports in ALE context */
#define CPSW_ALE_MACPORT_BASE                        (1U)

/*! \brief Macro to convert MAC port (#Cpsw_MacPort) to ALE port number */
#define CPSW_ALE_MACPORT_TO_ALEPORT(macPortNum)      (CPSW_NORMALIZE_MACPORT(macPortNum) + CPSW_ALE_MACPORT_BASE)

/*! \brief Macro to convert ALE port number to MAC port number(#Cpsw_MacPort) */
#define CPSW_ALE_ALEPORT_TO_MACPORT(alePortNum)      (CPSW_DENORMALIZE_MACPORT(alePortNum - CPSW_ALE_MACPORT_BASE))

/*! \brief Macro to convert MAC port (#Cpsw_MacPort) to ALE port mask */
#define CPSW_ALE_MACPORT_TO_PORTMASK(macPortNum)     CPSW_SET_BIT(CPSW_ALE_MACPORT_TO_ALEPORT(macPortNum))

/*! \brief Helper macro to create IOCTL commands for ALE submodule */
#define CPSW_ALE_PUBLIC_IOCTL(x)                     (CPSW_IOCTL_ALE_BASE | CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \addtogroup ALE_IOCTLS
 *
 * Runtime control commands supported by ALE
 * @{
 */

/*!
 * \brief The ALE module IOCTL enums
 */
typedef enum CPSW_ALE_IOCTL_TAG
{
    /*!
     * \brief Add unicast address entry to ALE table.
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetUcastEntryInArgs
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_UNICAST               = CPSW_ALE_PUBLIC_IOCTL(0U),

    /*!
     * \brief Add multicast address entry to ALE table
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetMcastEntryInArgs
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_MULTICAST             = CPSW_ALE_PUBLIC_IOCTL(1U),

    /*!
     * \brief Add inner/outer vlan entry to ALE table
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_VlanEntryInfo
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_VLAN                  = CPSW_ALE_PUBLIC_IOCTL(2U),

    /*!
     * \brief Add oui address entry to ALE table
     * IOCTL params:
     * -  inArgs: #CpswAle_OuiEntryInfo
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_OUI                   = CPSW_ALE_PUBLIC_IOCTL(3U),

    /*!
     * \brief Add IPv4 address to ALE table
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_Ipv4EntryInfo
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_IPV4ADDR              = CPSW_ALE_PUBLIC_IOCTL(4U),

    /*!
     * \brief Add IPv6 address to ALE table
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_Ipv6EntryInfo
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_IPV6ADDR              = CPSW_ALE_PUBLIC_IOCTL(5U),

    /*!
     * \brief Add ethertype entry to ALE table
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_EtherTypeEntryInfo
     * - outArgs: #CpswAle_AddEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_ADD_ETHERTYPE             = CPSW_ALE_PUBLIC_IOCTL(6U),

    /*!
     * \brief Return ALE entry info for given unicast address
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_MacAddrInfo
     * - outArgs: #CpswAle_GetUcastEntryOutArgs
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_LOOKUP_UNICAST            = CPSW_ALE_PUBLIC_IOCTL(7U),

    /*!
     * \brief Return ALE entry info for given multicast address
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_GetMcastEntryInArgs
     * - outArgs: #CpswAle_GetMcastEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_LOOKUP_MULTICAST          = CPSW_ALE_PUBLIC_IOCTL(8U),

    /*!
     * \brief Return ALE entry info for given vlanId
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_VlanIdInfo
     * - outArgs: #CpswAle_GetVlanEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_LOOKUP_VLAN               = CPSW_ALE_PUBLIC_IOCTL(9U),

    /*!
     * \brief Delete ALE entry info for given unicast/multicast addr
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_MacAddrInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_ADDR               = CPSW_ALE_PUBLIC_IOCTL(10U),

    /*!
     * \brief Delete ALE entry info for given vlanId
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_VlanIdInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_VLAN               = CPSW_ALE_PUBLIC_IOCTL(11U),

    /*!
     * \brief Delete ALE entry info for given oui addr
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_OuiEntryInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_OUI                = CPSW_ALE_PUBLIC_IOCTL(12U),

    /*!
     * \brief Delete ALE entry info for given IPv4 address
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_Ipv4EntryInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_IPV4ADDR           = CPSW_ALE_PUBLIC_IOCTL(13U),

    /*!
     * \brief Delete ALE entry info for given IPv6 address
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_Ipv6EntryInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_IPV6ADDR           = CPSW_ALE_PUBLIC_IOCTL(14U),

    /*!
     * \brief Delete ALE entry info for given Ethertype
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_EtherTypeEntryInfo
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_REMOVE_ETHERTYPE          = CPSW_ALE_PUBLIC_IOCTL(15U),

    /*!
     * \brief Set Policer/Classifier entry
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetPolicerEntryInArgs
     * - outArgs: #CpswAle_SetPolicerEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_POLICER               = CPSW_ALE_PUBLIC_IOCTL(16U),

    /*!
     * \brief Get Policer/Classifier entry info
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_PolicerMatchParams
     * - outArgs: #CpswAle_GetPolicerEntryOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_POLICER               = CPSW_ALE_PUBLIC_IOCTL(17U),

    /*! \brief Delete Policer/Classifier entry
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_PolicerMatchParams
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DEL_POLICER               = CPSW_ALE_PUBLIC_IOCTL(18U),

    /*!
     * \brief Dump ALE table entries
     *
     * Table entries will be printed using CpswUtils_Prms::printFxn param
     * passed at CPSW open time.
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DUMP_TABLE                = CPSW_ALE_PUBLIC_IOCTL(19U),

    /*!
     * \brief Dump ALE policer entries
     *
     * Table entries will be printed using CpswUtils_Prms::printFxn param
     * passed at CPSW open time.
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES      = CPSW_ALE_PUBLIC_IOCTL(20U),

    /*!
     * \brief Get ALE policer stats.
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_GetPolicerStatsInArgs
     * - outArgs: #CpswAle_GetPolicerStatsOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_POLICER_STATS         = CPSW_ALE_PUBLIC_IOCTL(21U),

    /*!
     * \brief Set Port Mirroring configuration
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetPortMirroringConfigInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_PORT_MIRROR_CFG       = CPSW_ALE_PUBLIC_IOCTL(22U),

    /*!
     * \brief Set Port Trunking configuration
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetTrunkConfigInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_TRUNK_CFG             = CPSW_ALE_PUBLIC_IOCTL(23U),

    /*!
     * \brief Enable OAM loopback for ports
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetOAMLoopbackInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_OAMLPBK_CFG           = CPSW_ALE_PUBLIC_IOCTL(24U),

    /*!
     * \brief Delete all ALE entries
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DEL_ALL_ENTRIES           = CPSW_ALE_PUBLIC_IOCTL(25U),

    /*!
     * \brief Set host port Rx filter
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_HostRxFilterParams
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_RX_FILTER             = CPSW_ALE_PUBLIC_IOCTL(26U),

    /*!
     * \brief Get current host port Rx filter
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: #CpswAle_HostRxFilterParams
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_RX_FILTER             = CPSW_ALE_PUBLIC_IOCTL(27U),

    /*!
     * \brief Set ALE port state
     * IOCTL params:
     * -  inArgs: #CpswAle_SetPortStateInArgs
     * - outArgs: None
     * Calling context: ISR/SWI/Task
     */
    CPSW_ALE_IOCTL_SET_PORT_STATE            = CPSW_ALE_PUBLIC_IOCTL(28U),

    /*!
     * \brief Set Broadcast/Multicast rate limit configuration
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetBcastMcastRateLimitInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT     = CPSW_ALE_PUBLIC_IOCTL(29U),

    /*!
     * \brief Set default thread configuration
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetDefaultThreadConfigInArgs
     * - outArgs: None
     *
     * Calling context:Task
     */
    CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG     = CPSW_ALE_PUBLIC_IOCTL(30U),

    /*!
     * \brief Age all ALE entries now
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context:ISR/SWI/Task
     */
    CPSW_ALE_IOCTL_AGE_ALE_ENTRIES           = CPSW_ALE_PUBLIC_IOCTL(31U),

    /*!
     * \brief Set InterVLAN routing configuration
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetInterVLANConfigInArgs
     * - outArgs: #CpswAle_SetInterVLANConfigOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_INTERVLAN_CFG         = CPSW_ALE_PUBLIC_IOCTL(32U),

    /*!
     * \brief Set InterVLAN routing configuration
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_INTERVLAN_CFG         = CPSW_ALE_PUBLIC_IOCTL(33U),

    /*!
     * \brief Delete all learnt ALE entries
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_DelLearntEntriesInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DEL_LEARNT_ENTRIES        = CPSW_ALE_PUBLIC_IOCTL(34U),

    /*!
     * \brief Set thread id for given classifier/policer
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_SetPolicerThreadConfigInArgs
     * - outArgs: #CpswAle_SetPolicerThreadConfigOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_POLICER_THREADCFG     = CPSW_ALE_PUBLIC_IOCTL(35U),

    /*!
     * \brief Get mac addresses reachable on given port
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_GetPortMacAddrInArgs
     * - outArgs: #CpswAle_GetPortMacAddrOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_PORT_MACADDR          = CPSW_ALE_PUBLIC_IOCTL(36U),

    /*!
     * \brief Delete port mirror configuration for a given port
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DISABLE_PORT_MIRROR       = CPSW_ALE_PUBLIC_IOCTL(37U),

    /*!
     * \brief Disable Broadcast/Multicast rate limit configuration
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DISABLE_BCAST_MCAST_LIMIT = CPSW_ALE_PUBLIC_IOCTL(38U),

    /*!
     * \brief Configure policer global settings
     * IOCTL params:
     * -  inArgs: #CpswAle_SetPolicerGlobalConfigInArgs
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_SET_POLICER_GLOBAL_CONFIG = CPSW_ALE_PUBLIC_IOCTL(39U),

    /*!
     * \brief Get configured Broadcast/Multicast rate limit params
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: #CpswAle_GetBcastMcastRateLimitOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT     = CPSW_ALE_PUBLIC_IOCTL(40U),

    /*!
     * \brief Set ALE port state
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_GetPortStateInArgs
     * - outArgs: #CpswAle_GetPortStateOutArgs
     *
     * Calling context: ISR/SWI/Task
     */
    CPSW_ALE_IOCTL_GET_PORT_STATE            = CPSW_ALE_PUBLIC_IOCTL(41U),

    /*!
     * \brief Get default thread configuration
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: #CpswAle_GetDefaultThreadConfigOutArgs
     *
     * Calling context:Task
     */
    CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG     = CPSW_ALE_PUBLIC_IOCTL(42U),

    /*!
     * \brief Dump ALE registers
     *
     * All ALE registers wil be printed using CpswUtils_Prms::printFxn param
     * passed at CPSW open time.
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_PRINT_REGS                = CPSW_ALE_PUBLIC_IOCTL(43U),

    /*!
     * \brief Configure policer global settings
     *
     * IOCTL params:
     * -  inArgs: None
     * - outArgs: #CpswAle_GetPolicerGlobalConfigOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_GET_POLICER_GLOBAL_CONFIG = CPSW_ALE_PUBLIC_IOCTL(44U),

    /*!
     * \brief Delete all policer entries and associated ALE entry with the
     *        given threadId
     * IOCTL params:
     * -  inArgs: uint32_t threadId
     * - outArgs: None
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_DEL_ALL_POLICER_THREADID  = CPSW_ALE_PUBLIC_IOCTL(45U),

    /*!
     * \brief Blacklist classifier to host port
     *
     * ALE supports feature to allow packets matching policer match criteria
     * that is destined for host port to be dropped.
     *
     * IOCTL params:
     * -  inArgs: #CpswAle_PolicerMatchParams
     * - outArgs: #CpswAle_BlockClassifierHostPortOutArgs
     *
     * Calling context: Task
     */
    CPSW_ALE_IOCTL_BLOCK_CLASSIFIER_HOSTPORT = CPSW_ALE_PUBLIC_IOCTL(46U),
} CPSW_ALE_IOCTL_E;
/* @} */

/*!
 * Forward State Level - Indicates the port state(s) required for the received
 * port on a destination address lookup in order for the multicast packet to
 * be forwarded to the transmit port(s). A transmit port must be in the
 * Forwarding state in order to forward the packet. If the transmit
 * port_mask has multiple set bits then each forward decision is
 * independent of the other transmit port(s) forward decision.
 * - 00 - Forwarding
 * - 01 - Blocking/Forwarding/Learning
 * - 10 - Forwarding/Learning
 * - 11 - Forwarding
 */
typedef enum CPSW_ALE_FWDSTLVL_TAG
{
    /*! Forwarding */
    CPSW_ALE_FWDSTLVL_FORWARDING  = 0,

    /*! Blocking/Forwarding/Learning */
    CPSW_ALE_FWDSTLVL_BLK_FWD_LRN = 1,

    /*! Forwarding/Learning */
    CPSW_ALE_FWDSTLVL_FWD_LRN     = 2,
} CPSW_ALE_FWDSTLVL_E;

/*!
 * \brief Yellow Threshold Value
 *
 * When set enables a portion of the yellow packets to be dropped based on the
 * "yellow_drop_en" enable.
*/
typedef enum
{
    /*! Drop 100% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_100,

    /*! Drop 50% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_50,

    /*! Drop 33% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_33,

    /*! Drop 25% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_25,

    /*! Drop 20% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_20,

    /*! Drop 17% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_17,

    /*! Drop 14% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_14,

    /*! Drop 13% of packets colored yellow by policer */
    CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_13,
} CPSW_ALE_POLICER_YELLOWTHRESH_E;

/*!
 * \brief Policing Match Mode
 *
 * This field determines what happens to packets that fail to hit any
 * policing/classifier entry.
 */
typedef enum CPSW_ALE_POLICER_NOMATCH_MODE_TAG
{
    /*! Color packets not matching any classifier entry as GREEN */
    CPSW_ALE_POLICER_NOMATCH_MODE_GREEN,

    /*! Color packets not matching any classifier entry as YELLOW */
    CPSW_ALE_POLICER_NOMATCH_MODE_YELLOW,

    /*! Color packets not matching any classifier entry as RED */
    CPSW_ALE_POLICER_NOMATCH_MODE_RED,

    /*! No Hit packets are marked based on policing for unregulated traffic */
    CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER,
} CPSW_ALE_POLICER_NOMATCH_MODE_E;

/*!
 * \brief Defines ALE port states
 */
typedef enum CPSW_ALE_PORTSTATE_TAG
{
    /*! Discard all packets */
    CPSW_ALE_PORTSTATE_DISABLED = 0,

    /*! Discard all packets except supervisory packets (BPDUs) */
    CPSW_ALE_PORTSTATE_BLOCKED,

    /*! Discard all packets except supervisory packets (BPDUs).
     *  Also add packet MAC SA as learned entries in ALE table */
    CPSW_ALE_PORTSTATE_LEARN,

    /*! Forward all packets.Learning also enabled */
    CPSW_ALE_PORTSTATE_FORWARD
} CPSW_ALE_PORTSTATE_E;

/*!
 * \brief Defines IP address classifier type
 */
typedef enum CPSW_ALE_IPADDR_CLASSIFIER_TAG
{
    /*! Classifier entry is IPv4 address */
    CPSW_ALE_IPADDR_CLASSIFIER_IPV4 = 0x00U,

    /*! Classifier entry is IPv6 address */
    CPSW_ALE_IPADDR_CLASSIFIER_IPV6 = 0x01U,
} CPSW_ALE_IPADDR_CLASSIFIER_E;

/*!
 * \brief Defines IP address classifier type
 */
typedef enum CPSW_ALE_TBLENTRYTYPE_TAG
{
    /*! ALE entry type Unicast/Multicast address */
    CPSW_ALE_TBLENTRYTYPE_ADDR,
 
   /*! ALE entry type Inner/outer vlan */
    CPSW_ALE_TBLENTRYTYPE_VLAN,

    /*! ALE entry type Ethertype */
    CPSW_ALE_TBLENTRYTYPE_ETHERTYPE,

    /*! ALE entry type OUI */
    CPSW_ALE_TBLENTRYTYPE_OUI,

    /*! ALE entry type IPv4 */
    CPSW_ALE_TBLENTRYTYPE_IPV4,

    /*! ALE entry type IPv6 */
    CPSW_ALE_TBLENTRYTYPE_IPV6
} CPSW_ALE_TBLENTRYTYPE_E;

/*!
 * \defgroup CPSW_ALE_CONFIG_MASK ALE Configuration Definitions
 *
 * This group contains mask to configure ALE mode of operation
 *
 * @{
 */

/*! \brief Enable ALE */
#define CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE            (CPSW_SET_BIT(0))

/*! Enable ALE  bypass */
#define CPSW_ALE_CONFIG_MASK_ALE_BYPASS_ENABLE            (CPSW_SET_BIT(1))

/*! Enable Unknown unicast packet flooding to host port */
#define CPSW_ALE_CONFIG_MASK_UNKNOWN_UNICAST_FLOOD2HOST   (CPSW_SET_BIT(2))

/*! @} */

/*!
 * \brief Packet Filtering Settings (cumulative)
 */
typedef enum CPSW_ALE_RXFILTER_TAG
{
    /*! Receive filter set to Nothing */
    CPSW_ALE_RXFILTER_NOTHING      = 0x0000U,

    /*! Receive filter set to Direct */
    CPSW_ALE_RXFILTER_DIRECT       = 0x0001U,

    /*! Receive filter set to Broadcast */
    CPSW_ALE_RXFILTER_BROADCAST    = 0x0002U,

    /*! Receive filter set to Multicast */
    CPSW_ALE_RXFILTER_MULTICAST    = 0x0003U,

    /*! Receive filter set to All Mcast */
    CPSW_ALE_RXFILTER_ALLMULTICAST = 0x0004U,

    /*! Receive filter set to All */
    CPSW_ALE_RXFILTER_ALL          = 0x0005U
} CPSW_ALE_RXFILTER_E;

/*!
 *  @}
 */

/*!
 * \brief Structure holding MAC address and VLAN Id
 */
typedef struct CpswAle_MacAddrInfo_s
{
    /*! 48-bit unicast/multicast mac address */
    uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS];

    /*! VLAN ID associated with the Adddress entry. vlanId of 0 indicates
     *  do not use VLAN */
    uint32_t vlanId;
} CpswAle_MacAddrInfo;

/*!
 * \brief IPv4 ALE entry info
 */
typedef struct CpswAle_Ipv4EntryInfo_s
{
    /*! IPv4 address */
    uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS];

    /*! CIDR mask in term of number of LSBs of IPaddr to ignore */
    uint32_t numLSBIgnoreBits;
} CpswAle_Ipv4EntryInfo;

/*!
 * \brief IPv6 ALE entry info
 */
typedef struct CpswAle_Ipv6EntryInfo_s
{
    /*! IPv6 address */
    uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS];

    /*! CIDR mask in term of number of LSBs of IPaddr to ignore */
    uint32_t numLSBIgnoreBits;
} CpswAle_Ipv6EntryInfo;

/*!
 * \brief ALE Multicast entry type info apart from MAC address
 */
typedef struct CpswAle_McastEntryInfo_s
{
    /*! Supervisory Flag - Indicates that this address can be received in port
     *  state other than Forwarding state */
    bool superFlag;

    /*! Forward State Level */
    CPSW_ALE_FWDSTLVL_E fwdState;

    /*! Pport bit mask that is returned with a found multicast destination
     *  address */
    uint32_t portMask;

    /*! Number of bits in the address to ignore. Up to 10 least significant bits
     *  can be ignored to form a multicast address range */
    uint32_t numIgnBits;
} CpswAle_McastEntryInfo;

/*!
 * \brief ALE unicast entry type info apart from MAC address that can be set
 */
typedef struct CpswAle_UcastEntryInfo_s
{
    /*! Port Number that the packet with a unicast destination address is
     *  forwarded to */
    uint32_t portNum;

    /*! Indicates that packet with a matching source or destination address
     *  should be dropped */
    bool blocked;

    /*! Indicates that packet with a matching source address should be dropped,
     *  if the received port number is not equal to the table entry port_number */
    bool secure;

    /*! Supervisory Flag - Indicates that this address can be received in port
     *  state other than Forwarding state */
    bool super;

    /*! Ageable Flag - Indicates if the Unicast entry is subject to aging */
    bool ageable;

    /*! Trunk flag - Indicates if the portNum refers to trunk number of
     *  port number */
    bool trunk;
} CpswAle_UcastEntryInfo;

/*!
 * \brief ALE IP address classifier info
 */
typedef struct CpswAle_IpAddrClassifierInfo_s
{
    /*! IPv4/IPv6 address type identifier */
    CPSW_ALE_IPADDR_CLASSIFIER_E ipAddrtype;

    /*! IPv4 address for classifier */
    CpswAle_Ipv4EntryInfo ipv4;

    /*! IPv6 address for classifier */
    CpswAle_Ipv6EntryInfo ipv6;
} CpswAle_IpAddrClassifierInfo;

/*!
 * \brief ALE OUI entry info
 */
typedef struct CpswAle_OuiEntryInfo_s
{
    /*! Ethernet frame Organization Unique Id portion of MAC address */
    uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS];
} CpswAle_OuiEntryInfo;

/*!
 * \brief ALE Ethertype entry info
 */
typedef struct CpswAle_EtherTypeEntryInfo_s
{
    /*! Ethernet frame Ethertype field value */
    uint32_t etherType;
} CpswAle_EtherTypeEntryInfo;

/*!
 * \brief ALE vlan identifier info
 */
typedef struct CpswAle_VlanIdInfo_s
{
    /*! VLAN Id to add */
    uint32_t vlanId;

    /*! Flag indicating VLAN type is outer or inner VLAN */
    bool outerVlanFlag;
} CpswAle_VlanIdInfo;

/*!
 * \brief ALE VLAN entry type info
 */
typedef struct CpswAle_VlanEntryInfo_s
{
    /*! VLAN id / VLAN type to be added */
    CpswAle_VlanIdInfo vlanIdInfo;

    /*! Port member mask for the VLAN entry being added */
    uint32_t vlanMemberList;

    /*! Unregistered multicast flood mask - Mask used for multicast when the
     *  multicast address is not found */
    uint32_t unregMcastFloodMask;

    /*! Registered multicast flood mask - Mask used for multicast when the
     *  multicast address is found */
    uint32_t regMcastFloodMask;

    /*! Force untagged egress bit flags - Causes the packet VLAN tag to be
     *  removed on egress */
    uint32_t forceUntaggedEgressMask;

    /*! VLAN No Learn Mask - When a bit is set in this mask, a packet with an
     *  unknown source address received on the associated port will not be
     *  learned (i.e. When a VLAN packet is received and the source address is
     *  not in the table, the source address will not be added to the table) */
    uint32_t noLearnMask;

    /*! VLAN Ingress Check - When set, if the receive port is not a member of
     *  this VLAN then the packet is dropped. This is similar to the
     *  vid_ingress_check bit in the ale_port_control register except this check
     *  is for this VLAN only (not all VLANs) */
    bool vidIngressCheck;

    /*! Limit IP NXT hdr field - When set IP packets only with configured
     *  NXTHDR will be allowed @sa CpswAle_IPPktSecurityConfig */
    bool limitIPNxtHdr;

    /*! VLAN No IPv4 Fragmented frames Control - Causes IPv4 fragmented IP
     *  frames to be dropped */
    bool disallowIPFragmentation;
} CpswAle_VlanEntryInfo;

/*!
 * \brief ALE classifier/policer match parameters for srcMacAddr
 */
typedef struct CpswAle_MacSrcAddrClassifierInfo_s
{
    /*! Source MAC  address to match */
    CpswAle_MacAddrInfo addr;

    /*! Ingress port number */
    uint32_t ingressPortNum;
} CpswAle_MacSrcAddrClassifierInfo;

/*!
 * \brief ALE classifier/policer match parameters for dstMacAddr
 */
typedef struct CpswAle_MacDstAddrClassifierInfo_s
{
    /*! Destination MAC address to match */
    CpswAle_MacAddrInfo addr;

    /*! Egress port number */
    uint32_t egressPortNum;
} CpswAle_MacDstAddrClassifierInfo;

/*!
 * \brief ALE Classifier/Policer Match parameters
 */
typedef struct CpswAle_PolicerMatchParams_s
{
    /*! Bitmask selecting classifier types to enable
     * @sa ALE classifier match types */
    uint32_t policerMatchEnableMask;

    /*! Port number to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_PORT in
     *  policerMatchEnableMask is set */
    uint32_t portNum;

    /*! Flag indicating port number is a trunk group */
    bool portIsTrunk;

    /*! Received packet priority to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_PRIORITY in
     *  policerMatchEnableMask is set */
    uint32_t priority;

    /*! OUI address to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_OUI in
     *  policerMatchEnableMask is set */
    CpswAle_OuiEntryInfo oui;

    /*! Source MAC  address to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_MACSRC in
     *  policerMatchEnableMask is set */
    CpswAle_MacSrcAddrClassifierInfo srcMacAddr;

    /*! Destination MAC address to match.
     *  Must valid if bit CPSW_ALE_POLICER_MATCH_MACDST in
     *  policerMatchEnableMask is set */
    CpswAle_MacDstAddrClassifierInfo dstMacAddr;

    /*! Inner VLAN ID to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_IVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ivlanId;

    /*! Outer VLAN ID to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_OVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ovlanId;

    /*! Frame Ethertype to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_ETHERTYPE in
     *  policerMatchEnableMask is set */
    CpswAle_EtherTypeEntryInfo etherType;

    /*! Source IP address to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_IPSRC in
     *  policerMatchEnableMask is set */
    CpswAle_IpAddrClassifierInfo srcIp;

    /*! Destination IP address to match.
     *  Must be valid if bit CPSW_ALE_POLICER_MATCH_IPDST in
     *  policerMatchEnableMask is set */
    CpswAle_IpAddrClassifierInfo dstIp;
} CpswAle_PolicerMatchParams;

/*!
 * \brief Host port receive packet type filter params
 */
typedef struct CpswAle_HostRxFilterParams_s
{
    /*! Receive packet filter type to be applied to host port */
    CPSW_ALE_RXFILTER_E rxFilter;
} CpswAle_HostRxFilterParams;

/*!
 * \brief ALE port mirroring configration for mirroring based on packet match
 */
typedef struct CpswAle_MirrorMatchParams_s
{
    /*! Type of ALE entry for matching and mirroring */
    CPSW_ALE_TBLENTRYTYPE_E entryType;

    /*! Unicast/Multicast address.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_ADDR */
    CpswAle_MacDstAddrClassifierInfo macAddrInfo;

    /*! VLAN id info.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_VLAN */
    CpswAle_VlanIdInfo vlanIdInfo;

    /*! IPv4 address info.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_IPV4 */
    CpswAle_Ipv4EntryInfo ipv4Info;

    /*! IPv6 address info.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_IPV6 */
    CpswAle_Ipv6EntryInfo ipv6Info;

    /*! OUI address info.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_OUI */
    CpswAle_OuiEntryInfo ouiInfo;

    /*! Ethertype address info.
     *  Must be valid if entryType is #CPSW_ALE_TBLENTRYTYPE_ETHERTYPE */
    CpswAle_EtherTypeEntryInfo etherTypeInfo;
} CpswAle_MirrorMatchParams;

/*!
 * \brief Structure for port default VLAN configuration
 */
typedef CpswAle_VlanEntryInfo CpswAle_PortVlanConfig;

/*!
 * \brief Structure for bcast/mcast bw limit configuration parameters
 */
typedef struct  CpswAle_PortBcastMcastRateLimitParams_s
{
    /*! Port number for which BCAST/MCAST rate limiting cfg is to be done */
    uint32_t portNum;

    /*! Enable Broadcast Rate Limit for port.
     *  If set to TRUE, #bcastLimitNumPktsPerSec should be non-zero */
    bool enableBRateLimitForPort;

    /*! Enable Multicast Rate Limit for port.
     *  If set to TRUE, #mcastLimitNumPktsPerSec should be non-zero */
    bool enableMRateLimitForPort;

    /*! Broadcast packet limit per sec */
    uint32_t bcastLimitNumPktsPerSec;

    /*! Multicast packet limit per sec */
    uint32_t mcastLimitNumPktsPerSec;
} CpswAle_PortBcastMcastRateLimitParams;

/*!
 * \brief Structure containing unregulated traffic (No classifier hit) policing
 *        params
 */
typedef struct CpswAle_UnregulatedTrafficPolicer_s
{
    /*! Peak rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t peakRateInBitsPerSec;

    /*! Commit rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t commitRateInBitsPerSec;
} CpswAle_UnregulatedTrafficPolicer;

/*!
 * \addtogroup CPSW_ALE_ENUMS
 *
 * @{
 */

/*!
 * \brief Output args for IOCTL cmds that add ALE entry, such as:
 *        #CPSW_ALE_IOCTL_ADD_UNICAST,
 *        #CPSW_ALE_IOCTL_ADD_VLAN,
 *        #CPSW_ALE_IOCTL_ADD_ETHERTYPE,
 *        #CPSW_ALE_IOCTL_ADD_IPV4ADDR,
 *        #CPSW_ALE_IOCTL_ADD_IPV6ADDR,
 *        #CPSW_ALE_IOCTL_ADD_OUI,
 *        #CPSW_ALE_IOCTL_ADD_MULTICAST
 */
typedef struct CpswAle_AddEntryOutArgs_s
{
    /*! ALE entry index */
    uint32_t aleEntryIndex;
} CpswAle_AddEntryOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_ADD_UNICAST IOCTL command
 */
typedef struct CpswAle_SetUcastEntryInArgs_s
{
    /*! Address info to be added */
    CpswAle_MacAddrInfo addr;

    /*! Unicast entry additional info to be added */
    CpswAle_UcastEntryInfo info;
} CpswAle_SetUcastEntryInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_LOOKUP_UNICAST IOCTL command
 */
typedef struct CpswAle_GetUcastEntryOutArgs_s
{
    /*! ALE table entry info associated with the given unicast address */
    CpswAle_UcastEntryInfo unicastInfo;

    /*! Indicates if packet with MAC SA = entry address has been received
     *  since last ALE aging interval */
    bool touched;

    /*! ALE entry index */
    uint32_t aleEntryIndex;
} CpswAle_GetUcastEntryOutArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_ADD_MULTICAST IOCTL command
 */
typedef struct CpswAle_SetMcastEntryInArgs_s
{
    /*! Multicast mac address to be added */
    CpswAle_MacAddrInfo addr;

    /*! Additional info to be added to ALE table for multicast entry */
    CpswAle_McastEntryInfo info;
} CpswAle_SetMcastEntryInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_LOOKUP_MULTICAST IOCTL command
 */
typedef struct CpswAle_GetMcastEntryInArgs_s
{
    /*! Multicast mac address to be added */
    CpswAle_MacAddrInfo addr;

    /*! Number of bits in the address to ignore */
    uint32_t numIgnBits;
} CpswAle_GetMcastEntryInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_LOOKUP_MULTICAST IOCTL command
 */
typedef struct CpswAle_GetMcastEntryOutArgs_s
{
    /*! ALE table entry info associated with the given multicast address */
    CpswAle_McastEntryInfo outInfo;

    /*! ALE entry index */
    uint32_t aleIndex;
} CpswAle_GetMcastEntryOutArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_LOOKUP_MULTICAST IOCTL command
 */
typedef struct CpswAle_GetVlanEntryOutArgs_s
{
    /*! Port member mask for the VLAN entry being queried */
    uint32_t vlanMemberList;

    /*! Unregistered multicast flood mask for queried VLAN */
    uint32_t unregMcastFloodMask;

    /*! Registered multicast flood mask for queried VLAN  */
    uint32_t regMcastFloodMask;

    /*! Force untagged egress port mask for queried VLAN */
    uint32_t forceUntaggedEgressMask;

    /*! VLAN No Learn Mask for queried VLAN */
    uint32_t noLearnMask;

    /*! VLAN Ingress Check enabled/disabled for queried VLAN */
    bool vidIngressCheck;

    /*! Limit IP NXT hdr field enabled/disabled for queried VLAN */
    bool limitIPNxtHdr;

    /*! VLAN No IPv4 Fragmented frames control enabled/disabled for queried VLAN */
    bool disallowIPFragmentation;

    /*! ALE entry index for queried VLAN */
    uint32_t aleEntryIndex;
} CpswAle_GetVlanEntryOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_POLICER command
 */
typedef struct CpswAle_SetPolicerEntryInArgs_s
{
    /*! Policer match config */
    CpswAle_PolicerMatchParams policerMatch;

    /*! Enable threadid setting for this policer entry */
    bool threadIdEnable;

    /*! Thread Id which will be enabled for this policer match */
    uint32_t threadId;

    /*! Peak rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t peakRateInBitsPerSec;

    /*! Commit rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t commitRateInBitsPerSec;
} CpswAle_SetPolicerEntryInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_POLICER command
 */
typedef struct CpswAle_SetPolicerEntryOutArgs_s
{
    /*! OUI ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_OUI in
     *  policerMatchEnableMask is set */
    uint32_t ouiAleEntryIndex;

    /*! MAC address ALE table entry index to match source MAC address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_MACSRC in
     *  policerMatchEnableMask is set */
    uint32_t srcMacAleEntryIndex;

    /*! MAC address ALE table entry index to match destination MAC address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_MACDST in
     *  policerMatchEnableMask is set */
    uint32_t dstMacAleEntryIndex;

    /*! Inner VLAN ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ivlanAleEntryIndex;

    /*! Outer VLAN ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_OVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ovlanAleEntryIndex;

    /*! Ethertype ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_ETHERTYPE in
     *  policerMatchEnableMask is set */
    uint32_t etherTypeAleEntryIndex;

    /*! IP address ALE table entry index to match as source IP address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IPSRC in
     *  policerMatchEnableMask is set */
    uint32_t srcIpAleEntryIndex;

    /*! IP address ALE table entry index to match as destination IP address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IPDST in
     *  policerMatchEnableMask is set */
    uint32_t dstIpAleEntryIndex;

    /*! ALE policer table entry index */
    uint32_t policerEntryIndex;
} CpswAle_SetPolicerEntryOutArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_POLICER command
 */
typedef struct CpswAle_GetPolicerEntryOutArgs_s
{
    /*! Bitmask selecting classifier types to enabled for this policer
     * @sa ALE classifier match types */
    uint32_t policerMatchEnableMask;

    /*! Port number to match.
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_PORT in
     *  policerMatchEnableMask is set */
    uint32_t port;

    /*! Flag indicating port number is a trunk group
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_PORT in
     *  policerMatchEnableMask is set */
    bool portIsTrunk;

    /*! Received packet priority to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_PRIORITY in
     *  policerMatchEnableMask is set */
    uint32_t priority;

    /*! OUI ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_OUI in
     *  policerMatchEnableMask is set */
    uint32_t ouiAleEntryIndex;

    /*! MAC address ALE table entry index to match source MAC address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_MACSRC in
     *  policerMatchEnableMask is set */
    uint32_t srcMacAleEntryIndex;

    /*! MAC address ALE table entry index to match destination MAC address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_MACDST in
     *  policerMatchEnableMask is set */
    uint32_t dstMacAleEntryIndex;

    /*! Inner VLAN ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ivlanAleEntryIndex;

    /*! Outer VLAN ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_OVLAN in
     *  policerMatchEnableMask is set */
    uint32_t ovlanAleEntryIndex;

    /*! Ethertype ALE table entry index to match
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_ETHERTYPE in
     *  policerMatchEnableMask is set */
    uint32_t etherTypeAleEntryIndex;

    /*! IP address ALE table entry index to match as source IP address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IPSRC in
     *  policerMatchEnableMask is set */
    uint32_t srcIpAleEntryIndex;

    /*! IP address ALE table entry index to match as destination IP address
     *  Is valid if bit CPSW_ALE_POLICER_MATCH_IPDST in
     *  policerMatchEnableMask is set */
    uint32_t dstIpAleEntryIndex;

    /*! Flag indicating if threadid setting is enabled for this policer entry */
    bool threadIdEnable;

    /*! Thread Id which will be enabled for this policer match.
     *  Valid if #threadIdEnable is TRUE */
    uint32_t threadId;

    /*! Peak rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t peakRateInBitsPerSec;

    /*! Commit rate in bits per second. 0 indicates rate limit is disabled */
    uint32_t commitRateInBitsPerSec;

    /*! ALE policer table entry index */
    uint32_t policerEntryIndex;

    /*! Egress Op Enabled */
    bool egressOpEnabled;

    /*! Egress Op code:
     *  - 0x00 - Egress Op disabled
     *  - 0xFF - Egress Op : OAM loopback enabled
     *  - 0x01 - #CPSW_MACPORT_INTERVLAN_ROUTEID_LAST - InterVLAN route enabled
     */
    uint32_t egressOpcode;

    /*! TTL check enabled for the interVLAN route. Valid only if
     *  egress op code has interVLan route enabled */
    bool enableTTLCheck;

    /*! Trunk index to be used if destPort for itnerVLAN route is a trunk.
     *  Valid only if egress op code has interVLan route enabled */
    uint32_t egressTrunkIndex;

    /*! Port mask to which the packet is to be routeed.
     *  Valid only if egress op code has interVLan route enabled */
    uint32_t destPortMask;
} CpswAle_GetPolicerEntryOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_DEL_POLICER IOCTL command
 */
typedef struct CpswAle_DelPolicerEntryInArgs_s
{
    /*! Policer match config */
    CpswAle_PolicerMatchParams policerMatch;

    /*! Bitmask to indicate which ALE entries associated with the policer
     *  have to be deleted
     *  @sa ALE classifier aleEntry delete bitmask */
    uint32_t delAleEntryMask;
} CpswAle_DelPolicerEntryInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_OAMLPBK_CFG IOCTL command
 */
typedef struct CpswAle_SetOAMLoopbackInArgs_s
{
    /*! Port mask on which ports OAM loopback is to be enabled */
    uint32_t enableLpbkPortMask;
} CpswAle_SetOAMLoopbackInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_PORT_STATE IOCTL command
 */
typedef struct  CpswAle_SetPortStateInArgs_s
{
    /*! Port number for which port state to be set */
    uint32_t portNum;

    /*! ALE port state to be set */
    CPSW_ALE_PORTSTATE_E portState;
} CpswAle_SetPortStateInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_GET_PORT_STATE IOCTL command
 */
typedef struct  CpswAle_GetPortStateInArgs_s
{
    /*! Port number for which port state to be set */
    uint32_t portNum;
} CpswAle_GetPortStateInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_PORT_STATE IOCTL command
 */
typedef struct  CpswAle_GetPortStateOutArgs_s
{
    /*! ALE port state of the queried port */
    CPSW_ALE_PORTSTATE_E portState;
} CpswAle_GetPortStateOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_PORT_MIRROR_CFG command
 */
typedef struct CpswAle_SetPortMirroringConfigInArgs_s
{
    /*! Enables the source port mirror option. When this bit is set any port
     *  with the pX_mirror_sp set in the ALE Port Control registers set will
     *  have its received traffic also sent to the mirror_top port */
    bool enableSrc;

    /*! Enables the destination port mirror option. When this bit is set any
     *  traffic destined for the mirror_dp port will have its transmit traffic
     *  also sent to the mirror_top port */
    bool enableDst;

    /*! Enables the match mirror option. When this bit is set any traffic whose
     *  destination, source, VLAN or OUI matches the mirror_midx entry index will
     *  have that traffic also sent to the mirror_top port */
    bool enableMatch;

    /*! The port to which destination traffic destined will be duplicated */
    uint32_t destPortNum;

    /*! The destination port for the mirror traffic */
    uint32_t toPortNum;

    /*! ALE lookup table entry info that when a match occurs will cause this
     *  traffic to be mirrored to the mirror_top port.
     *  Must be valid if #enableMatch is TRUE */
    CpswAle_MirrorMatchParams matchParams;

    /*! Source port number mirroring enable bitmask */
    uint32_t srcPortNumMask;
} CpswAle_SetPortMirroringConfigInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_TRUNK_CFG command
 */
typedef struct CpswAle_SetTrunkConfigInArgs_s
{
    /*! Trunk group ID value */
    uint32_t trunkId;

    /*! Number of ports in the trunk group */
    uint32_t numPorts;

    /*! Trunk Group index value */
    uint32_t trunkPortIndex[CPSW_ALE_NUM_PORTS];

    /*! Hash formula starting value */
    uint32_t trunkHashBase;

    /*! Enables the destination MAC address to be used with the hash function
     *  G(X) = 1 + X + X^3
     *  and affect the trunk port transmit link determination */
    bool enableDstIP;

    /*! Enables the source MAC address to be used with the hash function
     *  G(X) = 1 + X + X^3
     *  and affect the trunk port transmit link determination */
    bool enableSrcIP;

    /*! Enables the inner VLAN ID value (C-VLANID) to be used with the hash
     *  function
     *  G(X) = 1 + X+ X^3
     *  and affect the trunk port transmit link determination */
    bool enableInnerVLAN;

    /*! Enables the VLAN Priority bits to be used with the hash function
     *  G(X) = 1 + X + X^3
     *  and affect the trunk port transmit link determination */
    bool enablePri;

    /*! Enables the source IP address to be used with the hash function
     *  G(X) = 1 + X + X^3
     * and affect the trunk port transmit link determination */
    bool enableSrc;

    /*! enables the destination IP address to be used with the hash function
     * G(X) = 1 + X + X^3
     * and affect the trunk port transmit link determination */
    bool enableDst;
} CpswAle_SetTrunkConfigInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG IOCTL command
 */
typedef struct CpswAle_SetDefaultThreadConfigInArgs_s
{
    /*! Default thread enable - When set the switch will use the defthreadval for
     *  the host interface thread ID if no classifier is matched. If clear the
     *  switch will generate its own thread ID based on port and
     *  priority if there is no classifier match */
    bool defaultThreadEnable;

    /*! This field specifies the default thread ID value. */
    uint32_t threadId;

    /*! OR priority of packet with default thread */
    bool enablePriorityOR;

    /*! Do not use default thread for mac only port */
    bool disableMacPortDefaultThread;
} CpswAle_SetDefaultThreadConfigInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG IOCTL command
 */
typedef CpswAle_SetDefaultThreadConfigInArgs CpswAle_GetDefaultThreadConfigOutArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG IOCTL command
 *
 * InterVLAN routing configuration in ALE is associated with each
 * Policer/Classifier entry.
 */
typedef struct CpswAle_SetInterVLANConfigInArgs_s
{
    /*! Policer match config */
    CpswAle_PolicerMatchParams policerMatch;

    /*! Enable TTL field check for non-zero value */
    bool enableTTLCheck;

    /*! Destination Port mask to route the packet matching #policerMatch */
    uint32_t destPortMask;

    /*! Destination port route index which has info on the MAC SA/DA to be
     *  replaced for the out going packet */
    uint32_t routeIndex;

    /*! The Egress Trunk Index is the calculated trunk index from the SA,DA or
     *  VLAN if modified to that InterVLAN routing will work on trunks as well.
     *  The DA, SA and VLAN are ignored for trunk generation on InterVLAN
     *  Routing so that this field is the index generated from the Egress Op
     *  replacements elclusive or'd together into a three bit index */
    uint32_t egressTrunkIndex;
} CpswAle_SetInterVLANConfigInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_SET_INTERVLAN_CFG IOCTL command
 */
typedef CpswAle_GetPolicerEntryOutArgs CpswAle_SetInterVLANConfigOutArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_BLOCK_CLASSIFIER_HOSTPORT IOCTL command
 */
typedef CpswAle_GetPolicerEntryOutArgs CpswAle_BlockClassifierHostPortOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_GET_POLICER_STATS IOCTL command
 */
typedef struct CpswAle_GetPolicerStatsInArgs_s
{
    /*! Policer info for which statistics is queried */
    CpswAle_PolicerMatchParams policerInfo;

    /*! Flag for clearing the policer stats */
    bool clearStats;
} CpswAle_GetPolicerStatsInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_POLICER_STATS IOCTL command
 */
typedef struct CpswAle_GetPolicerStatsOutArgs_s
{
    /*! This indicates that the selected policing/classifier via the pol_test_idx
     *  field has been hit by a packet seen on any port that matches the
     *  policing/classifier entry match */
    bool policerHit;

    /*! This indicates that the selected policing/classifier via the pol_test_idx
     *  field has been hit during a RED condition by a packet seen on any port
     * that matches the policing/classifier entry match */
    bool policerRedHit;

    /*! This indicates that the selected policing/classifier via the pol_test_idx
     *  field has been hit during a YELLOW condition by a packet seen on any port
     *  that matches the policing/classifier entry match */
    bool policerYellowHit;
} CpswAle_GetPolicerStatsOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT IOCTL command
 */
typedef struct CpswAle_SetBcastMcastRateLimitInArgs_s
{
    /*! Rate Limit to be done at Transmit or Receive port */
    bool rateLimitAtTxPort;

    /*! Number of port for which BCAST/MCAST rate limitng config to be applied */
    uint32_t numPorts;

    /*! Per port BCAST/MCAST rate limit params */
    CpswAle_PortBcastMcastRateLimitParams portPrms[CPSW_ALE_NUM_PORTS];
} CpswAle_SetBcastMcastRateLimitInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT IOCTL command
 */
typedef struct CpswAle_GetBcastMcastRateLimitOutArgs_s
{
    /*! Flag indicating if BCAST/MCAST rate limit is enabled */
    bool rateLimitEnabled;

    /*! Rate Limit to be done at Transmit or Receive port */
    bool rateLimitAtTxPort;

    /*! Number of port for which BCAST/MCAST rate limitng config to be applied */
    uint32_t numPorts;

    /*! Per port BCAST/MCAST rate limit params */
    CpswAle_PortBcastMcastRateLimitParams portPrms[CPSW_ALE_NUM_PORTS];
} CpswAle_GetBcastMcastRateLimitOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_GET_PORT_MACADDR IOCTL command
 */
typedef struct CpswAle_GetPortMacAddrInArgs_s
{
    /*! Port number for which MAC addresses are being queried */
    uint32_t portNum;

    /*! Application provided buffer into which MAC addresses will be populated */
    CpswAle_MacAddrInfo *macAddr;

    /*! Max number of MAC entries that can be stored in provided buffer */
    uint32_t maxNumAddr;
} CpswAle_GetPortMacAddrInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_PORT_MACADDR IOCTL command
 */
typedef struct CpswAle_GetPortMacAddrOutArgs_s
{
    /*! Application provided buffer populated with  MAC addresses
     *  reachable on the given portNum */
    CpswAle_MacAddrInfo *macAddr;

    /*! Number of mac entries stored in provided buffer.Will not exceed
     *  CpswAle_GetPortMacAddrInArgs::maxNumAddr */
    uint32_t addrCount;

    /*! Total number of MAC addresses reachable on given portNum */
    uint32_t totalPortMacAddrCount;
} CpswAle_GetPortMacAddrOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_DEL_LEARNT_ENTRIES command
 */
typedef struct CpswAle_DelLearntEntriesInArgs_s
{
    /*! Port number for which MAC addresses are being queried */
    uint32_t portNum;
} CpswAle_DelLearntEntriesInArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_POLICER_THREADCFG command
 *
 * Configuration to set the thread/flow id associated with a classifier.
 */
typedef struct CpswAle_SetPolicerThreadConfigInArgs_s
{
    /*! Policer match config */
    CpswAle_PolicerMatchParams policerMatch;

    /*! Enable/Disable threadid setting for this policer entry */
    bool threadIdEnable;

    /*! Thread Id which will be enabled for this policer match */
    uint32_t threadId;
} CpswAle_SetPolicerThreadConfigInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_SET_POLICER_THREADCFG IOCTL command
 */
typedef CpswAle_GetPolicerEntryOutArgs CpswAle_SetPolicerThreadConfigOutArgs;

/*!
 * \brief Input args for #CPSW_ALE_IOCTL_SET_POLICER_GLOBAL_CONFIG IOCTL command
 */
typedef struct CpswAle_PolicerGlobalConfig_s CpswAle_SetPolicerGlobalConfigInArgs;

/*!
 * \brief Output args for #CPSW_ALE_IOCTL_GET_POLICER_GLOBAL_CONFIG IOCTL command
 */
typedef struct CpswAle_PolicerGlobalConfig_s CpswAle_GetPolicerGlobalConfigOutArgs;

/*  @} */

/*!
 * \addtogroup  CPSW_ALE_OPENPARAMS
 *
 * @{
 */

/*!
 * \brief Structure for policer global configuration parameters
 */
typedef struct CpswAle_PolicerGlobalConfig_s
{
    /*! Enables the policing to color the packets,this also enables red or
     *  yellow drop capabilities */
    bool policingEnable;

    /*! Enables the ALE to drop yellow packets based on the yellow thresh value */
    bool yellowDropEnable;

    /*! Enables the ALE to drop the red colored packets */
    bool redDropEnable;

    /*! Yellow Threshold Value */
    CPSW_ALE_POLICER_YELLOWTHRESH_E yellowThresh;

    /*! Policing match mode */
    CPSW_ALE_POLICER_NOMATCH_MODE_E policerNoMatchMode;

    /*! Policing params for no hit (unregulated traffic.
     *  Must be valid if policerNoMatchMode is set to
     *  #CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER
     */
    CpswAle_UnregulatedTrafficPolicer noMatchPolicer;
} CpswAle_PolicerGlobalConfig;

/*!
 * \brief Structure for ALE aging time configuration
 */
typedef struct CpswAle_AgingConfig_s
{
    /*! Flag to enable auto aging */
    bool enableAutoAging;

    /*! Aging period in ms */
    uint32_t agingPeriodInMs;
} CpswAle_AgingConfig;

/*!
 * \brief Structure for Init time VLAN configuration
 */
typedef struct CpswAle_InitVlanConfig_s
{
    /*! Configure ALE to operate in VLAN aware mode */
    bool aleVlanAwareMode;

    /*! Flag indicating if CPSW is in VLAN aware mode */
    bool cpswVlanAwareMode;

    /*! Learn VLAN with address when autolearning addresses */
    bool autoLearnWithVLAN;

    /*! Do not learn addresses with unknown VLAN */
    bool unknownVlanNoLearn;

    /*! Unknown VLAN Force Untagged Egress */
    uint32_t unknownForceUntaggedEgressMask;

    /*! Unknown VLAN Registered Multicast Flood Mask */
    uint32_t unknownRegMcastFloodMask;

    /*! Unknown VLAN Multicast Flood Mask */
    uint32_t unknownUnregMcastFloodMask;

    /*! Unknown VLAN Member List */
    uint32_t unknownVlanMemberListMask;
} CpswAle_InitVlanConfig;

/*!
 * \brief Structure for port MAC mode configuration
 */
typedef struct CpswAle_PortMacModeConfig_s
{
    /*! MAC-only Copy All Frames:
     *  - When set a Mac Only port will transfer all received good frames to the
     *    host
     *  - When clear a Mac Only port will transfer packets to the host based on
     *    ALE destination address lookup operation (which operates more like an
     *    Ethernet MAC) */
    bool macOnlyCafEnable;

    /*! MAC-only.
     *  When set enables this port be treated like a MAC port for the host.
     *  All traffic received is only sent to the host. The host must direct
     *  traffic to this port as the lookup engine will not send traffic to the
     *  ports with the p0_maconly bit set and the p0_no_learn also set.
     *  If p0_maconly bit is set and the p0_no_learn is not set, the host can
     *  send non-directed packets that can be sent to the destination of a
     *  MacOnly port. It is also possible that The host can broadcast to 
     *  ports including MacOnly ports in this mode */
    bool macOnlyEnable;
} CpswAle_PortMacModeConfig;

/*!
 * \brief Structure for MAC authentication configuration
 */
typedef struct CpswAle_MacAuthConfig_s
{
    /*! Enable MAC Authorization mode */
    bool enableAuthMode;

    /*! Disable Port authorization - When set will allow unknown addresses to
     *  arrive on a switch in authorization mode. It is intended for device to
     *  device network connection on ports which do not require MACSEC
     *  encryption */
    uint32_t macAuthDisableMask;
} CpswAle_MacAuthConfig;

/*!
 * \brief Structure for Port specific VLAN security configuration
 */
typedef struct CpswAle_PortVlanSecurityConfig_s
{
    /*! VLAN ID Ingress Check */
    bool vidIngressCheck;

    /*! Drop untagged packets */
    bool dropUntagged;

    /*! Drop Dual VLAN - When set will cause any received packet with dual VLAN
     *  stag followed by ctag to be dropped */
    bool dropDualVlan;

    /*! Drop Double VLAN - When set cause any received packet with double VLANs
     *  to be dropped. That is if there are two ctag or two stag fields in the
     * packet it will be dropped */
    bool dropDoubleVlan;
} CpswAle_PortVlanSecurityConfig;

/*!
 * \brief Structure for Port specific auto learning security configuration
 */
typedef struct CpswAle_PortLearningSecurityConfig_s
{
    /*! No-learn mode */
    bool noLearn;

    /*! No Source Address Update - When set will not update the source addresses
     *  for this port */
    bool noSaUpdateEnable;
} CpswAle_PortLearningSecurityConfig;

/*!
 * \brief Structure for IP packet security configuration
 */
typedef struct CpswAle_IPPktSecurityConfig_s
{
    /*! Enable dropping of packets if a VLAN entry is not found */
    bool enableDefaultNoFragment;

    /*! Enable dropping of packets if VLAN not found AND IP NXT HDR does not
     *  match one of #ipNxtHdrWhitelist */
    bool enableDefaultNxtHdrWhitelist;

    /*! Number of ip Nxt Hdr to whitelist.
     * Must be <= CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR */
    uint32_t ipNxtHdrWhitelistCount;

    /*! IP NXT HDR whitelist */
    uint8_t ipNxtHdrWhitelist[CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR];
} CpswAle_IPPktSecurityConfig;

/*!
 * \brief Structure for malformed packet handling configuration
 */
typedef struct CpswAle_MalformedPktSecurityConfig_s
{
    /*! Disable dropping of packets with Multicast address as SA */
    bool disableSrcMcastDrop;

    /*! Enable dropping of packets where the 802.3 length field is larger than
     *  the packet. Ethertypes 0-1500 are 802.3 lengths, all others are
     *  EtherTypes */
    bool enableBadLenPktDrop;
} CpswAle_MalformedPktSecurityConfig;

/*!
 * \brief Structure for ALE module level auto learning security configuration
 */
typedef struct CpswAle_LearningSecurityConfig_s
{
    /*! Update Static Entries - A static Entry is an entry that is not agable.
     *  - When clear this bit will prevent any static entry (agable bit clear)
     *    from being updated due to port change
     *  - When set it allows static entries (agable bit clear) to update the
     *    source port if required
     * This bit should normally be '0' for most switch configurations */
    bool updStatic;

    /*! Unknown VLAN No Learn - This field when set will prevent source addresses
     *  of unknown VLAN IDs from being automatically added into the look up table
     *  iflearning is enabled */
    bool uVlanNoLearn;

    /*! Learn No VID */
    bool learnNoVlanID;
} CpswAle_LearningSecurityConfig;

/*!
 * \brief High level structure with ALE network security configuration
 */
typedef struct CpswAle_NetworkSecurityConfig_s
{
    /*! When set, any packet with a non-matching OUI source address will be
     *  dropped to the host unless the packet destination address matches a
     *  supervisory destination address table entry. When cleared, any packet
     *  source address matching an OUI address table entry will be dropped to
     *  the host unless the destination address matches with a supervisory
     *  destination address table entry */
    bool hostOuiNoMatchDeny;

    /*! Enable VLAN ID = 0 Mode, When cleared process the priority tagged packet
     *  with VID = PORT_VLAN[11:0]. When set process the priority tagged packet
     *  with VID = 0 */
    bool enableVid0Mode;

    /*! Malformed packet configuration */
    CpswAle_MalformedPktSecurityConfig malformedPktCfg;

    /*! IP packet security configuration */
    CpswAle_IPPktSecurityConfig ipPktCfg;

    /*! MAC authentication configuration */
    CpswAle_MacAuthConfig macAuthCfg;
} CpswAle_NetworkSecurityConfig;

/*!
 * \brief ALE init time port specific configuration params
 */
typedef struct CpswAle_PortConfig_s
{
    /*! Port specific learning configuration */
    CpswAle_PortLearningSecurityConfig learningCfg;

    /*! Port specific learningvlan security configuration */
    CpswAle_PortVlanSecurityConfig vlanCfg;

    /*! Port specific mac only mode configuration  */
    CpswAle_PortMacModeConfig macModeCfg;

    /*! Default Vlan ID config for each port */
    CpswAle_PortVlanConfig pvidCfg;
} CpswAle_PortConfig;

/*!
 * \brief ALE configuration
 *
 * This data structure contains configuration items related to the
 * ALE_CONTROL register fields.
 */
typedef struct CpswAle_Config_s
{
    /*! ALE Control fields.Bit mask based on CPSW_ALE_CONFIG_MASK */
    uint32_t modeFlags;

    /*! ALE Policer Configuration */
    CpswAle_PolicerGlobalConfig policerGlobalConfig;

    /*! Aging config */
    CpswAle_AgingConfig agingConfig;

    /*! VLAN config */
    CpswAle_InitVlanConfig vlanConfig;

    /*! Network security related configuration of ALE */
    CpswAle_NetworkSecurityConfig nwSecCfg;

    /*! ALE Port Configs */
    CpswAle_PortConfig portCfg[CPSW_ALE_NUM_PORTS];
} CpswAle_Config;

/*!
 * \brief ALE restore configuration
 *
 * This data structure contains configuration items for CpswAle_restart
 * Configuration parameters are those params which have to be reconfigured on
 * Main domain reset or core watchdog reboot.
 * Original ALE open configuration related to register configuration is
 * retained due to CPSW IP reset isolation feature
 */
typedef struct CpswAle_RestartConfig_s
{
    /*! Configuration of CPSW IP type such as CPSW2G/CPSW9G. This is needed
     *  for SoCs such as J721E which support multiple CPSW instances in a single
     *  SoC */
    Cpsw_Type cpswType;

    /*! CPSW supports reset isolation and full availability even when a core
     *  goes down. This requires special handling in driver when restarting from
     *  such scenarios. Based on type of restart different handling is required */
    Cpsw_RestartType restartType;

    /*! Flag indicating if CPSW is in VLAN aware mode */
    bool cpswVlanAwareMode;

    /*! Default Vlan ID config for each port */
    CpswAle_PortVlanConfig pvidCfg[CPSW_ALE_NUM_PORTS];

    /*! MAC address for each port which will be added as host port ALE static
     *  entry*/
    uint8_t macAddr[CPSW_ALE_NUM_MAC_PORTS][CPSW_ALE_MACADDR_NUM_OCTETS];
} CpswAle_RestartConfig;
/*  @} */

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/*!
 * \brief Opaque handle for ALE State structure
 */
typedef struct CpswAle_Obj_s *CpswAle_Handle;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Initialize ALE configuration parameters
 *
 * Initializes ALE configuratiobn parameters with default values
 *
 * \param pALEConfig    [IN] ALE configuration to be initialized
 */
void CpswAle_initParams(CpswAle_Config *pALEConfig);

/*!
 * \brief ALE open configuration
 *
 * Opens CPSW ALE.
 *
 * \param cpswType     [IN] CPSW instance type
 * \param pALEConfig   [IN] ALE config parameters
 *
 * \return CPSW ALE opaque handle
 */
CpswAle_Handle CpswAle_open(Cpsw_Type cpswType,
                            const CpswAle_Config *pALEConfig);

/*!
 * \brief ALE close configuration
 *
 * Closes ALE.
 *
 * \param hAle     [IN] CPSW ALE opaque handle
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t  CpswAle_close(CpswAle_Handle hAle);

/*!
 * \brief ALE IOCTL configuration
 *
 * Issues a control operation on the CPSW ALE submodule.
 *
 * \param hAle     [IN] CPSW ALE opaque handle
 * \param cmd      [IN] IOCTL command Id
 * \param prms     [IN] IOCTL parameters
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t  CpswAle_ioctl(CpswAle_Handle hAle,
                       CPSW_ALE_IOCTL_E cmd,
                       Cpsw_IoctlPrms *prms);

/*!
 * \brief Get ALE handle to already opened ALE instance
 *
 * Gets the ALE opaque handle to an already opened ALE instance.
 *
 * \param cpswType     [IN] CPSW instance type
 *
 * \return CPSW ALE opaque handle
 */
CpswAle_Handle CpswAle_getHandle(Cpsw_Type cpswType);

/*!
 * \brief Restore ALE submodule driver after domain reboot
 *
 * ALE API to restore ALE software state after Main domain or watch dog reboot.
 * The ALE register contents are retained due to CPSW IP reset isolation feature.
 * The host port is disengaged and all software context is lost
 *
 * \param pALERestartConfig   [IN] ALE restart configuration parameters
 *
 * \return CPSW ALE opaque handle
 */
CpswAle_Handle CpswAle_restart(CpswAle_RestartConfig *pALERestartConfig);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_ALE_H_ */

/*!
 *  @}
 */
