/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_cpts.h
 *
 * \brief This file contains the main structure definitions and function
 *        prototypes of the CPTS submodule.
 */

/*!
 * \addtogroup CPSW_CPTS_API
 * @{
 */

#ifndef CPSW_CPTS_H_
#define CPSW_CPTS_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"
#include <ti/drv/cpsw/include/cpsw_queue.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Helper macro to create IOCTL commands for CPTS submodule */
#define CPSW_CPTS_PUBLIC_IOCTL(x)             (CPSW_IOCTL_CPTS_BASE | CPSW_IOCTL_MIN(x))

/*! \brief Helper macro to normalize CpswCpts_HwPush enum */
#define CPSW_CPTS_HW_PUSH_NORMALIZE(x)        ((uint32_t)((x) - CPSW_CPTS_HW_PUSH_FIRST))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief CPTS module IOCTL commands
 */
enum CpswCpts_Ioctl_e
{
    /*!
     * \brief Get the hardware version of the CPTS submodule
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #Cpsw_GetVersionOutArgs
     */
    CPSW_CPTS_IOCTL_GET_VERSION                     = CPSW_CPTS_PUBLIC_IOCTL(0U),

    /*!
     * \brief CPTS register stack
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_RegisterStackInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_REGISTER_STACK                  = CPSW_CPTS_PUBLIC_IOCTL(1U),

    /*!
     * \brief CPTS unregister stack
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_UNREGISTER_STACK                = CPSW_CPTS_PUBLIC_IOCTL(2U),

    /*!
     * \brief CPTS register hardware push event callback
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_RegisterHwPushCbInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_REGISTER_HW_PUSH_CALLBACK       = CPSW_CPTS_PUBLIC_IOCTL(3U),

    /*!
     * \brief CPTS unregister stack
     *
     * IOCTL parameters:
     * -  inArgs: CpswCpts_HwPush hwPushNum
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_UNREGISTER_HW_PUSH_CALLBACK     = CPSW_CPTS_PUBLIC_IOCTL(4U),

    /*!
     * \brief CPTS time stamp push
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: uint64_t tsVal
     */
    CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT            = CPSW_CPTS_PUBLIC_IOCTL(5U),

    /*!
     * \brief CPTS time stamp load
     *
     * IOCTL parameters:
     * -  inArgs: uint64_t tsVal
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_LOAD_TIMESTAMP                  = CPSW_CPTS_PUBLIC_IOCTL(6U),

    /*!
     * \brief CPTS set time stamp nudge
     *
     * IOCTL parameters:
     * -  inArgs: int32_t nudge
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_TS_NUDGE                    = CPSW_CPTS_PUBLIC_IOCTL(7U),

    /*!
     * \brief CPTS set time stamp PPM
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_SetTsPpmInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_TS_PPM                      = CPSW_CPTS_PUBLIC_IOCTL(8U),

    /*!
     * \brief CPTS set compare value
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_SetCompValInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_COMP                        = CPSW_CPTS_PUBLIC_IOCTL(9U),

    /*!
     * \brief CPTS set compare nudge
     *
     * IOCTL parameters:
     * -  inArgs: int32_t nudge
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_COMP_NUDGE                  = CPSW_CPTS_PUBLIC_IOCTL(10U),

    /*!
     * \brief CPTS set function generator
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_SetFxnGenInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_GENF                        = CPSW_CPTS_PUBLIC_IOCTL(11U),

    /*!
     * \brief CPTS set GenFn stamp nudge
     *
     * IOCTL parameters:
     * -  inArgs: #CptsSetFxnGenNudgeInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_GENF_NUDGE                  = CPSW_CPTS_PUBLIC_IOCTL(12U),

    /*!
     * \brief CPTS set EST function generator
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_SetFxnGenInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_ESTF                        = CPSW_CPTS_PUBLIC_IOCTL(13U),

    /*!
     * \brief CPTS set ESTFn stamp nudge
     *
     * IOCTL parameters:
     * -  inArgs: #CptsSetFxnGenNudgeInArgs
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SET_ESTF_NUDGE                  = CPSW_CPTS_PUBLIC_IOCTL(14U),

    /*!
     * \brief CPTS TS output bit select
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_OutputBitSel
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_SELECT_TS_OUTPUT_BIT            = CPSW_CPTS_PUBLIC_IOCTL(15U),

    /*!
     * \brief CPTS lookup event
     *
     * IOCTL parameters:
     * -  inArgs: #CpswCpts_LookUpEventInArgs
     * - outArgs: #CpswCpts_Event
     */
    CPSW_CPTS_IOCTL_LOOKUP_EVENT                    = CPSW_CPTS_PUBLIC_IOCTL(16U),

    /*!
     * \brief CPTS clear all events
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_CLEAR_ALL_EVENTS                = CPSW_CPTS_PUBLIC_IOCTL(17U),

    /*!
     * \brief Print CPTS registers
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_PRINT_REGS                      = CPSW_CPTS_PUBLIC_IOCTL(18U),

    /*!
     * \brief Print CPTS event statistics
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_CPTS_IOCTL_PRINT_STATS                     = CPSW_CPTS_PUBLIC_IOCTL(19U),
};

/*!
 * \brief CPTS clock frequency add value enumeration
 */
typedef enum CpswCpts_RftClkFreq_e
{
    /*! CPTS reference clock frequency: 1000MHz */
    CPSW_CPTS_RFTCLK_FREQ_1000MHZ   = 0U,

    /*! CPTS reference clock frequency: 500MHz */
    CPSW_CPTS_RFTCLK_FREQ_500MHZ    = 1U,

    /*! CPTS reference clock frequency: 333.33MHz */
    CPSW_CPTS_RFTCLK_FREQ_333_33MHZ = 2U,

    /*! CPTS reference clock frequency: 250MHz */
    CPSW_CPTS_RFTCLK_FREQ_250MHZ    = 3U,

    /*! CPTS reference clock frequency: 200MHz */
    CPSW_CPTS_RFTCLK_FREQ_200MHZ    = 4U,

    /*! CPTS reference clock frequency: 166.66MHz */
    CPSW_CPTS_RFTCLK_FREQ_166_66MHZ = 5U,

    /*! CPTS reference clock frequency: 142.85MHz */
    CPSW_CPTS_RFTCLK_FREQ_142_85MHZ = 6U,

    /*! CPTS reference clock frequency: 125MHz */
    CPSW_CPTS_RFTCLK_FREQ_125MHZ    = 7U,
} CpswCpts_RftClkFreq;

/*!
 * \brief CPTS event type enumeration
 */
typedef enum CpswCpts_EventType_e
{
    /*! CPTS software time stamp push event */
    CPSW_CPTS_EVENTTYPE_TS_PUSH         = 0U,

    /*! CPTS time stamp rollover event (32bit mode only) */
    CPSW_CPTS_EVENTTYPE_TS_ROLLOVER     = 1U,

    /*! CPTS time stamp half rollover event (32bit mode only) */
    CPSW_CPTS_EVENTTYPE_TS_HALFROLLOVER = 2U,

    /*! CPTS hardware time stamp push event */
    CPSW_CPTS_EVENTTYPE_HW_TS_PUSH      = 3U,

    /*! CPTS Ethernet receive event */
    CPSW_CPTS_EVENTTYPE_ETH_RECEIVE     = 4U,

    /*! CPTS Ethernet transmit event */
    CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT    = 5U,

    /*! CPTS time stamp compare event */
    CPSW_CPTS_EVENTTYPE_TS_COMP         = 6U,

    /*! CPTS time stamp host transmit event */
    CPSW_CPTS_EVENTTYPE_TS_HOST_TX      = 7U,

    /*! Invalid event */
    CPSW_CPTS_EVENTTYPE_INVALID         = -1,
} CpswCpts_EventType;

/*!
 * \brief CPTS Ethernet message type enumeration
 */
typedef enum CpswCpts_MsgType_e
{
    /*! CPTS sync message */
    CPSW_CPTS_MESSAGE_SYNC                  = 0x00U,

    /*! CPTS delay request message */
    CPSW_CPTS_MESSAGE_DELAY_REQ             = 0x01U,

    /*! CPTS Pdelay request message */
    CPSW_CPTS_MESSAGE_PDELAY_REQ            = 0x02U,

    /*! CPTS Pdelay response message */
    CPSW_CPTS_MESSAGE_PDELAY_RESP           = 0x03U,

    /*! CPTS follow up message */
    CPSW_CPTS_MESSAGE_FOLLOW_UP             = 0x08U,

    /*! CPTS delay response message */
    CPSW_CPTS_MESSAGE_DELAY_RESP            = 0x09U,

    /*! CPTS Pdelay response followup message */
    CPSW_CPTS_MESSAGE_PDELAY_RESP_FOLLOW_UP = 0x0AU,

    /*! CPTS announce message */
    CPSW_CPTS_MESSAGE_ANNOUNCE              = 0x0BU,

    /*! CPTS signaling message */
    CPSW_CPTS_MESSAGE_SIGNALING             = 0x0CU,

    /*! CPTS management message */
    CPSW_CPTS_MESSAGE_MANAGEMENT            = 0x0DU,

    /*! CPTS invalid message */
    CPSW_CPTS_MESSAGE_INVALID               = -1,
} CpswCpts_MsgType;

/*!
 * \brief CPTS time stamp output bit enumeration
 */
typedef enum CpswCpts_OutputBitSel_e
{
    /*! CPTS time stamp output bit selection disabled */
    CPSW_CPTS_TS_OUTPUT_BIT_DISABLED = 0U,

    /*! CPTS select time stamp output bit 17 */
    CPSW_CPTS_TS_OUTPUT_BIT_17       = 1U,

    /*! CPTS select time stamp output bit 18 */
    CPSW_CPTS_TS_OUTPUT_BIT_18       = 2U,

    /*! CPTS select time stamp output bit 19 */
    CPSW_CPTS_TS_OUTPUT_BIT_19       = 3U,

    /*! CPTS select time stamp output bit 20 */
    CPSW_CPTS_TS_OUTPUT_BIT_20       = 4U,

    /*! CPTS select time stamp output bit 21 */
    CPSW_CPTS_TS_OUTPUT_BIT_21       = 5U,

    /*! CPTS select time stamp output bit 22 */
    CPSW_CPTS_TS_OUTPUT_BIT_22       = 6U,

    /*! CPTS select time stamp output bit 23 */
    CPSW_CPTS_TS_OUTPUT_BIT_23       = 7U,

    /*! CPTS select time stamp output bit 24 */
    CPSW_CPTS_TS_OUTPUT_BIT_24       = 8U,

    /*! CPTS select time stamp output bit 25 */
    CPSW_CPTS_TS_OUTPUT_BIT_25       = 9U,

    /*! CPTS select time stamp output bit 26 */
    CPSW_CPTS_TS_OUTPUT_BIT_26       = 10U,

    /*! CPTS select time stamp output bit 27 */
    CPSW_CPTS_TS_OUTPUT_BIT_27       = 11U,

    /*! CPTS select time stamp output bit 28 */
    CPSW_CPTS_TS_OUTPUT_BIT_28       = 12U,

    /*! CPTS select time stamp output bit 29 */
    CPSW_CPTS_TS_OUTPUT_BIT_29       = 13U,

    /*! CPTS select time stamp output bit 30 */
    CPSW_CPTS_TS_OUTPUT_BIT_30       = 14U,

    /*! CPTS select time stamp output bit 31 */
    CPSW_CPTS_TS_OUTPUT_BIT_31       = 15U,
} CpswCpts_OutputBitSel;

/*!
 * \brief CPTS PPM mode enumeration
 */
typedef enum CpswCpts_PpmMode_e
{
    /*! PPM value is set to zero */
    CPSW_CPTS_PPM_DISABLE           = 0U,

    /*! PPM value is calculated for parts per million */
    CPSW_CPTS_PPM_PARTS_PER_MILLION = 1U,

    /*! PPM value is calculated for parts per hour */
    CPSW_CPTS_PPM_PARTS_PER_HOUR    = 2U,
} CpswCpts_PpmMode;

/*!
 * \brief CPTS PPM direction Enumeration
 */
typedef enum CpswCpts_PpmDir_e
{
    /*! Increase time stamp with PPM value */
    CPSW_CPTS_PPM_DIR_INCREASE = 0U,

    /*! Decrease time stamp with PPM value */
    CPSW_CPTS_PPM_DIR_DECREASE = 1U,
} CpswCpts_PpmDir;

/*!
 * \brief CPTS Hardware push instances Enumeration
 */
typedef enum CpswCpts_HwPush_e
{
    /*! First hardware push instance */
    CPSW_CPTS_HW_PUSH_FIRST    = 1U,

    /*! Hardware push instance 1 */
    CPSW_CPTS_HW_PUSH_1        = CPSW_CPTS_HW_PUSH_FIRST,

    /*! Hardware push instance 2 */
    CPSW_CPTS_HW_PUSH_2        = 2U,

    /*! Hardware push instance 3 */
    CPSW_CPTS_HW_PUSH_3        = 3U,

    /*! Hardware push instance 4 */
    CPSW_CPTS_HW_PUSH_4        = 4U,

    /*! Hardware push instance 5 */
    CPSW_CPTS_HW_PUSH_5        = 5U,

    /*! Hardware push instance 6 */
    CPSW_CPTS_HW_PUSH_6        = 6U,

    /*! Hardware push instance 7 */
    CPSW_CPTS_HW_PUSH_7        = 7U,

    /*! Hardware push instance 8 */
    CPSW_CPTS_HW_PUSH_8        = 8U,

    /*! Last hardware push instance */
    CPSW_CPTS_HW_PUSH_LAST     = CPSW_CPTS_HW_PUSH_8,

    /*! Invalid hardware push instance */
    CPSW_CPTS_HW_PUSH_INVALID  = -1,
} CpswCpts_HwPush;

/*!
 * \brief Max number of hardware push instances
 */
#define CPSW_CPTS_HW_PUSH_COUNT_MAX           ((uint32_t)(CPSW_CPTS_HW_PUSH_LAST))

/*!
 * \brief CPTS Event
 */
typedef struct CpswCpts_Event_s
{
    /*! Time stamp value when the event occurred */
    uint64_t tsVal;

    /*! CPTS event type information */
    CpswCpts_EventType eventType;

    /*! CPTS event message type */
    CpswCpts_MsgType msgType;

    /*! Sequence id of the event */
    uint32_t seqId;

    /*! MAC port number corresponding to the CPTS event */
    uint32_t portNum;

    /*! Domain number of the event */
    uint32_t domain;

    /*! Hardware push instance number which triggered the event */
    CpswCpts_HwPush hwPushNum;
} CpswCpts_Event;

/*!
 * \brief CPTS Lookup event IOCTL inArgs
 */
typedef CpswCpts_Event CpswCpts_LookUpEventInArgs;

/*!
 * \brief CPTS event statistics
 */
typedef struct CpswCpts_EventStats_s
{
    /*! Count of software time stamp push events */
    uint64_t swTsPushEventCnt;

    /*! Count of hardware time stamp push events */
    uint64_t hwTsPushEventCnt;

    /*! Count of time stamp compare events */
    uint64_t cmpEventCnt;

    /*! Count of ethernet Rx events */
    uint64_t ethRxEventCnt;

    /*! Count of ethernet Tx events */
    uint64_t ethTxEventCnt;

    /*! Count of ethernet host Tx events */
    uint64_t ethHostTxEventCnt;

    /*! Lookup count of hardware time stamp push events */
    uint64_t hwTsPushLookupCnt;

    /*! Lookup count of time stamp compare events */
    uint64_t cmpEventLookupCnt;

    /*! Lookup count of ethernet Rx events */
    uint64_t ethRxEventLookupCnt;

    /*! Lookup count of ethernet Tx events */
    uint64_t ethTxEventLookupCnt;

    /*! Lookup count of ethernet host Tx events */
    uint64_t ethHostTxEventLookupCnt;

    /*! Discard count of old hardware time stamp push events */
    uint64_t hwTsPushDiscardCnt;

    /*! Discard count of time stamp compare events */
    uint64_t cmpEventDiscardCnt;

    /*! Discard count of ethernet Rx events */
    uint64_t ethRxEventDiscardCnt;

    /*! Discard count of ethernet Tx events */
    uint64_t ethTxEventDiscardCnt;

    /*! Discard count of ethernet host Tx events */
    uint64_t ethHostTxEventDiscardCnt;
} CpswCpts_EventStats;

/*!
 * \brief CPTS Event notify callback type
 *
 * \param hEventNotifyCbArg     [IN] Callback argument
 * \param eventInfo             [IN] CPTS event information
 *
 */
typedef void (*CpswCpts_EventNotifyCb)(void *hEventNotifyCbArg,
                                       CpswCpts_Event *eventInfo);

/*!
 * \brief CPTS hardware push notify callback type
 *
 * \param pHwPushNotifyCbArg     [IN] Callback argument
 * \param hwPushNum              [IN] CPTS hardware push instance number
 *
 */
typedef void (*CpswCpts_HwPushNotifyCb)(void *pHwPushNotifyCbArg,
                                        CpswCpts_HwPush hwPushNum);

/*!
 * \brief CPTS Register Stack InArgs
 *
 * Configuration information for registering a stack with the CPTS module
 */
typedef struct CpswCpts_StackConfig_s
{
    /*! Pointer to event notify callback function */
    CpswCpts_EventNotifyCb pEventNotifyCb;

    /*! Pointer to event notify callback function argument */
    void *hEventNotifyCbArg;
} CpswCpts_RegisterStackInArgs;

/*!
 * \brief CPTS Register hardware push callback InArgs
 */
typedef struct CpswCpts_RegisterHwPushCbInArgs_s
{
    /*! Hardware push instance number */
    CpswCpts_HwPush hwPushNum;

    /*! Pointer to hardware push notify callback function */
    CpswCpts_HwPushNotifyCb pHwPushNotifyCb;

    /*! Pointer to hardware push notify callback function argument */
    void *pHwPushNotifyCbArg;
} CpswCpts_RegisterHwPushCbInArgs;

/*!
 * \brief Set time stamp PPM IOCTL inArgs
 */
typedef struct CpswCpts_SetTsPpmInArgs_s
{
    /*! Parts-per-million correction value for time stamp */
    uint64_t tsPpmVal;

    /*! Direction of parts-per-million correction value for time stamp */
    CpswCpts_PpmDir tsPpmDir;

    /*! Mode of correction value for time stamp: parts-per-million or parts-per-hour.
     *  If the ppmMode is set to CPSW_CPTS_PPM_DISABLE, the ppm value will be
     *  set to zero. */
    CpswCpts_PpmMode tsPpmMode;
} CpswCpts_SetTsPpmInArgs;

/*!
 * \brief Setup Compare Value IOCTL inArgs
 */
typedef struct CpswCpts_SetCompValInArgs_s
{
    /*! 64-bit time stamp compare value */
    uint64_t tsCompVal;

    /*! Time stamp comparison length value */
    uint32_t tsCompLen;

    /*! Time stamp compare toggle mode selection.
     *     true : Toggle mode
     *     false: Non-toggle mode */
    bool tsCompToggle;
} CpswCpts_SetCompValInArgs;

/*!
 * \brief Set Function Generator(same data structure for both Generic and Ethernet) IOCTL inArgs
 */
typedef struct CpswCpts_SetFxnGenInArgs_s
{
    /*! Function generator instance index value */
    uint32_t index;

    /*! Function generator length value */
    uint32_t length;

    /*! 64-bit function generator compare value */
    uint64_t compare;

    /*! Function generator polarity inversion flag
     *      true : Output of function generator asserts high
     *      false: Output of function generator asserts low */
    bool polarityInv;

    /*! Function generator parts-per-million value */
    uint64_t ppmVal;

    /*! Direction of parts-per-million correction value */
    CpswCpts_PpmDir ppmDir;

    /*! Mode of correction value for GENF PPM: parts-per-million or parts-per-hour.
     *  If the ppmMode is set to CPSW_CPTS_PPM_DISABLE, the ppm value will be
     *  set to zero. */
    CpswCpts_PpmMode ppmMode;
} CpswCpts_SetFxnGenInArgs;

/*!
 * \brief Setup Function generator(same data structure for both Generic and Ethernet) Nudge IOCTL inArgs
 */
typedef struct CptsSetFxnGenNudgeInArgs_s
{
    /*! Function generator instance index value */
    uint32_t index;

    /*! Function generator nudge value */
    int32_t tsNudge;
} CptsSetFxnGenNudgeInArgs;

/*!
 * \brief CPTS configuration
 *
 * Configuration information for the CPTS submodule.
 */
typedef struct CpswCpts_Config_s
{
    /*! Host Receive time stamp enable: When set,
     *  time stamps are enabled on received packets to host */
    bool hostRxTsEnable;

    /*! TS_COMP polarity:
     *   true:  TS_COMP is asserted high
     *   false: TS_COMP is asserted low */
    bool tsCompPolarity;

    /*! Disable all time stamp Ethernet receive events:
     *   true:  Receive events are disabled
     *   false: Receive events are enabled */
    bool tsRxEventsDisable;

    /*! GENF (and ESTF) clear enable:
     *   true:  A TS_GENFn (or TS_ESTFn) output is cleared when the associated
     *          ts_genf_length[31:0] (or ts_estf_length[31:0]) is cleared to zero
     *   false: A TS_GENFn (or TS_ESTFn) output is not cleared when the associated
     *          ts_genf_length[31:0] (or ts_estf_length[31:0]) is cleared to zero */
    bool tsGenfClrEnable;

    /*! CPTS RFT Clock frequency required to set TS_ADD VAL */
    CpswCpts_RftClkFreq cptsRftClkFreq;
} CpswCpts_Config;

/*!
 * \brief CPTS handle
 *
 * CPTS submodule opaque handle.
 */
typedef struct CpswCpts_Obj_s *CpswCpts_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_CPTS_H_ */

/* @} */
