/* =============================================================================
 *   Copyright (c) Texas Instruments Incorporated 2013-2015
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \ingroup DRV_CPSW_MODULE
 *  \defgroup CPSW_DMA_API CPSW Data Path Header
 *
 *  CPSW DMA submodule provides a API layer which allows applications to set up the
 *  UDMA peripheral to control the flow of packet data to/from the core from/to CPSW.
 *  CPSW DMA module is used to for
 *  1. Configuration of receive UDMA channel & flow for enabling reception of Ethernet
 *     packets matching the application criteria Ex - OUI, Ether Type, IPv4 SA/DA.
 *  2. Configuration of transmit UDMA channel for enabling transmit of Ethernet packets
 *
 *  The application using CPSW DMA submodule need to allocate memories for different
 *  DMA data structures like ring memories, host packet descriptors and packet buffer
 *  memories.
 *
 *  @{
 */

/*!
 *  \file cpsw_dma.h
 *
 *  \brief This file contains the structure definitions and function
 *            prototypes of the CPSW Data/DMA path set up.
 */

#ifndef CPSW_DMA_H_
#define CPSW_DMA_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_queue.h>
#include <ti/drv/cpsw/include/cpsw_dma_psi.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*!
 *  \name CPSW DMA Instance configuration
 *
 *  Configuration macros for CPSW DMA module
 *  @{
 */

/*! Maximum number of CPSW TX DMA channels */
#define CPSW_DMA_MAX_TX_CH                                  (8U)

/*! Maximum number of CPSW RX FLOWS channels */
#define CPSW_DMA_MAX_RX_FLOW                                (64U)

/*! Max. number for Rx free queues for CPSW RX DMA flow */
#define CPSW_DMA_MAX_RX_FREE_QUEUE                          (4U)

/*! CPSW DMA HPD packet size */
#define CPSW_DMA_HPD_SIZE                                   (128U)

/*! Extended Packet Info Block size */
#define CPSW_DMA_EXTENDED_PKT_INFO_BLOCK_SIZE               (16U)

/*! Extended Packet Info Block size */
#define CPSW_DMA_PROTOCOL_SPECIFIC_INFO_BLOCK_SIZE          (16U)

/*! Extended Packet Info Block size */
#define CPSW_DMA_PKT_DESC_RESERVED_SIZE                     (CPSW_DMA_HPD_SIZE - \
                                                             sizeof(CSL_UdmapCppi5HMPD) - \
                                                             (CPSW_DMA_EXTENDED_PKT_INFO_BLOCK_SIZE + \
                                                              CPSW_DMA_PROTOCOL_SPECIFIC_INFO_BLOCK_SIZE))

/*! CPSW DMA Desc size - this is used by applications to allocate DMA desc memory */
#define CPSW_DMA_RING_MEM_SIZE                              (sizeof(uint64_t))

/*! CPSW DMA Invalid RX flow id */
#define CPSW_DMA_RXFLOWIDX_INVALID                          (0xABCDABCDU)

/*!
 * \brief CPSW DMA RX MTU alignment. The RxFlow Mtu must be aligned to this value
 *
 * As per UDMAP spec
 * "Rx Packet Size Threshold 0: This value is left shifted by 5 bits and compared
 * against the packet size to determine which free descriptor queue should be used
 * for the SOP buffer in the packet"
 * The value programmed is hence supposed to be 32 bit aligned. Programming
 * non aligned values will result in truncation and expected MTU size will not be
 * programmed.
 * For example setting MTU size of 1518 (required MTU size) results in actual value
 * of 1504 getting programmed resulting in packet drop for frames of size
 * 1504 - 1518 which the app expects to less than the programmed MTU based on the
 * param passed. Refer ksdma_udmap spec section 4.4.2.6 for more details
*/
#define CPSW_DMA_RXMTU_ALIGN                                (1U << 5U)

/* @} */

/*!
 *  \name CPSW DMA Statistics configuration
 *
 *  No. of latest entries kept in the statistics
 *  @{
 */
#define CPSW_DMA_STATS_HISTORY_CNT                          ((uint32_t)32U)
/* @} */

/*!
 *  \name CPSW DMA driver opaque handles
 *
 *  Opaque handle typedefs for CPSW DMA driver objects.
 *  @{
 */

/*!
 * \brief Opaque structure to hold software state for CPSW DMA RX flow.
 */
typedef struct CpswDma_RxFlowObj_s *CpswDma_RxFlowHandle;

/*!
 * \brief Opaque structure to hold software state for CPSW TX DMA channel
 */
typedef struct CpswDma_TxChObj_s *CpswDma_TxChHandle;

/*!
 * \brief Opaque structure to hold software state for CPSW Packet Info
 */
typedef struct CpswDma_PktInfo_s *CpswDma_PktHandle;

/*!
 * \brief Opaque structure to hold DMA descriptor Q handle
 */
typedef struct CpswDma_DmaDescQ_s *CpswDma_DmaDescQHandle;


/* @} */

/*!
 *  \name CPSW DMA driver Callback function types
 *
 *   Callback function typedefs so that the CpswDma layer can call into the app layer
 *   and let it translate between the hardware packet descriptors and packets and the
 *   stack/translation layer's buffers and packets.
 *  @{
 */

/*! Function pointer type for DMAC packet allocation function  */
typedef uint8_t *
(*CpswDma_AllocRingMemFxn)(void *appPriv, uint32_t numRingEle, uint32_t alignSize);

/*! Function pointer type for DMAC packet free function  */
typedef void
(*CpswDma_FreeRingMemFxn)(void *appPriv, void *pRingMem, uint32_t numRingEle);

/*! Function pointer type for DMAC packet allocation function  */
typedef struct CpswDma_DmaDesc_s * (*CpswDma_AllocDmaDesccFxn)(void *appPriv, uint32_t alignSize);

/*! Function pointer type for DMAC packet free function  */
typedef void (*CpswDma_FreeDmaDescFxn)(void *appPriv, struct CpswDma_DmaDesc_s *dmaDescPtr);

/*!
 * Function pointer type for packet notify call back. This is called by driver
 * when packet is received on the Rx flow or transmission completed from
 * Tx channel.
 */
typedef void (*CpswDma_PktNotifyCb)(void *hCbArg);

/* @} */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 *  \defgroup CPSW_DMA_PKT_STATS Packet stats Definitions
 *
 *  This group contains details about CPSW DMA RX Flow and TX channel statistics
 *
 *  @{
 */

/*!
 * \brief Stats for packets submitted/retrieved in the DMA event callbacks functions
 */
typedef struct CpswDma_CbFxnStats_s
{
    /*! Total no. of times the packets were submitted or retrieved from DMA */
    uint64_t dataNotifyCnt;

    /*! Counts when DMA retrievePkts is called but no packets was retrived. This count
     *  can help in determining optimal pacing interval */
    uint64_t zeroNotifyCnt;
    /*! Total no. of packets submitted or retrieved from DMA */
    uint64_t totalPktCnt;
    /*! Total cycles consumed to submit or retrieve totalPktCnt packets from DMA */
    uint64_t totalCycleCnt;

    /*! Maximum packets retrieved in single Submit or retrieve call. This count
     *  can help in determining optimal no. of DMA descriptors and packet counts to be
     *  allocated */
    uint64_t pktsPerNotifyMax;

    /*! Packets retrieved per call to single Submit or retrieve call. Maintained for
     *   last CPSW_DMA_STATS_HISTORY_CNT packets */
    uint64_t pktsPerNotify[CPSW_DMA_STATS_HISTORY_CNT];
    /*! Maximum of cycles consumed of all Submit or retrieve call. */
    uint64_t cycleCntPerNotifyMax;

    /*! Cycles consumed per call to single Submit or retrieve call. Maintained for
     *   last CPSW_DMA_STATS_HISTORY_CNT packets */
    uint64_t cycleCntPerNotify[CPSW_DMA_STATS_HISTORY_CNT];
    /*! Maximum cycles for single packet in Submit or retrieve call */
    uint64_t cycleCntPerPktMax;

    /*! Cycles per packet in Submit or retrieve call. Maintained for
     *   last CPSW_DMA_STATS_HISTORY_CNT packets */
    uint64_t cycleCntPerPkt[CPSW_DMA_STATS_HISTORY_CNT];

    /*! No. of free DMA descriptors in the RX Flows or TX channels DMA descriptor Q. Maintained for
     *   last CPSW_DMA_STATS_HISTORY_CNT packets. This can be useful to detect and handle drops in
     *   packet bursts*/
    uint64_t readyDmaDescQCnt[CPSW_DMA_STATS_HISTORY_CNT];
} CpswDma_CbFxnStats;

/*!
 * \brief DMA descriptor stats for the flow and channel
 */
typedef struct CpswDma_DmaDescStats_s
{
    /*! Total no. of free/ready desc enq for RX flow or TX channel. */
    uint64_t readyDmaDescEnq;
    /*! Total no. of free/full desc deq for RX flow or TX channel. */
    uint64_t freeDmaDescDq;
    /*! Total no. of DMA desc Q underflows for RX flow or TX channel */
    uint64_t underFlowCnt;
}
CpswDma_DmaDescStats;

/*!
 * \brief Rx flow statistics
 */
typedef struct CpswDma_RxFlowStats_s
{
    /*! Stats for RX submit packet calls  */
    CpswDma_CbFxnStats submitPktStats;
    /*! Stats for RX retrieve packet calls  */
    CpswDma_CbFxnStats retrievePktStats;
    /*! RX Flows DMD desc Q stats  */
    CpswDma_DmaDescStats dmaDescStats;
    /*! Total no. of free packets enq for RX flow */
    uint64_t rxSubmitPktEnq;
    /*! Total no. of full packets deq for RX flow */
    uint64_t rxRetrievePktDq;
    /*! Total no. of packets freeQ underflows for RX flow */
    uint64_t rxSubmitPktUnderFlowCnt;
}CpswDma_RxFlowStats;

/*!
 * \brief TX channel statistics
 */
typedef struct CpswDma_TxChStats_s
{
    /*! Stats for TX submit packet calls  */
    CpswDma_CbFxnStats submitPktStats;
    /*! Stats for TX retrieve packet calls  */
    CpswDma_CbFxnStats retrievePktStats;
    /*! TX Channels DMD desc Q stats  */
    CpswDma_DmaDescStats dmaDescStats;
    /*! Total no. of free packets enq for TX Channel */
    uint64_t txSubmitPktEnq;
    /*! Total no. of full packets deq for TX Channel */
    uint64_t txRetrievePktDq;
    /*! Total no. of packets freeQ underflows for TX Channel */
    uint64_t txSubmitPktOverFlowCnt;
} CpswDma_TxChStats;

/* @} */

/*!
 *  \defgroup CPSW_DMA_PKT Packet Definitions
 *
 *  This group contains details about CPSW DMA packet Q structures. These Packet
 *  Q's are used to exchange data between application and driver module.
 *
 *  @{
 */

/*!
 * \brief CPPI Buffer timestamp info
 *
 * Buffer timestamp info structure.
 */
typedef struct CpswDma_PktTsInfo_s
{
    /*! Flag to enable/disable host Tx packet timestampping */
    bool enableHostTxTs;

    /*! Host Tx packet's sequence Id value */
    uint32_t txPktSeqId;

    /*! Host Tx packet's message type value */
    uint8_t txPktMsgType;

    /*! Host Tx packet's domain value */
    uint8_t txPktDomain;

    /*! Received packet's 64-bit ingress time stamp value */
    uint64_t rxPktTs;
} CpswDma_PktTsInfo;

/*!
 * \brief CPPI Buffer info
 *
 * Application buffer info structure.
 */
typedef struct CpswDma_PktInfo_s
{
    /*! Pointer to next Buffer
     * Note: Keep node_t as first member always as driver uses generic queue
     * functions and deferences to this member */
    node_t node;

    /*! Pointer to data buffer */
    uint8_t *bufPtr;
#ifdef QNX_OS
    /*! Pointer to Data Buffer Physical */
    uint8_t *bufPtrPhys;
#endif
    /*! Original length of the data buffer passed */
    uint32_t orgBufLen;

    /*! Actual filled buffer length while receiving/transmitting data with DMA */
    uint32_t userBufLen;

    /*! Pointer to application layer specific data blob */
    void *appPriv;

    /*! Packet State Info. Refer to #CpswUtils_PktStateModuleType,
     *  #CpswUtils_PktStateMemMgr, #CpswUtils_PktStateDma and
     *  #CpswUtils_PktStateApp.
     *
     *  This is only for runtime check and debug, not to be used by application */
    uint32_t pktState;

    /*! Protocol information word. Currently used for checksum offload related
     *  information.
     *
     *  For Rx it contains HW computed checksum. App should call appropriate
     *  CPSW CPPI helper macros to extract required fields.
     *
     *  For Tx it contains information passed to HW for checksum computation.
     *  App should call appropriate cpsw cppi helper macros to insert required
     *  fields */
    uint32_t chkSumInfo;

    /*! Packet time stamp information
     *
     * For Tx, if tsInfo.enableHostTxTs flag is set to true, packet will be timestampped on egress
     * and will trigger host transmit event in CPTS. The timestamp value is then stored in
     * CPTS FIFO with given sequence Id, message type and domain value. This can be used to
     * timestamp any packet (even non-PTP) sent from host.
     *
     * For Rx, the received packet's ingress timestamp is captured and is stored in tsInfo.rxPktTs.
     * All ingress packets are timestampped.
     */
    CpswDma_PktTsInfo tsInfo;

    /*! Directed port number
     *  The port number to which the packet is to be directed bypassing ALE lookup.
     *  This value is written to Dst Tag – Low bits of packet descriptor.
     *  If ALE lookup based switching is to be performed, this value should
     *  be set to CPSW_MAC_PORT_INV.
     *  Note: Directed packets go to the directed port, but an ALE lookup is performed to
     *  determine untagged egress in VLAN_AWARE mode. */
    Cpsw_MacPort txPortNum;

    /*! Packet's received port number
     *  For Rx, this variable contains the port number from which the packet was received.
     *  This value is obtained from the Source Tag – Low bits of packet descriptor. */
    Cpsw_MacPort rxPortNum;
}CpswDma_PktInfo;

/*! CPSW DMA Desc size - this is used by applications to allocate DMA desc memory */

/*!
 * \brief CPSW DMA UDMA descriptor format
 *
 * CPSW packet which contains information about the underlying UDMA packet
 * descriptor and can be used to form a queue of packets using CpswPktQ_*()
 * functions
 */
typedef struct CpswDma_cpswUdmaDesc_s
{
    /*! Host-mode packet descriptor (must be the first member of the structure */
    CSL_UdmapCppi5HMPD hostDesc  __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /*! Extended Packet Info Block
     *
     * Note:
     * - For Rx Flow extendedPktInfo will be present only if flowPrms.einfoPresent
     *   is set to true (TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_EINFO_PRESENT)
     * - For Tx channel extendedPktInfo will be present only if txPrms.filterEinfo
     *   is set to false(TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_EINFO_DISABLED)
     */
    uint8_t extendedPktInfo[CPSW_DMA_EXTENDED_PKT_INFO_BLOCK_SIZE];

    /*! Protocol specific info
     * Note:
     * - For Rx Flow psInfo will be present only if flowPrms.psInfoPresent
     *   is set to true (TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PSINFO_PRESENT)
     * - For Tx channel psInfo will be present only if txPrms.filterPsWords
     *   is set to false(TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_PSWORDS_DISABLED)
     */
    uint8_t psInfo[CPSW_DMA_PROTOCOL_SPECIFIC_INFO_BLOCK_SIZE];

    /*! Reserved area. Desc size is 128 bytes (HMPD + reserved = 128) */
    uint8_t reserved[CPSW_DMA_PKT_DESC_RESERVED_SIZE];
} CpswDma_CpswHpdDesc;

/*!
 * \brief CPPI DMA descriptor
 *
 * The format of a single DMA descriptor to be transmitted via the CpswDma.
 */
typedef struct CpswDma_DmaDesc_s
{
    /*! CPPI HPD descriptor.
     * Note: This MUST be first member as it is typecasted to
     * #CpswDma_CpswHpdDesc */
    CpswDma_CpswHpdDesc hpdDesc;

    /*! Pointer to next DMA Desc in the queue */
    struct CpswDma_DmaDesc_s *pNextDesc;

    /*! Applicatioon packet details - this is used by application to store
     *  original packet info. This is to avoid asking buffer address from
     *  driver descriptor */
    CpswDma_PktInfo *pAppPktInfo;
} CpswDma_DmaDesc;

/*! \brief Function pointer type for Ethernet packet allocation function  */
typedef CpswDma_PktInfo *(*CpswDma_AllocEthPktFxn)(void *appPriv,
                                                   uint32_t pktSize,
                                                   uint32_t alignSize);

/*! \brief Function pointer type for Ethernet packet free function  */
typedef void (*CpswDma_FreeEthPktFxn)(CpswDma_PktInfo *pPktInfo);

/*!
 * \brief Packet Queue
 *
 * A queue of packets, which are use for managing the active queue of packets
 * given to the hardware driver.  It can also be used by the translation or
 * application layer to manage packets that are not currently active.
 */
typedef queue_t CpswDma_PktInfoQ;

/* @} */

/*!
 * \defgroup CPSW_DMA_CONFIG_DEFINTIONS CPSW DMA Configuration Definitions
 *
 * This group contains structure and type definitions needed to properly
 * construct the CPSW DMA configuration structure which is part of the top-level
 * CPSW configuration structure.
 *
 *  @{
 */

/*!
 * \brief UDMA ring monitor config parameters
 *
 * UDMA Ring Monitor configuration paramters. This structure is stripped-down
 * version of #Udma_RingMonPrms.
 */
typedef struct
{
    /*! Ring monitor mode. Refer \ref tisci_msg_rm_ring_mon_cfg_req::mode */
    uint8_t mode;

    /*! When mode is TISCI_MSG_VALUE_RM_MON_MODE_PUSH_POP, this is read-only
     *  and represents number of pushes.
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_THRESHOLD, this is read-write
     *  and represents the low threshold value which should be programmed
     *  to generate the RM event when the threshold is crossed (goes below).
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_WATERMARK, this is read only
     *  and represents the low watermark.
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_STARVATION, this is read only
     *  and represents the starvation count */
    uint32_t data0;

    /*! When mode is TISCI_MSG_VALUE_RM_MON_MODE_PUSH_POP, this is read-only
     *  and represents number of pops.
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_THRESHOLD, this is read-write
     *  and represents the high threshold value which should be programmed
     *  to generate the RM event when the threshold is crossed (goes above).
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_WATERMARK, this is read only
     *  and represents the high watermark.
     *  When mode is TISCI_MSG_VALUE_RM_MON_MODE_STARVATION, this is not
     *  applicable */
    uint32_t data1;
} CpswDma_RingMonCfg;

/*!
 * \brief CPSW DMA channel ring parameters
 *
 * Ring configuration structure for RX flow and TX channel FQ & CQ rings.
 */
typedef struct
{
    /*! Ring bus order ID value to be programmed into the orderid field of
     *  the ring's RING_ORDERID register. */
    uint8_t orderId;

    /*! Ring mode. Refer \ref tisci_msg_rm_ring_cfg_req::mode */
    uint8_t mode;

    /*! Flag to indicate if ring monitor should be allocated for this ring.
     *  This should be enabled only for those flows that require very high
     *  throughput as there are limited ring monitors available in the
     *  system */
    bool useRingMon;

    /*! Ring monitor config (used only if useRingMon is true) */
    CpswDma_RingMonCfg ringMonCfg;
} CpswDma_UdmaRingPrms;

/*!
 *  \brief CPSW DMA Rx flow/TX channel ring configuration parameters
 *
 * The structure contains FQ and CQ ring configuration for RX flow/TX channel.
 * Teardown ring is allocated in the driver so we don't take it here.
 */
typedef struct
{
    /*! [IN] Free queue ring params where descriptors are queued */
    CpswDma_UdmaRingPrms fqRingPrms;

    /*! [IN] Completion queue ring params where descriptors are dequeued */
    CpswDma_UdmaRingPrms cqRingPrms;
} CpswDma_UdmaChPrms;

/*!
 * \brief UDMA RX channel flow parameters
 *
 * The structure is stripped down version of #Udma_FlowPrms.
 */
typedef struct
{
    /*! [IN] Set to 1 if extended packet info is present in the descriptor */
    uint8_t einfoPresent;

    /*! [IN] Set to 1 if protocol-specific info is present in the
    *    descriptor */
    uint8_t psInfoPresent;

    /*! [IN] Start of rx packet data (byte offset from the start of the SOP
     *  buffer) */
    uint16_t sopOffset;

    /*! [IN] Rx destination queue */
    uint16_t defaultRxCQ;

    /*! [IN] UDMAP receive flow source tag high byte constant configuration
     *  to be programmed into the rx_src_tag_hi field of the flow's RFLOW_RFB
     *  register */
    uint8_t srcTagHi;

    /*! [IN] UDMAP receive flow source tag low byte constant configuration
     *  to be programmed into the rx_src_tag_lo field of the flow's RFLOW_RFB
     *  register */
    uint8_t srcTagLo;

    /*! [IN] UDMAP receive flow source tag high byte selector configuration
     *  to be programmed into the rx_src_tag_hi_sel field of the RFLOW_RFC
     *  register. Refer #tisci_msg_rm_udmap_flow_cfg_req::rx_src_tag_hi_sel */
    uint8_t srcTagHiSel;

    /*! [IN] UDMAP receive flow source tag low byte selector configuration
     *  to be programmed into the rx_src_tag_low_sel field of the RFLOW_RFC
     *  register. Refer #tisci_msg_rm_udmap_flow_cfg_req::rx_src_tag_lo_sel */
    uint8_t srcTagLoSel;

    /*! [IN] UDMAP receive flow destination tag high byte constant configuration
     *  to be programmed into the rx_dest_tag_hi field of the flow's RFLOW_RFB
     *  register */
    uint8_t destTagHi;

    /*! [IN] UDMAP receive flow destination tag low byte constant configuration
     *  to be programmed into the rx_dest_tag_lo field of the flow's RFLOW_RFB
     *  register */
    uint8_t destTagLo;

    /*! [IN] UDMAP receive flow destination tag high byte selector configuration
     *  to be programmed into the rx_dest_tag_hi_sel field of the RFLOW_RFC
     *  register. Refer #tisci_msg_rm_udmap_flow_cfg_req::rx_dest_tag_hi_sel */
    uint8_t destTagHiSel;

    /*! [IN] UDMAP receive flow destination tag low byte selector configuration
     *  to be programmed into the rx_dest_tag_low_sel field of the RFLOW_RFC
     *  register. Refer #tisci_msg_rm_udmap_flow_cfg_req::rx_dest_tag_lo_sel */
    uint8_t destTagLoSel;
} CpswDma_UdmaFlowPrms;

/*!
 * \brief CPSW DMA TX channel parameters
 *
 * The structure is stripped down version of #Udma_ChTxPrms
 */
typedef struct
{
    /*! [IN] Bool: When set (TRUE), filter out extended info */
    uint8_t filterEinfo;

    /*! [IN] Bool: When set (TRUE), filter out protocl specific words */
    uint8_t filterPsWords;

    /*! [IN] Address type for this channel.
     *   Refer #tisci_msg_rm_udmap_tx_ch_cfg_req::tx_atype */
    uint8_t addrType;

    /*! [IN] Channel type. Refer #tisci_msg_rm_udmap_tx_ch_cfg_req::tx_chan_type */
    uint8_t chanType;

    /*! [IN] 3-bit priority value (0=highest, 7=lowest) */
    uint8_t busPriority;

    /*! [IN] 3-bit qos value (0=highest, 7=lowest) */
    uint8_t busQos;

    /*! [IN] 4-bit orderid value */
    uint8_t busOrderId;

    /*! [IN] This field selects which scheduling bin the channel will be placed
     *  in for bandwidth allocation of the Tx DMA units.
     *  Refer #tisci_msg_rm_udmap_tx_ch_cfg_req::tx_sched_priority */
    uint8_t dmaPriority;

    /*! [IN] TX credit for external channels */
    uint8_t txCredit;

    /*! [IN] The fifo depth is used to specify how many FIFO data phases deep
     *  the Tx per channel FIFO will be for the channel.
     *  While the maximum depth of the Tx FIFO is set at design time, the FIFO
     *  depth can be artificially reduced in order to control the maximum latency
     *  which can be introduced due to buffering effects */
    uint16_t fifoDepth;
} CpswDma_UdmaChTxPrms;

/*!
 * \brief CPSW Channel/Flow auto-reclaim config struct
 */
typedef struct
{
    /*! auto-reclaim enable flag
     *  Auto-reclaim or auto-recycling of buffer descriptors enables usage of same
     *  buffer descriptor between RX flow and TX channel.
     *  For example - by using auto-reclaim of RX FQ as TX CQ, we don't need to call
     *    CpswDma_retrieveTxDonePackets and CpswDma_submitRxReadyPackets
     *    in packet forwarding use-cases like SW interVLAN routing
     *  Note that current design offers/provides way to use TX FQ as RX CQ but not vice
     *  versa (RX CQ as TX FQ), similarly for RX FQ as TX CQ case. This is done due to
     *  UDMA HW design where FQ ring number is same as channel number. Basically any ring
     *  can't be attached as FQ ring to TX channel. Though this limitation doesn't exist
     *  for RX flow, we keep design consistent and enable auto-reclaim of RX FQ as TX CQ
     */
    bool enableFlag;

    /*! Handle to free DMA descriptor pool where a free descriptors retrieved in
     *  CpswDma_retrieveRxPackets() function. This is as we bypass
     *  CpswDma_retrieveTxDonePackets() in auto-reclaim */
    CpswDma_DmaDescQHandle hDmaDescPool;

    /*! handle to ring used in auto-reclaim. For example when using
     *  TX CQ --> RX FQ, the RX FQ ring handle is passed by an app to be used as
     *  TX CQ so default ring should not be allocated/used */
    Udma_RingHandle hReclaimRing;
} CpswDma_AutoReclaimPrms;

/*!
 * \brief Param struct for the get default flowId open parameters
 *
 * The Param struct for the Rx Flow open function, containing CPSW instance
 * Id for RX flow. The returned channel handle is an opaque variable used to
 * access the RX Flow in other function commands.
 */
typedef struct CpswDma_OpenRxFlowPrms_s
{
    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;

    /*! Flow start index*/
    uint32_t startIdx;

    /*! Flow index to be opened*/
    uint32_t flowIdx;

    /*!  UDMA specific channel params */
    CpswDma_UdmaChPrms udmaChPrms;

    /*! CPSW DMA event callback function - this function will be called when
     *   the registered packets are received on RX Flow*/
    CpswDma_PktNotifyCb notifyCb;

    /*!  UDMA Flow params */
    CpswDma_UdmaFlowPrms flowPrms;

    /*! Number of receive packets, used for allocating number of DMA descriptors
     *  Note - The HW ring element count field is 19-bit */
    uint32_t numRxPkts;

    /*! Flag to disable cache operations on the ring memory */
    bool disableCacheOpsFlag;

    /*! Maximum receive packet length for this flow. Make sure packets of at
     *  least this length are submitted in CpswDma_submitRxPackets() */
    uint32_t rxFlowMtu;

    /*! Callback functions.
     *  Callback function typedefs so that the CpswDma layer can call into the
     *  app layer and let it translate between the hardware buffer descriptors
     *  and packets and the stack/translation layer's buffers and packets  */

    /*! Ring memory allocation callback, this cannot be NULL */
    CpswDma_AllocRingMemFxn ringMemAllocFxn;

    /*! Ring memory free function callback, used in close flow */
    CpswDma_FreeRingMemFxn ringMemFreeFxn;

    /*! DMA HPD (host packet descriptor) memory allocation callback */
    CpswDma_AllocDmaDesccFxn dmaDescAllocFxn;

    /*! Transmit HPD (host packet descriptor) memory free callback */
    CpswDma_FreeDmaDescFxn dmaDescFreeFxn;

    /*! Argument to be used for the callback routines (it should mean something
     *  to layer into which the callback calls) */
    void *hCbArg;

    /*! Flag to indicate if a dedicated proxy should be allocated for the "FQ"
     *  for this flow.
     *  This should be enabled only for those flows that require very high
     *  throughput as there are limited proxy available in the system */
    bool useProxy;

    /*! Buffer Auto-reclaim params*/
    CpswDma_AutoReclaimPrms autoReclaimPrms;
}
CpswDma_OpenRxFlowPrms;

/*!
 * \brief Param struct for the Tx channel open function
 *
 * The parameter structure for the Tx channel open function, containing a
 * channel config structure.
 */
typedef struct CpswDma_OpenTxChPrms_s
{
    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;

    /*! UDMA channel to number allocated for transmit.
     *  Set to #UDMA_DMA_CH_ANY if the channel to allocate and open could be
     *  any from the free pool */
    uint32_t chNum;

    /*! UDMA channel params */
    CpswDma_UdmaChPrms udmaChPrms;

    /*! UDMA transmit channel params */
    CpswDma_UdmaChTxPrms udmaTxChPrms;

    /*! CPSW DMA event callback function - this function will be called when
     *  the registered packets are transmitted on TX channel */
    CpswDma_PktNotifyCb notifyCb;

    /*! Maximum number of transmit packets, used for allocating number of DMA descriptors
     *  Note - The HW ring element count field is 19-bit */
    uint32_t numTxPkts;

    /*! Flag to disable cache operations on the ring memory */
    bool disableCacheOpsFlag;

    /*! Callback functions.
     *  Callback function typedefs so that the CpswDma layer can call into the
     *  app layer and let it translate between the hardware buffer descriptors
     *  and packets and the stack/translation layer's buffers and packets */

    /*! Ring memory allocation callback, this cannot be NULL */
    CpswDma_AllocRingMemFxn ringMemAllocFxn;

    /*! Ring memory free function callback, used in close flow */
    CpswDma_FreeRingMemFxn ringMemFreeFxn;

    /*! DMA HPD (host packet descriptor) memory allocation callback */
    CpswDma_AllocDmaDesccFxn dmaDescAllocFxn;

    /*! Transmit HPD (host packet descriptor) memory free callback */
    CpswDma_FreeDmaDescFxn dmaDescFreeFxn;

    /*! Argument to be used for the callback routines (it should mean something
     *  to layer into which the callback calls) */
    void *hCbArg;

    /*! Flag to indicate if a dedicated proxy should be allocated for the "FQ"
     *  for this flow.
     *  This should be enabled only for those flows that require very high
     *  throughput as there are limited proxy available in the system */
    bool useProxy;

    /*! Buffer Auto-reclaim params*/
    CpswDma_AutoReclaimPrms autoReclaimPrms;

#ifdef QNX_OS
    /*! UDMA channel to number allocated for transmit.
     *  Set to #UDMA_DMA_CH_ANY if the channel to allocate and open could be
     *  any from the free pool */
    uint32_t preferredChNum;
#endif
}
CpswDma_OpenTxChPrms;

/* @} */

/*!
 *  \addtogroup CPSW_CONFIG_DEFINTIONS
 *
 *  @{
 *
 */

/*!
 * \brief Param struct for the Rx channel open
 *
 * The parameter structure for the Rx channel open, containing a channel config
 * structure.
 */
typedef struct CpswDma_RxChInitPrms_s
{
    /*! This field selects which scheduling bin the channel will be placed in
     *  for bandwidth allocation of the Tx DMA units.
     *  Refer \ref tisci_msg_rm_udmap_tx_ch_cfg_req::tx_sched_priority */
    uint8_t dmaPriority;

#ifdef QNX_OS
    /*! This field provided the preferred dma channel to use */
    uint32_t preferredChNum;
#endif
}
CpswDma_RxChInitPrms;

/*!
 * \brief Config structure for CPSW DMA
 *
 * The parameter structure for CPSW DMA configuration, containing a Rx channel
 * config and NAVSS instance id.
 */
typedef struct CpswDma_Config_s
{
    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;

    /*! RX channel configuration parameters */
    CpswDma_RxChInitPrms rxChInitPrms;
}
CpswDma_Config;

/*!
 * \brief Config structure for CPSW DMA Data Path initialization
 *
 * The parameter strcture for CPSW DMA data path init configuration.
 */
typedef struct CpswDma_DataPathConfig_s
{
    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;
}
CpswDma_DataPathConfig;

/*!
 * \brief Opaque handle for CPSW DMA driver object
 *
 * Opaque structure to hold software state for the CPSW DMA module.
 */
typedef struct CpswDma_DrvObj_s *CpswDma_Handle;

/* @} */

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/*!
 * \brief Initialize RX flow open parameters
 *
 * Initializes RX flow parameters with default values.
 *
 * \param pRxFlowPrm     [IN] Pointer to #CpswDma_OpenRxFlowPrms structure
 *
 */
void CpswDma_initRxFlowParams(CpswDma_OpenRxFlowPrms *pRxFlowPrm);

/*!
 * \brief Initialize TX channel open parameters
 *
 * Initializes TX channel open parameters with default values.
 *
 * \param pTxChCfg     [IN] Pointer to #CpswDma_OpenTxChPrms structure
 *
 */
void CpswDma_initTxChParams(CpswDma_OpenTxChPrms *pTxChCfg);

/*!
 * \brief Initialize packet information structure
 *
 * Initialize packet information structure with null/zero values.
 *
 * \param pktInfo     [IN] Pointer to #CpswDma_PktInfo structure
 */
void CpswDma_pktInfoInit(CpswDma_PktInfo *pktInfo);

/*!
 * \brief CPSW DMA open Rx flow
 *
 * Opens the CPSW DMA RX flow based on the flow parameters. This function
 * attaches flow using flowIdx provided in #CpswDma_OpenRxFlowPrms and configures
 * the UDMA flow. This also configures event if notifyCb is not null.
 *
 * Requirement:
 *
 * \param pRxFlowPrms       [IN] CPSW RX Flow parameters #CpswDma_OpenRxFlowPrms.
 *                          This parameter can't be NULL.
 *
 * \return RX flow opaque handle
 */
CpswDma_RxFlowHandle CpswDma_openRxFlow(CpswDma_OpenRxFlowPrms *pRxFlowPrms);

/*!
 * \brief CPSW DMA close Rx flow
 *
 * Closes the CPSW DMA RX flow and frees all associated resources. During close
 * operation, we flush FQ taking all DMA descriptors with packet submitted in
 * advance for reception and return to app. Also we retrieve all packets from
 * the CQ (packets received between last #CpswDma_retrieveRxPackets packets
 * call) and return those to app. App doesn't need to call #CpswDma_retrieveRxPackets
 * explicitly to retrieve these packets.
 *
 * Requirement:
 *
 * \param hRxFlow       [IN] CPSW DMA flow handle.
 *                      This parameter can't be NULL.
 * \param pFqPktInfoQ   [OUT] Pointer to #CpswDma_PktInfoQ structure where packets
 *                      from FQ (submitted for reception) are retrieved and
 *                      returned to application.
 *                      This parameter can't be NULL.
 *
 * \param pCqPktInfoQ   [OUT] Pointer to #CpswDma_PktInfoQ structure where packets
 *                      from CQ (received packets) are retrieved and returned to
 *                      application.
 *                      This parameter can't be NULL.
 *
 *  \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_closeRxFlow(CpswDma_RxFlowHandle hRxFlow,
                            CpswDma_PktInfoQ *pFqPktInfoQ,
                            CpswDma_PktInfoQ *pCqPktInfoQ);

/*!
 * \brief CPSW DMA open TX channel
 *
 * Opens the CPSW TX UDMA channel based on the channel parameters. This function
 * open TX channel using chNum provided in CpswDma_OpenTxChPrms() and configures
 * TX channel. This also configures event if notifyCb is not null.
 *
 * Requirement:
 *
 * \param pTxChCfg       [IN] CPSW TX channel parameters #CpswDma_OpenTxChPrms.
 *                       This parameter can't be NULL.
 *
 * \return TX channel opaque handle
 */
CpswDma_TxChHandle CpswDma_openTxCh(CpswDma_OpenTxChPrms *pTxChCfg);

/*!
 * \brief CPSW DMA close TX channel!
 *
 * Closes the CPSW DMA TX channel and frees all associated resources. During
 * close operation, we flush FQ taking all DMA descriptors with packet submitted
 * but not yet transmitted and return to app. Also we retrieve all packets from
 * the CQ (transmission completed packets) and return those to app. App doesn't
 * need to call CpswDma_retrieveTxDonePackets() explicitly to retrieve these
 * packets.
 *
 * Requirement:
 *
 * \param txChHandle    [IN] CPSW DMA TX Channel handle.
 *                      This parameter can't be NULL.
 * \param pFqPktInfoQ   [OUT] Pointer to #CpswDma_PktInfoQ structure where
 *                      packets from FQ (submitted but not transmitted) are
 *                      retrieved and returned to application.
 *                      This parameter can't be NULL.
 * \param pCqPktInfoQ   [OUT] Pointer to #CpswDma_PktInfoQ structure where
 *                      packets from CQ (transmission completed) are retrieved
 *                      and returned to application.
 *                      This parameter can't be NULL.
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_closeTxCh(CpswDma_TxChHandle txChHandle,
                          CpswDma_PktInfoQ *pFqPktInfoQ,
                          CpswDma_PktInfoQ *pCqPktInfoQ);

/*!
 * \brief Enable TX channel ring event
 *
 * Enables the ring completion event for TX channel. This allows application to
 * run time control ring events and retrieve packet once to do burst packet
 * processing.
 *
 * \param txChHandle  [IN] Handle to CPSW DMA TX channel
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_enableTxEvent(CpswDma_TxChHandle txChHandle);

/*!
 * \brief Disable TX channel ring event.
 *
 * Disable the ring completion event for TX channel. This allows application to
 * run time control ring events and retrieve packet once.
 *
 * \param txChHandle  [IN] Handle to CPSW DMA TX channel
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_disableTxEvent(CpswDma_TxChHandle txChHandle);

/*!
 * \brief Enable RX flow ring event
 *
 * Enables the ring completion event for RX Flow. This allows application to
 * run time control ring events and retrieve packet once to do burst packet
 * processing.
 *
 * \param hRxFlow  [IN] Handle to CPSW DMA RX Flow
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_enableRxEvent(CpswDma_RxFlowHandle hRxFlow);

/*!
 * \brief Disable RX flow ring event
 *
 * Disable the ring completion event for RX Flow. This allows application to
 * run time control ring events and retrieve packet once.
 *
 * \param hRxFlow [IN] Handle to CPSW DMA RX Flow
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_disableRxEvent(CpswDma_RxFlowHandle hRxFlow);

/*!
 * \brief Retrieve Rx ready (full) packet queue from Rx flow
 *
 * \param hRxFlow          [IN] flow handle for the flow we wish to retrieve
 *                         packets from
 * \param pFromFlowQueue   [OUT] Pointer to #CpswDma_PktInfoQ structure where
 *                         packets from hardware are retrieved and returned to
 *                         application
 *
 *  \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_retrieveRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                  CpswDma_PktInfoQ *pFromFlowQueue);

/*!
 * \brief Submit a queue of free packets for reception from RX flow
 *
 * \param hRxFlow           [IN] flow handle for the flow we wish to submit
 *                          packets to
 * \param pToFlowQueue      [OUT] Pointer to #CpswDma_PktInfoQ structure
 *                          containing packets to be submitted to flow
 * \param pUnUsedPktQueue  [OUT] Pointer to #CpswDma_PktInfoQ structure. This
 *                         queue is used by driver to return packets for which
 *                         it couldn't allocate the DMA descriptors
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                CpswDma_PktInfoQ *pToFlowQueue,
                                CpswDma_PktInfoQ *pUnUsedPktQueue);

/*!
 * \brief Retrieve Tx ready (empty) packet queue from Tx channel
 *
 * \param hTxCh           [IN] CPSW DMA TX channel handle for the TX channel we
 *                        wish to retrieve packets from
 * \param pFromTxChQueue  [OUT] Pointer to #CpswDma_PktInfoQ structure where
 *                        packets from hardware are retrieved and returned to
 *                        application
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_retrieveTxDonePackets(CpswDma_TxChHandle hTxCh,
                                      CpswDma_PktInfoQ *pFromTxChQueue);

/*!
 * \brief Submit a queue of ready (full) packet queue to TX channel
 *
 * \param hTxCh            [IN] TX Channel handle for the TX channel we wish to
 *                         submit packets to
 * \param pToTxChQueue     [OUT] Pointer to #CpswDma_PktInfoQ structure
 *                         containing packets to be submitted to TX channel
 * \param pUnUsedPktQueue  [OUT] Pointer to #CpswDma_PktInfoQ structure. This
 *                         queue is used by driver to return packets for which
 *                         it couldn't allocate the DMA descriptors
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_submitTxReadyPackets(CpswDma_TxChHandle hTxCh,
                                     CpswDma_PktInfoQ *pToTxChQueue,
                                     CpswDma_PktInfoQ *pUnUsedPktQueue);

/*!
 * \brief Check if any packet loss in RX flow FQ and CQ rings
 *
 * This development time error checks API checks DMA descriptor loss due to any
 * timing issues.  We check amount of descriptors with DMA and free matches the
 * number of allocated with some "margin" for packets in flight.
 *
 * Note: Margin can vary drastically based on Core (A72 vs R5), type of Udma
 * channel (NORMAL/HC etc.) and size of packet. This makes these API dependent
 * on application configuration.
 *
 * \param hRxFlow     [IN] flow handle for the flow we wish to check sanity
 * \param margin      [IN] Amount of packets which can be in flight in DMA
 *
 * \retval UDMA_SOK    Number of free descriptors and with DMA matches
 *                     (total allocated packets - margin) = (packets in CQ +
 *                     Packets in FQ)
 * \retval UDMA_EFAIL  Number offree descriptors and with DMA don't match
 */
int32_t CpswDma_checkRxFlowSanity(CpswDma_RxFlowHandle hRxFlow,
                                  uint32_t margin);

/*!
 * \brief Check if any packet loss in TX Channel FQ and CQ rings
 *
 * This development time error checks API checks dma desc loss due to any timing
 * issues. We check amount of descriptors with DMA and free matches the number
 * of allocated with some "margin" for packets in flight.
 *
 * Note: Margin can vary drastically based on Core (A72 vs R5), type of Udma
 * channel (NORMAL/HC etc.) and size of packet. This makes these API dependent
 * on application configuration.
 *
 * \param hTxCh       [IN] TX Channel handle for the TX channel we wish to check
 *                    sanity
 * \param margin      [IN] Amount of packets which can be in flight in DMA
 *
 * \retval UDMA_SOK    Number of free descriptors and with DMA matches (total
 *                     allocated packets - margin) = (packets in CQ + Packets
 *                     in FQ)
 * \retval UDMA_EFAIL  Number of free descriptors and with DMA don't match
 */
int32_t CpswDma_checkTxChSanity(CpswDma_TxChHandle hTxCh,
                                uint32_t margin);

/*!
 * \brief Get RX flow statistics
 *
 * \param hRxFlow      [IN] RX flow handle
 * \param rxFlowStats  [INOUT] Pointer to RX Flow stats object where stats would
 *                     be returned
 *
 * \retval CPSW_SOK            Stats are enabled and supported
 * \retval CPSW_ENOTSUPPORTED  Stats is not supported (enabled)
 */
int32_t CpswDma_getRxFlowStats(CpswDma_RxFlowHandle hRxFlow,
                               CpswDma_RxFlowStats *rxFlowStats);

/*!
 * \brief Get TX channel statistics
 *
 * \param hTxCh       [IN] TX channel handle
 * \param txChStats   [INOUT] Pointer to TX Channel stats object where stats
 *                    would be returned
 *
 * \retval CPSW_SOK            Stats are enabled and supported
 * \retval CPSW_ENOTSUPPORTED  Stats is not supported (enabled)
 */
int32_t CpswDma_getTxChStats(CpswDma_TxChHandle hTxCh,
                             CpswDma_TxChStats *txChStats);

/*!
 * \brief Reset RX flow statistics
 *
 * \param hRxFlow     [IN] RX flow handle
 *
 * \retval CPSW_SOK            Stats are enabled and supported
 * \retval CPSW_ENOTSUPPORTED  Stats is not supported (enabled)
 */
int32_t CpswDma_resetRxFlowStats(CpswDma_RxFlowHandle hRxFlow);

/*!
 * \brief Reset TX channel statistics
 *
 * \param hTxCh     [IN] TX Channel handle
 *
 * \retval CPSW_SOK            Stats are enabled and supported
 * \retval CPSW_ENOTSUPPORTED  Stats is not supported (enabled)
 */
int32_t CpswDma_resetTxChStats(CpswDma_TxChHandle hTxCh);

/*!
 * \brief Set default data path parameters
 *
 * \param pDmaConfig   [IN] Data path config structure to be initialized
 */
void CpswDma_initDataPathParams(CpswDma_DataPathConfig *pDmaConfig);

/*!
 * \brief Initialize data path
 *
 * Initialize DMA data path state on remote core.
 *
 * \param cpswType               [IN] Cpsw Type
 * \param pCpswDmaDataPathConfig [IN] Data Path Init params
 *
 * \return CPSW DMA opaque handle
 */
CpswDma_Handle CpswDma_initDataPath(Cpsw_Type cpswType,
                                    const CpswDma_DataPathConfig *pCpswDmaDataPathConfig);

/*!
 * \brief De-initialize data path
 *
 * De-initialize DMA data path state on remote core.
 *
 * \param hCpswDma     [IN] Cpsw DMA Handle
 *
 * \return \ref Udma_ErrorCodes
 */
int32_t CpswDma_deInitDataPath(CpswDma_Handle hCpswDma);

/*!
 * \brief Get TX channel FQ handle
 *
 * Returns FQ handle for TX Channel. This function can be used when auto-recycling
 * of buffers is used.
 *
 * \param hTxCh     [IN] TX Channel handle
 *
 * \return Udma_RingHandle - Handle to FQ ring
 */
Udma_RingHandle CpswDma_getTxChFqHandle(CpswDma_TxChHandle hTxCh);

/*!
 * \brief Get handle to DMA descriptor free pool for TX channel
 *
 * Returns DMA descriptor free pool for TX channel. This function can be used when auto-recycling
 * of buffers is used.
 *
 * \param hTxCh     [IN] TX channel handle
 *
 * \return CpswDma_DmaDescQHandle - Handle to DMA descriptor free pool
 */
CpswDma_DmaDescQHandle CpswDma_getTxChDescPoolHandle(CpswDma_TxChHandle hTxCh);

/*!
 * \brief Get RX flow FQ handle
 *
 * Returns FQ handle for RX flow. This function can be used when auto-recycling
 * of buffers is used.
 *
 * \param hRxFlow     [IN] RX flow handle
 *
 * \return Udma_RingHandle - Handle to FQ ring
 */
Udma_RingHandle CpswDma_getRxFlowFqHandle(CpswDma_RxFlowHandle hRxFlow);

/*!
 * \brief Get handle to DMA descriptor free pool for RX flow
 *
 * Returns DMA descriptor free pool for RX flow. This function can be used when auto-recycling
 * of buffers is used.
 *
 * \param hRxFlow     [IN] RX flow handle
 *
 * \return CpswDma_DmaDescQHandle - Handle to DMA descriptor free pool
 */
CpswDma_DmaDescQHandle CpswDma_getRxFlowDescPoolHandle(CpswDma_RxFlowHandle hRxFlow);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_DMA_H_ */

/*!
 *  @}
 */
