/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_dma_pktutils.h
 *
 *  \brief CPSW DMA utility API to check packet and descriptor sanity.
 */

/*!
 *  \ingroup CPSW_DMA_API
 *  \defgroup CPSW_DMA_PKT_UTILS CPSW DMA Packet Utils
 * @{
 */

#ifndef CPSW_DMA_PKTUTILS_H_
#define CPSW_DMA_PKTUTILS_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdarg.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#if (CPSW_CFG_DEV_ERROR == 1)
#define CPSW_UTILS_GET_STATE(state, mask, bitshift)    ((*(state) & (mask)) >> (bitshift))
#define CPSW_UTILS_SET_STATE(state, mask, bitshift, value) \
    ({                                                     \
         *(state) &= ~(mask);                              \
         *(state) |= ((value) << (bitshift));              \
     })
#else
#define CPSW_UTILS_GET_STATE(state, mask, bitshift)         ({while (false) {; } \
                                                             })
#define CPSW_UTILS_SET_STATE(state, mask, bitshift, value)  ({while (false) {; } \
                                                             })
#endif

#define CPSW_UTILS_DRIVER_STATE_MASK            (0x000000FFU)
#define CPSW_UTILS_DRIVER_STATE_BIT_SHIFT       (0U)

#define CPSW_UTILS_MEMUTILS_STATE_MASK          (0x0000FF00U)
#define CPSW_UTILS_MEMUTILS_STATE_BIT_SHIFT     (8U)

#define CPSW_UTILS_APP_STATE_MASK               (0x00FF0000U)
#define CPSW_UTILS_APP_STATE_BIT_SHIFT          (16U)

#define CPSW_UTILS_GET_PKT_DRIVER_STATE(state)          \
    (CPSW_UTILS_GET_STATE(state,                        \
                          CPSW_UTILS_DRIVER_STATE_MASK, \
                          CPSW_UTILS_DRIVER_STATE_BIT_SHIFT))

#define CPSW_UTILS_SET_PKT_DRIVER_STATE(state, value)        \
    (CPSW_UTILS_SET_STATE(state,                             \
                          CPSW_UTILS_DRIVER_STATE_MASK,      \
                          CPSW_UTILS_DRIVER_STATE_BIT_SHIFT, \
                          value))

#define CPSW_UTILS_GET_PKT_MEMUTILS_STATE(state)          \
    (CPSW_UTILS_GET_STATE(state,                          \
                          CPSW_UTILS_MEMUTILS_STATE_MASK, \
                          CPSW_UTILS_MEMUTILS_STATE_BIT_SHIFT))

#define CPSW_UTILS_SET_PKT_MEMUTILS_STATE(state, value)        \
    (CPSW_UTILS_SET_STATE(state,                               \
                          CPSW_UTILS_MEMUTILS_STATE_MASK,      \
                          CPSW_UTILS_MEMUTILS_STATE_BIT_SHIFT, \
                          value))

#define CPSW_UTILS_GET_PKT_APP_STATE(state)          \
    (CPSW_UTILS_GET_STATE(state,                     \
                          CPSW_UTILS_APP_STATE_MASK, \
                          CPSW_UTILS_APP_STATE_BIT_SHIFT))

#define CPSW_UTILS_SET_PKT_APP_STATE(state, value)        \
    (CPSW_UTILS_SET_STATE(state,                          \
                          CPSW_UTILS_APP_STATE_MASK,      \
                          CPSW_UTILS_APP_STATE_BIT_SHIFT, \
                          value))

#define CPSW_UTILS_MEMUTILS_DESCSTATE_MASK                      (0x000000FFU)
#define CPSW_UTILS_MEMUTILS_DESCSTATE_SHIFT                     (0U)

#define CPSW_UTILS_GET_DESC_MEMUTILS_STATE(state)             \
    (CPSW_UTILS_GET_STATE(state,                              \
                          CPSW_UTILS_MEMUTILS_DESCSTATE_MASK, \
                          CPSW_UTILS_MEMUTILS_DESCSTATE_SHIFT))

#define CPSW_UTILS_SET_DESC_MEMUTILS_STATE(state, value)       \
    (CPSW_UTILS_SET_STATE(state,                               \
                          CPSW_UTILS_MEMUTILS_DESCSTATE_MASK,  \
                          CPSW_UTILS_MEMUTILS_DESCSTATE_SHIFT, \
                          value))

/*!
 * \brief CPSW DMA Descriptor state
 *
 * DMA descriptor state used by CPSW DMA driver for sanity check.
 */
typedef enum CpswUtils_DescStateMemMgr_e
{
    /*! Desc free */
    CPSW_DESCSTATE_MEMMGR_FREE = 0U,

    /*! Desc Allocated */
    CPSW_DESCSTATE_MEMMGR_ALLOC
} CpswUtils_DescStateMemMgr;

/*!
 * \brief App Pkt state
 *
 * Packet state for application
 */
typedef enum CpswUtils_PktStateApp_e
{
    /*! Packet with APP Free Q*/
    CPSW_PKTSTATE_APP_WITH_FREEQ = 0U,

    /*! Packet with APP Ready Q*/
    CPSW_PKTSTATE_APP_WITH_READYQ,

    /*! Packet with Driver*/
    CPSW_PKTSTATE_APP_WITH_DRIVER
} CpswUtils_PktStateApp;

/*!
 * \brief DMA Pkt state
 *
 * Packet state for DMA
 */
typedef enum CpswUtils_PktStateDma_e
{
    /*! Packet with HW  */
    CPSW_PKTSTATE_DMA_WITH_HW = 0U,

    /*! Packet not with HW  */
    CPSW_PKTSTATE_DMA_NOT_WITH_HW
} CpswUtils_PktStateDma;

/*!
 * \brief CPSW DMA Pkt state
 *
 * Packet state for memory manager
 */
typedef enum CpswUtils_PktStateMemMgr_e
{
    /*! Packet free */
    CPSW_PKTSTATE_MEMMGR_FREE = 0U,

    /*! Packet Allocated */
    CPSW_PKTSTATE_MEMMGR_ALLOC,
} CpswUtils_PktStateMemMgr;

/*!
 * \brief CPSW DMA Pkt state
 *
 * Type of CPSW DMA Pkt state.
 */
typedef enum CpswUtils_PktStateModuleType_e
{
    /*! Module type is app */
    CPSW_PKTSTATE_MODULE_APP = 0U,

    /*! Module type is DMA Driver  */
    CPSW_PKTSTATE_MODULE_DRIVER,

    /*! Module type is MEM Mgr */
    CPSW_PKTSTATE_MODULE_MEMUTILS,
} CpswUtils_PktStateModuleType;

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Packet state check & set API for app, memory manager and dma driver.
 *
 * This development time API checks expected state and sets new packet state. This
 * is used by modules handling the packet to confirm the packet is returned and allocated
 * in correct state. Memutils checks (free, allocated), dma driver checks (with Hw, with driver)
 * and app checks which Q it belongs
 *
 *  \param pStateVar     [IN/OUT] Pointer to packet state variable of packet structure #CpswDma_PktInfo
 *  \param module        [IN] Module id. Refer to #CpswUtils_PktStateModuleType
 *  \param expectedState [IN] State packet is expected to be.
 *  \param newState      [IN] New state packet is set to.
 *
 * \note - Optional for MemMgr and app. DMA driver mandatory uses when CPSW_CFG_DEV_ERROR is defined.
 *
 */
void CpswUtils_checkPktState(uint32_t *pStateVar,
                             CpswUtils_PktStateModuleType module,
                             uint32_t expectedState,
                             uint32_t newState);

/*!
 * \brief Descriptor state check & set API for descriptor memory manager.
 *
 * This development time API checks expected state and sets new descriptor state. This
 * is used by modules handling the Desc to confirm the Desc is returned and allocated
 * in correct state. Only memutils modules uses it now for checking if desc is free or allocated
 * while allocating or freeing desc respectively.
 *
 *  \param pStateVar     [IN/OUT] Pointer to Desc state variable.
 *  \param expectedState [IN] State Desc is expected to be.
 *  \param newState      [IN] New state Desc is set to.
 *
 * \note - memory manager may not allocate the descState as this is optional check for MemMgr.
 */
void CpswUtils_checkDescState(uint32_t *pStateVar,
                              CpswUtils_DescStateMemMgr expectedState,
                              CpswUtils_DescStateMemMgr newState);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_DMA_PKTUTILS_H_ */

/*!
 *  @}
 */
