/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_hostport.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the Port module.
 */

/*!
 * \addtogroup CPSW_HOSTPORT_API
 * @{
 */

#ifndef CPSW_HOSTPORT_H_
#define CPSW_HOSTPORT_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"
#include "cpsw_port.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Helper macro to create IOCTL commands for host port submodule */
#define CPSW_HOSTPORT_PUBLIC_IOCTL(x)         (CPSW_IOCTL_HOSTPORT_BASE | CPSW_IOCTL_MIN(x))

/*! \brief Default value used for Host port RX MTU */
#define CPSW_HOSTPORT_RX_MTU_DEFAULT          (1518U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief Host port IOCTL commands
 */
enum CpswHostPort_Ioctl_e
{
    /*!
     * \brief Enable host port
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_ENABLE               = CPSW_HOSTPORT_PUBLIC_IOCTL(0U),

    /*!
     * \brief Disable host port
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_DISABLE              = CPSW_HOSTPORT_PUBLIC_IOCTL(1U),

    /*!
     * \brief Get host port's FIFO statistics
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #CpswHostPort_FifoStats
     */
    CPSW_HOSTPORT_IOCTL_GET_FIFO_STATS       = CPSW_HOSTPORT_PUBLIC_IOCTL(2U),

    /*!
     * \brief Set TX priority map (Header Packet Priority to Switch Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPort_TxPriority
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY      = CPSW_HOSTPORT_PUBLIC_IOCTL(3U),

    /*!
     * \brief Get TX priority map (Header Packet Priority to Switch Priority)
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #CpswPort_TxPriority
     */
    CPSW_HOSTPORT_IOCTL_GET_TX_PRIORITY      = CPSW_HOSTPORT_PUBLIC_IOCTL(4U),

    /*!
     * \brief Set RX priority map (Packet Priority to Header Packet Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPort_RxPriority
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY      = CPSW_HOSTPORT_PUBLIC_IOCTL(5U),

    /*!
     * \brief Get RX priority map (Packet Priority to Header Packet Priority)
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #CpswPort_RxPriority
     */
    CPSW_HOSTPORT_IOCTL_GET_RX_PRIORITY      = CPSW_HOSTPORT_PUBLIC_IOCTL(6U),

    /*!
     * \brief Set RX DSCP priority (TOS) map
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPort_DscpPriority
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY = CPSW_HOSTPORT_PUBLIC_IOCTL(7U),

    /*!
     * \brief Get RX DSCP priority (TOS) map
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t
     * - outArgs: #CpswPort_DscpPriority
     */
    CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY = CPSW_HOSTPORT_PUBLIC_IOCTL(8U),

    /*!
     * \brief Set traffic shapping configuration
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPort_TrafficShapingConfig
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_SET_TRAFFIC_SHAPING  = CPSW_HOSTPORT_PUBLIC_IOCTL(9U),

    /*!
     * \brief Get traffic shapping configuration
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #CpswPort_TrafficShapingConfig
     */
    CPSW_HOSTPORT_IOCTL_GET_TRAFFIC_SHAPING  = CPSW_HOSTPORT_PUBLIC_IOCTL(10U),

    /*!
     * \brief Print host port registers
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_HOSTPORT_IOCTL_PRINT_REGS           = CPSW_HOSTPORT_PUBLIC_IOCTL(11U),

    /*!
     * \brief Get host port Rx MTU
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: uint32_t
     */
    CPSW_HOSTPORT_IOCTL_GET_RX_MTU           = CPSW_HOSTPORT_PUBLIC_IOCTL(12U),

    /*!
     * \brief Get host port Flow Id offset
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: uint32_t
     */
    CPSW_HOSTPORT_GET_FLOW_ID_OFFSET         = CPSW_HOSTPORT_PUBLIC_IOCTL(13U),

    /*!
     * \brief Returns true if  checksum offload is enabled else false
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: bool
     */
    CPSW_HOSTPORT_IS_CSUM_OFFLOAD_ENABLE     = CPSW_HOSTPORT_PUBLIC_IOCTL(14U),
};

/*!
 * \brief Port priority type
 */
typedef enum CpswHostPort_RxPriorityType_e
{
    /*! Fixed priority */
    CPSW_RX_PRIORITY_TYPE_FIXED = 0U,

    /*! Round-robin priority */
    CPSW_RX_PRIORITY_TYPE_RR    = 1U,
} CpswHostPort_RxPriorityType;

/*!
 * \brief FIFO related statistics of the host port
 */
typedef struct CpswHostPort_FifoStats_s
{
    /*! Max throughput of the CPPI FIFO into the CPSW_NU */
    uint32_t rxThroughputRate;

    /*! Number of blocks allocated to the FIFO logical TX queues */
    uint32_t txBlockCount;

    /*! Number of blocks allocated in the RX FIFO Express MAC
     *  If IET is not enabled all traffic is express traffic
     */
    uint32_t rxBlockCountExpress;

    /*! Number of blocks allocated in the RX FIFO Preempt MAC */
    uint32_t rxBlockCountPreempt;

    /*! Whether a FIFO priority has one or more queued packets or not */
    bool txActiveFifo[CPSW_PRIORITY_NUM];
} CpswHostPort_FifoStats;

/*!
 * \brief Host port submodule configuration
 */
typedef struct CpswHostPort_Config_s
{
    /*! Type of CRC on all port 0 egress, regardless of the CRC type in Ethernet
     *  port ingress */
    Cpsw_CrcType crcType;

    /*! Whether or not CRC is removed on port 0 egress */
    bool removeCrc;

    /*! Whether short packets (ingress) are padded to 64-bytes or dropped */
    bool padShortPacket;

    /*! Whether packets with CRC errors (ingress) are dropped or transferred to
     *  the destination ports */
    bool passCrcErrors;

    /*! Max length of a received frame on ingress, including VLAN */
    uint32_t rxMtu;

    /*! Whether priority tagged packets should be passed unchanged (if set to
     *  true) or replaced with port's VID (if set to false) */
    bool passPriorityTaggedUnchanged;

    /*! Enable checksum offload feature which allows TCP checksum computation
     *  to be offloaded to the CPSW for frames transmitted by any core.
     *  The Protocol specific info needs to be populated in the descriptor
     *  to indicate the location in the packet where the computed checksum
     *  should be inserted */
    bool enableCsumOffload;

    /*! RX VLAN remap controls whether the hardware switch priority for VLAN
     *  tagged or priority tagged packets is determined from CPPI thread number
     *  (remap disabled) or via \ref CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY
     *  (remap enabled) */
    bool enableRxVlanRemap;

    /*! RX DSCP IPv4 remap controls whether the hardware switch priority for
     *  IPv4 packets is determined from CPPI thread number (remap disabled) or
     *  via \ref CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY (remap enabled) */
    bool enableRxDscpIPv4Remap;

    /*! RX DSCP IPv6 remap controls whether the hardware switch priority for
     *  IPv6 packets is determined from CPPI thread number (remap disabled) or
     *  via \ref CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY (remap enabled) */
    bool enableRxDscpIPv6Remap;

    /*! Port VLAN configuration */
    CpswPort_VlanConfig vlanCfg;

    /*! Ingress priority type */
    CpswHostPort_RxPriorityType rxPriorityType;

    /*! Egress priority type */
    CpswPort_TxPriorityType txPriorityType;
} CpswHostPort_Config;

/*!
 * \brief Host port submodule handle
 *
 * Host port submodule opaque handle used to call any host port related APIs.
 */
typedef struct CpswHostPort_Obj_s *CpswHostPort_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_HOSTPORT_H_ */

/* @} */
