/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_macport.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the MAC Port submodule.
 */

/*!
 * \addtogroup CPSW_MACPORT_API
 * @{
 */

#ifndef CPSW_MACPORT_H_
#define CPSW_MACPORT_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"
#include "cpsw_port.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Helper macro to create IOCTL commands for MAC port submodule */
#define CPSW_MACPORT_PUBLIC_IOCTL(x)          (CPSW_IOCTL_MACPORT_BASE | CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief MAC port IOCTL commands
 */
enum CpswMacPort_Ioctls_e
{
    /*!
     * \brief Get MAC port status
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswMacPort_PortStatus
     */
    CPSW_MACPORT_IOCTL_GET_STATUS           = CPSW_MACPORT_PUBLIC_IOCTL(0U),

    /*!
     * \brief Get MAC port FIFO statistics
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswMacPort_FifoStats
     */
    CPSW_MACPORT_IOCTL_GET_FIFO_STATS       = CPSW_MACPORT_PUBLIC_IOCTL(1U),

    /*!
     * \brief Get MAC link configuration (speed, duplexity)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswMacPort_LinkConfig
     */
    CPSW_MACPORT_IOCTL_GET_LINK_CONFIG      = CPSW_MACPORT_PUBLIC_IOCTL(2U),

    /*!
     * \brief Check if MAC port is in in-band mode
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: bool
     */
    CPSW_MACPORT_IOCTL_IS_INBAND            = CPSW_MACPORT_PUBLIC_IOCTL(3U),

    /*!
     * \brief Set TX priority map (Header Packet Priority to Switch Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_SetTxPriorityInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_TX_PRIORITY      = CPSW_MACPORT_PUBLIC_IOCTL(4U),

    /*!
     * \brief Get TX priority map (Header Packet Priority to Switch Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswPort_TxPriority
     */
    CPSW_MACPORT_IOCTL_GET_TX_PRIORITY      = CPSW_MACPORT_PUBLIC_IOCTL(5U),

    /*!
     * \brief Set Rx priority map (Packet Priority to Header Packet Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_SetRxPriorityInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_RX_PRIORITY      = CPSW_MACPORT_PUBLIC_IOCTL(6U),

    /*!
     * \brief Get Rx priority map (Packet Priority to Header Packet Priority)
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswPort_RxPriority
     */
    CPSW_MACPORT_IOCTL_GET_RX_PRIORITY      = CPSW_MACPORT_PUBLIC_IOCTL(7U),

    /*!
     * \brief Set RX DSCP priority (TOS) map
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_SetRxDscpPriorityInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY = CPSW_MACPORT_PUBLIC_IOCTL(8U),

    /*!
     * \brief Get RX DSCP priority (TOS) map
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswPort_DscpPriority
     */
    CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY = CPSW_MACPORT_PUBLIC_IOCTL(9U),

    /*!
     * \brief Set traffic shaping configuration
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_SetTrafficShapingInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_TRAFFIC_SHAPING  = CPSW_MACPORT_PUBLIC_IOCTL(10U),

    /*!
     * \brief Set traffic shaping configuration
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: #CpswPort_TrafficShapingConfig
     */
    CPSW_MACPORT_IOCTL_GET_TRAFFIC_SHAPING  = CPSW_MACPORT_PUBLIC_IOCTL(11U),

    /*!
     * \brief Set the number of unused blocks remaining in TX FIFOs for rate limiting
     *
     * Sets the number of blocks in each of the MAC port TX FIFOs that must be unused
     * for bulk traffic to be enabled.
     *
     * This IOCTL should be used along with CPSW_HOSTPORT_IOCTL_SET_TRAFFIC_SHAPING
     * for host port receive rate limiting.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_SetTxBlocksRemInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_TX_BLOCKS_REM    = CPSW_MACPORT_PUBLIC_IOCTL(12U),

    /*!
     * \brief Get the number of unused blocks remaining in TX FIFOs for rate limiting
     *
     * Gets the number of blocks in each of the MAC port TX FIFOs that must be unused
     * for bulk traffic to be enabled.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: uint32_t
     */
    CPSW_MACPORT_IOCTL_GET_TX_BLOCKS_REM    = CPSW_MACPORT_PUBLIC_IOCTL(13U),

    /*!
     * \brief Print MAC port registers
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_PRINT_REGS           = CPSW_MACPORT_PUBLIC_IOCTL(14U),

    /*!
     * \brief Get Rx MTU
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_GenericInArgs
     * - outArgs: uint32_t
     */
    CPSW_MACPORT_IOCTL_GET_RX_MTU           = CPSW_MACPORT_PUBLIC_IOCTL(15U),

    /*!
     * \brief Enable Ethernet Port CPTS Event
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMacPort_EnableTsEventInArgs
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT    = CPSW_MACPORT_PUBLIC_IOCTL(16U),

    /*!
     * \brief Disable Ethernet Port CPTS Event
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_MacPort
     * - outArgs: None
     */
    CPSW_MACPORT_IOCTL_DISABLE_CPTS_EVENT   = CPSW_MACPORT_PUBLIC_IOCTL(17U),
};

/*!
 * \brief MAC interface layer type
 */
typedef enum CpswMac_LayerType_e
{
    /*! Media-Independent Interface (MII) layer */
    CPSW_MAC_LAYER_MII   = 1U,

    /*! Gigabit Media-Independent Interface (GMII) layer */
    CPSW_MAC_LAYER_GMII  = 0U,

    /*! 10-Gigabit Media-Independent Interface (XGMII) layer */
    CPSW_MAC_LAYER_XGMII = 2U,
} CpswMac_LayerType;

/*!
 * \brief MAC interface sublayer type
 */
typedef enum CpswMac_SublayerType_e
{
    /*! Standard interface sublayer */
    CPSW_MAC_SUBLAYER_STANDARD             = 0U,

    /*! Reduced interface sublayer */
    CPSW_MAC_SUBLAYER_REDUCED              = 1U,

    /*! Serial interface sublayer */
    CPSW_MAC_SUBLAYER_SERIAL               = 2U,

    /*! Quad-serial interface sublayer (main port) */
    CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN     = 3U,

    /*! Quad-serial interface sublayer (sub port) */
    CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB      = 4U,
} CpswMac_SublayerType;

/*!
 * \brief MAC interface variant (only for specific interface types)
 */
typedef enum CpswMac_VariantType_e
{
    /*! No variant */
    CPSW_MAC_VARIANT_NONE       = 0U,

    /*! Forced mode (out of band) */
    CPSW_MAC_VARIANT_FORCED     = 1U,
} CpswMac_VariantType;

/*!
 * \brief MAC port interVLAN route identifier
 */
typedef enum CpswMacPort_InterVLANRouteId_e
{
    /*! First interVLAN route Id */
    CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST,

    /*! InterVLAN route Id 1 */
    CPSW_MACPORT_INTERVLAN_ROUTEID_1    = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST,

    /*! InterVLAN route Id 2 */
    CPSW_MACPORT_INTERVLAN_ROUTEID_2,

    /*! InterVLAN route Id 3 */
    CPSW_MACPORT_INTERVLAN_ROUTEID_3,

    /*! InterVLAN route Id 4 */
    CPSW_MACPORT_INTERVLAN_ROUTEID_4,

    /*! Last interVLAN route Id */
    CPSW_MACPORT_INTERVLAN_ROUTEID_LAST = CPSW_MACPORT_INTERVLAN_ROUTEID_4
} CpswMacPort_InterVLANRouteId;

/*!
 *  \brief Enum of VLAN usage type for CPTS packets
 */
typedef enum CpswMacPort_VlanType_e
{
    /*! No VLAN tag */
    CPSW_MACPORT_VLAN_TYPE_NONE                 = 0x00U,

    /*! Single VLAN tag */
    CPSW_MACPORT_VLAN_TYPE_SINGLE_TAG           = 0x01U,

    /*! Stacked VLAN tag */
    CPSW_MACPORT_VLAN_TYPE_STACKED_TAGS         = 0x02U,
} CpswMacPort_VlanType;

/*!
 * \brief MAC port interface
 */
typedef struct CpswMacPort_Interface_s
{
    /*! Interface layer type */
    CpswMac_LayerType layerType;

    /*! Interface sublayer type */
    CpswMac_SublayerType sublayerType;

    /*! Variant type */
    CpswMac_VariantType variantType;
} CpswMacPort_Interface;


/*!
 * \brief SGMII mode type
 */
typedef enum CpswMac_SgmiiModeType_e
{
    /*! Invalid mode. Set to when port is used in non Q/SGMII modes */
    CPSW_MAC_SGMIIMODE_INVALID              = 0U,

    /*! SGMII in fiber mode with PHY connection */
    CPSW_MAC_SGMIIMODE_FIBER_WITH_PHY       = 1U,

    /*! SGMII in SGMII mode with PHY connection */
    CPSW_MAC_SGMIIMODE_SGMII_WITH_PHY       = 2U,

    /*! SGMII in MAC to MAC with auto-neg master mode */
    CPSW_MAC_SGMIIMODE_SGMII_AUTONEG_MASTER = 3U,

    /*! SGMII in MAC to MAC with auto-neg slave mode */
    CPSW_MAC_SGMIIMODE_SGMII_AUTONEG_SLAVE  = 4U,

    /*! SGMII in MAC to MAC with forced master mode */
    CPSW_MAC_SGMIIMODE_SGMII_FORCEDLINK     = 5U,
} CpswMac_SgmiiModeType;

/*!
 * \brief SGMII configuration for MAC port
 */
typedef struct CpswMacPort_SgmiiConfig_s
{
    /*! SGMII mode of operation */
    CpswMac_SgmiiModeType  sgmiiMode;
} CpswMacPort_SgmiiConfig;

/*!
 * \brief MAC port configuration
 */
typedef struct CpswMacPort_Config_s
{
    /*! Whether loopback mode is enabled or not */
    bool enableLoopback;

    /*! Type of CRC */
    Cpsw_CrcType crcType;

    /*! Max length of a received frame on ingress. This max length includes
     *  VLAN */
    uint32_t rxMtu;

    /*! Whether priority tagged packets should be passed unchanged (if set to
     *  true) or replaced with port's VID (if set to false) */
    bool passPriorityTaggedUnchanged;

    /*! Port VLAN configuration */
    CpswPort_VlanConfig vlanConfig;

    /*! Egress priority type */
    CpswPort_TxPriorityType txPriorityType;

    /*! SGMII config when port is used in Q/SGMII mode */
    CpswMacPort_SgmiiConfig sgmiiConfig;
} CpswMacPort_Config;

/*!
 * \brief Generic input args
 */
typedef struct CpswMacPort_GenericInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;
} CpswMacPort_GenericInArgs;

/*!
 * \brief Port status
 */
typedef struct CpswMacPort_PortStatus_s
{
    /*! Whether the Ethernet port (express and preempt) is in idle state or not */
    bool portIdle;

    /*! Whether or not the express MAC is idle */
    bool expressMacIdle;

    /*! Whether or not the preempt MAC is idle */
    bool preemptMacIdle;

    /*! Whether or not the MAC transmit is idle (both, preempt and express MAC
     *  are in idle state */
    bool macTxIdle;

    /*! MAC address configured for the port */
    uint8_t macAddr[CPSW_MAC_ADDR_LEN];
} CpswMacPort_PortStatus;

/*!
 * \brief FIFO related statistics of a MAC port
 */
typedef struct CpswMacPort_FifoStats_s
{
    /*! Max throughput of the Ethernet ports to the crossbar SCR */
    uint32_t rxThroughputRate;

    /*! Number of 32-byte packet words in an Ethernet port transmit FIFO before
     *  packet egress will begin */
    uint32_t txStartWords;

    /*! Max number of blocks allowed on all TX FIFO priorities combined */
    uint32_t txMaxBlocks;

    /*! Max number of blocks allowed on the express and preempt RX FIFOs
     *  combined */
    uint32_t rxMaxBlocks;

    /*! Number of blocks allocated to the FIFO logical TX queues */
    uint32_t txBlockCount;

    /*! Number of blocks allocated to the FIFO express RX queue */
    uint32_t rxExpressBlockCount;

    /*! Number of blocks allocated to the FIFO preempt RX queue */
    uint32_t rxPreemptBlockCount;

    /*! Whether a FIFO priority has one or more queued packets or not */
    bool txActiveFifo[CPSW_PRIORITY_NUM];
} CpswMacPort_FifoStats;

/*!
 * \brief Link speed and duplexity configuration
 */
typedef struct CpswMacPort_LinkConfig_s
{
    /*! Link speed */
    Cpsw_Speed speed;

    /*! Duplexity */
    Cpsw_Duplexity duplexity;
} CpswMacPort_LinkConfig;

/*!
 * \brief Input args for #CPSW_MACPORT_IOCTL_SET_TX_PRIORITY command
 */
typedef struct CpswMacPort_SetTxPriorityInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! TX priority map (Header Packet Priority to Switch Priority) */
    CpswPort_TxPriority txPriority;
} CpswMacPort_SetTxPriorityInArgs;

/*!
 * \brief Input args for #CPSW_MACPORT_IOCTL_SET_RX_PRIORITY command
 */
typedef struct CpswMacPort_SetRxPriorityInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! RX priority map (Packet Priority to Header Packet Priority) */
    CpswPort_RxPriority rxPriority;
} CpswMacPort_SetRxPriorityInArgs;

/*!
 * \brief Input args for #CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY command
 */
typedef struct CpswMacPort_SetRxDscpPriorityInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! RX DSCP priority (TOS) map */
    CpswPort_DscpPriority dscpPriority;
} CpswMacPort_SetRxDscpPriorityInArgs;

/*!
 * \brief Input args for #CPSW_MACPORT_IOCTL_SET_TRAFFIC_SHAPING command
 */
typedef struct CpswMacPort_SetTrafficShapingInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Traffic shaping configuration for all priorities */
    CpswPort_TrafficShapingConfig trafficShapingConfig;
} CpswMacPort_SetTrafficShapingInArgs;

/*!
 * \brief Input args for #CPSW_MACPORT_IOCTL_SET_TX_BLOCKS_REM command
 */
typedef struct CpswMacPort_SetTxBlocksRemInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Transmit FIFO blocks that must be free before a non rate-limited
     *  host port 0 receive thread can begin sending a packet */
    uint32_t txBlocksRemaining;
} CpswMacPort_SetTxBlocksRemInArgs;

/*!
 * \brief Port interVLAN configuration info
 */
typedef struct CpswMacPort_InterVlanRoutingParams_s
{
    /*! Destination address */
    uint8_t dstAddr[CPSW_MAC_ADDR_LEN];

    /*! Source address */
    uint8_t srcAddr[CPSW_MAC_ADDR_LEN];

    /*! VLAN Id */
    uint32_t vlanId;

    /*! Whether routed packet's destination address will be replaced with
     *  dstAddr and packet's source address will be replaced with srcAddr */
    bool replaceDASA;

    /*! Destination VLAN Force Untagged Egress - When set, this bit indicates
     * that the VLAN should be removed on egress for the routed packet.
     */
    bool forceUntaggedEgress;

    /*! Decrement Time To Live.
     *  When set, the Time To Live (TTL) field in the header is decremented:
     *  - IPV4 - Decrement the TTL byte and update the Header Checksum
     *  - IPV6 - Decrement the Hop Limit
     *
     *  Note: When this bit is set, the ALE should be configured to send any
     *  IPv4/6 packet with a zero or one TTL field to the host.
     *  When this bit is cleared the TTL/Hop Limit fields are not checked or
     *  modified.
     */
    bool decrementTTL;
} CpswMacPort_InterVlanRoutingParams;

/*!
 * \brief Configuration of transmit short inter packet gap feature
 */
typedef struct CpswMacPort_TxShortGapConfig_s
{
    /*! Transmit Short Gap Enable.
     *  When set this flag causes transmit with a short IPG when the TX FIFO
     *  length for the port exceeds the configured CPSW gap threshold
     */
    bool enableTxShortGap;

    /*! Transmit Short Gap Limit Enable.
     *  When set this flag limits the number of short gap packets transmitted to
     *  100ppm.  Each time a short gap packet is sent, a counter is loaded with
     *  10,000 and decremented on each wireside clock.  Another short gap packet
     *  will not be sent out until the counter decrements to zero.
     *  This mode is included to preclude the host from filling up the FIFO and
     *  sending every packet out with short gap which would violate the maximum
     *  number of packets per second allowed.
     *
     *  This flag is used only with GMII (not XGMII)
     */
    bool enableTxShortGapLimit;
} CpswMacPort_TxShortGapConfig;

/*!
 * \brief Set transmit short inter packet gap feature input Args
 */
typedef struct CpswMacPort_PortTxShortGapConfig_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Short interpacket gap configuration */
    CpswMacPort_TxShortGapConfig shortGapCfg;
} CpswMacPort_PortTxShortGapConfig;

/*!
 * \brief Port IP configuration for time synchronization events
 *
 *  Common configuration for Annex D (IPv4) and Annex E (IPv6) time sync events
 */
typedef struct CpswMacPort_IpTsConfig_s
{
    /*! Enable Time Sync Time to Live Non-zero Enable */
    bool enableTTLNonzero;

    /*! Time Sync Unicast Enable */
    bool enableUnicast;

    /*! Time Sync Destination IP Address 129 Enable */
    bool enableTsIp129;

    /*! Time Sync Destination IP Address 130 Enable */
    bool enableTsIp130;

    /*! Time Sync Destination IP Address 131 Enable */
    bool enableTsIp131;

    /*! Time Sync Destination IP Address 132 Enable*/
    bool enableTsIp132;

    /*! Time Sync Destination IP Address 129 Enable */
    bool enableTsIp107;

    /*! Time Sync Destination port number 319 Enable */
    bool enableTsPort319;

    /*! Time Sync Destination port number 320 Enable*/
    bool enableTsPort320;

} CpswMacPort_IpTsConfig;

/*!
 *  \brief  Port Configuration for Time Synchronization
 *
 *  Configuration information for enabling Ethernet Rx & Tx Time sync events in a port
 */
typedef struct CpswMacPort_TsEventConfig_s
{
    /*! Enable Annex D(IPv4) Tx Time Synchronization */
    bool enableTxAnnexD;

    /*! Enable Annex D(IPv4) Rx Time Synchronization */
    bool enableRxAnnexD;

    /*! Enable Annex E(IPv6) Tx Time Synchronization */
    bool enableTxAnnexE;

   /*!  Enable Annex E(IPv6) Rx Time Synchronization */
    bool enableRxAnnexE;

    /*! Enable Annex F(IEEE802.3) Tx Time Synchronization */
    bool enableTxAnnexF;

    /*! Enable Annex F(IEEE802.3) Rx Time Synchronization*/
    bool enableRxAnnexF;

    /*! Time Sync Transmit Host Time Stamp Enable*/
    bool enableTxHostTs;

    /*! Tx VLAN Type */
    CpswMacPort_VlanType txVlanType;

    /*! Rx VLAN Type */
    CpswMacPort_VlanType rxVlanType;

    /*! VLAN Ltype 1 and 2 are common for both Tx and Rx and are valid only if
     *  #txVlanType or #rxVlanType is not #CPSW_MACPORT_VLAN_TYPE_NONE */
    uint32_t vlanLType1;

    /*! VLAN Ltype 1 and 2 are common for both Tx and Rx and are valid only if
     *  #txVlanType or #rxVlanType is not #CPSW_MACPORT_VLAN_TYPE_NONE */
    uint32_t vlanLType2;

    /*! Common Configuration for Annex D (IPv4) and Annex E (Ipv6) */
    CpswMacPort_IpTsConfig commonPortIpConfig;

    /*! Multicast Type value only for Annex E (IPv6) */
    uint32_t mcastType;

    /*! Message Type value mask is only for Annex F (PTP) */
    uint32_t messageType;

    /*! Ltype 2 is only for Annex F (PTP) */
    bool enableLType2;

    /*! Time Sync Sequence ID Offset */
    uint32_t seqIdOffset;

    /*! Time Sync Domain Offset */
    uint32_t domainOffset;
} CpswMacPort_TsEventConfig;

/*!
 *  \brief Enable Port Time Sync Event IOCTL inArgs
 */
typedef struct CpswMacPort_EnableTsEventInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Port number */
    CpswMacPort_TsEventConfig portCfg;
} CpswMacPort_EnableTsEventInArgs;

/*!
 * \brief MAC port submodule handle
 *
 * MAC port submodule opaque handle used to call any port related APIs on that
 * given MAC port.
 */
typedef struct CpswMacPort_Obj_s *CpswMacPort_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_MACPORT_H_ */

/* @} */
