/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_mdio.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the MDIO submodule.
 */

/*!
 * \addtogroup CPSW_MDIO_API
 * @{
 */

#ifndef CPSW_MDIO_H_
#define CPSW_MDIO_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Helper macro to create IOCTL commands for MDIO submodule */
#define CPSW_MDIO_PUBLIC_IOCTL(x)             (CPSW_IOCTL_MDIO_BASE | CPSW_IOCTL_MIN(x))

/*! \brief Create a MDIO PHY mask from a PHY address */
#define CPSW_MDIO_PHY_ADDR_MASK(addr)         (1U << (addr))

/*! \brief Check if the corresponding PHY address mask is set */
#define CPSW_MDIO_IS_PHY_ADDR_SET(mask, addr) (((mask) & (1U << (addr))) != 0U)

/*! \brief MDIO PHY address mask for no PHYs present */
#define CPSW_MDIO_PHY_ADDR_MASK_NONE          (0x00000000U)

/*! \brief MDIO PHY address mask for all PHYs present */
#define CPSW_MDIO_PHY_ADDR_MASK_ALL           (0xFFFFFFFFU)

/*! \brief Maximum number of PHYs supported on MDIO bus */
#define CPSW_MDIO_MAX_PHY_CNT                 (31U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief MDIO IOCTL commands
 */
enum CpswMdio_Ioctl_e
{
    /*!
     * \brief Get the hardware version of the MDIO submodule
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #CpswMdio_Version
     */
    CPSW_MDIO_IOCTL_GET_VERSION              = CPSW_MDIO_PUBLIC_IOCTL(0U),

    /*!
     * \brief Get PHY alive status
     *
     * IOCTL parameters:
     * -  inArgs: uint8_t
     * - outArgs: bool
     */
    CPSW_MDIO_IOCTL_GET_ALIVE_STATUS         = CPSW_MDIO_PUBLIC_IOCTL(1U),

    /*!
     * \brief Get PHY link status
     *
     * IOCTL parameters:
     * -  inArgs: uint8_t
     * - outArgs: bool
     */
    CPSW_MDIO_IOCTL_GET_LINK_STATUS          = CPSW_MDIO_PUBLIC_IOCTL(2U),

    /*!
     * \brief Read a PHY register using clause-22 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause22ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_MDIO_IOCTL_CLAUSE22_READ_REG        = CPSW_MDIO_PUBLIC_IOCTL(3U),

    /*!
     * \brief Write a PHY register using clause-22 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause22WriteRegInArgs
     * - outArgs: None
     */
    CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG       = CPSW_MDIO_PUBLIC_IOCTL(4U),

    /*!
     * \brief Read a PHY register using clause-45 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause45ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_MDIO_IOCTL_CLAUSE45_READ_REG        = CPSW_MDIO_PUBLIC_IOCTL(5U),

    /*!
     * \brief Write a PHY register using clause-45 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause45WriteRegInArgs
     * - outArgs: None
     */
    CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG       = CPSW_MDIO_PUBLIC_IOCTL(6U),

    /*!
     * \brief Read a PHY register asynchronously using clause-22 frame
     *
     * Reads a PHY register asynchronously using a clause-22 frame.  The
     * completion event is reported via CpswMdio_UserAccessCallback passed
     * to the MDIO submodule at open time.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause22ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_MDIO_IOCTL_CLAUSE22_READ_REG_ASYNC  = CPSW_MDIO_PUBLIC_IOCTL(7U),

    /*!
     * \brief Write a PHY register asynchronously using clause-22 frame
     *
     * Writes a PHY register asynchronously using a clause-22 frame.  The
     * completion event is reported via CpswMdio_UserAccessCallback passed
     * to the MDIO submodule at open time.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause22WriteRegInArgs
     * - outArgs: None
     */
    CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG_ASYNC = CPSW_MDIO_PUBLIC_IOCTL(8U),

    /*!
     * \brief Read a PHY register asynchronously using clause-45 frame
     *
     * Reads a PHY register asynchronously using a clause-45 frame.  The
     * completion event is reported via CpswMdio_UserAccessCallback passed
     * to the MDIO submodule at open time.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause45ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_MDIO_IOCTL_CLAUSE45_READ_REG_ASYNC  = CPSW_MDIO_PUBLIC_IOCTL(9U),

    /*!
     * \brief Write a PHY register asynchronously using clause-45 frame
     *
     * Writes a PHY register asynchronously using a clause-45 frame.  The
     * completion event is reported via CpswMdio_UserAccessCallback passed
     * to the MDIO submodule at open time.
     *
     * IOCTL parameters:
     * -  inArgs: #CpswMdio_Clause45WriteRegInArgs
     * - outArgs: None
     */
    CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG_ASYNC = CPSW_MDIO_PUBLIC_IOCTL(10U),

    /*!
     * \brief Get link state change poll enable mask
     *
     * Gets a mask of PHY addresses whose state change is being monitored
     * by MDIO (either in Normal Mode or State Change Mode).
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: uint32_t
     */
    CPSW_MDIO_IOCTL_GET_POLL_ENABLE_MASK     = CPSW_MDIO_PUBLIC_IOCTL(11U),

    /*!
     * \brief Print MDIO registers
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_MDIO_IOCTL_PRINT_REGS               = CPSW_MDIO_PUBLIC_IOCTL(12U),
};

/*!
 * \brief MDIO operating mode
 */
typedef enum CpswMdio_OperatingMode_e
{
    /*! Normal mode (used to monitor up to 2 PHY devices) */
    CPSW_MDIO_MODE_NORMAL       = 0U,

    /*! State change mode (used to detect changes on any PHY device) */
    CPSW_MDIO_MODE_STATE_CHANGE = 1U,

    /*! Manual mode (slow GPIO, not used for MDIO PHY operations) */
    CPSW_MDIO_MODE_MANUAL       = 2U
} CpswMdio_OperatingMode;

/*!
 * \brief MDIO user channels (also known as user groups)
 */
typedef enum CpswMdio_UserChannel_e
{
    /*! User channel (user group) 0 */
    CPSW_MDIO_USER_CHANNEL_0    = 0U,

    /*! User channel (user group) 1 */
    CPSW_MDIO_USER_CHANNEL_1    = 1U,

    /*! Last user chanenl */
    CPSW_MDIO_USER_CHANNEL_LAST = CPSW_MDIO_USER_CHANNEL_1,
} CpswMdio_UserChannel;

/*!
 * \brief Frame format
 */
typedef enum CpswMdio_FrameFormat_s
{
    /*! Clause 22 frame format */
    CPSW_MDIO_FRAME_FORMAT_CLAUSE22 = 0U,

    /*! Clause 45 frame format */
    CPSW_MDIO_FRAME_FORMAT_CLAUSE45 = 1U,
} CpswMdio_FrameFormat;

/*!
 * \brief Clause-45 MDIO Manageable Device (MMD) addresses
 */
typedef enum CpswMdio_Clause45Mmd_e
{
    /*! PMA/PMD */
    CPSW_MDIO_MMD_PMA_PMD  = 1U,

    /*! WIS */
    CPSW_MDIO_MMD_WIS      = 2U,

    /*! PCS */
    CPSW_MDIO_MMD_PCS      = 3U,

    /*! PHY XS */
    CPSW_MDIO_MMD_PHY_XS   = 4U,

    /*! DTE XS */
    CPSW_MDIO_MMD_DTE_XS   = 5U,

    /*! Vendor specific 1 */
    CPSW_MDIO_MMD_VENDOR_1 = 30U,

    /*! Vendor specific 2 */
    CPSW_MDIO_MMD_VENDOR_2 = 31U,
} CpswMdio_Clause45Mmd;

/*!
 * \brief MDIO PHY status
 *
 * PHY alive and linked status as detected by MDIO.
 */
typedef struct CpswMdio_PhyStatus_s
{
    /*! Bit mask of PHYs that have been detected as alive */
    uint32_t aliveMask;

    /*! Bit mask of PHYs that have been detected as linked */
    uint32_t linkedMask;
} CpswMdio_PhyStatus;

/*!
 * \brief Input args for #CPSW_MDIO_IOCTL_CLAUSE22_READ_REG command
 *
 * Clause-22 register read arguments.
 */
typedef struct CpswMdio_Clause22ReadRegInArgs_s
{
    /*! User channel to be used to perform the register access */
    CpswMdio_UserChannel userCh;

    /*! PHY address */
    uint8_t phyAddr;

    /*! Register address */
    uint16_t reg;
} CpswMdio_Clause22ReadRegInArgs;

/*!
 * \brief Input args for #CPSW_MDIO_IOCTL_CLAUSE45_READ_REG command
 *
 * Clause-45 register read arguments.
 */
typedef struct CpswMdio_Clause45ReadRegInArgs_s
{
    /*! User channel to be used to perform the register access */
    CpswMdio_UserChannel userCh;

    /*! PHY address */
    uint8_t phyAddr;

    /*! MMD */
    CpswMdio_Clause45Mmd mmd;

    /*! Register address */
    uint16_t reg;
} CpswMdio_Clause45ReadRegInArgs;

/*!
 * \brief Input args for #CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG command
 *
 * Clause-22 register write arguments.
 */
typedef struct CpswMdio_Clause22WriteRegInArgs_s
{
    /*! User channel to be used to perform the register access */
    CpswMdio_UserChannel userCh;

    /*! PHY address */
    uint8_t phyAddr;

    /*! Register address */
    uint16_t reg;

    /*! Value to be written */
    uint16_t val;
} CpswMdio_Clause22WriteRegInArgs;

/*!
 * \brief Input args for #CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG command
 *
 * Clause-45 register write arguments.
 */
typedef struct CpswMdio_Clause45WriteRegInArgs_s
{
    /*! User channel to be used to perform the register access */
    CpswMdio_UserChannel userCh;

    /*! PHY address */
    uint8_t phyAddr;

    /*! MMD */
    CpswMdio_Clause45Mmd mmd;

    /*! Register address */
    uint16_t reg;

    /*! Value to be written */
    uint16_t val;
} CpswMdio_Clause45WriteRegInArgs;

/*!
 * \brief MDIO hardware version
 *
 * Version of the MDIO hardware.
 */
typedef struct CpswMdio_Version_s
{
    /*! Major version */
    uint32_t major;

    /*! Minor version */
    uint32_t minor;

    /*! RTL version */
    uint32_t rtl;

    /*! Identification */
    uint32_t id;

    /*! Scheme identification */
    uint32_t scheme;

    /*! BU id */
    uint32_t bu;
} CpswMdio_Version;

/*!
 * \brief MDIO module configuration
 *
 * Configuration parameters for the MDIO module.
 */
typedef struct CpswMdio_Config_s
{
    /*! MDIO module operating mode */
    CpswMdio_OperatingMode mode;

    /*! MDIO bus clock (MDCLK) frequency in Hz */
    uint32_t mdioBusFreqHz;

    /*! Polling inter packet gap frequency in Hz */
    uint32_t phyStatePollFreqHz;

    /*! Indicates which PHY addresses have polling enabled.
     *  The CPSW_MDIO_PHY_ADDR_MASK macro should be used to create the mask.
     *  Only two PHYs can be monitored in Normal Mode, so only two bits must
     *  be set if operating in that mode. */
    uint32_t pollEnableMask;

    /*! Indicates which PHY addresses will use Clause-45 frame format.
     *  The CPSW_MDIO_PHY_ADDR_MASK macro should be used to create the mask */
    uint32_t clause45EnableMask;

    /*! User channel to use for MDIO PHY register access */
    CpswMdio_UserChannel userCh;
} CpswMdio_Config;

/*!
 * \brief MDIO submodule handle
 *
 * MDIO submodule opaque handle used to call any MDIO related APIs.
 */
typedef struct CpswMdio_Obj_s *CpswMdio_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_MDIO_H_ */

/* @} */
