/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_port.h
 *
 * \brief    This file contains the structure definitions common to the
 *           host and MAC ports.
 */

/*!
 * \addtogroup CPSW_MACPORT_API
 * @{
 */

#ifndef CPSW_PORT_H_
#define CPSW_PORT_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief CPSW packet priority */
#define CPSW_PACKET_PRIORITY_NUM              (8U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief Header Packet Priority
 */
typedef enum CpswPort_HeaderPacketPriority_e
{
    /*! Header Packet Priority 0 */
    CPSW_HEADER_PACKET_PRIORITY_0   = 0U,

    /*! Header Packet Priority 1 */
    CPSW_HEADER_PACKET_PRIORITY_1   = 1U,

    /*! Header Packet Priority 2 */
    CPSW_HEADER_PACKET_PRIORITY_2   = 2U,

    /*! Header Packet Priority 3 */
    CPSW_HEADER_PACKET_PRIORITY_3   = 3U,

    /*! Header Packet Priority 4 */
    CPSW_HEADER_PACKET_PRIORITY_4   = 4U,

    /*! Header Packet Priority 5 */
    CPSW_HEADER_PACKET_PRIORITY_5   = 5U,

    /*! Header Packet Priority 6 */
    CPSW_HEADER_PACKET_PRIORITY_6   = 6U,

    /*! Header Packet Priority 7 */
    CPSW_HEADER_PACKET_PRIORITY_7   = 7U,

    /*! Number of Header Packet Priority levels */
    CPSW_HEADER_PACKET_PRIORITY_NUM = 8U,
} CpswPort_HeaderPacketPriority;

/*!
 * \brief Switch Priority
 */
typedef enum CpswPort_SwitchPriority_e
{
    /* Switch Priority 0 */
    CPSW_SWITCH_PRIORITY_0   = 0U,

    /* Switch Priority 1 */
    CPSW_SWITCH_PRIORITY_1   = 1U,

    /* Switch Priority 2 */
    CPSW_SWITCH_PRIORITY_2   = 2U,

    /* Switch Priority 3 */
    CPSW_SWITCH_PRIORITY_3   = 3U,

    /* Switch Priority 4 */
    CPSW_SWITCH_PRIORITY_4   = 4U,

    /* Switch Priority 5 */
    CPSW_SWITCH_PRIORITY_5   = 5U,

    /* Switch Priority 6 */
    CPSW_SWITCH_PRIORITY_6   = 6U,

    /* Switch Priority 7 */
    CPSW_SWITCH_PRIORITY_7   = 7U,

    /* Number of Switch Priority levels */
    CPSW_SWITCH_PRIORITY_NUM = 8U,
} CpswPort_SwitchPriority;

/*!
 * \brief VLAN configuration parameters
 */
typedef struct CpswPort_VlanConfig_s
{
    /*! Port VLAN priority */
    uint32_t portPri;

    /*! Port CFI bit */
    uint32_t portCfi;

    /*! Port VLAN ID */
    uint32_t portVID;
} CpswPort_VlanConfig;

/*!
 * \brief Port priority type
 */
typedef enum CpswPort_TxPriorityType_e
{
    /*! Fixed priority */
    CPSW_TX_PRIORITY_TYPE_FIXED    = 0U,

    /*! Escalate priority */
    CPSW_TX_PRIORITY_TYPE_ESCALATE = 1U,
} CpswPort_TxPriorityType;

/*!
 * \brief Traffic shaping parameters (committed and excess rates)
 */
typedef struct CpswPort_TrafficShapingRates_s
{
    /*! Committed Information Rate (CIR) in bits-per-second */
    uint64_t committedRateBitsPerSec;

    /*! Excess Information Rate (EIR) in bits-per-second */
    uint64_t excessRateBitsPerSec;
} CpswPort_TrafficShapingRates;

/*!
 * \brief Traffic shaping configuration for all priorities
 */
typedef struct CpswPort_TrafficShapingConfig_s
{
    /*! Traffic shaping rates (committed and excess rates) for all priorities */
    CpswPort_TrafficShapingRates rates[CPSW_PRIORITY_NUM];
} CpswPort_TrafficShapingConfig;

/*!
 * \brief TX priority (header packet priority to switch priority)
 */
typedef struct CpswPort_TxPriority_s
{
    /*! Header packet priority to switch priority map */
    uint32_t priorityMap[CPSW_HEADER_PACKET_PRIORITY_NUM];
} CpswPort_TxPriority;

/*!
 * \brief RX priority (packet priority to header packet priority)
 */
typedef struct CpswPort_RxPriority_s
{
    /*! Packet priority to header packet priority map */
    uint32_t priorityMap[CPSW_PACKET_PRIORITY_NUM];
} CpswPort_RxPriority;

/*!
 * \brief DSCP priority
 */
typedef struct CpswPort_DscpPriority_s
{
    /*! DSCP IPv4/IPv6 to priority map */
    uint32_t tosMap[CPSW_TOS_PRIORITY_NUM];

    /*! Enable IPv4 based DSCP */
    bool enableDscpIPv4;

    /*! Enable IPv6 based DSCP */
    bool enableDscpIPv6;
} CpswPort_DscpPriority;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_PORT_H_ */

/* @} */
