/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_types.h
 *
 * \brief    This file contains the generic definitions and structure
 *           definitions of the CPSW driver.
 */

/*!
 * \addtogroup CPSW_MAIN_API
 * @{
 */

#ifndef CPSW_TYPES_H_
#define CPSW_TYPES_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/udma/udma.h>
#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! Macro to set bit at given bit position */
#define CPSW_SET_BIT(n)            (1U << (n))

/*! Macro to get bit at given bit position */
#define CPSW_GET_BIT(val, n)       (((val) & (1 << (n))) >> (n))

/*! Macro to check if bit at given bit position is set */
#define CPSW_IS_BIT_SET(val, n)            (((val) & (1U << (n))) != 0U)

/** \brief Field EXTract macro. We can't use CSL macro as it appends shift/mask with CSL_*/
#define CPSW_FEXT(reg, PER_REG_FIELD)                                        \
    (((reg) & PER_REG_FIELD##_MASK) >> PER_REG_FIELD##_SHIFT)

/** \brief Field INSert macro. We can't use CSL macro as it appends shift/mask with CSL_*/
#define CPSW_FINS(reg, PER_REG_FIELD, val)                                   \
    ((reg) = ((reg) & ~PER_REG_FIELD##_MASK)                                 \
      | (((val) << PER_REG_FIELD##_SHIFT) & PER_REG_FIELD##_MASK))

/*!
 * \anchor Cpsw_ErrorCodes
 * \name CPSW Error Codes
 *
 * Error codes returned by the CPSW driver APIs
 *
 * @{
 */

/* For simplicity, some CPSW error codes mirror the UDMA's */
/*! \brief Success */
#define CPSW_SOK                              (UDMA_SOK)

/*! \brief Generic failure error condition (typically caused by hardware) */
#define CPSW_EFAIL                            (UDMA_EFAIL)

/*! \brief Bad arguments (i.e. NULL pointer) */
#define CPSW_EBADARGS                         (UDMA_EBADARGS)

/*! \brief Invalid parameters (i.e. value out-of-range) */
#define CPSW_EINVALID_PARAMS                  (UDMA_EINVALID_PARAMS)

/*! \brief Time out while waiting for a given condition to happen */
#define CPSW_ETIMEOUT                         (UDMA_ETIMEOUT)

/*! \brief Allocation failure */
#define CPSW_EALLOC                           (UDMA_EALLOC)

/*! \brief Unexpected condition occurred (sometimes unrecoverable) */
#define CPSW_EUNEXPECTED                      (UDMA_EALLOC - 1)

/*! \brief The resource is currently busy performing an operation */
#define CPSW_EBUSY                            (UDMA_EALLOC - 2)

/*! \brief Malformed IOCTL (args pointer or size not as expected) */
#define CPSW_EMALFORMED_IOCTL                 (UDMA_EALLOC - 3)

/*! \brief Already open error */
#define CPSW_EALREADY_OPEN                    (UDMA_EALLOC - 4)

/*! \brief Operation not permitted */
#define CPSW_EPERM                            (UDMA_EALLOC - 5)

/*! \brief Operation not supported */
#define CPSW_ENOTSUPPORTED                    (UDMA_EALLOC - 6)

/*! \brief Unknown IOCTL */
#define CPSW_EUNKNOWN_IOCTL                   (UDMA_EALLOC - 7)

/*! \brief ALE Table full */
#define CPSW_EALE_TABLE_FULL                  (UDMA_EALLOC - 8)

/*! \brief ALE is in VLAN Unaware mode */
#define CPSW_EALE_VLAN_UNAWARE_MODE           (UDMA_EALLOC - 9)

/*! \brief Address not found in ALE entries */
#define CPSW_EALE_ADDRESS_NOT_FOUND           (UDMA_EALLOC - 10)

/*! \brief VLAN Entry not found */
#define CPSW_EALE_VLAN_NOT_FOUND              (UDMA_EALLOC - 11)

/*! \brief VLAN Mask Mux Entry not found */
#define CPSW_EALE_VLANMASK_NOT_FOUND          (UDMA_EALLOC - 12)

/*! \brief Policer entry table is full */
#define CPSW_EALE_POLICER_TABLE_FULL          (UDMA_EALLOC - 13)

/*! \brief Resource not found */
#define CPSW_ENOTFOUND                        (UDMA_EALLOC - 14)

/* @} */

/*! \brief IOCTL major number bit offset */
#define CPSW_IOCTL_TYPE_OFFSET                (24U)

/*! \brief IOCTL major number bit offset */
#define CPSW_IOCTL_MAJOR_OFFSET               (16U)

/*! \brief IOCTL minor number bit offset */
#define CPSW_IOCTL_MINOR_OFFSET               (0U)

/*! \brief Helper macro to get the IOCTL type (public/private) */
#define CPSW_IOCTL_GET_TYPE(x)                ((x) & 0xFF000000U)

/*! \brief Helper macro to get the IOCTL major number */
#define CPSW_IOCTL_GET_MAJ(x)                 ((x) & 0x00FF0000U)

/*! \brief Helper macro to get the IOCTL major number */
#define CPSW_IOCTL_GET_MIN(x)                 ((x) & 0x0000FFFFU)

/*! \brief Helper macro to set the IOCTL type */
#define CPSW_IOCTL_TYPE(x)                    ((x) << CPSW_IOCTL_TYPE_OFFSET)

/*! \brief Helper macro to set the IOCTL major number */
#define CPSW_IOCTL_MAJ(x)                     ((x) << CPSW_IOCTL_MAJOR_OFFSET)

/*! \brief Helper macro to set the IOCTL minor number */
#define CPSW_IOCTL_MIN(x)                     ((x) << CPSW_IOCTL_MINOR_OFFSET)

/*! \brief MAC address length (in bytes) */
#define CPSW_MAC_ADDR_LEN                     (6U)

/*! \brief CPSW priority levels */
#define CPSW_PRIORITY_NUM                     (8U)

/*! \brief Type of Service (ToS) priority */
#define CPSW_TOS_PRIORITY_NUM                 (64U)

/*! \brief Maximum value for VLAN ID */
#define CPSW_VLAN_ID_MAX                      (4095U)

/*! \brief 10-Mbps, half-duplex capability mask */
#define CPSW_LINK_CAP_HD10                    CPSW_SET_BIT(1)

/*! \brief 10-Mbps, full-duplex capability mask */
#define CPSW_LINK_CAP_FD10                    CPSW_SET_BIT(2)

/*! \brief 100-Mbps, half-duplex capability mask */
#define CPSW_LINK_CAP_HD100                   CPSW_SET_BIT(3)

/*! \brief 100-Mbps, full-duplex capability mask */
#define CPSW_LINK_CAP_FD100                   CPSW_SET_BIT(4)

/*! \brief 1-Gbps, half-duplex capability mask */
#define CPSW_LINK_CAP_HD1000                  CPSW_SET_BIT(5)

/*! \brief 1-Gbps, full-duplex capability mask */
#define CPSW_LINK_CAP_FD1000                  CPSW_SET_BIT(6)

/*! \brief 10-Mbps, full and half-duplex capability mask */
#define CPSW_LINK_CAP_10                      (CPSW_LINK_CAP_HD10 | \
                                               CPSW_LINK_CAP_FD10)

/*! \brief 100-Mbps, full and half-duplex capability mask */
#define CPSW_LINK_CAP_100                     (CPSW_LINK_CAP_HD100 | \
                                               CPSW_LINK_CAP_FD100)

/*! \brief 1-Gbps, full and half-duplex capability mask */
#define CPSW_LINK_CAP_1000                    (CPSW_LINK_CAP_HD1000 | \
                                               CPSW_LINK_CAP_FD1000)

/*! \brief Auto-negotiation mask with all duplexity and speed values set */
#define CPSW_LINK_CAP_ALL                     (CPSW_LINK_CAP_HD10 |   \
                                               CPSW_LINK_CAP_FD10 |   \
                                               CPSW_LINK_CAP_HD100 |  \
                                               CPSW_LINK_CAP_FD100 |  \
                                               CPSW_LINK_CAP_HD1000 | \
                                               CPSW_LINK_CAP_FD1000)

/*! \brief Ethertype value for PTP over Ethernet Annex F (IEEE 802.3)*/
#define CPSW_ETHERTYPE_PTP                    (0x88F7U)

/*! \brief Maximum number of client core that the CPSW driver can serve */
#define CPSW_REMOTE_CLIENT_CORES_MAX       (6U)

/*!
 * \brief CPSW virtual-to-physical address translation callback function.
 *
 * This function is used by the driver to convert virtual address to physical
 * address.
 *
 * \param virtAddr [IN] Virtual address
 * \param chNum    [IN] Channel number passed during channel open
 * \param appData  [IN] Callback pointer passed during channel open
 *
 * \return Corresponding physical address
 */
typedef uint64_t (*Cpsw_VirtToPhyFxn)(const void *virtAddr, void *appData);

/*!
 * \brief CPSW physical-to-virtual address translation callback function
 *
 * This function is used by the driver to convert physical address to virtual
 * address.
 *
 * \param phyAddr  [IN] Physical address
 * \param chNum    [IN] Channel number passed during channel open
 * \param appData  [IN] Callback pointer passed during channel open
 *
 * \return Corresponding virtual address
 */
typedef void *(*Cpsw_PhyToVirtFxn)(uint64_t phyAddr, void *appData);

/*!
 * \brief CPSW info/debug print function prototype
 *
 * This function is used by the driver to print info/debug messages.
 *
 * \param str      [OUT] Info string to print
 */
typedef void (*Cpsw_PrintFxnCb)(const char *str, ...);

/*!
 * \brief CPSW log trace function prototype
 *
 * This function is used by the driver to trace log messages.
 *
 * \param str      [OUT] Info string to print
 */
typedef void (*Cpsw_TraceFxnCb)(const char *str, ...);

/*!
 * \brief CPSW OSAL intr disable function prototype
 *
 * \return Cookie to be passed back to enable interrupt function
 */
typedef uintptr_t (*CpswOsal_DisableAllIntrFxn)(void);

/*!
 * \brief CPSW OSAL intr restore function prototype
 *
 * \param cookie       [IN] This is returned in disable interrupt function
 */
typedef void (*CpswOsal_RestoreAllIntrFxn)(uintptr_t cookie);

/*!
 * \brief CPSW OSAL intr restore function prototype
 *
 * \param coreIntrNum  [IN] Interrupt to restore
 */
typedef void (*CpswOsal_RestoreIntrFxn)(uint32_t coreIntrNum);

/*!
 * \brief CPSW OSAL intr disable function prototype
 *
 * \param coreIntrNum  [IN] Interrupt to disable
 */
typedef void (*CpswOsal_DisableIntrFxn)(uint32_t coreIntrNum);

/*!
 * \brief CPSW OSAL mutex create function prototype to protect critical section
 *
 * \return Pointer to mutex object
 */
typedef void *(*CpswOsal_MutexCreateFxn)(void);

/*!
 * \brief CPSW OSAL mutex delete function prototype
 *
 * \param mutexHandle  [IN] Pointer to mutex object returned during create
 */
typedef void (*CpswOsal_MutexDeleteFxn)(void *mutexHandle);

/*!
 * \brief CPSW OSAL mutex lock function prototype
 *
 * \param mutexHandle  [IN] Pointer to mutex object returned during create
 */
typedef void (*CpswOsal_MutexLockFxn)(void *mutexHandle);

/*!
 * \brief CPSW OSAL mutex lock function prototype
 *
 * \param mutexHandle  [IN] Pointer to mutex object returned during create
 */
typedef void (*CpswOsal_MutexUnlockFxn)(void *mutexHandle);

/*!
 * \brief CPSW OSAL ISR callback function prototype
 *
 * \param arg          [IN] App data
 */
typedef void (*CpswOsal_IsrFxn)(uintptr_t arg);

/*!
 * \brief CPSW OSAL ISR register function prototype
 *
 * \param isrFxn       [IN] ISR callback fxn pointer
 * \param coreIntrNum  [IN] Core interrupt number to register
 * \param intrPriority [IN] Priority
 * \param arg          [IN] Arg that will be passed back in the ISR
 *
 * \return Created HWI handle
 */
typedef void *(*CpswOsal_RegisterIntrFxn)(CpswOsal_IsrFxn isrFxn,
                                          uint32_t coreIntrNum,
                                          uint32_t intrPriority,
                                          void *arg);

/*!
 * \brief CPSW OSAL ISR unregister function prototype
 *
 * \param hwiHandle    [IN] HWI handle
 */
typedef void (*CpswOsal_UnRegisterIntrFxn)(void *hwiHandle);

/*!
 * \brief CPSW OSAL cache invalidate function prototype
 *
 * \param addr         [IN] Start address of the cache line/s
 * \param size         [IN] size (in bytes) of the memory to invalidate
 */
typedef void (*CpswOsal_CacheInv)(const void *addr, int32_t size);

/*!
 * \brief CPSW OSAL cache writeback function prototype
 *
 * \param addr         [IN] Start address of the cache line/s
 * \param size         [IN] size (in bytes) of the memory to be written back
 */
typedef void (*CpswOsal_CacheWb)(const void *addr, int32_t size);

/*!
 * \brief CPSW OSAL cache writeback invalidate function prototype
 *
 * \param addr         [IN] Start address of the cache line/s
 * \param size         [IN] size (in bytes) of the memory to be written back
 */
typedef void (*CpswOsal_CacheWbInv)(const void *addr, int32_t size);

/*!
 * \brief CPSW OSAL timer read function prototype
 *
 * \return Timer's read value
 */
typedef uint32_t (*CpswOsal_TimerReadFxn)(void);

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief IOCTL types
 */
enum Cpsw_IoctlType_e
{
    /*! Public IOCTL type (default type if omitted) */
    CPSW_IOCTL_TYPE_PUBLIC  = CPSW_IOCTL_TYPE(0U),

    /*! Private IOCTL type */
    CPSW_IOCTL_TYPE_PRIVATE = CPSW_IOCTL_TYPE(1U),
};

/*!
 * \brief IOCTL base number
 */
enum Cpsw_IoctlMajor_e
{
    /*! CPSW IOCTL base */
    CPSW_IOCTL_CPSW_BASE     = CPSW_IOCTL_MAJ(1U),

    /*! ALE submodule IOCTL base */
    CPSW_IOCTL_ALE_BASE      = CPSW_IOCTL_MAJ(2U),

    /*! CPTS submodule IOCTL base */
    CPSW_IOCTL_CPTS_BASE     = CPSW_IOCTL_MAJ(3U),

    /*! Host port submodule IOCTL base */
    CPSW_IOCTL_HOSTPORT_BASE = CPSW_IOCTL_MAJ(5U),

    /*! MAC port submodule IOCTL base */
    CPSW_IOCTL_MACPORT_BASE  = CPSW_IOCTL_MAJ(6U),

    /*! MDIO submodule IOCTL base */
    CPSW_IOCTL_MDIO_BASE     = CPSW_IOCTL_MAJ(7U),

    /*! SGMII submodule IOCTL base */
    CPSW_IOCTL_SGMII_BASE    = CPSW_IOCTL_MAJ(8U),

    /*! Statistics submodule IOCTL base */
    CPSW_IOCTL_STATS_BASE    = CPSW_IOCTL_MAJ(9U),

    /*! CPSW subsystem IOCTL base */
    CPSW_IOCTL_SS_BASE       = CPSW_IOCTL_MAJ(10U),

    /*! PHY IOCTL base */
    CPSW_IOCTL_PHY_BASE      = CPSW_IOCTL_MAJ(11U),

    /* RM submodule IOCTL base */
    CPSW_IOCTL_RM_BASE       = CPSW_IOCTL_MAJ(12U),
};

/*!
 * \brief CPSW instance type
 *
 * Type of CPSW instance.
 */
typedef enum Cpsw_Type_e
{
    /*! CPSW 2G: 1 host port, 1 MAC port */
    CPSW_2G,

    /*! CPSW 5G: 1 host port, 4 MAC ports */
    CPSW_5G,

    /*! CPSW 9G: 1 host port, 8 MAC ports */
    CPSW_9G,

    /*! CPSW last type - used for count of total number of types */
    CPSW_TYPE_LAST = CPSW_9G,
} Cpsw_Type;

/*! \brief Number of supported CPSW types */
#define CPSW_COUNT                            (CPSW_TYPE_LAST + 1U)

/*!
 * \brief CPSW MAC port
 */
typedef enum Cpsw_MacPort_e
{
    /*! MAC port 0 */
    CPSW_MAC_PORT_FIRST = 0U,

    /*! MAC port 0 */
    CPSW_MAC_PORT_0     = CPSW_MAC_PORT_FIRST,

    /*! MAC port 1 */
    CPSW_MAC_PORT_1     = 1U,

    /*! MAC port 2*/
    CPSW_MAC_PORT_2     = 2U,

    /*! MAC port 3 */
    CPSW_MAC_PORT_3     = 3U,

    /*! MAC port 4 */
    CPSW_MAC_PORT_4     = 4U,

    /*! MAC port 5 */
    CPSW_MAC_PORT_5     = 5U,

    /*! MAC port 6 */
    CPSW_MAC_PORT_6     = 6U,

    /*! MAC port 7 */
    CPSW_MAC_PORT_7     = 7U,

    /*! MAC last port - used for count of total number of ports. */
    CPSW_MAC_PORT_LAST  = CPSW_MAC_PORT_7,
} Cpsw_MacPort;

/*! Number of MAC ports*/
#define CPSW_MAC_PORT_NUM                 (CPSW_MAC_PORT_LAST + 1U)

/*! Macro to normalize #Cpsw_MacPort. It takes a #Cpsw_MacPort enum and converts
 *  it to a zero-based index */
#define CPSW_NORMALIZE_MACPORT(macPort)   ((macPort) - CPSW_MAC_PORT_FIRST)

/*! Macro to denormalize MAC port number. It takes a zero-based port number and converts
 *  it to a #Cpsw_MacPort enum  */
#define CPSW_DENORMALIZE_MACPORT(macPort) ((macPort) + CPSW_MAC_PORT_FIRST)

/*! MAC invalid port number. Used for error checks */
#define CPSW_MAC_PORT_INV                 ((Cpsw_MacPort)0xFFFFU)

/*!
 * \brief CRC type
 */
typedef enum Cpsw_CrcType_e
{
    /*! Ethernet CRC type */
    CPSW_CRC_ETHERNET   = 0U,

    /*! Castagnoli CRC type */
    CPSW_CRC_CASTAGNOLI = 1U,
} Cpsw_CrcType;

/*!
 * \brief MAC interface speed
 */
typedef enum Cpsw_Speed_e
{
    /*! 10 Mbps */
    CPSW_SPEED_10MBIT  = 0U,

    /*! 100 Mbps */
    CPSW_SPEED_100MBIT = 1U,

    /*! 1 Gbps */
    CPSW_SPEED_1GBIT   = 2U,

    /*! 2.5 Gbps */
    CPSW_SPEED_2p5GBIT = 3U,

    /*! 10 Gbps */
    CPSW_SPEED_10GBIT  = 4U,

    /*! Speed determined automatically */
    CPSW_SPEED_AUTO    = 5U,
} Cpsw_Speed;

/*!
 * \brief MAC interface duplexity
 */
typedef enum Cpsw_Duplexity_e
{
    /*! Half duplex */
    CPSW_DUPLEX_HALF = 0U,

    /*! Full duplex */
    CPSW_DUPLEX_FULL = 1U,

    /*! Duplexity determined automatically */
    CPSW_DUPLEX_AUTO = 2U,
} Cpsw_Duplexity;

/*!
 * \brief Double VLAN switch
 */
typedef enum Cpsw_VlanSwitch_e
{
    /*! Customer switch */
    CPSW_VLAN_SWITCH_CUSTOMER = 0U,

    /*! Service switch */
    CPSW_VLAN_SWITCH_SERVICE  = 1U
} Cpsw_VlanSwitch;

/*!
 * \brief CPSW Restart type
 *
 * CPSW supports reset isolation and full availability even when a core goes
 * down. This requires speacial handling in driver when restarting from such
 * scenarios. Based on type of restart different handling is required.
 * Enumerator identifies type of restart
 */
typedef enum
{
    /*! Restart after main domain reset */
    CPSW_RESTART_DOMAINRESET = 0U,

    /*! Restart after core reboot due to watchdog expiry or software initiated reboot */
    CPSW_RESTART_COREREBOOT  = 1U
} Cpsw_RestartType;

/*!
 * \brief CPSW hardware version
 *
 * Version of the CPSW hardware.
 */
typedef struct Cpsw_Version_s
{
    /*! Major version */
    uint32_t major;

    /*! Minor version */
    uint32_t minor;

    /*! RTL version */
    uint32_t rtl;

    /*! Identification */
    uint32_t id;
} Cpsw_Version;

/*!
 * \brief CPSW IOCTL param
 *
 * IOCTL params structure argument for all CPSW IOCTL commands
 */
typedef struct Cpsw_IoctlPrms_s
{
    /*! Pointer to an actual input args struct for the IOCTL cmd */
    const void *inArgs;

    /*! Size of #inArgs structure */
    uint32_t inArgsSize;

    /*! Pointer to an actual output args struct for the IOCTL cmd */
    void *outArgs;

    /*! Size of #outArgs structure */
    uint32_t outArgsSize;
} Cpsw_IoctlPrms;

/*!
 * \brief Output args for CPSW_*_IOCTL_GET_VERSION commands
 */
typedef Cpsw_Version *Cpsw_GetVersionOutArgs;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_TYPES_H_ */

/* @} */
