/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *   \file  ndk2cpsw.c
 *
 *   \brief
 *          Translation layer to go from NDK to CPSW hardware driver.
 *
 */

/*---------------------------------------------------------------------------*\
 |                                Header Files                                 |
 \*---------------------------------------------------------------------------*/

/* Standard language headers */
#include <stdio.h>
#include <assert.h>

/* xdc header - should be first file included as due to order dependency */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>

/* OS/Posix headers */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/utils/Load.h>
#include <ti/sysbios/hal/Cache.h>

/* NDK Dependencies */
#include <ti/ndk/inc/netmain.h>
#include <ti/ndk/inc/stkmain.h>

/* CPSW driver dependency headers */
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>

#include <ti/drv/uart/UART_stdio.h>

#include <ti/sysbios/interfaces/ITimer.h>

/* This module's header */
#include "ndk2cpsw.h"
#include "ndk2cpsw_appif.h"

/*---------------------------------------------------------------------------*\
 |                             Extern Declarations                             |
 \*---------------------------------------------------------------------------*/

extern PBMQ PBMQ_free;

/*---------------------------------------------------------------------------*\
 |                            Local Macros/Defines                             |
 \*---------------------------------------------------------------------------*/

#define NDK_RX_PACKET_TASK_PRI      (OS_TASKPRIHIGH)
#define NDK_RX_PACKET_TASK_STACK    (2048)

#define NDK_TX_PACKET_TASK_PRI      (OS_TASKPRIHIGH)
#define NDK_TX_PACKET_TASK_STACK    (2048)

#define NDK2CPSW_DIVIDER_US_TO_MS   (1000U)

/*---------------------------------------------------------------------------*\
 |                            Local Typedefs/Enums                             |
 \*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 |                         Local Function Declarations                         |
 \*---------------------------------------------------------------------------*/

static void Ndk2Cpsw_notifyRxPackets(void *hCbArg);

static void Ndk2Cpsw_notifyTxPackets(void *hCbArg);

static void Ndk2Cpsw_rxPacketTask(void *arg0);

static void Ndk2Cpsw_prepFreePktInfoQ(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_pbmQ2PktInfoQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                   PBMQ *PBMPktQ,
                                   CpswDma_PktInfoQ *pDmaPktInfoQ);

static void Ndk2Cpsw_pktInfoQ2PBMQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                   CpswDma_PktInfoQ *pDmaPktInfoQ,
                                   PBMQ *PBMPktQ);

static uint32_t Ndk2Cpsw_prepRxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                    CpswDma_PktInfoQ *pPktQ);

static void Ndk2Cpsw_submitRxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_txPacketTask(void *arg0);

static uint32_t Ndk2Cpsw_prepTxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                    CpswDma_PktInfoQ *pPktQ);

static void Ndk2Cpsw_freePBMPackets(CpswDma_PktInfoQ *tempQueue);

static void Ndk2Cpsw_updateRxNotifyStats(uint32_t packetCount,
                                         uint32_t timeDiff);

static void Ndk2Cpsw_updateTxNotifyStats(uint32_t packetCount,
                                         uint32_t timeDiff);

static void Ndk2Cpsw_print(const char *prnStr,
                           ...);

static int32_t Ndk2Cpsw_startRxTx(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_stopRxTx(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_submitTxPackets(Ndk2Cpsw_Handle hNdk2Cpsw,
                                     CpswDma_PktInfoQ *pSubmitQ);

static void Ndk2Cpsw_submitRxPackets(Ndk2Cpsw_Handle hNdk2Cpsw,
                                     CpswDma_PktInfoQ *pSubmitQ);

static void Ndk2Cpsw_retrieveTxPkts(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_timerCb(UArg arg);

static void Ndk2Cpsw_createTimer(Ndk2Cpsw_Handle hNdk2Cpsw);

static void Ndk2Cpsw_initGetHandleInArgs(Ndk2Cpsw_Handle hNdk2Cpsw,
                                         NimuCpswAppIf_GetHandleInArgs *inArgs);

static void Ndk2Cpsw_initReleaseHandleInArgs(Ndk2Cpsw_Handle hNdk2Cpsw,
                                             NimuCpswAppIf_ReleaseHandleInfo *inArgs);

/*---------------------------------------------------------------------------*\
 |                         Local Variable Declarations                         |
 \*---------------------------------------------------------------------------*/

static Ndk2Cpsw_Object gNdk2CpswObj;

/*---------------------------------------------------------------------------*\
 |                         Global Variable Declarations                        |
 \*---------------------------------------------------------------------------*/

Ndk2Cpsw_Stats gNdk2CpswStats;

/*---------------------------------------------------------------------------*\
 |                          Global Function Definitions                        |
 \*---------------------------------------------------------------------------*/

/**
 *  @b Ndk2Cpsw_open
 *  @n
 *      Opens and configures CPSW. Configures Interrupts, SGMII,
 *      and send and receive queues.
 *
 *  \param[in]  hNdk2Cpsw
 *      Ndk2Cpsw_object structure pointer.
 *
 *  \retval
 *      Success -  0
 *  \retval
 *      Error   -  Any non-zero value
 */
Ndk2Cpsw_Handle Ndk2Cpsw_open(STKEVENT_Handle hEvent)
{
    int32_t status;
    NimuCpswAppIf_GetHandleInArgs getHandleInArgs;
    Ndk2Cpsw_Handle hNdk2Cpsw;
    Semaphore_Params semaphoreParams;
    Task_Params params;
    Error_Block eb;
    uint32_t semInitCnt;

    hNdk2Cpsw = &gNdk2CpswObj;

    /* Clear instrumentation statistics structure */
    memset(&gNdk2CpswStats, 0, sizeof(Ndk2Cpsw_Stats));

    /* Initialize the allocated memory block. */
    memset(hNdk2Cpsw, 0, sizeof(Ndk2Cpsw_Object));

    /* Store total number of free PBM packets allocated. Used during close to
     * make sure all PBM packets are returned to NDK */
    hNdk2Cpsw->numAllocPBMPkts = PBMQ_count(&PBMQ_free);

    /* Initialize the Rx Queues & Data */
    PBMQ_init(&hNdk2Cpsw->rxReadyPBMPktQ);

    /* Initialize the Tx Queues & Data */
    PBMQ_init(&hNdk2Cpsw->txReadyPBMPktQ);
    PBMQ_init(&hNdk2Cpsw->txUnUsedPBMPktQ);

    /* Keep stack event handle to use to signal the NDK stack */
    hNdk2Cpsw->hEvent = hEvent;

    /* MCast List is EMPTY */
    hNdk2Cpsw->MCastCnt = 0;

    /* Init internal bookkeeping fields */
    hNdk2Cpsw->oldMCastCnt = 0;

    /* First init tasks, semaphores and clocks. This is required because
     * CpswDma event cb ISR can happen immediately after opening rx flow
     * in NimuCpswAppCb_getHandle and all semaphores, clocks and tasks should
     * be valid at that point
     */
    /* Create semaphore objects, init shutDownFlag status */
    hNdk2Cpsw->shutDownFlag = false;

    semInitCnt              = 0U;
    hNdk2Cpsw->hShutDownSem = Semaphore_create(semInitCnt, NULL, NULL);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hShutDownSem);

    Semaphore_Params_init(&semaphoreParams);
    semaphoreParams.mode    = Semaphore_Mode_BINARY;
    hNdk2Cpsw->hRxPacketSem = Semaphore_create(semInitCnt, &semaphoreParams, NULL);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hRxPacketSem);

    Semaphore_Params_init(&semaphoreParams);
    semaphoreParams.mode    = Semaphore_Mode_BINARY;
    hNdk2Cpsw->hTxPacketSem = Semaphore_create(semInitCnt, &semaphoreParams, NULL);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hTxPacketSem);

    /* Start the packet processing tasks now that the channels are open */
    Error_init(&eb);
    Task_Params_init(&params);
    params.instance->name = "Ndk2Cpsw_RxPacketTask";
    params.priority       = NDK_RX_PACKET_TASK_PRI;
    params.stackSize      = NDK_RX_PACKET_TASK_STACK;
    params.arg0           = (uintptr_t)hNdk2Cpsw;

    hNdk2Cpsw->hRxPacketTask = Task_create((Task_FuncPtr) & Ndk2Cpsw_rxPacketTask, &params, &eb);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hRxPacketTask);

    Error_init(&eb);
    Task_Params_init(&params);
    params.instance->name = "Ndk2Cpsw_txPacketTask";
    params.priority       = NDK_TX_PACKET_TASK_PRI;
    params.stackSize      = NDK_TX_PACKET_TASK_STACK;
    params.arg0           = (uintptr_t)hNdk2Cpsw;

    hNdk2Cpsw->hTxPacketTask = Task_create((Task_FuncPtr) & Ndk2Cpsw_txPacketTask, &params, &eb);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hTxPacketTask);

    /* Get CPSW & UDMA Drv Handle */
    Ndk2Cpsw_initGetHandleInArgs(hNdk2Cpsw, &getHandleInArgs);
    NimuCpswAppCb_getHandle(&getHandleInArgs,
                            &hNdk2Cpsw->appInfo);
    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.hCpsw != NULL);
    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.hUdmaDrv != NULL);
    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.isPortLinkedFxn != NULL);
    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.rxInfo.hRxFlow != NULL);
    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.txInfo.hTxChannel != NULL);

    if (NULL != hNdk2Cpsw->appInfo.printFxnCb)
    {
        /* set the print function callback if not null */
        hNdk2Cpsw->printFxnCb = hNdk2Cpsw->appInfo.printFxnCb;
    }
    else
    {
        hNdk2Cpsw->printFxnCb = (Cpsw_PrintFxnCb) & System_printf;
    }

    /* Open CPSW DMA driver */
    status = Ndk2Cpsw_startRxTx(hNdk2Cpsw);
    if (status != CPSW_SOK)
    {
        Ndk2Cpsw_print("Failed to open CPSW DMA: %d\n", status);
    }

    /* Get initial link/interface status from the driver */
    hNdk2Cpsw->linkIsUp = hNdk2Cpsw->appInfo.isPortLinkedFxn(hNdk2Cpsw->appInfo.hCpsw);

    if (hNdk2Cpsw->appInfo.disableTxEvent)
    {
        CpswDma_disableTxEvent(hNdk2Cpsw->appInfo.txInfo.hTxChannel);
    }

    if (false == hNdk2Cpsw->appInfo.isRingMonUsed)
    {
        CpswDma_disableRxEvent(hNdk2Cpsw->appInfo.rxInfo.hRxFlow);
    }

    /* assert if clk period is not valid  */
    Ndk2Cpsw_assert(0U != hNdk2Cpsw->appInfo.timerPeriodUs);
    Ndk2Cpsw_createTimer(hNdk2Cpsw);
    Timer_start(hNdk2Cpsw->hPacingTimer);

    hNdk2Cpsw->initDone = TRUE;

    return hNdk2Cpsw;
}

/*!
 *  @b Ndk2Cpsw_close
 *  @n
 *      Closes CPSW and disables interrupts.
 *
 *  \param[in]  hNdk2Cpsw
 *      Ndk2Cpsw_object structure pointer.
 *
 *  \retval
 *      void
 */
void Ndk2Cpsw_close(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    uint32_t pktInfoQCnt;
    NimuCpswAppIf_ReleaseHandleInfo releaseHandleInfo;

    Ndk2Cpsw_assert(NULL != hNdk2Cpsw);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hRxPacketSem);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hTxPacketSem);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hShutDownSem);

    /* Set the translation layer shutdown flag */
    hNdk2Cpsw->shutDownFlag = true;

    /* Stop and delete the tick timer */
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->hPacingTimer);
    Timer_stop (hNdk2Cpsw->hPacingTimer);
    Timer_delete ((Timer_Handle *)&hNdk2Cpsw->hPacingTimer);

    Ndk2Cpsw_stopRxTx(hNdk2Cpsw);

    Ndk2Cpsw_initReleaseHandleInArgs(hNdk2Cpsw, &releaseHandleInfo);
    NimuCpswAppCb_releaseHandle(&releaseHandleInfo);

    /* free txReadyPBMPktQ */
    while (PBMQ_count(&hNdk2Cpsw->txReadyPBMPktQ) != 0U)
    {
        PBM_Handle hPBMPacket = PBMQ_deq(&hNdk2Cpsw->txReadyPBMPktQ);
        PBM_free(hPBMPacket);
        Ndk2CpswStas_addOne(&gNdk2CpswStats.rxPBMFreeCnt);
    }

    pktInfoQCnt = CpswUtils_getQCount(&hNdk2Cpsw->txFreePktInfoQ) +
                  CpswUtils_getQCount(&hNdk2Cpsw->rxFreePktInfoQ);

    if (CPSW_UTILS_ARRAYSIZE(hNdk2Cpsw->pktInfoMem) != pktInfoQCnt)
    {
        Ndk2Cpsw_print("Ndk2Cpsw_stopRxTx: Failed to retrieve all PktInfo\n");
    }

#if !defined(NDK_MULTIPLE_STACKS)
    /* debug point */

    /* We can't enable below by default as below check will fail if there are
     * two drivers registered with NDK which share PBM buffers */

    /* Check for PBM packets allocated during init are freed correctly and packet
     * count before nimu driver open matches after close. */
    if (hNdk2Cpsw->numAllocPBMPkts != PBMQ_count(&PBMQ_free))
    {
        Ndk2Cpsw_print("Ndk2Cpsw_stopRxTx: Failed to retrieve all PBM packets\n");
    }
#endif

    memset((void *)&hNdk2Cpsw->appInfo, 0, sizeof(hNdk2Cpsw->appInfo));

    /* Pend on hShutDownSem (twice for two sub-tasks) */
    Semaphore_pend(hNdk2Cpsw->hShutDownSem, BIOS_WAIT_FOREVER);
    Semaphore_pend(hNdk2Cpsw->hShutDownSem, BIOS_WAIT_FOREVER);

    /* Delete the semaphore objects */
    Semaphore_delete(&hNdk2Cpsw->hRxPacketSem);
    Semaphore_delete(&hNdk2Cpsw->hTxPacketSem);
    Semaphore_delete(&hNdk2Cpsw->hShutDownSem);

    /*
     * We don't do task deletion since the NDK forces the
     * Task.deleteTerminatedTasks option be true, which means
     * the hRxPacketTask and hTxPacketTask are auto-deleted.
     */

    /* Clear the allocated translation */
    memset(hNdk2Cpsw, 0, sizeof(Ndk2Cpsw_Object));
}

/*!
 *  @b Ndk2Cpsw_setRx
 *  @n
 *      Sets the filter for CPSW. Sets up the multicast addresses in
 *      the ALE.
 *
 *  \param[in]  hNdk2Cpsw
 *      Ndk2Cpsw_object structure pointer.
 *
 *  \retval
 *      void
 */
void Ndk2Cpsw_setRx(Ndk2Cpsw_Handle hNdk2Cpsw)
{
}

/*!
 *  @b Ndk2Cpsw_sendTxPackets
 *  @n
 *      Routine to send out queued Tx packets to the hardware driver
 *
 *  \param[in]  hNdk2Cpsw
 *      Ndk2Cpsw_object structure pointer.
 *
 *  \retval
 *      void
 */
void Ndk2Cpsw_sendTxPackets(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    CpswDma_PktInfo *pCurrDmaPacket;
    PBM_Handle hPBMPacket;

    /* If link is not up, simply return */
    CpswDma_PktInfoQ txSubmitQ;
    CpswUtils_initQ(&txSubmitQ);
    if (PBMQ_count(&hNdk2Cpsw->txUnUsedPBMPktQ))
    {
        /* send any pending TX Q's */
        Ndk2Cpsw_pbmQ2PktInfoQ(hNdk2Cpsw, &hNdk2Cpsw->txUnUsedPBMPktQ, &txSubmitQ);
    }

    /* Check if there is anything to transmit, else simply return */
    while (PBMQ_count(&hNdk2Cpsw->txReadyPBMPktQ) != 0)
    {
        /* Dequeue one free TX Eth packet */
        pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->txFreePktInfoQ);

        if (pCurrDmaPacket == NULL)
        {
            /* If we run out of packet info Q, retrieve packets from HW
             * and try to dequeue free packet again */
            Ndk2Cpsw_retrieveTxPkts(hNdk2Cpsw);
            pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->txFreePktInfoQ);
        }

        if (NULL != pCurrDmaPacket)
        {
            hPBMPacket = PBMQ_deq(&hNdk2Cpsw->txReadyPBMPktQ);
            CpswDma_pktInfoInit(pCurrDmaPacket);

            pCurrDmaPacket->bufPtr = PBM_getDataBuffer(hPBMPacket) +
                                     PBM_getDataOffset(hPBMPacket);
            pCurrDmaPacket->appPriv    = hPBMPacket;
            pCurrDmaPacket->orgBufLen  = PBM_getBufferLen(hPBMPacket);
            pCurrDmaPacket->userBufLen = PBM_getValidLen(hPBMPacket);
            pCurrDmaPacket->node.pNext = NULL;

#ifdef NDK_CHECKSUM_SUPPORT
            PBM_Pkt *pbmPkt = (PBM_Pkt *)hPBMPacket;
            CPSWDMA_CPPIPSI_SET_CHKSUM_RES (pCurrDmaPacket->chkSumInfo, (pbmPkt->csInsertPos));
            CPSWDMA_CPPIPSI_SET_CHKSUM_STARTBYTE (pCurrDmaPacket->chkSumInfo, (pbmPkt->csStartPos));
            CPSWDMA_CPPIPSI_SET_CHKSUM_INV_FLAG (pCurrDmaPacket->chkSumInfo, true);
            CPSWDMA_CPPIPSI_SET_CHKSUM_BYTECNT (pCurrDmaPacket->chkSumInfo, (pbmPkt->csNumBytes));
#endif
            CPSW_UTILS_COMPILETIME_ASSERT(offsetof(CpswDma_PktInfo, node) == 0);
            CpswUtils_enQ(&txSubmitQ, &(pCurrDmaPacket->node));

            Ndk2CpswStas_addOne(&gNdk2CpswStats.txFreeAppPktDq);
            Ndk2CpswStas_addOne(&gNdk2CpswStats.txReadyPBMPktDq);
        }
        else
        {
            break;
        }
    }

    /* Give the accumulated packets to the hardware */
    /* Submit the CPSW packet for transmission */
    Ndk2Cpsw_submitTxPackets(hNdk2Cpsw, &txSubmitQ);
}

/*!
 *  @b Ndk2Cpsw_ioctl
 *  @n
 *  Low level driver Ioctl interface. This interface can be used for
 *  ALE configuration, control, statistics
 *
 *  \param[in]  hNdk2Cpsw
 *      Ndk2Cpsw_object structure pointer.
 *  \param[in]  cmd
 *      Ioctl command.
 *  \param[in]  pBuf
 *      Ioctl buffer with commands and params to set/get
 *      configuration from hardware.
 *  \param[in]  size
 *      Size of Ioctl buffer.
 *
 *  \retval
 *      void
 */
int32_t Ndk2Cpsw_ioctl(Ndk2Cpsw_Handle hNdk2Cpsw,
                       uint32_t cmd,
                       void *param,
                       uint32_t size)
{
    int32_t retVal = 0U;

    // TODO - make NIMU IOCTL design consistent with CPSW/CPSW modules

    /* Decode the command and act on it */
    switch (cmd)
    {
        /* Rx Task Load IOCTL */
        case NDK2CPSW_IOCTL_GET_RXTASK_LOAD:
        {
            Load_Stat stat;
            if (Load_getTaskLoad(hNdk2Cpsw->hRxPacketTask, &stat))
            {
                retVal = Load_calculateLoad(&stat);
            }

            break;
        }

        /* Adding Tx Task Load  */
        case NDK2CPSW_IOCTL_GET_TXTASK_LOAD:
        {
            Load_Stat stat;
            if (Load_getTaskLoad(hNdk2Cpsw->hTxPacketTask, &stat))
            {
                retVal = Load_calculateLoad(&stat);
            }

            break;
        }

        default:
        {
            /* TODO . To pass through IOCTL commands to Cpsw, we need to know
             *  outArgs type and pass the correct size.
             *  This is not possible with the current NIMU interface
             */
            Ndk2Cpsw_assert(FALSE);
        }
    }

    return retVal;
}

/*---------------------------------------------------------------------------*\
 |                           Local Function Definitions                        |
 \*---------------------------------------------------------------------------*/

void Ndk2Cpsw_periodicFxn(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    uint32_t prevLinkState     = hNdk2Cpsw->linkIsUp;
    uint32_t prevLinkInterface = hNdk2Cpsw->currLinkedIf;

#if (1U == CPSW_CFG_DEV_ERROR)
    int32_t status;

    CpswUtils_verifyQCount(&hNdk2Cpsw->txFreePktInfoQ);
    CpswUtils_verifyQCount(&hNdk2Cpsw->rxFreePktInfoQ);

    status = CpswDma_checkRxFlowSanity(hNdk2Cpsw->appInfo.rxInfo.hRxFlow, 5U);
    if (status != CPSW_SOK)
    {
        Ndk2Cpsw_print("CpswDma_checkRxFlowSanity Failed\n");
    }

    status = CpswDma_checkTxChSanity(hNdk2Cpsw->appInfo.txInfo.hTxChannel, 5U);
    if (status != CPSW_SOK)
    {
        Ndk2Cpsw_print("CpswDma_checkTxChSanity Failed\n");
    }
#endif

    /*
     * Return the same DMA packets back to the DMA channel (but now
     * associated with a new PBM Packet and buffer)
     */
    if ( (CpswUtils_getQCount(&hNdk2Cpsw->rxFreePktInfoQ) != 0U) &&
         (PBMQ_count(&PBMQ_free) != 0U) )
    {
        Ndk2Cpsw_submitRxPktQ(hNdk2Cpsw);
    }

#if defined(NIMU_INSTRUMENTATION_ENABLED)
    static uint32_t loadCount = 0;
    Load_Stat stat;
    gNdk2CpswStats.gCpuLoad[loadCount] = Load_getCPULoad();

    if (Load_getGlobalHwiLoad(&stat))
    {
        gNdk2CpswStats.gHwiLoad[loadCount] = Load_calculateLoad(&stat);
    }

    if (Load_getTaskLoad(hNdk2Cpsw->hRxPacketTask, &stat))
    {
        gNdk2CpswStats.rxStats.taskLoad[loadCount] = Load_calculateLoad(&stat);
    }

    if (Load_getTaskLoad(hNdk2Cpsw->hTxPacketTask, &stat))
    {
        gNdk2CpswStats.txStats.taskLoad[loadCount] = Load_calculateLoad(&stat);
    }

    loadCount = (loadCount + 1U) & (HISTORY_CNT - 1U);
#endif

    /* Get current link status as reported by the hardware driver */
    hNdk2Cpsw->linkIsUp = hNdk2Cpsw->appInfo.isPortLinkedFxn(hNdk2Cpsw->appInfo.hCpsw);

    /* If the interface changed, discard any queue packets (since the MAC would now be wrong) */
    if (prevLinkInterface != hNdk2Cpsw->currLinkedIf)
    {
        /* ToDo: Discard all queued packets */
    }

    /* If link status changed from down->up, then send any queued packets */
    if ((prevLinkState == 0) && (hNdk2Cpsw->linkIsUp))
    {
        Ndk2Cpsw_sendTxPackets(hNdk2Cpsw);
    }
}

static void Ndk2Cpsw_processRxUnusedQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                      CpswDma_PktInfoQ *unUsedQ)
{
    CpswDma_PktInfo *pDmaPacket;

    pDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(unUsedQ);
    while (pDmaPacket != NULL)
    {
        /* Get the full PBM packet that needs to be freed */
        PBM_Handle hPBMPacket = (PBM_Handle)pDmaPacket->appPriv;
        if (hPBMPacket)
        {
            /* Free the PBM */
            PBM_free(hPBMPacket);
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxPBMFreeCnt);

            /* Put packet info into free Q as we have removed the PBM buffers
             * from the it */
            CpswUtils_enQ(&hNdk2Cpsw->rxFreePktInfoQ, &pDmaPacket->node);
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxFreeAppPktEnq);

            pDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(unUsedQ);
        }
        else
        {
            /* should never happen as this is received from HW */
            Ndk2Cpsw_assert(FALSE);
        }
    }
}

static void Ndk2Cpsw_submitRxPackets(Ndk2Cpsw_Handle hNdk2Cpsw,
                                     CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitRxPackets(hNdk2Cpsw->appInfo.rxInfo.hRxFlow, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        Ndk2Cpsw_print("CpswDma_submitRxPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        /* Copy unused packets to back to readyQ */
        Ndk2Cpsw_processRxUnusedQ(hNdk2Cpsw, &unUsedQ);
    }
}

static void Ndk2Cpsw_pbmQ2PktInfoQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                   PBMQ *PBMPktQ,
                                   CpswDma_PktInfoQ *pDmaPktInfoQ)
{
    CpswDma_PktInfo *pCurrDmaPacket;
    PBM_Handle hPBMPacket;

    while (PBMQ_count(PBMPktQ) != 0U)
    {
        /* Dequeue one free TX Eth packet */
        pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->txFreePktInfoQ);

        if (pCurrDmaPacket == NULL)
        {
            /* If we run out of packet info Q, retrieve packets from HW
             * and try to dequeue free packet again */
            Ndk2Cpsw_retrieveTxPkts(hNdk2Cpsw);
            pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->txFreePktInfoQ);
        }

        if (NULL != pCurrDmaPacket)
        {
            hPBMPacket = PBMQ_deq(PBMPktQ);
            CpswDma_pktInfoInit(pCurrDmaPacket);

            pCurrDmaPacket->bufPtr = PBM_getDataBuffer(hPBMPacket) +
                                     PBM_getDataOffset(hPBMPacket);
            pCurrDmaPacket->appPriv    = hPBMPacket;
            pCurrDmaPacket->orgBufLen  = PBM_getBufferLen(hPBMPacket);
            pCurrDmaPacket->userBufLen = PBM_getValidLen(hPBMPacket);
            pCurrDmaPacket->node.pNext = NULL;

            CPSW_UTILS_COMPILETIME_ASSERT(offsetof(CpswDma_PktInfo, node) == 0);
            CpswUtils_enQ(pDmaPktInfoQ, &(pCurrDmaPacket->node));

            Ndk2CpswStas_addOne(&gNdk2CpswStats.txFreeAppPktDq);
        }
        else
        {
            break;
        }
    }
}

static void Ndk2Cpsw_pktInfoQ2PBMQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                   CpswDma_PktInfoQ *pDmaPktInfoQ,
                                   PBMQ *PBMPktQ)
{
    CpswDma_PktInfo *pDmaPacket;
    PBM_Handle hPBMPacket;

    while (CpswUtils_getQCount(pDmaPktInfoQ) != 0U)
    {
        pDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(pDmaPktInfoQ);
        hPBMPacket = (PBM_Handle)(pDmaPacket->appPriv);

        PBMQ_enq(PBMPktQ, hPBMPacket);
    }
}

static void Ndk2Cpsw_submitTxPackets(Ndk2Cpsw_Handle hNdk2Cpsw,
                                     CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitTxReadyPackets(hNdk2Cpsw->appInfo.txInfo.hTxChannel, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        Ndk2Cpsw_print("CpswDma_submitTxReadyPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        /* TODO: txUnUsedPBMPktQ can be removed once PBMQ_enqHead
         *       function in NDK (https://jira.itg.ti.com/browse/NDK-379)
         *       is implemented.
         *       PBMQ_enqHead(&hNdk2Cpsw->txReadyPBMPktQ, hPBMPacket);
         */
        Ndk2Cpsw_pktInfoQ2PBMQ(hNdk2Cpsw, &unUsedQ, &hNdk2Cpsw->txUnUsedPBMPktQ);
        CpswUtils_appendQ(&hNdk2Cpsw->txFreePktInfoQ, &unUsedQ);
        Ndk2CpswStats_addNum(&gNdk2CpswStats.txFreeAppPktEnq, CpswUtils_getQCount(&unUsedQ));
    }
}

static void Ndk2Cpsw_freePBMPackets(CpswDma_PktInfoQ *tempQueue)
{
    CpswDma_PktInfo *pCurrDmaPacket;
    PBM_Handle hPBMPacket;

    while (CpswUtils_getQCount(tempQueue))
    {
        pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(tempQueue);
        hPBMPacket     = pCurrDmaPacket->appPriv;
        PBM_free(hPBMPacket);
    }
}

static void Ndk2Cpsw_notifyRxPackets(void *hCbArg)
{
    /* Post semaphore to rx handling task */
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)hCbArg;

    /* do not post events if init not done or shutdown in progress */
    if ((hNdk2Cpsw->initDone) && (hNdk2Cpsw->shutDownFlag == false))
    {
        CpswDma_disableRxEvent(hNdk2Cpsw->appInfo.rxInfo.hRxFlow);
        Semaphore_post(hNdk2Cpsw->hRxPacketSem);
    }
}

static void Ndk2Cpsw_notifyTxPackets(void *hCbArg)
{
    /* Post semaphore to tx handling task */
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)hCbArg;

    /* do not post events if init not done or shutdown in progress */
    if ((hNdk2Cpsw->initDone) && (hNdk2Cpsw->shutDownFlag == false))
    {
        Semaphore_post(hNdk2Cpsw->hTxPacketSem);
    }
}

static void Ndk2Cpsw_rxPacketTask(void *arg0)
{
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)arg0;
    CpswDma_PktInfoQ tempQueue;
    int32_t retVal;
    uint32_t pktCnt;

    while (!hNdk2Cpsw->shutDownFlag)
    {
        /* Wait for the Rx ISR to notify us that packets are available with data */
        Semaphore_pend(hNdk2Cpsw->hRxPacketSem, BIOS_WAIT_FOREVER);

        if (hNdk2Cpsw->shutDownFlag)
        {
            /* This translation layer is shutting down, don't give anything else to the stack */
            break;
        }

        Ndk2CpswStas_addOne(&gNdk2CpswStats.rxStats.rawNotifyCnt);
        pktCnt = 0;

        /* Retrieve the used (filled) packets from the channel */
        {
            CpswUtils_initQ(&tempQueue);
            retVal = CpswDma_retrieveRxPackets(hNdk2Cpsw->appInfo.rxInfo.hRxFlow,
                                               &tempQueue);
            if (CPSW_SOK != retVal)
            {
                Ndk2Cpsw_print("Ndk2Cpsw_rxPacketTask: failed to retrieve RX pkts: %d\n", retVal);
            }
        }
        if (tempQueue.count == 0)
        {
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxStats.zeroNotifyCnt);
        }

        /*
         * Call Ndk2Cpsw_prepRxPktQ() even if no packets were received.
         * This allows new packets to be submitted if PBM buffers became
         * newly available and there were outstanding free packets.
         */
        {
            /*
             * Get all used Rx DMA packets from the hardware, then send the buffers
             * of those packets on to the NDK stack to be parsed/processed.
             */
            pktCnt = Ndk2Cpsw_prepRxPktQ(hNdk2Cpsw, &tempQueue);
        }

        /*
         * We don't want to time the semaphore post used to notify the NDK stack as that may cause a
         * task transition. We don't want to time the semaphore pend, since that would time us doing
         * nothing but waiting.
         */
        if (pktCnt != 0)
        {
            Ndk2Cpsw_updateRxNotifyStats(pktCnt, 0U);
        }

        /* Notify NDK stack that packets are available (this is basically a Semaphore_post) */
        if (PBMQ_count(&hNdk2Cpsw->rxReadyPBMPktQ) != 0U)
        {
            ((STKEVENT *)(hNdk2Cpsw->hEvent))->EventCodes[STKEVENT_ETHERNET] = 1;
            if (((STKEVENT *)(hNdk2Cpsw->hEvent))->hSemEvent != NULL)
            {
                SemPost(((STKEVENT *)(hNdk2Cpsw->hEvent))->hSemEvent);
            }

            Ndk2CpswStas_addOne(&gNdk2CpswStats.ndkStackNotifyCnt);
        }

        /* As we are operating in one shot mode, the Timer_start should be
         * called for the next timeout period
         */
        Timer_start(hNdk2Cpsw->hPacingTimer);
    }

    /* We are shutting down, notify that we are done */
    Semaphore_post(hNdk2Cpsw->hShutDownSem);
}

static void Ndk2Cpsw_txPacketTask(void *arg0)
{
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)arg0;

    while (!hNdk2Cpsw->shutDownFlag)
    {
        /*
         * Wait for the Tx ISR to notify us that empty packets are available
         * that were used to send data
         */
        Semaphore_pend(hNdk2Cpsw->hTxPacketSem, SEM_FOREVER);
        Ndk2Cpsw_retrieveTxPkts(hNdk2Cpsw);
    }

    /* We are shutting down, notify that we are done */
    Semaphore_post(hNdk2Cpsw->hShutDownSem);
}

static void Ndk2Cpsw_prepFreePktInfoQ(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    uint32_t i;
    CpswDma_PktInfo *pPktInfo;

    /* Initialize the DMA Free Queue*/
    CpswUtils_initQ(&hNdk2Cpsw->txFreePktInfoQ);
    CpswUtils_initQ(&hNdk2Cpsw->rxFreePktInfoQ);

    /*
     * Allocate the free pkt info Q. This is used to exchange buffer info from
     * DMA basically for submitting free packets and retrieving ready packets.
     */
    for (i = 0U; i < (uint32_t)(NDK2CPSW_TX_PACKETS); i++)
    {
        /* Initialize Pkt info Q from allocated memory */
        pPktInfo = &hNdk2Cpsw->pktInfoMem[i];
        Ndk2Cpsw_assert(pPktInfo != NULL);
        CpswDma_pktInfoInit(pPktInfo);
        CpswUtils_enQ(&hNdk2Cpsw->txFreePktInfoQ, &pPktInfo->node);
        Ndk2CpswStas_addOne(&gNdk2CpswStats.txFreeAppPktEnq);
    }

    for (i = NDK2CPSW_TX_PACKETS; i < (uint32_t)(NDK2CPSW_TX_PACKETS + NDK2CPSW_RX_PACKETS); i++)
    {
        /* Initialize Pkt info Q from allocated memory */
        pPktInfo = &hNdk2Cpsw->pktInfoMem[i];
        Ndk2Cpsw_assert(pPktInfo != NULL);
        CpswDma_pktInfoInit(pPktInfo);
        CpswUtils_enQ(&hNdk2Cpsw->rxFreePktInfoQ, &pPktInfo->node);
        Ndk2CpswStas_addOne(&gNdk2CpswStats.rxFreeAppPktEnq);
    }
}

/*
 * Enqueue a new packet and make sure that buffer descriptors are properly linked.
 * NOTE: Not thread safe
 */
static void Ndk2Cpsw_submitRxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    CpswDma_PktInfoQ resubmitPktQ;
    PBM_Handle hPBMPacket;
    CpswDma_PktInfo *pCurrDmaPacket;
    uint32_t bufSize;

    CpswUtils_initQ(&resubmitPktQ);

    /*
     * Fill in as many packets as we can with new PBM buffers so they can be
     * returned to the stack to be filled in.
     */
    pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->rxFreePktInfoQ);

    while (NULL != pCurrDmaPacket)
    {
        bufSize = NIMU_CPSW_GET_PBM_BUFSIZE(hNdk2Cpsw->appInfo.hostPortRxMtu);
        bufSize = CPSW_UTILS_ALIGN(bufSize, CPSW_UTILS_CACHELINE_SIZE);
        hPBMPacket = PBM_alloc(bufSize);

        if (hPBMPacket)
        {
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxPBMAllocCnt);
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxFreeAppPktDq);

            /* The PBM packet has a data offset */
            PBM_setDataOffset(hPBMPacket, PKT_PREPAD);
            CpswDma_pktInfoInit(pCurrDmaPacket);

            pCurrDmaPacket->bufPtr    = PBM_getDataBuffer(hPBMPacket) + PKT_PREPAD;
            pCurrDmaPacket->orgBufLen = PBM_getBufferLen(hPBMPacket) - PKT_PREPAD;
            /* Save off the PBM packet handle so it can be handled by this layer later */
            pCurrDmaPacket->appPriv = (void *)hPBMPacket;
            CpswUtils_enQ(&resubmitPktQ, &pCurrDmaPacket->node);

            pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(&hNdk2Cpsw->rxFreePktInfoQ);
        }
        else
        {
            /* As we allocate 2X number of PBM buffers to no. of descriptors, we should always
             * have free PBM buffer */
            Ndk2Cpsw_print("ERROR: PBM_alloc() failure...exiting!\n");
            Ndk2Cpsw_print("PBMQ_free count = %d\n", PBMQ_count(&PBMQ_free));
            Ndk2Cpsw_assert(FALSE);

            CpswUtils_enQ(&hNdk2Cpsw->rxFreePktInfoQ, &pCurrDmaPacket->node);
            break;
        }
    }

    /*
     * Return the same DMA packets back to the DMA channel (but now
     * associated with a new PBM Packet and buffer)
     */
    Ndk2Cpsw_submitRxPackets(hNdk2Cpsw, &resubmitPktQ);
}

static uint32_t Ndk2Cpsw_prepRxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                    CpswDma_PktInfoQ *pPktQ)
{
    uint32_t packetCount = 0;
    CpswDma_PktInfo *pCurrDmaPacket;
#ifdef NDK_CHECKSUM_SUPPORT
    uint32_t csumInfo;
#endif
    bool chkSumErr = false;

    pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(pPktQ);
    while (pCurrDmaPacket)
    {
        /* Get the full PBM packet that needs to be returned to the NDK stack */
        PBM_Handle hPBMPacket = (PBM_Handle)pCurrDmaPacket->appPriv;
        if (hPBMPacket)
        {
            uint32_t validLen = pCurrDmaPacket->userBufLen;

            /* Fill in PBM packet length field */
            PBM_setValidLen(hPBMPacket, validLen);

#ifdef NDK_CHECKSUM_SUPPORT
            /* We don't check if HW checksum offload is enabled while checking for checksum error
             * as default value of this field when offload not enabled is false */
            csumInfo = pCurrDmaPacket->chkSumInfo;
            if ( CPSWDMA_CPPIPSI_GET_IPV4_FLAG(csumInfo)||
                 CPSWDMA_CPPIPSI_GET_IPV6_FLAG(csumInfo) )
            {
                chkSumErr = CPSWDMA_CPPIPSI_GET_CHKSUM_ERR_FLAG(csumInfo);
            }
#endif

            if (!chkSumErr)
            {
                /* Enqueue the received packet so that it can be consumed by the NDK stack*/
                PBMQ_enq(&hNdk2Cpsw->rxReadyPBMPktQ, hPBMPacket);
                Ndk2CpswStas_addOne(&gNdk2CpswStats.rxReadyPBMPktEnq);
                packetCount++;
            }
            else
            {
                /* Put PBM buffer in free Q as we are not passing to stack */
                PBM_free(hPBMPacket);
                Ndk2CpswStas_addOne(&gNdk2CpswStats.rxPBMFreeCnt);
                Ndk2CpswStas_addOne(&gNdk2CpswStats.rxChkSumErr);
            }

            /* Put packet info into free Q as we have removed the PBM buffers
             * from the it */
            CpswUtils_enQ(&hNdk2Cpsw->rxFreePktInfoQ, &pCurrDmaPacket->node);
            Ndk2CpswStas_addOne(&gNdk2CpswStats.rxFreeAppPktEnq);

            pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(pPktQ);
        }
        else
        {
            /* Should never happen as this is received from HW */
            Ndk2Cpsw_assert(FALSE);
        }
    }

    /* return as many packets to CPSW as we can */
    Ndk2Cpsw_submitRxPktQ(hNdk2Cpsw);

    return packetCount;
}

static uint32_t Ndk2Cpsw_prepTxPktQ(Ndk2Cpsw_Handle hNdk2Cpsw,
                                    CpswDma_PktInfoQ *pPktQ)
{
    uint32_t packetCount;
    CpswDma_PktInfo *pCurrDmaPacket;
    PBM_Handle hPBMPacket;

    packetCount = CpswUtils_getQCount(pPktQ);

    pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(pPktQ);
    while (pCurrDmaPacket)
    {
        hPBMPacket = pCurrDmaPacket->appPriv;

        /* Free PBM buffer as it is transmitted by DMA now */
        PBM_free(hPBMPacket);
        /* Return packet info to free pool */
        CpswUtils_enQ(&hNdk2Cpsw->txFreePktInfoQ, &pCurrDmaPacket->node);
        pCurrDmaPacket = (CpswDma_PktInfo *)CpswUtils_deQ(pPktQ);
    }

    Ndk2CpswStats_addNum(&gNdk2CpswStats.txFreeAppPktEnq, packetCount);

    return packetCount;
}

static void Ndk2Cpsw_updateTxNotifyStats(uint32_t packetCount,
                                         uint32_t timeDiff)
{
#if defined(NIMU_INSTRUMENTATION_ENABLED)
    uint32_t notificationCount;
    uint32_t timePerPacket = timeDiff / packetCount;

    notificationCount = gNdk2CpswStats.txStats.dataNotifyCnt & (HISTORY_CNT - 1U);
    gNdk2CpswStats.txStats.dataNotifyCnt++;

    gNdk2CpswStats.txStats.totalPktCnt   += packetCount;
    gNdk2CpswStats.txStats.totalCycleCnt += timeDiff;

    gNdk2CpswStats.txStats.cycleCntPerNotify[notificationCount] = timeDiff;
    if (timeDiff > gNdk2CpswStats.txStats.cycleCntPerNotifyMax)
    {
        gNdk2CpswStats.txStats.cycleCntPerNotifyMax = timeDiff;
    }

    gNdk2CpswStats.txStats.pktsPerNotify[notificationCount] = packetCount;
    if (packetCount > gNdk2CpswStats.txStats.pktsPerNotifyMax)
    {
        gNdk2CpswStats.txStats.pktsPerNotifyMax = packetCount;
    }

    gNdk2CpswStats.txStats.cycleCntPerPkt[notificationCount] = timePerPacket;
    if (timePerPacket > gNdk2CpswStats.txStats.cycleCntPerPktMax)
    {
        gNdk2CpswStats.txStats.cycleCntPerPktMax = timePerPacket;
    }
#endif
}

static void Ndk2Cpsw_updateRxNotifyStats(uint32_t packetCount,
                                         uint32_t timeDiff)
{
#if defined(NIMU_INSTRUMENTATION_ENABLED)
    uint32_t notificationCount;
    uint32_t timePerPacket = timeDiff / packetCount;

    notificationCount = gNdk2CpswStats.rxStats.dataNotifyCnt & (HISTORY_CNT - 1U);
    gNdk2CpswStats.rxStats.dataNotifyCnt++;

    gNdk2CpswStats.rxStats.totalPktCnt   += packetCount;
    gNdk2CpswStats.rxStats.totalCycleCnt += timeDiff;

    gNdk2CpswStats.rxStats.cycleCntPerNotify[notificationCount] = timeDiff;
    if (timeDiff > gNdk2CpswStats.rxStats.cycleCntPerNotifyMax)
    {
        gNdk2CpswStats.rxStats.cycleCntPerNotifyMax = timeDiff;
    }

    gNdk2CpswStats.rxStats.pktsPerNotify[notificationCount] = packetCount;
    if (packetCount > gNdk2CpswStats.rxStats.pktsPerNotifyMax)
    {
        gNdk2CpswStats.rxStats.pktsPerNotifyMax = packetCount;
    }

    gNdk2CpswStats.rxStats.cycleCntPerPkt[notificationCount] = timePerPacket;
    if (timePerPacket > gNdk2CpswStats.rxStats.cycleCntPerPktMax)
    {
        gNdk2CpswStats.rxStats.cycleCntPerPktMax = timePerPacket;
    }
#endif
}

static void Ndk2Cpsw_print(const char *prnStr,
                           ...)
{
    // TODO fix me
    if (NULL != gNdk2CpswObj.printFxnCb)
    {
        /* Function is non- reentrant */
        va_list vaArgPtr;
        char *buf;

        buf = &gNdk2CpswObj.printBuf[0];
        va_start(vaArgPtr, prnStr);
        vsnprintf(buf, CPSW_CFG_PRINT_BUF_LEN, (const char *)prnStr, vaArgPtr);
        va_end(vaArgPtr);

        (*gNdk2CpswObj.printFxnCb)("[NDK2CPSW] ");
        (*gNdk2CpswObj.printFxnCb)(gNdk2CpswObj.printBuf);
    }
}

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static int32_t Ndk2Cpsw_startRxTx(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    int32_t status = CPSW_SOK;

    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->appInfo.txInfo.hTxChannel);
    Ndk2Cpsw_assert(NULL != hNdk2Cpsw->appInfo.rxInfo.hRxFlow);

    status = CpswDma_enableTxEvent(hNdk2Cpsw->appInfo.txInfo.hTxChannel);

    /* Prepare a Queue of DmaPktInfo pointers pointing to DmaPktInfoMem */
    Ndk2Cpsw_prepFreePktInfoQ(hNdk2Cpsw);

    Ndk2Cpsw_assert(hNdk2Cpsw->appInfo.hostPortRxMtu != 0);

    if (PBMQ_count(&PBMQ_free) < NDK2CPSW_RX_PACKETS * (uint32_t)2U)
    {
        /*
         * PBM buffers should be allocated twice as DMA descriptors.
         * The idea here is that even if we fill up all the DMA descriptors submit queue,
         * we will have another complete set to swap in right away.
         */
        Ndk2Cpsw_print("Not enough PBM buffers allocated = %d, expected = %d\n",
                       PBMQ_count(&PBMQ_free),
                       NDK2CPSW_RX_PACKETS * 2U);
        Ndk2Cpsw_assert(false);
    }

    /* Submit all allocated packets to DMA so it can use for packet RX */
    Ndk2Cpsw_submitRxPktQ(hNdk2Cpsw);

    return status;
}

static void Ndk2Cpsw_stopRxTx(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    /* Close RX channel */
    /* Post to rx packet task so that it will terminate (shutDownFlag flag is already set) */
    if (hNdk2Cpsw->hRxPacketTask != NULL)
    {
        Semaphore_post(hNdk2Cpsw->hRxPacketSem);

        while (Task_Mode_TERMINATED != Task_getMode(hNdk2Cpsw->hRxPacketTask))
        {
            /* Sleep so Rx packet task gets scheduled & terminates itself */
            Task_sleep(1U);
        }
    }

    /* Close TX channel */
    /* Post to tx packet task so that it will terminate (shutDownFlag flag is already set) */
    if (hNdk2Cpsw->hTxPacketTask != NULL)
    {
        Semaphore_post(hNdk2Cpsw->hTxPacketSem);

        while (Task_Mode_TERMINATED != Task_getMode(hNdk2Cpsw->hTxPacketTask))
        {
            /* Sleep so Tx packet task gets scheduled & terminates itself */
            Task_sleep(1U);
        }
    }
}

static void Ndk2Cpsw_freeTxPktCb(void *cbArg,
                                 CpswDma_PktInfoQ *fqPktInfoQ,
                                 CpswDma_PktInfoQ *cqPktInfoQ)
{
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)cbArg;

    CpswUtils_appendQ(&hNdk2Cpsw->txFreePktInfoQ, fqPktInfoQ);
    Ndk2CpswStats_addNum(&gNdk2CpswStats.txFreeAppPktEnq, CpswUtils_getQCount(fqPktInfoQ));
    Ndk2Cpsw_freePBMPackets(fqPktInfoQ);

    CpswUtils_appendQ(&hNdk2Cpsw->txFreePktInfoQ, cqPktInfoQ);
    Ndk2CpswStats_addNum(&gNdk2CpswStats.txFreeAppPktEnq, CpswUtils_getQCount(cqPktInfoQ));
    Ndk2Cpsw_freePBMPackets(cqPktInfoQ);
}

static void Ndk2Cpsw_freeRxPktCb(void *cbArg,
                                 CpswDma_PktInfoQ *fqPktInfoQ,
                                 CpswDma_PktInfoQ *cqPktInfoQ)
{
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)cbArg;

    /* Now that we free PBM buffers, push all freed pktInfo's into Rx
     * free Q */
    CpswUtils_appendQ(&hNdk2Cpsw->rxFreePktInfoQ, fqPktInfoQ);
    Ndk2CpswStats_addNum(&gNdk2CpswStats.rxFreeAppPktEnq, CpswUtils_getQCount(fqPktInfoQ));
    Ndk2CpswStats_addNum(&gNdk2CpswStats.rxPBMFreeCnt, CpswUtils_getQCount(fqPktInfoQ));
    Ndk2Cpsw_freePBMPackets(fqPktInfoQ);

    CpswUtils_appendQ(&hNdk2Cpsw->rxFreePktInfoQ, cqPktInfoQ);
    Ndk2CpswStats_addNum(&gNdk2CpswStats.rxFreeAppPktEnq, CpswUtils_getQCount(cqPktInfoQ));
    Ndk2CpswStats_addNum(&gNdk2CpswStats.rxPBMFreeCnt, CpswUtils_getQCount(fqPktInfoQ));
    Ndk2Cpsw_freePBMPackets(cqPktInfoQ);

    /* Flush out our pending receive queues */
    while (PBMQ_count(&hNdk2Cpsw->rxReadyPBMPktQ) != 0)
    {
        PBM_free(PBMQ_deq(&hNdk2Cpsw->rxReadyPBMPktQ));
        Ndk2CpswStas_addOne(&gNdk2CpswStats.rxReadyPBMPktDq);
        Ndk2CpswStas_addOne(&gNdk2CpswStats.rxPBMFreeCnt);
    }
}

static void Ndk2Cpsw_retrieveTxPkts(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    CpswDma_PktInfoQ tempQueue;
    uint32_t packetCount = 0;
    int32_t retVal;

    Ndk2CpswStas_addOne(&gNdk2CpswStats.txStats.rawNotifyCnt);
    packetCount = 0;

    /* Retrieve the used (sent/empty) packets from the channel */
    {
        CpswUtils_initQ(&tempQueue);
        /* Retrieve all TX packets and keep them locally */
        retVal = CpswDma_retrieveTxDonePackets(hNdk2Cpsw->appInfo.txInfo.hTxChannel,
                                               &tempQueue);
        if (CPSW_SOK != retVal)
        {
            Ndk2Cpsw_print("Ndk2Cpsw_retrieveTxPkts: Failed to retrieve TX pkts: %d\n", retVal);
        }
    }

    if (tempQueue.count != 0U)
    {
        /*
         * Get all used Tx DMA packets from the hardware, then return those
         * buffers to the txFreePktQ so they can be used later to send with.
         */
        packetCount = Ndk2Cpsw_prepTxPktQ(hNdk2Cpsw, &tempQueue);
    }
    else
    {
        Ndk2CpswStas_addOne(&gNdk2CpswStats.txStats.zeroNotifyCnt);
    }

    if (packetCount != 0)
    {
        Ndk2Cpsw_updateTxNotifyStats(packetCount, 0U);
    }
}

static Void Ndk2Cpsw_timerCb(UArg arg)
{
    /* Post semaphore to rx handling task */
    Ndk2Cpsw_Handle hNdk2Cpsw = (Ndk2Cpsw_Handle)arg;

    if ((hNdk2Cpsw->initDone) && (hNdk2Cpsw->shutDownFlag == false))
    {
        Semaphore_post(hNdk2Cpsw->hRxPacketSem);
    }
}

static void Ndk2Cpsw_createTimer(Ndk2Cpsw_Handle hNdk2Cpsw)
{
    Timer_Params timerPrms;

    Timer_Params_init(&timerPrms);
    timerPrms.periodType = ITimer_PeriodType_MICROSECS;
    timerPrms.period     = hNdk2Cpsw->appInfo.timerPeriodUs;
    timerPrms.runMode    = Timer_RunMode_ONESHOT;
    timerPrms.startMode  = Timer_StartMode_USER;
    timerPrms.arg        = (UArg)hNdk2Cpsw;

    hNdk2Cpsw->hPacingTimer =  Timer_create(Timer_ANY  ,
                                            Ndk2Cpsw_timerCb,
                                            &timerPrms,
                                            NULL);
    Ndk2Cpsw_assert(hNdk2Cpsw->hPacingTimer != NULL);
}

static void  Ndk2Cpsw_initGetHandleInArgs(Ndk2Cpsw_Handle hNdk2Cpsw,
                                          NimuCpswAppIf_GetHandleInArgs *inArgs)
{
    inArgs->txCfg.cbArg      = hNdk2Cpsw;
    inArgs->txCfg.notifyCb   = &Ndk2Cpsw_notifyTxPackets;
    inArgs->txCfg.numPackets = NDK2CPSW_TX_PACKETS;

    inArgs->rxCfg.cbArg      = hNdk2Cpsw;
    inArgs->rxCfg.notifyCb   = &Ndk2Cpsw_notifyRxPackets;
    inArgs->rxCfg.numPackets = NDK2CPSW_RX_PACKETS;
}

static void Ndk2Cpsw_initReleaseHandleInArgs(Ndk2Cpsw_Handle hNdk2Cpsw,
                                             NimuCpswAppIf_ReleaseHandleInfo *inArgs)
{
    inArgs->coreId       = hNdk2Cpsw->appInfo.coreId;
    inArgs->coreKey      = hNdk2Cpsw->appInfo.coreKey;
    inArgs->hCpsw        = hNdk2Cpsw->appInfo.hCpsw;
    inArgs->hUdmaDrv     = hNdk2Cpsw->appInfo.hUdmaDrv;
    inArgs->rxFreePktCb  = &Ndk2Cpsw_freeRxPktCb;
    inArgs->rxInfo       = hNdk2Cpsw->appInfo.rxInfo;
    inArgs->txFreePktCb  = &Ndk2Cpsw_freeTxPktCb;
    inArgs->txInfo       = hNdk2Cpsw->appInfo.txInfo;
    inArgs->freePktCbArg = hNdk2Cpsw;
}

/*---------------------------------------------------------------------------*\
 |                                 End of File                                 |
 \*---------------------------------------------------------------------------*/
