/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw.c
 *
 * \brief This file contains the implementation of the main interface
 *        layer as well as CPSW_SS.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/cpsw.h>
#include "priv/cpsw_priv.h"
#include "priv/cpsw_ale_priv.h"
#include "priv/cpsw_cpts_priv.h"
#include "priv/cpsw_dma_priv.h"
#include "priv/cpsw_hostport_priv.h"
#include "priv/cpsw_macport_priv.h"
#include "priv/cpsw_mdio_priv.h"
#include "priv/cpsw_stats_priv.h"
#include "priv/cpsw_phy_priv.h"
#include "priv/cpsw_trace_priv.h"
#include "priv/cpsw_rm_priv.h"
#include "priv/cpsw_osal.h"
/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_COMMON_TX_MTU_DEFAULT                               (2024U)

#define CSL_CPSW_SS_MOD_VER_MODULE_ID_RESETVAL                   (0x00006BA0U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/*!
 * \brief MDIO status change (MDIO_LINKINT) context
 */
typedef struct Cpsw_MdioLinkIntCtx_s
{
    /*! Alive PHYs, updated upon MDIO_LINKINT interrupt */
    uint32_t aliveMask;

    /*! Linked PHYs, updated upon MDIO_LINKINT interrupt */
    uint32_t linkedMask;

    /*! Mask of PHYs with status change poll enabled */
    uint32_t pollEnableMask;

    /*! Link state change callback function pointer */
    Cpsw_MdioLinkStateChangeCb linkStateChangeCb;

    /*! Application data to be passed to the link state change callback */
    void *linkStateChangeCbArg;
} Cpsw_MdioLinkIntCtx;

/*!
 * \brief Port link state (link up/down, tick enabled)
 */
typedef struct Cpsw_PortLinkState_s
{
    /*! PHY address */
    uint32_t phyAddr;

    /*! Whether link port has been opened */
    bool isOpen;

    /*! Port link up/down state as reported by driver's FSM */
    bool isLinkUp;

    /*! Whether MDIO state change polling is enabled on this port's PHY */
    bool isPollEnabled;

    /*! Whether PHY periodic tick is enabled on the port link's PHY */
    bool isTickEnabled;
} Cpsw_PortLinkState;

/*!
 * \brief CPSW driver object
 */
typedef struct Cpsw_Obj_s
{
    /*! Magic value to identify whether object handle is valid or not */
    Cpsw_Magic magic;

    /*! CPSW instance type */
    Cpsw_Type cpswType;

    /*! Core on which #Cpsw_Open() is executed */
    uint32_t selfCoreId;

    /*! Main, API-level lock */
    void *lock;

    /*! DMA handle */
    CpswDma_Handle dma;

    /*! Host port handle */
    CpswHostPort_Handle hostPort;

    /*! MAC port handles */
    CpswMacPort_Handle macPort[CPSW_MAC_PORT_NUM];

    /*! PHY handles */
    CpswPhy_Handle phy[CPSW_MAC_PORT_NUM];

    /*! ALE handle */
    CpswAle_Handle ale;

    /*! CPTS handle */
    CpswCpts_Handle cpts;

    /*! MDIO handle */
    CpswMdio_Handle mdio;

    /*! Network statistics handle */
    CpswStats_Handle stats;

    /*! Resource Manager Handle */
    CpswRm_Handle hRm;

    /*! CPSW DMA Rx Reserved flow handle*/
    CpswDma_RxFlowHandle hRxRsvdFlow;

    /*! CPSW DMA Rx Reserved flow Id */
    uint32_t rsvdFlowId;

    /*! CPSW susbsystem register overlay */
    CSL_Xge_cpsw_ss_sRegs *cpswSsRegs;

    /*! CPSW register overlay */
    CSL_Xge_cpswRegs *cpswRegs;

    /*! PHY link up/down state */
    Cpsw_PortLinkState portLinkState[CPSW_MAC_PORT_NUM];

    /*! Statistics interrupt handle */
    void *hStatsIntr;

    /*! MDIO interrupt handle */
    void *hMdioIntr;

    /*! CPTS interrupt handle */
    void *hCptsIntr;

    /*! MDIO link state change interrupt context */
    Cpsw_MdioLinkIntCtx mdioLinkIntCtx;

    /* Port link status change callback function pointer */
    Cpsw_PortLinkStatusChangeCb portLinkStatusChangeCb;

    /*! Application data to be passed to the port link status change callback */
    void *portLinkStatusChangeCbArg;

    /*! Maximum of MTUs for TX Priority 0 to 7 */
    uint32_t maxPerPrioMtu;
} Cpsw_Obj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static Cpsw_Handle Cpsw_getObject(Cpsw_Type cpswType);

static int32_t Cpsw_openInternal(Cpsw_Handle hCpsw,
                                 Cpsw_Type cpswType,
                                 const Cpsw_Config *cfg);

static void Cpsw_closeInternal(Cpsw_Handle hCpsw);

static int32_t Cpsw_ioctlInternal(Cpsw_Handle hCpsw,
                                  uint32_t coreId,
                                  uint32_t cmd,
                                  Cpsw_IoctlPrms *prms);

static int32_t Cpsw_rmIrqSet(uint32_t dmscDevIdCpsw,
                             uint32_t srcIdx,
                             uint32_t coreIntNum);

static int32_t Cpsw_registerIntrs(Cpsw_Handle hCpsw,
                                  const Cpsw_Config *cfg);

static int32_t Cpsw_rmIrqRelease(uint32_t dmscDevIdCpsw,
                                 uint32_t srcIdx,
                                 uint32_t coreIntNum);

static void Cpsw_unregisterIntrs(Cpsw_Handle hCpsw);

static void Cpsw_statsIsr(uintptr_t arg);

static void Cpsw_mdioIsr(uintptr_t arg);

static void Cpsw_cptsIsr(uintptr_t arg);

static void Cpsw_handleMdioLinkStateChange(CpswMdio_UserChannel userCh,
                                           CpswMdio_PhyStatus *phyStatus,
                                           void *appData);

static int32_t Cpsw_printRegisters(Cpsw_Handle hCpsw);

static int32_t Cpsw_getVersion(Cpsw_Handle hCpsw,
                               Cpsw_GetVersionsOutArgs *version);

static int32_t Cpsw_setHostPortTrafficShaping(Cpsw_Handle hCpsw,
                                              CpswPort_TrafficShapingConfig *shapingCfg,
                                              uint32_t txBlocksRem);

static uint32_t Cpsw_mapPort2XgmiiId(uint32_t portNum);

static int32_t Cpsw_getPortLinkConfig(Cpsw_Handle hCpsw,
                                      Cpsw_MacPort portNum,
                                      CpswMacPort_LinkConfig *linkCfg);

static int32_t Cpsw_isPortLinkUp(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum,
                                 bool *linked);

static int32_t Cpsw_openPortLinkWithPhy(Cpsw_Handle hCpsw,
                                        Cpsw_MacPort portNum,
                                        CpswMacPort_Interface *interface,
                                        CpswMacPort_LinkConfig *linkConfig,
                                        CpswMacPort_Config *macConfig,
                                        CpswPhy_Config *phyConfig);

static int32_t Cpsw_openPortLinkNoPhy(Cpsw_Handle hCpsw,
                                      Cpsw_MacPort portNum,
                                      CpswMacPort_Interface *interface,
                                      CpswMacPort_LinkConfig *linkConfig,
                                      CpswMacPort_Config *macConfig);

static int32_t Cpsw_openPortLink(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Interface *interface,
                                 CpswMacPort_LinkConfig *linkConfig,
                                 CpswMacPort_Config *macConfig,
                                 CpswPhy_Config *phyConfig);

static void Cpsw_closePortLink(Cpsw_Handle hCpsw,
                               Cpsw_MacPort portNum);

static Cpsw_PortLinkState *Cpsw_getPortLinkState(Cpsw_Handle hCpsw,
                                                 uint32_t phyAddr);

static int32_t Cpsw_runPhyTick(Cpsw_Handle hCpsw,
                               Cpsw_MacPort portNum,
                               bool *linkChanged,
                               bool *linked);

static int32_t Cpsw_handleLinkUp(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum);

static int32_t Cpsw_handleLinkDown(Cpsw_Handle hCpsw,
                                   Cpsw_MacPort portNum);

static void Cpsw_validateSsIPVersion(CSL_Xge_cpsw_ss_sRegs *cpswSsRegBase,
                                     Cpsw_Type cpswType);

static int32_t Cpsw_setInterVlanRouteUniEgress(const Cpsw_Handle hCpsw,
                                               const Cpsw_SetInterVlanRouteUniEgressInArgs *inArgs,
                                               Cpsw_SetInterVlanRouteUniEgressOutArgs *outArgs);

static int32_t Cpsw_setInterVlanRouteMultiEgress(const Cpsw_Handle hCpsw,
                                                 const Cpsw_SetInterVlanRouteMultiEgressInArgs *inArgs,
                                                 Cpsw_SetInterVlanRouteMultiEgressOutArgs *outArgs);

static uint32_t Cpsw_normalizeRouteId(CpswMacPort_InterVLANRouteId routeId);

static int32_t Cpsw_clearInterVlanRouteUniEgress(const Cpsw_Handle hCpsw,
                                                 const Cpsw_ClearInterVlanRouteUniEgressInArgs *inArgs,
                                                 Cpsw_ClearInterVlanRouteUniEgressOutArgs *outArgs);

static int32_t Cpsw_clearInterVlanRouteMultiEgress(const Cpsw_Handle hCpsw,
                                                   const Cpsw_ClearInterVlanRouteMultiEgressInArgs *inArgs,
                                                   Cpsw_ClearInterVlanRouteMultiEgressOutArgs *outArgs);

static int32_t Cpsw_setTxShortIPGConfig(const Cpsw_Handle hCpsw,
                                        const Cpsw_SetShortTxIPGConfigInArgs *inArgs);

static int32_t Cpsw_getTxShortIPGConfig(const Cpsw_Handle hCpsw,
                                        Cpsw_GetShortTxIPGConfigOutArgs *outArgs);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW objects */
static Cpsw_Obj cpswObjs[CPSW_COUNT];

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_ERROR)
static const char *Cpsw_gSpeedNames[] =
{
    [CPSW_SPEED_10MBIT]  = "10-Mbps",
    [CPSW_SPEED_100MBIT] = "100-Mbps",
    [CPSW_SPEED_1GBIT]   = "1-Gbps",
    [CPSW_SPEED_2p5GBIT] = "2.5-Gbps",
    [CPSW_SPEED_10GBIT]  = "10-Gbps",
    [CPSW_SPEED_AUTO]    = "auto",
};

static const char *Cpsw_gDuplexNames[] =
{
    [CPSW_DUPLEX_HALF] = "Half-Duplex",
    [CPSW_DUPLEX_FULL] = "Full-Duplex",
    [CPSW_DUPLEX_AUTO] = "auto",
};
#endif

/* Default Host and MAC port TX Priority MTUs*/
static const uint32_t Cpsw_txPriMtuDefault[] =
{
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 0 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 1 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 2 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 3 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 4 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 5 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT, /* TX Priority 6 MTU */
    CPSW_COMMON_TX_MTU_DEFAULT  /* TX Priority 7 MTU */
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void Cpsw_initParams(Cpsw_Config *cfg)
{
    uint32_t i;

    cfg->escalatePriorityLoadVal = CPSW_CFG_ESC_PRI_LD_VAL;
    cfg->vlanConfig.vlanAware    = false;
    cfg->vlanConfig.vlanSwitch   = CPSW_VLAN_SWITCH_CUSTOMER;
    cfg->vlanConfig.outerVlan    = 0x88A8U;
    cfg->vlanConfig.innerVlan    = 0x8100U;

    CPSW_UTILS_COMPILETIME_ASSERT
        (CPSW_UTILS_ARRAYSIZE(cfg->txMtu) == CPSW_PRIORITY_NUM);

    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        cfg->txMtu[i] = Cpsw_txPriMtuDefault[i];
    }

    /* Initialize host port config params */
    CpswHostPort_initParams(&cfg->hostPortConfig);

    /* Initialize CPTS config params */
    CpswCpts_initParams(&cfg->cptsConfig);

    /* Initialize MDIO config params */
    CpswMdio_initParams(&cfg->mdioConfig);

    /* Initialize ALE params */
    CpswAle_initParams(&cfg->aleConfig);

    /* Initialize DMA params */
    CpswDma_initParams(&cfg->dmaConfig);

    cfg->intrPriority              = 1U;
    cfg->mdioLinkStateChangeCb     = NULL;
    cfg->mdioLinkStateChangeCbArg  = NULL;
    cfg->portLinkStatusChangeCb    = NULL;
    cfg->portLinkStatusChangeCbArg = NULL;

    cfg->enableQsgmii0RDC = false;
    cfg->enableQsgmii1RDC = false;

}

void Cpsw_initMacPortParams(CpswMacPort_Config *cfg)
{
    CpswMacPort_initParams(cfg);
}

void Cpsw_initPhyParams(CpswPhy_Config *cfg)
{
    CpswPhy_initParams(cfg);
}

void Cpsw_setPhyExtendedParams(CpswPhy_Config *cfg,
                               const void *extendedCfg,
                               uint32_t extendedCfgSize)
{
    CpswPhy_setExtendedParams(cfg, extendedCfg, extendedCfgSize);
}

void Cpsw_initOsalPrms(CpswOsal_Prms *pOsalPrms)
{
    CpswOsal_initPrms(pOsalPrms);
}

Cpsw_Handle Cpsw_getHandle(Cpsw_Type cpswType)
{
    Cpsw_Handle hCpsw = NULL;

    if (cpswObjs[cpswType].magic == CPSW_DRV_MAGIC)
    {
        hCpsw = &cpswObjs[cpswType];
    }

    return hCpsw;
}

void Cpsw_init(Cpsw_Type cpswType,
               const CpswOsal_Prms *pOsalPrms,
               CpswUtils_Prms *pUtilsPrms)
{
    Cpsw_Handle hCpsw = &cpswObjs[cpswType];
    CpswOsal_Prms defaultOsalPrms;
    CpswUtils_Prms defaultUtilsPrms =
    {
        .printFxn     = NULL,
        .virtToPhyFxn = NULL,
        .phyToVirtFxn = NULL,
        .traceFxn     = NULL
    };

    memset(hCpsw, 0U, sizeof(*hCpsw));

    /* Open OSAL, use defaults if NULL params passed */
    if (pOsalPrms != NULL)
    {
        CpswOsal_open(pOsalPrms);
    }
    else
    {
        CpswOsal_initPrms(&defaultOsalPrms);
        CpswOsal_open(&defaultOsalPrms);
    }

    /* Initialize utils, use defaults if NULL params passed */
    if (pUtilsPrms != NULL)
    {
        CpswUtils_init(pUtilsPrms);
    }
    else
    {
        CpswUtils_init(&defaultUtilsPrms);
    }

    /* Create lock */
    hCpsw->lock = CpswOsal_mutexCreate();
    Cpsw_assert(hCpsw->lock != NULL);
}

void Cpsw_deinit(Cpsw_Type cpswType)
{
    Cpsw_Handle hCpsw = &cpswObjs[cpswType];

    /* Destroy lock */
    CpswOsal_mutexDelete(hCpsw->lock);
    hCpsw->lock = NULL;

    /* Deinitialize utils */
    CpswUtils_deInit();

    /* Close OSAL */
    CpswOsal_close();
}

CpswTrace_TraceLevel Cpsw_setTraceLevel(CpswTrace_TraceLevel level)
{
    return CpswTrace_setLevel(level);
}

CpswTrace_TraceLevel Cpsw_getTraceLevel(void)
{
    return CpswTrace_getLevel();
}

static int32_t Cpsw_setDefaultThreadCfg(Cpsw_Handle hCpsw,
                                        uint32_t flowId)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    CpswAle_SetDefaultThreadConfigInArgs setDefaultThreadConfigInArgs;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &setDefaultThreadConfigInArgs);
    status = CpswAle_ioctl(hCpsw->ale,
                           CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG,
                           &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG failed: %d\n", status);
    if (status == CPSW_SOK)
    {
        setDefaultThreadConfigInArgs.defaultThreadEnable = TRUE;
        setDefaultThreadConfigInArgs.threadId            = flowId;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setDefaultThreadConfigInArgs);

        status = CpswAle_ioctl(hCpsw->ale,
                               CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG,
                               &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set default thread in ALE: %d\n", status);
    }

    return status;
}

static int32_t Cpsw_validateCpswType(Cpsw_Type cpswType)
{
    int32_t status = CPSW_SOK;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    if (NULL == ipInfo)
    {
        CPSWTRACE_ERROR("CPSWALE:Unsupported CPSW_TYPE:%d", cpswType);
        Cpsw_assert(false);
    }

    return status;
}

Cpsw_Handle Cpsw_open(Cpsw_Type cpswType,
                      const Cpsw_Config *cfg)
{
    Cpsw_Handle hCpsw;
    Cpsw_IoctlPrms prms;
    CSL_CPSW_PTYPE pType;
    uintptr_t key;
    uint32_t i;
    int32_t status = CPSW_SOK;

    Cpsw_assert(cfg != NULL);
    Cpsw_validateCpswType(cpswType);

    hCpsw = Cpsw_getObject(cpswType);

    CpswOsal_mutexLock(hCpsw->lock);

    if (hCpsw->magic == CPSW_DRV_MAGIC)
    {
        CPSWTRACE_WARN("CPSW is already open\n");
        status = CPSW_EALREADY_OPEN;
    }

    if (status == CPSW_SOK)
    {
        hCpsw->cpswSsRegs = Cpsw_getCslSsRegs(cpswType);
        Cpsw_assert(hCpsw->cpswSsRegs != NULL);
        Cpsw_validateSsIPVersion(hCpsw->cpswSsRegs, cpswType);

        hCpsw->cpswRegs = Cpsw_getCslRegs(cpswType);
        Cpsw_assert(hCpsw->cpswRegs != NULL);

        CSL_CPSW_getPTypeReg(hCpsw->cpswRegs, &pType);
        pType.escPriLoadVal = cfg->escalatePriorityLoadVal;
        CSL_CPSW_setPTypeReg(hCpsw->cpswRegs, &pType);

        hCpsw->cpswType = cpswType;

        /* Set VLAN config */
        if (cfg->vlanConfig.vlanAware == TRUE)
        {
            CSL_CPSW_enableVlanAware(hCpsw->cpswRegs);
        }
        else
        {
            CSL_CPSW_disableVlanAware(hCpsw->cpswRegs);
        }

        CSL_CPSW_setVlanType(hCpsw->cpswRegs,
                             (uint32_t)cfg->vlanConfig.vlanSwitch);

        CSL_CPSW_setVlanLTypeReg(hCpsw->cpswRegs,
                                 cfg->vlanConfig.innerVlan,
                                 cfg->vlanConfig.outerVlan);

        /* Set port global config */
        for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
        {
            CSL_CPSW_setTxMaxLenPerPriority(hCpsw->cpswRegs,
                                            i,
                                            cfg->txMtu[i]);

            /* Save largest of per priority port egress MTUs to check against
             * host port and MAC port RX MTUs */
            if (cfg->txMtu[i] > hCpsw->maxPerPrioMtu)
            {
                hCpsw->maxPerPrioMtu = cfg->txMtu[i];
            }
        }

        /* Configure QSGMII RDCD */
        CSL_CPSW_SS_setQSGMIIControlRdCd(hCpsw->cpswSsRegs,
                                        0U, /* QSGMII 0 */
                                        (cfg->enableQsgmii0RDC ? 1U : 0U));
        CSL_CPSW_SS_setQSGMIIControlRdCd(hCpsw->cpswSsRegs,
                                        1U, /* QSGMII 1 */
                                        (cfg->enableQsgmii1RDC ? 1U : 0U));
    }

    /* Initialize all CPSW submodules */
    if (status == CPSW_SOK)
    {
        status = Cpsw_openInternal(hCpsw, cpswType, cfg);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to open submodules: %d\n", status);
    }

    /* Set host port flowIdOffset which is added to the switch transmit
     * (egress/Host receive) Flow Id */
    if (status == CPSW_SOK)
    {
        /* Get the flow startIdx for CPSW RX channel */
        uint32_t flowIdOffset = CpswDma_getRxChFlowStartIdx(hCpsw->dma);

        CPSW_IOCTL_SET_IN_ARGS(&prms, &flowIdOffset);
        status = CpswHostPort_ioctl(hCpsw->hostPort,
                                    CPSW_HOSTPORT_SET_FLOW_ID_OFFSET,
                                    &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set flow ID: %d\n", status);
    }

    /* Register interrupts */
    key = CpswOsal_disableAllIntr();
    if (status == CPSW_SOK)
    {
        status = Cpsw_registerIntrs(hCpsw, cfg);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to register interrupts: %d\n", status);
    }

    /* Enable CPTS interrupt */
    if (status == CPSW_SOK && hCpsw->cpts != NULL)
    {
        status = CpswCpts_ioctl(hCpsw->cpts,
                                CPSW_CPTS_IOCTL_ENABLE_INTR,
                                NULL);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to enable CPTS Interrupt: %d\n", status);
    }

    /* All initialization is complete, set magic number now */
    if (status == CPSW_SOK)
    {
        hCpsw->magic      = CPSW_DRV_MAGIC;
        hCpsw->cpswType   = cpswType;
        hCpsw->selfCoreId = cfg->resourceConfig.selfCoreId;
    }
    else if (status != CPSW_EALREADY_OPEN)
    {
        /* Rollback if any error other than trying to open while already open */
        Cpsw_unregisterIntrs(hCpsw);
        Cpsw_closeInternal(hCpsw);
        hCpsw->cpswSsRegs = NULL;
        hCpsw->cpswRegs   = NULL;
    }
    else
    {
        CPSWTRACE_ERROR("unexpected status: %d\n", status);
    }

    CpswOsal_mutexUnlock(hCpsw->lock);

    CpswOsal_restoreAllIntr(key);

    return (status == CPSW_SOK) ? hCpsw : NULL;
}

Cpsw_Handle Cpsw_reopen(Cpsw_Type cpswType)
{
    uint32_t priTxMtu, i;
    Cpsw_Handle hCpsw = Cpsw_getObject(cpswType);

    hCpsw->cpswSsRegs = Cpsw_getCslSsRegs(cpswType);
    Cpsw_assert(hCpsw->cpswSsRegs != NULL);

    hCpsw->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hCpsw->cpswRegs != NULL);

    /* Recompute the maximum of per priority MTUs */
    hCpsw->maxPerPrioMtu = 0U;
    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        priTxMtu = CSL_CPSW_getTxMaxLenPerPriority(hCpsw->cpswRegs, i);
        if (priTxMtu > hCpsw->maxPerPrioMtu)
        {
            hCpsw->maxPerPrioMtu = priTxMtu;
        }
    }

    hCpsw->magic    = CPSW_DRV_MAGIC;
    hCpsw->cpswType = cpswType;

    return hCpsw;
}

void Cpsw_close(Cpsw_Handle hCpsw)
{
    uintptr_t key;
    int32_t status = CPSW_SOK;

    if (hCpsw != NULL)
    {
        CpswOsal_mutexLock(hCpsw->lock);

        key = CpswOsal_disableAllIntr();

        /* Disable CPTS interrupt */
        if (status == CPSW_SOK && hCpsw->cpts != NULL)
        {
            status = CpswCpts_ioctl(hCpsw->cpts,
                                    CPSW_CPTS_IOCTL_DISABLE_INTR,
                                    NULL);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to disable CPTS Interrupt: %d\n", status);
        }

        if (status == CPSW_SOK)
        {
            Cpsw_unregisterIntrs(hCpsw);

            if (hCpsw->magic == CPSW_DRV_MAGIC)
            {
                Cpsw_setDefaultThreadCfg(hCpsw, hCpsw->rsvdFlowId);
                Cpsw_closeInternal(hCpsw);

                hCpsw->magic      = CPSW_NO_MAGIC;
                hCpsw->cpswSsRegs = NULL;
                hCpsw->cpswRegs   = NULL;
            }

            CpswOsal_restoreAllIntr(key);
            CpswOsal_mutexUnlock(hCpsw->lock);
        }
    }
}

int32_t Cpsw_ioctl(Cpsw_Handle hCpsw,
                   uint32_t coreId,
                   uint32_t cmd,
                   Cpsw_IoctlPrms *prms)
{
    uint32_t major = CPSW_IOCTL_GET_MAJ(cmd);
    int32_t status = CPSW_SOK;

    Cpsw_assert(hCpsw != NULL);

    if (CPSW_IOCTL_GET_TYPE(cmd) != CPSW_IOCTL_TYPE_PUBLIC)
    {
        CPSWTRACE_ERROR("IOCTL cmd %u is not public\n", cmd);
        status = CPSW_EINVALID_PARAMS;
    }

    if ((hCpsw->magic == CPSW_DRV_MAGIC) &&
        (status == CPSW_SOK))
    {
        Cpsw_IoctlPrms rmPrms;
        CpswRm_ValidatePermissionInArgs permInArgs;

        permInArgs.cmd    = cmd;
        permInArgs.coreId = coreId;
        CPSW_IOCTL_SET_IN_ARGS(&rmPrms, &permInArgs);

        status = CpswRm_ioctlPriv(hCpsw->hRm,
                                  CPSW_RM_IOCTL_VALIDATE_PERMISSION,
                                  &rmPrms);
    }

    if ((hCpsw->magic == CPSW_DRV_MAGIC) &&
        (status == CPSW_SOK))
    {
        CpswOsal_mutexLock(hCpsw->lock);

        switch (major)
        {
            case CPSW_IOCTL_CPSW_BASE:
            {
                status = Cpsw_ioctlInternal(hCpsw, coreId, cmd, prms);
            }
            break;

            case CPSW_IOCTL_ALE_BASE:
            {
                CPSW_ALE_IOCTL_E aleCmd;

                aleCmd = (CPSW_ALE_IOCTL_E)cmd;
                status = CpswAle_ioctl(hCpsw->ale, aleCmd, prms);
            }
            break;

            case CPSW_IOCTL_HOSTPORT_BASE:
            {
                status = CpswHostPort_ioctl(hCpsw->hostPort, cmd, prms);
            }
            break;

            case CPSW_IOCTL_MACPORT_BASE:
            {
                /* Note: Typecast to GenericInArgs is possible because all
                 * public MAC port IOCTL input args have portNum as
                 * their first member */
                CpswMacPort_GenericInArgs *inArgs = (CpswMacPort_GenericInArgs *)prms->inArgs;
                uint32_t portNum = inArgs->portNum;

                if ((portNum < Cpsw_getMacPortMax(hCpsw->cpswType)) &&
                    (hCpsw->macPort[portNum] != NULL))
                {
                    status = CpswMacPort_ioctl(hCpsw->macPort[portNum], cmd,
                                               prms);
                }
                else
                {
                    status = CPSW_EINVALID_PARAMS;
                }
            }
            break;

            case CPSW_IOCTL_MDIO_BASE:
            {
                status = CpswMdio_ioctl(hCpsw->mdio, cmd, prms);
            }
            break;

            case CPSW_IOCTL_STATS_BASE:
            {
                status = CpswStats_ioctl(hCpsw->stats, cmd, prms);
            }
            break;

            case CPSW_IOCTL_PHY_BASE:
            {
                /* Note: Typecast to GenericInArgs is possible because all
                 * public PHY IOCTL input args have portNum as
                 * their first member */
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                uint32_t portNum = inArgs->portNum;

                if ((portNum < Cpsw_getMacPortMax(hCpsw->cpswType)) &&
                    (hCpsw->phy[portNum] != NULL))
                {
                    status = CpswPhy_ioctl(hCpsw->phy[portNum], cmd, prms);
                }
                else
                {
                    status = CPSW_EINVALID_PARAMS;
                }
            }
            break;

            case CPSW_IOCTL_RM_BASE:
            {
                if (hCpsw->hRm != NULL)
                {
                    status = CpswRm_ioctl(hCpsw->hRm, (CpswRm_Ioctls_e)cmd, prms);
                }
                else
                {
                    status = CPSW_EINVALID_PARAMS;
                }
            }
            break;

            case CPSW_IOCTL_CPTS_BASE:
            {
                if (hCpsw->cpts != NULL)
                {
                    status = CpswCpts_ioctl(hCpsw->cpts, cmd, prms);
                }
                else
                {
                    status = CPSW_EINVALID_PARAMS;
                }

            }
            break;
            default:
                status = CPSW_EINVALID_PARAMS;
                break;
        }

        CpswOsal_mutexUnlock(hCpsw->lock);
    }

    return status;
}

void Cpsw_periodicTick(Cpsw_Handle hCpsw)
{
    Cpsw_MacPort portNum;
    Cpsw_PortLinkState *portLinkState;
    bool linkChanged;
    bool linked;
    uint32_t maxPorts = 0U;
    uint32_t i;
    int32_t status;
    uint32_t numPortCb;
    struct Cpsw_PortLinkCbInfoList_s
    {
        bool linked;
        Cpsw_MacPort portNum;
    } portCbInfoList[CPSW_MAC_PORT_NUM];

    Cpsw_assert(hCpsw != NULL);

    numPortCb = 0;
    /* Get the max number of ports */
    CpswOsal_mutexLock(hCpsw->lock);
    if (hCpsw->magic == CPSW_DRV_MAGIC)
    {
        maxPorts = Cpsw_getMacPortMax(hCpsw->cpswType);
    }

    CpswOsal_mutexUnlock(hCpsw->lock);

    /* Run PHY tick and handle link up/down events */
    for (i = 0U; i < maxPorts; i++)
    {
        portLinkState = &hCpsw->portLinkState[i];

        /* Check if the corresponding PHY is enabled */
        if ((hCpsw->magic == CPSW_DRV_MAGIC) &&
            portLinkState->isTickEnabled)
        {
            CpswOsal_mutexLock(hCpsw->lock);

            portNum = (Cpsw_MacPort)i;

            if (hCpsw->phy[i] != NULL)
            {
                /* Run PHY tick */
                status = Cpsw_runPhyTick(hCpsw, portNum, &linkChanged, &linked);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to run port %d PHY tick: %d\n",
                                   portNum, status);

                /* Handle link up/down events */
                if (linkChanged && (status == CPSW_SOK))
                {
                    status = linked ? Cpsw_handleLinkUp(hCpsw, portNum) :
                             Cpsw_handleLinkDown(hCpsw, portNum);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to handle port %d link change: %d\n",
                                       portNum, status);

                    /* Call application callback when port link is up - at this point app can
                     * start data flow */
                    if ((status == CPSW_SOK) && (hCpsw->portLinkStatusChangeCb != NULL))
                    {
                        /* Add port's to callback info list.
                         * All portLinkStatus Cb functions are invoked at the
                         * end of function after relinquishing locks
                         */
                        Cpsw_assert(numPortCb < CPSW_UTILS_ARRAYSIZE(portCbInfoList));
                        portCbInfoList[numPortCb].portNum = portNum;
                        portCbInfoList[numPortCb].linked  = linked;
                        numPortCb++;
                    }
                }

                /* All checks cleared, link state can be updated now */
                if (linkChanged && (status == CPSW_SOK))
                {
                    portLinkState->isLinkUp = linked;

                    /* Disable periodic tick while link is up. It will be re-enabled
                     * by MDIO_LINKINT upon link down detection */
                    if (portLinkState->isPollEnabled && linked)
                    {
                        hCpsw->portLinkState[i].isTickEnabled = false;
                    }
                }
            }
            else if (hCpsw->macPort[i] != NULL)
            {
                /* If port is in noPhy mode , invoke the portLinkUp Cb */
                hCpsw->portLinkState[portNum].isLinkUp      = true;
                hCpsw->portLinkState[portNum].isTickEnabled = false;
                if (hCpsw->portLinkStatusChangeCb != NULL)
                {
                    /* Add port's to callback info list.
                     * All portLinkStatus Cb functions are invoked at the
                     * end of function after relinquishing locks
                     */
                    Cpsw_assert(numPortCb < CPSW_UTILS_ARRAYSIZE(portCbInfoList));
                    portCbInfoList[numPortCb].portNum = portNum;
                    portCbInfoList[numPortCb].linked  = true;
                    numPortCb++;
                }
            }

            CpswOsal_mutexUnlock(hCpsw->lock);
        }
    }

    for (i = 0U; i < numPortCb; i++)
    {
        if (hCpsw->portLinkStatusChangeCb != NULL)
        {
            /* Call application's port link status change callback */
            hCpsw->portLinkStatusChangeCb(portCbInfoList[i].portNum,
                                          portCbInfoList[i].linked,
                                          hCpsw->portLinkStatusChangeCbArg);
        }
    }

    return;
}

uint32_t Cpsw_getMacPortMax(Cpsw_Type cpswType)
{
    uint32_t ports = 0U;

    switch (cpswType)
    {
        case CPSW_2G:
            ports = 1U;
            break;

        case CPSW_5G:
            ports = 4U;
            break;

        case CPSW_9G:
            ports = 8U;
            break;

        default:
            Cpsw_assert(false);
    }

    return ports;
}

static Cpsw_Handle Cpsw_getObject(Cpsw_Type cpswType)
{
    Cpsw_Handle hCpsw = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hCpsw = &cpswObjs[cpswType];
            break;

        default:
            Cpsw_assert(false);
    }

    return hCpsw;
}

static uint32_t Cpsw_getTxMtuPerPriority(Cpsw_Handle hCpsw,
                                         uint32_t priority)
{
    uint32_t txMtu = 0U;

    if (hCpsw != NULL)
    {
        txMtu = CSL_CPSW_getTxMaxLenPerPriority(hCpsw->cpswRegs, priority);
    }

    return txMtu;
}

static uint32_t Cpsw_getRxMtuPort0(Cpsw_Handle hCpsw)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    uint32_t rxMtu;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &rxMtu);
    status = CpswHostPort_ioctl(hCpsw->hostPort,
                                CPSW_HOSTPORT_IOCTL_GET_RX_MTU,
                                &prms);
    if (status != CPSW_SOK)
    {
        CPSWTRACE_ERROR("failed to get host port's RX MTU: %d\n", status);
        rxMtu = 0U;
    }

    return rxMtu;
}

static void Cpsw_setRxRsvdflowPrms(Cpsw_Handle hCpsw,
                                   Udma_DrvHandle hUdmaDrv,
                                   CpswDma_RsvdRxFlowPrms *pRxFlowPrms,
                                   uint32_t rsvdFlowId)
{
    pRxFlowPrms->startIdx = CpswDma_getRxChFlowStartIdx(hCpsw->dma);
    pRxFlowPrms->flowIdx  = rsvdFlowId;
    pRxFlowPrms->hUdmaDrv = hUdmaDrv;
}

static uint32_t Cpsw_getTotalRxFlowCount(const CpswRm_ResourcePrms *rscPrms)
{
    uint32_t i;
    uint32_t rxFlowCount;

    rxFlowCount = 0;
    for (i = 0; i < rscPrms->numCores; i++)
    {
        rxFlowCount += rscPrms->coreDmaResourceInfo[i].numRxFlows;
    }

    return rxFlowCount;
}

static int32_t Cpsw_openInternal(Cpsw_Handle hCpsw,
                                 Cpsw_Type cpswType,
                                 const Cpsw_Config *cfg)
{
    const CpswSoc_Config *pSocCfg;
    Cpsw_MdioLinkIntCtx *linkIntCtx = &hCpsw->mdioLinkIntCtx;
    Cpsw_PortLinkState *portLinkState;
    uint32_t i;
    int32_t status = CPSW_SOK;

    /* Verify CPSW main clock frequency in SoC config is valid */
    pSocCfg = CpswSoc_getConfig(cpswType);
    Cpsw_assert(pSocCfg->cppiClkFreqHz != 0U);

    /* Clear PHY link state */
    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(hCpsw->portLinkState); i++)
    {
        portLinkState = &hCpsw->portLinkState[i];

        portLinkState->phyAddr       = CPSW_PHY_INVALID_PHYADDR;
        portLinkState->isOpen        = false;
        portLinkState->isLinkUp      = false;
        portLinkState->isPollEnabled = false;
        portLinkState->isTickEnabled = true;
    }

    linkIntCtx->aliveMask            = CPSW_MDIO_PHY_ADDR_MASK_NONE;
    linkIntCtx->linkedMask           = CPSW_MDIO_PHY_ADDR_MASK_NONE;
    linkIntCtx->pollEnableMask       = cfg->mdioConfig.pollEnableMask;
    linkIntCtx->linkStateChangeCb    = cfg->mdioLinkStateChangeCb;
    linkIntCtx->linkStateChangeCbArg = cfg->mdioLinkStateChangeCbArg;
    hCpsw->portLinkStatusChangeCb    = cfg->portLinkStatusChangeCb;
    hCpsw->portLinkStatusChangeCbArg = cfg->portLinkStatusChangeCbArg;

    if (cfg->hostPortConfig.rxMtu > hCpsw->maxPerPrioMtu)
    {
        /* Host port & MAC port MTUs should not be greater than largest of the port
         * egress per priority MTU as packet would get dropped by Switch.
         * Though it is valid from HW configuration, we return an error as it serves
         * no purpose */
        CPSWTRACE_ERROR("Host Port RX MTU (%d) exceeds max of TX Priority 0-7 MTU(%d)\n ",
                        cfg->hostPortConfig.rxMtu, hCpsw->maxPerPrioMtu);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        /* Open host port */
        hCpsw->hostPort = CpswHostPort_open(cpswType, &cfg->hostPortConfig);
        if (hCpsw->hostPort == NULL)
        {
            CPSWTRACE_ERROR("failed to open host port\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open ALE */
    if (status == CPSW_SOK)
    {
        hCpsw->ale = CpswAle_open(cpswType, &cfg->aleConfig);
        if (hCpsw->ale == NULL)
        {
            CPSWTRACE_ERROR("failed to open ALE\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open CPTS */
    if (status == CPSW_SOK)
    {
        hCpsw->cpts = CpswCpts_open(cpswType, &cfg->cptsConfig);
        if (hCpsw->cpts == NULL)
        {
            CPSWTRACE_ERROR("failed to open CPTS\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open MDIO */
    if (status == CPSW_SOK)
    {
        hCpsw->mdio = CpswMdio_open(cpswType, &cfg->mdioConfig);
        if (hCpsw->mdio == NULL)
        {
            CPSWTRACE_ERROR("failed to open MDIO\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open statistics */
    if (status == CPSW_SOK)
    {
        hCpsw->stats = CpswStats_open(cpswType);
        if (hCpsw->stats == NULL)
        {
            CPSWTRACE_ERROR("failed to open STATS\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open DMA */
    if (status == CPSW_SOK)
    {
        /* Open UDMA for CPSW NAVSS instance type */
        hCpsw->dma = CpswDma_open(cpswType, &cfg->dmaConfig);
        if (NULL != hCpsw->dma)
        {
            uint32_t totalRxFlowCount;

            /* Compute the total number of rx flows required for the Rx channel
             * by summing up the total number of rx flows allocated for the
             * different cores in the CPSW RM resourceConfig
             */
            totalRxFlowCount = Cpsw_getTotalRxFlowCount(&cfg->resourceConfig.resourcePartitionInfo);
            Cpsw_assert(totalRxFlowCount <= (CSL_ALE_THREADMAPVAL_THREADVAL_MAX + 1));

            /* Open & start channel immediately as we have not opened default flow
             * UDMA should not get any packets in CPSW channel */
            status = CpswDma_openRxCh(hCpsw->dma,
                                      &cfg->dmaConfig.rxChInitPrms,
                                      totalRxFlowCount);
        }
    }

    /* Open Resource Manager if Rx Channel open succeeded */
    if (status == CPSW_SOK)
    {
        CpswRm_Config rmCfg;

        rmCfg.cpswType              = cpswType;
        rmCfg.rxStartFlowIdx        = CpswDma_getRxChFlowStartIdx(hCpsw->dma);
        rmCfg.rxFlowIdxCnt          = CpswDma_getRxFlowCnt(hCpsw->dma);
        rmCfg.ioctlPermissionInfo   = cfg->resourceConfig.ioctlPermissionInfo;
        rmCfg.macList               = cfg->resourceConfig.macList;
        rmCfg.resourcePartitionInfo = cfg->resourceConfig.resourcePartitionInfo;

        hCpsw->hRm = CpswRm_Open(&rmCfg);
        if (hCpsw->hRm == NULL)
        {
            CPSWTRACE_ERROR("failed to open RM\n");
            status = CPSW_EFAIL;
        }
    }

    /* Open  Reserved flow if Rx channel & RM open succeeded */
    if (status == CPSW_SOK)
    {
        CpswDma_RsvdRxFlowPrms cpswRxFlowCfg;
        Cpsw_IoctlPrms prms;
        CpswRm_AllocRxFlowOutArgs allocRxFlowOutArgs;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &cfg->resourceConfig.selfCoreId, &allocRxFlowOutArgs);

        status = CpswRm_ioctlPriv(hCpsw->hRm,
                                  CPSW_RM_IOCTL_INTERNAL_ALLOC_RX_FLOW,
                                  &prms);
        if (status == CPSW_SOK)
        {
            Cpsw_setRxRsvdflowPrms(hCpsw,
                                   cfg->dmaConfig.hUdmaDrv,
                                   &cpswRxFlowCfg,
                                   allocRxFlowOutArgs.flowIdx);
            hCpsw->hRxRsvdFlow = CpswDma_openRxRsvdFlow(&cpswRxFlowCfg);
            Cpsw_assert(NULL != hCpsw->hRxRsvdFlow);
            hCpsw->rsvdFlowId = allocRxFlowOutArgs.flowIdx;

            status = Cpsw_setDefaultThreadCfg(hCpsw, hCpsw->rsvdFlowId);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to set default thread config: %d\n", status);
        }
        else
        {
            CPSWTRACE_ERROR("failed to allocate RX flow via RM: %d\n", status);
            hCpsw->rsvdFlowId = CPSW_DMA_RXFLOWIDX_INVALID;
        }
    }

    if (status != CPSW_SOK)
    {
        Cpsw_closeInternal(hCpsw);
    }

    return status;
}

static void Cpsw_closeInternal(Cpsw_Handle hCpsw)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeInternalRxFlowInArgs freeRsvdFlowInArgs;

    if (hCpsw->hostPort != NULL)
    {
        CpswHostPort_close(hCpsw->hostPort);
        hCpsw->hostPort = NULL;
    }

    if (hCpsw->ale != NULL)
    {
        CpswAle_close(hCpsw->ale);
        hCpsw->ale = NULL;
    }

    if (hCpsw->cpts != NULL)
    {
        CpswCpts_close(hCpsw->cpts);
        hCpsw->cpts = NULL;
    }

    if (hCpsw->mdio != NULL)
    {
        CpswMdio_close(hCpsw->mdio);
        hCpsw->mdio = NULL;
    }

    if (hCpsw->stats != NULL)
    {
        CpswStats_close(hCpsw->stats);
        hCpsw->stats = NULL;
    }

    if (hCpsw->hRxRsvdFlow != NULL)
    {
        status = CpswDma_closeRxRsvdFlow(hCpsw->hRxRsvdFlow);
        Cpsw_assert(status == CPSW_SOK);

        freeRsvdFlowInArgs.flowIdx = hCpsw->rsvdFlowId;
        freeRsvdFlowInArgs.coreId  = hCpsw->selfCoreId;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &freeRsvdFlowInArgs);

        status = CpswRm_ioctlPriv(hCpsw->hRm,
                                  CPSW_RM_IOCTL_INTERNAL_FREE_RX_FLOW,
                                  &prms);
        Cpsw_assert(status == CPSW_SOK);
    }

    if (hCpsw->dma->initFlag)
    {
        /* Stop and check RX channel */
        status = CpswDma_closeRxCh(hCpsw->dma);
        Cpsw_assert(status == CPSW_SOK);
    }

    if (hCpsw->hRm != NULL)
    {
        CpswRm_close(hCpsw->hRm);
        hCpsw->hRm = NULL;
    }

    if (hCpsw->dma->initFlag)
    {
        /* Close dma module */
        status = CpswDma_close(hCpsw->dma);
        Cpsw_assert(status == CPSW_SOK);
    }
}

static int32_t Cpsw_validateFlowId(Cpsw_Handle hCpsw,
                                   uint32_t coreKey,
                                   uint32_t startIdx,
                                   uint32_t flowIdx)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    uint32_t p0FlowIdOffset;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &p0FlowIdOffset);
    status = CpswHostPort_ioctl(hCpsw->hostPort,
                                CPSW_HOSTPORT_GET_FLOW_ID_OFFSET,
                                &prms);

    if (status == CPSW_SOK)
    {
        if (startIdx != p0FlowIdOffset)
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    if (status == CPSW_SOK)
    {
        CpswRm_ValidateRxFlowInArgs rmInArgs;

        rmInArgs.coreKey = coreKey;
        rmInArgs.flowIdx = flowIdx;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &rmInArgs);

        status = CpswRm_ioctlPriv(hCpsw->hRm,
                                  CPSW_RM_IOCTL_VALIDATE_RX_FLOW,
                                  &prms);
    }

    return status;
}

static int32_t Cpsw_validateDefaultFlow(Cpsw_Handle hCpsw,
                                        Cpsw_DefaultFlowInfo *defaultFlowInfo,
                                        uint32_t expectedDefaultFlowId)
{
    int32_t status;
    Cpsw_IoctlPrms prms;

    status = Cpsw_validateFlowId(hCpsw,
                                 defaultFlowInfo->coreKey,
                                 defaultFlowInfo->startIdx,
                                 defaultFlowInfo->flowIdx);
    if (status == CPSW_SOK)
    {
        CpswAle_GetDefaultThreadConfigOutArgs getDefaultThreadConfigOutArgs;

        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getDefaultThreadConfigOutArgs);

        status = CpswAle_ioctl(hCpsw->ale,
                               CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG,
                               &prms);
        if (status == CPSW_SOK)
        {
            if ((getDefaultThreadConfigOutArgs.defaultThreadEnable == true)
                &&
                (getDefaultThreadConfigOutArgs.threadId == expectedDefaultFlowId))
            {
                status = CPSW_SOK;
            }
            else
            {
                status = CPSW_EINVALID_PARAMS;
            }
        }
    }

    return status;
}

static int32_t Cpsw_ioctlInternal(Cpsw_Handle hCpsw,
                                  uint32_t coreId,
                                  uint32_t cmd,
                                  Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    switch (cmd)
    {
        case CPSW_IOCTL_GET_VERSION:
        {
            status = Cpsw_checkOutArgs(prms,
                                       sizeof(Cpsw_GetVersionsOutArgs));
            if (status == CPSW_SOK)
            {
                Cpsw_GetVersionsOutArgs *version =
                    (Cpsw_GetVersionsOutArgs *)prms->outArgs;

                status = Cpsw_getVersion(hCpsw, version);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to get CPSW version: %d\n", status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid GET_VERSION params\n");
            }
        }
        break;

        case CPSW_IOCTL_OPEN_PORT_LINK:
        {
            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_OpenPortLinkInArgs));
            if (status == CPSW_SOK)
            {
                Cpsw_OpenPortLinkInArgs *inArgs =
                    (Cpsw_OpenPortLinkInArgs *)prms->inArgs;

                if (inArgs->macConfig.rxMtu > hCpsw->maxPerPrioMtu)
                {
                    /* Host port and MAC port MTU should not be greater than largest of the port
                     * egress per priority MTU as packet would get dropped by Switch.
                     * Though it is valid from HW configuration, we return an error as it serves
                     * no purpose */
                    CPSWTRACE_ERROR("Port %u RX MTU (%d) exceeds max of TX Priority 0-7 MTU(%d)\n ",
                                    inArgs->portNum, inArgs->macConfig.rxMtu, hCpsw->maxPerPrioMtu);
                    status = CPSW_EINVALID_PARAMS;
                }
                else
                {
                    status = Cpsw_openPortLink(hCpsw,
                                               inArgs->portNum,
                                               &inArgs->interface,
                                               &inArgs->linkConfig,
                                               &inArgs->macConfig,
                                               &inArgs->phyConfig);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to open port %u link: %d\n",
                                       inArgs->portNum, status);
                }
            }
            else
            {
                CPSWTRACE_ERROR("invalid OPEN_PORT_LINK params\n");
            }
        }
        break;

        case CPSW_IOCTL_CLOSE_PORT_LINK:
        {
            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_GenericPortLinkInArgs));
            if (status == CPSW_SOK)
            {
                Cpsw_GenericPortLinkInArgs *inArgs =
                    (Cpsw_GenericPortLinkInArgs *)prms->inArgs;

                Cpsw_closePortLink(hCpsw, inArgs->portNum);
            }
            else
            {
                CPSWTRACE_ERROR("invalid CLOSE_PORT_LINK params\n");
            }
        }
        break;

        case CPSW_IOCTL_IS_PORT_LINK_UP:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_GenericPortLinkInArgs),
                                         sizeof(bool));
            if (status == CPSW_SOK)
            {
                Cpsw_GenericPortLinkInArgs *inArgs =
                    (Cpsw_GenericPortLinkInArgs *)prms->inArgs;
                bool *linked = (bool *)prms->outArgs;

                *linked = hCpsw->portLinkState[inArgs->portNum].isLinkUp;
            }
            else
            {
                CPSWTRACE_ERROR("invalid IS_PORT_LINK_UP params\n");
            }
        }
        break;

        case CPSW_IOCTL_GET_PORT_LINK_CONFIG:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_GenericPortLinkInArgs),
                                         sizeof(CpswMacPort_LinkConfig));
            if (status == CPSW_SOK)
            {
                Cpsw_GenericPortLinkInArgs *inArgs =
                    (Cpsw_GenericPortLinkInArgs *)prms->inArgs;
                CpswMacPort_LinkConfig *linkCfg =
                    (CpswMacPort_LinkConfig *)prms->outArgs;

                status = Cpsw_getPortLinkConfig(hCpsw,
                                                inArgs->portNum,
                                                linkCfg);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to get port %u link config: %d\n",
                                   inArgs->portNum, status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid GET_PORT_LINK_CONFIG params\n");
            }
        }
        break;

        case CPSW_IOCTL_SET_HOSTPORT_TRAFFIC_SHAPING:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(Cpsw_SetHostPortTrafficShapingInArgs));
            if (status == CPSW_SOK)
            {
                Cpsw_SetHostPortTrafficShapingInArgs *inArgs =
                    (Cpsw_SetHostPortTrafficShapingInArgs *)prms->inArgs;

                status = Cpsw_setHostPortTrafficShaping(hCpsw,
                                                        &inArgs->trafficShapingConfig,
                                                        inArgs->txBlocksRem);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to set host port traffic shaping: %d\n",
                                   status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid SET_HOSTPORT_TRAFFIC_SHAPING params\n");
            }
        }
        break;

        case CPSW_IOCTL_PRINT_REGISTERS:
        {
            status = Cpsw_printRegisters(hCpsw);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to print regs: %d\n", status);
        }
        break;

        case CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_SetInterVlanRouteUniEgressInArgs),
                                         sizeof(Cpsw_SetInterVlanRouteUniEgressOutArgs));
            if (status == CPSW_SOK)
            {
                const Cpsw_SetInterVlanRouteUniEgressInArgs *inArgs =
                    (const Cpsw_SetInterVlanRouteUniEgressInArgs *)prms->inArgs;
                Cpsw_SetInterVlanRouteUniEgressOutArgs *outArgs =
                    (Cpsw_SetInterVlanRouteUniEgressOutArgs *)prms->outArgs;

                status = Cpsw_setInterVlanRouteUniEgress(hCpsw,
                                                         inArgs,
                                                         outArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to set interVLAN route unicast egress: %d\n",
                                   status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid SET_INTERVLAN_ROUTE_UNI_EGRESS params\n");
            }
        }
        break;

        case CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_UNI_EGRESS:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_ClearInterVlanRouteUniEgressInArgs),
                                         sizeof(Cpsw_ClearInterVlanRouteUniEgressOutArgs));
            if (status == CPSW_SOK)
            {
                const Cpsw_ClearInterVlanRouteUniEgressInArgs *inArgs =
                    (const Cpsw_ClearInterVlanRouteUniEgressInArgs *)prms->inArgs;
                Cpsw_ClearInterVlanRouteUniEgressOutArgs *outArgs =
                    (Cpsw_ClearInterVlanRouteUniEgressOutArgs *)prms->outArgs;

                status = Cpsw_clearInterVlanRouteUniEgress(hCpsw,
                                                           inArgs,
                                                           outArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to clear interVLAN route unicast egress: %d\n",
                                   status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid CLEAR_INTERVLAN_ROUTE_UNI_EGRESS params\n");
            }
        }
        break;

        case CPSW_IOCTL_SET_INTERVLAN_ROUTE_MULTI_EGRESS:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_SetInterVlanRouteMultiEgressInArgs),
                                         sizeof(Cpsw_SetInterVlanRouteMultiEgressOutArgs));
            if (status == CPSW_SOK)
            {
                const Cpsw_SetInterVlanRouteMultiEgressInArgs *inArgs =
                    (const Cpsw_SetInterVlanRouteMultiEgressInArgs *)prms->inArgs;
                Cpsw_SetInterVlanRouteMultiEgressOutArgs *outArgs =
                    (Cpsw_SetInterVlanRouteMultiEgressOutArgs *)prms->outArgs;

                status = Cpsw_setInterVlanRouteMultiEgress(hCpsw,
                                                           inArgs,
                                                           outArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to set interVLAN route multicast egress: %d\n",
                                   status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid SET_INTERVLAN_ROUTE_MULTI_EGRESS params\n");
            }
        }
        break;

        case CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_MULTI_EGRESS:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(Cpsw_ClearInterVlanRouteMultiEgressInArgs),
                                         sizeof(Cpsw_ClearInterVlanRouteMultiEgressOutArgs));
            if (status == CPSW_SOK)
            {
                const Cpsw_ClearInterVlanRouteMultiEgressInArgs *inArgs =
                    (const Cpsw_ClearInterVlanRouteMultiEgressInArgs *)prms->inArgs;
                Cpsw_ClearInterVlanRouteMultiEgressOutArgs *outArgs =
                    (Cpsw_ClearInterVlanRouteMultiEgressOutArgs *)prms->outArgs;

                status = Cpsw_clearInterVlanRouteMultiEgress(hCpsw,
                                                             inArgs,
                                                             outArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to clear interVLAN route multicast egress: %d\n",
                                   status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid CLEAR_INTERVLAN_ROUTE_MULTI_EGRESS params\n");
            }
        }
        break;

        case CPSW_IOCTL_SET_SHORT_IPG_CONFIG:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(Cpsw_SetShortTxIPGConfigInArgs));
            if (status == CPSW_SOK)
            {
                const Cpsw_SetShortTxIPGConfigInArgs *inArgs =
                    (const Cpsw_SetShortTxIPGConfigInArgs *)prms->inArgs;

                status = Cpsw_setTxShortIPGConfig(hCpsw, inArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to set TX short IPG config: %d\n", status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid SET_SHORT_IPG_CONFIG params\n");
            }
        }
        break;

        case CPSW_IOCTL_GET_SHORT_IPG_CONFIG:
        {
            status = Cpsw_checkOutArgs(prms,
                                       sizeof(Cpsw_GetShortTxIPGConfigOutArgs));
            if (status == CPSW_SOK)
            {
                Cpsw_GetShortTxIPGConfigOutArgs *outArgs =
                    (Cpsw_GetShortTxIPGConfigOutArgs *)prms->outArgs;

                status = Cpsw_getTxShortIPGConfig(hCpsw, outArgs);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to get TX short IPG config: %d\n", status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid GET_SHORT_IPG_CONFIG params\n");
            }
        }
        break;

        case CPSW_IOCTL_ATTACH_CORE:
        {
            uint32_t i;
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(Cpsw_AttachCoreOutArgs));
            if (status == CPSW_SOK)
            {
                uint32_t coreId                 = *((uint32_t *)prms->inArgs);
                Cpsw_AttachCoreOutArgs *outArgs = (Cpsw_AttachCoreOutArgs *)prms->outArgs;
                Cpsw_IoctlPrms rmPrms;

                CPSW_IOCTL_SET_INOUT_ARGS(&rmPrms, &coreId, &outArgs->coreKey);
                status = CpswRm_ioctlPriv(hCpsw->hRm,
                                          CPSW_RM_IOCTL_ATTACH,
                                          &rmPrms);

                if (status == CPSW_SOK)
                {
                    /* Get mtu values */
                    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
                    {
                        outArgs->txMtu[i] = Cpsw_getTxMtuPerPriority(hCpsw, i);
                    }

                    outArgs->rxMtu = Cpsw_getRxMtuPort0(hCpsw);
                }
                else
                {
                    CPSWTRACE_ERROR("failed to attach core %u: %d\n", coreId, status);
                }
            }
            else
            {
                CPSWTRACE_ERROR("invalid ATTACH_CORE params\n");
            }
        }
        break;

        case CPSW_IOCTL_DETACH_CORE:
        {
            status = Cpsw_checkInArgs(prms, sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                uint32_t coreKey = *((uint32_t *)prms->inArgs);
                Cpsw_IoctlPrms rmPrms;

                CPSW_IOCTL_SET_IN_ARGS(&rmPrms, &coreKey);

                status = CpswRm_ioctlPriv(hCpsw->hRm,
                                          CPSW_RM_IOCTL_DETACH,
                                          &rmPrms);

                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to detach core: %d\n", status);
            }
            else
            {
                CPSWTRACE_ERROR("invalid DETACH_CORE params\n");
            }
        }
        break;

        case CPSW_IOCTL_REGISTER_RX_DEFAULT_FLOW:
        {
            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_DefaultFlowInfo));
            if (status == CPSW_SOK)
            {
                Cpsw_DefaultFlowInfo *defaultFlowInfo = (Cpsw_DefaultFlowInfo *)prms->inArgs;

                status = Cpsw_validateDefaultFlow(hCpsw, defaultFlowInfo, hCpsw->rsvdFlowId);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid Default Flow: %d\n", defaultFlowInfo->flowIdx);
                if (status == CPSW_SOK)
                {
                    status = Cpsw_setDefaultThreadCfg(hCpsw, defaultFlowInfo->flowIdx);
                }
            }
            else
            {
                CPSWTRACE_ERROR("invalid register default params\n");
            }
        }
        break;

        case CPSW_IOCTL_UNREGISTER_RX_DEFAULT_FLOW:
        {
            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_DefaultFlowInfo));
            if (status == CPSW_SOK)
            {
                Cpsw_DefaultFlowInfo *defaultFlowInfo = (Cpsw_DefaultFlowInfo *)prms->inArgs;

                status = Cpsw_validateDefaultFlow(hCpsw, defaultFlowInfo, defaultFlowInfo->flowIdx);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid Default Flow: %d\n", defaultFlowInfo->flowIdx);
                if (status == CPSW_SOK)
                {
                    status = Cpsw_setDefaultThreadCfg(hCpsw, hCpsw->rsvdFlowId);
                }
            }
            else
            {
                CPSWTRACE_ERROR("invalid register default params\n");
            }
        }
        break;

        case CPSW_IOCTL_REGISTER_DSTMAC_RX_FLOW:
        {
            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_MacDstFlowInfo));
            if (status == CPSW_SOK)
            {
                Cpsw_MacDstFlowInfo *macDstFlowInfo = (Cpsw_MacDstFlowInfo *)prms->inArgs;

                status = Cpsw_validateFlowId(hCpsw, macDstFlowInfo->coreKey, macDstFlowInfo->startIdx, macDstFlowInfo->flowIdx);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid Flow: %d\n", macDstFlowInfo->flowIdx);
                if (status == CPSW_SOK)
                {
                    Cpsw_IoctlPrms alePrms;
                    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
                    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;

                    memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));

                    setPolicerInArgs.policerMatch.policerMatchEnableMask = CPSW_ALE_POLICER_MATCH_MACDST;
                    memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], macDstFlowInfo->macAddress,
                           sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
                    setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
                    setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_HOST_PORT_NUM;
                    setPolicerInArgs.threadIdEnable                        = TRUE;
                    setPolicerInArgs.threadId                              = macDstFlowInfo->flowIdx;
                    setPolicerInArgs.peakRateInBitsPerSec                  = 0;
                    setPolicerInArgs.commitRateInBitsPerSec                = 0;

                    CPSW_IOCTL_SET_INOUT_ARGS(&alePrms, &setPolicerInArgs, &setPolicerOutArgs);

                    status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_SET_POLICER,
                                           &alePrms);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "CPSW_ALE_IOCTL_SET_POLICER failed: %d\n", status);
                    CPSWTRACE_INFO_IF((status == CPSW_SOK),
                                      "CPSW: Registered MAC address.ALE entry:%u, Policer Entry:%u",
                                      setPolicerOutArgs.dstMacAleEntryIndex, setPolicerOutArgs.policerEntryIndex);
                }
            }
            else
            {
                CPSWTRACE_ERROR("invalid register default params\n");
            }
        }
        break;

        case CPSW_IOCTL_UNREGISTER_DSTMAC_RX_FLOW:
        {
            Cpsw_IoctlPrms alePrms;
            CpswAle_DelPolicerEntryInArgs delPolicerInArgs;
            CpswAle_PolicerMatchParams policerMatchPrms;
            CpswAle_GetPolicerEntryOutArgs policerEntryOutArgs;
            Cpsw_MacDstFlowInfo *macDstFlowInfo = (Cpsw_MacDstFlowInfo *)prms->inArgs;

            status = Cpsw_checkInArgs(prms, sizeof(Cpsw_MacDstFlowInfo));
            if (status == CPSW_SOK)
            {
                status = Cpsw_validateFlowId(hCpsw, macDstFlowInfo->coreKey, macDstFlowInfo->startIdx, macDstFlowInfo->flowIdx);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid Flow: %d\n", macDstFlowInfo->flowIdx);
            }

            if (status == CPSW_SOK)
            {
                memset(&policerMatchPrms, 0, sizeof(policerMatchPrms));
                policerMatchPrms.policerMatchEnableMask = CPSW_ALE_POLICER_MATCH_MACDST;
                memcpy(&policerMatchPrms.dstMacAddr.addr.addr[0U], macDstFlowInfo->macAddress,
                       sizeof(policerMatchPrms.dstMacAddr.addr.addr));
                policerMatchPrms.dstMacAddr.addr.vlanId   = 0;
                policerMatchPrms.dstMacAddr.egressPortNum = CPSW_ALE_HOST_PORT_NUM;

                CPSW_IOCTL_SET_INOUT_ARGS(&alePrms, &policerMatchPrms, &policerEntryOutArgs);

                status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_GET_POLICER,
                                       &alePrms);
                if (status == CPSW_SOK)
                {
                    if ((policerEntryOutArgs.threadIdEnable == TRUE) &&
                        (policerEntryOutArgs.threadId == macDstFlowInfo->flowIdx))
                    {
                        status = CPSW_SOK;
                    }
                    else
                    {
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
            }

            if (status == CPSW_SOK)
            {
                memset(&delPolicerInArgs, 0, sizeof(delPolicerInArgs));

                delPolicerInArgs.policerMatch    = policerMatchPrms;
                delPolicerInArgs.delAleEntryMask = CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST;

                CPSW_IOCTL_SET_IN_ARGS(&alePrms, &delPolicerInArgs);

                status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_DEL_POLICER,
                                       &alePrms);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid Flow: %d\n", macDstFlowInfo->flowIdx);
            }
        }
        break;

        default:
        {
            CPSWTRACE_ERROR("invalid IOCTL cmd: %u\n", cmd);
            status = CPSW_EINVALID_PARAMS;
        }
        break;
    }

    return status;
}

static int32_t Cpsw_rmIrqSet(uint32_t dmscDevIdCpsw,
                             uint32_t srcIdx,
                             uint32_t coreIntNum)
{
    int32_t status;
    struct tisci_msg_rm_irq_set_req rmIrqReq;
    struct tisci_msg_rm_irq_set_resp rmIrqResp;

    rmIrqReq.valid_params  = TISCI_MSG_VALUE_RM_DST_ID_VALID;
    rmIrqReq.valid_params |= TISCI_MSG_VALUE_RM_DST_HOST_IRQ_VALID;

    rmIrqReq.global_event          = 0U;
    rmIrqReq.src_id                = dmscDevIdCpsw;
    rmIrqReq.src_index             = srcIdx;
    rmIrqReq.dst_id                = CpswSoc_getCoreDevId();
    rmIrqReq.dst_host_irq          = coreIntNum;
    rmIrqReq.ia_id                 = 0U;
    rmIrqReq.vint                  = 0U;
    rmIrqReq.vint_status_bit_index = 0U;
    rmIrqReq.secondary_host        = TISCI_MSG_VALUE_RM_UNUSED_SECONDARY_HOST;

    status = Sciclient_rmIrqSet(&rmIrqReq, &rmIrqResp, SCICLIENT_SERVICE_WAIT_FOREVER);
    CPSWTRACE_ERROR_IF((status != CSL_PASS),
                       "Error in SciClient Interrupt Params Configuration: %d\n", status);

    return status;
}

static int32_t Cpsw_registerIntrs(Cpsw_Handle hCpsw,
                                  const Cpsw_Config *cfg)
{
    int32_t status = CPSW_SOK;
    const CpswSoc_Config *pSocCfg;
    const CpswSoc_IntrConnCfg *pCptsIntr, *pStatIntr, *pMdioIntr;

    pSocCfg   = CpswSoc_getConfig(hCpsw->cpswType);
    pCptsIntr = &pSocCfg->intrs.evntPend;
    pStatIntr = &pSocCfg->intrs.statPend0;
    pMdioIntr = &pSocCfg->intrs.mdioIntr;

    if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pStatIntr->srcIdx)
    {
        status = Cpsw_rmIrqSet(pSocCfg->dmscDevIdCpsw,
                               pStatIntr->srcIdx,
                               pStatIntr->coreIntNum);
    }

    if (status == CPSW_SOK)
    {
        /* Register statistics interrupt */
        hCpsw->hStatsIntr = CpswOsal_registerIntr(Cpsw_statsIsr,
                                                  pStatIntr->coreIntNum,
                                                  cfg->intrPriority,
                                                  hCpsw->stats);
        if (hCpsw->hStatsIntr == NULL)
        {
            CPSWTRACE_ERROR("failed to register STATS intr\n");
            if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pStatIntr->srcIdx)
            {
                status = Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                                           pStatIntr->srcIdx,
                                           pStatIntr->coreIntNum);
                /* failure in failure, what can we do but assert */
                Cpsw_assert(status == CPSW_SOK);
            }

            status = CPSW_EFAIL;
        }
    }

    /* Register MDIO interrupt */
    if ((status == CPSW_SOK) &&
        (hCpsw->mdioLinkIntCtx.linkStateChangeCb != NULL))
    {
        if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pMdioIntr->srcIdx)
        {
            status = Cpsw_rmIrqSet(pSocCfg->dmscDevIdCpsw,
                                   pMdioIntr->srcIdx,
                                   pMdioIntr->coreIntNum);
        }

        if (status == CPSW_SOK)
        {
            hCpsw->hMdioIntr = CpswOsal_registerIntr(Cpsw_mdioIsr,
                                                     pMdioIntr->coreIntNum,
                                                     cfg->intrPriority,
                                                     hCpsw);
            if (hCpsw->hMdioIntr == NULL)
            {
                CPSWTRACE_ERROR("failed to register MDIO intr\n");
                if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pMdioIntr->srcIdx)
                {
                    status = Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                                               pMdioIntr->srcIdx,
                                               pMdioIntr->coreIntNum);
                    /* failure in failure, what can we do but assert */
                    Cpsw_assert(status == CPSW_SOK);
                }

                status = CPSW_EFAIL;
            }
        }
    }

    /* Register CPTS interrupt */
    if (status == CPSW_SOK)
    {
        if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pCptsIntr->srcIdx)
        {
            status = Cpsw_rmIrqSet(pSocCfg->dmscDevIdCpsw,
                                   pCptsIntr->srcIdx,
                                   pCptsIntr->coreIntNum);
        }

        if (status == CPSW_SOK)
        {
            hCpsw->hCptsIntr = CpswOsal_registerIntr(Cpsw_cptsIsr,
                                                     pCptsIntr->coreIntNum,
                                                     cfg->intrPriority,
                                                     hCpsw);
            if (hCpsw->hCptsIntr == NULL)
            {
                CPSWTRACE_ERROR("failed to register CPTS intr\n");
                if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pCptsIntr->srcIdx)
                {
                    status = Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                                               pCptsIntr->srcIdx,
                                               pCptsIntr->coreIntNum);
                    /* failure in failure, what can we do but assert */
                    Cpsw_assert(status == CPSW_SOK);
                }

                status = CPSW_EFAIL;
            }
        }
    }

    return status;
}

static int32_t Cpsw_rmIrqRelease(uint32_t dmscDevIdCpsw,
                                 uint32_t srcIdx,
                                 uint32_t coreIntNum)
{
    int32_t status;
    struct tisci_msg_rm_irq_release_req rmIrqRel;

    rmIrqRel.valid_params  = TISCI_MSG_VALUE_RM_DST_ID_VALID;
    rmIrqRel.valid_params |= TISCI_MSG_VALUE_RM_DST_HOST_IRQ_VALID;

    rmIrqRel.global_event          = 0U;
    rmIrqRel.src_id                = dmscDevIdCpsw;
    rmIrqRel.src_index             = srcIdx;
    rmIrqRel.dst_id                = CpswSoc_getCoreDevId();
    rmIrqRel.dst_host_irq          = coreIntNum;
    rmIrqRel.ia_id                 = 0U;
    rmIrqRel.vint                  = 0U;
    rmIrqRel.vint_status_bit_index = 0U;
    rmIrqRel.secondary_host        = TISCI_MSG_VALUE_RM_UNUSED_SECONDARY_HOST;

    status = Sciclient_rmIrqRelease(&rmIrqRel, SCICLIENT_SERVICE_WAIT_FOREVER);
    CPSWTRACE_ERROR_IF((status != CSL_PASS),
                       "SciClient Interrupt Release failed: %d\n", status);

    return status;
}

static void Cpsw_unregisterIntrs(Cpsw_Handle hCpsw)
{
    const CpswSoc_Config *pSocCfg;
    const CpswSoc_IntrConnCfg *pStatIntr, *pMdioIntr, *pCptsIntr;

    pSocCfg   = CpswSoc_getConfig(hCpsw->cpswType);
    pStatIntr = &pSocCfg->intrs.statPend0;
    pMdioIntr = &pSocCfg->intrs.mdioIntr;
    pCptsIntr = &pSocCfg->intrs.evntPend;

    /* Unregister MDIO interrupt */
    if (hCpsw->hMdioIntr != NULL)
    {
        CpswOsal_unRegisterIntr(hCpsw->hMdioIntr);
        if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pMdioIntr->srcIdx)
        {
            Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                              pMdioIntr->srcIdx,
                              pMdioIntr->coreIntNum);
        }

        hCpsw->hMdioIntr = NULL;
    }

    /* Unregister statistics interrupt */
    if (hCpsw->hStatsIntr != NULL)
    {
        if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pStatIntr->srcIdx)
        {
            Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                              pStatIntr->srcIdx,
                              pStatIntr->coreIntNum);
        }

        CpswOsal_unRegisterIntr(hCpsw->hStatsIntr);
        hCpsw->hStatsIntr = NULL;
    }
    /* Unregister CPTS interrupt */
    if (hCpsw->hCptsIntr != NULL)
    {
        if (CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID != pCptsIntr->srcIdx)
        {
            Cpsw_rmIrqRelease(pSocCfg->dmscDevIdCpsw,
                              pCptsIntr->srcIdx,
                              pCptsIntr->coreIntNum);
        }

        CpswOsal_unRegisterIntr(hCpsw->hCptsIntr);
        hCpsw->hCptsIntr = NULL;
    }
}

static void Cpsw_statsIsr(uintptr_t arg)
{
    CpswStats_Handle hStats = (CpswStats_Handle)arg;
    int32_t status;

    Cpsw_assert(hStats != NULL);

    status = CpswStats_ioctl(hStats, CPSW_STATS_IOCTL_SYNC, NULL);

    /* TODO: Add ISR safe error:
     * failed to sync statistics counters: %d\n", status); */
    Cpsw_assert(status == CPSW_SOK);
}

static void Cpsw_mdioIsr(uintptr_t arg)
{
    Cpsw_Handle hCpsw     = (Cpsw_Handle)arg;
    CpswMdio_Handle hMdio = hCpsw->mdio;
    Cpsw_IoctlPrms prms;
    CpswMdio_Callbacks callbacks =
    {
        .linkStateCb  = Cpsw_handleMdioLinkStateChange,
        .userAccessCb = NULL,
        .appData      = hCpsw,
    };
    int32_t status;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &callbacks);
    status = CpswMdio_ioctl(hMdio, CPSW_MDIO_IOCTL_HANDLE_INTR, &prms);

    /* TODO: Add ISR safe error:
     * failed to handle MDIO intr: %d\n", status); */
    Cpsw_assert(status == CPSW_SOK);
}

static void Cpsw_cptsIsr(uintptr_t arg)
{
    int32_t status;

    Cpsw_Handle hCpsw     = (Cpsw_Handle)arg;
    CpswCpts_Handle hCpts = hCpsw->cpts;

    Cpsw_assert(hCpts != NULL);

    status = CpswCpts_ioctl(hCpts, CPSW_CPTS_IOCTL_HANDLE_INTR, NULL);

    /* TODO: Add ISR safe error:
     * failed to handle CPTS intr: %d\n", status); */
    Cpsw_assert(status == CPSW_SOK);
}

static void Cpsw_handleMdioLinkStateChange(CpswMdio_UserChannel userCh,
                                           CpswMdio_PhyStatus *phyStatus,
                                           void *appData)
{
    Cpsw_Handle hCpsw               = (Cpsw_Handle)appData;
    Cpsw_MdioLinkIntCtx *linkIntCtx = &hCpsw->mdioLinkIntCtx;
    Cpsw_PortLinkState *portLinkState;
    Cpsw_MdioLinkStateChangeInfo info;
    uint32_t aliveMaskChange;
    uint32_t linkedMaskChange;
    uint32_t i;

    aliveMaskChange  = phyStatus->aliveMask ^ linkIntCtx->aliveMask;
    linkedMaskChange = phyStatus->linkedMask ^ linkIntCtx->linkedMask;

    for (i = 0U; i <= CPSW_MDIO_MAX_PHY_CNT; i++)
    {
        info.aliveChanged = CPSW_IS_BIT_SET(aliveMaskChange, i);
        info.linkChanged  = CPSW_IS_BIT_SET(linkedMaskChange, i);

        if (info.aliveChanged || info.linkChanged)
        {
            info.phyAddr  = i;
            info.isAlive  = CPSW_IS_BIT_SET(phyStatus->aliveMask, i);
            info.isLinked = CPSW_IS_BIT_SET(phyStatus->linkedMask, i);

            /* Re-enable periodic tick when link down is detected */
            if (info.linkChanged && !info.isLinked)
            {
                portLinkState = Cpsw_getPortLinkState(hCpsw, i);
                if (portLinkState != NULL)
                {
                    portLinkState->isTickEnabled = true;
                }
            }

            linkIntCtx->linkStateChangeCb(&info,
                                          linkIntCtx->linkStateChangeCbArg);
        }
    }

    linkIntCtx->aliveMask  = phyStatus->aliveMask;
    linkIntCtx->linkedMask = phyStatus->linkedMask;
}

static int32_t Cpsw_printRegisters(Cpsw_Handle hCpsw)
{
    uint32_t i;
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    /* TODO: Print CPSW SS + CPSW NU registers */

    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = CpswAle_ioctl(hCpsw->ale,
                               CPSW_ALE_IOCTL_PRINT_REGS,
                               &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to print ALE regs: %d\n", status);
    }

    if (status == CPSW_SOK)
    {
        status = CpswHostPort_ioctl(hCpsw->hostPort,
                                    CPSW_HOSTPORT_IOCTL_PRINT_REGS,
                                    NULL);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to print host port regs: %d\n", status);
    }

    if (status == CPSW_SOK)
    {
        for (i = 0U; i < Cpsw_getMacPortMax(hCpsw->cpswType); i++)
        {
            if (hCpsw->macPort[i] != NULL)
            {
                CpswMacPort_GenericInArgs inArgs;

                inArgs.portNum = hCpsw->macPort[i]->portNum;
                CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs);
                status = CpswMacPort_ioctl(hCpsw->macPort[i],
                                           CPSW_MACPORT_IOCTL_PRINT_REGS,
                                           &prms);
                if (status != CPSW_SOK)
                {
                    CPSWTRACE_ERROR("failed to print MAC port %u regs: %d\n",
                                    i, status);
                    break;
                }
            }
        }
    }

    if (status == CPSW_SOK)
    {
        status = CpswMdio_ioctl(hCpsw->mdio,
                                CPSW_MDIO_IOCTL_PRINT_REGS,
                                NULL);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to print MDIO regs: %d\n", status);
    }

    if (status == CPSW_SOK)
    {
        status = CpswCpts_ioctl(hCpsw->cpts,
                                CPSW_CPTS_IOCTL_PRINT_REGS,
                                NULL);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to print CPTS regs: %d\n", status);
    }

    return status;
}

static int32_t Cpsw_getVersion(Cpsw_Handle hCpsw,
                               Cpsw_GetVersionsOutArgs *version)
{
    CSL_CPSW_SS_VERSION ssVer;
    CSL_CPSW_VERSION ver;
    Cpsw_IoctlPrms prms;
    int32_t status = CPSW_SOK;

    /* CPSW subsystem version */
    CSL_CPSW_SS_getVersionInfo(hCpsw->cpswSsRegs, &ssVer);
    version->ss.major = ssVer.majorVer;
    version->ss.minor = ssVer.minorVer;
    version->ss.rtl   = ssVer.rtlVer;
    version->ss.id    = ssVer.id;

    /* CPSW module version */
    CSL_CPSW_getCpswVersionInfo(hCpsw->cpswRegs, &ver);
    version->cpsw.major = ver.majorVer;
    version->cpsw.minor = ver.minorVer;
    version->cpsw.rtl   = ver.rtlVer;
    version->cpsw.id    = ver.id;

    if (CPSW_SOK == status)
    {
        /* CPTS version */
        CPSW_IOCTL_SET_OUT_ARGS(&prms, &version->cpts);
        status = CpswCpts_ioctl(hCpsw->cpts,
                                CPSW_CPTS_IOCTL_GET_VERSION,
                                &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to get CPTS version: %d\n", status);
    }

    if (CPSW_SOK == status)
    {
        /* MDIO version */
        CPSW_IOCTL_SET_OUT_ARGS(&prms, &version->mdio);
        status = CpswMdio_ioctl(hCpsw->mdio,
                                CPSW_MDIO_IOCTL_GET_VERSION,
                                &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to get MDIO version: %d\n", status);
    }

    /* TODO: Add ALE versions */

    return status;
}

static int32_t Cpsw_setHostPortTrafficShaping(Cpsw_Handle hCpsw,
                                              CpswPort_TrafficShapingConfig *shapingCfg,
                                              uint32_t txBlocksRem)
{
    Cpsw_IoctlPrms portPrms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    CPSW_IOCTL_SET_IN_ARGS(&portPrms, shapingCfg);
    status = CpswHostPort_ioctl(hCpsw->hostPort,
                                CPSW_HOSTPORT_IOCTL_SET_TRAFFIC_SHAPING,
                                &portPrms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to set host port traffic shaping: %d\n", status);

    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_IN_ARGS(&portPrms, &txBlocksRem);

        for (i = 0U; i < Cpsw_getMacPortMax(hCpsw->cpswType); i++)
        {
            if (hCpsw->macPort[i] != NULL)
            {
                status = CpswMacPort_ioctl(hCpsw->macPort[i],
                                           CPSW_MACPORT_IOCTL_SET_TX_BLOCKS_REM,
                                           &portPrms);
                if (status != CPSW_SOK)
                {
                    CPSWTRACE_ERROR("failed to set MAC port %u TX blocks rem: %d\n",
                                    i, status);
                    break;
                }
            }
        }
    }

    return status;
}

static uint32_t Cpsw_mapPort2XgmiiId(uint32_t portNum)
{
    /* TODO: Needs correct logic to translate portNum to QSGMII Id */
    return 0;
}

static int32_t Cpsw_getPortLinkConfig(Cpsw_Handle hCpsw,
                                      Cpsw_MacPort portNum,
                                      CpswMacPort_LinkConfig *linkCfg)
{
    CpswMac_LayerType enetLayer;
    CpswMac_SublayerType enetSublayer;
    int32_t status;

    status = CpswSoc_getEnetControl(hCpsw->cpswType,
                                    portNum,
                                    &enetLayer,
                                    &enetSublayer);
    if (status == CPSW_SOK)
    {
        /* Link configuration is only available for RGMII */
        if ((enetLayer == CPSW_MAC_LAYER_GMII) &&
            (enetSublayer == CPSW_MAC_SUBLAYER_REDUCED))
        {
            CSL_CPSW_SS_RGMIISTATUS rgmiiStatus;

            CSL_CPSW_SS_getRGMIIStatus(hCpsw->cpswSsRegs,
                                       portNum,
                                       &rgmiiStatus);

            /* Get link speed */
            if (rgmiiStatus.speed == 0U)
            {
                linkCfg->speed = CPSW_SPEED_10MBIT;
            }
            else if (rgmiiStatus.speed == 1U)
            {
                linkCfg->speed = CPSW_SPEED_100MBIT;
            }
            else if (rgmiiStatus.speed == 2U)
            {
                linkCfg->speed = CPSW_SPEED_1GBIT;
            }
            else
            {
                CPSWTRACE_ERROR("invalid RGMII speed: %d\n", rgmiiStatus.speed);
                Cpsw_assert(FALSE);
            }

            /* Get link duplexity */
            if (rgmiiStatus.fullDuplex == 0U)
            {
                linkCfg->duplexity = CPSW_DUPLEX_HALF;
            }
            else
            {
                linkCfg->duplexity = CPSW_DUPLEX_FULL;
            }
        }
        else
        {
            CPSWTRACE_ERROR("link config not available for layer %u sublayer %u\n",
                            enetLayer, enetSublayer);
            status = CPSW_ENOTSUPPORTED;
        }
    }
    else
    {
        CPSWTRACE_ERROR("failed to get ENET_CTRL %u: %d\n", portNum, status);
    }

    return status;
}

static int32_t Cpsw_isPortLinkUp(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum,
                                 bool *linked)
{
    CpswMacPort_Handle hPort = hCpsw->macPort[portNum];
    Cpsw_IoctlPrms prms;
    CpswMacPort_GenericInArgs macPortInArgs;
    CpswMac_LayerType enetLayer;
    CpswMac_SublayerType enetSublayer;
    CSL_CPSW_SS_RGMIISTATUS rgmiiStatus;
    bool inBand, sgmiiLink;
    int32_t status;
    uint32_t qsgmiiId;

    macPortInArgs.portNum = portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macPortInArgs, &inBand);
    status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_IS_INBAND, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to get MAC port %d in-band state: %d\n",
                       portNum, status);

    if (status == CPSW_SOK)
    {
        status = CpswSoc_getEnetControl(hCpsw->cpswType,
                                        portNum,
                                        &enetLayer,
                                        &enetSublayer);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to get MAC port %d ENET_CTRL: %d\n",
                           portNum, status);
    }

    if (status == CPSW_SOK)
    {
        if (enetLayer == CPSW_MAC_LAYER_GMII)
        {
            if (enetSublayer == CPSW_MAC_SUBLAYER_REDUCED)
            {
                /* Link status indicator is valid only in in-band mode.
                 * In forced mode, it always reports linked so to prevent
                 * misleading the caller, return not supported error code */
                if (inBand)
                {
                    CSL_CPSW_SS_getRGMIIStatus(hCpsw->cpswSsRegs,
                                               portNum,
                                               &rgmiiStatus);
                    *linked = rgmiiStatus.link;
                }
                else
                {
                    CPSWTRACE_DEBUG("RGMII link status not used in forced mode\n");
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            else if (enetSublayer == CPSW_MAC_SUBLAYER_SERIAL)
            {
                /* We have link status in CPSW subsystem regs as well as in SGMII regs,
                 * check SGMII regs as link status in CPSW subsystem regs is only for 
                 * QSGMII link status */
                macPortInArgs.portNum = portNum;
                CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macPortInArgs, &sgmiiLink);
                status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_GET_SGMII_LINK_STATUS, &prms);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to get SGMII port %d link state: %d\n",
                                   portNum, status);
                *linked = sgmiiLink;
            }
            else if ( (enetSublayer == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN)||
                      (enetSublayer == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB) )
            {
                status = CpswSoc_mapPort2QsgmiiId(hCpsw->cpswType, portNum, &qsgmiiId);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Map QSGMII Id failed for port %d: %d\n", portNum, status);
                Cpsw_assert(status == CPSW_SOK);

                if (CPSW_SOK == status)
                {
                    *linked = CSL_CPSW_SS_getQSGMIIStatusRxSync(hCpsw->cpswSsRegs,
                                                                qsgmiiId);

                    /* Check SGMII link status as well for QSGMII as it is possible that rx sync is set
                     * because of single port */
                    macPortInArgs.portNum = portNum;
                    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macPortInArgs, &sgmiiLink);
                    status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_GET_SGMII_AN_LINK_STATUS, &prms);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to get SGMII port %d link state: %d\n",
                                       portNum, status);
                    CPSWTRACE_WARN_IF((!sgmiiLink), "SGMII port %d link not up\n", portNum);
                    if (!sgmiiLink)
                    {
                        *linked = false;
                    }
                }
            }
            else
            {
                CPSWTRACE_DEBUG("sublayer %u doesn't support link status\n", enetSublayer);
                status = CPSW_ENOTSUPPORTED;
            }
        }
        else if (enetLayer == CPSW_MAC_LAYER_XGMII)
        {
            *linked = CSL_CPSW_SS_getXGMIILinkStatus(hCpsw->cpswSsRegs,
                                                     Cpsw_mapPort2XgmiiId(portNum));
        }
        else
        {
            CPSWTRACE_DEBUG("layer %u doesn't support link status\n", enetLayer);
            status = CPSW_ENOTSUPPORTED;
        }
    }

    return status;
}

static int32_t Cpsw_setSgmiiMode(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Interface *interface,
                                 CpswMacPort_Config *macConfig)
{
    int32_t status = CPSW_EINVALID_PARAMS;

    if ((interface->sublayerType == CPSW_MAC_SUBLAYER_SERIAL) &&
        (0U != (CPSW_FEATURE_SGMII & CpswUtils_getIpFeatureBitMap(hCpsw->cpswType))))
    {
        status = CPSW_SOK;
    }
    else if (((interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN) ||
              (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB))
             &&
             (0U != (CPSW_FEATURE_QSGMII & CpswUtils_getIpFeatureBitMap(hCpsw->cpswType))))
    {
        status = CPSW_SOK;
    }
    else
    {
        /* Invalid configuration */
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        if (macConfig->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_FIBER_WITH_PHY)
        {
            CSL_CPSW_SS_setSGMIIMode(hCpsw->cpswSsRegs, portNum, CSL_SGMII_MODE_FIBER);
        }
        else
        {
            CSL_CPSW_SS_setSGMIIMode(hCpsw->cpswSsRegs, portNum, CSL_SGMII_MODE_SGMII);
        }
    }

    return status;
}

static int32_t Cpsw_openPortLinkWithPhy(Cpsw_Handle hCpsw,
                                        Cpsw_MacPort portNum,
                                        CpswMacPort_Interface *interface,
                                        CpswMacPort_LinkConfig *linkConfig,
                                        CpswMacPort_Config *macConfig,
                                        CpswPhy_Config *phyConfig)
{
    int32_t status = CPSW_SOK;

    /* Open MAC port */
    hCpsw->macPort[portNum] = CpswMacPort_open(hCpsw->cpswType,
                                               portNum,
                                               macConfig,
                                               interface,
                                               linkConfig);
    if (hCpsw->macPort[portNum] == NULL)
    {
        CPSWTRACE_ERROR("failed to open MAC port %u\n", portNum);
        status = CPSW_EFAIL;
    }

    /* Open PHY */
    if (status == CPSW_SOK)
    {
        hCpsw->phy[portNum] = CpswPhy_open(hCpsw->cpswType,
                                           hCpsw->mdio,
                                           portNum,
                                           interface,
                                           phyConfig,
                                           linkConfig);
        if (hCpsw->phy[portNum] == NULL)
        {
            CPSWTRACE_ERROR("failed to open PHY for MAC port %u\n", portNum);
            CpswMacPort_close(hCpsw->macPort[portNum]);
            hCpsw->macPort[portNum] = NULL;
            status                  = CPSW_EFAIL;
        }
    }

    return status;
}

static int32_t Cpsw_openPortLinkNoPhy(Cpsw_Handle hCpsw,
                                      Cpsw_MacPort portNum,
                                      CpswMacPort_Interface *interface,
                                      CpswMacPort_LinkConfig *linkConfig,
                                      CpswMacPort_Config *macConfig)
{
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    int32_t status = CPSW_SOK;

    hCpsw->phy[portNum] = NULL;

    /* Explicit speed and duplexity must be provided, can't be discovered */
    if ((linkConfig->speed == CPSW_SPEED_AUTO) ||
        (linkConfig->duplexity == CPSW_DUPLEX_AUTO))
    {
        CPSWTRACE_ERROR("auto-speed/duplexity is not valid for PHY-less links\n");
        status = CPSW_EINVALID_PARAMS;
    }

    /* Open MAC port in force mode */
    if (status == CPSW_SOK)
    {
        if ((interface->layerType == CPSW_MAC_LAYER_GMII) &&
            (interface->sublayerType == CPSW_MAC_SUBLAYER_REDUCED) &&
            (interface->variantType != CPSW_MAC_VARIANT_FORCED))
        {
            CPSWTRACE_WARN("override variant to force type\n");
            interface->variantType = CPSW_MAC_VARIANT_FORCED;

            if ((macConfig->enableLoopback == true) &&
                (linkConfig->speed != CPSW_SPEED_1GBIT))
            {
                /* PDK-4633 - RGMII loopback fails in 100Mbps mode  */
                CPSWTRACE_WARN("RGMII loopback in 1G mode not supported, "
                               "overriding speed to 1Gbps\n");
                linkConfig->speed = CPSW_SPEED_1GBIT;
            }
        }

        hCpsw->macPort[portNum] = CpswMacPort_open(hCpsw->cpswType,
                                                   portNum,
                                                   macConfig,
                                                   interface,
                                                   linkConfig);
        if (hCpsw->macPort[portNum] == NULL)
        {
            CPSWTRACE_ERROR("failed to open MAC port %u\n", portNum);
            status = CPSW_EFAIL;
        }
    }

    /* Enable MAC port with requested speed/duplexity */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_IN_ARGS(&prms, linkConfig);
        status = CpswMacPort_ioctl(hCpsw->macPort[portNum], CPSW_MACPORT_IOCTL_ENABLE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to enable MAC port: %d\n", status);
    }

    /* Set ALE port in forward state */
    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_MACPORT_TO_ALEPORT(portNum);
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);

        status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_SET_PORT_STATE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set ALE port %u to forward state: %d\n",
                           setPortStateInArgs.portNum, status);
    }

    if (status != CPSW_SOK)
    {
        Cpsw_closePortLink(hCpsw, portNum);
    }

    return status;
}

static int32_t Cpsw_openPortLink(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Interface *interface,
                                 CpswMacPort_LinkConfig *linkConfig,
                                 CpswMacPort_Config *macConfig,
                                 CpswPhy_Config *phyConfig)
{
    Cpsw_PortLinkState *portLinkState = &hCpsw->portLinkState[portNum];
    uint32_t pollEnableMask           = hCpsw->mdioLinkIntCtx.pollEnableMask;
    uint32_t phyAddr                  = phyConfig->phyAddr;
    int32_t status                    = CPSW_SOK;

    if ( (interface->layerType == CPSW_MAC_LAYER_GMII)
          &&
         ((interface->sublayerType == CPSW_MAC_SUBLAYER_SERIAL) ||
          (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN)||
          (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB)) )
    {
        /* Configure SGMII mode in CPSW SS */
        status = Cpsw_setSgmiiMode(hCpsw, portNum, interface, macConfig);

        if ((interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN)||
            (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB))
        {
            /* Make sure the QSGMII ports are configured correctly */
            status += CpswSoc_validateQsgmiiCfg(hCpsw->cpswType);
        }
    }

    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "Invalid SGMII/QSGMII port %u configuration: %d\n", portNum, status);

    if ((portNum >= Cpsw_getMacPortMax(hCpsw->cpswType)) ||
        (hCpsw->macPort[portNum] != NULL) ||
        (hCpsw->phy[portNum] != NULL))
    {
        CPSWTRACE_ERROR("failed to open port %u (MAC port %s, PHY %s)\n",
                        portNum,
                        (hCpsw->macPort[portNum] != NULL) ? "open" : "closed",
                        (hCpsw->phy[portNum] != NULL) ? "open" : "closed");
        status = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == status)
    {
        if (phyAddr == CPSW_PHY_INVALID_PHYADDR)
        {
            status = Cpsw_openPortLinkNoPhy(hCpsw,
                                            portNum,
                                            interface,
                                            linkConfig,
                                            macConfig);
        }
        else
        {
            status = Cpsw_openPortLinkWithPhy(hCpsw,
                                              portNum,
                                              interface,
                                              linkConfig,
                                              macConfig,
                                              phyConfig);
        }
    }

    if (status == CPSW_SOK)
    {
        portLinkState->isOpen  = true;
        portLinkState->phyAddr = phyAddr;

        if ((hCpsw->hMdioIntr != NULL) &&
            CPSW_PHY_IS_ADDR_VALID(phyAddr))
        {
            portLinkState->isPollEnabled = CPSW_MDIO_IS_PHY_ADDR_SET(pollEnableMask,
                                                                     phyAddr);
        }
    }

    return status;
}

static void Cpsw_closePortLink(Cpsw_Handle hCpsw,
                               Cpsw_MacPort portNum)
{
    bool linked;

    if (portNum < Cpsw_getMacPortMax(hCpsw->cpswType))
    {
        if (NULL == hCpsw->phy[portNum])
        {
            /* phy handle is NULL in NOPHY mode, linked is always true */
            linked = true;
        }
        else
        {
            linked = hCpsw->portLinkState[portNum].isLinkUp;
        }

        /* Force link-down handling to undo settings done for this port link
         * i.e. ALE port state, MAC port disable, etc */
        if (linked)
        {
            Cpsw_handleLinkDown(hCpsw, portNum);
        }

        if (hCpsw->macPort[portNum] != NULL)
        {
            CpswMacPort_close(hCpsw->macPort[portNum]);
            hCpsw->macPort[portNum] = NULL;
        }

        if (hCpsw->phy[portNum] != NULL)
        {
            CpswPhy_close(hCpsw->phy[portNum]);
            hCpsw->phy[portNum] = NULL;
        }

        hCpsw->portLinkState[portNum].isOpen        = false;
        hCpsw->portLinkState[portNum].isPollEnabled = false;
        hCpsw->portLinkState[portNum].isTickEnabled = true;
    }
}

static Cpsw_PortLinkState *Cpsw_getPortLinkState(Cpsw_Handle hCpsw,
                                                 uint32_t phyAddr)
{
    Cpsw_PortLinkState *portLinkState = NULL;
    uint32_t i;

    for (i = 0U; i < CPSW_MAC_PORT_NUM; i++)
    {
        if (hCpsw->portLinkState[i].isOpen &&
            (phyAddr == hCpsw->portLinkState[i].phyAddr))
        {
            portLinkState = &hCpsw->portLinkState[i];
            break;
        }
    }

    return portLinkState;
}

static int32_t Cpsw_runPhyTick(Cpsw_Handle hCpsw,
                               Cpsw_MacPort portNum,
                               bool *linkChanged,
                               bool *linked)
{
    CpswPhy_Handle hPhy = hCpsw->phy[portNum];
    Cpsw_IoctlPrms prms;
    CpswPhy_GenericInArgs phyInArgs =
    {
        .portNum = portNum,
    };
    uint32_t mdioStatus;
    int32_t status;

    *linkChanged = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyInArgs, &mdioStatus);
    status = CpswPhy_ioctl(hPhy, CPSW_PHY_IOCTL_PHY_TICK, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "port %d's PHY tick: %d\n", portNum, status);

    /* Check for errors during PHY driver tick */
    if (status == CPSW_SOK)
    {
        if (mdioStatus == CPSW_PHY_EVENT_PHYERROR)
        {
            CPSWTRACE_ERROR("port %d's PHY tick detected error event\n", portNum);
            status = CPSW_EFAIL;
        }
    }

    /* Check that link state changed */
    if (status == CPSW_SOK)
    {
        if ((mdioStatus == CPSW_PHY_EVENT_LINKDOWN) ||
            (mdioStatus == CPSW_PHY_EVENT_LINKUP))
        {
            *linkChanged = true;

            /* Get PHY link state (linked or not) */
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyInArgs, linked);
            status = CpswPhy_ioctl(hPhy, CPSW_PHY_IOCTL_PHY_GET_LINK_STATE, &prms);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to get port %d's PHY link state: %d\n",
                               portNum, status);
        }
    }

    return status;
}

static int32_t Cpsw_matchRgmiiInBandCfg(CpswMacPort_Handle hPort,
                                        Cpsw_MacPort portNum,
                                        CpswMacPort_LinkConfig phyLinkCfg)
{
    int32_t status;
    bool inBand;
    bool match;
    CpswMacPort_GenericInArgs macPortInArgs;

    macPortInArgs.portNum = portNum;
    CpswMacPort_LinkConfig macLinkCfg;
    Cpsw_IoctlPrms prms;

    /* Check if MAC port is in forced mode or in-band - only applies to RGMII mode */
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macPortInArgs, &inBand);
    status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_IS_INBAND, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to get MAC port %d in-band state: %d\n",
                       portNum, status);

    if ((status == CPSW_SOK) && inBand)
    {
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macPortInArgs, &macLinkCfg);
        status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_GET_LINK_CONFIG, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "port %d in-band link params check failed\n", portNum);

        if (status == CPSW_SOK)
        {
            /* Check that link duplexity matches */
            match = (macLinkCfg.duplexity == phyLinkCfg.duplexity);

            /* Check that link speed matches */
            if (phyLinkCfg.speed == CPSW_SPEED_10MBIT)
            {
                /* MAC port link status doesn't differentiate between 10 and
                 * 100-Mbps, both are reported as 100-Mbps */
                match &= (macLinkCfg.speed == CPSW_SPEED_100MBIT);
            }
            else
            {
                match &= (macLinkCfg.speed == phyLinkCfg.speed);
            }

            CPSWTRACE_ERROR_IF(!match,
                               "port %u mode mismatch:\n  MAC = %s %s\n  PHY = %s %s\n",
                               portNum,
                               Cpsw_gSpeedNames[macLinkCfg.speed],
                               Cpsw_gDuplexNames[macLinkCfg.duplexity],
                               Cpsw_gSpeedNames[phyLinkCfg.speed],
                               Cpsw_gDuplexNames[phyLinkCfg.duplexity]);
            Cpsw_assert(match);
        }
    }

    return status;
}

static int32_t Cpsw_handleLinkUp(Cpsw_Handle hCpsw,
                                 Cpsw_MacPort portNum)
{
    CpswMacPort_Handle hPort = hCpsw->macPort[portNum];
    CpswPhy_Handle hPhy = hCpsw->phy[portNum];
    Cpsw_IoctlPrms prms;
    CpswPhy_GenericInArgs phyInArgs;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    CpswMacPort_LinkConfig phyLinkCfg;
    bool isPortLinked = false;
    int32_t status;

    phyInArgs.portNum = portNum;
    /* Check that port status also detected link up */
    do {
        /* Need to repeatedly check port status as it does not reflect
         * PHY status immediately
         */
        status = Cpsw_isPortLinkUp(hCpsw, portNum, &isPortLinked);
    } while ((CPSW_SOK == status) && (isPortLinked == false));

    if (status == CPSW_SOK)
    {
        CPSWTRACE_ERROR_IF(!isPortLinked,
                           "port %d's link indicator detected link down\n", portNum);
        Cpsw_assert(isPortLinked);
    }
    else if (status == CPSW_ENOTSUPPORTED)
    {
        /* It's not an error if port interface doesn't support link indicator */
        status = CPSW_SOK;
    }
    CPSWTRACE_ERROR_IF((status != CPSW_SOK), "failed to get port %d's link indicator: %d\n", portNum, status);

    /* Get link parameters (speed/duplexity) from PHY state machine */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyInArgs, &phyLinkCfg);
        status = CpswPhy_ioctl(hPhy, CPSW_PHY_IOCTL_PHY_GET_LINK_MODE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to get port %d's PHY state: %d\n", portNum, status);
    }

    /* Enable MAC port */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_IN_ARGS(&prms, &phyLinkCfg);
        status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_ENABLE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to enable MAC port %d: %d\n", portNum, status);
    }

    /* Check that PHY and MAC link configuration parameters match if in in-band mode */
    if (status == CPSW_SOK)
    {
        status = Cpsw_matchRgmiiInBandCfg(hPort, portNum, phyLinkCfg);
    }

    /* Set ALE port state to 'Forward' */
    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_MACPORT_TO_ALEPORT(portNum);
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);

        status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_SET_PORT_STATE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set ALE port %u to forward state: %d\n",
                           setPortStateInArgs.portNum, status);
    }

    CPSWTRACE_INFO_IF((status == CPSW_SOK),
                      "port %d: Link up: %s %s\n",
                      portNum,
                      Cpsw_gSpeedNames[phyLinkCfg.speed],
                      Cpsw_gDuplexNames[phyLinkCfg.duplexity]);

    return status;
}

static int32_t Cpsw_handleLinkDown(Cpsw_Handle hCpsw,
                                   Cpsw_MacPort portNum)
{
    CpswMacPort_Handle hPort = hCpsw->macPort[portNum];
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    CpswAle_DelLearntEntriesInArgs delLearntEntriesInArgs;
    int32_t status;

    CPSWTRACE_INFO("port %d: Link down\n", portNum);

    CPSW_IOCTL_SET_NO_ARGS(&prms);
    status = CpswMacPort_ioctl(hPort, CPSW_MACPORT_IOCTL_DISABLE, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to disable MAC port %u: %d\n", portNum, status);

    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_MACPORT_TO_ALEPORT(portNum);
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_DISABLED;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);

        status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_SET_PORT_STATE, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set ALE port %u to disabled state: %d\n",
                           setPortStateInArgs.portNum, status);
    }

    if (status == CPSW_SOK)
    {
        delLearntEntriesInArgs.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(portNum);
        CPSW_IOCTL_SET_IN_ARGS(&prms, &delLearntEntriesInArgs);

        status = CpswAle_ioctl(hCpsw->ale, CPSW_ALE_IOCTL_DEL_LEARNT_ENTRIES, &prms);
        if (status == CPSW_EALE_ADDRESS_NOT_FOUND)
        {
            CPSWTRACE_DEBUG("no ALE learned entries found for MAC port %u\n", portNum);
            status = CPSW_SOK;
        }

        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to delete learnt ALE entries for port %u: %d\n",
                           delLearntEntriesInArgs.portNum, status);
    }

    return status;
}

static void Cpsw_validateSsIPVersion(CSL_Xge_cpsw_ss_sRegs *cpswSsRegBase,
                                     Cpsw_Type cpswType)
{
    CSL_CPSW_SS_VERSION version;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    CSL_CPSW_SS_getVersionInfo(cpswSsRegBase, &version);

    Cpsw_assert(version.id == ipInfo->cpswSSInfo.versionInfo.moduleId);
    Cpsw_assert(version.majorVer == ipInfo->cpswSSInfo.versionInfo.majorVer);
    Cpsw_assert(version.minorVer == ipInfo->cpswSSInfo.versionInfo.minorVer);
    Cpsw_assert(version.rtlVer == ipInfo->cpswSSInfo.versionInfo.rtlVer);
}

static void Cpsw_setInterVlanAlePolicerParams(CpswAle_PolicerMatchParams *aleMatchPrms,
                                              const Cpsw_InterVlanRouteIngressPktMatchCfg *cpswMatchPrms)
{
    aleMatchPrms->policerMatchEnableMask = 0;

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_PORT;
        aleMatchPrms->portNum                 = CPSW_ALE_MACPORT_TO_ALEPORT(cpswMatchPrms->ingressPort);
        aleMatchPrms->portIsTrunk             = FALSE;
    }

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_MACSRC;
        aleMatchPrms->srcMacAddr              = cpswMatchPrms->srcMacAddr;
    }

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_MACDST;
        aleMatchPrms->dstMacAddr              = cpswMatchPrms->dstMacAddr;
    }

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_ETHERTYPE;
        aleMatchPrms->etherType               = cpswMatchPrms->etherType;
    }

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IPSRC;
        aleMatchPrms->srcIp                   = cpswMatchPrms->srcIp;
    }

    if (cpswMatchPrms->packetMatchEnableMask & CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST)
    {
        aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IPDST;
        aleMatchPrms->dstIp                   = cpswMatchPrms->dstIp;
    }

    /* Always set match params for ingress vlan id */
    aleMatchPrms->policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IVLAN;
    aleMatchPrms->ivlanId                 = cpswMatchPrms->vlanId;
}

static int32_t Cpsw_setInterVlanAleConfig(const CpswAle_Handle hAle,
                                          uint32_t destPortMask,
                                          CpswMacPort_InterVLANRouteId routeId,
                                          const Cpsw_InterVlanRouteIngressPktMatchCfg *pktMatchCfg,
                                          CpswAle_SetInterVLANConfigOutArgs *outArgs)
{
    Cpsw_IoctlPrms prms;
    CpswAle_SetInterVLANConfigInArgs aleInterVlanCfgInArgs;
    int32_t status;

    aleInterVlanCfgInArgs.destPortMask = destPortMask;
    aleInterVlanCfgInArgs.routeIndex   = CpswMacPort_mapInterVLANRouteId2Pointer(routeId);
    /* TODO: Support for egress port that is operating as a trunk */
    aleInterVlanCfgInArgs.egressTrunkIndex = 0;
    aleInterVlanCfgInArgs.enableTTLCheck   = pktMatchCfg->enableTTLCheck;
    Cpsw_setInterVlanAlePolicerParams(&aleInterVlanCfgInArgs.policerMatch, pktMatchCfg);
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &aleInterVlanCfgInArgs, outArgs);

    status = CpswAle_ioctl(hAle,
                           CPSW_ALE_IOCTL_SET_INTERVLAN_CFG,
                           &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to set interVLAN ALE config: %d\n", status);

    return status;
}

static int32_t Cpsw_clearInterVlanAleConfig(const CpswAle_Handle hAle,
                                            const Cpsw_InterVlanRouteIngressPktMatchCfg *pktMatchCfg,
                                            uint32_t delAleEntryMask)
{
    Cpsw_IoctlPrms prms;
    CpswAle_DelPolicerEntryInArgs aleDelPolicerInArgs;
    int32_t status;

    Cpsw_setInterVlanAlePolicerParams(&aleDelPolicerInArgs.policerMatch, pktMatchCfg);
    aleDelPolicerInArgs.delAleEntryMask = delAleEntryMask;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &aleDelPolicerInArgs);

    status = CpswAle_ioctl(hAle,
                           CPSW_ALE_IOCTL_DEL_POLICER,
                           &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "failed to clear interVLAN ALE config: %d\n", status);

    return status;
}

static int32_t Cpsw_setInterVlanGlobalConfig(const Cpsw_Handle hCpsw)
{
    Cpsw_IoctlPrms prms;
    int32_t status;
    CpswAle_GetPolicerGlobalConfigOutArgs alePolicerGlobalCfg;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &alePolicerGlobalCfg);
    status = CpswAle_ioctl(hCpsw->ale,
                           CPSW_ALE_IOCTL_GET_POLICER_GLOBAL_CONFIG,
                           &prms);
    if (CPSW_SOK == status)
    {
        if (alePolicerGlobalCfg.policingEnable == false)
        {
            CPSWTRACE_WARN("CPSW:Intervlan requires ALE POLICECONTROL.POLICING_EN to be true"
                           "Enabling POLICECONTROL.POLICING_EN ");
            alePolicerGlobalCfg.policingEnable = true;
            CPSW_IOCTL_SET_IN_ARGS(&prms, &alePolicerGlobalCfg);

            status = CpswAle_ioctl(hCpsw->ale,
                                   CPSW_ALE_IOCTL_SET_POLICER_GLOBAL_CONFIG,
                                   &prms);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to set policer global config: %d\n", status);
        }
    }
    else
    {
        CPSWTRACE_ERROR("failed to get policer global config: %d\n", status);
    }

    return status;
}

static int32_t Cpsw_validateInterVlanEgressConfig(const Cpsw_Handle hCpsw,
                                                  const Cpsw_InterVlanEgressPortCfg *egressCfg)
{
    int32_t status = CPSW_SOK;

    if (egressCfg->outPktModCfg.forceUntaggedEgress == false)
    {
        /* If forceUntagged egress is false, it means we need to replace VID
         * Check CPSW is in vlan aware mdoe else VID replace will not work
         * in interVLan routing */
        if (CSL_CPSW_isVlanAwareEnabled(hCpsw->cpswRegs) == 0)
        {
            CPSWTRACE_ERROR("InterVLAN Vlan Id replacement on egress requires "
                            "CpswConfig.vlanConfig.vlanAware to be true in Cpsw_Open");
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

static int32_t Cpsw_setInterVlanRouteUniEgress(const Cpsw_Handle hCpsw,
                                               const Cpsw_SetInterVlanRouteUniEgressInArgs *inArgs,
                                               Cpsw_SetInterVlanRouteUniEgressOutArgs *outArgs)
{
    CpswMacPort_InterVlanRouteIdArgs routeIdArg;
    Cpsw_IoctlPrms prms;
    int32_t status;

    status = Cpsw_validateInterVlanEgressConfig(hCpsw, &inArgs->egressCfg);

    if (CPSW_SOK == status)
    {
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg.outPktModCfg, &routeIdArg);

        status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg.egressPort],
                                   CPSW_MACPORT_IOCTL_SET_INTERVLAN_ROUTE,
                                   &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set interVLAN route: %d\n", status);
    }

    if (CPSW_SOK == status)
    {
        uint32_t aleInterVlanDestPortMask;

        Cpsw_assert(routeIdArg.routeId <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST);
        outArgs->egressPortRouteId = routeIdArg.routeId;
        aleInterVlanDestPortMask   = CPSW_ALE_MACPORT_TO_PORTMASK(inArgs->egressCfg.egressPort);
        status                     = Cpsw_setInterVlanAleConfig(hCpsw->ale,
                                                                aleInterVlanDestPortMask,
                                                                routeIdArg.routeId,
                                                                &inArgs->inPktMatchCfg,
                                                                &outArgs->ingressPacketClassifierInfo);
        if (status != CPSW_SOK)
        {
            CpswMacPort_InterVlanRouteIdArgs delRouteId;

            /* ALE IOCTL failed. Rollback the MAC Port interVLAN route allocation */
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg.outPktModCfg, &delRouteId);

            status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg.egressPort],
                                       CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE,
                                       &prms);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to delete interVLAN route: %d\n", status);
            Cpsw_assert(status == CPSW_SOK);
            Cpsw_assert(routeIdArg.routeId == delRouteId.routeId);
        }
    }

    if (status == CPSW_SOK)
    {
        status = Cpsw_setInterVlanGlobalConfig(hCpsw);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to set interVLAN global config: %d\n", status);
    }

    return status;
}

static int32_t Cpsw_clearInterVlanRouteUniEgress(const Cpsw_Handle hCpsw,
                                                 const Cpsw_ClearInterVlanRouteUniEgressInArgs *inArgs,
                                                 Cpsw_ClearInterVlanRouteUniEgressOutArgs *outArgs)
{
    Cpsw_IoctlPrms prms;
    int32_t status;

    status = Cpsw_clearInterVlanAleConfig(hCpsw->ale, &inArgs->inPktMatchCfg, inArgs->delAleEntryMask);

    if (CPSW_SOK == status)
    {
        CpswMacPort_InterVlanRouteIdArgs routeIdArgs;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg.outPktModCfg, &routeIdArgs);
        status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg.egressPort],
                                   CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE,
                                   &prms);
        if (CPSW_SOK == status)
        {
            outArgs->egressPortRouteId = routeIdArgs.routeId;
        }
        else
        {
            CPSWTRACE_ERROR("failed to delete interVLAN route: %d\n", status);
        }
    }

    return status;
}

static int32_t Cpsw_findCommonFreeSlot(const Cpsw_Handle hCpsw,
                                       uint32_t numEgressPorts,
                                       const Cpsw_InterVlanEgressPortCfg *egressCfgList,
                                       CpswMacPort_InterVLANRouteId *pFreeRouteId)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    CpswMacPort_InterVLANRouteId routeIter;
    uint32_t freeRouteMask;

    if (numEgressPorts > 0)
    {
        freeRouteMask = CPSW_UTILS_MK_ONES(Cpsw_normalizeRouteId(CPSW_MACPORT_INTERVLAN_ROUTEID_LAST), 0);
    }
    else
    {
        freeRouteMask = 0;
    }

    for (i = 0; i < numEgressPorts; i++)
    {
        CpswMacPort_InterVlanFreeRouteInfo freeRoute;
        uint32_t routeIndex;
        uint32_t portFreeRouteMask;
        const Cpsw_InterVlanEgressPortCfg *curEgressCfg;

        curEgressCfg = &egressCfgList[i];
        CPSW_IOCTL_SET_OUT_ARGS(&prms, &freeRoute);
        Cpsw_assert(CPSW_NORMALIZE_MACPORT(curEgressCfg->egressPort) <
                    CPSW_UTILS_ARRAYSIZE(hCpsw->macPort));

        status = CpswMacPort_ioctl(hCpsw->macPort[CPSW_NORMALIZE_MACPORT(curEgressCfg->egressPort)],
                                   CPSW_MACPORT_IOCTL_GET_INTERVLAN_FREEROUTES,
                                   &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to get interVLAN free routes: %d\n", status);

        portFreeRouteMask = 0;
        for (routeIndex = 0; routeIndex < freeRoute.numFreeRoutes; routeIndex++)
        {
            portFreeRouteMask |= CPSW_SET_BIT(Cpsw_normalizeRouteId(freeRoute.freeRouteId[routeIndex]));
        }

        freeRouteMask &= portFreeRouteMask;
    }

    if (freeRouteMask)
    {
        status = CPSW_SOK;
        for (routeIter = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST;
             routeIter <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;
             routeIter = (CpswMacPort_InterVLANRouteId)(routeIter + 1))
        {
            if (freeRouteMask & CPSW_SET_BIT(Cpsw_normalizeRouteId(routeIter)))
            {
                *pFreeRouteId = routeIter;
                break;
            }
        }

        /* loop to programtically validate free found common free slot across ports */
        for (i = 0; i < numEgressPorts; i++)
        {
            CpswMacPort_InterVlanRouteIdArgs routeIdArgs;
            bool isRouteFree;
            const Cpsw_InterVlanEgressPortCfg *curEgressCfg;

            curEgressCfg        = &egressCfgList[i];
            routeIdArgs.routeId = *pFreeRouteId;
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &routeIdArgs, &isRouteFree);

            status = CpswMacPort_ioctl(hCpsw->macPort[CPSW_NORMALIZE_MACPORT(curEgressCfg->egressPort)],
                                       CPSW_MACPORT_IOCTL_IS_INTERVLAN_ROUTE_FREE,
                                       &prms);
            CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                               "failed to check interVLAN route free state: %d\n", status);

            Cpsw_assert(status == CPSW_SOK);
            Cpsw_assert(isRouteFree == TRUE);
        }
    }
    else
    {
        CPSWTRACE_ERROR("no free routes\n");
        status = CPSW_EFAIL;
    }

    return status;
}

static int32_t Cpsw_validateInterVlanMultiEgressConfig(const Cpsw_Handle hCpsw,
                                                       uint32_t numEgressPorts,
                                                       const Cpsw_InterVlanEgressPortCfg *egressCfgList)
{
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0; i < numEgressPorts; i++)
    {
        const Cpsw_InterVlanEgressPortCfg *curEgressCfg;

        curEgressCfg = &egressCfgList[i];
        status       = Cpsw_validateInterVlanEgressConfig(hCpsw, curEgressCfg);
        if (status != CPSW_SOK)
        {
            CPSWTRACE_ERROR("failed to validate interVLAN egress config for port %u: %d\n",
                            i, status);
            break;
        }
    }

    return status;
}

static int32_t Cpsw_setInterVlanRouteMultiEgress(const Cpsw_Handle hCpsw,
                                                 const Cpsw_SetInterVlanRouteMultiEgressInArgs *inArgs,
                                                 Cpsw_SetInterVlanRouteMultiEgressOutArgs *outArgs)
{
    int32_t status;
    uint32_t i;
    CpswMacPort_InterVLANRouteId commonFreeRouteId = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST;
    Cpsw_IoctlPrms prms;

    status = Cpsw_validateInterVlanMultiEgressConfig(hCpsw, inArgs->numEgressPorts, inArgs->egressCfg);

    if (CPSW_SOK == status)
    {
        status = Cpsw_findCommonFreeSlot(hCpsw,
                                         inArgs->numEgressPorts,
                                         inArgs->egressCfg,
                                         &commonFreeRouteId);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to find common free slot: %d\n", status);
    }

    if (CPSW_SOK == status)
    {
        for (i = 0; i < inArgs->numEgressPorts; i++)
        {
            CpswMacPort_SetSpecificInterVlanRouteInArgs portInArgs;

            portInArgs.routeId  = commonFreeRouteId;
            portInArgs.routeCfg = inArgs->egressCfg[i].outPktModCfg;
            CPSW_IOCTL_SET_IN_ARGS(&prms, &portInArgs);

            status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg[i].egressPort],
                                       CPSW_MACPORT_IOCTL_SET_SPECIFIC_INTERVLAN_ROUTE,
                                       &prms);
            if (status != CPSW_SOK)
            {
                CPSWTRACE_ERROR("failed to set specific interVLAN route: %d\n");
                break;
            }
        }
    }

    if (CPSW_SOK == status)
    {
        uint32_t aleInterVlanDestPortMask;

        Cpsw_assert(commonFreeRouteId <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST);
        outArgs->egressPortRouteId = commonFreeRouteId;
        aleInterVlanDestPortMask   = 0;
        for (i = 0; i < inArgs->numEgressPorts; i++)
        {
            aleInterVlanDestPortMask |= CPSW_ALE_MACPORT_TO_PORTMASK(inArgs->egressCfg[i].egressPort);
        }

        status = Cpsw_setInterVlanAleConfig(hCpsw->ale,
                                            aleInterVlanDestPortMask,
                                            commonFreeRouteId,
                                            &inArgs->inPktMatchCfg,
                                            &outArgs->ingressPacketClassifierInfo);
        if (status != CPSW_SOK)
        {
            /* ALE IOCTL failed. Rollback the MAC Port interVLAN route allocation */
            for (i = 0; i < inArgs->numEgressPorts; i++)
            {
                CpswMacPort_InterVlanRouteIdArgs delRouteId;

                CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg[i].outPktModCfg, &delRouteId);

                status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg[i].egressPort],
                                           CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE,
                                           &prms);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to delete interVLAN route: %d\n");
                Cpsw_assert(status == CPSW_SOK);
                Cpsw_assert(commonFreeRouteId == delRouteId.routeId);
            }
        }
    }

    if (CPSW_SOK == status)
    {
        Cpsw_setInterVlanGlobalConfig(hCpsw);
    }

    return status;
}

static int32_t Cpsw_validateClearInterVlanRouteMultiEgress(const Cpsw_Handle hCpsw,
                                                           const Cpsw_ClearInterVlanRouteMultiEgressInArgs *inArgs)
{
    uint32_t i;
    CpswMacPort_InterVlanRouteIdArgs routeIdArgs;
    CpswMacPort_InterVLANRouteId commonRouteId;
    bool validateFailed = FALSE;
    Cpsw_IoctlPrms prms;
    int32_t status;

    for (i = 0; ((i < inArgs->numEgressPorts) && (validateFailed == FALSE)); i++)
    {
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg[i].outPktModCfg, &routeIdArgs);
        status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg[i].egressPort],
                                   CPSW_MACPORT_IOCTL_FIND_INTERVLAN_ROUTE,
                                   &prms);
        if (CPSW_SOK == status)
        {
            if (0 == i)
            {
                commonRouteId = routeIdArgs.routeId;
            }
            else
            {
                if (routeIdArgs.routeId != commonRouteId)
                {
                    validateFailed = TRUE;
                }
            }
        }
        else
        {
            CPSWTRACE_ERROR("failed to find interVLAN route: %d\n", status);
            validateFailed = TRUE;
        }
    }

    if (validateFailed == TRUE)
    {
        status = CPSW_EFAIL;
    }
    else
    {
        status = CPSW_SOK;
    }

    return status;
}

static int32_t Cpsw_clearInterVlanRouteMultiEgress(const Cpsw_Handle hCpsw,
                                                   const Cpsw_ClearInterVlanRouteMultiEgressInArgs *inArgs,
                                                   Cpsw_ClearInterVlanRouteMultiEgressOutArgs *outArgs)
{
    Cpsw_IoctlPrms prms;
    int32_t status;
    CpswMacPort_InterVlanRouteIdArgs routeIdArgs;
    uint32_t i;

    status = Cpsw_validateClearInterVlanRouteMultiEgress(hCpsw, inArgs);
    if (CPSW_SOK == status)
    {
        status = Cpsw_clearInterVlanAleConfig(hCpsw->ale,
                                              &inArgs->inPktMatchCfg,
                                              inArgs->delAleEntryMask);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "failed to clear interVLAN ALE config: %d\n", status);
    }

    if (CPSW_SOK == status)
    {
        for (i = 0; i < inArgs->numEgressPorts; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs->egressCfg[i].outPktModCfg, &routeIdArgs);
            status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->egressCfg[i].egressPort],
                                       CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE,
                                       &prms);
            if (CPSW_SOK != status)
            {
                CPSWTRACE_ERROR("failed to delete interVLAN route: %d\n", status);
                break;
            }
            else
            {
                outArgs->egressPortRouteId = routeIdArgs.routeId;
            }
        }
    }

    return status;
}

static uint32_t Cpsw_normalizeRouteId(CpswMacPort_InterVLANRouteId routeId)
{
    return((routeId) - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST);
}

static int32_t Cpsw_validateTxShortIPGConfig(const Cpsw_Handle hCpsw,
                                             const Cpsw_SetShortTxIPGConfigInArgs *inArgs)
{
    uint32_t i;
    int32_t status = CPSW_SOK;

    if ((inArgs->configureGapThreshold) &&
        (inArgs->ipgTriggerThresholdBlkCount > CSL_XGE_CPSW_GAP_THRESH_REG_GAP_THRESH_MAX))
    {
        CPSWTRACE_ERROR("IPG trigger thres block count %u exceeded max",
                        inArgs->ipgTriggerThresholdBlkCount);
        status = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == status)
    {
        if (inArgs->numMacPorts > CPSW_UTILS_ARRAYSIZE(inArgs->portIPGCfg))
        {
            CPSWTRACE_ERROR("invalid number of MAC ports %u exceeded %u\n",
                            inArgs->numMacPorts, CPSW_UTILS_ARRAYSIZE(inArgs->portIPGCfg));
            status = CPSW_EINVALID_PARAMS;
        }
    }

    if (CPSW_SOK == status)
    {
        uint32_t macPortList = 0;

        for (i = 0; i < inArgs->numMacPorts; i++)
        {
            if ((CPSW_NORMALIZE_MACPORT(inArgs->portIPGCfg[i].portNum) > Cpsw_getMacPortMax(hCpsw->cpswType))
                ||
                (hCpsw->macPort[CPSW_NORMALIZE_MACPORT(inArgs->portIPGCfg[i].portNum)] == NULL))
            {
                CPSWTRACE_ERROR("CPSW:Invalid MAC port:%u", inArgs->portIPGCfg[i].portNum);
                status = CPSW_EINVALID_PARAMS;
                break;
            }

            if (CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(inArgs->portIPGCfg[i].portNum)) & macPortList)
            {
                /* inArgs->portIPGCfg entries should have list of unique mac ports.
                 * Flag error as previously seen macPort in inArgs->portIPGCfg
                 * is encountered again */
                CPSWTRACE_ERROR("CPSW:Multiple occurence of same macport:%u", inArgs->portIPGCfg[i].portNum);
                status = CPSW_EINVALID_PARAMS;
                break;
            }

            if ((inArgs->portIPGCfg[i].shortGapCfg.enableTxShortGap == false)
                &&
                (inArgs->portIPGCfg[i].shortGapCfg.enableTxShortGapLimit == true))
            {
                CPSWTRACE_ERROR("CPSW:Short gap limit can be enabled only if short gap is enabled:%u", inArgs->portIPGCfg[i].portNum);
                status = CPSW_EINVALID_PARAMS;
                break;
            }

            macPortList |= CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(inArgs->portIPGCfg[i].portNum));
        }
    }

    return status;
}

static int32_t Cpsw_setTxShortIPGConfig(const Cpsw_Handle hCpsw,
                                        const Cpsw_SetShortTxIPGConfigInArgs *inArgs)
{
    uint32_t i;
    int32_t status;
    Cpsw_IoctlPrms prms;

    status = Cpsw_validateTxShortIPGConfig(hCpsw, inArgs);
    if (CPSW_SOK == status)
    {
        if (inArgs->configureGapThreshold)
        {
            CSL_CPSW_setGapThreshold(hCpsw->cpswRegs, inArgs->ipgTriggerThresholdBlkCount);
        }

        for (i = 0; i < inArgs->numMacPorts; i++)
        {
            CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs->portIPGCfg[i]);
            status = CpswMacPort_ioctl(hCpsw->macPort[inArgs->portIPGCfg[i].portNum],
                                       CPSW_MACPORT_IOCTL_SET_SHORT_IPG,
                                       &prms);
            if (CPSW_SOK != status)
            {
                CPSWTRACE_ERROR("failed to set short IPG: %d\n", status);
                break;
            }
        }
    }
    else
    {
        CPSWTRACE_ERROR("failed to validated short IPG config: %d\n", status);
    }

    return status;
}

static int32_t Cpsw_getTxShortIPGConfig(const Cpsw_Handle hCpsw,
                                        Cpsw_GetShortTxIPGConfigOutArgs *outArgs)
{
    uint32_t i;
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    outArgs->ipgTriggerThresholdBlkCount = CSL_CPSW_getGapThreshold(hCpsw->cpswRegs);
    outArgs->numMacPorts                 = 0;
    for (i = 0; i < Cpsw_getMacPortMax(hCpsw->cpswType); i++)
    {
        if (hCpsw->macPort[i])
        {
            CpswMacPort_GenericInArgs inArgs;

            inArgs.portNum = hCpsw->macPort[i]->portNum;
            outArgs->portIPGCfg[outArgs->numMacPorts].portNum = inArgs.portNum;
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs->portIPGCfg[outArgs->numMacPorts].shortGapCfg);
            status = CpswMacPort_ioctl(hCpsw->macPort[i],
                                       CPSW_MACPORT_IOCTL_GET_SHORT_IPG,
                                       &prms);
            if (CPSW_SOK != status)
            {
                CPSWTRACE_ERROR("failed to get short IPG: %d\n", status);
                break;
            }

            outArgs->numMacPorts++;
        }
    }

    return status;
}
