/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_cpts.c
 *
 * \brief This file contains the implementation of the CPTS submodule.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_cpts.h>
#include <ti/drv/cpsw/src/priv/cpsw_osal.h>

#include "priv/cpsw_cpts_priv.h"
#include "priv/cpsw_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static CpswCpts_Handle CpswCpts_getObject(Cpsw_Type cpswType);

static void CpswCpts_validateIPVersion(CSL_cptsRegs *cpswCptsRegBase,
                                       Cpsw_Type cpswType);

static int32_t CpswCpts_handleEvents(CpswCpts_Handle hCpts);

static int32_t CpswCpts_clearEventMemPools(CpswCpts_Handle hCpts);

static void CpswCpts_printRegs(CSL_cptsRegs *cptsRegs);

static void CpswCpts_printStats(CpswCpts_Handle hCpts);

static void CpswCpts_calcPpmVal(uint64_t *tsPpmVal,
                                CpswCpts_PpmMode ppmMode);

static int32_t CpswCpts_lookUpEvent(CpswCpts_Handle hCpts,
                                    CpswCpts_LookUpEventInArgs *matchPrms,
                                    CpswCpts_Event *event);

static int32_t CpswCpts_checkTsCompCompatibility(CpswCpts_Handle hCpts);

static int32_t CpswCpts_checkGenfEstfErrata(CpswCpts_Handle hCpts,
                                            uint32_t index,
                                            bool isGenf);

static int32_t CpswCpts_checkSetGenfInArgs(CpswCpts_Handle hCpts,
                                           CpswCpts_SetFxnGenInArgs *setGenFInArgs);

static int32_t CpswCpts_checkSetEstfInArgs(CpswCpts_Handle hCpts,
                                           CpswCpts_SetFxnGenInArgs *setESTFInArgs);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW CPTS objects */
CpswCpts_Obj cptsObjs[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswCpts_initParams(CpswCpts_Config *cfg)
{
    cfg->hostRxTsEnable = true;
    cfg->tsCompPolarity = true;
    cfg->tsRxEventsDisable = false;
    cfg->tsGenfClrEnable = false;
    cfg->cptsRftClkFreq  = CPSW_CPTS_RFTCLK_FREQ_500MHZ;
}

CpswCpts_Handle CpswCpts_open(Cpsw_Type cpswType,
                              const CpswCpts_Config *cfg)
{
    CSL_CPTS_CONTROL control;
    CpswCpts_Handle hCpts;
    const CpswSoc_Config *pSocCfg;
    uint8_t i = 0U;

    hCpts = CpswCpts_getObject(cpswType);
    memset(hCpts, 0U, sizeof(*hCpts));
    hCpts->cpswType = cpswType;
    hCpts->cptsRegs = Cpsw_getCslCptsRegs(cpswType);
    Cpsw_assert(hCpts->cptsRegs != NULL);

    CpswCpts_validateIPVersion(hCpts->cptsRegs, cpswType);

    hCpts->tsPushInFifo = false;

    memset(&control, 0U, sizeof(CSL_CPTS_CONTROL));

    /* Set default values for control register fields */
    pSocCfg = CpswSoc_getConfig(cpswType);
    hCpts->hwPushCount = pSocCfg->cptsHwPushCount;
    control.cptsEn  = true;
    control.intTest = false;
    control.ts64bMode = true;
    control.tsOutputBitSel = CPSW_CPTS_TS_OUTPUT_BIT_DISABLED;
    control.seqEn = false;
    control.tsCompToggle = false;
    for (i = 0U; i < hCpts->hwPushCount; i++)
    {
        control.tsHwPushEn[i] = true;
    }

    /* Set Application based CPTS control configurations */
    control.tsCompPolarity = cfg->tsCompPolarity;
    control.tsDisableRxEvents = cfg->tsRxEventsDisable;
    control.tsGenfClrEn = cfg->tsGenfClrEnable;
    control.tstampEn = cfg->hostRxTsEnable;

    CSL_CPTS_disableCpts(hCpts->cptsRegs);

    /* Configure Time stamp add value to enable 1ns operations */
    CSL_CPTS_setTSAddVal(hCpts->cptsRegs, cfg->cptsRftClkFreq);

    CSL_CPTS_setCntlReg(hCpts->cptsRegs, &control);

    CSL_CPTS_enableCpts(hCpts->cptsRegs);

    hCpts->magic = CPSW_CPTS_MAGIC;

    return hCpts;
}

CpswCpts_Handle CpswCpts_reopen(Cpsw_Type cpswType)
{
    CpswCpts_Handle hCpts = CpswCpts_getObject(cpswType);

    memset(hCpts, 0U, sizeof(*hCpts));

    hCpts->cptsRegs = Cpsw_getCslCptsRegs(cpswType);
    Cpsw_assert(hCpts->cptsRegs != NULL);

    hCpts->magic = CPSW_CPTS_MAGIC;

    return hCpts;
}

void CpswCpts_close(CpswCpts_Handle hCpts)
{
    if (hCpts->magic == CPSW_CPTS_MAGIC)
    {
        /* Disable CPTS module to disable events*/
        CSL_CPTS_disableCpts(hCpts->cptsRegs);

        hCpts->magic = CPSW_NO_MAGIC;
        hCpts->cptsRegs = NULL;
    }
    else
    {
        CPSWTRACE_ERROR("CPTS is not open (handle=%p)\n", hCpts);
    }
}

int32_t CpswCpts_ioctl(CpswCpts_Handle hCpts,
                       uint32_t cmd,
                       Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    if (hCpts->magic != CPSW_CPTS_MAGIC)
    {
        CPSWTRACE_ERROR("CPTS is not open (handle=%p)\n", hCpts);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        switch (cmd)
        {
            case CPSW_CPTS_IOCTL_GET_VERSION:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(Cpsw_GetVersionOutArgs));
                if (status == CPSW_SOK)
                {
                    Cpsw_Version *version = (Cpsw_Version *)prms->outArgs;
                    CSL_CPTS_VERSION cptsVer;

                    CSL_CPTS_getCptsVersionInfo(hCpts->cptsRegs, &cptsVer);
                    version->major = cptsVer.majorVer;
                    version->minor = cptsVer.minorVer;
                    version->rtl = cptsVer.rtlVer;
                    version->id  = cptsVer.id;
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_VERSION params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_REGISTER_STACK:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_RegisterStackInArgs));
                if (status == CPSW_SOK)
                {
                    CpswCpts_RegisterStackInArgs *stackInArgs = (CpswCpts_RegisterStackInArgs *)prms->inArgs;
                    if (hCpts->pEventNotifyCb == NULL)
                    {
                        if (stackInArgs->pEventNotifyCb != NULL)
                        {
                            hCpts->pEventNotifyCb = stackInArgs->pEventNotifyCb;
                            hCpts->hEventNotifyCbArg = stackInArgs->hEventNotifyCbArg;
                        }
                        else
                        {
                            CPSWTRACE_ERROR("invalid REGISTER_STACK params\n");
                            status = CPSW_EINVALID_PARAMS;
                        }
                    }
                    else
                    {
                        CPSWTRACE_ERROR("Stack already registered\n");
                        status = CPSW_EALREADY_OPEN;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid REGISTER_STACK params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_UNREGISTER_STACK:
            {
                /* Check if the stack is registered or not */
                if (hCpts->pEventNotifyCb != NULL)
                {
                    hCpts->pEventNotifyCb = NULL;
                    hCpts->hEventNotifyCbArg = NULL;
                }
                else
                {
                    CPSWTRACE_ERROR("No stack is registered \n");
                    status = CPSW_EUNEXPECTED;
                }
            }
            break;

            case CPSW_CPTS_IOCTL_REGISTER_HW_PUSH_CALLBACK:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_RegisterHwPushCbInArgs));
                if (status == CPSW_SOK)
                {
                    CpswCpts_RegisterHwPushCbInArgs *hwPushRegisterCbInArgs = (CpswCpts_RegisterHwPushCbInArgs *)prms->inArgs;
                    uint32_t hwPushIdx = 0U;

                    if ((hwPushRegisterCbInArgs->hwPushNum <= hCpts->hwPushCount) &&
                        (hwPushRegisterCbInArgs->hwPushNum >= CPSW_CPTS_HW_PUSH_FIRST))
                    {
                        hwPushIdx = CPSW_CPTS_HW_PUSH_NORMALIZE(hwPushRegisterCbInArgs->hwPushNum);
                        if ((hwPushIdx < CPSW_CPTS_HW_PUSH_COUNT_MAX) && (hCpts->pHwPushNotifyCb[hwPushIdx] == NULL))
                        {
                            if (hwPushRegisterCbInArgs->pHwPushNotifyCb != NULL)
                            {
                                hCpts->pHwPushNotifyCb[hwPushIdx] = hwPushRegisterCbInArgs->pHwPushNotifyCb;
                                hCpts->pHwPushNotifyCbArg[hwPushIdx] = hwPushRegisterCbInArgs->pHwPushNotifyCbArg;
                            }
                            else
                            {
                                CPSWTRACE_ERROR("invalid REGISTER_HW_PUSH_CALLBACK params\n");
                                status = CPSW_EINVALID_PARAMS;
                            }
                        }
                        else
                        {
                            CPSWTRACE_ERROR("Hardware push callback for instance %d already registered\n",
                                            hwPushRegisterCbInArgs->hwPushNum);
                            status = CPSW_EALREADY_OPEN;
                        }
                    }
                    else
                    {
                        CPSWTRACE_ERROR("invalid hardware push instance number %d\n",
                                        hwPushRegisterCbInArgs->hwPushNum);
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid REGISTER_HW_PUSH_CALLBACK params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_UNREGISTER_HW_PUSH_CALLBACK:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_HwPush));
                if (status == CPSW_SOK)
                {
                    CpswCpts_HwPush hwPushNum = *((CpswCpts_HwPush *)prms->inArgs);
                    uint32_t hwPushIdx = 0U;

                    if ((hwPushNum <= hCpts->hwPushCount) &&
                        (hwPushNum >= CPSW_CPTS_HW_PUSH_FIRST))
                    {
                        hwPushIdx = CPSW_CPTS_HW_PUSH_NORMALIZE(hwPushNum);

                        /* Check if the callback is registered or not */
                        if ((hwPushIdx < CPSW_CPTS_HW_PUSH_COUNT_MAX) && (hCpts->pHwPushNotifyCb[hwPushIdx] != NULL))
                        {
                            hCpts->pHwPushNotifyCb[hwPushIdx] = NULL;
                            hCpts->pHwPushNotifyCbArg[hwPushIdx] = NULL;
                        }
                        else
                        {
                            CPSWTRACE_WARN("No callback for hardware push instance %d is registered\n",
                                           hwPushNum);
                            status = CPSW_SOK;
                        }
                    }
                    else
                    {
                        CPSWTRACE_ERROR("invalid hardware push instance number %d\n",
                                        hwPushNum);
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid UNREGISTER_HW_PUSH_CALLBACK params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_ENABLE_INTR:
            {
                CSL_CPTS_enableInterrupt(hCpts->cptsRegs);
            }
            break;

            case CPSW_CPTS_IOCTL_DISABLE_INTR:
            {
                CSL_CPTS_disableInterrupt(hCpts->cptsRegs);
            }
            break;

            case CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(uint64_t));
                if (status == CPSW_SOK)
                {
                    uint64_t *tsVal = (uint64_t *)prms->outArgs;
                    int32_t loop = CPSW_CPTS_POLL_INTERVAL;

                    if (false == hCpts->tsPushInFifo)
                    {
                        hCpts->tsPushInFifo = true;
                        CSL_CPTS_TSEventPush(hCpts->cptsRegs);

                        /* Poll until the Time stamp event occurs */
                        while (loop > 0U)
                        {
                            if (hCpts->tsPushInFifo == false)
                            {
                                *tsVal = hCpts->tsVal;
                                break;
                            }
#ifdef QNX_OS
                            delay(0);
#endif
                            loop--;
                        }

                        if (loop == 0U)
                        {
                            *tsVal = 0U;
                            CPSWTRACE_ERROR("Timeout!! Failed to retrieve event, try after sometime \n");
                            status = CPSW_ENOTFOUND;
                        }
                    }
                    else
                    {
                        CPSWTRACE_WARN("Time stamp push event is already in FIFO, try pushing the event after sometime...\n");
                        status = CPSW_EBUSY;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid PUSH_TIMESTAMP params \n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_LOAD_TIMESTAMP:
            {
                status = Cpsw_checkInArgs(prms, sizeof(uint64_t));
                if (status == CPSW_SOK)
                {
                    uint64_t tsLoadVal = *(uint64_t *)prms->inArgs;
                    uint32_t tsLoadValHi = 0U;
                    uint32_t tsLoadValLo = 0U;

                    /* Disable CPTS Interrupt */
                    CSL_CPTS_disableInterrupt(hCpts->cptsRegs);

                    tsLoadValHi = (uint32_t)(tsLoadVal >> 32U);
                    tsLoadValLo = (uint32_t)(tsLoadVal & 0xFFFFFFFF);
                    CSL_CPTS_setTSVal(hCpts->cptsRegs,
                                      tsLoadValLo,
                                      tsLoadValHi);

                    /* Re-enable CPTS Interrupt */
                    CSL_CPTS_enableInterrupt(hCpts->cptsRegs);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid LOAD_TIMESTAMP params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_TS_NUDGE:
            {
                status = Cpsw_checkInArgs(prms, sizeof(int32_t));
                if (status == CPSW_SOK)
                {
                    int32_t tsNudge = *(int32_t *)prms->inArgs;

                    if ((tsNudge >= CPSW_CPTS_NUDGE_MIN_VAL) &&
                        (tsNudge <= CPSW_CPTS_NUDGE_MAX_VAL))
                    {
                        CSL_CPTS_setTSNudge(hCpts->cptsRegs, tsNudge);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("Nudge value %u is out of range (%u, %u)\n",
                                        tsNudge, CPSW_CPTS_NUDGE_MIN_VAL, CPSW_CPTS_NUDGE_MAX_VAL);
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_TS_NUDGE params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_TS_PPM:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_SetTsPpmInArgs));
                if (status == CPSW_SOK)
                {
                    CpswCpts_SetTsPpmInArgs *tsPpmInArgs = (CpswCpts_SetTsPpmInArgs *)prms->inArgs;

                    if (tsPpmInArgs->tsPpmMode != CPSW_CPTS_PPM_DISABLE)
                    {
                        CpswCpts_calcPpmVal(&tsPpmInArgs->tsPpmVal,
                                            tsPpmInArgs->tsPpmMode);

                        uint32_t tsPpmValHi = (uint32_t)(tsPpmInArgs->tsPpmVal >> 32U);
                        uint32_t tsPpmValLo = (uint32_t)(tsPpmInArgs->tsPpmVal & 0xFFFFFFFF);

                        CSL_CPTS_setTSPpm(hCpts->cptsRegs,
                                          tsPpmValLo,
                                          tsPpmValHi,
                                          tsPpmInArgs->tsPpmDir);
                    }
                    else
                    {
                        CSL_CPTS_setTSPpm(hCpts->cptsRegs,
                                          0U,
                                          0U,
                                          tsPpmInArgs->tsPpmDir);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_TS_PPM params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_COMP:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_SetCompValInArgs));
                if (status == CPSW_SOK)
                {
                    status = CpswCpts_checkTsCompCompatibility(hCpts);

                    if (status == CPSW_SOK)
                    {
                        CSL_CPTS_CONTROL cptsCntlCfg;
                        uint32_t tsCompValHi = 0U;
                        uint32_t tsCompValLo = 0U;

                        CpswCpts_SetCompValInArgs *tsCompValInArgs = (CpswCpts_SetCompValInArgs *)prms->inArgs;

                        CSL_CPTS_getCntlReg(hCpts->cptsRegs, &cptsCntlCfg);
                        cptsCntlCfg.tsCompToggle = tsCompValInArgs->tsCompToggle;
                        CSL_CPTS_setCntlReg(hCpts->cptsRegs, &cptsCntlCfg);

                        tsCompValHi = (uint32_t)(tsCompValInArgs->tsCompVal >> 32U);
                        tsCompValLo = (uint32_t)(tsCompValInArgs->tsCompVal & 0xFFFFFFFF);
                        CSL_CPTS_setTSCompVal(hCpts->cptsRegs,
                                              tsCompValLo,
                                              tsCompValHi,
                                              tsCompValInArgs->tsCompLen);
                    }
                    else
                    {
                        CPSWTRACE_ERROR(" SETUP_COMP not supported with TS Add Value or TS PPM, try GENFn instead for same functionality \n");
                        status = CPSW_ENOTSUPPORTED;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SETUP_COMP params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_COMP_NUDGE:
            {
                status = Cpsw_checkInArgs(prms, sizeof(int32_t));
                if (status == CPSW_SOK)
                {
                    int32_t tsCompNudge = *(int32_t *)prms->inArgs;

                    if ((tsCompNudge >= CPSW_CPTS_NUDGE_MIN_VAL) &&
                        (tsCompNudge <= CPSW_CPTS_NUDGE_MAX_VAL))
                    {
                        CSL_CPTS_setTSCompNudge(hCpts->cptsRegs, tsCompNudge);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("Nudge value %u is out of range (%u, %u)\n",
                                        tsCompNudge, CPSW_CPTS_NUDGE_MIN_VAL, CPSW_CPTS_NUDGE_MAX_VAL);
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_COMP_NUDGE params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_GENF:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_SetFxnGenInArgs));
                if (status == CPSW_SOK)
                {
                    CpswCpts_SetFxnGenInArgs *setGenFInArgs = (CpswCpts_SetFxnGenInArgs *)prms->inArgs;

                    status = CpswCpts_checkSetGenfInArgs(hCpts, setGenFInArgs);
                    if (status == CPSW_SOK)
                    {
                        status = CSL_CPTS_setupGENFn(hCpts->cptsRegs,
                                                     setGenFInArgs->index,
                                                     setGenFInArgs->length,
                                                     setGenFInArgs->compare,
                                                     setGenFInArgs->polarityInv,
                                                     setGenFInArgs->ppmVal,
                                                     setGenFInArgs->ppmDir);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "Wrong GENFn index value, error: %d\n", status);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SETUP_GENF params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_GENF_NUDGE:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CptsSetFxnGenNudgeInArgs));
                if (status == CPSW_SOK)
                {
                    CptsSetFxnGenNudgeInArgs *setGenFNudgeInArgs = (CptsSetFxnGenNudgeInArgs *)prms->inArgs;

                    if ((setGenFNudgeInArgs->tsNudge >= CPSW_CPTS_NUDGE_MIN_VAL) &&
                        (setGenFNudgeInArgs->tsNudge <= CPSW_CPTS_NUDGE_MAX_VAL))
                    {
                        status = CSL_CPTS_setGENFnNudge(hCpts->cptsRegs,
                                                        setGenFNudgeInArgs->index,
                                                        setGenFNudgeInArgs->tsNudge);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "Wrong GENFn index value, error: %d\n", status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("Nudge value %u is out of range (%u, %u)\n",
                                        setGenFNudgeInArgs->tsNudge, CPSW_CPTS_NUDGE_MIN_VAL, CPSW_CPTS_NUDGE_MAX_VAL);
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_GENF_NUDGE params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_ESTF:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_SetFxnGenInArgs));
                if (status == CPSW_SOK)
                {
                    CpswCpts_SetFxnGenInArgs *setESTFInArgs = (CpswCpts_SetFxnGenInArgs *)prms->inArgs;

                    status = CpswCpts_checkSetEstfInArgs(hCpts, setESTFInArgs);
                    if (status == CPSW_SOK)
                    {
                        status = CSL_CPTS_setupESTFn(hCpts->cptsRegs,
                                                     setESTFInArgs->index,
                                                     setESTFInArgs->length,
                                                     setESTFInArgs->compare,
                                                     setESTFInArgs->polarityInv,
                                                     setESTFInArgs->ppmVal,
                                                     setESTFInArgs->ppmDir);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "Wrong ESTFn index value, error: %d\n", status);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SETUP_ESTF params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SET_ESTF_NUDGE:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CptsSetFxnGenNudgeInArgs));
                if (status == CPSW_SOK)
                {
                    CptsSetFxnGenNudgeInArgs *setESTFNudgeInArgs = (CptsSetFxnGenNudgeInArgs *)prms->inArgs;
                    if ((setESTFNudgeInArgs->tsNudge >= CPSW_CPTS_NUDGE_MIN_VAL) &&
                        (setESTFNudgeInArgs->tsNudge <= CPSW_CPTS_NUDGE_MAX_VAL))
                    {
                        status = CSL_CPTS_setESTFnNudge(hCpts->cptsRegs,
                                                        setESTFNudgeInArgs->index,
                                                        setESTFNudgeInArgs->tsNudge);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "Wrong ESTFn index value, error: %d\n", status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("Nudge value is not in the acceptable range of -128 to 127\n");
                        status = CPSW_EINVALID_PARAMS;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_ESTF_NUDGE params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_SELECT_TS_OUTPUT_BIT:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswCpts_OutputBitSel));
                if (status == CPSW_SOK)
                {
                    CSL_CPTS_CONTROL cptsCntlCfg;
                    CpswCpts_OutputBitSel bitSelect = *((CpswCpts_OutputBitSel *)prms->inArgs);

                    CSL_CPTS_getCntlReg(hCpts->cptsRegs, &cptsCntlCfg);

                    if (CPSW_CPTS_TS_OUTPUT_BIT_DISABLED != cptsCntlCfg.tsOutputBitSel)
                    {
                        cptsCntlCfg.tsOutputBitSel = CPSW_CPTS_TS_OUTPUT_BIT_DISABLED;
                        CSL_CPTS_setCntlReg(hCpts->cptsRegs, &cptsCntlCfg);
                    }

                    cptsCntlCfg.tsOutputBitSel = bitSelect;
                    CSL_CPTS_setCntlReg(hCpts->cptsRegs, &cptsCntlCfg);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SELECT_TS_OUTPUT_BIT params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_LOOKUP_EVENT:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswCpts_LookUpEventInArgs),
                                             sizeof(CpswCpts_Event));
                if (status == CPSW_SOK)
                {
                    CpswCpts_LookUpEventInArgs *matchPrms = (CpswCpts_LookUpEventInArgs *)prms->inArgs;
                    CpswCpts_Event *event = (CpswCpts_Event *)prms->outArgs;

                    status = CpswCpts_lookUpEvent(hCpts, matchPrms, event);
#ifndef QNX_OS
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to lookup for event: %d\n", status);
#endif
                }
                else
                {
                    CPSWTRACE_ERROR("invalid LOOKUP_EVENT params\n");
                }
            }
            break;

            case CPSW_CPTS_IOCTL_CLEAR_ALL_EVENTS:
            {
                status = CpswCpts_clearEventMemPools(hCpts);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "failed to clear all events: %d\n", status);
            }
            break;

            case CPSW_CPTS_IOCTL_HANDLE_INTR:
            {
                /* Call CPTS Handle Events function and notify the respective callbacks */
                status = CpswCpts_handleEvents(hCpts);
                if (status != CPSW_SOK)
                {
                    Cpsw_assert(status != CPSW_SOK);
                    CpswUtils_printf("Error in handling CPTS Interrupt: %d\n", status);
                }
            }
            break;

            case CPSW_CPTS_IOCTL_PRINT_REGS:
            {
                CpswCpts_printRegs(hCpts->cptsRegs);
            }
            break;

            case CPSW_CPTS_IOCTL_PRINT_STATS:
            {
                CpswCpts_printStats(hCpts);
            }
            break;

            default:
            {
                CPSWTRACE_ERROR("invalid IOCTL cmd: %u\n", cmd);
                status = CPSW_EINVALID_PARAMS;
            }
            break;
        }
    }

    return status;
}

static CpswCpts_Handle CpswCpts_getObject(Cpsw_Type cpswType)
{
    CpswCpts_Handle hCpts = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hCpts = &cptsObjs[cpswType];
            break;

        default:
            Cpsw_assert(false);
    }

    return hCpts;
}

static void CpswCpts_validateIPVersion(CSL_cptsRegs *cpswCptsRegBase,
                                       Cpsw_Type cpswType)
{
    CSL_CPTS_VERSION version;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    CSL_CPTS_getCptsVersionInfo(cpswCptsRegBase, &version);
    Cpsw_assert(version.id == ipInfo->cptsInfo.versionInfo.moduleId);
    Cpsw_assert(version.majorVer == ipInfo->cptsInfo.versionInfo.majorVer);
    Cpsw_assert(version.minorVer == ipInfo->cptsInfo.versionInfo.minorVer);
    Cpsw_assert(version.rtlVer == ipInfo->cptsInfo.versionInfo.rtlVer);
}

static int32_t CpswCpts_handleEvents(CpswCpts_Handle hCpts)
{
    CSL_CPTS_EVENTINFO eventInfo;
    volatile bool intrPendStatus = false;
    CpswCpts_Event *buff = NULL;
    uint32_t *idx  = NULL;
    int32_t status = CPSW_SOK;
    uint64_t tsVal = 0U;
    uint32_t hwPushIdx = 0U;

    /* Disable CPTS Interrupt */
    CSL_CPTS_disableInterrupt(hCpts->cptsRegs);

    do
    {
        CSL_CPTS_getEventInfo(hCpts->cptsRegs, &eventInfo);
        /* Pop the previously read value off of the event FIFO */
        CSL_CPTS_popEvent(hCpts->cptsRegs);

        tsVal = ((((uint64_t)(eventInfo.timeStampHi)) << 32U) |
                 ((uint64_t)(eventInfo.timeStamp)));

        if (eventInfo.eventType == CPSW_CPTS_EVENTTYPE_TS_PUSH)
        {
            hCpts->tsVal = tsVal;
            /* Clear the flag that indicates a TS_PUSH event is in the FIFO */
            hCpts->tsPushInFifo = false;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
            hCpts->eventStats.swTsPushEventCnt++;
#endif
        }
        else
        {
            switch (eventInfo.eventType)
            {
                case CPSW_CPTS_EVENTTYPE_HW_TS_PUSH:
                {
                    buff = &hCpts->hwPushEventPool.eventMemPool[0U];
                    idx  = &hCpts->hwPushEventPool.index;

                    /* Get the HW PUSH input number(1-8) which caused the event */
                    buff[*idx].eventType = (CpswCpts_EventType)eventInfo.eventType;
                    buff[*idx].hwPushNum = (CpswCpts_HwPush)eventInfo.portNo;

                    /* These fields are meaningless for this event type */
                    buff[*idx].seqId   = 0U;
                    buff[*idx].domain  = 0U;
                    buff[*idx].msgType = CPSW_CPTS_MESSAGE_INVALID;

                    /* Check for registered calback and call it */
                    hwPushIdx = CPSW_CPTS_HW_PUSH_NORMALIZE(buff[*idx].hwPushNum);
                    if ((hwPushIdx < CPSW_CPTS_HW_PUSH_COUNT_MAX) && (hCpts->pHwPushNotifyCb[hwPushIdx] != NULL))
                    {
                        hCpts->pHwPushNotifyCb[hwPushIdx](hCpts->pHwPushNotifyCbArg[hwPushIdx],
                                                          buff[*idx].hwPushNum);
                    }
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                    hCpts->eventStats.hwTsPushEventCnt++;
                    if (buff[*idx].tsVal != 0U)
                    {
                        hCpts->eventStats.hwTsPushDiscardCnt++;
                    }
#endif
                }
                break;

                case CPSW_CPTS_EVENTTYPE_ETH_RECEIVE:
                {
                    buff = &hCpts->ethRxEventPool.eventMemPool[0U];
                    idx  = &hCpts->ethRxEventPool.index;

                    /* For Ethernet events, get the messageType, sequence ID, and port number */
                    buff[*idx].eventType = (CpswCpts_EventType)eventInfo.eventType;
                    buff[*idx].msgType = (CpswCpts_MsgType)eventInfo.msgType;
                    buff[*idx].seqId = eventInfo.seqId;
                    buff[*idx].portNum = eventInfo.portNo;
                    buff[*idx].domain  = eventInfo.domain;

                    /* These fields are meaningless for this event type */
                    buff[*idx].hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                    hCpts->eventStats.ethRxEventCnt++;
                    if (buff[*idx].tsVal != 0U)
                    {
                        hCpts->eventStats.ethRxEventDiscardCnt++;
                    }
#endif
                }
                break;

                case CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT:
                {
                    buff = &hCpts->ethTxEventPool.eventMemPool[0U];
                    idx  = &hCpts->ethTxEventPool.index;

                    /* For Ethernet events, get the messageType, sequence ID, and port number */
                    buff[*idx].eventType = (CpswCpts_EventType)eventInfo.eventType;
                    buff[*idx].msgType = (CpswCpts_MsgType)eventInfo.msgType;
                    buff[*idx].seqId = eventInfo.seqId;
                    buff[*idx].portNum = eventInfo.portNo;
                    buff[*idx].domain  = eventInfo.domain;

                    /* These fields are meaningless for this event type */
                    buff[*idx].hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                    hCpts->eventStats.ethTxEventCnt++;
                    if (buff[*idx].tsVal != 0U)
                    {
                        hCpts->eventStats.ethTxEventDiscardCnt++;
                    }
#endif
                }
                break;

                case CPSW_CPTS_EVENTTYPE_TS_COMP:
                {
                    buff = &hCpts->cmpEventPool.eventMemPool[0U];
                    idx  = &hCpts->cmpEventPool.index;

                    buff[*idx].eventType = (CpswCpts_EventType)eventInfo.eventType;

                    /* These fields are meaningless for this event type */
                    buff[*idx].portNum = 0U;
                    buff[*idx].seqId  = 0U;
                    buff[*idx].domain = 0U;
                    buff[*idx].hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
                    buff[*idx].msgType = CPSW_CPTS_MESSAGE_INVALID;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                    hCpts->eventStats.cmpEventCnt++;
                    if (buff[*idx].tsVal != 0U)
                    {
                        hCpts->eventStats.cmpEventDiscardCnt++;
                    }
#endif
                }
                break;

                case CPSW_CPTS_EVENTTYPE_TS_HOST_TX:
                {
                    buff = &hCpts->hostTxEventPool.eventMemPool[0U];
                    idx  = &hCpts->hostTxEventPool.index;

                    buff[*idx].eventType = (CpswCpts_EventType)eventInfo.eventType;
                    buff[*idx].msgType = (CpswCpts_MsgType)eventInfo.msgType;
                    buff[*idx].seqId = eventInfo.seqId;
                    buff[*idx].domain  = eventInfo.domain;

                    /* These fields are meaningless for this event type */
                    buff[*idx].portNum = 0U;
                    buff[*idx].hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                    hCpts->eventStats.ethHostTxEventCnt++;
                    if (buff[*idx].tsVal != 0U)
                    {
                        hCpts->eventStats.ethHostTxEventDiscardCnt++;
                    }
#endif
                }
                break;

                default:
                {
                    status = CPSW_EINVALID_PARAMS;
                }
                break;
            }

            if (CPSW_SOK == status)
            {
                /* Copy the 64-bit timestamp value to memory pool*/
                buff[*idx].tsVal = tsVal;

                if (NULL != hCpts->pEventNotifyCb)
                {
                    /* Notify the registered stack that a CPTS event has been detected */
                    (*hCpts->pEventNotifyCb)(hCpts->hEventNotifyCbArg, &buff[*idx]);
                }

                /* Increment the Index */
                *idx = (((*idx) + 1U) % CPSW_CFG_CPTS_EVENTS_POOL_SIZE);
            }
        }

        intrPendStatus = CSL_CPTS_isRawInterruptStatusBitSet(hCpts->cptsRegs);
    }
    while (intrPendStatus == true);

    /* Re-enable CPTS Interrupt */
    CSL_CPTS_enableInterrupt(hCpts->cptsRegs);
    return status;
}

static int32_t CpswCpts_clearEventMemPools(CpswCpts_Handle hCpts)
{
    int32_t status = CPSW_SOK;
    uintptr_t key = CpswOsal_disableAllIntr();

    if (NULL != hCpts)
    {
        memset(&hCpts->hwPushEventPool.eventMemPool[0U],
               0U,
               sizeof(hCpts->hwPushEventPool.eventMemPool));

        memset(&hCpts->ethTxEventPool.eventMemPool[0U],
               0U,
               sizeof(hCpts->ethTxEventPool.eventMemPool));

        memset(&hCpts->ethRxEventPool.eventMemPool[0U],
               0U,
               sizeof(hCpts->ethRxEventPool.eventMemPool));

        memset(&hCpts->cmpEventPool.eventMemPool[0U],
               0U,
               sizeof(hCpts->cmpEventPool.eventMemPool));

        memset(&hCpts->hostTxEventPool.eventMemPool[0U],
               0U,
               sizeof(hCpts->hostTxEventPool.eventMemPool));

        hCpts->hwPushEventPool.index = 0U;
        hCpts->ethTxEventPool.index  = 0U;
        hCpts->ethRxEventPool.index  = 0U;
        hCpts->cmpEventPool.index = 0U;
        hCpts->hostTxEventPool.index = 0U;
    }
    else
    {
        status = CPSW_EBADARGS;
    }

    CpswOsal_restoreAllIntr(key);
    return status;
}

static void CpswCpts_printRegs(CSL_cptsRegs *cptsRegs)
{
    uint32_t *regAddr = (uint32_t *)cptsRegs;
    uint32_t regIdx = 0;

    while (((uintptr_t)regAddr) < ((uintptr_t)cptsRegs + sizeof(CSL_cptsRegs)))
    {
        if (*regAddr)
        {
            CpswUtils_printf("CPTS: %u: 0x%08x\n", regIdx, *regAddr);
        }

        regAddr++;
        regIdx++;
    }
}

static void CpswCpts_printStats(CpswCpts_Handle hCpts)
{
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
    CpswUtils_printf("\nSoftware time stamp event count              : %lld \n", hCpts->eventStats.swTsPushEventCnt);
    CpswUtils_printf("Hardware time stamp push event count           : %lld \n", hCpts->eventStats.hwTsPushEventCnt);
    CpswUtils_printf("Compare time stamp event count                 : %lld \n", hCpts->eventStats.cmpEventCnt);
    CpswUtils_printf("Ethernet Rx time stamp count                   : %lld \n", hCpts->eventStats.ethRxEventCnt);
    CpswUtils_printf("Ethernet Tx time stamp count                   : %lld \n", hCpts->eventStats.ethTxEventCnt);
    CpswUtils_printf("Ethernet host Tx time stamp count              : %lld \n", hCpts->eventStats.ethHostTxEventCnt);
    CpswUtils_printf("Hardware time stamp lookup count               : %lld \n", hCpts->eventStats.hwTsPushLookupCnt);
    CpswUtils_printf("Compare event lookup count                     : %lld \n", hCpts->eventStats.cmpEventLookupCnt);
    CpswUtils_printf("Ethernet Rx time stamp lookup event count      : %lld \n", hCpts->eventStats.ethRxEventLookupCnt);
    CpswUtils_printf("Ethernet Tx time stamp lookup event count      : %lld \n", hCpts->eventStats.ethTxEventLookupCnt);
    CpswUtils_printf("Ethernet host Tx time stamp lookup event count : %lld \n", hCpts->eventStats.ethHostTxEventLookupCnt);
    CpswUtils_printf("Hardware time stamp discard count              : %lld \n", hCpts->eventStats.hwTsPushDiscardCnt);
    CpswUtils_printf("Compare event discard count                    : %lld \n", hCpts->eventStats.cmpEventDiscardCnt);
    CpswUtils_printf("Ethernet Rx time stamp discard event count     : %lld \n", hCpts->eventStats.ethRxEventDiscardCnt);
    CpswUtils_printf("Ethernet Tx time stamp discard event count     : %lld \n", hCpts->eventStats.ethTxEventDiscardCnt);
    CpswUtils_printf("Ethernet host Tx time stamp discard event count: %lld \n", hCpts->eventStats.ethHostTxEventDiscardCnt);
#endif
}

static void CpswCpts_calcPpmVal(uint64_t *tsPpmVal,
                                CpswCpts_PpmMode ppmMode)
{
    if (ppmMode == CPSW_CPTS_PPM_PARTS_PER_MILLION)
    {
        *tsPpmVal = (CPSW_CPTS_PPM / *tsPpmVal);
    }

    if (ppmMode == CPSW_CPTS_PPM_PARTS_PER_HOUR)
    {
        *tsPpmVal = (CPSW_CPTS_PPM_GIGAHERTZ_VAL / *tsPpmVal) * CPSW_CPTS_SECONDS_PER_HOUR;
    }

    if (*tsPpmVal <= CPSW_CPTS_PPM_MIN_VAL)
    {
        *tsPpmVal = CPSW_CPTS_PPM_MIN_VAL;
        CPSWTRACE_WARN(" Setting PPM to %u, as %u is less than minimum value(1024)\n",
                       CPSW_CPTS_PPM_MIN_VAL,
                       *tsPpmVal);
    }
}

static int32_t CpswCpts_lookUpEvent(CpswCpts_Handle hCpts,
                                    CpswCpts_LookUpEventInArgs *matchPrms,
                                    CpswCpts_Event *event)
{
    int32_t status = CPSW_SOK;
    uint32_t i = 0U;
    bool matchFound = false;
    CpswCpts_Event *buff = NULL;
    uintptr_t key = CpswOsal_disableAllIntr();

    switch (matchPrms->eventType)
    {
        case CPSW_CPTS_EVENTTYPE_HW_TS_PUSH:
        {
            buff = &hCpts->hwPushEventPool.eventMemPool[0U];
        }
        break;

        case CPSW_CPTS_EVENTTYPE_ETH_RECEIVE:
        {
            buff = &hCpts->ethRxEventPool.eventMemPool[0U];
        }
        break;

        case CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT:
        {
            buff = &hCpts->ethTxEventPool.eventMemPool[0U];
        }
        break;

        case CPSW_CPTS_EVENTTYPE_TS_COMP:
        {
            buff = &hCpts->cmpEventPool.eventMemPool[0U];
        }
        break;

        case CPSW_CPTS_EVENTTYPE_TS_HOST_TX:
        {
            buff = &hCpts->hostTxEventPool.eventMemPool[0U];
        }
        break;

        default:
        {
            status = CPSW_EINVALID_PARAMS;
        }
        break;
    }

    if (status == CPSW_SOK)
    {
        /* Traverse through memory pool for a matching event*/
        for (i = 0U; i < CPSW_CFG_CPTS_EVENTS_POOL_SIZE; i++)
        {
            switch (buff[i].eventType)
            {
                case CPSW_CPTS_EVENTTYPE_HW_TS_PUSH:
                {
                    if (matchPrms->hwPushNum == buff[i].hwPushNum)
                    {
                        matchFound = true;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                        hCpts->eventStats.hwTsPushLookupCnt++;
#endif
                    }
                }
                break;

                case CPSW_CPTS_EVENTTYPE_ETH_RECEIVE:
                case CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT:
                case CPSW_CPTS_EVENTTYPE_TS_HOST_TX:
                {
                    if ((matchPrms->msgType == buff[i].msgType) &&
                        (matchPrms->seqId == buff[i].seqId) &&
                        (matchPrms->domain == buff[i].domain) &&
                        (matchPrms->portNum == (buff[i].portNum - 1U)))
                    {
                        matchFound = true;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                        if (matchPrms->eventType == CPSW_CPTS_EVENTTYPE_ETH_RECEIVE)
                        {
                            hCpts->eventStats.ethRxEventLookupCnt++;
                        }
                        else if (matchPrms->eventType == CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT)
                        {
                            hCpts->eventStats.ethTxEventLookupCnt++;
                        }
                        else if (matchPrms->eventType == CPSW_CPTS_EVENTTYPE_TS_HOST_TX)
                        {
                            hCpts->eventStats.ethHostTxEventLookupCnt++;
                        }
#endif
                    }
                }
                break;

                case CPSW_CPTS_EVENTTYPE_TS_COMP:
                {
                    if (matchPrms->eventType == buff[i].eventType)
                    {
                        matchFound = true;
#if defined(CPSWCPTS_INSTRUMENTATION_ENABLED)
                        hCpts->eventStats.cmpEventLookupCnt++;
#endif
                    }
                }
                break;

                case CPSW_CPTS_EVENTTYPE_TS_PUSH:
                case CPSW_CPTS_EVENTTYPE_TS_ROLLOVER:
                case CPSW_CPTS_EVENTTYPE_TS_HALFROLLOVER:
                case CPSW_CPTS_EVENTTYPE_INVALID:
                {
                    status = CPSW_ENOTSUPPORTED;
                }
                break;
            }

            if (matchFound == true)
            {
                break;
            }
        }

        if (matchFound == true)
        {
            /* Copy the event information */
            *event = buff[i];

            /* Remove the Event from pool after copying */
            memset(&buff[i], 0U, sizeof(CpswCpts_Event));
            status = CPSW_SOK;
        }
        else
        {
            status = CPSW_ENOTFOUND;
        }
    }

    CpswOsal_restoreAllIntr(key);
    return status;
}

static int32_t CpswCpts_checkTsCompCompatibility(CpswCpts_Handle hCpts)
{
    int32_t status = CPSW_SOK;
    uint32_t tsAddVal = 0U;
    uint32_t tsPpmVal[2] = {0U, 0U};

    tsAddVal = CSL_CPTS_getTSAddVal(hCpts->cptsRegs);
    CSL_CPTS_getTSPpm(hCpts->cptsRegs, tsPpmVal);

    /* Time stamp compare feature is not compatible with
     * time stamp add value or time stamp PPM
     */
    if ((tsAddVal != 0U) ||
        (tsPpmVal[0] != 0U) ||
        (tsPpmVal[1] != 0U))
    {
        status = CPSW_ENOTSUPPORTED;
    }

    return status;
}

static int32_t CpswCpts_checkGenfEstfErrata(CpswCpts_Handle hCpts,
                                            uint32_t index,
                                            bool isGenf)
{
    int32_t status = CPSW_SOK;
    uint32_t ipErrataBitMap = 0U;
    uint32_t length = 0U;

    ipErrataBitMap = CpswUtils_getIpErrataBitMap(hCpts->cpswType);

    if ((CPSW_ERRATA_GENFN_RECONFIG & ipErrataBitMap) != 0U)
    {
        if (isGenf == true)
        {
            CSL_CPTS_getGENFnLength(hCpts->cptsRegs, index, &length);
        }
        else
        {
            CSL_CPTS_getESTFnLength(hCpts->cptsRegs, index, &length);
        }

        if (length != 0U)
        {
            status = CPSW_ENOTSUPPORTED;
            CPSWTRACE_WARN("Re-configuring GENFn/ESTFn is not supported for this "
                                "CPTS IP version\n");
        }
    }

    return status;
}

static int32_t CpswCpts_checkSetGenfInArgs(CpswCpts_Handle hCpts, CpswCpts_SetFxnGenInArgs *setGenFInArgs)
{
    int32_t status = CPSW_SOK;

    status = CpswCpts_checkGenfEstfErrata(hCpts, setGenFInArgs->index, true);
    if (status == CPSW_SOK)
    {
        if (setGenFInArgs->ppmMode != CPSW_CPTS_PPM_DISABLE)
        {
            CpswCpts_calcPpmVal(&setGenFInArgs->ppmVal,
                                setGenFInArgs->ppmMode);

            if (setGenFInArgs->length < CPSW_CPTS_GENFN_LENGTH_MIN_VAL)
            {
                status = CPSW_EINVALID_PARAMS;
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "Invalid %u GENFn length value, it should be > 5 \n", setGenFInArgs->length);
            }
        }
        else
        {
            setGenFInArgs->ppmVal = 0U;
        }
    }

    return status;
}

static int32_t CpswCpts_checkSetEstfInArgs(CpswCpts_Handle hCpts, CpswCpts_SetFxnGenInArgs *setESTFInArgs)
{
    int32_t status = CPSW_SOK;

    status = CpswCpts_checkGenfEstfErrata(hCpts, setESTFInArgs->index, false);
    if (status == CPSW_SOK)
    {
        if (setESTFInArgs->ppmMode != CPSW_CPTS_PPM_DISABLE)
        {
            CpswCpts_calcPpmVal(&setESTFInArgs->ppmVal,
                                setESTFInArgs->ppmMode);
        }
        else
        {
            setESTFInArgs->ppmVal = 0U;
        }
    }

    return status;
}
