/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_dma.c
 *
 *  \brief This file contains CPPI-related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <ti/csl/arch/csl_arch.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/cpsw/include/cpsw_queue.h>
#include "priv/cpsw_priv.h"
#include "priv/cpsw_osal.h"
#include "priv/cpsw_dma_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static CpswDma_DrvObj *CpswDma_getDmaObj(Cpsw_Type cpswType);

static int32_t CpswDma_checkTxChParams(CpswDma_OpenTxChPrms *pTxChPrms);

static int32_t CpswDma_checkRxFlowParams(CpswDma_OpenRxFlowPrms *pRxFlowPrms);

static void CpswDma_drainIsrCq(queue_t *dstQ,
                               queue_t *isrCq);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW DMA objects */
CpswDma_DrvObj gCpswDmaObj[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswDma_initParams(CpswDma_Config *pDmaConfig)
{
    pDmaConfig->hUdmaDrv                 = NULL;
    pDmaConfig->rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;
#ifdef QNX_OS
    pDmaConfig->rxChInitPrms.preferredChNum = UDMA_DMA_CH_ANY;
#endif
}

CpswDma_Handle CpswDma_open(Cpsw_Type cpswType,
                            const CpswDma_Config *pCpswDmaConfig)
{
    CpswDma_DrvObj *pCpswDmaObj = NULL;
    int32_t retVal;

    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_DMA_RING_MEM_SIZE == sizeof(uint64_t));

    /* Error check */
    retVal = UDMA_EBADARGS;
    if (NULL != pCpswDmaConfig)
    {
        if (NULL != pCpswDmaConfig->hUdmaDrv)
        {
            retVal = UDMA_SOK;
        }
        else
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] pCpswDmaConfig has null UDMA handle!!\n");
        }
    }
    else
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] pCpswDmaConfig NULL !!\n");
    }

    if (UDMA_SOK == retVal)
    {
        pCpswDmaObj = CpswDma_getDmaObj(cpswType);
        memset(pCpswDmaObj, 0U, sizeof(*pCpswDmaObj));

        pCpswDmaObj->hUdmaDrv = pCpswDmaConfig->hUdmaDrv;

        pCpswDmaObj->initFlag         = true;
        pCpswDmaObj->cpswType         = cpswType;
        pCpswDmaObj->rxChObj.cpswType = cpswType;

        /* Initialize memory manager module managing driver object memories */
        CpswDma_memMgrInit();
    }

    return pCpswDmaObj;
}

int32_t CpswDma_close(CpswDma_Handle hCpswDma)
{
    int32_t retVal = UDMA_SOK;

    // TODO check all Rx & Tx channels & are closed
    /* Error check */
    if (hCpswDma == NULL)
    {
        CPSWTRACE_ERROR("[CPSW DMA] CPSW DMA handle is NULL!! \n");
        retVal = UDMA_EBADARGS;
    }

    if (UDMA_SOK == retVal)
    {
        if (hCpswDma->initFlag == false)
        {
            CPSWTRACE_ERROR("[CPSW DMA] DMA is not initialized before close !! \n");
            retVal = UDMA_EFAIL;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* De-initialize driver memory manager */
        CpswDma_memMgrDeInit();
        hCpswDma->initFlag = false;
    }

    return retVal;
}

int32_t CpswDma_openRxCh(CpswDma_Handle hCpswDma,
                         const CpswDma_RxChInitPrms *pRxChInitPrms,
                         uint32_t totalRxFlowCount)
{
    int32_t retVal;
    Udma_DrvHandle hUdmaDrv;
    CpswDma_RxChObj *pRxCh;
    Udma_ChHandle hUdmaCh;
    Udma_FlowHandle hFlow;
    Udma_ChPrms chPrms;
    Udma_ChRxPrms rxPrms;

    bool allocFlow = false, chOpenFlag = false;
    uint32_t chType;
    uintptr_t intrKey;
    const CpswSoc_Config *pSocCfg;
    void *pTdCqRingMem;

    /* Error check */
    retVal = UDMA_EBADARGS;
    if (hCpswDma != NULL)
    {
        if (hCpswDma->initFlag)
        {
            retVal = UDMA_SOK;
        }
        else
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] CPSW DMA driver not initialized !!\n");
        }
    }
    else
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] CPSW DMA handle is NULL !!\n");
    }

    if (UDMA_SOK == retVal)
    {
        pSocCfg = CpswSoc_getConfig(hCpswDma->cpswType);
        if (totalRxFlowCount > pSocCfg->rxFlowCount)
        {
            CPSWTRACE_ERROR("[CPSW DMA] Invalid Rx channel flow count "
                            "Requested = %d, SOC Available = %d !! \n",
                            totalRxFlowCount, pSocCfg->rxFlowCount);
            /* flows num greater than CPSW instance max. supported */
            retVal = UDMA_EBADARGS;
        }
    }

    intrKey = CpswOsal_disableAllIntr();

    if (UDMA_SOK == retVal)
    {
        hUdmaDrv = hCpswDma->hUdmaDrv;
        pRxCh    = &hCpswDma->rxChObj;
        hUdmaCh  = &pRxCh->udmaChObj;
        chType   = UDMA_CH_TYPE_RX;

        /* Initialize channel params (PSI-L thread and RingAcc memories) */
        UdmaChPrms_init(&chPrms, chType);
        chPrms.peerChNum = CpswDma_getRxChPeerThreadId(hCpswDma->cpswType);
#ifdef QNX_OS
        chPrms.chNum     = pRxChInitPrms->preferredChNum;
#else
        chPrms.chNum     = UDMA_DMA_CH_ANY;
#endif

        UdmaRingPrms_init(&chPrms.tdCqRingPrms);
        pTdCqRingMem = CpswDma_memMgrAllocTdCqRingMemObj();
        if (pTdCqRingMem != NULL)
        {
            chPrms.tdCqRingPrms.ringMem = pTdCqRingMem;
            chPrms.tdCqRingPrms.elemCnt = CPSW_DMA_TDCQ_RING_ELE_CNT;
            chPrms.tdCqRingPrms.mode    = CSL_RINGACC_RING_MODE_RING;

            /* Open the UDMA channel */
            retVal = Udma_chOpen(hUdmaDrv, hUdmaCh, chType, &chPrms);
            CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                               "[CPSW DMA] UDMA RX Channel open failed: 0x%x\n", retVal);
            if (UDMA_SOK == retVal)
            {
                chOpenFlag = true;
            }
        }
        else
        {
            CPSWTRACE_ERROR("[CPSW DMA] Teardown CQ ring memory allocation for Rx channel failed\n");
            retVal = UDMA_EALLOC;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Configure RX parameters */
        hFlow  = &pRxCh->flowUdmaObj;
        retVal = Udma_flowAlloc(hUdmaDrv, hFlow, totalRxFlowCount);

        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] UDMA flow alloc failed!!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocFlow = true;

            if (totalRxFlowCount != Udma_flowGetCount(hFlow))
            {
                CPSWTRACE_ERROR("[CPSW DMA] Allocated flow Count doesn't match num requested!!\n");
                retVal = UDMA_EALLOC;
            }
        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaChRxPrms_init(&rxPrms, chType);

        rxPrms.dmaPriority        = pRxChInitPrms->dmaPriority;
        rxPrms.flowIdFwRangeStart = Udma_flowGetNum(hFlow);
        rxPrms.flowIdFwRangeCnt   = Udma_flowGetCount(hFlow);

        rxPrms.configDefaultFlow = FALSE;

        retVal = Udma_chConfigRx(hUdmaCh, &rxPrms);
        if (UDMA_SOK == retVal)
        {
            /* Enable the UDMA channel */
            retVal = Udma_chEnable(&pRxCh->udmaChObj);
        }
        else
        {
            CPSWTRACE_ERROR("[CPSW DMA] UDMA RX channel enable failed, closing channel !!\n");
            retVal     = Udma_chClose(hUdmaCh);
            chOpenFlag = false;
        }
    }

    /* Get teardown cqRing handle */
    if (UDMA_SOK == retVal)
    {
        pRxCh->tdCqRing = Udma_chGetTdCqRingHandle(hUdmaCh);

        if(NULL == pRxCh->tdCqRing)
        {
            retVal = UDMA_EALLOC;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Save channel config */
        hCpswDma->rxChObj.rxChInitPrms = *pRxChInitPrms;
        pRxCh->initFlag                = true;
        pRxCh->cpswType                = hCpswDma->cpswType;
    }
    else
    {
        /* Error. Free-up resource if allocated */
        if (chOpenFlag)
        {
            retVal = Udma_chClose(hUdmaCh);
        }

        if (allocFlow)
        {
            retVal = Udma_flowFree(&pRxCh->flowUdmaObj);
        }

        /* return error as channel open has failed */
        retVal = UDMA_EFAIL;
    }

    CpswOsal_restoreAllIntr(intrKey);

    return retVal;
}

int32_t CpswDma_closeRxCh(CpswDma_Handle hCpswDma)
{
    int32_t retVal = UDMA_SOK;
    CpswDma_RxChObj *pRxCh;
    CSL_UdmapTdResponse tdResp;
    uintptr_t intrKey;
    void *tdCqRingMemPtr = NULL;

    // TODO how to check all flows are closed?
    // TODO check all Rx flows & Tx channels & are closed

    /* Error check */
    if (hCpswDma == NULL)
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] DMA handle is NULL !!\n");
        retVal = UDMA_EBADARGS;
    }

    if ((UDMA_SOK == retVal) && (!hCpswDma->initFlag))
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] DMA not initialized!!\n");
        retVal = UDMA_EBADARGS;
    }

    if (UDMA_SOK == retVal)
    {
        pRxCh = &hCpswDma->rxChObj;
        if (!pRxCh->initFlag)
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] CPSW RX Channel not initialized!!\n");
            retVal = UDMA_EFAIL;
        }
    }

    intrKey = CpswOsal_disableAllIntr();

    if (UDMA_SOK == retVal)
    {
        /* Disable the UDMA channel */
        retVal += Udma_chDisable(&pRxCh->udmaChObj, CPSW_UDMA_TEARDOWN_TIME_MS);

        /* Dequeue the teardown response */
        if (UDMA_SOK == retVal)
        {
            tdCqRingMemPtr = Udma_ringGetMemPtr(pRxCh->tdCqRing);

            do
            {
                retVal = Udma_chDequeueTdResponse(&pRxCh->udmaChObj, &tdResp);
            }
            while (UDMA_SOK != retVal);

            if (FALSE == tdResp.tdIndicator)
            {
                CPSWTRACE_ERROR("[CPSW DMA] RX channel TearDown failed!!\n");
                retVal += UDMA_EFAIL;
            }

            /* Free tdCq ring memory */
            CpswDma_memMgrFreeTdCqRingMemObj(tdCqRingMemPtr);
        }
    }

    // TODO flush FQ ring

    if (UDMA_SOK == retVal)
    {
        /* free all flow ids allocated for this channel */
        retVal += Udma_flowFree(&pRxCh->flowUdmaObj);
    }

    if (UDMA_SOK == retVal)
    {
        /* Close the UDMA channel */
        retVal += Udma_chClose(&pRxCh->udmaChObj);
    }

    CpswOsal_restoreAllIntr(intrKey);

    return retVal;
}

void CpswDma_initRxFlowParams(CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    Udma_FlowPrms flowPrms;

    memset(&flowPrms, 0, sizeof(flowPrms));
    UdmaFlowPrms_init(&flowPrms, UDMA_CH_TYPE_RX);

    pRxFlowPrms->flowPrms.psInfoPresent = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PSINFO_PRESENT;
    pRxFlowPrms->flowPrms.einfoPresent  = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_EINFO_PRESENT;
    pRxFlowPrms->flowPrms.sopOffset    = flowPrms.sopOffset;
    pRxFlowPrms->flowPrms.defaultRxCQ  = flowPrms.defaultRxCQ;
    pRxFlowPrms->flowPrms.srcTagHi     = flowPrms.srcTagHi;
    pRxFlowPrms->flowPrms.srcTagLo     = flowPrms.srcTagLo;
    pRxFlowPrms->flowPrms.srcTagHiSel  = flowPrms.srcTagHiSel;
    pRxFlowPrms->flowPrms.srcTagLoSel  = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_SRC_SELECT_SRC_TAG;
    pRxFlowPrms->flowPrms.destTagHi    = flowPrms.destTagHi;
    pRxFlowPrms->flowPrms.destTagLo    = flowPrms.destTagLo;
    pRxFlowPrms->flowPrms.destTagHiSel = flowPrms.destTagHiSel;
    pRxFlowPrms->flowPrms.destTagLoSel = flowPrms.destTagLoSel;

    pRxFlowPrms->udmaChPrms.fqRingPrms.orderId    = UDMA_DEFAULT_RING_ORDER_ID;
    /* Use message mode for Rx FQ so at channel/flow teardown time UDMA IP can push
     * inflow descriptors back into the ring.*/
    pRxFlowPrms->udmaChPrms.fqRingPrms.mode       = CSL_RINGACC_RING_MODE_MESSAGE;
    pRxFlowPrms->udmaChPrms.fqRingPrms.useRingMon = false;

    pRxFlowPrms->udmaChPrms.cqRingPrms.orderId    = UDMA_DEFAULT_RING_ORDER_ID;
    pRxFlowPrms->udmaChPrms.cqRingPrms.mode       = CSL_RINGACC_RING_MODE_RING;
    pRxFlowPrms->udmaChPrms.cqRingPrms.useRingMon = false;

    /* We initialize callback function pointers to NULL so we check app pass
     * uninitialized  value */
    pRxFlowPrms->hUdmaDrv        = NULL;
    pRxFlowPrms->ringMemAllocFxn = NULL;
    pRxFlowPrms->ringMemFreeFxn  = NULL;
    pRxFlowPrms->dmaDescAllocFxn = NULL;
    pRxFlowPrms->dmaDescFreeFxn  = NULL;

    pRxFlowPrms->notifyCb = NULL;
    pRxFlowPrms->hCbArg   = NULL;

    pRxFlowPrms->useProxy = false;

    /* auto-reclaim initialization */
    pRxFlowPrms->autoReclaimPrms.enableFlag       = false;
    pRxFlowPrms->autoReclaimPrms.hDmaDescPool          = NULL;
    pRxFlowPrms->autoReclaimPrms.hReclaimRing = NULL;

    return;
}

CpswDma_RxFlowHandle CpswDma_openRxFlow(CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    int32_t retVal;
    Udma_RingHandle dropFqRing = NULL, localHandle;
    Udma_FlowPrms flowPrms;
    Udma_RingPrms ringPrms;
    CpswDma_RxFlowObj *pRxFlow;
    CpswDma_udmaInfo udmaInfo;
    CpswDma_ringAllocInfo ringAllocInfo;
    uint32_t eventType;
    CpswDma_UdmaRingPrms *pFqRingPrms = &pRxFlowPrms->udmaChPrms.fqRingPrms;
    CpswDma_AutoReclaimPrms *reclaimPrms = &pRxFlowPrms->autoReclaimPrms;

    bool allocFlowObj = false, allocFqRing = false, allocDropFqRing = false;
    bool allocCqRing  = false, flowAttachFlag = false;
    bool allocProxy   = false, allocRingMon = false;
    uintptr_t intrKey;
    uint8_t flowStart;

    intrKey = CpswOsal_disableAllIntr();

    /* Set to NULL so if error condition we return NULL */
    pRxFlow = NULL;

    /* Error check */
    retVal = CpswDma_checkRxFlowParams(pRxFlowPrms);

    if (UDMA_SOK == retVal)
    {
        pRxFlow = CpswDma_memMgrAllocRxFlowObj();

        if (pRxFlow == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA] Memory allocation for Rx flow object failed !!\n");
            retVal = UDMA_EALLOC;
        }
        else
        {
            allocFlowObj = true;
            memset(pRxFlow, 0U, sizeof(*pRxFlow));
            /* Save Flow config */
            pRxFlow->rxFlowPrms = *pRxFlowPrms;
            /* Initialize local variables to be used in reminder of this function  */

            pRxFlow->hUdmaDrv  = pRxFlowPrms->hUdmaDrv;
            dropFqRing         = &pRxFlow->rxFlowMemObj.dropFqRingObj;
            pRxFlow->hUdmaFlow = &pRxFlow->rxFlowMemObj.flowUdmaObj;
            pRxFlow->fqRing    = &pRxFlow->rxFlowMemObj.fqRingObj;
            pRxFlow->hUdmaEvt  = &pRxFlow->rxFlowMemObj.udmaEvtObj;

            if (!reclaimPrms->enableFlag)
            {
                pRxFlow->cqRing       = &pRxFlow->rxFlowMemObj.cqRingObj;
                pRxFlow->hDmaDescPool = &pRxFlow->rxFlowMemObj.rxReadyDmaDescQ;
            }
            else
            {
#if (1U == CPSW_CFG_DEV_ERROR)
                Cpsw_assert(NULL != reclaimPrms->hDmaDescPool);
                Cpsw_assert(NULL != reclaimPrms->hReclaimRing);
                /* Ring mode should be message as both RX flow (pushes to CQ) and
                 * TX channel (pops from FQ) uses this ring */
                Cpsw_assert(CSL_RINGACC_RING_MODE_MESSAGE == Udma_ringGetMode(reclaimPrms->hReclaimRing));
                /* Filtering of protocol specific and extended info should be enabled as this
                 * information might corrupt TX channel (as the TX PSI and EINFO gets mapped to
                 * RX flow PSI and EINFO which has entirely different meaning)*/
                Cpsw_assert(pRxFlowPrms->flowPrms.psInfoPresent ==
                            TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PSINFO_PRESENT);
                Cpsw_assert(pRxFlowPrms->flowPrms.einfoPresent ==
                            TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_EINFO_PRESENT);
                /* Notify callback should be NULL when auto recycle is enabled as the
                 * CQ event should be disabled */
                Cpsw_assert(pRxFlowPrms->notifyCb == NULL);
#endif
                pRxFlow->cqRing       = reclaimPrms->hReclaimRing;
                pRxFlow->hDmaDescPool = reclaimPrms->hDmaDescPool;
            }
        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaRingPrms_init(&ringPrms);
        ringPrms.elemCnt = pRxFlowPrms->numRxPkts;
        ringPrms.orderId = pRxFlowPrms->udmaChPrms.fqRingPrms.orderId;
        ringPrms.mode    = pRxFlowPrms->udmaChPrms.fqRingPrms.mode;

        ringAllocInfo.allocRingMem    = true;
        ringAllocInfo.ringMemAllocFxn = pRxFlowPrms->ringMemAllocFxn;
        ringAllocInfo.ringMemFreeFxn  = pRxFlowPrms->ringMemFreeFxn;
        ringAllocInfo.hCbArg          = pRxFlowPrms->hCbArg;
        ringAllocInfo.ringNum         = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pRxFlow->hUdmaDrv,
                                   pRxFlow->fqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Fq ring allocation failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocFqRing = true;
        }
    }

    if ((UDMA_SOK == retVal) && (pFqRingPrms->useRingMon))
    {
        pRxFlow->hUdmaRingMon = CpswDma_allocRingMon(pRxFlow->hUdmaDrv,
                                                     pRxFlow->fqRing,
                                                     pRxFlowPrms->numRxPkts,
                                                     &pFqRingPrms->ringMonCfg);
        if (NULL == pRxFlow->hUdmaRingMon)
        {
            CPSWTRACE_ERROR("[CPSW DMA] RX FQ Ring Mon Config failed !!\n");
            retVal = UDMA_EALLOC;
        }
        else
        {
            allocRingMon = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* PDK-3723 We use drop ring for routing packets larger than flow max. allowed size
         * to avoid invalid data getting spilled into multiple descriptors. */
        UdmaRingPrms_init(&ringPrms);
        ringPrms.elemCnt = 1U;
        ringPrms.mode    = CSL_RINGACC_RING_MODE_RING;

        ringAllocInfo.allocRingMem    = true;
        ringAllocInfo.ringMemAllocFxn = pRxFlowPrms->ringMemAllocFxn;
        ringAllocInfo.ringMemFreeFxn  = pRxFlowPrms->ringMemFreeFxn;
        ringAllocInfo.hCbArg          = pRxFlowPrms->hCbArg;
        ringAllocInfo.ringNum         = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pRxFlow->hUdmaDrv,
                                   dropFqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Drop Fq ring allocation failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocDropFqRing = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        if (pRxFlowPrms->useProxy)
        {
            retVal = CpswDma_allocProxy(pRxFlow->hUdmaDrv, &pRxFlow->rxFlowMemObj.proxyObj, pRxFlow->fqRing);
            if (UDMA_SOK == retVal)
            {
                allocProxy          = true;
                pRxFlow->hUdmaProxy = &pRxFlow->rxFlowMemObj.proxyObj;
            }
        }
        else
        {
            pRxFlow->hUdmaProxy = NULL;
        }
    }

    if ((!reclaimPrms->enableFlag) && (UDMA_SOK == retVal))
    {
        UdmaRingPrms_init(&ringPrms);
        ringPrms.elemCnt = pRxFlowPrms->numRxPkts;
        ringPrms.orderId = pRxFlowPrms->udmaChPrms.cqRingPrms.orderId;
        ringPrms.mode    = pRxFlowPrms->udmaChPrms.cqRingPrms.mode;

        ringAllocInfo.allocRingMem    = true;
        ringAllocInfo.ringMemAllocFxn = pRxFlowPrms->ringMemAllocFxn;
        ringAllocInfo.ringMemFreeFxn  = pRxFlowPrms->ringMemFreeFxn;
        ringAllocInfo.hCbArg          = pRxFlowPrms->hCbArg;
        ringAllocInfo.ringNum         = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pRxFlow->hUdmaDrv,
                                   pRxFlow->cqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Cq ring allocation failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocCqRing = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        memset(&flowPrms, 0, sizeof(flowPrms));
        UdmaFlowPrms_init(&flowPrms, UDMA_CH_TYPE_RX);

        /* UdmaFlowPrms_init sets errorHandling to retry(1) by default,
         * in case of descriptor/buffer starvation UDMA retries unless it
         * gets a descriptor. In case of multi flow, this results in bottom
         * of FIFO drop, to avoid this errorHandling must be set to drop(0).
         */
        flowPrms.errorHandling = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_ERR_DROP;

        /* CPPI TX Status Data Word [0..3] are mapped to Host Packet Descriptor Protocol Specific
         * Words if RFLOW[a]_RFA.rx_psinfo_present = 1 and RFLOW[a]_RFA.rx_ps_location = 0 (end) */
        flowPrms.psInfoPresent = pRxFlowPrms->flowPrms.psInfoPresent;
        /* Extended packet info block present CPPI TX info words are mapped */
        flowPrms.einfoPresent  = pRxFlowPrms->flowPrms.einfoPresent;
        flowPrms.psLocation    = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PS_END_PD;

        /* Set Src tag low selection to get port number of the received packet (or)
         * flow Id (or) config tag from CPPI. */
        flowPrms.srcTagLoSel   = pRxFlowPrms->flowPrms.srcTagLoSel;

        flowPrms.defaultRxCQ = pRxFlow->cqRing->ringNum;

        /* Enable Rx Packet Size Threshold so we can route packets larger than
         * pRxFlowPrms->rxFlowMtu to drop ring */
        flowPrms.sizeThreshEn = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_SIZE_THRESH_MAX;

        /* Align the RX MTU size to avoid truncation. Refer to CPSW_DMA_RXMTU_ALIGN description for
         * more details */
        flowPrms.sizeThresh0 = CPSW_UTILS_ALIGN(pRxFlowPrms->rxFlowMtu, CPSW_DMA_RXMTU_ALIGN);
        flowPrms.sizeThresh1 = CPSW_UTILS_ALIGN(CPSW_DMA_JUMBO_PACKET_SIZE, CPSW_DMA_RXMTU_ALIGN);
        flowPrms.sizeThresh2 = CPSW_UTILS_ALIGN(CPSW_DMA_JUMBO_PACKET_SIZE, CPSW_DMA_RXMTU_ALIGN);

        flowPrms.fdq0Sz0Qnum = pRxFlow->fqRing->ringNum;
        flowPrms.fdq0Sz1Qnum = dropFqRing->ringNum;
        flowPrms.fdq0Sz2Qnum = dropFqRing->ringNum;
        flowPrms.fdq0Sz3Qnum = dropFqRing->ringNum;
        flowPrms.fdq1Qnum    = dropFqRing->ringNum;
        flowPrms.fdq2Qnum    = dropFqRing->ringNum;
        flowPrms.fdq3Qnum    = dropFqRing->ringNum;

        flowStart = pRxFlowPrms->startIdx + pRxFlowPrms->flowIdx;

        /* Attach and configure the flows */
        retVal = Udma_flowAttach(pRxFlow->hUdmaDrv,
                                 pRxFlow->hUdmaFlow,
                                 flowStart,
                                 1U /* flowCnt */);

        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Flow attach failed!!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            flowAttachFlag = true;
            retVal         = Udma_flowConfig(pRxFlow->hUdmaFlow,
                                             0U, /* Flow Index */
                                             &flowPrms);
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Initialize the completion queue.
         * Must be done before enabling CQ events */
        CpswUtils_initQ(&pRxFlow->cqIsrQ);
    }

    if (UDMA_SOK == retVal)
    {
        if (!reclaimPrms->enableFlag)
        {
            CpswDma_initRxFreeDescQ(pRxFlow);
        }

        CPSWTRACE_VERBOSE_IF((pTxChPrms->notifyCb == NULL),
                    "[CPSW DMA] Event callback function is null !!\n");
        if (pRxFlowPrms->notifyCb != NULL)
        {
            if (pFqRingPrms->useRingMon && allocRingMon)
            {
                /* If ring mon is enabled on FQ, we register event for FQ low
                 * threshold and not CQ */
                eventType   = UDMA_EVENT_TYPE_RING_MON;
                localHandle = pRxFlow->fqRing;
            }
            else
            {
                eventType   = UDMA_EVENT_TYPE_RING;
                localHandle = pRxFlow->cqRing;
            }

            CpswDma_getRxFlowUdmaInfo(pRxFlow, &udmaInfo);
            retVal = CpswDma_registerEvent(&udmaInfo,
                                           localHandle,
                                           CpswDma_rxCqIsr,
                                           pRxFlow,
                                           eventType);
            CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Rx Flow Open: event registration failed!!: 0x%x\n", retVal);
            if (CPSW_SOK == retVal)
            {
                pRxFlow->evtInitFlag = true;
            }
        }
    }

    if (UDMA_SOK != retVal)
    {
        if (pRxFlow->evtInitFlag)
        {
            retVal = CpswDma_unregisterEvent(pRxFlow->hUdmaEvt);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        /* Error. Free-up resource if allocated */
        if (allocFqRing)
        {
            retVal = CpswDma_freeRing(pRxFlow->fqRing,
                                      pRxFlow->rxFlowPrms.numRxPkts,
                                      pRxFlow->rxFlowPrms.ringMemFreeFxn,
                                      &pRxFlow->rxFlowPrms.hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocDropFqRing)
        {
            retVal = CpswDma_freeRing(&pRxFlow->rxFlowMemObj.dropFqRingObj,
                                      1U,
                                      pRxFlow->rxFlowPrms.ringMemFreeFxn,
                                      &pRxFlow->rxFlowPrms.hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocCqRing)
        {
            /* We free cqRing without retrieving packets here as packets are
             * not submitted by application yet so there wouldn't be any packets
             * in the cqRing */
            retVal = CpswDma_freeRing(pRxFlow->cqRing,
                                      pRxFlow->rxFlowPrms.numRxPkts,
                                      pRxFlow->rxFlowPrms.ringMemFreeFxn,
                                      &pRxFlow->rxFlowPrms.hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocRingMon)
        {
            retVal = CpswDma_freeRingMon(pRxFlow->hUdmaRingMon);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (flowAttachFlag)
        {
            retVal = Udma_flowDetach(pRxFlow->hUdmaFlow);
            Cpsw_assert(UDMA_SOK == retVal);

            /* Free up ready DMA descriptors */
            if (!reclaimPrms->enableFlag)
            {
                CpswDma_deInitRxFreeDescQ(pRxFlow);
            }
        }

        if (allocProxy)
        {
            retVal = CpswDma_freeProxy(pRxFlow->hUdmaProxy);
            Cpsw_assert(retVal == UDMA_SOK);
            pRxFlow->hUdmaProxy = NULL;
        }

        /* Free the rxFlowObj last */
        if (allocFlowObj)
        {
            CpswDma_memMgrFreeRxFlowObj(pRxFlow);
        }

        /* As flow open is failed return NULL */
        pRxFlow = NULL;
    }
    else
    {
        pRxFlow->initFlag    = true;
    }

    CpswOsal_restoreAllIntr(intrKey);

    return pRxFlow;
}

int32_t CpswDma_closeRxFlow(CpswDma_RxFlowHandle hRxFlow,
                            CpswDma_PktInfoQ *pFqPktInfoQ,
                            CpswDma_PktInfoQ *pCqPktInfoQ)
{
    int32_t retVal = UDMA_SOK;
    uintptr_t intrKey;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hRxFlow) ||
        (NULL == pFqPktInfoQ) ||
        (NULL == pCqPktInfoQ))
    {
        CPSWTRACE_ERROR_IF((NULL == hRxFlow), "[CPSW DMA] hRxFlow is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFqPktInfoQ), "[CPSW DMA] pFqPktInfoQ is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFqPktInfoQ), "[CPSW DMA] pFqPktInfoQ is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
        intrKey = CpswOsal_disableAllIntr();

        /* Error check */
        if (hRxFlow == NULL)
        {
            retVal = UDMA_EBADARGS;
        }

        if (UDMA_SOK == retVal)
        {
            if (hRxFlow->evtInitFlag)
            {
                /* Disable event */
                retVal = CpswDma_disableRxEvent(hRxFlow);
                Cpsw_assert(UDMA_SOK == retVal);
            }

            /* Flush packets in Rx FQ - retrieve any packets that haven't been
             * processed yet */
            retVal += CpswDma_flushRxFlowRing(hRxFlow, hRxFlow->fqRing, pFqPktInfoQ);
            if (UDMA_SOK == retVal)
            {
                retVal = CpswDma_freeRing(hRxFlow->fqRing,
                                          hRxFlow->rxFlowPrms.numRxPkts,
                                          hRxFlow->rxFlowPrms.ringMemFreeFxn,
                                          &hRxFlow->rxFlowPrms.hCbArg);
                Cpsw_assert(UDMA_SOK == retVal);

                retVal = CpswDma_freeRing(&hRxFlow->rxFlowMemObj.dropFqRingObj,
                                          1U,
                                          hRxFlow->rxFlowPrms.ringMemFreeFxn,
                                          &hRxFlow->rxFlowPrms.hCbArg);
                Cpsw_assert(UDMA_SOK == retVal);
            }
        }

        if (UDMA_SOK == retVal)
        {
            /* Retrieve packets in Rx CQ so app can free them */
            retVal += CpswDma_flushRxFlowRing(hRxFlow, hRxFlow->cqRing, pCqPktInfoQ);
            if ((UDMA_SOK == retVal) && (hRxFlow->evtInitFlag))
            {
                /* Return if any packets retrieved in the ISR */
                CpswUtils_appendQ(pCqPktInfoQ, &hRxFlow->cqIsrQ);

                /* Unregister event - make sure you unregister the event after flushing
                 * the ring. This is because unregister event resets the ring and all the
                 * descriptors in the ring gets discarded causing resource leakage */
                retVal              += CpswDma_unregisterEvent(hRxFlow->hUdmaEvt);
                hRxFlow->evtInitFlag = false;
            }

            if (UDMA_SOK == retVal)
            {
                retVal = CpswDma_freeRing(hRxFlow->cqRing,
                                          hRxFlow->rxFlowPrms.numRxPkts,
                                          hRxFlow->rxFlowPrms.ringMemFreeFxn,
                                          &hRxFlow->rxFlowPrms.hCbArg);
                Cpsw_assert(UDMA_SOK == retVal);
            }
        }

        if (UDMA_SOK == retVal)
        {
            /* Free if any DMA descriptors in the flows ready Q which application has
             * not consumed */
            if (!hRxFlow->rxFlowPrms.autoReclaimPrms.enableFlag)
            {
                CpswDma_deInitRxFreeDescQ(hRxFlow);
            }

            if (hRxFlow->rxFlowPrms.useProxy)
            {
                retVal += CpswDma_freeProxy(hRxFlow->hUdmaProxy);
                // TODO remove assert and return this error without continuing.
                Cpsw_assert(retVal == UDMA_SOK);
            }

            if (NULL != hRxFlow->hUdmaRingMon)
            {
                retVal = CpswDma_freeRingMon(hRxFlow->hUdmaRingMon);
                Cpsw_assert(UDMA_SOK == retVal);
            }

            // TODO should detach happen at beginning?
            /* Detach UDMA flow so data reception starts */
            retVal           += Udma_flowDetach(hRxFlow->hUdmaFlow);
            hRxFlow->initFlag = false;

            /* Free Tx channel driver object memory */
            CpswDma_memMgrFreeRxFlowObj(hRxFlow);
        }

        CpswOsal_restoreAllIntr(intrKey);
    }

    return retVal;
}

void CpswDma_initTxChParams(CpswDma_OpenTxChPrms *pTxChPrms)
{
    Udma_ChTxPrms txPrms;

    UdmaChTxPrms_init(&txPrms, UDMA_CH_TYPE_TX);
    pTxChPrms->udmaTxChPrms.filterPsWords = TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_PSWORDS_DISABLED;
    /* By default - don't filter extended packet info block where CPPI RX info words are mapped.
       App can enable filtering in case auto-reclaim use-case */
    pTxChPrms->udmaTxChPrms.filterEinfo   = TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_EINFO_DISABLED;
    pTxChPrms->udmaTxChPrms.addrType    = txPrms.addrType;
    pTxChPrms->udmaTxChPrms.chanType    = txPrms.chanType;
    pTxChPrms->udmaTxChPrms.busPriority = txPrms.busPriority;
    pTxChPrms->udmaTxChPrms.busQos      = txPrms.busQos;
    pTxChPrms->udmaTxChPrms.busOrderId  = txPrms.busOrderId;
    pTxChPrms->udmaTxChPrms.dmaPriority = txPrms.dmaPriority;
    pTxChPrms->udmaTxChPrms.txCredit    = txPrms.txCredit;
    pTxChPrms->udmaTxChPrms.fifoDepth   = txPrms.fifoDepth;

    pTxChPrms->chNum = UDMA_DMA_CH_INVALID;

    pTxChPrms->udmaChPrms.fqRingPrms.orderId    = UDMA_DEFAULT_RING_ORDER_ID;
    /* Use message mode for Tx FQ so at channel/flow teardown time UDMA IP can push
     * inflow descriptors back into the ring.*/
    pTxChPrms->udmaChPrms.fqRingPrms.mode       = CSL_RINGACC_RING_MODE_MESSAGE;
    pTxChPrms->udmaChPrms.fqRingPrms.useRingMon = false;

    pTxChPrms->udmaChPrms.cqRingPrms.orderId    = UDMA_DEFAULT_RING_ORDER_ID;
    pTxChPrms->udmaChPrms.cqRingPrms.mode       = CSL_RINGACC_RING_MODE_RING;
    pTxChPrms->udmaChPrms.cqRingPrms.useRingMon = false;

    /* We initialize callback function pointers to NULL so we check app pass
     * uninitialized  value */
    pTxChPrms->hUdmaDrv        = NULL;
    pTxChPrms->ringMemAllocFxn = NULL;
    pTxChPrms->ringMemFreeFxn  = NULL;
    pTxChPrms->dmaDescAllocFxn = NULL;
    pTxChPrms->dmaDescFreeFxn  = NULL;

    pTxChPrms->notifyCb = NULL;
    pTxChPrms->hCbArg   = NULL;

    pTxChPrms->useProxy = false;

    /* auto-reclaim initialization */
    pTxChPrms->autoReclaimPrms.enableFlag       = false;
    pTxChPrms->autoReclaimPrms.hDmaDescPool          = NULL;
    pTxChPrms->autoReclaimPrms.hReclaimRing = NULL;

#ifdef QNX_OS
    pTxChPrms->preferredChNum = UDMA_DMA_CH_ANY;
#endif
    return;
}

CpswDma_TxChHandle CpswDma_openTxCh(CpswDma_OpenTxChPrms *pTxChPrms)
{
    int32_t retVal         = UDMA_SOK;
    CpswDma_TxChObj *pTxCh = NULL;
    Udma_ChPrms chPrms;
    Udma_ChTxPrms txPrms;
    Udma_RingPrms ringPrms;
    CpswDma_udmaInfo udmaInfo;
    CpswDma_ringAllocInfo ringAllocInfo;
    uint32_t chType;
    bool allocChObj    = false, chOpenFlag = false, chEnFlag = false;
    bool allocFqRing = false, allocCqRing  = false;
    bool allocProxy    = false;
    CpswDma_AutoReclaimPrms *reclaimPrms = &pTxChPrms->autoReclaimPrms;
    void *pTdCqRingMem;
    uintptr_t intrKey;

    intrKey = CpswOsal_disableAllIntr();

    /* Set to NULL so if error condition we return NULL */
    pTxCh = NULL;

    /* Error check */
    retVal = CpswDma_checkTxChParams(pTxChPrms);

    if (UDMA_SOK == retVal)
    {
        /* allocate Tx channel object */
        pTxCh = CpswDma_memMgrAllocTxChObj();
        if (pTxCh == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA] Memory allocation for TX channel object failed !!\n");
            retVal = UDMA_EALLOC;
        }
        else
        {
            allocChObj = true;
            memset(pTxCh, 0U, sizeof(*pTxCh));
            /* Save channel config */
            pTxCh->txChPrms = *pTxChPrms;
            /* Initialize local variables to be used in reminder of this function  */
            chType          = UDMA_CH_TYPE_TX;
            pTxCh->hUdmaDrv = pTxChPrms->hUdmaDrv;
            pTxCh->hUdmaCh  = &pTxCh->txChMemObj.udmaChObj;
            pTxCh->fqRing   = &pTxCh->txChMemObj.fqRingObj;
            pTxCh->hUdmaEvt = &pTxCh->txChMemObj.udmaEvtObj;

            if (!reclaimPrms->enableFlag)
            {
                pTxCh->cqRing       = &pTxCh->txChMemObj.cqRingObj;
                pTxCh->hDmaDescPool = &pTxCh->txChMemObj.txFreeDmaDescQ;
            }
            else
            {
#if (1U == CPSW_CFG_DEV_ERROR)
                Cpsw_assert(NULL != reclaimPrms->hDmaDescPool);
                Cpsw_assert(NULL != reclaimPrms->hReclaimRing);
                /* Ring mode should be message as both TX channel (pushes to CQ) and RX flow
                 * (pops from FQ) uses this ring */
                Cpsw_assert(CSL_RINGACC_RING_MODE_MESSAGE == Udma_ringGetMode(reclaimPrms->hReclaimRing));
                /* Filtering of protocol specific and extended info should be enabled as this
                 * information might corrupt RX flow (as the RX flow PSI and EINFO gets mapped to
                 * Tx channel PSI and EINFO which has entirely different meaning) */
                Cpsw_assert(pTxChPrms->udmaTxChPrms.filterEinfo ==
                            TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_EINFO_DISABLED);
                Cpsw_assert(pTxChPrms->udmaTxChPrms.filterPsWords ==
                            TISCI_MSG_VALUE_RM_UDMAP_TX_CH_FILT_PSWORDS_ENABLED);
                /* Notify callback should be NULL when auto recycle is enabled as the
                 * CQ event should be disabled */
                Cpsw_assert(pTxChPrms->notifyCb == NULL);
#endif
                pTxCh->cqRing       = reclaimPrms->hReclaimRing;
                pTxCh->hDmaDescPool = reclaimPrms->hDmaDescPool;
            }
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Initialize channel params (PSI-L thread and RingAcc memories) */
        UdmaChPrms_init(&chPrms, chType);
        chPrms.peerChNum          = pTxChPrms->chNum;
    }

    if (UDMA_SOK == retVal)
    {
        pTdCqRingMem = CpswDma_memMgrAllocTdCqRingMemObj();
        if (pTdCqRingMem != NULL)
        {
            chPrms.tdCqRingPrms.ringMem = pTdCqRingMem;
            chPrms.tdCqRingPrms.elemCnt = CPSW_DMA_TDCQ_RING_ELE_CNT;
            chPrms.tdCqRingPrms.mode    = CSL_RINGACC_RING_MODE_RING;
#ifdef QNX_OS
            chPrms.chNum                = pTxChPrms->preferredChNum;
#else
            chPrms.chNum                = UDMA_DMA_CH_ANY;
#endif

            /* Open the UDMA channel */
            retVal = Udma_chOpen(pTxCh->hUdmaDrv, pTxCh->hUdmaCh, chType, &chPrms);
            CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                               "[CPSW DMA] UDMA Tx channel failed !!: 0x%x\n", retVal);
            if (UDMA_SOK == retVal)
            {
                chOpenFlag = true;
            }
        }
        else
        {
            CPSWTRACE_ERROR_IF((pTdCqRingMem == NULL),
                               "[CPSW DMA] Teardown CQ ring memory allocation for Tx channel failed\n");
            retVal = UDMA_EALLOC;
        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaRingPrms_init(&ringPrms);
        ringPrms.elemCnt = pTxChPrms->numTxPkts;
        ringPrms.orderId = pTxChPrms->udmaChPrms.fqRingPrms.orderId;
        ringPrms.mode    = pTxChPrms->udmaChPrms.fqRingPrms.mode;

        ringAllocInfo.allocRingMem    = true;
        ringAllocInfo.ringMemAllocFxn = pTxChPrms->ringMemAllocFxn;
        ringAllocInfo.ringMemFreeFxn  = pTxChPrms->ringMemFreeFxn;
        ringAllocInfo.hCbArg          = pTxChPrms->hCbArg;
        ringAllocInfo.ringNum         = Udma_chGetNum(pTxCh->hUdmaCh);

        retVal = CpswDma_allocRing(pTxCh->hUdmaDrv,
                                   pTxCh->fqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Fq ring allocation failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocFqRing = true;
        }
    }

    if ((!reclaimPrms->enableFlag) && (UDMA_SOK == retVal))
    {
        UdmaRingPrms_init(&ringPrms);
        ringPrms.elemCnt = pTxChPrms->numTxPkts;
        ringPrms.orderId = pTxChPrms->udmaChPrms.cqRingPrms.orderId;
        ringPrms.mode    = pTxChPrms->udmaChPrms.cqRingPrms.mode;

        ringAllocInfo.allocRingMem    = true;
        ringAllocInfo.ringMemAllocFxn = pTxChPrms->ringMemAllocFxn;
        ringAllocInfo.ringMemFreeFxn  = pTxChPrms->ringMemFreeFxn;
        ringAllocInfo.hCbArg          = pTxChPrms->hCbArg;
        ringAllocInfo.ringNum         = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pTxCh->hUdmaDrv,
                                   pTxCh->cqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] Cq ring allocation failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocCqRing = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Configure TX channel */
        UdmaChTxPrms_init(&txPrms, chType);

        /* TODO - Should this be enabled? */
        txPrms.pauseOnError    = TISCI_MSG_VALUE_RM_UDMAP_CH_PAUSE_ON_ERROR_DISABLED;

        /* CPPI RX (host TX) Control Data Words [0..2] are mapped to Host Packet Descriptor
         * Protocol Specific Words if (TCHAN[a]_TCFG.tx_filt_pswords = 0) and
         * (Host PD Word 1.Protocol Specific Region Location.bit[28] = 0h) and
         * (Host PD Word 1.Protocol Specific Valid Word Count.bits[22-27] =4h)
         */
        txPrms.filterPsWords = pTxChPrms->udmaTxChPrms.filterPsWords;
        txPrms.filterEinfo   = pTxChPrms->udmaTxChPrms.filterEinfo;
        txPrms.dmaPriority   = pTxChPrms->udmaTxChPrms.dmaPriority;
        retVal               = Udma_chConfigTx(pTxCh->hUdmaCh, &txPrms);

        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] UDMA Tx channel config failed !!: 0x%x\n", retVal);
        if (UDMA_SOK == retVal)
        {
            chOpenFlag = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        /* Initialize the completion queue.
         * Must be done before enabling CQ events */
        CpswUtils_initQ(&pTxCh->cqIsrQ);
    }

    if (UDMA_SOK == retVal)
    {
        if (NULL != pTxCh->fqRing)
        {
            if (pTxChPrms->useProxy)
            {
                retVal = CpswDma_allocProxy(pTxCh->hUdmaDrv, &pTxCh->txChMemObj.proxyObj, pTxCh->fqRing);
                if (UDMA_SOK == retVal)
                {
                    allocProxy        = true;
                    pTxCh->hUdmaProxy = &pTxCh->txChMemObj.proxyObj;
                }
            }
            else
            {
                pTxCh->hUdmaProxy = NULL;
            }
        }
    }

    if (UDMA_SOK == retVal)
    {
        if (NULL != pTxCh->cqRing)
        {
            if (reclaimPrms->enableFlag)
            {
                /* In auto-reclaim case we "re" configure the app passed desc Q
                 * to change return Q index (retQIdx) configured in descriptor */
                CpswDma_reconfigureDescQ(pTxCh);
            }
            else
            {
                /* Initialize the descriptor Q.*/
                CpswDma_initTxFreeDescQ(pTxCh);
            }

            /* Enable the UDMA channel */
            retVal = Udma_chEnable(pTxCh->hUdmaCh);
        }

        CPSWTRACE_VERBOSE_IF((pTxChPrms->notifyCb == NULL),
                    "[CPSW DMA] Event callback function is null !!\n");
        if ((UDMA_SOK == retVal) && (pTxChPrms->notifyCb != NULL))
        {
            chEnFlag = true;
            CpswDma_getTxChUdmaInfo(pTxCh, &udmaInfo);
            retVal = CpswDma_registerEvent(&udmaInfo,
                                           pTxCh->cqRing,
                                           &CpswDma_txCqIsr,
                                           pTxCh,
                                           UDMA_EVENT_TYPE_RING);
            CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA] TX Ch Open: event registration failed!!: 0x%x\n", retVal);
            if (CPSW_SOK == retVal)
            {
                pTxCh->evtInitFlag = true;
            }
        }
    }

    /* Get teardown cqRing handle */
    if (UDMA_SOK == retVal)
    {
        pTxCh->tdCqRing = Udma_chGetTdCqRingHandle(pTxCh->hUdmaCh);
        if(NULL == pTxCh->tdCqRing)
        {
            retVal = UDMA_EALLOC;
        }
    }

    if ((UDMA_SOK != retVal) && allocChObj)
    {
        /* Error. Free-up resource if allocated - order for freeing is important due
         * to dependencies */
        if (pTxCh->evtInitFlag)
        {
            retVal = CpswDma_unregisterEvent(pTxCh->hUdmaEvt);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocFqRing)
        {
            retVal = CpswDma_freeRing(pTxCh->fqRing,
                                      pTxChPrms->numTxPkts,
                                      pTxChPrms->ringMemFreeFxn,
                                      &pTxChPrms->hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocCqRing)
        {
            /* We free cqRing without retrieving packets here as packets are
             * not submitted by application yet so there wouldn't be any packets
             * in the cqRing */
            retVal = CpswDma_freeRing(pTxCh->cqRing,
                                      pTxChPrms->numTxPkts,
                                      pTxChPrms->ringMemFreeFxn,
                                      &pTxChPrms->hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        /* Free up ready DMA descriptors */
        if (!reclaimPrms->enableFlag)
        {
            /* Free DMA descriptors in txFreeDmaDescQ */
            CpswDma_deInitTxFreeDescQ(pTxCh);
        }

        if (allocProxy)
        {
            retVal = CpswDma_freeProxy(pTxCh->hUdmaProxy);
            Cpsw_assert(retVal == UDMA_SOK);
            pTxCh->hUdmaProxy = NULL;
        }

        if (chEnFlag)
        {
            retVal = Udma_chDisable(pTxCh->hUdmaCh, CPSW_UDMA_TEARDOWN_TIME_MS);
            Cpsw_assert(retVal == UDMA_SOK);
        }

        if (chOpenFlag)
        {
            retVal = Udma_chClose(pTxCh->hUdmaCh);
            Cpsw_assert(retVal == UDMA_SOK);
        }

        if (allocChObj)
        {
            CpswDma_memMgrFreeTxChObj(pTxCh);
        }
    }
    else
    {
        pTxCh->initFlag    = true;
    }

    CpswOsal_restoreAllIntr(intrKey);

    return pTxCh;
}

int32_t CpswDma_closeTxCh(CpswDma_TxChHandle hTxCh,
                          CpswDma_PktInfoQ *pFqPktInfoQ,
                          CpswDma_PktInfoQ *pCqPktInfoQ)
{
    int32_t retVal = CPSW_SOK;
    CSL_UdmapTdResponse tdResp;
    void *tdCqRingMemPtr = NULL;
    uintptr_t intrKey;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hTxCh) ||
        (NULL == pFqPktInfoQ) ||
        (NULL == pCqPktInfoQ))
    {
        CPSWTRACE_ERROR_IF((NULL == hTxCh), "[CPSW DMA] hTxCh is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFqPktInfoQ), "[CPSW DMA] pFqPktInfoQ is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFqPktInfoQ), "[CPSW DMA] pFqPktInfoQ is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
        intrKey = CpswOsal_disableAllIntr();

        /* Error check */
        if (hTxCh == NULL)
        {
            retVal = UDMA_EBADARGS;
        }

        if (UDMA_SOK == retVal)
        {
            if (!hTxCh->initFlag)
            {
                retVal = UDMA_EFAIL;
            }
        }

        /* TODO: Decode response information. Return an error if tear down
         * is incomplete or if tear-down was not graceful and data was
         * potentially lost */
        if (UDMA_SOK == retVal)
        {
            /* store the ring memory pointers so we can free ring memory when UDMA frees
             * Ring via Udma_ringFree internally in Udma_chClose */
            tdCqRingMemPtr = Udma_ringGetMemPtr(hTxCh->tdCqRing);

            /* Disable the UDMA channel */
            retVal += Udma_chDisable(hTxCh->hUdmaCh, CPSW_UDMA_TEARDOWN_TIME_MS);
        }

        /* Dequeue the teardown response */
        if (UDMA_SOK == retVal)
        {
            do
            {
                retVal = Udma_chDequeueTdResponse(hTxCh->hUdmaCh, &tdResp);
            }
            while (UDMA_SOK != retVal);

            if (FALSE == tdResp.tdIndicator)
            {
                retVal += UDMA_EFAIL;
            }

            /* Free tdCq ring memory */
            CpswDma_memMgrFreeTdCqRingMemObj(tdCqRingMemPtr);
        }

        if (UDMA_SOK == retVal)
        {
            if (hTxCh->evtInitFlag)
            {
                /* Disable channel Event */
                retVal += CpswDma_disableTxEvent(hTxCh);
            }

            /* Flush packets in Rx FQ - retrieve any packets that haven't been
             * processed yet */
            retVal += CpswDma_flushTxChRing(hTxCh, hTxCh->fqRing, pFqPktInfoQ);
            retVal += CpswDma_freeRing(hTxCh->fqRing,
                                       hTxCh->txChPrms.numTxPkts,
                                       hTxCh->txChPrms.ringMemFreeFxn,
                                       &hTxCh->txChPrms.hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (UDMA_SOK == retVal)
        {
            /* Retrieve packets in Rx CQ */
            retVal += CpswDma_flushTxChRing(hTxCh, hTxCh->cqRing, pCqPktInfoQ);

            if ((UDMA_SOK == retVal) && (hTxCh->evtInitFlag))
            {
                /* Return if any packets retrieved in the ISR */
                CpswUtils_appendQ(pCqPktInfoQ, &hTxCh->cqIsrQ);

                /* Unregister event - make sure you unregister the event after flushing the ring.
                 * This is because unregister event resets the ring and all the descriptors
                 * in the ring gets discarded causing resource leakage */
                retVal            += CpswDma_unregisterEvent(hTxCh->hUdmaEvt);
                hTxCh->evtInitFlag = false;
            }

            retVal += CpswDma_freeRing(hTxCh->cqRing,
                                       hTxCh->txChPrms.numTxPkts,
                                       hTxCh->txChPrms.ringMemFreeFxn,
                                       &hTxCh->txChPrms.hCbArg);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (UDMA_SOK == retVal)
        {
            /* Free up ready DMA descriptors */
            if (!hTxCh->txChPrms.autoReclaimPrms.enableFlag)
            {
                /* Free DMA descriptors in txFreeDmaDescQ */
                CpswDma_deInitTxFreeDescQ(hTxCh);
            }

            /* Free proxies if allocated */
            if (hTxCh->txChPrms.useProxy)
            {
                retVal += CpswDma_freeProxy(hTxCh->hUdmaProxy);
                // TODO remove assert and return this error without continuing.
                Cpsw_assert(retVal == UDMA_SOK);
            }

            if (UDMA_SOK == retVal)
            {
                /* Close the UDMA channel */
                retVal         += Udma_chClose(hTxCh->hUdmaCh);
                hTxCh->initFlag = false;
            }

            /* Free Tx channel driver object memory */
            CpswDma_memMgrFreeTxChObj(hTxCh);
        }

        CpswOsal_restoreAllIntr(intrKey);
    }

    return retVal;
}

int32_t CpswDma_enableRxEvent(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal;

    if ((NULL != hRxFlow) && (true == hRxFlow->evtInitFlag))
    {
        retVal = Udma_eventEnable(hRxFlow->hUdmaEvt);
    }
    else
    {
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

int32_t CpswDma_disableRxEvent(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal;

    if ((NULL != hRxFlow) && (true == hRxFlow->evtInitFlag))
    {
        retVal = Udma_eventDisable(hRxFlow->hUdmaEvt);
    }
    else
    {
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

Udma_RingHandle CpswDma_getTxChFqHandle(CpswDma_TxChHandle hTxCh)
{
    Cpsw_assert(NULL != hTxCh);
    Cpsw_assert(NULL != hTxCh->fqRing);

    return hTxCh->fqRing;
}

CpswDma_DmaDescQHandle CpswDma_getTxChDescPoolHandle(CpswDma_TxChHandle hTxCh)
{
    Cpsw_assert(NULL != hTxCh);
    return hTxCh->hDmaDescPool;
}

Udma_RingHandle CpswDma_getRxFlowFqHandle(CpswDma_RxFlowHandle hRxFlow)
{
    Cpsw_assert(NULL != hRxFlow);
    Cpsw_assert(NULL != hRxFlow->fqRing);

    return hRxFlow->fqRing;
}

CpswDma_DmaDescQHandle CpswDma_getRxFlowDescPoolHandle(CpswDma_RxFlowHandle hRxFlow)
{
    Cpsw_assert(NULL != hRxFlow);
    return hRxFlow->hDmaDescPool;
}


int32_t CpswDma_enableTxEvent(CpswDma_TxChHandle hTxCh)
{
    int32_t retVal;

    if ((NULL != hTxCh) && (true == hTxCh->evtInitFlag))
    {
        retVal = Udma_eventEnable(hTxCh->hUdmaEvt);
    }
    else
    {
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

int32_t CpswDma_disableTxEvent(CpswDma_TxChHandle hTxCh)
{
    int32_t retVal;

    if ((NULL != hTxCh) && (true == hTxCh->evtInitFlag))
    {
        retVal = Udma_eventDisable(hTxCh->hUdmaEvt);
    }
    else
    {
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

int32_t CpswDma_retrieveRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                  CpswDma_PktInfoQ *pFromFlowQueue)
{
    int32_t retVal = UDMA_SOK;
    queue_t tempQ;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hRxFlow) ||
        (NULL == pFromFlowQueue))
    {
        CPSWTRACE_ERROR_IF((NULL == hRxFlow), "[CPSW DMA] hRxFlow is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFromFlowQueue), "[CPSW DMA] pFromFlowQueue is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        uint32_t startTime, diffTime;
        uint32_t pktCnt;
        startTime = CpswOsal_timerRead();
#endif

        CpswUtils_initQ(pFromFlowQueue);
        CpswDma_drainIsrCq(pFromFlowQueue, &hRxFlow->cqIsrQ);
        CpswUtils_initQ(&tempQ);

#ifndef QNX_OS
        /* CpswDma_retrievePkts initializes the queue so cannot pass
         * pFromFlowQueue as it contains packets drained from isrq
         */
        retVal = CpswDma_retrievePkts(hRxFlow->cqRing,
                                      &tempQ,
                                      hRxFlow->hDmaDescPool,
                                      hRxFlow->rxFlowPrms.disableCacheOpsFlag,
                                      CPSW_DMA_DIR_RX);

        if (CPSW_SOK == retVal)
        {
            CpswUtils_appendQ(pFromFlowQueue, &tempQ);
        }
#endif

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        pktCnt = CpswUtils_getQCount(pFromFlowQueue);
        CpswDmaStats_addCnt(&hRxFlow->stats.rxRetrievePktDq, pktCnt);
        diffTime = CpswOsal_timerGetDiff(startTime);
        CpswDmaStats_updateNotifyStats(&hRxFlow->stats.retrievePktStats, pktCnt, diffTime);
#endif
    }

    return retVal;
}

int32_t CpswDma_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                CpswDma_PktInfoQ *pToFlowQueue,
                                CpswDma_PktInfoQ *pUnUsedPktQueue)
{
    int32_t retVal = UDMA_SOK;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hRxFlow) ||
        (NULL == pToFlowQueue) ||
        (NULL == pUnUsedPktQueue))
    {
        CPSWTRACE_ERROR_IF((NULL == hRxFlow), "[CPSW DMA] hRxFlow is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pToFlowQueue), "[CPSW DMA] pToFlowQueue is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pUnUsedPktQueue), "[CPSW DMA] pUnUsedPktQueue is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        uint32_t startTime, diffTime;
        uint32_t pktCnt, notifyCount;
        startTime = CpswOsal_timerRead();
        pktCnt    = CpswUtils_getQCount(pToFlowQueue);
#endif

        /* Enqueue descs to fqRing regardless of caller's queue state */
        if (CpswUtils_getQCount(pToFlowQueue) > 0U)
        {
            retVal = CpswDma_submitPkts(hRxFlow->fqRing,
                                        pToFlowQueue,
                                        hRxFlow->hDmaDescPool,
                                        hRxFlow->rxFlowPrms.disableCacheOpsFlag,
                                        CPSW_DMA_DIR_RX,
                                        hRxFlow->hUdmaProxy);
        }

        if (0U != CpswUtils_getQCount(pToFlowQueue))
        {
            CpswUtils_initQ(pUnUsedPktQueue);
            CpswUtils_copyQ(pUnUsedPktQueue, pToFlowQueue);
            CpswUtils_initQ(pToFlowQueue);
        }

        /* If fqRing ran out of space is not an error, packets will be re-submitted from application */
        if (UDMA_ETIMEOUT == retVal)
        {
            CPSWTRACE_INFO("RX FLOW FQ underflow had occurred \n");
            retVal = UDMA_SOK;
        }

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        CpswDmaStats_addCnt(&hRxFlow->stats.rxSubmitPktUnderFlowCnt,
                            CpswUtils_getQCount(pUnUsedPktQueue));
        pktCnt -= CpswUtils_getQCount(pUnUsedPktQueue);
        CpswDmaStats_addCnt(&hRxFlow->stats.rxSubmitPktEnq, pktCnt);
        diffTime                                                    = CpswOsal_timerGetDiff(startTime);
        notifyCount                                                 = hRxFlow->stats.submitPktStats.dataNotifyCnt & (CPSW_DMA_STATS_HISTORY_CNT - 1U);
        hRxFlow->stats.submitPktStats.readyDmaDescQCnt[notifyCount] = CpswDma_dmaDescQCount(hRxFlow->hDmaDescPool);

        CpswDmaStats_updateNotifyStats(&hRxFlow->stats.submitPktStats, pktCnt, diffTime);
#endif
    }

    return retVal;
}

int32_t CpswDma_retrieveTxDonePackets(CpswDma_TxChHandle hTxCh,
                                      CpswDma_PktInfoQ *pFromChannelQueue)
{
    int32_t retVal = UDMA_SOK;
    queue_t tempQ;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hTxCh) ||
        (NULL == pFromChannelQueue))
    {
        CPSWTRACE_ERROR_IF((NULL == hTxCh), "[CPSW DMA] hTxCh is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pFromChannelQueue), "[CPSW DMA] pFromChannelQueue is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        uint32_t startTime, diffTime;
        uint32_t pktCnt;
        startTime = CpswOsal_timerRead();
#endif

        CpswUtils_initQ(pFromChannelQueue);
        CpswDma_drainIsrCq(pFromChannelQueue, &hTxCh->cqIsrQ);
        CpswUtils_initQ(&tempQ);

        /* CpswDma_retrievePkts initializes the queue so cannot pass
         * pFromChannelQueue as it contains packets drained from isrq
         */
        retVal = CpswDma_retrievePkts(hTxCh->cqRing,
                                      &tempQ,
                                      hTxCh->hDmaDescPool,
                                      hTxCh->txChPrms.disableCacheOpsFlag,
                                      CPSW_DMA_DIR_TX);

        if (CPSW_SOK == retVal)
        {
            CpswUtils_appendQ(pFromChannelQueue, &tempQ);
        }

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        pktCnt = CpswUtils_getQCount(pFromChannelQueue);
        CpswDmaStats_addCnt(&hTxCh->stats.txRetrievePktDq, pktCnt);
        diffTime = CpswOsal_timerGetDiff(startTime);
        CpswDmaStats_updateNotifyStats(&hTxCh->stats.retrievePktStats, pktCnt, diffTime);
#endif
    }

    return retVal;
}

int32_t CpswDma_submitTxReadyPackets(CpswDma_TxChHandle hTxCh,
                                     CpswDma_PktInfoQ *pToChannelQueue,
                                     CpswDma_PktInfoQ *pUnUsedPktQueue)
{
    int32_t retVal = UDMA_SOK;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hTxCh) ||
        (NULL == pToChannelQueue) ||
        (NULL == pUnUsedPktQueue))
    {
        CPSWTRACE_ERROR_IF((NULL == hTxCh), "[CPSW DMA] hTxCh is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pToChannelQueue), "[CPSW DMA] pToChannelQueue is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == pUnUsedPktQueue), "[CPSW DMA] pUnUsedPktQueue is NULL!!\n");
        Cpsw_assert(FALSE);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        uint32_t startTime, diffTime;
        uint32_t pktCnt, notifyCount;
        startTime = CpswOsal_timerRead();
        pktCnt    = CpswUtils_getQCount(pToChannelQueue);
#endif

        /* Enqueue descs to fqRing regardless of caller's queue state */
        if (CpswUtils_getQCount(pToChannelQueue) > 0U)
        {
            retVal = CpswDma_submitPkts(hTxCh->fqRing,
                                        pToChannelQueue,
                                        hTxCh->hDmaDescPool,
                                        hTxCh->txChPrms.disableCacheOpsFlag,
                                        CPSW_DMA_DIR_TX,
                                        hTxCh->hUdmaProxy);
        }

        if (0U != CpswUtils_getQCount(pToChannelQueue))
        {
            CpswUtils_initQ(pUnUsedPktQueue);
            CpswUtils_copyQ(pUnUsedPktQueue, pToChannelQueue);
            /* Init the app queue as we have copied queue elements to unused queue */
            CpswUtils_initQ(pToChannelQueue);
        }

        /* If fqRing ran out of space it is not an error, packets will be re-submitted by application*/
        if (UDMA_ETIMEOUT == retVal)
        {
            CPSWTRACE_INFO("TX Channel FQ underflow had occurred \n");
            retVal = UDMA_SOK;
        }

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
        CpswDmaStats_addCnt(&hTxCh->stats.txSubmitPktOverFlowCnt, CpswUtils_getQCount(pUnUsedPktQueue));
        pktCnt -= CpswUtils_getQCount(pUnUsedPktQueue);
        CpswDmaStats_addCnt(&hTxCh->stats.txSubmitPktEnq, pktCnt);
        diffTime                                                  = CpswOsal_timerGetDiff(startTime);
        notifyCount                                               = hTxCh->stats.submitPktStats.dataNotifyCnt & (CPSW_DMA_STATS_HISTORY_CNT - 1U);
        hTxCh->stats.submitPktStats.readyDmaDescQCnt[notifyCount] = hTxCh->hDmaDescPool->count;

        CpswDmaStats_updateNotifyStats(&hTxCh->stats.submitPktStats, pktCnt, diffTime);
#endif
    }

    return retVal;
}

void CpswDma_pktInfoInit(CpswDma_PktInfo *pktInfo)
{
    memset(&pktInfo->node, 0U, sizeof(pktInfo->node));
    pktInfo->bufPtr                = NULL;
#ifdef QNX_OS
    pktInfo->bufPtrPhys            = NULL;
#endif
    pktInfo->orgBufLen             = 0U;
    pktInfo->userBufLen            = 0U;
    pktInfo->appPriv               = NULL;
    pktInfo->tsInfo.enableHostTxTs = false;
    pktInfo->txPortNum             = CPSW_MAC_PORT_INV;
    CPSW_UTILS_SET_PKT_DRIVER_STATE(&pktInfo->pktState,
                                    (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW);
}

uint32_t CpswDma_getRxChFlowStartIdx(CpswDma_Handle hCpswDma)
{
    CpswDma_RxChObj *pRxCh = &hCpswDma->rxChObj;
    uint32_t flowStartIdx;

    flowStartIdx = Udma_flowGetNum(&pRxCh->flowUdmaObj);

    return flowStartIdx;
}

uint32_t CpswDma_getRxFlowCnt(CpswDma_Handle hCpswDma)
{
    CpswDma_RxChObj *pRxCh = &hCpswDma->rxChObj;
    uint32_t flowCnt;

    flowCnt = Udma_flowGetCount(&pRxCh->flowUdmaObj);

    return flowCnt;
}

int32_t CpswDma_checkRxFlowSanity(CpswDma_RxFlowHandle hRxFlow,
                                  uint32_t margin)
{
    int32_t retVal       = CPSW_SOK;
    uint32_t allocPkts   = hRxFlow->rxFlowPrms.numRxPkts;
    uint32_t fqRingHwOcc = CSL_ringaccGetRingHwOcc(&hRxFlow->fqRing->drvHandle->raRegs,
                                                   hRxFlow->fqRing->ringNum);
    uint32_t cqRingHwOcc = CSL_ringaccGetRingHwOcc(&hRxFlow->cqRing->drvHandle->raRegs,
                                                   hRxFlow->cqRing->ringNum);
    uint32_t withHwPkts = fqRingHwOcc + cqRingHwOcc;

    /* App allocated should match packets with DMA module and HW.
     * Also we consider margin packets in fly while reading the HW status */
    if ((allocPkts - margin) < (CpswDma_dmaDescQCount(hRxFlow->hDmaDescPool) + withHwPkts))
    {
        retVal = CPSW_SOK;
    }
    else
    {
        CPSWTRACE_WARN("hRxFlow->rxFlowPrms.numTxPkts   = %d \n", hRxFlow->rxFlowPrms.numRxPkts);
        CPSWTRACE_WARN("CpswDma_dmaDescQCount(hRxFlow->hDmaDescPool)  = %d \n", CpswDma_dmaDescQCount(hRxFlow->hDmaDescPool));
        CPSWTRACE_WARN("hRxFlow->fqRing->pRtRegs->HWOCC = %d \n", fqRingHwOcc);
        CPSWTRACE_WARN("hRxFlow->cqRing->pRtRegs->HWOCC = %d \n", cqRingHwOcc);

        retVal = CPSW_EFAIL;
    }

    return retVal;
}

int32_t CpswDma_checkTxChSanity(CpswDma_TxChHandle hTxCh,
                                uint32_t margin)
{
    int32_t retVal       = CPSW_SOK;
    uint32_t allocPkts   = hTxCh->txChPrms.numTxPkts;
    uint32_t fqRingHwOcc = CSL_ringaccGetRingHwOcc(&hTxCh->fqRing->drvHandle->raRegs,
                                                   hTxCh->fqRing->ringNum);
    uint32_t cqRingHwOcc = CSL_ringaccGetRingHwOcc(&hTxCh->cqRing->drvHandle->raRegs,
                                                   hTxCh->cqRing->ringNum);
    uint32_t withHwPkts = fqRingHwOcc + cqRingHwOcc;

    /* App allocated should match packets with DMA module and HW.
     * Also we consider margin packets in fly while reading the HW status */
    if ((allocPkts - margin) < (hTxCh->hDmaDescPool->count + withHwPkts))
    {
        retVal = CPSW_SOK;
    }
    else
    {
        CPSWTRACE_WARN("hTxCh->txChPrms.numTxPkts     = %d \n", hTxCh->txChPrms.numTxPkts);
        CPSWTRACE_WARN("hTxCh->hDmaDescPool->count   = %d \n", hTxCh->hDmaDescPool->count);
        CPSWTRACE_WARN("hTxCh->fqRing->pRtRegs->HWOCC = %d \n", fqRingHwOcc);
        CPSWTRACE_WARN("hTxCh->cqRing->pRtRegs->HWOCC = %d \n", cqRingHwOcc);
        retVal = CPSW_EFAIL;
    }

    return retVal;
}

CpswDma_DrvObj *CpswDma_getDmaObj(Cpsw_Type cpswType)
{
    return &gCpswDmaObj[cpswType];
}

int32_t CpswDma_getRxFlowStats(CpswDma_RxFlowHandle hRxFlow,
                               CpswDma_RxFlowStats *rxFlowStats)
{
    int32_t retVal = CPSW_ENOTSUPPORTED;

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    *rxFlowStats = hRxFlow->stats;
#endif

    return retVal;
}

int32_t CpswDma_getTxChStats(CpswDma_TxChHandle hTxCh,
                             CpswDma_TxChStats *txChStats)
{
    int32_t retVal = CPSW_ENOTSUPPORTED;

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    *txChStats = hTxCh->stats;
#endif

    return retVal;
}

int32_t CpswDma_resetRxFlowStats(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal = CPSW_ENOTSUPPORTED;

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    memset(&hRxFlow->stats, 0, sizeof(hRxFlow->stats));
#endif

    return retVal;
}

int32_t CpswDma_resetTxChStats(CpswDma_TxChHandle hTxCh)
{
    int32_t retVal = CPSW_ENOTSUPPORTED;

#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    memset(&hTxCh->stats, 0, sizeof(hTxCh->stats));
#endif

    return retVal;
}

static int32_t CpswDma_checkRxFlowParams(CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    int32_t retVal = UDMA_SOK;

    if (NULL == pRxFlowPrms)
    {
        CPSWTRACE_ERROR("[CPSW DMA] Flow params is NULL!!\n");
        retVal = UDMA_EBADARGS;
    }
    else
    {
        if (NULL == pRxFlowPrms->hUdmaDrv)
        {
            CPSWTRACE_ERROR("[CPSW DMA] UDMA not opened !!\n");
            retVal = UDMA_EBADARGS;
        }
    }

    if (UDMA_SOK == retVal)
    {
        if ((NULL == pRxFlowPrms->ringMemAllocFxn) ||
            (NULL == pRxFlowPrms->ringMemFreeFxn) ||
            (NULL == pRxFlowPrms->dmaDescAllocFxn) ||
            (NULL == pRxFlowPrms->dmaDescFreeFxn))
        {
            CPSWTRACE_ERROR_IF((NULL == pRxFlowPrms->ringMemAllocFxn), "[CPSW DMA] ringMemAllocFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pRxFlowPrms->ringMemFreeFxn), "[CPSW DMA] ringMemFreeFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pRxFlowPrms->dmaDescAllocFxn), "[CPSW DMA] dmaDescAllocFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pRxFlowPrms->dmaDescFreeFxn), "[CPSW DMA] dmaDescFreeFxn is NULL!!\n");
            retVal = UDMA_EBADARGS;
        }

        if (CPSW_RM_RXFLOWIDX_INVALID == pRxFlowPrms->flowIdx)
        {
            CPSWTRACE_ERROR("[CPSW DMA] Invalid flow id for CPSW RX flow open !!\n", pRxFlowPrms->flowIdx);
            retVal = UDMA_EBADARGS;
        }
    }

    if ((UDMA_SOK == retVal) && (pRxFlowPrms->udmaChPrms.cqRingPrms.useRingMon))
    {
        CPSWTRACE_ERROR("[CPSW DMA] Driver supports Ring monitor only for RX FQ !!\n");
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

static int32_t CpswDma_checkTxChParams(CpswDma_OpenTxChPrms *pTxChPrms)
{
    int32_t retVal = UDMA_SOK;

    if (NULL == pTxChPrms)
    {
        CPSWTRACE_ERROR("[CPSW DMA] TX channel params can't be NULL !!\n");
        retVal = UDMA_EBADARGS;
    }
    else
    {
        if (NULL == pTxChPrms->hUdmaDrv)
        {
            /* DMA should be opened before opening Ch/Flow */
            CPSWTRACE_ERROR("[CPSW DMA] UDMA not opened !!\n");
            retVal = UDMA_EBADARGS;
        }
    }

    if (UDMA_SOK == retVal)
    {
        if ((NULL == pTxChPrms->ringMemAllocFxn) ||
            (NULL == pTxChPrms->ringMemFreeFxn) ||
            (NULL == pTxChPrms->dmaDescAllocFxn) ||
            (NULL == pTxChPrms->dmaDescFreeFxn))
        {
            CPSWTRACE_ERROR_IF((NULL == pTxChPrms->ringMemAllocFxn), "[CPSW DMA] ringMemAllocFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pTxChPrms->ringMemFreeFxn), "[CPSW DMA] ringMemFreeFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pTxChPrms->dmaDescAllocFxn), "[CPSW DMA] dmaDescAllocFxn is NULL!!\n");
            CPSWTRACE_ERROR_IF((NULL == pTxChPrms->dmaDescFreeFxn), "[CPSW DMA] dmaDescFreeFxn is NULL!!\n");
            retVal = UDMA_EBADARGS;
        }

        if (CPSW_RM_TXCHNUM_INVALID == pTxChPrms->chNum)
        {
            CPSWTRACE_ERROR("[CPSW DMA] Invalid channel number for CPSW TX channel open !!: %d\n", pTxChPrms->chNum);
            retVal = UDMA_EBADARGS;
        }
    }

    if ((UDMA_SOK == retVal) &&
        ((pTxChPrms->udmaChPrms.fqRingPrms.useRingMon) || (pTxChPrms->udmaChPrms.cqRingPrms.useRingMon)))
    {
        CPSWTRACE_ERROR("[CPSW DMA] Driver supports Ring monitor only for RX FQ !!\n");
        retVal = UDMA_EBADARGS;
    }

    return retVal;
}

/*
 * Drain packets in ISR due to ring level fix
 * Udma ring completion (CQ) interrupt is a self clearing level interrupt.
 * The CQ ISR must drain the ring in Udma callback. This requires a fix in Udma driver
 * to not clear ring interrupt explicitly and change in CpswDma to drain buffers in ISR callback
 * Without this change, Ring interrupt will randomly stop occurring
 */
static void CpswDma_drainIsrCq(queue_t *dstQ,
                               queue_t *isrCq)
{
    uintptr_t key = CpswOsal_disableAllIntr();

    CpswUtils_appendQ(dstQ, isrCq);
    CpswUtils_initQ(isrCq);

    CpswOsal_restoreAllIntr(key);
}

void CpswDma_initDataPathParams(CpswDma_DataPathConfig *pDmaConfig)
{
    pDmaConfig->hUdmaDrv = NULL;
}

CpswDma_Handle CpswDma_initDataPath(Cpsw_Type cpswType,
                                    const CpswDma_DataPathConfig *pCpswDmaDataPathConfig)
{
    CpswDma_Config cfg;
    CpswDma_Handle hDmaHandle;

    CpswDma_initParams(&cfg);
    cfg.hUdmaDrv = pCpswDmaDataPathConfig->hUdmaDrv;
    hDmaHandle   = CpswDma_open(cpswType, &cfg);
    return hDmaHandle;
}

int32_t CpswDma_deInitDataPath(CpswDma_Handle hCpswDma)
{
    int32_t status;

    status = CpswDma_close(hCpswDma);
    return status;
}
