/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_mdio.c
 *
 * \brief This file contains MDIO related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_mdio.h>
#include "priv/cpsw_mdio_priv.h"
#include "priv/cpsw_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */


/*! \brief Max number of PHYs that can be monitored by MDIO in normal mode */
#define CPSW_MDIO_PHY_MONITOR_MAX             (CPSW_MDIO_USER_CHANNEL_LAST + 1U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static CpswMdio_Handle CpswMdio_getObject(Cpsw_Type cpswType);

static inline CSL_mdioHandle CpswMdio_getCslRegs(Cpsw_Type cpswType);

static void CpswMdio_validateIPVersion(CSL_mdioHandle handle,
                                       Cpsw_Type cpswType);

static int32_t CpswMdio_setupNormalMode(CpswMdio_Handle hMdio,
                                        const CpswMdio_Config *cfg);

static void CpswMdio_setupStatusChangeMode(CpswMdio_Handle hMdio,
                                           const CpswMdio_Config *cfg);

static void CpswMdio_handleIntr(CpswMdio_Handle hMdio,
                                CpswMdio_Callbacks *callbacks);

static int32_t CpswMdio_readRegC45(CpswMdio_Handle hMdio,
                                   uint8_t phyAddr,
                                   CpswMdio_Clause45Mmd mmd,
                                   uint16_t reg,
                                   uint16_t *val);

static int32_t CpswMdio_writeRegC45(CpswMdio_Handle hMdio,
                                    uint8_t phyAddr,
                                    CpswMdio_Clause45Mmd mmd,
                                    uint16_t reg,
                                    uint16_t val);

static void CpswMdio_printRegs(CSL_mdioHandle mdioRegs);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW MDIO objects */
CpswMdio_Obj mdioObjs[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswMdio_initParams(CpswMdio_Config *cfg)
{
    cfg->mode               = CPSW_MDIO_MODE_STATE_CHANGE;
    cfg->mdioBusFreqHz      = 2200000U;
    cfg->phyStatePollFreqHz = cfg->mdioBusFreqHz;
    cfg->pollEnableMask     = CPSW_MDIO_PHY_ADDR_MASK_ALL;
    cfg->clause45EnableMask = CPSW_MDIO_PHY_ADDR_MASK_NONE;
    cfg->userCh             = CPSW_MDIO_USER_CHANNEL_0;
}

CpswMdio_Handle CpswMdio_open(Cpsw_Type cpswType,
                              const CpswMdio_Config *cfg)
{
    CpswMdio_Handle hMdio = CpswMdio_getObject(cpswType);
    const CpswSoc_Config *pSocCfg;
    uint32_t clkdiv;
    uint32_t ipgRatio;
    int32_t status = CPSW_SOK;

    memset(hMdio, 0U, sizeof(*hMdio));

    hMdio->mdioRegs = CpswMdio_getCslRegs(cpswType);
    Cpsw_assert(hMdio->mdioRegs != NULL);

    pSocCfg = CpswSoc_getConfig(cpswType);
    CpswMdio_validateIPVersion(hMdio->mdioRegs, cpswType);

    if (cfg->mdioBusFreqHz != 0U)
    {
        clkdiv = (pSocCfg->cppiClkFreqHz / cfg->mdioBusFreqHz) - 1U;
        if (clkdiv > 0xFFFFU)
        {
            CPSWTRACE_ERROR("unsupported clk div %u (CPPI %u Hz, MDIO bus %u Hz)\n",
                            clkdiv, pSocCfg->cppiClkFreqHz, cfg->mdioBusFreqHz);
            status = CPSW_EFAIL;
        }

        ipgRatio = cfg->mdioBusFreqHz / cfg->phyStatePollFreqHz;
        if (ipgRatio > 0xFFU)
        {
            CPSWTRACE_ERROR("unsupported IPG ratio %u (MDIO bus %u Hz, PHY poll %u Hz)\n",
                            ipgRatio, cfg->mdioBusFreqHz, cfg->phyStatePollFreqHz);
            status = CPSW_EFAIL;
        }
    }
    else
    {
        CPSWTRACE_ERROR("invalid MDIO bus freq %u Hz\n", cfg->mdioBusFreqHz);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 0U);
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 1U);

        switch (cfg->mode)
        {
            case CPSW_MDIO_MODE_NORMAL:
                status = CpswMdio_setupNormalMode(hMdio, cfg);
                break;

            case CPSW_MDIO_MODE_STATE_CHANGE:
                CpswMdio_setupStatusChangeMode(hMdio, cfg);
                break;

            case CPSW_MDIO_MODE_MANUAL:
                CPSWTRACE_ERROR("manual mode is not supported\n");
                status = CPSW_ENOTSUPPORTED;
                break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_MDIO_setPollIPG(hMdio->mdioRegs, (uint8_t)ipgRatio);
        CSL_MDIO_setClkDivVal(hMdio->mdioRegs, (uint16_t)clkdiv);
        CSL_MDIO_enableStateMachine(hMdio->mdioRegs);
        CSL_MDIO_setClause45EnableMask(hMdio->mdioRegs, cfg->clause45EnableMask);
        hMdio->magic = CPSW_MDIO_MAGIC;
    }
    else
    {
        hMdio = NULL;
    }

    return hMdio;
}

CpswMdio_Handle CpswMdio_reopen(Cpsw_Type cpswType)
{
    CpswMdio_Handle hMdio = CpswMdio_getObject(cpswType);

    memset(hMdio, 0U, sizeof(*hMdio));
    hMdio->mdioRegs = CpswMdio_getCslRegs(cpswType);
    Cpsw_assert(hMdio->mdioRegs != NULL);

    hMdio->magic = CPSW_MDIO_MAGIC;

    return hMdio;
}

void CpswMdio_close(CpswMdio_Handle hMdio)
{
    uint32_t i;

    if (hMdio->magic == CPSW_MDIO_MAGIC)
    {
        for (i = 0U; i < CPSW_MDIO_PHY_MONITOR_MAX; i++)
        {
            CSL_MDIO_disableLinkStatusChangeInterrupt(hMdio->mdioRegs, i, 0U);
        }

        CSL_MDIO_disableStatusChangeModeInterrupt(hMdio->mdioRegs);
        CSL_MDIO_disableStateMachine(hMdio->mdioRegs);
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 0U);
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 1U);
        hMdio->magic = CPSW_NO_MAGIC;
    }
    else
    {
        CPSWTRACE_ERROR("MDIO is not open (handle=%p)\n", hMdio);
    }
}

int32_t CpswMdio_ioctl(CpswMdio_Handle hMdio,
                       uint32_t cmd,
                       Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    if (hMdio->magic != CPSW_MDIO_MAGIC)
    {
        CPSWTRACE_ERROR("MDIO is not open (handle=%p)\n", hMdio);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        switch (cmd)
        {
            case CPSW_MDIO_IOCTL_GET_VERSION:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswMdio_Version));
                if (status == CPSW_SOK)
                {
                    CSL_MDIO_VERSION ver;
                    CpswMdio_Version *version;

                    CSL_MDIO_getVersionInfo(hMdio->mdioRegs, &ver);
                    version         = (CpswMdio_Version *)prms->outArgs;
                    version->major  = ver.revMaj;
                    version->minor  = ver.revMin;
                    version->rtl    = 0U;
                    version->id     = ver.modId;
                    version->scheme = ver.scheme;
                    version->bu     = ver.bu;
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_VERSION params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_GET_ALIVE_STATUS:
            {
                status = Cpsw_checkInOutArgs(prms, sizeof(uint8_t), sizeof(bool));
                if (status == CPSW_SOK)
                {
                    uint8_t *phyAddr = (uint8_t *)prms->inArgs;
                    bool *alive      = (bool *)prms->outArgs;

                    *alive = (CSL_MDIO_isPhyAlive(hMdio->mdioRegs,
                                                  *phyAddr) == 0) ? false : true;
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_ALIVE_STATUS params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_GET_LINK_STATUS:
            {
                status = Cpsw_checkInOutArgs(prms, sizeof(uint8_t), sizeof(bool));
                if (status == CPSW_SOK)
                {
                    uint8_t *phyAddr = (uint8_t *)prms->inArgs;
                    bool *linked     = (bool *)prms->outArgs;

                    *linked = (CSL_MDIO_isPhyLinked(hMdio->mdioRegs,
                                                    *phyAddr) == 0) ? false : true;
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_LINK_STATUS params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_CLAUSE22_READ_REG:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMdio_Clause22ReadRegInArgs),
                                             sizeof(uint16_t));
                if (status == CPSW_SOK)
                {
                    CpswMdio_Clause22ReadRegInArgs *inArgs =
                        (CpswMdio_Clause22ReadRegInArgs *)prms->inArgs;
                    uint16_t *val = (uint16_t *)prms->outArgs;
                    uint32_t ack;
                    uint32_t c45EnableMask;

                    c45EnableMask = CSL_MDIO_getClause45EnableMask(hMdio->mdioRegs);
                    if (!CPSW_IS_BIT_SET(c45EnableMask, inArgs->phyAddr))
                    {
                        ack = CSL_MDIO_phyRegRead2(hMdio->mdioRegs,
                                                   hMdio->userCh,
                                                   inArgs->phyAddr,
                                                   inArgs->reg,
                                                   val);
                        status = (ack == TRUE) ? CPSW_SOK : CPSW_EFAIL;
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "failed to read C22 reg (PHY=%u reg=%u): %d\n",
                                           inArgs->phyAddr, inArgs->reg, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("PHY %u is not configured for Clause-22 access\n",
                                        inArgs->phyAddr);
                        status = CPSW_EPERM;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid CLAUSE22_READ_REG params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswMdio_Clause22WriteRegInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMdio_Clause22WriteRegInArgs *inArgs =
                        (CpswMdio_Clause22WriteRegInArgs *)prms->inArgs;
                    uint32_t c45EnableMask;

                    c45EnableMask = CSL_MDIO_getClause45EnableMask(hMdio->mdioRegs);
                    if (!CPSW_IS_BIT_SET(c45EnableMask, inArgs->phyAddr))
                    {
                        CSL_MDIO_phyRegWrite2(hMdio->mdioRegs,
                                              hMdio->userCh,
                                              inArgs->phyAddr,
                                              inArgs->reg,
                                              inArgs->val);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("PHY %u is not configured for Clause-22 access\n",
                                        inArgs->phyAddr);
                        status = CPSW_EPERM;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid CLAUSE22_WRITE_REG params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_CLAUSE45_READ_REG:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMdio_Clause45ReadRegInArgs),
                                             sizeof(uint16_t));
                if (status == CPSW_SOK)
                {
                    CpswMdio_Clause45ReadRegInArgs *inArgs =
                        (CpswMdio_Clause45ReadRegInArgs *)prms->inArgs;
                    uint16_t *val = (uint16_t *)prms->outArgs;
                    uint32_t c45EnableMask;

                    c45EnableMask = CSL_MDIO_getClause45EnableMask(hMdio->mdioRegs);
                    if (CPSW_IS_BIT_SET(c45EnableMask, inArgs->phyAddr))
                    {
                        status = CpswMdio_readRegC45(hMdio,
                                                     inArgs->phyAddr,
                                                     inArgs->mmd,
                                                     inArgs->reg,
                                                     val);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "failed to read C45 reg (PHY=%u MMD=%d reg=%u): %d\n",
                                           inArgs->phyAddr, inArgs->mmd, inArgs->reg, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("PHY %u is not configured for Clause-45 access\n",
                                        inArgs->phyAddr);
                        status = CPSW_EPERM;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid CLAUSE45_READ_REG params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswMdio_Clause45WriteRegInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMdio_Clause45WriteRegInArgs *inArgs =
                        (CpswMdio_Clause45WriteRegInArgs *)prms->inArgs;
                    uint32_t c45EnableMask;

                    c45EnableMask = CSL_MDIO_getClause45EnableMask(hMdio->mdioRegs);
                    if (CPSW_IS_BIT_SET(c45EnableMask, inArgs->phyAddr))
                    {
                        status = CpswMdio_writeRegC45(hMdio,
                                                      inArgs->phyAddr,
                                                      inArgs->mmd,
                                                      inArgs->reg,
                                                      inArgs->val);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "failed to write C45 reg (PHY=%u MMD=%d reg=%u): %d\n",
                                           inArgs->phyAddr, inArgs->mmd, inArgs->reg, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("PHY %u is not configured for Clause-45 access\n",
                                        inArgs->phyAddr);
                        status = CPSW_EPERM;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid CLAUSE45_WRITE_REG params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_GET_POLL_ENABLE_MASK:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    uint32_t *pollMask = (uint32_t *)prms->outArgs;
                    uint32_t isStatusChangeMode;
                    uint32_t phyAddr;
                    uint32_t i;

                    isStatusChangeMode = CSL_MDIO_isStateChangeModeEnabled(hMdio->mdioRegs);
                    if (isStatusChangeMode == 1U)
                    {
                        *pollMask = CSL_MDIO_getPollEnableMask(hMdio->mdioRegs);
                    }
                    else
                    {
                        *pollMask = CPSW_MDIO_PHY_ADDR_MASK_NONE;
                        for (i = 0U; i < CPSW_MDIO_PHY_MONITOR_MAX; i++)
                        {
                            phyAddr    = CSL_MDIO_getLinkStatusChangePhyAddr(hMdio->mdioRegs, i);
                            *pollMask |= CPSW_MDIO_PHY_ADDR_MASK(phyAddr);
                        }
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_POLL_ENABLE_MASK params\n");
                }
            }
            break;

            case CPSW_MDIO_IOCTL_HANDLE_INTR:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswMdio_Callbacks));
                if (status == CPSW_SOK)
                {
                    CpswMdio_Callbacks *callbacks = (CpswMdio_Callbacks *)prms->inArgs;

                    CpswMdio_handleIntr(hMdio, callbacks);
                }
            }
            break;

            case CPSW_MDIO_IOCTL_PRINT_REGS:
            {
                CpswMdio_printRegs(hMdio->mdioRegs);
            }
            break;

            default:
            {
                CPSWTRACE_ERROR("invalid IOCTL cmd: %u\n", cmd);
                status = CPSW_EINVALID_PARAMS;
            }
            break;
        }
    }

    return status;
}

static CpswMdio_Handle CpswMdio_getObject(Cpsw_Type cpswType)
{
    CpswMdio_Handle hMdio = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hMdio = &mdioObjs[cpswType];
            break;

        default:
            Cpsw_assert(false);
    }

    return hMdio;
}

static inline CSL_mdioHandle CpswMdio_getCslRegs(Cpsw_Type cpswType)
{
    CSL_mdioHandle regs = NULL;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
#ifdef QNX_OS
    uint32_t length = sizeof(CSL_MdioRegs);
    regs = (CSL_mdioHandle) mmap_device_memory(0,
        length, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)ipInfo->mdioInfo.regBaseAddr);
#else
    regs = (CSL_mdioHandle)ipInfo->mdioInfo.regBaseAddr;
#endif

    return regs;
}

static void CpswMdio_validateIPVersion(CSL_mdioHandle handle,
                                       Cpsw_Type cpswType)
{
    CSL_MDIO_VERSION version;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    CSL_MDIO_getVersionInfo(handle, &version);
    Cpsw_assert(version.modId ==  ipInfo->mdioInfo.versionInfo.moduleId);
    Cpsw_assert(version.revMaj == ipInfo->mdioInfo.versionInfo.majorVer);
    Cpsw_assert(version.revMin == ipInfo->mdioInfo.versionInfo.minorVer);
    Cpsw_assert(version.revRtl == ipInfo->mdioInfo.versionInfo.rtlVer);
}

static int32_t CpswMdio_setupNormalMode(CpswMdio_Handle hMdio,
                                        const CpswMdio_Config *cfg)
{
    uint32_t phyMonIndex = 0U;
    uint32_t i;
    int32_t status = CPSW_SOK;

    for (i = 0U; i <= CPSW_MDIO_MAX_PHY_CNT; i++)
    {
        if (CPSW_IS_BIT_SET(cfg->pollEnableMask, i))
        {
            if (phyMonIndex < CPSW_MDIO_PHY_MONITOR_MAX)
            {
                CSL_MDIO_enableLinkStatusChangeInterrupt(hMdio->mdioRegs,
                                                         phyMonIndex,
                                                         i);
            }

            phyMonIndex++;
        }
    }

    if (phyMonIndex <= CPSW_MDIO_PHY_MONITOR_MAX)
    {
        /* Normal Mode implies State Change Mode is disabled */
        CSL_MDIO_disableStateChangeMode(hMdio->mdioRegs);
    }
    else
    {
        CPSWTRACE_ERROR("invalid PHY monitor count %d\n", phyMonIndex);
        for (i = 0U; i < CPSW_MDIO_PHY_MONITOR_MAX; i++)
        {
            CSL_MDIO_disableLinkStatusChangeInterrupt(hMdio->mdioRegs, i, 0U);
        }

        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 0U);
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 1U);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static void CpswMdio_setupStatusChangeMode(CpswMdio_Handle hMdio,
                                           const CpswMdio_Config *cfg)
{
    CSL_MDIO_setPollEnableMask(hMdio->mdioRegs, cfg->pollEnableMask);
    CSL_MDIO_enableStatusChangeModeInterrupt(hMdio->mdioRegs);
    CSL_MDIO_enableStateChangeMode(hMdio->mdioRegs);
}

static void CpswMdio_handleIntr(CpswMdio_Handle hMdio,
                                CpswMdio_Callbacks *callbacks)
{
    CpswMdio_PhyStatus phyStatus;
    uint32_t phyAddr;
    uint32_t isStatusChangeMode;
    uint32_t pollEnableMask = CPSW_MDIO_PHY_ADDR_MASK_NONE;
    uint32_t linkChanged;
    uint32_t i;
    bool linkInt0Changed = false;
    bool linkInt1Changed = false;

    linkChanged = CSL_MDIO_isUnmaskedLinkStatusChangeIntSet(hMdio->mdioRegs, 0U);
    if (linkChanged == 1U)
    {
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 0U);
        linkInt0Changed = true;
    }

    linkChanged = CSL_MDIO_isUnmaskedLinkStatusChangeIntSet(hMdio->mdioRegs, 1U);
    if (linkChanged == 1U)
    {
        CSL_MDIO_clearUnmaskedLinkStatusChangeInt(hMdio->mdioRegs, 1U);
        linkInt1Changed = true;
    }

    /* Only report state change of PHYs being polled */
    isStatusChangeMode = CSL_MDIO_isStateChangeModeEnabled(hMdio->mdioRegs);
    if (isStatusChangeMode == 0)
    {
        for (i = 0U; i < CPSW_MDIO_PHY_MONITOR_MAX; i++)
        {
            phyAddr         = CSL_MDIO_getLinkStatusChangePhyAddr(hMdio->mdioRegs, i);
            pollEnableMask |= CPSW_MDIO_PHY_ADDR_MASK(phyAddr);
        }

        phyStatus.aliveMask  = CPSW_MDIO_PHY_ADDR_MASK_NONE;
        phyStatus.linkedMask = hMdio->mdioRegs->LINK_REG & pollEnableMask;
    }
    else
    {
        pollEnableMask = CSL_MDIO_getPollEnableMask(hMdio->mdioRegs);

        phyStatus.aliveMask  = hMdio->mdioRegs->ALIVE_REG & pollEnableMask;
        phyStatus.linkedMask = hMdio->mdioRegs->LINK_REG & pollEnableMask;
    }

    /* MDIO_LINKINT[0] event handling */
    if (linkInt0Changed)
    {
        if (callbacks->linkStateCb != NULL)
        {
            callbacks->linkStateCb(CPSW_MDIO_USER_CHANNEL_0,
                                   &phyStatus,
                                   callbacks->appData);
        }
    }

    /* MDIO_LINKINT[1] event handling */
    if (linkInt1Changed)
    {
        if (isStatusChangeMode == 0U)
        {
            if (callbacks->linkStateCb != NULL)
            {
                callbacks->linkStateCb(CPSW_MDIO_USER_CHANNEL_1,
                                       &phyStatus,
                                       callbacks->appData);
            }
        }
        else
        {
            /* MDIO_LINKINT[1] is unexpected in State Change Mode */
            /* TODO: Report an error message */
        }
    }
}

static int32_t CpswMdio_readRegC45(CpswMdio_Handle hMdio,
                                   uint8_t phyAddr,
                                   CpswMdio_Clause45Mmd mmd,
                                   uint16_t reg,
                                   uint16_t *val)
{
    int32_t status;
    bool isComplete;

    /* Wait for any ongoing transaction to complete */
    do
    {
        isComplete = CSL_MDIO_isPhyRegAccessComplete(hMdio->mdioRegs, hMdio->userCh);
    } while (isComplete == FALSE);

    /* Initiate register read */
    status = CSL_MDIO_phyInitiateRegReadC45(hMdio->mdioRegs,
                                            hMdio->userCh,
                                            phyAddr,
                                            (uint32_t)mmd,
                                            reg);
    CPSWTRACE_ERROR_IF((status != CSL_PASS),
                       "failed to initiate C45 register read: %d\n", status);

    /* Wait for register read transaction to complete */
    if (status == CSL_PASS)
    {
        do
        {
            isComplete = CSL_MDIO_isPhyRegAccessComplete(hMdio->mdioRegs, hMdio->userCh);
        } while (isComplete == FALSE);
    }

    /* Get the value read from PHY register once transaction is complete */
    if (status == CSL_PASS)
    {
        status = CSL_MDIO_phyGetRegReadVal(hMdio->mdioRegs, hMdio->userCh, val);
        CPSWTRACE_ERROR_IF((status == CSL_ETIMEOUT),
                           "C45 register read was not acknowledged by PHY: %d\n", status);
        CPSWTRACE_ERROR_IF((status == CSL_EFAIL),
                           "failed to read C45 register: %d\n", status);
    }

    return status;
}

static int32_t CpswMdio_writeRegC45(CpswMdio_Handle hMdio,
                                    uint8_t phyAddr,
                                    CpswMdio_Clause45Mmd mmd,
                                    uint16_t reg,
                                    uint16_t val)
{
    int32_t status;
    bool isComplete;

    /* Wait for any ongoing transaction to complete */
    do
    {
        isComplete = CSL_MDIO_isPhyRegAccessComplete(hMdio->mdioRegs, hMdio->userCh);
    } while (isComplete == FALSE);

    /* Initiate register write */
    status = CSL_MDIO_phyInitiateRegWriteC45(hMdio->mdioRegs,
                                            hMdio->userCh,
                                            phyAddr,
                                            (uint32_t)mmd,
                                             reg,
                                             val);
    CPSWTRACE_ERROR_IF((status != CSL_PASS),
                       "failed to initiate C45 register write: %d\n", status);

    /* Wait for register write transaction to complete */
    if (status == CSL_PASS)
    {
        do
        {
            isComplete = CSL_MDIO_isPhyRegAccessComplete(hMdio->mdioRegs, hMdio->userCh);
        } while (isComplete == FALSE);
    }

    return status;
}

static void CpswMdio_printRegs(CSL_mdioHandle mdioRegs)
{
    uint32_t *regAddr = (uint32_t *)mdioRegs;
    uint32_t regIdx   = 0;

    while (((uintptr_t)regAddr) < ((uintptr_t)mdioRegs + sizeof(*mdioRegs)))
    {
        if (*regAddr)
        {
            CpswUtils_printf("MDIO: %u: 0x%08x\n", regIdx, *regAddr);
        }

        regAddr++;
        regIdx++;
    }
}
