/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_rm.c
 *
 *  \brief This file contains CPSW Resource management functionalities.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_rm.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>

#include "priv/cpsw_priv.h"
#include "priv/cpsw_rm_priv.h"
#include "priv/cpsw_trace_priv.h"
/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

typedef enum  CpswRm_ResourceType_tag
{
    CPSW_RM_RESOURCE_TXCH,
    CPSW_RM_RESOURCE_RXFLOW,
    CPSW_RM_RESOURCE_MACADDRESS,
    CPSW_RM_RESOURCE_LAST = CPSW_RM_RESOURCE_MACADDRESS,
} CpswRm_ResourceType_e;

#define CPSW_RM_RESOURCE_COUNT                      (CPSW_RM_RESOURCE_LAST + 1)
#define CPSW_RM_RESOURCE_ID_ANY                     (~0U)
#define CPSW_RM_RESOURCE_INTERNAL_FLAG              (1U << 31U)
#define CPSWRM_GET_INTERNAL_ALLOC_COREID(coreId)   ((coreId) | CPSW_RM_RESOURCE_INTERNAL_FLAG)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct CpswRm_ResourceEntry_s
{
    node_t node;
    uint32_t ownerCoreId;
    uint32_t id;
}  CpswRm_ResourceEntry_t;

typedef struct CpswRm_CoreResourceInfo_s
{
    uint32_t coreId;
    queue_t resourceQ;
} CpswRm_CoreResourceInfo_t;

typedef struct CpswRm_CoreResourceTable_s
{
    uint32_t numCores;
    CpswRm_CoreResourceInfo_t coreResourceTbl[CPSW_REMOTE_CLIENT_CORES_MAX];
} CpswRm_CoreResourceTable_t;

typedef struct CpswRm_TxChObj_s
{
    uint32_t txPSILThreadIdOffset;
    uint32_t resourceCount;
    CpswRm_CoreResourceTable_t txResourceTable;
    CpswRm_ResourceEntry_t txResource[CPSW_DMA_MAX_TX_CH];
} CpswRm_TxChObj;

typedef struct CpswRm_RxFlowIdxObj_s
{
    uint32_t internalAllocCoreId;
    uint32_t internalAllocCount;
    uint32_t resourceCount;
    CpswRm_CoreResourceTable_t rxResourceTable;
    CpswRm_ResourceEntry_t rxResource[CPSW_DMA_MAX_RX_FLOW];
} CpswRm_RxFlowIdxObj;

typedef struct CpswRm_MacAddressObj_s
{
    uint32_t resourceCount;
    CpswRm_CoreResourceTable_t macResourceTable;
    CpswRm_ResourceEntry_t macResource[CPSW_RM_NUM_MACADDRESS_MAX];
} CpswRm_MacAddressObj;

typedef struct CpswRm_CoreAttachInfo_s
{
    uint32_t numCores;
    uint32_t attachedCores[CPSW_REMOTE_CLIENT_CORES_MAX];
} CpswRm_CoreAttachInfo;

typedef struct CpswRm_Obj_s
{
    CpswRm_Config cfg;

    CpswRm_TxChObj txObj;

    CpswRm_RxFlowIdxObj rxObj;

    CpswRm_MacAddressObj macObj;

    CpswRm_CoreAttachInfo coreAttachObj;

    uint32_t magic;
} CpswRm_Obj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static int32_t CpswRm_attachCore(CpswRm_Handle hRm,
                                 uint32_t coreId,
                                 uint32_t *coreKey);

static int32_t CpswRm_detachCore(CpswRm_Handle hRm,
                                 uint32_t coreKey);

static int32_t CpswRm_allocRxFlowIdx(CpswRm_Handle hRm,
                                     uint32_t coreKey,
                                     uint32_t *startIdx,
                                     uint32_t *flowIdx);

static int32_t CpswRm_freeRxFlowIdx(CpswRm_Handle hRm,
                                    uint32_t coreKey,
                                    uint32_t flowIdx);

static int32_t CpswRm_allocTxChNum(CpswRm_Handle hRm,
                                   uint32_t coreKey,
                                   uint32_t *txPSILThreadId);

static int32_t CpswRm_freeTxChNum(CpswRm_Handle hRm,
                                  uint32_t coreKey,
                                  uint32_t txChNum);

static int32_t CpswRm_allocMacAddr(CpswRm_Handle hRm,
                                   uint32_t coreKey,
                                   uint8_t macAddr[]);

static int32_t CpswRm_freeMacAddr(CpswRm_Handle hRm,
                                  uint32_t coreKey,
                                  uint8_t macAddr[]);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Cpsw DMA Resource manager object */
CpswRm_Obj gCpswRmObj[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
static node_t *CpswRm_allocResourceNode(CpswRm_CoreResourceTable_t *rscTbl,
                                        uint32_t coreId)
{
    uint32_t i;
    node_t *resource;

    Cpsw_assert(rscTbl->numCores <= CPSW_UTILS_ARRAYSIZE(rscTbl->coreResourceTbl));
    for (i = 0; i < rscTbl->numCores; i++)
    {
        if (coreId == rscTbl->coreResourceTbl[i].coreId)
        {
            break;
        }
    }

    if (i < rscTbl->numCores)
    {
        resource = CpswUtils_deQ(&rscTbl->coreResourceTbl[i].resourceQ);
    }
    else
    {
        resource = NULL;
    }

    return resource;
}

static CpswRm_ResourceEntry_t *CpswRm_allocResource(CpswRm_CoreResourceTable_t *rscTbl,
                                                    uint32_t coreId)
{
    node_t *rscNode;
    CpswRm_ResourceEntry_t *resource;

    rscNode = CpswRm_allocResourceNode(rscTbl, coreId);
    if (rscNode != NULL)
    {
        resource = container_of(rscNode, CpswRm_ResourceEntry_t, node);
        memset(&resource->node, 0, sizeof(resource->node));
        Cpsw_assert(resource->ownerCoreId == CPSW_RM_INVALIDCORE);
        resource->ownerCoreId = coreId;
    }
    else
    {
        resource = NULL;
    }

    return resource;
}

static CpswRm_ResourceEntry_t *CpswRm_findResource(CpswRm_ResourceEntry_t *rscTable,
                                                   uint32_t numTableEntries,
                                                   uint32_t coreId,
                                                   uint32_t rscId)
{
    CpswRm_ResourceEntry_t *resource;
    uint32_t i;

    for (i = 0; i < numTableEntries; i++)
    {
        if (rscTable[i].ownerCoreId == coreId)
        {
            if ((rscTable[i].id == rscId)
                ||
                (rscId == CPSW_RM_RESOURCE_ID_ANY))
            {
                break;
            }
        }
    }

    if (i < numTableEntries)
    {
        resource = &rscTable[i];
    }
    else
    {
        resource = NULL;
    }

    return resource;
}

static void CpswRm_initResourceNode(CpswRm_ResourceEntry_t *rsc,
                                    uint32_t rscId)
{
    memset(rsc, 0, sizeof(*rsc));
    rsc->ownerCoreId = CPSW_RM_INVALIDCORE;
    rsc->id          = rscId;
}

static void  CpswRm_initResourceQ(queue_t *rscQ,
                                  CpswRm_ResourceEntry_t *rscNodes,
                                  uint32_t rscNodeLen,
                                  uint32_t startRscIdx,
                                  uint32_t numResource)
{
    uint32_t i;

    CpswUtils_initQ(rscQ);
    for (i = 0; i < numResource; i++)
    {
        Cpsw_assert(i < rscNodeLen);
        CpswRm_initResourceNode(&rscNodes[i], (startRscIdx + i));
        CpswUtils_enQ(rscQ, &rscNodes[i].node);
    }
}

static void  CpswRm_initCoreResourceQ(CpswRm_CoreResourceInfo_t *coreRscTbl,
                                      uint32_t coreId,
                                      CpswRm_ResourceEntry_t *rscNodes,
                                      uint32_t rscNodeLen,
                                      uint32_t startRscIdx,
                                      uint32_t numResource)
{
    coreRscTbl->coreId = coreId;
    CpswRm_initResourceQ(&coreRscTbl->resourceQ, rscNodes, rscNodeLen, startRscIdx, numResource);
}

static queue_t *CpswRm_getFreeQ(CpswRm_CoreResourceTable_t *rscTbl,
                                uint32_t coreId)
{
    uint32_t i;
    queue_t *queue;

    Cpsw_assert(rscTbl->numCores <= CPSW_UTILS_ARRAYSIZE(rscTbl->coreResourceTbl));
    for (i = 0; i < rscTbl->numCores; i++)
    {
        if ((coreId == rscTbl->coreResourceTbl[i].coreId)
            ||
            (coreId == CPSWRM_GET_INTERNAL_ALLOC_COREID(rscTbl->coreResourceTbl[i].coreId)))
        {
            break;
        }
    }

    if (i < rscTbl->numCores)
    {
        queue = &rscTbl->coreResourceTbl[i].resourceQ;
    }
    else
    {
        queue = NULL;
    }

    return queue;
}

static int32_t CpswRm_freeResource(CpswRm_CoreResourceTable_t *rscTbl,
                                   CpswRm_ResourceEntry_t *rscList,
                                   uint32_t rscCount,
                                   uint32_t coreId,
                                   uint32_t rscId)
{
    CpswRm_ResourceEntry_t *resource;
    int32_t status;
    queue_t *freeQ;

    freeQ = CpswRm_getFreeQ(rscTbl, coreId);
    if (freeQ == NULL)
    {
        status = CPSW_EFAIL;
    }
    else
    {
        status = CPSW_SOK;
    }

    if (CPSW_SOK == status)
    {
        resource = CpswRm_findResource(rscList, rscCount, coreId, rscId);
        if (resource != NULL)
        {
            Cpsw_assert(resource->ownerCoreId == coreId);

            /* Do not use argument rscId to init resource as it may be RSC_ANY
             * which is an invalid id
             */
            CpswRm_initResourceNode(resource, resource->id);
            CpswUtils_enQ(freeQ, &resource->node);
            status = CPSW_SOK;
        }
        else
        {
            status = CPSW_EFAIL;
        }
    }

    return status;
}

static void CpswRm_freeCoreResource(CpswRm_CoreResourceTable_t *rscTbl,
                                    CpswRm_ResourceEntry_t *rscList,
                                    uint32_t numTableEntries,
                                    uint32_t coreId)
{
    int32_t status;

    do
    {
        status = CpswRm_freeResource(rscTbl, rscList, numTableEntries, coreId, CPSW_RM_RESOURCE_ID_ANY);
    }
    while (status == CPSW_SOK);
}

static void    CpswRm_initTxObj(CpswRm_TxChObj *txObj,
                                const CpswRm_Config *rmCfg)
{
    const CpswSoc_Config *pSocCfg;
    uint32_t i;
    uint32_t allocRscCount;
    uint32_t coreCount;
    const uint32_t rscTableSize = CPSW_UTILS_ARRAYSIZE(txObj->txResource);

    pSocCfg                         = CpswSoc_getConfig(rmCfg->cpswType);
    txObj->txPSILThreadIdOffset     = pSocCfg->txChPeerThreadId;
    txObj->txResourceTable.numCores = rmCfg->resourcePartitionInfo.numCores;
    allocRscCount                   = 0;
    for (i = 0; i < rmCfg->resourcePartitionInfo.numCores; i++)
    {
        coreCount = rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].numTxCh;
        Cpsw_assert((allocRscCount + coreCount) <= pSocCfg->txChCount);
        Cpsw_assert((allocRscCount + coreCount) <= rscTableSize);
        CpswRm_initCoreResourceQ(&txObj->txResourceTable.coreResourceTbl[i],
                                 rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].coreId,
                                 &txObj->txResource[allocRscCount],
                                 (rscTableSize - allocRscCount),
                                 allocRscCount,
                                 coreCount);
        allocRscCount += coreCount;
    }

    txObj->resourceCount = allocRscCount;
}

static void    CpswRm_initRxObj(CpswRm_RxFlowIdxObj *rxObj,
                                const CpswRm_Config *rmCfg)
{
    const CpswSoc_Config *pSocCfg;
    uint32_t i;
    uint32_t allocRscCount;
    uint32_t coreCount;
    const uint32_t rscTableSize = CPSW_UTILS_ARRAYSIZE(rxObj->rxResource);

    pSocCfg                         = CpswSoc_getConfig(rmCfg->cpswType);
    rxObj->rxResourceTable.numCores = rmCfg->resourcePartitionInfo.numCores;
    allocRscCount                   = 0;
    for (i = 0; i < rmCfg->resourcePartitionInfo.numCores; i++)
    {
        coreCount = rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].numRxFlows;
        Cpsw_assert((allocRscCount + coreCount) <= pSocCfg->rxFlowCount);
        Cpsw_assert((allocRscCount + coreCount) < rscTableSize);
        CpswRm_initCoreResourceQ(&rxObj->rxResourceTable.coreResourceTbl[i],
                                 rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].coreId,
                                 &rxObj->rxResource[allocRscCount],
                                 (rscTableSize - allocRscCount),
                                 allocRscCount,
                                 coreCount);
        allocRscCount += coreCount;
    }

    rxObj->resourceCount       = allocRscCount;
    rxObj->internalAllocCount  = 0;
    rxObj->internalAllocCoreId = CPSW_RM_INVALIDCORE;
}

static void    CpswRm_initMacObj(CpswRm_MacAddressObj *macObj,
                                 const CpswRm_Config *rmCfg)
{
    uint32_t i;
    uint32_t allocRscCount;
    uint32_t coreCount;
    const uint32_t rscTableSize = CPSW_UTILS_ARRAYSIZE(macObj->macResource);

    macObj->macResourceTable.numCores = rmCfg->resourcePartitionInfo.numCores;
    allocRscCount                     = 0;
    for (i = 0; i < rmCfg->resourcePartitionInfo.numCores; i++)
    {
        coreCount = rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].numMacAddress;
        Cpsw_assert((allocRscCount + coreCount) < rscTableSize);
        Cpsw_assert((allocRscCount + coreCount) <= rmCfg->macList.numMacAddress);
        CpswRm_initCoreResourceQ(&macObj->macResourceTable.coreResourceTbl[i],
                                 rmCfg->resourcePartitionInfo.coreDmaResourceInfo[i].coreId,
                                 &macObj->macResource[allocRscCount],
                                 (rscTableSize - allocRscCount),
                                 allocRscCount,
                                 coreCount);
        allocRscCount += coreCount;
    }

    macObj->resourceCount = allocRscCount;
}

static void CpswRm_initAttachObj(CpswRm_CoreAttachInfo *coreAttachObj)
{
    coreAttachObj->numCores = 0;
}

static int32_t CpswRm_mapMacAddr2ResourceIdx(CpswRm_MacAddressPool *macPool,
                                             uint8_t macAddr[],
                                             uint32_t *rscIndex)
{
    uint32_t i;
    int32_t status;

    for (i = 0; i < macPool->numMacAddress; i++)
    {
        if (memcmp(&macPool->macAddress[i][0], macAddr, CPSW_MAC_ADDR_LEN) == 0)
        {
            break;
        }
    }

    if (i < macPool->numMacAddress)
    {
        *rscIndex = i;
        status    = CPSW_SOK;
    }
    else
    {
        status = CPSW_EFAIL;
    }

    return status;
}

static CpswRm_Handle CpswRm_getRmObj(Cpsw_Type cpswType)
{
    CpswRm_Obj *pRm = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            pRm = &gCpswRmObj[cpswType];
            break;

        default:
            Cpsw_assert(FALSE);
    }

    return pRm;
}

static int32_t CpswRm_validateCpswType(Cpsw_Type cpswType)
{
    int32_t status = CPSW_SOK;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    if (NULL == ipInfo)
    {
        CPSWTRACE_ERROR("CPSWALE:Unsupported CPSW_TYPE:%d", cpswType);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static int32_t CpswRm_validateResourcePartitionInfo(const CpswRm_Config *rmConfig,
                                                    uint32_t maxTxRscCount,
                                                    uint32_t maxRxRscCount,
                                                    uint32_t maxMacRscCount)
{
    int32_t status = CPSW_SOK;
    uint32_t i, txChCount = 0U, rxFlowCount = 0U, macAddressCount = 0U;
    const CpswSoc_Config *pSocCfg;
    const CpswRm_ResourcePrms *resourcePartitionInfo = &rmConfig->resourcePartitionInfo;

    pSocCfg = CpswSoc_getConfig(rmConfig->cpswType);

    if (resourcePartitionInfo->numCores >
        CPSW_UTILS_ARRAYSIZE(resourcePartitionInfo->coreDmaResourceInfo))
    {
        status = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == status)
    {
        for (i = 0U; i < resourcePartitionInfo->numCores; i++)
        {
            txChCount       += resourcePartitionInfo->coreDmaResourceInfo[i].numTxCh;
            rxFlowCount     += resourcePartitionInfo->coreDmaResourceInfo[i].numRxFlows;
            macAddressCount += resourcePartitionInfo->coreDmaResourceInfo[i].numMacAddress;
        }

        if ((txChCount > pSocCfg->txChCount)
            ||
            (txChCount > maxTxRscCount))
        {
            status = CPSW_EINVALID_PARAMS;
        }

        if ((rxFlowCount > rmConfig->rxFlowIdxCnt)
            ||
            (rxFlowCount > maxRxRscCount))
        {
            status = CPSW_EINVALID_PARAMS;
        }

        if ((macAddressCount > rmConfig->macList.numMacAddress)
            ||
            (macAddressCount > maxMacRscCount)
            ||
            (macAddressCount > CPSW_UTILS_ARRAYSIZE(rmConfig->macList.macAddress))
            ||
            (rmConfig->macList.numMacAddress > CPSW_UTILS_ARRAYSIZE(rmConfig->macList.macAddress)))
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

static int32_t CpswRm_validateConfig(const CpswRm_Config *rmConfig,
                                     uint32_t maxTxRscCount,
                                     uint32_t maxRxRscCount,
                                     uint32_t maxMacRscCount)
{
    int32_t status = CPSW_SOK;

    if (NULL != rmConfig)
    {
        status = CpswRm_validateCpswType(rmConfig->cpswType);
    }
    else
    {
        status = CPSW_EBADARGS;
    }

    if (CPSW_SOK == status)
    {
        status = CpswRm_validateResourcePartitionInfo(rmConfig, maxTxRscCount, maxRxRscCount, maxMacRscCount);
    }

    return status;
}

static bool CpswRm_isCoreAttached(CpswRm_CoreAttachInfo *coreAttachInfo,
                                  uint32_t coreId,
                                  uint32_t *attachIndex)
{
    bool coreAttached;
    uint32_t i;

    /* Initialize to avoid compiler warning.*/
    if (NULL != attachIndex)
    {
        *attachIndex = 0;
    }

    for (i = 0; i < coreAttachInfo->numCores; i++)
    {
        if (coreAttachInfo->attachedCores[i] == coreId)
        {
            break;
        }
    }

    if (i < coreAttachInfo->numCores)
    {
        coreAttached = true;
        if (attachIndex != NULL)
        {
            *attachIndex = i;
        }
    }
    else
    {
        coreAttached = false;
    }

    return coreAttached;
}

static int32_t CpswRm_validateCoreKey(CpswRm_Handle hRm,
                                      uint32_t coreKey)
{
    int32_t status = CPSW_SOK;
    uint32_t coreId;
    CpswRm_Obj *pRm = hRm;

    coreId = CPSW_COREKEY_2_COREID(coreKey);

    if (CpswRm_isCoreAttached(&pRm->coreAttachObj, coreId, (uint32_t *)NULL_PTR) == false)
    {
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static void CpswRm_addCoreAttached(CpswRm_CoreAttachInfo *coreAttachInfo,
                                   uint32_t coreId)
{
    bool coreAttached;

    coreAttached = CpswRm_isCoreAttached(coreAttachInfo, coreId, (uint32_t *)NULL_PTR);
    Cpsw_assert(coreAttached == false);
    Cpsw_assert(coreAttachInfo->numCores < CPSW_UTILS_ARRAYSIZE(coreAttachInfo->attachedCores));
    coreAttachInfo->attachedCores[coreAttachInfo->numCores] = coreId;
    coreAttachInfo->numCores++;
}

static void CpswRm_delCoreAttached(CpswRm_CoreAttachInfo *coreAttachInfo,
                                   uint32_t coreId)
{
    bool coreAttached;
    uint32_t attachIndex;

    coreAttached = CpswRm_isCoreAttached(coreAttachInfo, coreId, &attachIndex);
    Cpsw_assert(coreAttached == true);
    Cpsw_assert(attachIndex < CPSW_UTILS_ARRAYSIZE(coreAttachInfo->attachedCores));
    Cpsw_assert(coreAttachInfo->numCores > 0);
    /* move the last entry to the detached core slot */
    coreAttachInfo->numCores--;
    coreAttachInfo->attachedCores[attachIndex] =
        coreAttachInfo->attachedCores[coreAttachInfo->numCores];
}

static CpswRm_ResourceInfo *CpswRm_getCoreResourceInfo(CpswRm_Config *cfg,
                                                       uint32_t coreId)
{
    uint32_t i;
    CpswRm_ResourceInfo *rscInfo;

    Cpsw_assert(cfg->resourcePartitionInfo.numCores <= CPSW_UTILS_ARRAYSIZE(cfg->resourcePartitionInfo.coreDmaResourceInfo));
    for (i = 0; i < cfg->resourcePartitionInfo.numCores; i++)
    {
        if (cfg->resourcePartitionInfo.coreDmaResourceInfo[i].coreId == coreId)
        {
            break;
        }
    }

    if (i < cfg->resourcePartitionInfo.numCores)
    {
        rscInfo = &cfg->resourcePartitionInfo.coreDmaResourceInfo[i];
    }
    else
    {
        rscInfo = NULL;
    }

    return rscInfo;
}

static int32_t CpswRm_validateRxFlow(CpswRm_Handle hRm,
                                     uint32_t coreKey,
                                     uint32_t flowIdx)
{
    int32_t status             = CPSW_SOK;
    CpswRm_RxFlowIdxObj *rxObj = &hRm->rxObj;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (CPSW_SOK == status)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);
        CpswRm_ResourceEntry_t *rsc;

        rsc =
            CpswRm_findResource(rxObj->rxResource,
                                CPSW_UTILS_ARRAYSIZE(rxObj->rxResource),
                                coreId,
                                flowIdx);
        if (rsc != NULL)
        {
            status = CPSW_SOK;
        }
        else
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

static int32_t CpswRm_allocInternalRxFlowIdx(CpswRm_RxFlowIdxObj *rxObj,
                                             uint32_t coreId,
                                             uint32_t *flowIdx,
                                             bool internalAllocation)
{
    CpswRm_ResourceEntry_t *rxResource;
    int32_t status;

    rxResource = CpswRm_allocResource(&rxObj->rxResourceTable, coreId);
    if (rxResource != NULL)
    {
        Cpsw_assert(rxResource->ownerCoreId == coreId);
        if (internalAllocation)
        {
            rxResource->ownerCoreId = CPSWRM_GET_INTERNAL_ALLOC_COREID(rxResource->ownerCoreId);
        }

        Cpsw_assert(rxResource->id < rxObj->resourceCount);
        *flowIdx = rxResource->id;
        status   = CPSW_SOK;
    }
    else
    {
        status = CPSW_EALLOC;
    }

    return status;
}

static int32_t CpswRm_allocRxFlowIdx(CpswRm_Handle hRm,
                                     uint32_t coreKey,
                                     uint32_t *startIdx,
                                     uint32_t *flowIdx)
{
    CpswRm_Obj *pRm = hRm;
    int32_t status;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);

        status = CpswRm_allocInternalRxFlowIdx(&hRm->rxObj, coreId, flowIdx, false);
        if (CPSW_SOK == status)
        {
            *startIdx = pRm->cfg.rxStartFlowIdx;
            Cpsw_assert(*flowIdx < pRm->cfg.rxFlowIdxCnt);
        }
    }

    return status;
}

static int32_t CpswRm_allocTxChNum(CpswRm_Handle hRm,
                                   uint32_t coreKey,
                                   uint32_t *txPSILThreadId)
{
    int32_t status;
    CpswRm_Obj *pRm = hRm;

    *txPSILThreadId = CPSW_RM_TXCHNUM_INVALID;
    status          = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        CpswRm_ResourceEntry_t *txResource;
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);

        txResource = CpswRm_allocResource(&pRm->txObj.txResourceTable, coreId);
        if (txResource != NULL)
        {
            const CpswSoc_Config *pSocCfg;

            pSocCfg = CpswSoc_getConfig(hRm->cfg.cpswType);
            Cpsw_assert(txResource->ownerCoreId == coreId);
            Cpsw_assert(txResource->id < pRm->txObj.resourceCount);
            Cpsw_assert(txResource->id < pSocCfg->txChCount);
            *txPSILThreadId = txResource->id + pRm->txObj.txPSILThreadIdOffset;
            status          = CPSW_SOK;
        }
        else
        {
            status = CPSW_EALLOC;
        }
    }

    return status;
}

static int32_t CpswRm_allocMacAddr(CpswRm_Handle hRm,
                                   uint32_t coreKey,
                                   uint8_t macAddr[])
{
    int32_t status;
    CpswRm_Obj *pRm = hRm;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        CpswRm_ResourceEntry_t *macResource;
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);

        macResource = CpswRm_allocResource(&pRm->macObj.macResourceTable, coreId);
        if (macResource != NULL)
        {
            Cpsw_assert(macResource->ownerCoreId == coreId);
            Cpsw_assert(macResource->id < pRm->cfg.macList.numMacAddress);
            Cpsw_assert(macResource->id < pRm->macObj.resourceCount);
            Cpsw_assert(macResource->id < CPSW_UTILS_ARRAYSIZE(pRm->cfg.macList.macAddress));
            memcpy(macAddr, &pRm->cfg.macList.macAddress[macResource->id][0], CPSW_MAC_ADDR_LEN);
            status = CPSW_SOK;
        }
        else
        {
            status = CPSW_EALLOC;
        }
    }

    return status;
}

static int32_t CpswRm_freeInternalRxFlowIdx(CpswRm_RxFlowIdxObj *rxObj,
                                            uint32_t coreId,
                                            uint32_t flowIdx,
                                            bool internalAllocation)
{
    int32_t status = CPSW_SOK;
    uint32_t matchCoreId;

    if (internalAllocation)
    {
        matchCoreId = CPSWRM_GET_INTERNAL_ALLOC_COREID(coreId);
    }
    else
    {
        matchCoreId = coreId;
    }

    Cpsw_assert(rxObj->resourceCount <= CPSW_UTILS_ARRAYSIZE(rxObj->rxResource));
    status = CpswRm_freeResource(&rxObj->rxResourceTable,
                                 rxObj->rxResource,
                                 rxObj->resourceCount,
                                 matchCoreId,
                                 flowIdx);
    return status;
}

static int32_t CpswRm_freeRxFlowIdx(CpswRm_Handle hRm,
                                    uint32_t coreKey,
                                    uint32_t flowIdx)
{
    int32_t status  = CPSW_SOK;
    CpswRm_Obj *pRm = hRm;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);

        CpswRm_freeInternalRxFlowIdx(&pRm->rxObj, coreId, flowIdx, false);
    }

    return status;
}

static int32_t CpswRm_freeTxChNum(CpswRm_Handle hRm,
                                  uint32_t coreKey,
                                  uint32_t txChNum)
{
    int32_t status          = CPSW_SOK;
    CpswRm_Obj *pRm         = hRm;
    CpswRm_TxChObj *txChObj = &pRm->txObj;
    uint32_t txChOffset     = txChNum - txChObj->txPSILThreadIdOffset;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);

        Cpsw_assert(pRm->txObj.resourceCount <= CPSW_UTILS_ARRAYSIZE(pRm->txObj.txResource));

        status = CpswRm_freeResource(&pRm->txObj.txResourceTable,
                                     pRm->txObj.txResource,
                                     pRm->txObj.resourceCount,
                                     coreId,
                                     txChOffset);
    }

    return status;
}

static int32_t CpswRm_freeMacAddr(CpswRm_Handle hRm,
                                  uint32_t coreKey,
                                  uint8_t macAddr[])
{
    int32_t status  = CPSW_SOK;
    CpswRm_Obj *pRm = hRm;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);
        uint32_t macIdx;

        status = CpswRm_mapMacAddr2ResourceIdx(&hRm->cfg.macList, macAddr, &macIdx);

        if (status == CPSW_SOK)
        {
            Cpsw_assert(pRm->macObj.resourceCount <= CPSW_UTILS_ARRAYSIZE(pRm->macObj.macResource));
            status = CpswRm_freeResource(&pRm->macObj.macResourceTable,
                                         pRm->macObj.macResource,
                                         pRm->macObj.resourceCount,
                                         coreId,
                                         macIdx);
        }
    }

    return status;
}

static int32_t CpswRm_validateCoreIoctlPrivilege(CpswRm_Obj *pRm,
                                                 uint32_t cmd,
                                                 uint32_t coreId)
{
    uint32_t i;
    bool ioctlPermitted;
    int32_t status;

    for (i = 0; i < pRm->cfg.ioctlPermissionInfo.numEntries; i++)
    {
        if (cmd == pRm->cfg.ioctlPermissionInfo.entry[i].cmd)
        {
            break;
        }
    }

    if (i < pRm->cfg.ioctlPermissionInfo.numEntries)
    {
        ioctlPermitted = CPSW_IS_BIT_SET(pRm->cfg.ioctlPermissionInfo.entry[i].permittedCoreMask, coreId);
    }
    else
    {
        ioctlPermitted = CPSW_IS_BIT_SET(pRm->cfg.ioctlPermissionInfo.defaultPermittedCoreMask, coreId);
    }

    if (ioctlPermitted)
    {
        status = CPSW_SOK;
    }
    else
    {
        status = CPSW_EPERM;
    }

    return status;
}

static int32_t CpswRm_attachCore(CpswRm_Handle hRm,
                                 uint32_t coreId,
                                 uint32_t *coreKey)
{
    int32_t status = CPSW_SOK;
    bool coreAttached;

    coreAttached = CpswRm_isCoreAttached(&hRm->coreAttachObj, coreId, (uint32_t *)NULL_PTR);
    if (coreAttached == false)
    {
        CpswRm_addCoreAttached(&hRm->coreAttachObj, coreId);
        *coreKey = CPSW_COREID_2_COREKEY(coreId);
    }
    else
    {
        status = CPSW_EALREADY_OPEN;
    }

    return status;
}

static int32_t CpswRm_detachCore(CpswRm_Handle hRm,
                                 uint32_t coreKey)
{
    int32_t status  = CPSW_SOK;
    CpswRm_Obj *pRm = hRm;

    status = CpswRm_validateCoreKey(hRm, coreKey);

    if (status == CPSW_SOK)
    {
        uint32_t coreId = CPSW_COREKEY_2_COREID(coreKey);
        bool coreAttached;

        coreAttached = CpswRm_isCoreAttached(&pRm->coreAttachObj, coreId, (uint32_t *)NULL_PTR);
        if (coreAttached == true)
        {
            queue_t *txFreeQ             = CpswRm_getFreeQ(&pRm->txObj.txResourceTable, coreId);
            queue_t *rxFreeQ             = CpswRm_getFreeQ(&pRm->rxObj.rxResourceTable, coreId);
            queue_t *macFreeQ            = CpswRm_getFreeQ(&pRm->macObj.macResourceTable, coreId);
            CpswRm_ResourceInfo *rscInfo = CpswRm_getCoreResourceInfo(&pRm->cfg, coreId);
            uint32_t internalRxAllocCount;

            if (rscInfo == NULL)
            {
                Cpsw_assert(false);
            }
            else
            {
                Cpsw_assert(pRm->txObj.resourceCount <= CPSW_UTILS_ARRAYSIZE(pRm->txObj.txResource));
                CpswRm_freeCoreResource(&pRm->txObj.txResourceTable,
                                        pRm->txObj.txResource,
                                        pRm->txObj.resourceCount,
                                        coreId);
                if (NULL != txFreeQ)
                {
                    Cpsw_assert(CpswUtils_getQCount(txFreeQ) == rscInfo->numTxCh);
                }

                Cpsw_assert(pRm->rxObj.resourceCount <= CPSW_UTILS_ARRAYSIZE(pRm->rxObj.rxResource));
                CpswRm_freeCoreResource(&pRm->rxObj.rxResourceTable,
                                        pRm->rxObj.rxResource,
                                        pRm->rxObj.resourceCount,
                                        coreId);
                internalRxAllocCount = 0;
                if (coreId == pRm->rxObj.internalAllocCoreId)
                {
                    internalRxAllocCount = pRm->rxObj.internalAllocCount;
                }

                if (NULL != rxFreeQ)
                {
                    Cpsw_assert((CpswUtils_getQCount(rxFreeQ) + internalRxAllocCount) == rscInfo->numRxFlows);
                }

                Cpsw_assert(pRm->macObj.resourceCount <= CPSW_UTILS_ARRAYSIZE(pRm->macObj.macResource));
                CpswRm_freeCoreResource(&pRm->macObj.macResourceTable,
                                        pRm->macObj.macResource,
                                        pRm->macObj.resourceCount,
                                        coreId);
                if (NULL != macFreeQ)
                {
                    Cpsw_assert(CpswUtils_getQCount(macFreeQ) == rscInfo->numMacAddress);
                }

                CpswRm_delCoreAttached(&pRm->coreAttachObj, coreId);
            }
        }
        else
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

uint32_t CpswRm_ioctl(CpswRm_Handle hRm,
                      CpswRm_Ioctls_e cmd,
                      Cpsw_IoctlPrms *prms)
{
    int32_t status  = CPSW_SOK;
    CpswRm_Obj *pRm = hRm;

    /* Decode the command and act on it */
    switch (cmd)
    {
        case CPSW_RM_IOCTL_ALLOC_RX_FLOW:
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(CpswRm_AllocRxFlowOutArgs));
            if (status == CPSW_SOK)
            {
                uint32_t coreKey                   = *((uint32_t *)prms->inArgs);
                CpswRm_AllocRxFlowOutArgs *outArgs = (CpswRm_AllocRxFlowOutArgs *)prms->outArgs;

                status = CpswRm_allocRxFlowIdx(pRm,
                                               coreKey,
                                               &outArgs->startIdx,
                                               &outArgs->flowIdx);
            }

            break;

        case CPSW_RM_IOCTL_ALLOC_TX_CH_PEERID:
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                uint32_t coreKey         = *((uint32_t *)prms->inArgs);
                uint32_t *txPSILThreadId = (uint32_t *)prms->outArgs;

                status = CpswRm_allocTxChNum(pRm,
                                             coreKey,
                                             txPSILThreadId);
            }

            break;

        case CPSW_RM_IOCTL_ALLOC_MAC_ADDR:
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(CpswRm_AllocMacAddrOutArgs));
            if (status == CPSW_SOK)
            {
                uint32_t coreKey                    = *((uint32_t *)prms->inArgs);
                CpswRm_AllocMacAddrOutArgs *outArgs = (CpswRm_AllocMacAddrOutArgs *)prms->outArgs;

                status = CpswRm_allocMacAddr(pRm,
                                             coreKey,
                                             &outArgs->macAddr[0U]);
            }

            break;

        case CPSW_RM_IOCTL_FREE_RX_FLOW:
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_FreeRxFlowInArgs));
            if (status == CPSW_SOK)
            {
                CpswRm_FreeRxFlowInArgs *inArgs = (CpswRm_FreeRxFlowInArgs *)prms->inArgs;

                status = CpswRm_freeRxFlowIdx(pRm,
                                              inArgs->coreKey,
                                              inArgs->flowIdx);
            }

            break;

        case CPSW_RM_IOCTL_FREE_TX_CH_PEERID:
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_FreeTxChInArgs));
            if (status == CPSW_SOK)
            {
                CpswRm_FreeTxChInArgs *inArgs = (CpswRm_FreeTxChInArgs *)prms->inArgs;

                status = CpswRm_freeTxChNum(pRm,
                                            inArgs->coreKey,
                                            inArgs->txChNum);
            }

            break;

        case CPSW_RM_IOCTL_FREE_MAC_ADDR:
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_FreeMacAddrInArgs));
            if (status == CPSW_SOK)
            {
                CpswRm_FreeMacAddrInArgs *inArgs = (CpswRm_FreeMacAddrInArgs *)prms->inArgs;

                status = CpswRm_freeMacAddr(pRm,
                                            inArgs->coreKey,
                                            &inArgs->macAddr[0U]);
            }

            break;

        default:
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

int32_t CpswRm_ioctlPriv(CpswRm_Handle hRm,
                         CpswRm_PrivIoctls_e cmd,
                         Cpsw_IoctlPrms *prms)
{
    CpswRm_Obj *pRm = hRm;
    int32_t status  = CPSW_SOK;

    /* Decode the command and act on it */
    switch (cmd)
    {
        case CPSW_RM_IOCTL_VALIDATE_PERMISSION:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_ValidatePermissionInArgs));
            if (status == CPSW_SOK)
            {
                const CpswRm_ValidatePermissionInArgs *inArgs = (const CpswRm_ValidatePermissionInArgs *)prms->inArgs;

                status = CpswRm_validateCoreIoctlPrivilege(pRm,
                                                           inArgs->cmd,
                                                           inArgs->coreId);
            }

            break;
        }

        case CPSW_RM_IOCTL_DETACH:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                uint32_t *pCoreKey = (uint32_t *)prms->inArgs;

                status = CpswRm_detachCore(pRm, *pCoreKey);
            }

            break;
        }

        case CPSW_RM_IOCTL_ATTACH:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                uint32_t *pCoreId  = (uint32_t *)prms->inArgs;
                uint32_t *pCoreKey = (uint32_t *)prms->outArgs;

                status = CpswRm_attachCore(pRm, *pCoreId, pCoreKey);
            }

            break;
        }

        case CPSW_RM_IOCTL_INTERNAL_ALLOC_RX_FLOW:
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(uint32_t),
                                         sizeof(CpswRm_AllocRxFlowOutArgs));
            if (status == CPSW_SOK)
            {
                uint32_t *coreId                   = (uint32_t *)prms->inArgs;
                CpswRm_AllocRxFlowOutArgs *outArgs = (CpswRm_AllocRxFlowOutArgs *)prms->outArgs;

                Cpsw_assert((pRm->rxObj.internalAllocCoreId == CPSW_RM_INVALIDCORE)
                            ||
                            (pRm->rxObj.internalAllocCoreId == *coreId));
                status = CpswRm_allocInternalRxFlowIdx(&pRm->rxObj,
                                                       *coreId,
                                                       &outArgs->flowIdx,
                                                       true);
                if (CPSW_SOK == status)
                {
                    outArgs->startIdx = pRm->cfg.rxStartFlowIdx;
                    Cpsw_assert(outArgs->flowIdx < pRm->cfg.rxFlowIdxCnt);
                    pRm->rxObj.internalAllocCoreId = *coreId;
                    pRm->rxObj.internalAllocCount++;
                }
            }

            break;

        case CPSW_RM_IOCTL_INTERNAL_FREE_RX_FLOW:
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_FreeInternalRxFlowInArgs));
            if (status == CPSW_SOK)
            {
                CpswRm_FreeInternalRxFlowInArgs *inArgs = (CpswRm_FreeInternalRxFlowInArgs *)prms->inArgs;

                status = CpswRm_freeInternalRxFlowIdx(&pRm->rxObj,
                                                      inArgs->coreId,
                                                      inArgs->flowIdx,
                                                      true);
                if (status == CPSW_SOK)
                {
                    Cpsw_assert(pRm->rxObj.internalAllocCount > 0);
                    pRm->rxObj.internalAllocCount--;
                    if (pRm->rxObj.internalAllocCount == 0)
                    {
                        pRm->rxObj.internalAllocCoreId = CPSW_RM_INVALIDCORE;
                    }
                }
            }

            break;

        case CPSW_RM_IOCTL_VALIDATE_RX_FLOW:
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswRm_ValidateRxFlowInArgs));
            if (status == CPSW_SOK)
            {
                CpswRm_ValidateRxFlowInArgs *inArgs = (CpswRm_ValidateRxFlowInArgs *)prms->inArgs;

                status = CpswRm_validateRxFlow(hRm,
                                               inArgs->coreKey,
                                               inArgs->flowIdx);
            }

            break;

        default:
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }

    return status;
}

CpswRm_Handle CpswRm_Open(const CpswRm_Config *rmConfig)
{
    int32_t status;
    CpswRm_Obj *pRm = NULL;

    status = CpswRm_validateConfig(rmConfig,
                                   CPSW_UTILS_ARRAYSIZE(gCpswRmObj[0].txObj.txResource),
                                   CPSW_UTILS_ARRAYSIZE(gCpswRmObj[0].rxObj.rxResource),
                                   CPSW_UTILS_ARRAYSIZE(gCpswRmObj[0].macObj.macResource));

    if (CPSW_SOK == status)
    {
        pRm = CpswRm_getRmObj(rmConfig->cpswType);

        memset(pRm, 0U, sizeof(CpswRm_Obj));
        pRm->cfg = *rmConfig;
        CpswRm_initTxObj(&pRm->txObj, rmConfig);
        CpswRm_initRxObj(&pRm->rxObj, rmConfig);
        CpswRm_initMacObj(&pRm->macObj, rmConfig);
        CpswRm_initAttachObj(&pRm->coreAttachObj);
        pRm->magic = CPSW_RM_MAGIC;
    }

    return pRm;
}

void CpswRm_close(CpswRm_Handle hRm)
{
    uint32_t i      = 0U;
    CpswRm_Obj *pRm = hRm;
    uint32_t txFreeRscCount, rxFreeRscCount, macFreeRscCount;
    CpswRm_ResourcePrms *resourcePartitionInfo = &pRm->cfg.resourcePartitionInfo;
    queue_t *pQ;

    /* Assert all internally allocated flows are freed */
    Cpsw_assert((pRm->rxObj.internalAllocCoreId == CPSW_RM_INVALIDCORE)
                &&
                (pRm->rxObj.internalAllocCount == 0));
    Cpsw_assert(resourcePartitionInfo->numCores <=
                CPSW_UTILS_ARRAYSIZE(resourcePartitionInfo->coreDmaResourceInfo));
    txFreeRscCount  = 0;
    rxFreeRscCount  = 0;
    macFreeRscCount = 0;

    for (i = 0; i < resourcePartitionInfo->numCores; i++)
    {
        CpswRm_detachCore(hRm, CPSW_COREID_2_COREKEY(resourcePartitionInfo->coreDmaResourceInfo[i].coreId));

        pQ = CpswRm_getFreeQ(&pRm->txObj.txResourceTable, resourcePartitionInfo->coreDmaResourceInfo[i].coreId);
        if (NULL != pQ)
        {
            txFreeRscCount += CpswUtils_getQCount(pQ);
        }

        pQ = CpswRm_getFreeQ(&pRm->rxObj.rxResourceTable, resourcePartitionInfo->coreDmaResourceInfo[i].coreId);
        if (NULL != pQ)
        {
            rxFreeRscCount += CpswUtils_getQCount(pQ);
        }

        pQ = CpswRm_getFreeQ(&pRm->macObj.macResourceTable, resourcePartitionInfo->coreDmaResourceInfo[i].coreId);
        if (NULL != pQ)
        {
            macFreeRscCount += CpswUtils_getQCount(pQ);
        }
    }

    Cpsw_assert(txFreeRscCount == pRm->txObj.resourceCount);
    Cpsw_assert(rxFreeRscCount == pRm->rxObj.resourceCount);
    Cpsw_assert(macFreeRscCount == pRm->macObj.resourceCount);
    memset(pRm, 0, sizeof(*pRm));
}

/*!
 * @}
 */
