/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_utils.c
 *
 *  \brief File containing the CPSW driver utility functions.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* This is needed for vsnprintf */
#include <stdio.h>
#include <stdarg.h>

/* This is needed for memset/memcpy */
#include <string.h>

#include <ti/drv/cpsw/include/cpsw_cfg.h>

#include "priv/cpsw_priv.h"
#include "priv/cpsw_osal.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct CpswUtils_Obj_s
{
    Cpsw_VirtToPhyFxn virtToPhyFxn;

    Cpsw_PhyToVirtFxn phyToVirtFxn;

    Cpsw_PrintFxnCb printFxn;

    Cpsw_TraceFxnCb traceFxn;

    void *printLock;

    char printBuf[CPSW_CFG_PRINT_BUF_LEN];

    void *traceLock;

    char traceBuf[CPSW_CFG_TRACE_BUF_LEN];
} CpswUtils_Obj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static uint64_t CpswUtils_defaultVirtToPhyFxn(const void *virtAddr,
                                              void *appData);

static void *CpswUtils_defaultPhyToVirtFxn(uint64_t phyAddr,
                                           void *appData);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswUtils_Obj gCpswUtilsObj;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswUtils_printf(const char *format,
                      ...)
{
#if defined(CPSW_CFG_PRINT_ENABLE)
    va_list vaArgPtr;
    char *buf;

    if (gCpswUtilsObj.printFxn != NULL)
    {
        CpswOsal_mutexLock(gCpswUtilsObj.printLock);

        buf = &gCpswUtilsObj.printBuf[0];
        (void)va_start(vaArgPtr, format);
        (void)vsnprintf(buf, CPSW_CFG_PRINT_BUF_LEN, (const char *)format, vaArgPtr);
        va_end(vaArgPtr);

        gCpswUtilsObj.printFxn(buf);

        CpswOsal_mutexUnlock(gCpswUtilsObj.printLock);
    }
#endif

    return;
}

uint64_t CpswUtils_virtToPhyFxn(const void *virtAddr,
                                void *appData)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswUtilsObj.virtToPhyFxn != NULL);
#endif
    return gCpswUtilsObj.virtToPhyFxn(virtAddr, appData);
}

void *CpswUtils_phyToVirtFxn(uint64_t phyAddr,
                             void *appData)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswUtilsObj.phyToVirtFxn != NULL);
#endif
    return gCpswUtilsObj.phyToVirtFxn(phyAddr, appData);
}

void *CpswUtils_registerIntr(CpswOsal_IsrFxn isrFxn,
                             uint32_t coreIntrNum,
                             uint32_t intrPriority,
                             void *arg)
{
    OsalRegisterIntrParams_t intrPrms;
    OsalInterruptRetCode_e osalRetVal;
    HwiP_Handle hwiHandle = NULL;

    Osal_RegisterInterrupt_initParams(&intrPrms);

    /* Populate the interrupt parameters */
    intrPrms.corepacConfig.arg             = (uintptr_t)arg;
    intrPrms.corepacConfig.isrRoutine      = isrFxn;
    intrPrms.corepacConfig.priority        = intrPriority;
    intrPrms.corepacConfig.corepacEventNum = 0U;    // TODO: Should be set based on core?
    intrPrms.corepacConfig.intVecNum       = coreIntrNum;
#ifdef QNX_OS
    intrPrms.corepacConfig.intAutoEnable    = 1;
#endif

    /* Register interrupts */
    osalRetVal = Osal_RegisterInterrupt(&intrPrms, &hwiHandle);
    if (OSAL_INT_SUCCESS != osalRetVal)
    {
        hwiHandle = NULL;
    }

    return(hwiHandle);
}

void CpswUtils_unRegisterIntr(void *hwiHandle)
{
    int32_t corepacEventNum = 0U;       // TODO: Should be set based on core?

    /* Delete interrupts */
    Osal_DeleteInterrupt((HwiP_Handle)hwiHandle, corepacEventNum);

    return;
}

void CpswUtils_enableIntr(uint32_t coreIntrNum)
{
    int32_t corepacEventNum = 0U;       // TODO: Should be set based on core?

    Osal_EnableInterrupt(corepacEventNum, coreIntrNum);

    return;
}

void CpswUtils_disableIntr(uint32_t coreIntrNum)
{
    int32_t corepacEventNum = 0U;       // TODO: Should be set based on core?

    Osal_DisableInterrupt(corepacEventNum, coreIntrNum);

    return;
}

/* We need only mutex in CPSW driver. Use semaphore in binary mode */
void *CpswUtils_mutexCreate(void)
{
    SemaphoreP_Params semPrms;
    SemaphoreP_Handle mutexHandle;

    SemaphoreP_Params_init(&semPrms);
    semPrms.mode = SemaphoreP_Mode_BINARY;
    mutexHandle  = (void *)SemaphoreP_create(1U, &semPrms);

    return(mutexHandle);
}

void CpswUtils_mutexDelete(void *mutexHandle)
{
    SemaphoreP_delete((SemaphoreP_Handle)mutexHandle);
}

void CpswUtils_mutexLock(void *mutexHandle)
{
    SemaphoreP_pend((SemaphoreP_Handle)mutexHandle, SemaphoreP_WAIT_FOREVER);
}

void CpswUtils_mutexUnlock(void *mutexHandle)
{
    SemaphoreP_post((SemaphoreP_Handle)mutexHandle);
}

void CpswUtils_init(CpswUtils_Prms *pUtilsPrms)
{
    memset(&gCpswUtilsObj, 0U, sizeof(CpswUtils_Obj));

    gCpswUtilsObj.printFxn = pUtilsPrms->printFxn;

    gCpswUtilsObj.printLock = CpswOsal_mutexCreate();
    gCpswUtilsObj.traceLock = CpswOsal_mutexCreate();

    if (NULL != pUtilsPrms->traceFxn)
    {
        gCpswUtilsObj.traceFxn = pUtilsPrms->traceFxn;
    }

    if (NULL != pUtilsPrms->virtToPhyFxn)
    {
        gCpswUtilsObj.virtToPhyFxn = pUtilsPrms->virtToPhyFxn;
    }
    else
    {
        gCpswUtilsObj.virtToPhyFxn = &CpswUtils_defaultVirtToPhyFxn;
    }

    if (NULL != pUtilsPrms->phyToVirtFxn)
    {
        gCpswUtilsObj.phyToVirtFxn = pUtilsPrms->phyToVirtFxn;
    }
    else
    {
        gCpswUtilsObj.phyToVirtFxn = &CpswUtils_defaultPhyToVirtFxn;
    }

    return;
}

void CpswUtils_deInit(void)
{
    CpswOsal_mutexDelete(gCpswUtilsObj.printLock);

    CpswOsal_mutexDelete(gCpswUtilsObj.traceLock);

    gCpswUtilsObj.printFxn = NULL;

    if (NULL != gCpswUtilsObj.traceFxn)
    {
        gCpswUtilsObj.traceFxn = NULL;
    }

    return;
}

static uint64_t CpswUtils_defaultVirtToPhyFxn(const void *virtAddr,
                                              void *appData)
{
    return((uint64_t)virtAddr);
}

static void *CpswUtils_defaultPhyToVirtFxn(uint64_t phyAddr,
                                           void *appData)
{
#if defined(__aarch64__)
    uint64_t temp = phyAddr;
#else
    /* R5 is 32-bit machine, need to truncate to avoid void * typecast error */
    uint32_t temp = (uint32_t)phyAddr;
#endif

    return((void *)temp);
}

void CpswUtils_trace(const char *format,
                     va_list vaArgPtr)
{
#if (CPSW_TRACE_CFG_TRACE_LEVEL > CPSW_TRACE_CFG_LEVEL_NONE)
    char *buf;

    if (gCpswUtilsObj.traceFxn != NULL)
    {
        CpswOsal_mutexLock(gCpswUtilsObj.traceLock);

        buf = &gCpswUtilsObj.traceBuf[0];
        vsnprintf(buf, CPSW_CFG_TRACE_BUF_LEN, (const char *)format, vaArgPtr);

        gCpswUtilsObj.traceFxn(buf);

        CpswOsal_mutexUnlock(gCpswUtilsObj.traceLock);
    }
#endif
}

void CpswUtils_checkDescState(uint32_t *pStateVar,
                              CpswUtils_DescStateMemMgr expectedState,
                              CpswUtils_DescStateMemMgr newState)
{
#if (CPSW_CFG_DEV_ERROR == 1)
    Cpsw_assert(CPSW_UTILS_GET_DESC_MEMUTILS_STATE(pStateVar) == expectedState);
    CPSW_UTILS_SET_DESC_MEMUTILS_STATE(pStateVar, newState);
#endif
}

void CpswUtils_checkPktState(uint32_t *pStateVar,
                             CpswUtils_PktStateModuleType module,
                             uint32_t expectedState,
                             uint32_t newState)
{
#if (CPSW_CFG_DEV_ERROR == 1)
    switch (module)
    {
        case CPSW_PKTSTATE_MODULE_APP:
            Cpsw_assert(CPSW_UTILS_GET_PKT_APP_STATE(pStateVar) == expectedState);
            CPSW_UTILS_SET_PKT_APP_STATE(pStateVar, newState);
            break;

        case CPSW_PKTSTATE_MODULE_DRIVER:
            Cpsw_assert(CPSW_UTILS_GET_PKT_DRIVER_STATE(pStateVar) == expectedState);
            CPSW_UTILS_SET_PKT_DRIVER_STATE(pStateVar, newState);
            break;

        case CPSW_PKTSTATE_MODULE_MEMUTILS:
            Cpsw_assert(CPSW_UTILS_GET_PKT_MEMUTILS_STATE(pStateVar) == expectedState);
            CPSW_UTILS_SET_PKT_MEMUTILS_STATE(pStateVar, newState);
            break;

        default:
            break;
    }
#endif
}

int32_t CpswUtils_timerInit(void)
{
    return CPSW_SOK;
}

int32_t CpswUtils_timerDeInit(void)
{
    return CPSW_SOK;
}

uint32_t CpswUtils_timerRead(void)
{
    uint32_t time = 0U;

    // TODO - add timer code once profiling branch is merged
    return time;
}

uint32_t CpswUtils_min(uint32_t num1,
                       uint32_t num2)
{
    return((num1 < num2) ? num1 : num2);
}

uint32_t CpswUtils_max(uint32_t num1,
                       uint32_t num2)
{
    return((num1 > num2) ? num1 : num2);
}

uint32_t CpswUtils_getIpFeatureBitMap(Cpsw_Type cpswType)
{
    uint32_t bitmap = 0U;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    if (ipInfo != NULL)
    {
        bitmap = ipInfo->featureBitmap;
    }
    else
    {
        Cpsw_assert(false);
    }

    return bitmap;
}

uint32_t CpswUtils_getIpErrataBitMap(Cpsw_Type cpswType)
{
    uint32_t bitmap = 0U;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    if (ipInfo != NULL)
    {
        bitmap = ipInfo->errataBitmap;
    }
    else
    {
        Cpsw_assert(false);
    }

    return bitmap;
}

const CpswSoc_CpswIpInfo *CpswUtils_getIpInfo(Cpsw_Type cpswType)
{
    uint32_t i;
    const CpswSoc_CpswIpInfo *ipInfo = NULL;
    const CpswSoc_CpswIpInfoTbl *ipInfoTbl = CpswSoc_getIPInfoTbl();
    bool foundIp = false;

    for (i = 0; i < ipInfoTbl->numCpswIPs; i++)
    {
        if (ipInfoTbl->info[i].cpswType == cpswType)
        {
            foundIp = true;
            break;
        }
    }

    if (foundIp)
    {
        ipInfo = &ipInfoTbl->info[i];
    }

    return ipInfo;
}

void CpswUtils_delay(uint32_t delayVal)
{
    uint32_t i = 0U;
    while(i++ < delayVal);
}

/* end of file */
