/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file vsc8514.c
 *
 * \brief VSC8514 PHY specific implementation
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/src/priv/cpsw_priv.h>
#include <ti/drv/cpsw/src/priv/cpsw_phy_priv.h>
#include <ti/drv/cpsw/src/priv/cpsw_trace_priv.h>
#include <ti/drv/cpsw/include/vsc8514.h>
#include "generic_phy.h"
#include "vsc8514_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define VSC8514_OUI                           (0x000001C1U)
#define VSC8514_MODEL                         (0x27U)
#define VSC8514_REV                           (0x00U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static bool Vsc8514_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version);

static bool Vsc8514_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode);

static int32_t Vsc8514_writeMicroCmd(CpswPhy_Handle hPhyDev);

static int32_t Vsc8514_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode);

static void Vsc8514_reset(CpswPhy_Handle hPhyDev);

static bool Vsc8514_isResetComplete(CpswPhy_Handle hPhyDev);

static void Vsc8514_rmwExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t mask,
                              uint16_t shift,
                              uint16_t val);

static void Vsc8514_readExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t *val);

static void Vsc8514_writeExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t val);

static void Vsc8514_printRegs(CpswPhy_Handle hPhyDev);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswPhy_Drv gCpswPhyDrvVsc8514 =
{
    .name               = "vsc8514",
    .isPhyDevSupported  = Vsc8514_isPhyDevSupported,
    .isMacModeSupported = Vsc8514_isMacModeSupported,
    .config             = Vsc8514_config,
    .reset              = Vsc8514_reset,
    .isResetComplete    = Vsc8514_isResetComplete,
    .runComplianceTest  = NULL,
    .readExtReg         = GenericPhy_readExtReg,
    .writeExtReg        = GenericPhy_writeExtReg,
    .printRegs          = Vsc8514_printRegs,
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void Vsc8514_initParams(Vsc8514_Config *config)
{

}

static bool Vsc8514_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version)
{
    bool supported = false;

    if ((version->oui == VSC8514_OUI) &&
        (version->model == VSC8514_MODEL) &&
        (version->revision == VSC8514_REV))
    {
        supported = true;
    }

    return supported;
}

static bool Vsc8514_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode)
{
    bool supported;

    switch (mode)
    {
        case CPSW_PHY_MAC_MODE_QSGMII:
            supported = true;
            break;

        default:
            supported = false;
            break;
    }

    return supported;
}

static int32_t Vsc8514_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode)
{
    const Vsc8514_Config *extendedCfg = (const Vsc8514_Config *)cfg->extendedConfig;
    uint32_t extendedCfgSize          = cfg->extendedConfigSize;
    int32_t status                    = CPSW_SOK;

    if ((extendedCfg == NULL) ||
        (extendedCfgSize != sizeof(*extendedCfg)))
    {
        CPSWTRACE_ERROR("PHY %u: invalid config params (cfg=%p, size=%u)\n",
                        hPhyDev->addr, extendedCfg, extendedCfgSize);
        status = CPSW_EINVALID_PARAMS;
    }

    /* Configure MAC configuration register to set QSGMII mac mode */
    Vsc8514_rmwExtReg(hPhyDev,
                      VSC8514_EXTREGSET_GPIO,
                      VSC8514_GPIO_MACMODE,
                      VSC8514_GPIO_MACMODE_MACMODE_MASK,
                      VSC8514_GPIO_MACMODE_MACMODE_SHIFT,
                      VSC8514_GPIO_MACMODE_MACMODE_QSGMII);

    /* Enable 4 ports MAC QSGMII using micro command register */
    status = Vsc8514_writeMicroCmd(hPhyDev);
    CPSWTRACE_ERROR_IF((CPSW_SOK != status), "PHY %u: Micro command write failed \n",
                       hPhyDev->addr);
    Cpsw_assert(CPSW_SOK == status);

    /* Configure media operating mode in PHY extended control register */
    CpswPhy_writeReg(hPhyDev, VSC8514_EXTENDED_PHYCTRL1, 0x0U);

    /* Software reset PHY to apply initial SGMII Configure */
    Vsc8514_reset(hPhyDev);
    while(false == Vsc8514_isResetComplete(hPhyDev))
    {
    }

    /* Enable MAC interface restart on media link change */
    Vsc8514_rmwExtReg(hPhyDev,
                      VSC8514_EXTREGSET_PAGE3,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_MACIF_RESTART_MASK,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_MACIF_RESTART_SHIFT,
                      1U);

    /* Enable Auto-negotiation */
    Vsc8514_rmwExtReg(hPhyDev,
                      VSC8514_EXTREGSET_PAGE3,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_AUTONEG_EN_MASK,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_AUTONEG_EN_SHIFT,
                      1U);

    /* Restart Auto-negotiation */
    Vsc8514_rmwExtReg(hPhyDev,
                      VSC8514_EXTREGSET_PAGE3,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_AUTONEG_RESTART_MASK,
                      VSC8514_EXT3_MACSERDES_PCS_CTRL_AUTONEG_RESTART_SHIFT,
                      1U);

    return status;
}

static void Vsc8514_reset(CpswPhy_Handle hPhyDev)
{
    GenericPhy_reset(hPhyDev);
}

static bool Vsc8514_isResetComplete(CpswPhy_Handle hPhyDev)
{
    return GenericPhy_isResetComplete(hPhyDev);
}

static void Vsc8514_readExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t *val)
{
    CPSWTRACE_VERBOSE("PHY %u: Read reg %u val 0x%04x\n",
                      hPhyDev->addr, reg, val);

    /* set page select */
    CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, extRegSet);

    CpswPhy_readReg(hPhyDev, reg, val);

    /* Set page select register back to 0x0 to access main register space  */
    CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, VSC8514_EXTREGSET_MAIN);
}

static void Vsc8514_writeExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t val)
{
    CPSWTRACE_VERBOSE("PHY %u: write reg %u val 0x%04x\n",
                      hPhyDev->addr, reg, val);

    /* set page select */
    CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, extRegSet);

    CpswPhy_writeReg(hPhyDev, reg, val);

    /* Set page select register back to 0x0 to access main register space  */
    CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, VSC8514_EXTREGSET_MAIN);
}

static void Vsc8514_rmwExtReg(CpswPhy_Handle hPhyDev,
                              Vsc8514_ExtRegSetType extRegSet,
                              uint32_t reg,
                              uint16_t mask,
                              uint16_t shift,
                              uint16_t val)
{
    uint16_t data;

    CPSWTRACE_VERBOSE("PHY %u: write extended reg %u mask 0x%04x val 0x%04x\n",
                      hPhyDev->addr, reg, mask, val);

    Vsc8514_readExtReg(hPhyDev, extRegSet, reg, &data);

    CSL_REG16_FINS_RAW(&data, mask, shift, val);

    Vsc8514_writeExtReg(hPhyDev, extRegSet, reg, data);
}

static int32_t Vsc8514_writeMicroCmd(CpswPhy_Handle hPhyDev)
{
    uint16_t regVal;
    int32_t status = CPSW_SOK;

    /* set page select */
    status = CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, VSC8514_EXTREGSET_GPIO);

    if(CPSW_SOK == status)
    {
        while(true)
        {
            status = CpswPhy_readReg(hPhyDev, VSC8514_GPIO_COMMAND, &regVal);

            /* wait until not busy */
            if(0U == (regVal & VSC8514_GPIO_COMMAND_GO_MASK))
            {
                break;
            }
            CpswUtils_delay(100U);
        }
    }

    if(CPSW_SOK == status)
    {
        /* Enable 4 ports MAC QSGMII */
        regVal = (VSC8514_GPIO_COMMAND_GO_SET|VSC8514_GPIO_COMMAND_QSGMII);
        status = CpswPhy_writeReg(hPhyDev, VSC8514_GPIO_COMMAND, regVal);
    }

    if(CPSW_SOK == status)
    {
        while(true)
        {
            /* Commands may take up to 25 ms to complete before command complete bit
             * (bit 15) changes to 0 */
            CpswUtils_delay(2500U);
            status = CpswPhy_readReg(hPhyDev, VSC8514_GPIO_COMMAND, &regVal);

            /* wait until not busy */
            if(VSC8514_GPIO_COMMAND_QSGMII == regVal)
            {
                break;
            }
        }
    }

    if(CPSW_SOK == status)
    {
        /* Set page select register back to 0x0 to access main register space  */
        status = CpswPhy_writeReg(hPhyDev, VSC8514_PAGE_ACCESS, VSC8514_EXTREGSET_MAIN);
    }

    return status;
}


static void Vsc8514_printRegs(CpswPhy_Handle hPhyDev)
{
    uint32_t phyAddr = hPhyDev->addr;
    uint16_t val;

    CpswPhy_readReg(hPhyDev, PHY_BMCR, &val);
    CpswUtils_printf("PHY %u: BMCR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);
    CpswUtils_printf("PHY %u: BMSR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR1, &val);
    CpswUtils_printf("PHY %u: PHYIDR1 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR2, &val);
    CpswUtils_printf("PHY %u: PHYIDR2 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANAR, &val);
    CpswUtils_printf("PHY %u: ANAR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANLPAR, &val);
    CpswUtils_printf("PHY %u: ANLPAR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANER, &val);
    CpswUtils_printf("PHY %u: ANER    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPTR, &val);
    CpswUtils_printf("PHY %u: ANNPTR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPRR, &val);
    CpswUtils_printf("PHY %u: ANNPRR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGCR, &val);
    CpswUtils_printf("PHY %u: CFG1    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGSR, &val);
    CpswUtils_printf("PHY %u: STS1    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGESR, &val);
    CpswUtils_printf("PHY %u: 1KSCR   = 0x%04x\n", phyAddr, val);
}
