/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_dma_priv.h
 *
 *  \brief CPSW DMA private header file.
 */

#ifndef CPSW_DMA_PRIV_H_
#define CPSW_DMA_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_dma.h>
#include <ti/drv/cpsw/include/cpsw_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/** \brief UDMAP descriptor type for CPSW channels */
#define CPSW_DMA_UDMAP_DESC_TYPE     (uint32_t)CSL_UDMAP_CPPI5_PD_DESCINFO_DTYPE_VAL_HOST

/** \brief UDMA max teardown time (in msecs) */
#define CPSW_UDMA_TEARDOWN_TIME_MS   (100U)

/*! \brief Ethernet Jumbo frame size */
#define CPSW_DMA_JUMBO_PACKET_SIZE   (9000U)

/*! \brief UDMA teardown CQ ring memory size */
#define CPSW_DMA_TDCQ_RING_ELE_CNT   (1U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */
/*!
 * \brief CPPI TX (host RX) Info words structure
 * Note - This maps to EPI (extended packet info) block of host descriptor.
 */
typedef struct
{
    /*! TX info word 0 */
    uint32_t word0;

    /*! TX info word 1 */
    uint32_t word1;

    /*! Reserved word */
    uint32_t rsvd;

    /*! TX info word 3 */
    uint32_t word3;
}CpswDma_CppiTxInfo;

/*!
 * \brief CPPI TX (host RX) Status words structure
 * Note - This maps to PSI (protocol specific info) block of host descriptor.
 */
typedef struct
{
    /*! Lower word of 64-bit timestamp */
    uint32_t tsLow;

    /*! Higher word of 64-bit timestamp */
    uint32_t tsHigh;

    /*! pV4/IpV6 protocol information info status word */
    uint32_t chkSumInfo;

    /*! Reserved */
    uint32_t rsvd;
}CpswDma_CppiTxStatus;

/*!
 * \brief CPPI RX (host TX) Info words structure
 * Note - This maps to EPI (extended packet info) block of host descriptor.
 */
typedef struct
{
    /*! RX info word 0 */
    uint32_t word0;

    /*! Reserved word */
    uint32_t rsvd1;

    /*! TX info word 2 */
    uint32_t word2;

    /*! Reserved word */
    uint32_t rsvd2;
}CpswDma_CppiRxInfo;

/*!
 * \brief CPPI RX (host TX) Control words structure
 * Note - This maps to PSI (protocol specific info) block of host descriptor.
 */
typedef struct
{
    /*! Reserved */
    uint32_t rsvd1;

    /*! Timestamp info */
    uint32_t tsInfo;

    /*! IpV4/IpV6 protocol information  */
    uint32_t chkSumInfo;

    /*! Reserved */
    uint32_t rsvd2;
}CpswDma_CppiRxControl;

/*!
 * \brief DMA transfer direction (rx/tx) enum
 */
typedef enum CpswDma_Dir_s
{
    /*! DMA RX flow  */
    CPSW_DMA_DIR_RX = 0,

    /*! DMA Tx channel   */
    CPSW_DMA_DIR_TX = 1
} CpswDma_Dir;

/*!
 * \brief DMA Desc Queue
 *
 * \details
 *      A queue of DMA Descs, which are use for managing the active queue of
 *      of DMA Descs given to the hardware driver.
 */
typedef struct CpswDma_DmaDescQ_s
{
    /*! Number of DMA Descs in queue */
    uint32_t count;

    /*! Pointer to first DMA desc */
    CpswDma_DmaDesc *pHead;

    /*! Pointer to last DMA Descs */
    CpswDma_DmaDesc *pTail;
}
CpswDma_DmaDescQ;

/*!
 * \brief Structure holding UDMA handles for CPSW DMA Rx flow and TX channel.
 */
typedef struct
{
    /*! UDMA driver handle */
    Udma_DrvHandle hUdmaDrv;

    /*! UDMA Channel handle */
    Udma_ChHandle hUdmaCh;

    /*! UDMA Event handle */
    Udma_EventHandle hUdmaEvt;

    /*! UDMA Ring monitor handle (null if ringMon not used)*/
    Udma_RingMonHandle hUdmaRingMon;
} CpswDma_udmaInfo;

/*!
 * \brief Structure holding Ring allocation info.
 */
typedef struct
{
    /*! Ring number to be used in ring alloc function. Use #UDMA_RING_ANY if use
     *   any ring */
    uint16_t ringNum;

    /*! Flag to allocate ring memory. If set, ringPrms.ringMem can not be NULL */
    bool allocRingMem;

    /*! Allocate ring memory callback function */
    CpswDma_AllocRingMemFxn ringMemAllocFxn;

    /*! Free ring memory callback function */
    CpswDma_FreeRingMemFxn ringMemFreeFxn;

    /*! Callback argument for the ring callback functions */
    void *hCbArg;
} CpswDma_ringAllocInfo;


/*!
 * \brief Structure for reserved flow params.
 */
typedef struct CpswDma_OpenRsvdRxFlowPrms_s
{
    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;

    /*! Flow start index*/
    uint32_t startIdx;

    /*! Flow index to be opened*/
    uint32_t flowIdx;
} CpswDma_RsvdRxFlowPrms;

/*!
 * \brief Structure for RX flow memory objects
 */
typedef struct CpswDma_RxFlowMemObj_s
{
    /*! UDMA Rx flow object */
    struct Udma_FlowObj flowUdmaObj;

    /*! UDMA event object */
    struct Udma_EventObj udmaEvtObj;

    /*! Free queue ring object */
    struct Udma_RingObj fqRingObj;

    /*! This FQ is used for when received packet length is larger than app
     *   allocated packet size (PDK-3723) */
    struct Udma_RingObj dropFqRingObj;

    /*! Complete queue ring handle */
    struct Udma_RingObj cqRingObj;

    /*! Udma dedicated proxy object used if rxFlowPrms.useProxy is TRUE */
    // TODO - remove proxyObj and allocate it via dma memory manager
    struct Udma_ProxyObj proxyObj;

    /*! Ready DMA descriptor Q */
    CpswDma_DmaDescQ rxReadyDmaDescQ;
} CpswDma_RxFlowMemObj;

typedef struct CpswDma_RxFlowObj_s
{
    /*! Channel init params */
    CpswDma_OpenRxFlowPrms rxFlowPrms;

    /*! Struct holding RX flow object memories */
    CpswDma_RxFlowMemObj rxFlowMemObj;

    /*! UDMA driver handle */
    Udma_DrvHandle hUdmaDrv;

    /*! RX flow handle */
    Udma_FlowHandle hUdmaFlow;

    /*! UDMA Event object */
    Udma_EventHandle hUdmaEvt;

    /*! Free queue ring object */
    Udma_RingHandle fqRing;

    /*! Complete queue ring handle */
    Udma_RingHandle cqRing;

    /*! DMA descriptor free pool handle. This is used when auto-reclaim of TX FQ as
     *  RX CQ is enabled to use app passed free pool instead of rxReadyDmaDescQ pool */
    CpswDma_DmaDescQ *hDmaDescPool;

    /*! Event initialization state */
    bool evtInitFlag;

    /*! Channel initialization state */
    bool initFlag;

    /*! App private */
    void *appPriv;

    /*! Packet queue into which CQ Ring ISR will drain packets into */
    queue_t cqIsrQ;

    /*! Udma proxy handle. Will be NULL or pointer to proxyObj based on
     *   useProxy is set or not.
     */
    Udma_ProxyHandle hUdmaProxy;

    /*! Udma ring monitor handle. Will be NULL based if useRingMon is not set.
     */
    Udma_RingMonHandle hUdmaRingMon;

    /*! Rx flow app packet and DMA desc stats */
    CpswDma_RxFlowStats stats;
} CpswDma_RxFlowObj;

/*!
 * \brief Structure for Tx channel memory objects
 */
typedef struct CpswDma_TxChMemObj_s
{
    /*! UDMA channel object */
    struct Udma_ChObj udmaChObj;

    /*! UDMA event object */
    struct Udma_EventObj udmaEvtObj;

    /*! Free queue ring object */
    struct Udma_RingObj fqRingObj;

    /*! Complete queue ring handle */
    struct Udma_RingObj cqRingObj;

    /*! Udma dedicated proxy object used if rxFlowPrms.useProxy is TRUE */
    // TODO - remove proxyObj and allocate it via dma memory manager
    struct Udma_ProxyObj proxyObj;

    /*! Ready DMA descriptor Q */
    CpswDma_DmaDescQ txFreeDmaDescQ;
} CpswDma_TxChMemObj;

typedef struct CpswDma_TxChObj_s
{
    /*! Channel init params */
    CpswDma_OpenTxChPrms txChPrms;

    /*! Struct holding TX Channel object memories */
    CpswDma_TxChMemObj txChMemObj;

    /*! UDMA driver handle */
    Udma_DrvHandle hUdmaDrv;

    /*! TX channel handle */
    Udma_ChHandle hUdmaCh;

    /*! UDMA Event object */
    Udma_EventHandle hUdmaEvt;

    /*! Free queue ring handle */
    Udma_RingHandle fqRing;

    /*! Complete queue ring handle */
    Udma_RingHandle cqRing;

    /*! Teardown complete queue ring handle */
    Udma_RingHandle tdCqRing;

    /*! DMA descriptor free pool handle. This is used when auto-reclaim of RX FQ as
     *  TX CQ is enabled to use app passed free pool instead of rxReadyDmaDescQ pool */
    CpswDma_DmaDescQ *hDmaDescPool;

    /*! Event initialization state */
    bool evtInitFlag;

    /*! Channel initialization state */
    bool initFlag;

    /*! App private */
    void *appPriv;

    /*! Packet queue into which CQ Ring ISR will drain packets into */
    queue_t cqIsrQ;

    /*! Udma proxy handle. Will be NULL or pointer to proxyObj based on
     *   useProxy is set or not.
     */
    Udma_ProxyHandle hUdmaProxy;

    /*! Udma ring monitor handle. Will be NULL based if useRingMon is not set.
     */
    Udma_RingMonHandle hUdmaRingMon;

    /*! Tx channel app packet and DMA desc stats */
    CpswDma_TxChStats stats;
} CpswDma_TxChObj;

typedef struct CpswDma_RxChObj_s
{
    /*! CPSW NAVSS instance */
    Cpsw_Type cpswType;

    /*! Channel init params */
    CpswDma_RxChInitPrms rxChInitPrms;

    /*! UDMA channel object */
    struct Udma_ChObj udmaChObj;

    /*! UDMA Rx flow object - as we don't use default flow and allocate new using
     *  Udma_flowAlloc, we need to allocate flow object to our RX channel */
    struct Udma_FlowObj flowUdmaObj;

    /*! Teardown complete queue ring handle */
    Udma_RingHandle tdCqRing;

    /*! Channel initialization state */
    bool initFlag;
} CpswDma_RxChObj;

/**
 *  \brief DMA module
 *
 *  This is an internal/private driver structure and should not be used
 *  or modified by the caller.
 */
typedef struct CpswDma_DrvObj_s
{
    /*! CPSW NAVSS instance */
    Cpsw_Type cpswType;

    /*! UDMA driver handle*/
    Udma_DrvHandle hUdmaDrv;

    /*! Module initialization state */
    bool initFlag;

    /*! Receive channel Obj */
    CpswDma_RxChObj rxChObj;
} CpswDma_DrvObj;

/* ========================================================================== */
/*                         Global Variables                                   */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

void CpswDma_initParams(CpswDma_Config *pDmaConfig);

CpswDma_Handle CpswDma_open(Cpsw_Type cpswType,
                            const CpswDma_Config *pCpswDmaConfig);

int32_t CpswDma_close(CpswDma_Handle hCpswDma);

uint32_t CpswDma_getRxChFlowStartIdx(CpswDma_Handle hCpswDma);

uint32_t CpswDma_getRxFlowCnt(CpswDma_Handle hCpswDma);

/**
 *  \brief CPSW DMA Open function.
 *
 *  This API opens the CPSW DMA RX channel and sets data path module configuration.
 *  This function should be called only by Master core and should be called before
 *  calling any IOCTLs.
 *
 *  \param hCpswDma  [IN] CPSW DMA module handle.
 *  \param rxChInitPrms  [IN] CPSW RX channel init parameters.
 *  \param totalRxFlowCount [IN] Total number of Rx flows to be allocated to the
 *                               channel
 *
 *  \return \ref Cpsw_ErrorCodes
 */
int32_t CpswDma_openRxCh(CpswDma_Handle hCpswDma,
                         const CpswDma_RxChInitPrms *pRxChInitPrms,
                         uint32_t totalRxFlowCount);

/**
 *  \brief CPSW DMA close function.
 *
 *  Stop and close the Rx channel and set deinit DMA module configuration.
 *
 *  \param hRxCh  [IN] Handle to CPSW DMA RX channel returned from call to
 *                   CpswDma_open.
 *
 *  \return \ref Cpsw_ErrorCodes
 */
int32_t CpswDma_closeRxCh(CpswDma_Handle hCpswDma);

int32_t CpswDma_registerEvent(CpswDma_udmaInfo *pUdmaInfo,
                              Udma_RingHandle hUdmaRing,
                              Udma_EventCallback eventCb,
                              void *hCbArg,
                              uint32_t eventType);

int32_t CpswDma_unregisterEvent(Udma_EventHandle hUdmaEvt);

uint32_t CpswDma_getTxChPeerThreadId(Cpsw_Type cpswType,
                                     uint32_t chNum);

uint32_t CpswDma_getRxChPeerThreadId(Cpsw_Type cpswType);

void CpswDma_getRxFlowUdmaInfo(CpswDma_RxFlowHandle hRxFlow,
                               CpswDma_udmaInfo *pUdmaInfo);

void CpswDma_getTxChUdmaInfo(CpswDma_TxChHandle hTxCh,
                             CpswDma_udmaInfo *pUdmaInfo);

int32_t CpswDma_allocRing(Udma_DrvHandle hUdmaDrv,
                            Udma_RingHandle hUdmaRing,
                            Udma_RingPrms *pRingPrms,
                            CpswDma_ringAllocInfo *pRingAllocInfo);

void CpswDma_initRxFreeDescQ(CpswDma_RxFlowObj *pRxFlow);

void CpswDma_deInitRxFreeDescQ(CpswDma_RxFlowObj *pRxFlow);

void CpswDma_initTxFreeDescQ(CpswDma_TxChObj *pTxChObj);

void CpswDma_deInitTxFreeDescQ(CpswDma_TxChObj *pTxChObj);

void CpswDma_reconfigureDescQ(CpswDma_TxChObj *pTxCh);

int32_t CpswDma_ringDequeue(Udma_RingHandle hUdmaRing,
                            CpswDma_DmaDesc **pDmaDesc,
                            bool disableCacheOpsFlag,
                            CpswDma_Dir transferDir);

int32_t CpswDma_ringEnqueue(Udma_RingHandle hUdmaRing,
                            CpswDma_DmaDesc *pDmaDesc,
                            uint32_t packetSize,
                            bool disableCacheOpsFlag,
                            CpswDma_Dir transferDir,
                            Udma_ProxyHandle hUdmaProxy);

int32_t CpswDma_freeTxRing(Udma_RingHandle hUdmaRing,
                           CpswDma_UdmaRingPrms *pRingPrms,
                           CpswDma_OpenTxChPrms *pTxChPrms);

int32_t CpswDma_flushRxFlowRing(CpswDma_RxFlowHandle hRxFlow,
                                Udma_RingHandle hUdmaRing,
                                CpswDma_PktInfoQ *pFqPktInfoQ);

int32_t CpswDma_flushTxChRing(CpswDma_TxChHandle hTxCh,
                              Udma_RingHandle hUdmaRing,
                              CpswDma_PktInfoQ *pFqPktInfoQ);

int32_t CpswDma_freeRing(Udma_RingHandle hUdmaRing,
                           uint32_t numPkts,
                           CpswDma_FreeRingMemFxn ringMemFreeFxn,
                           void *hCbArg);

int32_t CpswDma_submitPkts(Udma_RingHandle hUdmaRing,
                           CpswDma_PktInfoQ *pToHwQueue,
                           CpswDma_DmaDescQ *pDmaDescQ,
                           bool disableCacheOpsFlag,
                           CpswDma_Dir transferDir,
                           Udma_ProxyHandle hUdmaProxy);

int32_t CpswDma_retrievePkts(Udma_RingHandle hUdmaRing,
                             CpswDma_PktInfoQ *pFromHwQueue,
                             CpswDma_DmaDescQ *pDmaDescQ,
                             bool disableCacheOpsFlag,
                             CpswDma_Dir transferDir);

void CpswDma_buffDescInit(CpswDma_CpswHpdDesc *hpdDesc);

void CpswDma_dmaDescInit(CpswDma_DmaDesc *pCpswDmaPacket);

void CpswDma_dmaDescQInit(CpswDma_DmaDescQ *pDmaDescQ);

uint32_t CpswDma_dmaDescQCount(CpswDma_DmaDescQ *pDmaDescQ);

CpswDma_DmaDesc *CpswDma_dmaDescDeque(CpswDma_DmaDescQ *pDmaDescQ);

void CpswDma_dmaDescCopyQ(CpswDma_DmaDescQ *pDstQueue,
                          const CpswDma_DmaDescQ *pSrcQueue);

void CpswDma_dmaDescEnque(CpswDma_DmaDescQ *pDmaDescQ,
                          CpswDma_DmaDesc *pDmaDesc);

int32_t CpswDma_dmaDescCheck(CpswDma_DmaDesc *dmaDesc,
                             CpswDma_Dir transferDir);

CpswDma_RxFlowHandle CpswDma_openRxRsvdFlow(CpswDma_RsvdRxFlowPrms *pRxFlowPrms);

int32_t CpswDma_closeRxRsvdFlow(CpswDma_RxFlowHandle hRxFlow);

/*! Rx flow object allocation function  */
CpswDma_RxFlowObj *CpswDma_memMgrAllocRxFlowObj(void);

/*! Rx flow object free function  */
void CpswDma_memMgrFreeRxFlowObj(CpswDma_RxFlowObj *pRxFlowObj);

/*! TX channel object allocation function  */
CpswDma_TxChObj *CpswDma_memMgrAllocTxChObj(void);

/*! TX channel object free function  */
void CpswDma_memMgrFreeTxChObj(CpswDma_TxChObj *pTxChObj);

/*! Ring Monitor object allocation function  */
Udma_RingMonHandle CpswDma_memMgrAllocRingMonObj(void);

/*! Ring Monitor object free function  */
void CpswDma_memMgrFreeRingMonObj(Udma_RingMonHandle hUdmaRingMon);

/*! TD Cq Ring memory object allocation function  */
void *CpswDma_memMgrAllocTdCqRingMemObj(void);

/*! TD Cq Ring memory object free function  */
void CpswDma_memMgrFreeTdCqRingMemObj(void *pTdCqRingMem);

/*! Drv object memory manager init  */
void CpswDma_memMgrInit(void);

/*! Drv object memory manager deinit function  */
void CpswDma_memMgrDeInit(void);

void CpswDma_txCqIsr(Udma_EventHandle hUdmaEvt,
                     uint32_t eventType,
                     void *appData);

void CpswDma_rxCqIsr(Udma_EventHandle hUdmaEvt,
                     uint32_t eventType,
                     void *appData);

int32_t CpswDma_allocProxy(Udma_DrvHandle hUdmaDrv,
                           Udma_ProxyHandle hUdmaProxy,
                           Udma_RingHandle hUdmaRing);

int32_t CpswDma_freeProxy(Udma_ProxyHandle hUdmaProxy);

Udma_RingMonHandle CpswDma_allocRingMon(Udma_DrvHandle hUdmaDrv,
                                        Udma_RingHandle hUdmaRing,
                                        uint32_t elemCnt,
                                        CpswDma_RingMonCfg *pRingMonCfg);

int32_t CpswDma_freeRingMon(Udma_RingMonHandle hRingMon);

void CpswDmaStats_updateNotifyStats(CpswDma_CbFxnStats *pktStats,
                                    uint32_t pktCnt,
                                    uint32_t timeDiff);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

static inline void CpswDmaStats_addCnt(uint64_t *statCnt,
                                       uint64_t addCnt)
{
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    *statCnt += addCnt;
#endif
}

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_DMA_PRIV_H_ */
