/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_rm_priv.h
 *
 * \brief    This file contains the private definitions of the resource manager
 */

/*!
 * \addtogroup CPSW_RM_API
 * @{
 */

#ifndef CPSW_RM_PRIV_H_
#define CPSW_RM_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
#include "cpsw_osal.h"

#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------------*\
 |                                Header Files                                 |
 \*---------------------------------------------------------------------------*/

/* Project dependency headers */
#include "cpsw_rm.h"

/*---------------------------------------------------------------------------*\
 |                           Global Macros/Defines                             |
 \*---------------------------------------------------------------------------*/
#define CPSW_RM_PRIVATE_IOCTL(x)         (CPSW_IOCTL_RM_BASE |      \
                                          CPSW_IOCTL_TYPE_PRIVATE | \
                                          CPSW_IOCTL_MIN(x))

/*---------------------------------------------------------------------------*\
 |                           Global Typedefs/Enums                             |
 \*---------------------------------------------------------------------------*/

/*!
 * \brief CPSW RM Private IOCTL commands
 */
typedef enum CpswRm_PrivIoctls_tag
{
    /*!
     * \brief Validate IOCTL permission
     *
     * IOCTL parameters:
     *   inArgs:  CpswRm_ValidatePermissionInArgs inArgs
     *   outArgs: none
     */
    CPSW_RM_IOCTL_VALIDATE_PERMISSION    = CPSW_RM_PRIVATE_IOCTL(0U),

    /*!
     * \brief Detach IOCTL
     *
     * IOCTL parameters:
     *   inArgs:  uint32_t coreKey
     *   outArgs: none
     */
    CPSW_RM_IOCTL_DETACH                 = CPSW_RM_PRIVATE_IOCTL(1U),

    /*!
     * \brief Attach IOCTL
     *
     * IOCTL parameters:
     *   inArgs:  uint32_t coreId
     *   outArgs: uint32_t coreKey
     */
    CPSW_RM_IOCTL_ATTACH                 = CPSW_RM_PRIVATE_IOCTL(2U),

    /*!
     * \brief Driver internal rx alloc
     *
     * IOCTL parameters:
     *   inArgs:  uint32_t coreId
     *   outArgs: CpswRm_AllocRxFlowOutArgs rxAllocOutArgs
     */
    CPSW_RM_IOCTL_INTERNAL_ALLOC_RX_FLOW = CPSW_RM_PRIVATE_IOCTL(3U),

    /*!
     * \brief  Driver internal rx free
     *
     * IOCTL parameters:
     *   inArgs:  CpswRm_FreeInternalRxFlowInArgs
     *   outArgs: NONE
     */
    CPSW_RM_IOCTL_INTERNAL_FREE_RX_FLOW  = CPSW_RM_PRIVATE_IOCTL(4U),

    /*!
     * \brief  Validate Rx Flow is allocated resource and is from a attached core
     *
     * IOCTL parameters:
     *   inArgs:  CpswRm_ValidateRxFlowInArgs
     *   outArgs: NONE
     */
    CPSW_RM_IOCTL_VALIDATE_RX_FLOW       = CPSW_RM_PRIVATE_IOCTL(5U),
}CpswRm_PrivIoctls_e;

/*---------------------------------------------------------------------------*\
 |                         Global Variable Declarations                        |
 \*---------------------------------------------------------------------------*/
/*! inArgs for CPSW_RM_IOCTL_VALIDATE_PERMISSION IOCTL CMD */
typedef struct CpswRm_ValidatePermissionInArgs_s
{
    /*! IOCTL command */
    uint32_t cmd;

    /*! Core id from which the cmd is invoked  */
    uint32_t coreId;
}CpswRm_ValidatePermissionInArgs;

/*! inArgs for CPSW_RM_IOCTL_FREE_RX_FLOW IOCTL CMD */
typedef struct CpswRm_FreeInternalRxFlowInArgs_s
{
    /*! Core Key */
    uint32_t coreId;

    /*! Rx Flow Index */
    uint32_t flowIdx;
} CpswRm_FreeInternalRxFlowInArgs;

/*!
 * \brief CPSW RM Configuration
 *        This structure has the configuration
 *        used during initialization.
 */
typedef struct CpswRm_Config_s
{
    /*! CPSW TYPE */
    Cpsw_Type cpswType;

    /*! Rx Flow Start Index */
    uint32_t rxStartFlowIdx;

    /*! Number of Rx Flows allocated */
    uint32_t rxFlowIdxCnt;

    /*! Resource Partition Information */
    CpswRm_ResourcePrms resourcePartitionInfo;

    /*! IOCTL permission info */
    CpswRm_IoctlPermissionTable ioctlPermissionInfo;

    /*! Mac Address List to be managed by RM */
    CpswRm_MacAddressPool macList;
}CpswRm_Config;

/*! inArgs for CPSW_RM_IOCTL_V CMD */
typedef CpswRm_FreeRxFlowInArgs CpswRm_ValidateRxFlowInArgs;

/*---------------------------------------------------------------------------*\
 |                   Global Inline Function Definitions                        |
 \*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 |                         Global Function Declarations                        |
 \*---------------------------------------------------------------------------*/

/*!
 * \brief Open the CPSW Resource Manager
 *
 * \param rmConfig    [IN] Pointer to the RM config structure
 */
CpswRm_Handle CpswRm_Open(const CpswRm_Config *rmConfig);

/*! CPSW RM IOCTL configuration */
uint32_t CpswRm_ioctl(CpswRm_Handle hRm,
                      CpswRm_Ioctls_e cmd,
                      Cpsw_IoctlPrms *prms);

void CpswRm_close(CpswRm_Handle hRm);

/*! CPSW RM IOCTL configuration */
int32_t CpswRm_ioctlPriv(CpswRm_Handle hRm,
                         CpswRm_PrivIoctls_e cmd,
                         Cpsw_IoctlPrms *prms);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_RM_PRIV_H_ */
