/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_switching_logic.c
 *
 * \brief    This file contains the cpsw_switching_logic test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_aging_maxlength.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_TEST_MACPORT_RX_MTU                        (1000U)
#define CPSW_TEST_HOSTPORT_RX_MTU                       (1000U)

/* TX priority MTU should always be higher than mac port MTU */
#define CPSW_TXPRI0_7_MTU                          (1200U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Test ALE Entry Address */
static uint8_t testAddr[CPSW_MAC_ADDR_LEN] =
{
    0x40U, 0xcdU, 0xefU, 0xfeU, 0xdcU, 0xbaU
};

static volatile uint32_t gRxHostMtu = 0U;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
void CpswTestAgingMaxPktLength_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                                 Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
    pLinkArgs->macConfig.rxMtu          = CPSW_TEST_MACPORT_RX_MTU;
}

void CpswTestAgingMaxPktLength_setOpenPrms(CpswTestTaskObj *taskObj,
                                           Cpsw_Config *pCpswCfg,
                                           CpswOsal_Prms *pOsalPrms,
                                           CpswUtils_Prms *pUtilsPrms)
{
    uint32_t i;

    pCpswCfg->aleConfig.agingConfig.enableAutoAging = TRUE;
    pCpswCfg->aleConfig.agingConfig.agingPeriodInMs = 1000;
    pCpswCfg->hostPortConfig.rxMtu                  = CPSW_TEST_HOSTPORT_RX_MTU;
    gRxHostMtu                                      = CPSW_TEST_HOSTPORT_RX_MTU;

    /* Set MAC port global config */
    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        pCpswCfg->txMtu[i] = CPSW_TXPRI0_7_MTU;
    }
}

static void CpswTestAgingMaxPktLength_showStats(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status = CPSW_SOK;
    CpswStats_HostPort_2g *hostPortStats;
    CpswStats_MacPort_2g *macPortStats;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_STATS_IOCTL_GET_HOSTPORT_STATS,
                        &prms);
    if (status == CPSW_SOK)
    {
        hostPortStats = (CpswStats_HostPort_2g *)&portStats;

        CpswAppUtils_print("\n Port 0 Statistics\n");
        CpswAppUtils_print("-----------------------------------------\n");
        CpswAppUtils_print(" rxOversizedFrames = %llu\n", hostPortStats->rxOversizedFrames);
        CpswAppUtils_print(" portMaskDrop      = %llu\n", hostPortStats->portMaskDrop);
        CpswAppUtils_print("\n");
    }
    else
    {
        CpswAppUtils_print("CpswApp_showStats() failed to get host stats: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        inArgs.portNum = CPSW_MAC_PORT_0;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
        status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                            taskObj->stateObj.coreId,
                            CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                            &prms);
        if (status == CPSW_SOK)
        {
            macPortStats = (CpswStats_MacPort_2g *)&portStats;
            CpswAppUtils_print("\n Port 1 Statistics\n");
            CpswAppUtils_print("-----------------------------------------\n");
            CpswAppUtils_print("  rxOversizedFrames = %llu\n", macPortStats->rxOversizedFrames);
            CpswAppUtils_print(" portMaskDrop       = %llu\n", macPortStats->portMaskDrop);
            CpswAppUtils_print("\n");
        }
        else
        {
            CpswAppUtils_print("CpswApp_showStats() failed to get MAC stats: %d\n",
                               status);
        }
    }
}

static int32_t CpswTestAgingMaxPktLength_setAleEntry(CpswTestTaskObj *taskObj,
                                                     uint8_t macAddr[],
                                                     uint32_t portNum)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], macAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = portNum;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = true;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(taskObj->stateObj.hCpsw, taskObj->stateObj.coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setAleEntry() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    return status;
}

static int32_t CpswTestAgingMaxPktLength_lookupAleEntry(CpswTestTaskObj *taskObj,
                                                        uint8_t
                                                        macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_MacAddrInfo macAddrInfo;
    CpswAle_GetUcastEntryOutArgs getUcastInArgs;

    memcpy(&macAddrInfo.addr[0], macAddr, sizeof(macAddrInfo.addr));
    macAddrInfo.vlanId = 0U;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macAddrInfo, &getUcastInArgs);

    status = Cpsw_ioctl(taskObj->stateObj.hCpsw, taskObj->stateObj.coreId, CPSW_ALE_IOCTL_LOOKUP_UNICAST,
                        &prms);
    return status;
}

static int32_t CpswTestAgingMaxPktLength_validateAging(CpswTestTaskObj *taskObj)
{
    int32_t status;

    status = CpswTestAgingMaxPktLength_setAleEntry(taskObj, &testAddr[0U], 1);

    if (status == CPSW_SOK)
    {
        CpswUtApp_wait(2000);
        status = CpswTestAgingMaxPktLength_lookupAleEntry(taskObj,
                                                          &testAddr[0U]);
        if (status != CPSW_SOK)
        {
            /* If lookup failed, it indcates entry was aged out successfully */
            CpswAppUtils_print("CpswApp_lookupAleEntry() Test Passed: Entry got aged \n");
            status = CPSW_SOK;
        }
        else
        {
            /* If lookup succeeded, it indcates entry was not aged out
             *successfully */
            CpswAppUtils_print("CpswApp_lookupAleEntry() Test Failed: Entry still exists in the ALE table \n");
            status = CPSW_EFAIL;
        }
    }

    return status;
}

static int32_t CpswTestAgingMaxPktLength_pktRxTx(CpswTestTaskObj *taskObj,
                                                 uint32_t txCfgIdx,
                                                 uint32_t rxCfgIdx)
{
    CpswDma_PktInfoQ txSubmitQ;
    CpswDma_PktInfo *pktInfo;
    EthFrame *frame;
    uint32_t txRetrievePktCnt;
    uint32_t rxReadyCnt;
    uint32_t pktCnt, loopRxPktCnt, loopTxPktCnt;
    int32_t status               = CPSW_SOK;
    CpswTestStateObj *stateObj   = &taskObj->stateObj;
    CpswTestRxFlowObj *rxFlowObj = &stateObj->rxFlowObj[rxCfgIdx];
    CpswTestTxChObj *txFlowObj   = &stateObj->txChObj[txCfgIdx];

    /* Broadcast address */
    uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] = {0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU};

    pktCnt = 0U;
    while (pktCnt < stateObj->txChCfgInfo[txCfgIdx]->pktSendCount)
    {
        loopRxPktCnt = loopTxPktCnt = 0U;
        /* Transmit a single packet */
        CpswUtils_initQ(&txSubmitQ);

        /* Dequeue one free TX Eth packet */
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFlowObj->txFreePktInfoQ);

        while (NULL != pktInfo)
        {
            pktCnt++;
            /* Fill the TX Eth frame with test content */
            frame = (EthFrame *)pktInfo->bufPtr;
            memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
            memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U],
                   CPSW_MAC_ADDR_LEN);
            frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);
            memset(&frame->payload[0U], (uint8_t)(0xA5 + pktCnt),
                   stateObj->txChCfgInfo[txCfgIdx]->maxTxPktLen);

            /* send bigger packet so host port truncates it */
            pktInfo->userBufLen = gRxHostMtu + 100U;
            pktInfo->appPriv    = taskObj;
            /* Enqueue the packet for later transmission */
            CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

            if (pktCnt >= stateObj->txChCfgInfo[txCfgIdx]->pktSendCount)
            {
                break;
            }

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFlowObj->txFreePktInfoQ);
        }

        loopTxPktCnt = CpswUtils_getQCount(&txSubmitQ);
        while (0U != CpswUtils_getQCount(&txSubmitQ))
        {
            uint32_t txCnt = CpswUtils_getQCount(&txSubmitQ);
            status = CpswAppUtils_submitTxPackets(txFlowObj->hTxCh, &txSubmitQ);
            /* Retrieve TX free packets */
            if (status == CPSW_SOK)
            {
                txCnt            = txCnt - CpswUtils_getQCount(&txSubmitQ);
                txRetrievePktCnt = 0U;
                while (txRetrievePktCnt != txCnt)
                {
                    /* TODO this is not failure as HW is busy sending packets, we need
                     * to wait and again call retrieve packets */
                    CpswUtApp_wait(1);
                    txRetrievePktCnt += CpswTestCommon_retrieveFreeTxPkts(taskObj,
                                                                          txCfgIdx);
#if DEBUG
                    CpswAppUtils_print("Failed to retrieve consumed transmit packets: %d\n", status);
#endif
                }
            }
            else
            {
                break;
            }
        }

        /* Wait for packet reception */
        do
        {
            /* Get the packets received so far */
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, rxCfgIdx);
            if (rxReadyCnt > 0U)
            {
                /* Consume the received packets and release them */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxFlowObj->rxReadyQ);

                while (NULL != pktInfo)
                {
                    /* Consume the packet by just printing its content */
                    frame = (EthFrame *)pktInfo->bufPtr;
#ifdef ENABLE_PRINTFRAME
                    CpswAppUtils_printFrame(frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif
                    /* Release the received packet */
                    CpswUtils_enQ(&rxFlowObj->rxFreeQ, &pktInfo->node);
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxFlowObj->rxReadyQ);
                }

                /*Submit now processed buffers */
                if (status == CPSW_SOK)
                {
                    CpswAppUtils_submitRxPackets(rxFlowObj->hRxFlow, &rxFlowObj->rxFreeQ);
                }
            }

            loopRxPktCnt += rxReadyCnt;
        }
        while (loopRxPktCnt < loopTxPktCnt);

        CpswAppUtils_print("Transmitted %d & Received %d packets\n", loopTxPktCnt, loopRxPktCnt);
    }

    return status;
}

int32_t CpswTestAgingMaxPktLength_Run(CpswTestTaskObj *taskObj)
{
    int32_t status;

    CpswTestCommon_waitForPortLink(taskObj);

    status = CpswTestAgingMaxPktLength_validateAging(taskObj);
    if (status == CPSW_SOK)
    {
        CpswAppUtils_assert(taskObj->taskCfg->numTxCh == 1);
        CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);

        status = CpswTestAgingMaxPktLength_pktRxTx(taskObj, 0, 0);
        if (status == CPSW_SOK)
        {
            /* Print CPSW statistics of all ports */
            CpswTestAgingMaxPktLength_showStats(taskObj);
        }
    }

    return status;
}
