/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_autolearn_vlan.c
 *
 * \brief    This file contains the cpsw_test_autolearn_vlan  implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_autolearn_vlan.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */
/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

static void CpswUtApp_setOUIEntry(CpswTestTaskObj *taskObj,
                                  uint8_t macAddr[]);

static void CpswUtApp_setIpv4Entry(CpswTestTaskObj *taskObj,
                                   uint8_t ipv4Addr[]);

static void CpswUtApp_setIpv6Entry(CpswTestTaskObj *taskObj,
                                   uint8_t ipv6Addr[]);

static void CpswUtApp_setEtherTypeEntry(CpswTestTaskObj *taskObj,
                                        uint32_t etherType);

static void CpswUtApp_setAleMulticastEntry(CpswTestTaskObj *taskObj,
                                           uint8_t macAddr[]);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/* Test ALE Entry Multicast Address */
static uint8_t testMCastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x01, 0x80, 0xC2, 0x00, 0x00, 0x02
};

static uint8_t testOuiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
{
    0xab, 0xcd, 0xef
};

static uint8_t testIpv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
{
    172, 24, 190, 113
};

static uint8_t testIpv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
{
    0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70, 0x80, 0x90, 0x00, 0x01, 0x02,
    0x03, 0x04, 0x05, 0x06
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
void CpswTestAutoLearnVlan_setVLANentry(CpswTestTaskObj *taskObj,
                                        uint32_t vlanId,
                                        uint32_t vlanMemberMask)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_VlanEntryInfo vlanEntry;
    CpswAle_AddEntryOutArgs vlanOutInfo;
    CpswAle_GetVlanEntryOutArgs getVlanOutArgs;
    CpswAle_VlanIdInfo getVlanInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    vlanEntry.disallowIPFragmentation  = FALSE;
    vlanEntry.forceUntaggedEgressMask  = 0x0;
    vlanEntry.limitIPNxtHdr            = FALSE;
    vlanEntry.noLearnMask              = 0x0;
    vlanEntry.regMcastFloodMask        = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    vlanEntry.unregMcastFloodMask      = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    vlanEntry.vidIngressCheck          = FALSE;
    vlanEntry.vlanMemberList           = vlanMemberMask;
    vlanEntry.vlanIdInfo.vlanId        = vlanId;
    vlanEntry.vlanIdInfo.outerVlanFlag = FALSE;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &vlanEntry, &vlanOutInfo);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setAleEntry() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    getVlanInArgs.vlanId        = vlanId;
    getVlanInArgs.outerVlanFlag = FALSE;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getVlanInArgs, &getVlanOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_LOOKUP_VLAN,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestAutoLearnVlan_setVLANentry() failed CPSW_ALE_IOCTL_LOOKUP_VLAN: %d\n",
                           status);
    }
    else
    {
        CpswAppUtils_assert(vlanOutInfo.aleEntryIndex == getVlanOutArgs.aleEntryIndex);
        CpswAppUtils_assert(vlanEntry.disallowIPFragmentation == getVlanOutArgs.disallowIPFragmentation);
        CpswAppUtils_assert(vlanEntry.forceUntaggedEgressMask == getVlanOutArgs.forceUntaggedEgressMask);
        CpswAppUtils_assert(vlanEntry.limitIPNxtHdr == getVlanOutArgs.limitIPNxtHdr);
        CpswAppUtils_assert(vlanEntry.noLearnMask == getVlanOutArgs.noLearnMask);
        CpswAppUtils_assert(vlanEntry.regMcastFloodMask == getVlanOutArgs.regMcastFloodMask);
        CpswAppUtils_assert(vlanEntry.unregMcastFloodMask == getVlanOutArgs.unregMcastFloodMask);
        CpswAppUtils_assert(vlanEntry.vidIngressCheck == getVlanOutArgs.vidIngressCheck);
        CpswAppUtils_assert(vlanEntry.vlanMemberList == getVlanOutArgs.vlanMemberList);
    }
}

static int32_t CpswTestAutoLearnVlan_pktRxTx(CpswTestTaskObj *taskObj,
                                             uint32_t txCfgIndex,
                                             uint32_t rxCfgIndex)
{
    CpswDma_PktInfoQ txSubmitQ;
    CpswDma_PktInfo *pktInfo;
    uint32_t txRetrievePktCnt, loopTxPktCnt, totalTxCnt;
    uint32_t rxReadyCnt;
    uint32_t loopCnt, pktCnt, loopRxPktCnt, zeroRxCount;
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    bool testComplete;
    uint32_t rxPktNum;

    totalTxCnt = 0U;
    for (loopCnt = 0U;
         loopCnt < stateObj->txChCfgInfo[txCfgIndex]->pktSendLoopCount;
         loopCnt++)
    {
        pktCnt = 0U;

        loopRxPktCnt = loopTxPktCnt = 0U;
        /* Transmit a single packet */
        CpswUtils_initQ(&txSubmitQ);

        /* Dequeue one free TX Eth packet */
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);

        while (NULL != pktInfo)
        {
            pktCnt++;
            /* Fill the TX Eth frame with test content */
            CpswTestCommon_setTxPktInfo(taskObj, txCfgIndex, pktCnt,
                                        pktInfo, &testComplete);

            /* Enqueue the packet for later transmission */
            CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

            if (pktCnt >= stateObj->txChCfgInfo[txCfgIndex]->pktSendCount)
            {
                break;
            }

            /* Dequeue one free TX Eth packet */
            pktInfo =
                (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);
        }

        loopTxPktCnt = CpswUtils_getQCount(&txSubmitQ);
        while (0U != CpswUtils_getQCount(&txSubmitQ))
        {
            uint32_t txCnt = CpswUtils_getQCount(&txSubmitQ);
            status = CpswAppUtils_submitTxPackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                                  &txSubmitQ);
            /* Retrieve TX free packets */
            if (status == CPSW_SOK)
            {
                txCnt            = txCnt - CpswUtils_getQCount(&txSubmitQ);
                txRetrievePktCnt = 0U;
                while (txRetrievePktCnt != txCnt)
                {
                    /* TODO this is not failure as HW is busy sending packets, we need
                     * to wait and again call retrieve packets */
                    CpswUtApp_wait(1);
                    txRetrievePktCnt += CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
#if DEBUG
                    CpswAppUtils_print("Failed to retrieve consumed transmit packets: %d\n",
                                       status);
#endif
                }
            }
            else
            {
                break;
            }
        }

        CpswTestAutoLearnVlan_setVLANentry(taskObj, 0x10, 3);
        CpswUtApp_setAleMulticastEntry(taskObj, &testMCastAddr[0]);
        CpswUtApp_setEtherTypeEntry(taskObj, 0x800);
        CpswUtApp_setIpv4Entry(taskObj, testIpv4Addr);
        CpswUtApp_setIpv6Entry(taskObj, testIpv6Addr);
        CpswUtApp_setOUIEntry(taskObj, testOuiAddr);

        rxPktNum = 0;
        /* Wait for packet reception */
        do
        {
            /* Get the packets received so far */
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, rxCfgIndex);
            if (rxReadyCnt > 0U)
            {
                /* Consume the received packets and release them */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
                while (NULL != pktInfo)
                {
                    CpswTestCommon_processRxPkt(taskObj,
                                                rxCfgIndex,
                                                rxPktNum,
                                                pktInfo,
                                                &testComplete);
                    rxPktNum++;
                    /* Release the received packet */
                    CpswUtils_enQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ,
                                  &pktInfo->node);
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
                }

                /*Submit now processed buffers */
                if (status == CPSW_SOK)
                {
                    CpswAppUtils_submitRxPackets(stateObj->rxFlowObj[rxCfgIndex].hRxFlow,
                                                 &stateObj->rxFlowObj[rxCfgIndex].rxFreeQ);
                }
            }

            loopRxPktCnt += rxReadyCnt;
        }
        while (loopRxPktCnt < 500);

        CpswTestAutoLearnVlan_setVLANentry(taskObj, 0x10, 2);
        CpswAppUtils_print("Transmitted & Received 500 packets & VLAN Entry modified\n");
        loopRxPktCnt = 0;
        zeroRxCount  = 0;

        /* Wait for packet reception */
        do
        {
            /* Get the packets received so far */
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, rxCfgIndex);
            if (rxReadyCnt > 0U)
            {
                /* Consume the received packets and release them */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
                while (NULL != pktInfo)
                {
                    CpswTestCommon_processRxPkt(taskObj, rxCfgIndex,  rxPktNum,
                                                pktInfo, &testComplete);
                    rxPktNum++;
                    /* Release the received packet */
                    CpswUtils_enQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ,
                                  &pktInfo->node);
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
                }

                /*Submit now processed buffers */
                if (status == CPSW_SOK)
                {
                    CpswAppUtils_submitRxPackets(stateObj->rxFlowObj[rxCfgIndex].hRxFlow,
                                                 &stateObj->rxFlowObj[rxCfgIndex].
                                                 rxFreeQ);
                }
            }
            else
            {
                zeroRxCount++;
            }

            loopRxPktCnt += rxReadyCnt;
        }
        while (zeroRxCount < 1000);

        /* Reset ready count */
        totalTxCnt += loopTxPktCnt;
    }

    CpswAppUtils_print("VLAN Test Succes \n ");
    CPSW_IOCTL_SET_NO_ARGS(&prms);
    Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
               &prms);

    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("Transmitted & Received %d packets\n", totalTxCnt);
    }
    else
    {
        CpswAppUtils_print("Failed to transmit/receive packets: %d, transmitted: %d \n",
                           stateObj->txChCfgInfo[txCfgIndex]->pktSendLoopCount, totalTxCnt);
    }

    return status;
}

int32_t CpswTestAutoLearnVlan_Run(CpswTestTaskObj *taskObj)
{
    int32_t status;

    CpswTestCommon_waitForPortLink(taskObj);

    CpswAppUtils_assert(taskObj->taskCfg->numTxCh == 1);
    CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);

    status = CpswTestAutoLearnVlan_pktRxTx(taskObj, 0, 0);

    return status;
}

void CpswTestAutoLearnVlan_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                             Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestAutoLearnVlan_setOpenPrms(CpswTestTaskObj *taskObj,
                                       Cpsw_Config *pCpswCfg,
                                       CpswOsal_Prms *pOsalPrms,
                                       CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->aleConfig.modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE |
         CPSW_ALE_CONFIG_MASK_UNKNOWN_UNICAST_FLOOD2HOST);
    pCpswCfg->aleConfig.agingConfig.enableAutoAging           = FALSE;
    pCpswCfg->aleConfig.portCfg[0].learningCfg.noLearn        = FALSE;
    pCpswCfg->aleConfig.portCfg[1].learningCfg.noLearn        = FALSE;
    pCpswCfg->aleConfig.portCfg[0].learningCfg.noLearn        = TRUE;
    pCpswCfg->aleConfig.portCfg[1].learningCfg.noLearn        = TRUE;
    pCpswCfg->aleConfig.portCfg[1].vlanCfg.dropUntagged       = TRUE;
    pCpswCfg->aleConfig.portCfg[0].vlanCfg.dropUntagged       = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask = 0x0;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask   = 0x0;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask  = 0x0;
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode           = TRUE;
    pCpswCfg->aleConfig.vlanConfig.autoLearnWithVLAN          = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.enableVid0Mode               = FALSE;
}

static void CpswUtApp_setOUIEntry(CpswTestTaskObj *taskObj,
                                  uint8_t macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setOuiOutArgs;
    CpswAle_OuiEntryInfo setOuiInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setOuiInArgs.ouiAddr[0], macAddr, sizeof(setOuiInArgs.ouiAddr));

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setOuiInArgs, &setOuiOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_OUI,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setOUIEntry() failed : %d\n", status);
    }
}

static void CpswUtApp_setIpv4Entry(CpswTestTaskObj *taskObj,
                                   uint8_t ipv4Addr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setIpv4OutArgs;
    CpswAle_Ipv4EntryInfo setIpv4InArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setIpv4InArgs.ipv4Addr[0], ipv4Addr, sizeof(setIpv4InArgs.ipv4Addr));
    setIpv4InArgs.numLSBIgnoreBits = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setIpv4InArgs, &setIpv4OutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_IPV4ADDR,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setIpv4Entry() failed : %d\n", status);
    }
}

static void CpswUtApp_setIpv6Entry(CpswTestTaskObj *taskObj,
                                   uint8_t ipv6Addr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setIpv6OutArgs;
    CpswAle_Ipv6EntryInfo setIpv6InArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setIpv6InArgs.ipv6Addr[0], ipv6Addr, sizeof(setIpv6InArgs.ipv6Addr));
    setIpv6InArgs.numLSBIgnoreBits = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setIpv6InArgs, &setIpv6OutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_IPV6ADDR,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setIpv6Entry() failed : %d\n", status);
    }
}

static void CpswUtApp_setEtherTypeEntry(CpswTestTaskObj *taskObj,
                                        uint32_t etherType)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setEtherTypeOutArgs;
    CpswAle_EtherTypeEntryInfo setEtherTypeInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    setEtherTypeInArgs.etherType = etherType;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setEtherTypeInArgs, &setEtherTypeOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_ETHERTYPE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setEtherTypeEntry() failed : %d\n", status);
    }
}

static void CpswUtApp_setAleMulticastEntry(CpswTestTaskObj *taskObj,
                                           uint8_t macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setMcastOutArgs;
    CpswAle_SetMcastEntryInArgs setMcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setMcastInArgs.addr.addr[0], macAddr,
           sizeof(setMcastInArgs.addr.addr));
    setMcastInArgs.addr.vlanId = 0U;

    setMcastInArgs.info.superFlag  = false;
    setMcastInArgs.info.fwdState   = CPSW_ALE_FWDSTLVL_FWD_LRN;
    setMcastInArgs.info.portMask   = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    setMcastInArgs.info.numIgnBits = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_MULTICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setAleMulticastEntry() failed CPSW_ALE_IOCTL_ADD_MULTICAST: %d\n",
                           status);
    }
}

/* end of file */
