/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_base.c
 *
 * \brief    This file contains the common code test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Mailbox.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Semaphore.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apprm.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_board.h"
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_TEST_COMMON_TX_TASK_STACK_SIZE (10U * 1024U)
#define CPSW_TEST_COMMON_TX_TASK_PRI (7)
#define CPSW_TEST_COMMON_TX_TEST_STATUS_COUNT  (4)

#define CPSW_TEST_COMMON_RX_TASK_STACK_SIZE (10U * 1024U)
#define CPSW_TEST_COMMON_RX_TASK_PRI (8)
#define CPSW_TEST_COMMON_RX_TEST_STATUS_COUNT  (4)

#define CPSW_TEST_COMMON_MAIN_TASK_PRI (9)
#define CPSW_TEST_COMMON_PHYLINK_STATE_POLL_INTERVAL_MS   (1U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */
typedef struct CpswTestCommon_TxTestStatus_s
{
    uint32_t testStatus;
    uint32_t stackUsed;
} CpswTestCommon_TxTestStatus_t;

typedef struct CpswTestCommon_RxTestStatus_s
{
    uint32_t testStatus;
    uint32_t stackUsed;
} CpswTestCommon_RxTestStatus_t;

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

static void CpswTestCommon_InitUdmaDrv(CpswTestTaskObj *taskObj);

static void CpswTestCommon_DeInitUdmaDrv(CpswTestTaskObj *taskObj);

static void     CpswTestCommon_showStats(CpswTestTaskObj *taskObj);

static int32_t CpswTestCommon_openCpsw(CpswTestTaskObj *taskObj,
                                       bool *pDisableHostAddrSecureFlag);

static void     CpswTestCommon_closeCpsw(CpswTestTaskObj *taskObj);

static int32_t CpswTestCommon_openDma(CpswTestTaskObj *taskObj,
                                      bool disableHostAddrSecureFlag);

static void     CpswTestCommon_closeDma(CpswTestTaskObj *taskObj);

static int32_t CpswTestCommon_setRxflowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms,
                                            CpswTestTaskObj *taskObj,
                                            uint32_t rxFlowCfgIdx);
static void CpswTestCommon_recvTask(UArg a0, UArg a1);
static void  CpswTestCommon_createTxSem(CpswTestTaskObj *taskObj,
                                        uint32_t txCfgIndex);

static void CpswTestCommon_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                           CpswDma_PktInfoQ *pSubmitQ);

static int32_t CpswTestCommon_submitTxPackets(CpswDma_TxChHandle hTxChannel,
                                              CpswDma_PktInfoQ *pSubmitQ);

static uint32_t CpswTestCommon_waitForXmit(CpswTestTaskObj *taskObj,
                                           uint32_t txCfgIndex);

static void CpswTestCommon_deleteRxSem(CpswTestTaskObj *taskObj,
                                       uint32_t rxCfgIndex);

static void CpswTestCommon_deleteTxSem(CpswTestTaskObj *taskObj,
                                       uint32_t txCfgIndex);

static void CpswTestCommon_setPortLinkPrms(Cpsw_OpenPortLinkInArgs *linkArgs,
                                           Cpsw_MacPort portNum,
                                           CpswTestTaskObj *taskObj);

static int32_t CpswTestCommon_setTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms,
                                   CpswTestTaskObj *taskObj,
                                   uint32_t txPSILThreadId,
                                   uint32_t txChCfgIdx);

static int32_t CpswTestCommon_addDefaultHostPortEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
static uint8_t gDefaultHostMacAddr[] = {0x02, 0x01, 0x02, 0x03, 0x04, 0x5};
/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

/* PDK-4761 - The CpswAppUtils_enableClocks / CpswAppUtils_disableClocks sequence causes
 * core to lock up on accessing CPSW register after clock enable when executed
 * mutliple iterations.
 * To avoid CPU lockup disabling clock disable/enable during each test execution
 * TODO: Fix the CpswAppUtils_enableClocks/CpswAppUtils_disableClocks and remove
 * the workaround
 */
bool gDisableClockDisable = true;

int32_t CpswTestCommon_Init(CpswTestTaskObj *taskObj)
{
    int32_t status;
    CpswTestStateObj *stateObj     = &taskObj->stateObj;
    static bool clockInitDone      = false;
    bool disableHostAddrSecureFlag = false;

    /* App should open UDMA first as UDMA handle is needed to initialize
     * CPSW RX channel */
    CpswTestCommon_InitUdmaDrv(taskObj);

    if ((gDisableClockDisable == false) || (!clockInitDone))
    {
        CpswAppUtils_enableClocks(taskObj->taskCfg->cpswType);
        clockInitDone = true;
    }

    /* Open the CPSW */
    status = CpswTestCommon_openCpsw(taskObj, &disableHostAddrSecureFlag);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to open CPSW: %d\n", status);
    }

    /* Open legacy DMA driver */
    if (status == CPSW_SOK)
    {
        status = CpswTestCommon_openDma(taskObj, disableHostAddrSecureFlag);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to open legacy DMA: %d\n", status);
        }
    }

    /* Enable host port */
    if (status == CPSW_SOK)
    {
        Cpsw_IoctlPrms prms;
        CpswAle_SetPortStateInArgs setPortStateInArgs;

        setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
        prms.outArgs = NULL;
        status       = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_PORT_STATE, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestCommon_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
        }

        if (status == CPSW_SOK)
        {
            CPSW_IOCTL_SET_NO_ARGS(&prms);
            status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_HOSTPORT_IOCTL_ENABLE, &prms);
            if (status != CPSW_SOK)
            {
                CpswAppUtils_print("Failed to enable host port: %d\n", status);
            }
        }
    }

    /* Show alive PHYs */
    if (status == CPSW_SOK)
    {
        Cpsw_IoctlPrms prms;
        bool alive;
        int8_t i;

        for (i = 0U; i < CPSW_MDIO_MAX_PHY_CNT; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                                CPSW_MDIO_IOCTL_GET_ALIVE_STATUS,
                                &prms);
            if (status == CPSW_SOK)
            {
                if (alive == true)
                {
                    CpswAppUtils_print("PHY %d is alive\n", i);
                }
            }
            else
            {
                CpswAppUtils_print("Failed to get PHY %d alive status: %d\n", i,
                                   status);
            }
        }
    }

    if (status == CPSW_SOK)
    {
        status = CpswTestCommon_addDefaultHostPortEntry(taskObj);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to add default host port address: %d\n", status);
        }
    }

    return status;
}

void  CpswTestCommon_getMacPortList(uint32_t macPortBitMask,
                                    CpswTestMacPortList_t *macPortList)
{
    Cpsw_MacPort portIndex;
    uint32_t numMacPorts = 0U;

    for (portIndex = CPSW_MAC_PORT_FIRST; portIndex <= CPSW_MAC_PORT_LAST; portIndex++)
    {
        if (CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(portIndex)) & macPortBitMask)
        {
            CpswAppUtils_assert(numMacPorts < CPSW_UTILS_ARRAYSIZE(macPortList->macPortList));
            macPortList->macPortList[numMacPorts] = portIndex;
            numMacPorts++;
        }
    }

    macPortList->numMacPorts = numMacPorts;
}

static void CpswTestCommon_waitForPhyLinkUp(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t phyPortMask       = taskObj->taskCfg->macPortMask;

    phyPortMask &= ~(stateObj->noPhyPortMask);
    CpswAppUtils_print("Waiting for PHY link on enabled ports mask:0x%X \n", phyPortMask);
    CpswAppUtils_assert(stateObj->phyEvent != NULL);

    /* Wait on event only if phyPortMask is non-zero */
    if (phyPortMask)
    {
        Event_pend(stateObj->phyEvent, phyPortMask, 0, BIOS_WAIT_FOREVER);
    }
}

void CpswTestCommon_waitForPortLink(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswTestCommon_waitForPhyLinkUp(taskObj);

    CpswAppUtils_assert(stateObj->portEvent != NULL);
    CpswAppUtils_print("Waiting for PORT link on enabled ports \n");
    Event_pend(stateObj->portEvent, taskObj->taskCfg->macPortMask, 0, BIOS_WAIT_FOREVER);
}

int32_t CpswTestCommon_Run(CpswTestTaskObj *taskObj)
{
    int32_t status;

    CpswTestCommon_waitForPortLink(taskObj);
    status = CpswTestCommon_pktTxRx(taskObj);

    return status;
}

int32_t CpswTestCommon_DeInit(CpswTestTaskObj *taskObj)
{
    int32_t status;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_IoctlPrms prms;
    uint32_t i;

    /* Print CPSW statistics of all ports */
    CpswTestCommon_showStats(taskObj);

    /* DeInit Portion */
    /* Disable host port */
    CPSW_IOCTL_SET_NO_ARGS(&prms);
    status = Cpsw_ioctl(stateObj->hCpsw,
                        stateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_DISABLE,
                        &prms);
    /* Host port disable should not fail */
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Close legacy DMA driver */
    CpswTestCommon_closeDma(taskObj);

    /* Close CPSW */
    CpswTestCommon_closeCpsw(taskObj);

    /* Delete the Tx and Rx semaphore */
    for (i = 0; i < taskObj->taskCfg->numTxCh; i++)
    {
        CpswTestCommon_deleteTxSem(taskObj, i);
    }

    /* Delete the Tx and Rx semaphore */
    for (i = 0; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswTestCommon_deleteRxSem(taskObj, i);
    }

    CpswTestCommon_DeInitUdmaDrv(taskObj);

    if (!gDisableClockDisable)
    {
        CpswAppUtils_disableClocks(taskObj->taskCfg->cpswType);
    }

    return CPSW_SOK;
}

void CpswTestCommon_initAlePrms(CpswAle_Config *aleConfig,
                                CpswTestTaskObj *taskObj)
{
    aleConfig->modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging           = TRUE;
    aleConfig->agingConfig.agingPeriodInMs           = 1000;
    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
}

void CpswTestCommon_initDmaPrms(CpswDma_Config *dmaConfig,
                                CpswTestTaskObj *taskObj)
{
    dmaConfig->rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;

    dmaConfig->hUdmaDrv = taskObj->stateObj.hUdmaDrv;
}

static void CpswTestCommon_initHostPortPrms(CpswHostPort_Config *hostPrms)
{
    hostPrms->removeCrc      = true;
    hostPrms->padShortPacket = true;
    hostPrms->passCrcErrors  = true;
}

static void CpswTestCommon_initCpswVlanPrms(Cpsw_VlanConfig *vlanPrms)
{
    vlanPrms->vlanAware = false;
}

static void CpswTestCommon_initCptsPrms(CpswCpts_Config *cptsPrms)
{
    cptsPrms->hostRxTsEnable = false;
}

static Cpsw_MacPort CpswTestCommon_mapPhyAddr2MacPort(CpswTestTaskObj *taskObj,
                                                      uint32_t matchPhyAddr)
{
    CpswTestMacPortList_t enabledPorts;
    uint32_t i;

    /* Initialized to avoid warning from GCC about uninitialized variable.
     * However if matching port is not found it will assert and abort execution
     * rather than return MAC_PORT_0
     */
    Cpsw_MacPort matchPort = CPSW_MAC_PORT_0;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    for (i = 0; i < enabledPorts.numMacPorts; i++)
    {
        uint32_t portPhyAddr = CpswTestBoard_getPhyAddr(taskObj->taskCfg->cpswType,
                                                        enabledPorts.macPortList[i]);
        if (portPhyAddr == matchPhyAddr)
        {
            break;
        }
    }

    if (i < enabledPorts.numMacPorts)
    {
        matchPort = enabledPorts.macPortList[i];
    }
    else
    {
        CpswAppUtils_assert(false);
    }

    return matchPort;
}

static uint32_t CpswTestCommon_getEnablePhyMask(CpswTestTaskObj *taskObj)
{
    CpswTestMacPortList_t enabledPorts;
    uint32_t i;
    uint32_t enabledPhyMask;

    enabledPhyMask = 0;
    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    for (i = 0; i < enabledPorts.numMacPorts; i++)
    {
        uint32_t portPhyAddr = CpswTestBoard_getPhyAddr(taskObj->taskCfg->cpswType,
                                                        enabledPorts.macPortList[i]);
        enabledPhyMask |= CPSW_MDIO_PHY_ADDR_MASK(portPhyAddr);
    }

    return enabledPhyMask;
}

static void CpswTestCommon_mdioLinkStatusChangeCb(Cpsw_MdioLinkStateChangeInfo *info,
                                                  void *appArg)
{
    CpswTestTaskObj *taskObj = (CpswTestTaskObj *)appArg;
    Cpsw_MacPort linkStateChangePort;

    linkStateChangePort = CpswTestCommon_mapPhyAddr2MacPort(taskObj, info->phyAddr);

    if ((info->linkChanged) && (info->isLinked))
    {
        CpswAppUtils_assert(taskObj->stateObj.phyEvent != NULL);
        Event_post(taskObj->stateObj.phyEvent, CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(linkStateChangePort)));
    }
}

static void CpswTestCommon_portLinkStatusChangeCb(Cpsw_MacPort portNum,
                                                  bool isLinkUp,
                                                  void *appArg)
{
    CpswTestTaskObj *taskObj = (CpswTestTaskObj *)appArg;

    CpswAppUtils_print("MAC Port %u: link %s\n", portNum, isLinkUp ? "up" : "down");

    if (isLinkUp)
    {
        CpswAppUtils_assert(taskObj->stateObj.portEvent != NULL);
        Event_post(taskObj->stateObj.portEvent, CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(portNum)));
    }
}

static void CpswTestCommon_initCpswLinkCbPrms(Cpsw_Config *cpswCfg,
                                              CpswTestTaskObj *taskObj)
{
    cpswCfg->mdioLinkStateChangeCb     = &CpswTestCommon_mdioLinkStatusChangeCb;
    cpswCfg->mdioLinkStateChangeCbArg  = taskObj;
    cpswCfg->portLinkStatusChangeCb    = &CpswTestCommon_portLinkStatusChangeCb;
    cpswCfg->portLinkStatusChangeCbArg = taskObj;
}

static void CpswTestCommon_initMdioPrms(CpswMdio_Config *mdioCfg,
                                        CpswTestTaskObj *taskObj)
{
    if (taskObj->taskCfg->cpswType == CPSW_2G)
    {
        mdioCfg->mode = CPSW_MDIO_MODE_NORMAL;
    }
    else if ((taskObj->taskCfg->cpswType == CPSW_9G)||
             (taskObj->taskCfg->cpswType == CPSW_5G))
    {
        mdioCfg->mode = CPSW_MDIO_MODE_STATE_CHANGE;
    }
    else
    {
        CpswAppUtils_assert(false);
    }

    mdioCfg->pollEnableMask = CpswTestCommon_getEnablePhyMask(taskObj);
}

static void CpswTestCommon_setPortLinkPrms(Cpsw_OpenPortLinkInArgs *linkArgs,
                                           Cpsw_MacPort portNum,
                                           CpswTestTaskObj *taskObj)
{
    CpswMacPort_Config *macConfig      = &linkArgs->macConfig;
    CpswMacPort_LinkConfig *linkConfig = &linkArgs->linkConfig;
    CpswPhy_Config *phyConfig          = &linkArgs->phyConfig;
    CpswMacPort_Interface *interface   = &linkArgs->interface;

    linkArgs->portNum    = portNum;
    interface->layerType = CPSW_MAC_LAYER_GMII;

    Cpsw_initMacPortParams(macConfig);
    macConfig->enableLoopback = false;

    Cpsw_initPhyParams(phyConfig);
    CpswTestBoard_setPhyConfig(taskObj, portNum, macConfig, interface, phyConfig);
    linkConfig->speed       = CPSW_SPEED_AUTO;
    linkConfig->duplexity   = CPSW_DUPLEX_AUTO;
}

static int32_t CpswTestCommon_openCpsw(CpswTestTaskObj *taskObj,
                                       bool *pDisableHostAddrSecureFlag)
{
    Cpsw_IoctlPrms prms;
    int32_t status = CPSW_SOK;
    uint32_t index;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_Config cpswCfg;
    CpswUtils_Prms utilsPrms;
    CpswOsal_Prms osalPrms;
    Cpsw_OpenPortLinkInArgs linkArgs;

    *pDisableHostAddrSecureFlag = FALSE;
    /* Set configuration parameters */
    Cpsw_initParams(&cpswCfg);
    CpswTestCommon_initCpswLinkCbPrms(&cpswCfg, taskObj);
    CpswTestCommon_initCpswVlanPrms(&cpswCfg.vlanConfig);

    CpswAppUtils_initResourceConfig(taskObj->taskCfg->cpswType, taskObj->stateObj.coreId, &cpswCfg.resourceConfig);

    CpswTestCommon_initMdioPrms(&cpswCfg.mdioConfig, taskObj);
    CpswTestCommon_initHostPortPrms(&cpswCfg.hostPortConfig);
    CpswTestCommon_initAlePrms(&cpswCfg.aleConfig, taskObj);
    CpswTestCommon_initDmaPrms(&cpswCfg.dmaConfig, taskObj);
    CpswTestCommon_initCptsPrms(&cpswCfg.cptsConfig);

    if (CPSW_9G == taskObj->taskCfg->cpswType)
    {
        CpswAppUtils_print("CPSW_9G Test on MAIN NAVSS\n");
    }
    if (CPSW_5G == taskObj->taskCfg->cpswType)
    {
        CpswAppUtils_print("CPSW_5G Test on MAIN NAVSS\n");
    }
    else if (CPSW_2G == taskObj->taskCfg->cpswType)
    {
        CpswAppUtils_print("CPSW_2G Test on MCU NAVSS\n");
    }

    /* Initialize CPSW driver with default OSAL, utils */
    utilsPrms.printFxn     = CpswAppUtils_print;
    utilsPrms.traceFxn     = CpswAppUtils_print;
    utilsPrms.phyToVirtFxn = &CpswAppUtils_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswAppUtils_virtToPhyFxn;

    /* Open the CPSW driver */
    if (taskObj->taskCfg->openPrmsUpdate != NULL)
    {
        taskObj->taskCfg->openPrmsUpdate(taskObj, &cpswCfg, &osalPrms,
                                         &utilsPrms);
    }

    Cpsw_initOsalPrms(&osalPrms);
    Cpsw_init(taskObj->taskCfg->cpswType, &osalPrms, &utilsPrms);
    stateObj->hCpsw = Cpsw_open(taskObj->taskCfg->cpswType, &cpswCfg);
    if (stateObj->hCpsw == NULL)
    {
        CpswAppUtils_print("CpswTestCommon_openCpsw() failed to open: %d\n",
                           status);
        status = CPSW_EFAIL;
    }

    if (status == CPSW_SOK)
    {
        /* memutils open should happen after Cpsw is opened as it uses CpswUtils_Q
         * functions */
        CpswAppMemUtils_init();
    }

    if (status == CPSW_SOK)
    {
        CpswTestMacPortList_t enabledPorts;
        CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
        for (index = 0; index < enabledPorts.numMacPorts; index++)
        {
            CpswTestCommon_setPortLinkPrms(&linkArgs,
                                           enabledPorts.macPortList[index],
                                           taskObj);
            if (taskObj->taskCfg->portLinkPrmsUpdate != NULL)
            {
                /* Override test specific port link init params */
                taskObj->taskCfg->portLinkPrmsUpdate(&linkArgs,
                                                     enabledPorts.macPortList[index]);
            }

            CPSW_IOCTL_SET_IN_ARGS(&prms, &linkArgs);
            status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_OPEN_PORT_LINK,
                                &prms);
            if (status != CPSW_SOK)
            {
                CpswAppUtils_print("CpswTestCommon_openCpsw() failed to open MAC port: %d\n", status);
                break;
            }

            if (linkArgs.macConfig.enableLoopback)
            {
                *pDisableHostAddrSecureFlag = TRUE;
            }

            if (CPSW_PHY_INVALID_PHYADDR == linkArgs.phyConfig.phyAddr)
            {
                /* If PHY is disabled for the port set the noPhyMask */
                taskObj->stateObj.noPhyPortMask |= CPSW_SET_BIT(CPSW_NORMALIZE_MACPORT(linkArgs.portNum));
            }
        }
    }

    return status;
}

static void CpswTestCommon_closeCpsw(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t index;
    int32_t status;
    CpswTestMacPortList_t enabledPorts;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    for (index = 0; index < enabledPorts.numMacPorts; index++)
    {
        Cpsw_GenericPortLinkInArgs args =
        {
            .portNum = enabledPorts.macPortList[index],
        };

        CPSW_IOCTL_SET_IN_ARGS(&prms, &args);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_CLOSE_PORT_LINK, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("close() failed to close MAC port: %d\n", status);
            /* Cannot handle port close failing */
            CpswAppUtils_assert(FALSE);
        }
    }

    Cpsw_close(stateObj->hCpsw);

    Cpsw_deinit(taskObj->taskCfg->cpswType);
}

void CpswTestCommon_showStats(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status =
        Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_STATS_IOCTL_GET_HOSTPORT_STATS,
                   &prms);
    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("\n Port 0 Statistics\n");
        CpswAppUtils_print("-----------------------------------------\n");
        switch (taskObj->taskCfg->cpswType)
        {
            case CPSW_2G:
            {
                CpswStats_HostPort_2g *st;

                st = (CpswStats_HostPort_2g *)&portStats;
                CpswAppUtils_printHostPortStats2G(st);
                break;
            }

            case CPSW_9G:
            case CPSW_5G:
            {
                CpswStats_HostPort_9g *st;

                st = (CpswStats_HostPort_9g *)&portStats;
                CpswAppUtils_printHostPortStats9G(st);
                break;
            }
        }

        CpswAppUtils_print("\n");
    }
    else
    {
        CpswAppUtils_print("CpswTestCommon_showStats() failed to get host stats: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        uint32_t index;
        CpswTestMacPortList_t enabledPorts;

        CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
        for (index = 0; index < enabledPorts.numMacPorts; index++)
        {
            inArgs.portNum = enabledPorts.macPortList[index];
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
            status =
                Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                           &prms);
            if (status == CPSW_SOK)
            {
                CpswAppUtils_print("\n External Port %d Statistics\n", CPSW_NORMALIZE_MACPORT(enabledPorts.macPortList[index]));
                CpswAppUtils_print("-----------------------------------------\n");
                switch (taskObj->taskCfg->cpswType)
                {
                    case CPSW_2G:
                    {
                        CpswStats_MacPort_2g *st;

                        st = (CpswStats_MacPort_2g *)&portStats;
                        CpswAppUtils_printMacPortStats2G(st);
                        break;
                    }

                    case CPSW_5G:
                    case CPSW_9G:
                    {
                        CpswStats_MacPort_9g *st;

                        st = (CpswStats_MacPort_9g *)&portStats;
                        CpswAppUtils_printMacPortStats9G(st);
                        break;
                    }
                }

                CpswAppUtils_print("\n");
            }
            else
            {
                CpswAppUtils_print("CpswTestCommon_showStats() failed to get MAC stats: %d\n",
                                   status);
            }
        }
    }
}

void CpswTestCommon_rxIsrFxn(void *appData)
{
    /* Post semaphore to tx handling task */
    CpswTestRxFlowObj *pRxFlowObj = (CpswTestRxFlowObj *)appData;

    Semaphore_post(pRxFlowObj->hRxCQSem);
}

void CpswTestCommon_txIsrFxn(void *appData)
{
    /* Post semaphore to tx handling task */
    CpswTestTxChObj *pTxChObj = (CpswTestTxChObj *)appData;

    Semaphore_post(pTxChObj->hTxCQSem);
}

void CpswTestCommon_initTxFreePktQ(CpswTestTaskObj *taskObj,
                                   uint32_t txChCfgIdx)
{
    CpswDma_PktInfo *pktInfo;
    uint32_t index;
    CpswTestStateObj *stateObj       = &taskObj->stateObj;
    CpswTestTxChObj *txChObj         = &stateObj->txChObj[txChCfgIdx];
    CpswTestTxChCfgInfo *txChCfgInfo = stateObj->txChCfgInfo[txChCfgIdx];
    uint32_t maxPktsPerCh;

    /* Delibrately allocate more packets than number of descriptors to
     * exercise unused queue logic when submitting packets
     */
    maxPktsPerCh = (CPSW_APPMEMUTILS_NUM_TX_PKTS * CPSW_CFG_TX_CHANNELS_NUM) / taskObj->taskCfg->numTxCh;

    /* Initialize all queues */
    CpswUtils_initQ(&txChObj->txFreePktInfoQ);

    /* Initialize TX EthPkts and queue them to txFreePktInfoQ */
    for (index = 0U; index < CpswUtils_min(txChCfgInfo->pktSendCount, maxPktsPerCh); index++)
    {
        pktInfo = CpswAppMemUtils_allocEthPktFxn(NULL, txChCfgInfo->maxTxPktLen, UDMA_CACHELINE_ALIGNMENT);
        if (NULL == pktInfo)
        {
            break;
        }

        memset(&pktInfo->node, 0U, sizeof(pktInfo->node));
        pktInfo->orgBufLen  = txChCfgInfo->maxTxPktLen;
        pktInfo->userBufLen = txChCfgInfo->maxTxPktLen;
        pktInfo->appPriv    = &txChCfgInfo;

        CpswUtils_enQ(&txChObj->txFreePktInfoQ, &pktInfo->node);
    }

    CpswAppUtils_assert(CpswUtils_getQCount(&txChObj->txFreePktInfoQ) == CpswUtils_min(txChCfgInfo->pktSendCount, maxPktsPerCh));
    CpswAppUtils_print("initQs() txFreePktInfoQ initialized with %d pkts\n", CpswUtils_getQCount(&txChObj->txFreePktInfoQ));
}

static int32_t CpswTestCommon_setTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms,
                                          CpswTestTaskObj *taskObj,
                                          uint32_t txPSILThreadId,
                                          uint32_t txChCfgIdx)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_setCommonTxChPrms(pTxChPrms);

    pTxChPrms->chNum = txPSILThreadId;

    pTxChPrms->notifyCb = &CpswTestCommon_txIsrFxn;
    pTxChPrms->hCbArg   = &stateObj->txChObj[txChCfgIdx];

    pTxChPrms->hUdmaDrv = taskObj->stateObj.hUdmaDrv;

    return status;
}

static int32_t CpswTestCommon_setRxflowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms,
                                            CpswTestTaskObj *taskObj,
                                            uint32_t rxFlowCfgIdx)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_setCommonRxFlowPrms(pRxFlowPrms);

    pRxFlowPrms->notifyCb = &CpswTestCommon_rxIsrFxn;
    pRxFlowPrms->hCbArg   = &stateObj->rxFlowObj[rxFlowCfgIdx];

    pRxFlowPrms->hUdmaDrv  = stateObj->hUdmaDrv;
    pRxFlowPrms->rxFlowMtu =
    CPSW_UTILS_ALIGN((stateObj->rxFlowCfgInfo[rxFlowCfgIdx]->rxFlowMtu
                      + sizeof(EthVlanFrameHeader)), CPSW_DMA_RXMTU_ALIGN);

    return status;
}

void CpswTestCommon_initRxReadyPktQ(CpswTestTaskObj *taskObj,
                                    uint32_t rxFlowCfgIdx)
{
    CpswDma_PktInfoQ rxReadyQ;
    int32_t status;
    uint32_t i;
    CpswDma_PktInfo *pPktInfo;
    CpswTestStateObj *stateObj           = &taskObj->stateObj;
    CpswTestRxFlowObj *rxFlowObj         = &stateObj->rxFlowObj[rxFlowCfgIdx];
    CpswTestRxFlowCfgInfo *rxFlowCfgInfo = stateObj->rxFlowCfgInfo[rxFlowCfgIdx];

    uint32_t maxPkts;

    /* Delibrately allocate more packets than number of descriptors to
     * exercise unused queue logic when submitting packets
     */
    maxPkts = (CPSW_APPMEMUTILS_NUM_RX_PKTS * CPSW_CFG_RX_FLOWS_NUM) / taskObj->taskCfg->numRxFlow;

    CpswUtils_initQ(&rxFlowObj->rxFreeQ);
    CpswUtils_initQ(&rxReadyQ);

    for (i = 0U; i < CpswUtils_min(rxFlowCfgInfo->pktRecvCount, maxPkts); i++)
    {
        pPktInfo = CpswAppMemUtils_allocEthPktFxn(&stateObj,
                                                  rxFlowCfgInfo->rxFlowMtu,
                                                  UDMA_CACHELINE_ALIGNMENT);
        if (pPktInfo == NULL)
        {
            break;
        }

        CpswUtils_enQ(&rxFlowObj->rxFreeQ, &pPktInfo->node);
    }

    CpswAppUtils_assert(CpswUtils_getQCount(&rxFlowObj->rxFreeQ) ==
                        CpswUtils_min(rxFlowCfgInfo->pktRecvCount, maxPkts));

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(rxFlowObj->hRxFlow, &rxReadyQ);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* There should not be any packet with DMA during init */
    CpswAppUtils_assert(CpswUtils_getQCount(&rxReadyQ) == 0U);

    CpswTestCommon_submitRxPackets(rxFlowObj->hRxFlow, &rxFlowObj->rxFreeQ);
}

static int32_t CpswTestCommon_openDma(CpswTestTaskObj *taskObj,
                                      bool disableHostAddrSecureFlag)
{
    int32_t status = CPSW_SOK;
    int32_t i;
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswDma_OpenTxChPrms cpswTxChCfg;
    CpswDma_OpenRxFlowPrms cpswRxFlowCfg;
    Cpsw_AttachCoreOutArgs attachCoreOutArgs;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &stateObj->coreId, &attachCoreOutArgs);
    status = Cpsw_ioctl(stateObj->hCpsw,
                        stateObj->coreId,
                        CPSW_IOCTL_ATTACH_CORE,
                        &prms);

    CpswAppUtils_assert(status == CPSW_SOK);

    stateObj->coreKey = attachCoreOutArgs.coreKey;

    for (i = 0U; (i < taskObj->taskCfg->numTxCh) && (i < CPSW_TEST_CFG_MAX_NUM_TX_CH); i++)
    {
        CpswTestCommon_initTxFreePktQ(taskObj, i);

        /* Set configuration parameters */
        CpswDma_initTxChParams(&cpswTxChCfg);

        CpswTestCommon_createTxSem(taskObj, i);

        status = CpswAppUtils_allocTxCh(stateObj->hCpsw,
                                        stateObj->coreKey,
                                        stateObj->coreId,
                                        &stateObj->txChObj[i].txPSILThreadId);
        CpswAppUtils_assert(status == CPSW_SOK);

        CpswTestCommon_setTxChPrms(&cpswTxChCfg,
                                   taskObj,
                                   stateObj->txChObj[i].txPSILThreadId,
                                   i);
        if (stateObj->txChCfgInfo[i]->updateTxChPrms != NULL)
        {
            stateObj->txChCfgInfo[i]->updateTxChPrms(&cpswTxChCfg,
                                                     taskObj,
                                                     stateObj->txChObj[i].txPSILThreadId,
                                                     i);
        }

        stateObj->txChObj[i].hTxCh = CpswDma_openTxCh(&cpswTxChCfg);
        if (NULL != stateObj->txChObj[i].hTxCh)
        {
            status = CpswDma_enableTxEvent(stateObj->txChObj[i].hTxCh);
            if (CPSW_SOK != status)
            {
                CpswAppUtils_print("CpswDma_startTxCh() failed: %d\n", status);
                status = CPSW_EFAIL;
            }
        }
        else
        {
            CpswAppUtils_print("CpswDma_openTxCh() failed to open: %d\n",
                               status);
            status = CPSW_EFAIL;
        }

        CpswAppUtils_assert(status == CPSW_SOK);
    }

    for (i = 0U; (i < taskObj->taskCfg->numRxFlow) && (i < CPSW_TEST_CFG_MAX_NUM_RX_FLOWS); i++)
    {
        /* Open the CPSW RX flow  */
        if (status == CPSW_SOK)
        {
            CpswDma_initRxFlowParams(&cpswRxFlowCfg);

            CpswTestCommon_createRxSem(taskObj, i);

            status = CpswAppUtils_allocRxFlow(stateObj->hCpsw,
                                              stateObj->coreKey,
                                              stateObj->coreId,
                                              &cpswRxFlowCfg.startIdx,
                                              &cpswRxFlowCfg.flowIdx);

            CpswTestCommon_setRxflowPrms(&cpswRxFlowCfg, taskObj, i);
            if (stateObj->rxFlowCfgInfo[i]->updateRxFlowPrms != NULL)
            {
                stateObj->rxFlowCfgInfo[i]->updateRxFlowPrms(&cpswRxFlowCfg, taskObj, i);
            }

            CpswAppUtils_assert(status == CPSW_SOK);

            stateObj->rxFlowObj[i].hRxFlow = CpswDma_openRxFlow(&cpswRxFlowCfg);

            if (NULL == stateObj->rxFlowObj[i].hRxFlow)
            {
                CpswAppUtils_print("CpswDma_openRxFlow() failed to open: %d\n", status);
                CpswAppUtils_freeRxFlow(stateObj->hCpsw,
                                        stateObj->coreKey,
                                        stateObj->coreId,
                                        cpswRxFlowCfg.flowIdx);
                CpswAppUtils_assert(NULL != stateObj->rxFlowObj[i].hRxFlow);
            }
            else
            {
                /* save flow idx */
                stateObj->rxFlowObj[i].rxFlowIdx      = cpswRxFlowCfg.flowIdx;
                stateObj->rxFlowObj[i].rxFlowStartIdx = cpswRxFlowCfg.startIdx;
                /* Submit all ready RX buffers to DMA.*/
                CpswTestCommon_initRxReadyPktQ(taskObj, i);
            }

            if (stateObj->rxFlowCfgInfo[i]->useDefaultFlow)
            {
                status = CpswAppUtils_registerDefaultRxFlow(stateObj->hCpsw,
                                                            stateObj->coreKey,
                                                            stateObj->coreId,
                                                            stateObj->rxFlowObj[i].rxFlowStartIdx,
                                                            stateObj->rxFlowObj[i].rxFlowIdx);
                CpswAppUtils_assert(status == CPSW_SOK);
            }
        }

        CpswAppUtils_assert(status == CPSW_SOK);
    }

    CpswAppUtils_allocMac(stateObj->hCpsw, stateObj->coreKey, stateObj->coreId, &stateObj->hostMacAddr[0U]);
    CpswAppUtils_print("Host MAC address: ");
    CpswAppUtils_printMacAddr(&stateObj->hostMacAddr[0U]);
    CpswAppUtils_addHostPortEntry(stateObj->hCpsw,
                                  stateObj->coreId,
                                  stateObj->hostMacAddr);
    if ((disableHostAddrSecureFlag) && (status == CPSW_SOK))
    {
        /* For mac loopback disable secure flag for the host port entry */
        CpswTestCommon_disableHostAleEntrySecure(taskObj);
    }

    return status;
}

static void CpswTestCommon_freePkts(CpswDma_PktInfoQ *pPktQ)
{
    uint32_t i;
    uint32_t pktQCnt;
    CpswDma_PktInfo *pktInfo;

    pktQCnt = CpswUtils_getQCount(pPktQ);
    for (i = 0U; i < pktQCnt; i++)
    {
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(pPktQ);
        CpswAppMemUtils_freeEthPktFxn(pktInfo);
    }
}

static void CpswTestCommon_unregisterFlows(CpswTestTaskObj *taskObj,
                                           uint32_t rxFlowCfgIndex)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_IoctlPrms prms;
    int32_t status;

    if (stateObj->rxFlowCfgInfo[rxFlowCfgIndex]->useDefaultFlow)
    {
        status = CpswAppUtils_unregisterDefaultRxFlow(stateObj->hCpsw,
                                                      stateObj->coreKey,
                                                      stateObj->coreId,
                                                      stateObj->rxFlowObj[rxFlowCfgIndex].rxFlowStartIdx,
                                                      stateObj->rxFlowObj[rxFlowCfgIndex].rxFlowIdx);
    }
    else
    {
        CPSW_IOCTL_SET_IN_ARGS(&prms, &stateObj->rxFlowObj[rxFlowCfgIndex].rxFlowIdx);

        status = Cpsw_ioctl(stateObj->hCpsw,
                            stateObj->coreId,
                            CPSW_ALE_IOCTL_DEL_ALL_POLICER_THREADID,
                            &prms);
    }

    CpswAppUtils_assert(status == CPSW_SOK);
}

static void CpswTestCommon_closeDma(CpswTestTaskObj *taskObj)
{
    CpswDma_PktInfoQ rxReadyQ;
    CpswDma_PktInfoQ teardownFQ, teardownCQ;
    int32_t status = CPSW_SOK;
    uint32_t i;
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Retrieve any CPSW packets which are ready */
    for (i = 0U; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswUtils_initQ(&rxReadyQ);
        status = CpswDma_retrieveRxPackets(stateObj->rxFlowObj[i].hRxFlow,
                                           &rxReadyQ);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswTestCommon_freePkts(&rxReadyQ);
    }

    /* Close RX channel */
    for (i = 0U; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswUtils_initQ(&teardownFQ);
        CpswUtils_initQ(&teardownCQ);
        CpswTestCommon_unregisterFlows(taskObj, i);

        status = CpswDma_closeRxFlow(stateObj->rxFlowObj[i].hRxFlow,
                                     &teardownFQ,
                                     &teardownCQ);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswTestCommon_freePkts(&teardownFQ);
        CpswTestCommon_freePkts(&teardownCQ);
        status = CpswAppUtils_freeRxFlow(stateObj->hCpsw, stateObj->coreKey, stateObj->coreId, stateObj->rxFlowObj[i].rxFlowIdx);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswTestCommon_freePkts(&stateObj->rxFlowObj[i].rxFreeQ);
        CpswTestCommon_freePkts(&stateObj->rxFlowObj[i].rxReadyQ);
    }

    /* Close TX channel */
    for (i = 0U; i < taskObj->taskCfg->numTxCh; i++)
    {
        status = CpswDma_disableTxEvent(stateObj->txChObj[i].hTxCh);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswUtils_initQ(&teardownFQ);
        CpswUtils_initQ(&teardownCQ);
        status += CpswDma_closeTxCh(stateObj->txChObj[i].hTxCh, &teardownFQ,
                                    &teardownCQ);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswTestCommon_freePkts(&teardownFQ);
        CpswTestCommon_freePkts(&teardownCQ);

        status = CpswAppUtils_freeTxCh(stateObj->hCpsw,
                                       stateObj->coreKey,
                                       stateObj->coreId,
                                       stateObj->txChObj[i].txPSILThreadId);
        CpswAppUtils_assert(status == CPSW_SOK);
        CpswTestCommon_freePkts(&stateObj->txChObj[i].txFreePktInfoQ);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &stateObj->coreKey);

    status = Cpsw_ioctl(stateObj->hCpsw,
                        stateObj->coreId,
                        CPSW_IOCTL_DETACH_CORE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("close() failed CPSW_IOCTL_DETACH_CORE: %d\n", status);
    }

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestCommon_closeDma() failed: %d\n", status);
    }

    CpswAppMemUtils_deInit();
}

void CpswTestCommon_disableHostAleEntrySecure(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setUcastInArgs.addr.addr[0U], &stateObj->hostMacAddr[0U],
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw,
                        stateObj->coreId,
                        CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestCommon_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
        CpswAppUtils_assert(false);
    }
}

int32_t CpswTestCommon_setTxPktInfo(CpswTestTaskObj *taskObj,
                                    uint32_t         txChIndex,
                                    uint32_t         pktNum,
                                    CpswDma_PktInfo *pktInfo,
                                    bool            *testComplete)
{
    EthFrame *frame;
    /* Broadcast address */
    uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] = {0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU};
    CpswTestStateObj *stateObj          = &taskObj->stateObj;

    frame = (EthFrame *)pktInfo->bufPtr;

    memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
    memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
    frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);

    memset(&frame->payload[0U],
           (uint8_t)(0xA5 + pktNum),
           stateObj->txChCfgInfo[txChIndex]->maxTxPktLen);
    pktInfo->userBufLen = stateObj->txChCfgInfo[txChIndex]->maxTxPktLen +
                          sizeof(EthFrameHeader);
    pktInfo->appPriv = stateObj;
    *testComplete    = FALSE;

    return CPSW_SOK;
}

int32_t CpswTestCommon_processRxPkt(CpswTestTaskObj *taskObj,
                                    uint32_t rxFlowId,
                                    uint32_t pktNum,
                                    CpswDma_PktInfo *pktInfo,
                                    bool *testComplete)
{
#ifdef ENABLE_PRINTFRAME
    EthFrame *frame;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;
    CpswAppUtils_printFrame(frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif
    *testComplete = FALSE;

    return CPSW_SOK;
}

uint32_t CpswTestCommon_retrieveFreeTxPkts(CpswTestTaskObj *taskObj,
                                           uint32_t txCfgIndex)
{
    CpswDma_PktInfoQ txFreeQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t txFreeQCnt        = 0U;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswUtils_initQ(&txFreeQ);

    /* Retrieve any CPSW packets that may be free now */
    status = CpswDma_retrieveTxDonePackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                           &txFreeQ);
    if (status == CPSW_SOK)
    {
        txFreeQCnt = CpswUtils_getQCount(&txFreeQ);

        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        while (NULL != pktInfo)
        {
            CpswUtils_enQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ,
                          &pktInfo->node);

            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        }
    }
    else
    {
        CpswAppUtils_print("retrieveFreeTxPkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return txFreeQCnt;
}

uint32_t CpswTestCommon_receivePkts(CpswTestTaskObj *taskObj,
                                    uint32_t rxCfgIndex)
{
    CpswDma_PktInfoQ rxReadyQ;
    int32_t status;
    CpswTestStateObj *stateObj   = &taskObj->stateObj;
    CpswTestRxFlowObj *rxFlowObj = &stateObj->rxFlowObj[rxCfgIndex];

    CpswUtils_initQ(&rxReadyQ);

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(rxFlowObj->hRxFlow,
                                       &rxReadyQ);
    if (status == CPSW_SOK)
    {
        CpswUtils_appendQ(&rxFlowObj->rxReadyQ, &rxReadyQ);
    }
    else
    {
        CpswAppUtils_print("receivePkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return CpswUtils_getQCount(&rxFlowObj->rxReadyQ);
}

Cpsw_Handle CpswTestCommon_getCpswHandle(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_assert(stateObj->hCpsw != NULL);
    return stateObj->hCpsw;
}

static void CpswTestCommon_setUdmaRmConfig(Udma_RmInitPrms *rmInitPrms)
{
    uint32_t defaultVintr = rmInitPrms->numVintr;

    if (defaultVintr < CPSW_TEST_VINTR_MAX)
    {
        CpswAppUtils_print("[Error] Insuffient no. of VINTR as per defaultBoardCfg!! \n");
        CpswAppUtils_print("Available no. of VINTR: %d!! \n",
                            defaultVintr);
    }
}

static void CpswTestCommon_InitUdmaDrv(CpswTestTaskObj *taskObj)
{
    Udma_InitPrms initPrms, *pInitPrms = NULL;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t instId;

    if ((taskObj->taskCfg->cpswType == CPSW_9G)||
        (taskObj->taskCfg->cpswType == CPSW_5G))
    {
        instId = CpswAppUtils_getNavSSInstanceId(taskObj->taskCfg->cpswType);
        /* Initialize the UDMA driver based on NAVSS instance */
        UdmaInitPrms_init(instId, &initPrms);
        initPrms.printFxn = (Udma_PrintFxn) & CpswAppUtils_print;

        /* Need custom Udma open function for J721E as the default RM config
         * is not sufficient to allocates more than total of 5 Rx+Tx which is
         * not sufficient for testcases like policer test
         */
        CpswTestCommon_setUdmaRmConfig(&initPrms.rmInitPrms);

        pInitPrms = &initPrms;
    }

    stateObj->hUdmaDrv = CpswAppUtils_udmaOpen(taskObj->taskCfg->cpswType, pInitPrms);
    CpswAppUtils_assert(NULL != stateObj->hUdmaDrv);
}

static void CpswTestCommon_DeInitUdmaDrv(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_udmaclose(stateObj->hUdmaDrv);
}

static int32_t CpswTestCommon_transmitPkts(CpswTestTaskObj *taskObj,
                                           uint32_t txCfgIndex,
                                           CpswDma_PktInfoQ *pTxSubmitQ,
                                           uint32_t *pCurRetrieveCnt)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    *pCurRetrieveCnt = 0;
    while (0U != CpswUtils_getQCount(pTxSubmitQ))
    {
        status = CpswTestCommon_submitTxPackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                                pTxSubmitQ);
        if (status != CPSW_SOK)
        {
            break;
        }

        if (0U != CpswUtils_getQCount(pTxSubmitQ))
        {
            *pCurRetrieveCnt += CpswTestCommon_waitForXmit(taskObj, txCfgIndex);
        }
    }

    return status;
}

static uint32_t CpswTestCommon_waitForXmit(CpswTestTaskObj *taskObj,
                                           uint32_t txCfgIndex)
{
    bool semStatus;
    uint32_t txFreeQCnt;
    CpswTestTxChObj *txChObj;

    txChObj    = &taskObj->stateObj.txChObj[txCfgIndex];
    txFreeQCnt = CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
    while (txFreeQCnt == 0)
    {
        semStatus = Semaphore_pend(txChObj->hTxCQSem, BIOS_WAIT_FOREVER);
        CpswAppUtils_assert(semStatus == TRUE);
        txFreeQCnt = CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
    }

    return txFreeQCnt;
}

static void CpswTestCommon_xmitTask(UArg a0,
                                    UArg a1)
{
    int32_t retVal           = CPSW_SOK;
    CpswTestTaskObj *taskObj = (CpswTestTaskObj *)a0;
    uint32_t txCfgIndex      = (uint32_t)a1;
    CpswDma_PktInfo *pktInfo;
    CpswDma_PktInfoQ txSubmitQ;
    CpswTestTxChCfgInfo *txChCfgInfo;
    CpswTestTxChObj *txChObj;
    CpswTestCommon_TxTestStatus_t testStatusMsg;
    CpswTest_SetTxPktInfoFxn setTxPktInfo;
    bool mbxStatus;
    bool testComplete;
    Task_Stat stat;
    uint32_t startTime;
    int32_t byteCount;

    txChCfgInfo = taskObj->stateObj.txChCfgInfo[txCfgIndex];
    txChObj     = &taskObj->stateObj.txChObj[txCfgIndex];

    /* Open the CPSW TX channel  */
    if (txChCfgInfo->setTxPktInfo == NULL)
    {
        setTxPktInfo = &CpswTestCommon_setTxPktInfo;
    }
    else
    {
        setTxPktInfo = txChCfgInfo->setTxPktInfo;
    }

    if (txChCfgInfo->preTxSend != NULL)
    {
        retVal = txChCfgInfo->preTxSend(taskObj, txCfgIndex);
    }

    CpswDma_enableTxEvent(txChObj->hTxCh);
    txChObj->processCount = 0;
    txChObj->retrieveCnt  = 0;
    byteCount             = 0;
    startTime             = 0;
    if (CPSW_SOK == retVal)
    {
        testComplete = FALSE;
        while ((txChObj->processCount < txChCfgInfo->pktSendCount)
               &&
               (CPSW_SOK == retVal)
               &&
               (testComplete == FALSE))
        {
            CpswUtils_initQ(&txSubmitQ);

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txChObj->txFreePktInfoQ);
            while (NULL != pktInfo)
            {
                retVal = setTxPktInfo(taskObj,
                                      txCfgIndex,
                                      txChObj->processCount,
                                      pktInfo,
                                      &testComplete);
                if (txChObj->processCount == 0)
                {
                    startTime = AppUtils_getCurTimeInMsec();
                }

                txChObj->processCount++;
                byteCount += pktInfo->userBufLen;

                /* Enqueue the packet for later transmission */
                CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

                if (txChObj->processCount >= txChCfgInfo->pktSendCount)
                {
                    break;
                }

                if (testComplete)
                {
                    CpswAppUtils_print("Test Completed:%s,Id:%u,Frame:%u\n", __func__, txCfgIndex, txChObj->processCount);
                    break;
                }

                if (retVal != CPSW_SOK)
                {
                    CpswAppUtils_print("Test Fail:%s,Id:%u,Frame:%u\n", __func__, txCfgIndex, txChObj->processCount);
                    break;
                }

                /* Dequeue one free TX Eth packet */
                pktInfo =
                    (CpswDma_PktInfo *)CpswUtils_deQ(&txChObj->txFreePktInfoQ);
            }

            if (CpswUtils_getQCount(&txSubmitQ) == 0)
            {
                txChObj->retrieveCnt += CpswTestCommon_waitForXmit(taskObj, txCfgIndex);
            }
            else
            {
                uint32_t curRetrieveCnt = 0;

                retVal = CpswTestCommon_transmitPkts(taskObj, txCfgIndex,
                                                     &txSubmitQ,
                                                     &curRetrieveCnt);
                txChObj->retrieveCnt += curRetrieveCnt;
            }

            if (retVal != CPSW_SOK)
            {
                break;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        while (txChObj->retrieveCnt < txChCfgInfo->pktSendCount)
        {
            txChObj->retrieveCnt += CpswTestCommon_waitForXmit(taskObj, txCfgIndex);
        }
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t elapsedMSecs = AppUtils_getCurTimeInMsec() - startTime;
        txChObj->transmitBitRate = (byteCount * 8) / elapsedMSecs;

        if (txChCfgInfo->postTxSend != NULL)
        {
            retVal = txChCfgInfo->postTxSend(taskObj, txCfgIndex);
        }
    }

    CpswAppUtils_print("Transmit Completed:%s,Id:%u,Frame:%u\n", __func__, txCfgIndex, txChObj->retrieveCnt);
    CpswDma_disableTxEvent(txChObj->hTxCh);
    Task_stat(Task_self(), &stat);
    CpswAppUtils_assert(stat.used < stat.stackSize);
    testStatusMsg.stackUsed  = stat.used;
    testStatusMsg.testStatus = retVal;
    mbxStatus                =
        Mailbox_post(taskObj->stateObj.txChObj[txCfgIndex].hTxTestStatusMbx,
                     &testStatusMsg,
                     BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(mbxStatus == TRUE);
}

static void CpswTestCommon_deleteTxTask(CpswTestTaskObj *taskObj,
                                        uint32_t txCfgIndex)
{
    Task_Stat statbuf;

    Task_stat(taskObj->stateObj.txChObj[txCfgIndex].hTxTask, &statbuf);
    Task_delete(&taskObj->stateObj.txChObj[txCfgIndex].hTxTask);

    Utils_memFree(UTILS_MEM_HEAP_ID_DDR, statbuf.stack, statbuf.stackSize);
}

static void CpswTestCommon_deleteTxSem(CpswTestTaskObj *taskObj,
                                       uint32_t txCfgIndex)
{
    if (taskObj->stateObj.txChObj[txCfgIndex].hTxCQSem != NULL)
    {
        Semaphore_delete(&taskObj->stateObj.txChObj[txCfgIndex].hTxCQSem);
    }
}

static void CpswTestCommon_deleteTxMbx(CpswTestTaskObj *taskObj,
                                       uint32_t txCfgIndex)
{
    Mailbox_delete(&taskObj->stateObj.txChObj[txCfgIndex].hTxTestStatusMbx);
}

static void  CpswTestCommon_createTxTask(CpswTestTaskObj *taskObj,
                                         uint32_t txCfgIndex)
{
    Task_Params params;
    Error_Block eb;

    Error_init(&eb);

    Task_Params_init(&params);

    params.priority = CPSW_TEST_COMMON_TX_TASK_PRI;
    params.stack    = Utils_memAlloc(UTILS_MEM_HEAP_ID_DDR,
                                     CPSW_TEST_COMMON_TX_TASK_STACK_SIZE,
                                     sizeof(uint64_t));
    CpswAppUtils_assert(params.stack != NULL);
    params.stackSize      = CPSW_TEST_COMMON_TX_TASK_STACK_SIZE;
    params.arg0           = (UArg)taskObj;
    params.arg1           = (UArg)txCfgIndex;
    params.instance->name = "Test common Transmit task";

    taskObj->stateObj.txChObj[txCfgIndex].hTxTask = Task_create(CpswTestCommon_xmitTask,
                                                                &params,
                                                                &eb);
    CpswAppUtils_assert(Error_check(&eb) == FALSE);
}

static void  CpswTestCommon_createTxSem(CpswTestTaskObj *taskObj,
                                        uint32_t txCfgIndex)
{
    Semaphore_Params semParams;

    Semaphore_Params_init(&semParams);
    semParams.mode                                 = Semaphore_Mode_BINARY;
    taskObj->stateObj.txChObj[txCfgIndex].hTxCQSem = Semaphore_create(0, &semParams, NULL);
    CpswAppUtils_assert(NULL != taskObj->stateObj.txChObj[txCfgIndex].hTxCQSem);
}

static void  CpswTestCommon_createTxMbx(CpswTestTaskObj *taskObj,
                                        uint32_t txCfgIndex)
{
    Mailbox_Params mbxParams;

    Mailbox_Params_init(&mbxParams);
    taskObj->stateObj.txChObj[txCfgIndex].hTxTestStatusMbx =
        Mailbox_create(sizeof(CpswTestCommon_TxTestStatus_t),
                       CPSW_TEST_COMMON_TX_TEST_STATUS_COUNT,
                       &mbxParams,
                       NULL);
    CpswAppUtils_assert(taskObj->stateObj.txChObj[txCfgIndex].hTxTestStatusMbx != NULL);
}

static void CpswTestCommon_waitTxTaskTermination(CpswTestTaskObj *taskObj,
                                                 uint32_t txCfgIndex)
{
    CpswTestCommon_TxTestStatus_t testStatusMsg;
    bool mbxStatus;

    mbxStatus =
        Mailbox_pend(taskObj->stateObj.txChObj[txCfgIndex].hTxTestStatusMbx,
                     &testStatusMsg,
                     BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(mbxStatus == TRUE);

    taskObj->stateObj.txChObj[txCfgIndex].txTaskStatus =
        testStatusMsg.testStatus;
    taskObj->stateObj.txChObj[txCfgIndex].stackUsed =
        testStatusMsg.stackUsed;

    while (Task_getMode(taskObj->stateObj.txChObj[txCfgIndex].hTxTask) !=
           Task_Mode_TERMINATED)
    {
        Task_sleep(1);
    }
}

static void CpswTestCommon_spawnTxChTasks(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    for (i = 0; i < taskObj->taskCfg->numTxCh; i++)
    {
        CpswTestCommon_createTxMbx(taskObj, i);
        CpswTestCommon_createTxTask(taskObj, i);
    }
}

static void CpswTestCommon_terminateTxChTasks(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    for (i = 0; i < taskObj->taskCfg->numTxCh; i++)
    {
        CpswTestCommon_waitTxTaskTermination(taskObj, i);
        CpswTestCommon_deleteTxTask(taskObj, i);
        CpswTestCommon_deleteTxMbx(taskObj, i);
    }
}

static void CpswTestCommon_deleteRxTask(CpswTestTaskObj *taskObj,
                                        uint32_t rxCfgIndex)
{
    Task_Stat statbuf;

    Task_stat(taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTask, &statbuf);

    Task_delete(&taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTask);

    Utils_memFree(UTILS_MEM_HEAP_ID_DDR, statbuf.stack, statbuf.stackSize);
}

static void CpswTestCommon_deleteRxSem(CpswTestTaskObj *taskObj,
                                       uint32_t rxCfgIndex)
{
    if (taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxCQSem != NULL)
    {
        Semaphore_delete(&taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxCQSem);
    }
}

static void CpswTestCommon_deleteRxMbx(CpswTestTaskObj *taskObj,
                                       uint32_t rxCfgIndex)
{
    Mailbox_delete(&taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTestStatusMbx);
}

static void  CpswTestCommon_createRxTask(CpswTestTaskObj *taskObj,
                                         uint32_t rxCfgIndex)
{
    Task_Params params;
    Error_Block eb;

    Error_init(&eb);

    Task_Params_init(&params);

    params.priority = CPSW_TEST_COMMON_RX_TASK_PRI;
    params.stack    = Utils_memAlloc(UTILS_MEM_HEAP_ID_DDR,
                                     CPSW_TEST_COMMON_RX_TASK_STACK_SIZE,
                                     sizeof(uint64_t));
    CpswAppUtils_assert(params.stack != NULL);

    params.stackSize      = CPSW_TEST_COMMON_RX_TASK_STACK_SIZE;
    params.arg0           = (UArg)taskObj;
    params.arg1           = (UArg)rxCfgIndex;
    params.instance->name = "Test common Receive task";

    taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTask = Task_create(CpswTestCommon_recvTask,
                                                                  &params,
                                                                  &eb);
    CpswAppUtils_assert(Error_check(&eb) == FALSE);
}

void  CpswTestCommon_createRxSem(CpswTestTaskObj *taskObj,
                                 uint32_t rxCfgIndex)
{
    Semaphore_Params semParams;

    Semaphore_Params_init(&semParams);
    semParams.mode                                   = Semaphore_Mode_BINARY;
    taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxCQSem = Semaphore_create(0, &semParams, NULL);
    CpswAppUtils_assert(NULL != taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxCQSem);
}

static void  CpswTestCommon_createRxMbx(CpswTestTaskObj *taskObj,
                                        uint32_t rxCfgIndex)
{
    Mailbox_Params mbxParams;

    Mailbox_Params_init(&mbxParams);
    taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTestStatusMbx =
        Mailbox_create(sizeof(CpswTestCommon_RxTestStatus_t),
                       CPSW_TEST_COMMON_RX_TEST_STATUS_COUNT,
                       &mbxParams,
                       NULL);
}

static void CpswTestCommon_waitRxTaskTermination(CpswTestTaskObj *taskObj,
                                                 uint32_t rxCfgIndex)
{
    CpswTestCommon_RxTestStatus_t testStatusMsg;
    bool mbxStatus;

    mbxStatus = Mailbox_pend(taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTestStatusMbx,
                             &testStatusMsg,
                             BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(mbxStatus == TRUE);

    taskObj->stateObj.rxFlowObj[rxCfgIndex].rxTaskStatus = testStatusMsg.testStatus;
    taskObj->stateObj.rxFlowObj[rxCfgIndex].stackUsed    = testStatusMsg.stackUsed;

    while (Task_getMode(taskObj->stateObj.rxFlowObj[rxCfgIndex].hRxTask) !=
           Task_Mode_TERMINATED)
    {
        Task_sleep(1);
    }
}

void CpswTestCommon_spawnRxFlowTasks(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    for (i = 0; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswTestCommon_createRxMbx(taskObj, i);
        CpswTestCommon_createRxTask(taskObj, i);
    }
}

void CpswTestCommon_terminateRxFlowTasks(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    for (i = 0; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswTestCommon_waitRxTaskTermination(taskObj, i);
        CpswTestCommon_deleteRxTask(taskObj, i);
        CpswTestCommon_deleteRxMbx(taskObj, i);
    }
}

static void CpswTestCommon_recvTask(UArg a0,
                                    UArg a1)
{
    CpswTestTaskObj *taskObj = (CpswTestTaskObj *)a0;
    uint32_t rxCfgIndex      = (uint32_t)a1;
    CpswDma_PktInfo *pktInfo;
    uint32_t rxReadyCnt;
    CpswTestRxFlowCfgInfo *rxFlowCfgInfo;
    CpswTestRxFlowObj *rxFlowObj;
    bool semStatus;
    bool testComplete;
    int32_t retVal;
    CpswTest_ProcessRxPktFxn processRxPkt;
    CpswTestCommon_RxTestStatus_t testStatusMsg;
    Bool mbxStatus;
    uint32_t startTime;
    int32_t byteCount;
    Task_Stat stat;

    rxFlowCfgInfo = taskObj->stateObj.rxFlowCfgInfo[rxCfgIndex];
    rxFlowObj     = &taskObj->stateObj.rxFlowObj[rxCfgIndex];
    retVal        = CPSW_SOK;
    byteCount     = 0;
    startTime     = 0;

    /* Set process packet function to common function if application passes null */
    if (rxFlowCfgInfo->processRxPkt == NULL)
    {
        processRxPkt = &CpswTestCommon_processRxPkt;
    }
    else
    {
        processRxPkt = rxFlowCfgInfo->processRxPkt;
    }

    if (rxFlowCfgInfo->preRxProcess != NULL)
    {
        retVal = rxFlowCfgInfo->preRxProcess(taskObj, rxCfgIndex);
    }

    if (CPSW_SOK == retVal)
    {
        rxFlowObj->processCount = 0;
        testComplete            = FALSE;
        while ((rxFlowObj->processCount < rxFlowCfgInfo->pktRecvCount)
               &&
               (CPSW_SOK == retVal)
               &&
               (testComplete == FALSE))
        {
            /* Get the packets received so far */
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, rxCfgIndex);
            if (rxReadyCnt > 0U)
            {
                /* Consume the received packets and release them */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxFlowObj->rxReadyQ);
                while (NULL != pktInfo)
                {
                    if (rxFlowObj->processCount == 0)
                    {
                        startTime = AppUtils_getCurTimeInMsec();
                    }

                    byteCount += pktInfo->userBufLen;

                    retVal = processRxPkt(taskObj,
                                          rxCfgIndex,
                                          rxFlowObj->processCount,
                                          pktInfo,
                                          &testComplete);
                    if (retVal != CPSW_SOK)
                    {
                        CpswAppUtils_print("Receive Fail:%s,Id:%u,Frame:%u\n",
                                           __func__, rxCfgIndex, rxFlowObj->processCount);
                        CpswUtils_enQ(&rxFlowObj->rxFreeQ, &pktInfo->node);
                        break;
                    }

                    CpswUtils_enQ(&rxFlowObj->rxFreeQ, &pktInfo->node);

                    rxFlowObj->processCount++;
                    if (!testComplete)
                    {
                        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxFlowObj->rxReadyQ);
                    }
                    else
                    {
                        pktInfo = NULL;
                    }
                }

                CpswTestCommon_submitRxPackets(rxFlowObj->hRxFlow, &rxFlowObj->rxFreeQ);
            }
            else
            {
                semStatus = Semaphore_pend(rxFlowObj->hRxCQSem, BIOS_WAIT_FOREVER);
                CpswAppUtils_assert(semStatus == TRUE);
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t elapsedMSecs = AppUtils_getCurTimeInMsec() - startTime;
        rxFlowObj->receivedBitRate = (byteCount / elapsedMSecs) * 8 * 1000;

        if (rxFlowCfgInfo->postRxProcess != NULL)
        {
            retVal = rxFlowCfgInfo->postRxProcess(taskObj, rxCfgIndex);
        }
    }

    CpswAppUtils_print("Receive Complete:%s,Id:%u,Frame:%u\n", __func__, rxCfgIndex, rxFlowObj->processCount);

    Task_stat(Task_self(), &stat);
    CpswAppUtils_assert(stat.used < stat.stackSize);

    testStatusMsg.stackUsed  = stat.used;
    testStatusMsg.testStatus = retVal;
    mbxStatus                = Mailbox_post(rxFlowObj->hRxTestStatusMbx,
                                            &testStatusMsg,
                                            BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(mbxStatus == TRUE);
}

static bool CpswTestCommon_checkTxError(CpswTestTaskObj *taskObj)
{
    uint32_t i;
    bool error;

    for (i = 0; i < taskObj->taskCfg->numTxCh; i++)
    {
        if (taskObj->stateObj.txChObj[i].txTaskStatus != CPSW_SOK)
        {
            break;
        }
    }

    if (i < taskObj->taskCfg->numTxCh)
    {
        error = TRUE;
    }
    else
    {
        error = FALSE;
    }

    return error;
}

static bool CpswTestCommon_checkRxError(CpswTestTaskObj *taskObj)
{
    uint32_t i;
    bool error;

    for (i = 0; i < taskObj->taskCfg->numRxFlow; i++)
    {
        if (taskObj->stateObj.rxFlowObj[i].rxTaskStatus != CPSW_SOK)
        {
            break;
        }
    }

    if (i < taskObj->taskCfg->numRxFlow)
    {
        error = TRUE;
    }
    else
    {
        error = FALSE;
    }

    return error;
}

static void CpswTestCommon_printTxStackUsage(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    CpswAppUtils_print("\r\n");
    for (i = 0; i < taskObj->taskCfg->numTxCh; i++)
    {
        CpswAppUtils_print("Tx stack usage.Tx Cfg Id:%d, Stack Used:%d\n",
                           i, taskObj->stateObj.txChObj[i].stackUsed);
    }
}

static void CpswTestCommon_printRxStackUsage(CpswTestTaskObj *taskObj)
{
    uint32_t i;

    CpswAppUtils_print("\r\n");
    for (i = 0; i < taskObj->taskCfg->numRxFlow; i++)
    {
        CpswAppUtils_print("Rx stack usage.Rx Cfg Id:%d, Stack Used:%d\n",
                           i, taskObj->stateObj.rxFlowObj[i].stackUsed);
    }
}



int32_t CpswTestCommon_pktTxRx(CpswTestTaskObj *taskObj)
{
    int32_t retVal;
    Int prevPri;

    prevPri = Task_setPri(Task_self(), CPSW_TEST_COMMON_MAIN_TASK_PRI);

    CpswTestCommon_spawnTxChTasks(taskObj);
    CpswTestCommon_spawnRxFlowTasks(taskObj);

    CpswTestCommon_terminateTxChTasks(taskObj);
    CpswTestCommon_terminateRxFlowTasks(taskObj);

    retVal = CpswTestCommon_checkTxError(taskObj);
    if (CPSW_SOK == retVal)
    {
        retVal = CpswTestCommon_checkRxError(taskObj);
    }

    CpswTestCommon_printTxStackUsage(taskObj);
    CpswTestCommon_printRxStackUsage(taskObj);

    /* Restore task pritority to original */
    Task_setPri(Task_self(), prevPri);
    return retVal;
}

int32_t CpswTestCommon_setAleMulticastEntry(CpswTestTaskObj *taskObj,
                                            uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                                            uint32_t vlanId,
                                            uint32_t numIgnBits,
                                            uint32_t portMask)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setMcastOutArgs;
    CpswAle_SetMcastEntryInArgs setMcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setMcastInArgs.addr.addr[0], macAddr,
           sizeof(setMcastInArgs.addr.addr));
    setMcastInArgs.addr.vlanId = vlanId;

    setMcastInArgs.info.superFlag  = false;
    setMcastInArgs.info.fwdState   = CPSW_ALE_FWDSTLVL_FORWARDING;
    setMcastInArgs.info.portMask   = portMask;
    setMcastInArgs.info.numIgnBits = numIgnBits;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_MULTICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setAleMulticastEntry() failed CPSW_ALE_IOCTL_ADD_MULTICAST: %d\n",
                           status);
    }

    return status;
}

int32_t CpswTestCommon_getAleMulticastEntry(CpswTestTaskObj *taskObj,
                                            uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                                            uint32_t vlanId,
                                            uint32_t numIgnBits,
                                            uint32_t *pPortMask)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_GetMcastEntryInArgs inArgs;
    CpswAle_GetMcastEntryOutArgs getMcastOutArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&inArgs.addr.addr[0], macAddr, sizeof(inArgs.addr.addr));
    inArgs.addr.vlanId = vlanId;
    inArgs.numIgnBits  = numIgnBits;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &getMcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_LOOKUP_MULTICAST,
                        &prms);
    if (status == CPSW_SOK)
    {
        // CpswAppUtils_print(ALE ENTRY INDEX: %d\n", getMcastOutArgs.aleIndex);
        // CpswAppUtils_print("portMask: %d\n", getMcastOutArgs.outInfo.portMask);
        // CpswAppUtils_print("numIgnBits: %d\n", getMcastOutArgs.outInfo.numIgnBits);
        CpswAppUtils_assert(numIgnBits == getMcastOutArgs.outInfo.numIgnBits);
    }
    else
    {
        CpswAppUtils_print("CPSW_ALE_IOCTL_LOOKUP_MULTICAST IOCTL Failed: %d\n",
                           status);
    }

    *pPortMask = getMcastOutArgs.outInfo.portMask;
    return status;
}

static void CpswTestCommon_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                                           CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitRxPackets(hRxFlow, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswAppUtils_print("CpswDma_submitRxPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        /* Copy unused packets to back to readyQ */
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }
}

static int32_t CpswTestCommon_submitTxPackets(CpswDma_TxChHandle hTxChannel,
                                              CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitTxReadyPackets(hTxChannel, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswAppUtils_print("CpswDma_submitTxReadyPackets: failed to submit pkts: %d\n", retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }

    return retVal;
}

void CpswTestCommon_discardRxPkts(CpswTestTaskObj *taskObj,
                                  uint32_t rxCfgIndex)
{
    CpswTestStateObj *stateObj   = &taskObj->stateObj;
    CpswTestRxFlowObj *rxFlowObj = &stateObj->rxFlowObj[rxCfgIndex];

    CpswUtils_appendQ(&rxFlowObj->rxFreeQ, &rxFlowObj->rxReadyQ);
    CpswUtils_initQ(&rxFlowObj->rxReadyQ);
    while (CpswTestCommon_receivePkts(taskObj, rxCfgIndex))
    {
        CpswUtils_appendQ(&rxFlowObj->rxFreeQ, &rxFlowObj->rxReadyQ);
        CpswUtils_initQ(&rxFlowObj->rxReadyQ);
    }

    CpswTestCommon_submitRxPackets(rxFlowObj->hRxFlow, &rxFlowObj->rxFreeQ);
}

uint32_t CpswTestCommon_getMacPortAleMask(CpswTestTaskObj *taskObj)
{
    CpswTestMacPortList_t enabledPorts;
    uint32_t i;
    uint32_t alePortMask;

    alePortMask = 0;
    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    for (i = 0; i < enabledPorts.numMacPorts; i++)
    {
        alePortMask |=
            CPSW_SET_BIT(CPSW_ALE_MACPORT_TO_ALEPORT(enabledPorts.macPortList[i]));
    }

    return alePortMask;
}

static int32_t CpswTestCommon_addDefaultHostPortEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], gDefaultHostMacAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.trunk   = false;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw,
                        stateObj->coreId,
                        CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestCommon_addDefaultHostPortEntry() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n", status);
        CpswAppUtils_assert(false);
    }

    return status;
}


void CpswTestCommon_getDefaultHostMacAddr(uint8_t *macAddr)
{
    memcpy(macAddr, gDefaultHostMacAddr, sizeof(gDefaultHostMacAddr));
}


/* end of file */
