/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_switching_logic.c
 *
 * \brief    This file contains the cpsw_switching_logic test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_basicswitching.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
#define CPSW_TEST_SWITCHING_DEFAULT_SHORTIPG_THRESHOLD                    (11)

static int32_t CpswTestSwitching_setShortIPG(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_SetShortTxIPGConfigInArgs setShortIPGInArgs;
    CpswTestMacPortList_t enabledPorts;
    int32_t status;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    CpswAppUtils_print("Enabling IPG Config - Ingress port %u, Egress port %u \n",
                       enabledPorts.macPortList[0U],
                       enabledPorts.macPortList[1U]);

    CPSW_IOCTL_SET_IN_ARGS(&prms, &setShortIPGInArgs);
    setShortIPGInArgs.configureGapThreshold                           = TRUE;
    setShortIPGInArgs.ipgTriggerThresholdBlkCount                     = 0;
    setShortIPGInArgs.numMacPorts                                     = 2;
    setShortIPGInArgs.portIPGCfg[0].portNum                           = enabledPorts.macPortList[0U];
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGapLimit = false;

    setShortIPGInArgs.portIPGCfg[1].portNum                           = enabledPorts.macPortList[1U];
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGapLimit = false;

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_SHORT_IPG_CONFIG,
                        &prms);
    if (CPSW_SOK == status)
    {
        Cpsw_GetShortTxIPGConfigOutArgs getShortIPGOutArgs;

        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getShortIPGOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_GET_SHORT_IPG_CONFIG,
                            &prms);
        if (CPSW_SOK == status)
        {
            uint32_t i;

            CpswAppUtils_assert(getShortIPGOutArgs.ipgTriggerThresholdBlkCount == 0);
            for (i = 0; i < getShortIPGOutArgs.numMacPorts; i++)
            {
                if ((getShortIPGOutArgs.portIPGCfg[i].portNum == enabledPorts.macPortList[1U])
                    ||
                    (getShortIPGOutArgs.portIPGCfg[i].portNum == enabledPorts.macPortList[1U]))
                {
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGap == true);
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGapLimit == false);
                }
            }
        }
    }

    return status;
}

int32_t CpswTestSwitching_RunWithIPG(CpswTestTaskObj *taskObj)
{
    int32_t status;

    CpswTestSwitching_setShortIPG(taskObj);

    status = CpswTestCommon_Run(taskObj);
    return status;
}

int32_t CpswTestSwitching_Run(CpswTestTaskObj *taskObj)
{
    int32_t status;

    status = CpswTestCommon_Run(taskObj);

    return status;
}
