/*
 *  Copyright (c) Texas Instruments Incorporated 2020
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_bcastmcast_limit.c
 *
 * \brief    This file contains the cpsw_test_BCASTMCAST_LIMIT test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_bcastmcast_limit.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT    (CPSW_MAC_PORT_3)
#define CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT     (CPSW_MAC_PORT_2)

#define CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_INPORT_NUMPKTSPERSEC      (40000)
#define CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_INPORT_NUMPKTSPERSEC      (20000)

#define CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_OUTPORT_NUMPKTSPERSEC     (60000)
#define CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_OUTPORT_NUMPKTSPERSEC     (30000)

#define CPSW_TEST_BCASTMCAST_LIMIT_RATELIMIT_AT_TX                       (true)


/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswTestBcastMcastLimit_Config(CpswTestTaskObj *taskObj);

static int32_t CpswTestBcastMcastLimit_AddAleEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
static uint8_t testMCastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x01, 0xFF, 0xFF, 0xFF, 0x00, 0x00
};

static uint8_t testBCastAddr[CPSW_MAC_ADDR_LEN] =
{
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
int32_t CpswTestBcastMcastLimit_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    status = CpswTestBcastMcastLimit_Config(taskObj);

    if (status == CPSW_SOK)
    {
        status = CpswTestBcastMcastLimit_AddAleEntry(taskObj);
    }

    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        status = CpswTestCommon_Run(taskObj);
    }

    return status;
}

static uint32_t CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(CpswTestTaskObj *taskObj,
                                                                   uint32_t numPktsPerSec)
{
    const CpswSoc_Config *socCfg;
    const CpswSoc_CpswIpInfo *ipInfo;
    uint32_t alePrescale;
    uint32_t preScalerPulseFreq;

    socCfg        = CpswSoc_getConfig(taskObj->taskCfg->cpswType);
    ipInfo        = CpswUtils_getIpInfo(taskObj->taskCfg->cpswType);
    alePrescale   = CSL_CPSW_getAlePrescaleReg((CSL_AleRegs *)ipInfo->aleInfo.regBaseAddr);
    if (alePrescale == 0)
    {
        alePrescale = 1;
    }
    preScalerPulseFreq = socCfg->cppiClkFreqHz / alePrescale;
    return ((numPktsPerSec / preScalerPulseFreq) * preScalerPulseFreq);
}


static int32_t CpswTestBcastMcastLimit_Config(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_IoctlPrms prms;
    CpswAle_SetBcastMcastRateLimitInArgs setInArgs;
    CpswAle_GetBcastMcastRateLimitOutArgs getOutArgs;
    int32_t status;

    /* CPSW_ALE_IOCTL_DISABLE_BCAST_MCAST_LIMIT
     */
    memset(&setInArgs, 0, sizeof(setInArgs));
    setInArgs.numPorts = 2;
    setInArgs.rateLimitAtTxPort = CPSW_TEST_BCASTMCAST_LIMIT_RATELIMIT_AT_TX;

    /* BCAST/MCAST Rate Limit for Ingress Port */
    setInArgs.portPrms[0].portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT);
    setInArgs.portPrms[0].enableBRateLimitForPort = true;
    setInArgs.portPrms[0].enableMRateLimitForPort = true;
    setInArgs.portPrms[0].mcastLimitNumPktsPerSec = CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_INPORT_NUMPKTSPERSEC;
    setInArgs.portPrms[0].bcastLimitNumPktsPerSec = CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_INPORT_NUMPKTSPERSEC;

    /* BCAST/MCAST Rate Limit for Egress Port */
    setInArgs.portPrms[1].portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT);
    setInArgs.portPrms[1].enableBRateLimitForPort = true;
    setInArgs.portPrms[1].enableMRateLimitForPort = true;
    setInArgs.portPrms[1].mcastLimitNumPktsPerSec = CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_OUTPORT_NUMPKTSPERSEC;
    setInArgs.portPrms[1].bcastLimitNumPktsPerSec = CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_OUTPORT_NUMPKTSPERSEC;


    CPSW_IOCTL_SET_IN_ARGS(&prms, &setInArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                        CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestBcastMcastLimit_Config() failed CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        memset(&getOutArgs, 0, sizeof(getOutArgs));
        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                            CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestBcastMcastLimit_Config() failed CPSW_MACPORT_IOCTL_GET_BCASTMCAST_LIMIT: %d\n",
                               status);
        }
    }
    /* Compare set and retrieved Bcast Mcast Rate Limit configuration */
    if (status == CPSW_SOK)
    {
        CpswAppUtils_assert(getOutArgs.numPorts == 2);
        CpswAppUtils_assert(getOutArgs.rateLimitEnabled == true);
        CpswAppUtils_assert(getOutArgs.rateLimitAtTxPort == CPSW_TEST_BCASTMCAST_LIMIT_RATELIMIT_AT_TX);

        CpswAppUtils_assert(getOutArgs.portPrms[0].enableMRateLimitForPort == true);
        CpswAppUtils_assert(getOutArgs.portPrms[0].enableBRateLimitForPort == true);
        CpswAppUtils_assert((getOutArgs.portPrms[0].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT))
                            ||
                            (getOutArgs.portPrms[0].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT)));
        if (getOutArgs.portPrms[0].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT))
        {
            CpswAppUtils_assert(getOutArgs.portPrms[0].bcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_INPORT_NUMPKTSPERSEC));
            CpswAppUtils_assert(getOutArgs.portPrms[0].mcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_INPORT_NUMPKTSPERSEC));

        }
        else
        {
            CpswAppUtils_assert(getOutArgs.portPrms[0].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT));
            CpswAppUtils_assert(getOutArgs.portPrms[0].bcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_OUTPORT_NUMPKTSPERSEC));
            CpswAppUtils_assert(getOutArgs.portPrms[0].mcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_OUTPORT_NUMPKTSPERSEC));
        }

        CpswAppUtils_assert(getOutArgs.portPrms[1].enableMRateLimitForPort == true);
        CpswAppUtils_assert(getOutArgs.portPrms[1].enableBRateLimitForPort == true);
        CpswAppUtils_assert((getOutArgs.portPrms[1].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT))
                            ||
                            (getOutArgs.portPrms[1].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT)));
        if (getOutArgs.portPrms[1].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_INGRESS_PORT))
        {
            CpswAppUtils_assert(getOutArgs.portPrms[1].bcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_INPORT_NUMPKTSPERSEC));
            CpswAppUtils_assert(getOutArgs.portPrms[1].mcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_INPORT_NUMPKTSPERSEC));

        }
        else
        {
            CpswAppUtils_assert(getOutArgs.portPrms[1].portNum == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_BCASTMCAST_LIMIT_EGRESS_PORT));
            CpswAppUtils_assert(getOutArgs.portPrms[1].bcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_BCASTLIMIT_OUTPORT_NUMPKTSPERSEC));
            CpswAppUtils_assert(getOutArgs.portPrms[1].mcastLimitNumPktsPerSec == CpswTestBcastMcastLimit_mapNumPktsPerSec2RateLimit(taskObj, CPSW_TEST_BCASTMCAST_LIMIT_MCASTLIMIT_OUTPORT_NUMPKTSPERSEC));
        }
    }

    return status;
}

static int32_t CpswTestBcastMcastLimit_AddAleEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setMcastOutArgs;
    CpswAle_SetMcastEntryInArgs setMcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setMcastInArgs, 0, sizeof(setMcastInArgs));
    memcpy(&setMcastInArgs.addr.addr[0U], testMCastAddr,
           sizeof(setMcastInArgs.addr.addr));
    setMcastInArgs.addr.vlanId  = 0;
    setMcastInArgs.info.superFlag = false;
    setMcastInArgs.info.numIgnBits = 8;
    setMcastInArgs.info.fwdState = CPSW_ALE_FWDSTLVL_FORWARDING;
    setMcastInArgs.info.portMask = CpswTestCommon_getMacPortAleMask(taskObj);
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);
    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_MULTICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestBcastMcastLimit_AddAleEntry() failed CPSW_ALE_IOCTL_ADD_MULTICAST: %d\n",
                           status);
    }
    if (status == CPSW_SOK)
    {
        memset(&setMcastInArgs, 0, sizeof(setMcastInArgs));
        memcpy(&setMcastInArgs.addr.addr[0U], testBCastAddr,
               sizeof(setMcastInArgs.addr.addr));
        setMcastInArgs.addr.vlanId  = 0;
        setMcastInArgs.info.superFlag = false;
        setMcastInArgs.info.numIgnBits = 8;
        setMcastInArgs.info.fwdState = CPSW_ALE_FWDSTLVL_FORWARDING;
        setMcastInArgs.info.portMask = CpswTestCommon_getMacPortAleMask(taskObj);
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_MULTICAST, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestBcastMcastLimit_AddAleEntry() failed CPSW_ALE_IOCTL_ADD_MULTICAST: %d\n",
                               status);
        }
    }
    return status;
}

void CpswTestBcastMcastLimit_setOpenPrms(CpswTestTaskObj *taskObj,
                                        Cpsw_Config *pCpswCfg,
                                        CpswOsal_Prms *pOsalPrms,
                                        CpswUtils_Prms *pUtilsPrms)
{

    pCpswCfg->aleConfig.policerGlobalConfig.policingEnable = TRUE;
    pCpswCfg->hostPortConfig.passPriorityTaggedUnchanged   = TRUE;

    pCpswCfg->aleConfig.modeFlags                             = CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE;
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask = 0x0;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask   = CpswTestCommon_getMacPortAleMask(taskObj);
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask  = (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK);
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode           = TRUE;
    pCpswCfg->aleConfig.vlanConfig.autoLearnWithVLAN          = FALSE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode          = FALSE;
    pCpswCfg->aleConfig.nwSecCfg.enableVid0Mode               = FALSE;

    pCpswCfg->vlanConfig.vlanAware           = FALSE;

    pCpswCfg->aleConfig.policerGlobalConfig.policingEnable     = FALSE;
    pCpswCfg->aleConfig.policerGlobalConfig.yellowDropEnable   = FALSE;
    pCpswCfg->aleConfig.policerGlobalConfig.redDropEnable      = FALSE;
    pCpswCfg->aleConfig.policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_GREEN;
}


