/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_cpts_event.c
 *
 * \brief    This file contains the cpts event test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Semaphore.h>

#include <ti/drv/uart/UART_stdio.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_cpts_event.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_TEST_CPTS_EVENT_MEM_COUNT                          (16U)
#define CPSW_TEST_CPTS_EVENT_TS_LOAD_VAL                        (90000000000U)
#define CPSW_TEST_CPTS_EVENT_TS_NUDGE                           (3U)
#define CPSW_TEST_CPTS_EVENT_ETH_TX_SEQ_ID                      (7600U)
#define CPSW_TEST_CPTS_EVENT_ETH_RX_SEQ_ID                      (7606U)
#define CPSW_TEST_CPTS_EVENT_HOST_TX_SEQ_ID                     (100U)
#define CPSW_TEST_CPTS_EVENT_HOST_TX_MSG_TYPE                   (1U)
#define CPSW_TEST_CPTS_EVENT_HOST_TX_DOMAIN                     (5U)
#define CPSW_TEST_CPTS_EVENT_GENF_COMPARE_VAL                   (95000000000U)
#define CPSW_TEST_CPTS_EVENT_GENF_COMPARE_VAL_2                 (98000000000U)
#define CPSW_TEST_CPTS_EVENT_GENF_LENGTH_VAL                    (1000000000U)
#define CPSW_TEST_CPTS_EVENT_HW_PUSH_INSTANCE_NUM               (CPSW_CPTS_HW_PUSH_1)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                         Function Declarations                              */
/* ========================================================================== */

void CptsTestCptsEvent_HwPushNotifyCallback(void *cbArg,
                                            CpswCpts_HwPush hwPushNum);

void Cpts_EventNotifyCallback(void *handle,
                              CpswCpts_Event *eventInfo);

static void CpswTestCptsEvent_setDefaultPortEventPrms(CpswMacPort_TsEventConfig *tsPortEventCfg);

static int32_t CpswTestCptsEvent_setTimeSyncRouter(Cpsw_Type cpswType);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswCpts_Event cptsEventMem[CPSW_TEST_CPTS_EVENT_MEM_COUNT];

uint8_t cptsPtpPkt[] = {0x00U, 0x02U, 0x00U, 0x2cU, 0x00U, 0x00U, 0x00U, 0x00U,
                        0x00U, 0x00U, 0x00U, 0x00U, 0x00U, 0x00U, 0x00U, 0x00U,
                        0x00U, 0x00U, 0x00U, 0x00U, 0xfcU, 0xafU, 0x6aU, 0xffU,
                        0xfeU, 0x02U, 0x52U, 0x15U, 0x00U, 0x01U, 0x1dU, 0xb0U,
                        0x00U, 0x00U, 0x00U, 0x00U, 0x00U, 0x06U, 0x94U, 0x01U,
                        0x02U, 0xc7U, 0x55U, 0xb7U};

Semaphore_Handle hPktRxSem;
volatile uint8_t hwPushCount[8U] = { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
void CpswTestCptsEvent_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                         Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestCptsEvent_setOpenPrms(CpswTestTaskObj *taskObj,
                                   Cpsw_Config *pCpswCfg,
                                   CpswOsal_Prms *pOsalPrms,
                                   CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->cptsConfig.hostRxTsEnable = true;
    pCpswCfg->cptsConfig.cptsRftClkFreq = CPSW_CPTS_RFTCLK_FREQ_500MHZ;
}

int32_t CpswTestCptsEvent_setTxPktInfo(CpswTestTaskObj *taskObj,
                                       uint32_t txChIndex,
                                       uint32_t pktNum,
                                       CpswDma_PktInfo *pktInfo,
                                       bool *testComplete)
{
    EthFrame *frame;
    /* Broadcast address */
    uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
    {
        0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU
    };
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    frame = (EthFrame *)pktInfo->bufPtr;
    memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
    memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
    frame->hdr.etherType = htons(CPSW_ETHERTYPE_PTP);
    memcpy(&frame->payload[0U], &cptsPtpPkt[0U], sizeof(cptsPtpPkt));
    /* Change Sequence Id for every packet */
    frame->payload[31U] += pktNum;
    pktInfo->userBufLen  = taskObj->stateObj.txChCfgInfo[txChIndex]->maxTxPktLen +
                           sizeof(EthFrameHeader);

    if (pktNum == 2U)
    {
        memset(&frame->payload[0U], 0xABU, sizeof(cptsPtpPkt));
        pktInfo->tsInfo.enableHostTxTs = true;
        pktInfo->tsInfo.txPktSeqId = CPSW_TEST_CPTS_EVENT_HOST_TX_SEQ_ID;
        pktInfo->tsInfo.txPktMsgType = CPSW_TEST_CPTS_EVENT_HOST_TX_MSG_TYPE;
        pktInfo->tsInfo.txPktDomain = CPSW_TEST_CPTS_EVENT_HOST_TX_DOMAIN;
    }

    pktInfo->appPriv = stateObj;
    *testComplete = FALSE;

    return CPSW_SOK;
}

int32_t CpswTestCptsEvent_Run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;
    uint64_t tsVal = 0U, tsVal1 = 0U, tsVal2 = 0U;
    Cpsw_IoctlPrms prms;
    CpswCpts_RegisterStackInArgs regStackInArgs;
    CpswMacPort_EnableTsEventInArgs enableTsEventInArgs;
    CpswCpts_LookUpEventInArgs lookupEventInArgs;
    CpswCpts_Event lookupEventOutArgs;
    CpswCpts_RegisterHwPushCbInArgs hwPushCbInArgs;
    CpswCpts_HwPush hwPushNum;
    CpswCpts_SetTsPpmInArgs setTsPpmInArgs;
    int32_t nudge = 0;
    uint64_t loadTsVal = 0U;
    CpswCpts_OutputBitSel tsSyncOutputSel;
    CpswCpts_SetFxnGenInArgs setGenFInArgs;
    Semaphore_Params semParams;

    CpswTestCommon_waitForPortLink(taskObj);
    CpswAppUtils_assert(taskObj->taskCfg->numTxCh == 1);
    CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);

    Semaphore_Params_init(&semParams);
    semParams.mode = Semaphore_Mode_BINARY;
    hPktRxSem = Semaphore_create(0, &semParams, NULL);
    CpswAppUtils_assert(NULL != hPktRxSem);

    /* Configure Timesync Router to enable Hardware push events */
    CpswTestCptsEvent_setTimeSyncRouter(taskObj->taskCfg->cpswType);

    regStackInArgs.pEventNotifyCb = Cpts_EventNotifyCallback;
    regStackInArgs.hEventNotifyCbArg = (void *)taskObj;
    /* Register a stack with call back */
    CPSW_IOCTL_SET_IN_ARGS(&prms, &regStackInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_REGISTER_STACK,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Software Time stamp Push event */
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_print("\nTimeStamp Value before Nudge: %lld \n", tsVal);

    /* TS Nudge */
    nudge = CPSW_TEST_CPTS_EVENT_TS_NUDGE;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &nudge);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SET_TS_NUDGE,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_print("TimeStamp Value after Nudge: %lld \n", tsVal);
    CpswAppUtils_print("TimeStamp Nudge successful \n");

    /* Load Time Stamp Test */
    loadTsVal = CPSW_TEST_CPTS_EVENT_TS_LOAD_VAL;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &loadTsVal);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_LOAD_TIMESTAMP,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_print("\nTimeStamp Value after Load TS: %lld \n", tsVal);
    CpswAppUtils_assert(tsVal > CPSW_TEST_CPTS_EVENT_TS_LOAD_VAL);
    CpswAppUtils_print("Load TimeStamp Value successful \n");

    /* TS PPM test */
    setTsPpmInArgs.tsPpmMode = CPSW_CPTS_PPM_PARTS_PER_MILLION;
    setTsPpmInArgs.tsPpmDir  = CPSW_CPTS_PPM_DIR_DECREASE;
    setTsPpmInArgs.tsPpmVal  = 500U;
    CpswAppUtils_print("\nTimeStamp PPM set to : %lld \n", setTsPpmInArgs.tsPpmVal);

    CPSW_IOCTL_SET_IN_ARGS(&prms, &setTsPpmInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SET_TS_PPM,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Check for Time stamps after enabling PPM */
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal1);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal2);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_print("TimeStamp Value 1: %lld \n", tsVal1);
    CpswAppUtils_print("TimeStamp Value 2: %lld \n", tsVal2);

    /* Set Ts PPM value back to zero */
    setTsPpmInArgs.tsPpmMode = CPSW_CPTS_PPM_DISABLE;
    setTsPpmInArgs.tsPpmVal  = 0U;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setTsPpmInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SET_TS_PPM,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal1);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CPSW_IOCTL_SET_OUT_ARGS(&prms, &tsVal2);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    CpswAppUtils_print("\nTimeStamp PPM is removed \n");
    CpswAppUtils_print("TimeStamp Value 1: %lld \n", tsVal1);
    CpswAppUtils_print("TimeStamp Value 2: %lld \n", tsVal2);

    /* Enable PTP over Bare Ethernet - Annex F (IEEE802.3) */
    enableTsEventInArgs.portNum = CPSW_MAC_PORT_FIRST;
    CpswTestCptsEvent_setDefaultPortEventPrms(&enableTsEventInArgs.portCfg);

    CPSW_IOCTL_SET_IN_ARGS(&prms, &enableTsEventInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswTestCommon_pktTxRx(taskObj);
    Semaphore_pend(hPktRxSem, BIOS_WAIT_FOREVER);

    /* Event Lookup for Tx Event */
    lookupEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT;
    lookupEventInArgs.hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
    lookupEventInArgs.msgType = CPSW_CPTS_MESSAGE_SYNC;
    lookupEventInArgs.portNum = CPSW_MAC_PORT_FIRST;
    lookupEventInArgs.seqId = CPSW_TEST_CPTS_EVENT_ETH_TX_SEQ_ID;
    lookupEventInArgs.domain = 0U;
    CpswAppUtils_print("\nTx Event Lookup for sequence Id: %d", lookupEventInArgs.seqId);
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &lookupEventInArgs, &lookupEventOutArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_LOOKUP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* Utilize Call back function to print Event Information */
    Cpts_EventNotifyCallback(NULL, &lookupEventOutArgs);

    /* Event Lookup for Rx Event */
    lookupEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_RECEIVE;
    lookupEventInArgs.hwPushNum = CPSW_CPTS_HW_PUSH_INVALID;
    lookupEventInArgs.msgType = CPSW_CPTS_MESSAGE_SYNC;
    lookupEventInArgs.portNum = CPSW_MAC_PORT_FIRST;
    lookupEventInArgs.seqId = CPSW_TEST_CPTS_EVENT_ETH_RX_SEQ_ID;
    lookupEventInArgs.domain = 0U;
    CpswAppUtils_print("\nRx Event Lookup for sequence Id: %d", lookupEventInArgs.seqId);
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &lookupEventInArgs, &lookupEventOutArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_LOOKUP_EVENT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* Utilize Call back function to print Event Information */
    Cpts_EventNotifyCallback(NULL, &lookupEventOutArgs);

    /* Register hardware push 1 callback */
    hwPushCbInArgs.hwPushNum = CPSW_TEST_CPTS_EVENT_HW_PUSH_INSTANCE_NUM;
    hwPushCbInArgs.pHwPushNotifyCb = CptsTestCptsEvent_HwPushNotifyCallback;
    hwPushCbInArgs.pHwPushNotifyCbArg = (void *)taskObj;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &hwPushCbInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_REGISTER_HW_PUSH_CALLBACK,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Enable TS_OUTPUT_BIT to get Hardware push events fired */
    tsSyncOutputSel = CPSW_CPTS_TS_OUTPUT_BIT_31;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &tsSyncOutputSel);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SELECT_TS_OUTPUT_BIT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    while (hwPushCount[0U] == 0U);
    while (hwPushCount[1U] == 0U);
    while (hwPushCount[2U] == 0U);
    while (hwPushCount[3U] == 0U);

    CpswAppUtils_print("\nHardware Push events test successful \n");

    /* Disable TS_OUTPUT_BIT to stop Hardware push events */
    tsSyncOutputSel = CPSW_CPTS_TS_OUTPUT_BIT_DISABLED;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &tsSyncOutputSel);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SELECT_TS_OUTPUT_BIT,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Enable GenF0 and GenF1 */

    /* Load Time Stamp to bring the time stamp value in the comparison value range */
    loadTsVal = CPSW_TEST_CPTS_EVENT_TS_LOAD_VAL;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &loadTsVal);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_LOAD_TIMESTAMP,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Configure GENF0  */
    setGenFInArgs.index   = 0U;
    setGenFInArgs.length  = CPSW_TEST_CPTS_EVENT_GENF_LENGTH_VAL;
    setGenFInArgs.compare = CPSW_TEST_CPTS_EVENT_GENF_COMPARE_VAL;
    setGenFInArgs.polarityInv = true;
    setGenFInArgs.ppmVal  = 0U;
    setGenFInArgs.ppmDir  = CPSW_CPTS_PPM_DIR_INCREASE;
    setGenFInArgs.ppmMode = CPSW_CPTS_PPM_DISABLE;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setGenFInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SET_GENF,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* Configure GENF1  */
    setGenFInArgs.index   = 1U;
    setGenFInArgs.length  = CPSW_TEST_CPTS_EVENT_GENF_LENGTH_VAL;
    setGenFInArgs.compare = CPSW_TEST_CPTS_EVENT_GENF_COMPARE_VAL_2;
    setGenFInArgs.polarityInv = true;
    setGenFInArgs.ppmVal  = 0U;
    setGenFInArgs.ppmDir  = CPSW_CPTS_PPM_DIR_INCREASE;
    setGenFInArgs.ppmMode = CPSW_CPTS_PPM_DISABLE;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setGenFInArgs);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_SET_GENF,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);
    while (hwPushCount[4U] == 0U);
    while (hwPushCount[5U] == 0U);
    while (hwPushCount[6U] == 0U);
    while (hwPushCount[7U] == 0U);

    /* Un-register hardware push call back */
    hwPushNum = CPSW_TEST_CPTS_EVENT_HW_PUSH_INSTANCE_NUM;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &hwPushNum);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_UNREGISTER_HW_PUSH_CALLBACK,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

    /* Print CPTS event statistics */
    CPSW_IOCTL_SET_NO_ARGS(&prms);
    status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                        taskObj->stateObj.coreId,
                        CPSW_CPTS_IOCTL_PRINT_STATS,
                        &prms);
    CpswAppUtils_assert(status == CPSW_SOK);

   /* Disable GenF0 and GenF1 */
   /* Disable GenF0 */
   setGenFInArgs.index  = 0U;
   setGenFInArgs.length = 0U;
   CPSW_IOCTL_SET_IN_ARGS(&prms, &setGenFInArgs);
   status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                       taskObj->stateObj.coreId,
                       CPSW_CPTS_IOCTL_SET_GENF,
                       &prms);
   /* Not supported is not an error, so set the status
    * to CPSW_SOK */
   if (status == CPSW_ENOTSUPPORTED)
   {
       status = CPSW_SOK;
   }
   CpswAppUtils_assert(status == CPSW_SOK);

   /* Disable GenF1 */
   setGenFInArgs.index  = 1U;
   setGenFInArgs.length = 0U;
   CPSW_IOCTL_SET_IN_ARGS(&prms, &setGenFInArgs);
   status = Cpsw_ioctl(taskObj->stateObj.hCpsw,
                       taskObj->stateObj.coreId,
                       CPSW_CPTS_IOCTL_SET_GENF,
                       &prms);
   /* Not supported is not an error, so set the status
    * to CPSW_SOK */
   if (status == CPSW_ENOTSUPPORTED)
   {
       status = CPSW_SOK;
   }
   CpswAppUtils_assert(status == CPSW_SOK);

    if (hPktRxSem != NULL)
    {
        Semaphore_delete(&hPktRxSem);
    }

   return status;
}

void CptsTestCptsEvent_HwPushNotifyCallback(void *cbArg, CpswCpts_HwPush hwPushNum)
{
    if (hwPushNum == CPSW_TEST_CPTS_EVENT_HW_PUSH_INSTANCE_NUM)
    {
        CpswAppUtils_print("\nHardware push notify callback test passed \n");
    }
    else
    {
        CpswAppUtils_print("\nHardware push notify callback test failed \n");
    }
}

void Cpts_EventNotifyCallback(void *handle,
                              CpswCpts_Event *eventInfo)
{
    static uint32_t ptpRxCount = 0U;

    /* Print Event Information */
    if (eventInfo != NULL)
    {
        CpswAppUtils_print("\n=================================\n");
        CpswAppUtils_print("    CPTS Event Information \n");
        CpswAppUtils_print("=================================\n");

        switch (eventInfo->eventType)
        {
            case CPSW_CPTS_EVENTTYPE_TS_PUSH:
            {
                CpswAppUtils_print("Event Type          : Software Time Stamp Push \n");
            }
            break;

            case CPSW_CPTS_EVENTTYPE_HW_TS_PUSH:
            {
                CpswAppUtils_print("Event Type          : Hardware Time Stamp Push \n");
                CpswAppUtils_print("Hardware Push Number: %d \n", eventInfo->hwPushNum);
                hwPushCount[CPSW_CPTS_HW_PUSH_NORMALIZE(eventInfo->hwPushNum)]++;
            }
            break;

            case CPSW_CPTS_EVENTTYPE_ETH_RECEIVE:
            {
                CpswAppUtils_print("Event Type          : Ethernet Receive \n");
                CpswAppUtils_print("Message Type        : %d \n", eventInfo->msgType);
                CpswAppUtils_print("Sequence Id         : %d \n", eventInfo->seqId);
                CpswAppUtils_print("Port Number         : %d \n", eventInfo->portNum);
                CpswAppUtils_print("Domain              : %d \n", eventInfo->domain);
                if (2U == ++ptpRxCount)
                {
                    Semaphore_post(hPktRxSem);
                }
            }
            break;

            case CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT:
            {
                CpswAppUtils_print("Event Type          : Ethernet Transmit \n");
                CpswAppUtils_print("Message Type        : %d \n", eventInfo->msgType);
                CpswAppUtils_print("Sequence Id         : %d \n", eventInfo->seqId);
                CpswAppUtils_print("Port Number         : %d \n", eventInfo->portNum);
                CpswAppUtils_print("Domain              : %d \n", eventInfo->domain);
            }
            break;

            case CPSW_CPTS_EVENTTYPE_TS_COMP:
            {
                CpswAppUtils_print("Event Type          : Time Stamp Compare \n");
            }
            break;

            case CPSW_CPTS_EVENTTYPE_TS_HOST_TX:
            {
                CpswAppUtils_print("Event Type          : Ethernet Host Transmit \n");
                CpswAppUtils_print("Message Type        : %d \n", eventInfo->msgType);
                CpswAppUtils_print("Sequence Id         : %d \n", eventInfo->seqId);
                CpswAppUtils_print("Domain              : %d \n", eventInfo->domain);
            }
            break;

            default:
            {
                CpswAppUtils_print("Event Type          : Invalid \n");
            }
            break;
        }

        /* Print Time stamp value */
        CpswAppUtils_print("Time Stamp Value    : %lld \n", eventInfo->tsVal);
    }
}

static void CpswTestCptsEvent_setDefaultPortEventPrms(CpswMacPort_TsEventConfig *tsPortEventCfg)
{
    tsPortEventCfg->commonPortIpConfig.enableTTLNonzero = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp107 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsIp129 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsIp130 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsIp131 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsIp132 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsPort319 = false;
    tsPortEventCfg->commonPortIpConfig.enableTsPort320 = false;
    tsPortEventCfg->commonPortIpConfig.enableUnicast = true;
    tsPortEventCfg->domainOffset = 4U;
    tsPortEventCfg->enableLType2 = false;
    tsPortEventCfg->enableRxAnnexD = true;
    tsPortEventCfg->enableRxAnnexE = true;
    tsPortEventCfg->enableRxAnnexF = true;
    tsPortEventCfg->enableTxAnnexD = true;
    tsPortEventCfg->enableTxAnnexE = true;
    tsPortEventCfg->enableTxAnnexF = true;
    tsPortEventCfg->enableTxHostTs = true;
    tsPortEventCfg->mcastType = 0U;
    tsPortEventCfg->messageType = 0xFFFFU;
    tsPortEventCfg->rxVlanType  = CPSW_MACPORT_VLAN_TYPE_NONE;
    tsPortEventCfg->seqIdOffset = 30U;
    tsPortEventCfg->txVlanType  = CPSW_MACPORT_VLAN_TYPE_NONE;
    tsPortEventCfg->vlanLType1  = 0U;
    tsPortEventCfg->vlanLType2  = 0U;
}

static int32_t CpswTestCptsEvent_setTimeSyncRouter(Cpsw_Type cpswType)
{
    int32_t status = CPSW_SOK;

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_SYNC_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW1_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_SYNC_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW2_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_SYNC_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW3_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_SYNC_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW4_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_GENF0_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW5_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_GENF0_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW6_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_GENF1_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW7_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswAppUtils_setTimeSyncRouter(cpswType,
                                            CSLR_TIMESYNC_INTRTR0_IN_CPSW0_CPTS_GENF1_0,
                                            CSLR_TIMESYNC_INTRTR0_OUTL_CPSW0_CPTS_HW8_PUSH_0);
    CpswAppUtils_assert(status == CPSW_SOK);

    return status;
}
