/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_switching_logic.c
 *
 * \brief    This file contains the cpsw_switching_logic test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_crcstrip.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
void CpswTestCrcStrip_setOpenPrms(CpswTestTaskObj *taskObj,
                                  Cpsw_Config *pCpswCfg,
                                  CpswOsal_Prms *pOsalPrms,
                                  CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->hostPortConfig.removeCrc = FALSE;
}

void CpswTestCrcStrip_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                        Cpsw_MacPort portNum)
{
    CpswAppUtils_setNoPhyCfgRgmii(&pLinkArgs->interface, &pLinkArgs->phyConfig);
    pLinkArgs->macConfig.enableLoopback = true;
    pLinkArgs->linkConfig.speed     = CPSW_SPEED_1GBIT;
    pLinkArgs->linkConfig.duplexity = CPSW_DUPLEX_FULL;
    
}

int32_t CpswTestCrcStrip_setTxPktInfo(CpswTestTaskObj *taskObj,
                                      uint32_t txChIndex,
                                      uint32_t pktNum,
                                      CpswDma_PktInfo *pktInfo,
                                      bool *testComplete)
{
    EthFrame *frame;
    /* Broadcast address */
    uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
    {
        0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU
    };
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    frame = (EthFrame *)pktInfo->bufPtr;
    memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
    memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
    frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);
    pktInfo->userBufLen  =
        taskObj->stateObj.txChCfgInfo[txChIndex]->maxTxPktLen +
        sizeof(EthFrameHeader);

    frame->payload[0U] = pktInfo->userBufLen & 0xFF;
    frame->payload[1U] = (pktInfo->userBufLen >> 8) & 0xFF;
    frame->payload[2U] = (pktInfo->userBufLen >> 16) & 0xFF;
    frame->payload[3U] = (pktInfo->userBufLen >> 24) & 0xFF;

    pktInfo->appPriv = stateObj;
    *testComplete    = FALSE;
    return CPSW_SOK;
}

int32_t CpswTestCrcStrip_processRxPkt(CpswTestTaskObj *taskObj,
                                      uint32_t rxFlowId,
                                      uint32_t pktNum,
                                      CpswDma_PktInfo *pktInfo,
                                      bool *testComplete)
{
    uint32_t txBufLength;
    EthFrame *frame;
    int32_t retVal = CPSW_SOK;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;

#ifdef ENABLE_PRINTFRAME
    CpswAppUtils_printFrame
        (frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif
    txBufLength  = frame->payload[0U];
    txBufLength |= (frame->payload[1U] & 0xFF) << 8;
    txBufLength |= (frame->payload[2U] & 0xFF) << 16;
    txBufLength |= (frame->payload[3U] & 0xFF) << 24;

    /* Rx packet should not have FCS (4 bytes stripped) */
    if ((txBufLength + 4) != pktInfo->userBufLen)
    {
        retVal = CPSW_EFAIL;
    }

    *testComplete = FALSE;
    return retVal;
}
