/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_default_priority.c
 *
 * \brief    This file contains the cpsw_test_default_priority test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>
#include <ti/drv/ipc/ipc.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_default_priority.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define ETH_TEST_MAC_PORT_VLAN_ID_BASE            (100U)
#define ETH_TEST_HOST_PORT_VLAN_ID                (200U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswTestDefaultPriority_setPolicerAleEntry(CpswTestTaskObj *taskObj,
                                                          uint32_t rxFlowId);

static int32_t CpswTestDefaultPriority_setRxDscpPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestDefaultPriority_setRxPriority(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestDefaultPriority_Run(CpswTestTaskObj *taskObj)
{
    int32_t status      = CPSW_SOK;
    uint32_t priority[] = {CPSWTESTDEFAULTPRIORITY1_RXFLOWID,
                           CPSWTESTDEFAULTPRIORITY2_RXFLOWID,
                           CPSWTESTDEFAULTPRIORITY3_RXFLOWID};
    int32_t i;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    Cpsw_IoctlPrms prms;
    CpswAle_SetDefaultThreadConfigInArgs setInArgs;
    CpswAle_GetDefaultThreadConfigOutArgs getOutArgs;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &getOutArgs);
    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                        CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG,
                        &prms);
    if (status == CPSW_SOK)
    {
        setInArgs.defaultThreadEnable         = TRUE;
        setInArgs.threadId                    = taskObj->stateObj.rxFlowObj[CPSWTESTDEFAULTPRIORITY_DEFAULTFLOWID].rxFlowIdx;
        setInArgs.enablePriorityOR            = TRUE;
        setInArgs.disableMacPortDefaultThread = FALSE;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &setInArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                            CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG,
                            &prms);
        if (status == CPSW_SOK)
        {
            status = CpswTestDefaultPriority_setRxPriority(taskObj);
        }

        if (status == CPSW_SOK)
        {
            status = CpswTestDefaultPriority_setRxDscpPriority(taskObj);
        }

        if (status == CPSW_SOK)
        {
            if (CPSW_2G == taskObj->taskCfg->cpswType)
            {
                for (i = 0; i < CPSW_UTILS_ARRAYSIZE(priority); i++)
                {
                    status = CpswTestDefaultPriority_setPolicerAleEntry(taskObj, priority[i]);
                    CpswAppUtils_assert(status == CPSW_SOK);
                }
            }

            status = CpswTestCommon_Run(taskObj);
        }
    }

    return status;
}

void CpswTestDefaultPriority_setOpenPrms(CpswTestTaskObj *taskObj,
                                         Cpsw_Config *pCpswCfg,
                                         CpswOsal_Prms *pOsalPrms,
                                         CpswUtils_Prms *pUtilsPrms)
{
    Cpsw_MacPort i;

    pCpswCfg->aleConfig.policerGlobalConfig.policingEnable = TRUE;
    pCpswCfg->hostPortConfig.passPriorityTaggedUnchanged   = TRUE;

    pCpswCfg->aleConfig.modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE | CPSW_ALE_CONFIG_MASK_UNKNOWN_UNICAST_FLOOD2HOST);
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask = 0x0;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask   = 0x0;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode           = TRUE;
    pCpswCfg->aleConfig.vlanConfig.autoLearnWithVLAN          = FALSE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode          = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.enableVid0Mode               = TRUE;

    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].learningCfg.noLearn              = FALSE;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].vlanCfg.dropUntagged             = FALSE;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.unregMcastFloodMask      = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.regMcastFloodMask        = CPSW_ALE_ALL_PORTS_MASK;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.forceUntaggedEgressMask  = 0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.noLearnMask              = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vidIngressCheck          = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.limitIPNxtHdr            = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.disallowIPFragmentation  = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanIdInfo.outerVlanFlag = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanIdInfo.vlanId        = ETH_TEST_HOST_PORT_VLAN_ID;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanMemberList           = CPSW_ALE_ALL_PORTS_MASK;

    for (i = CPSW_MAC_PORT_FIRST; i <= CPSW_MAC_PORT_LAST; i++)
    {
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].learningCfg.noLearn              = FALSE;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].vlanCfg.dropUntagged             = FALSE;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.unregMcastFloodMask      = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.regMcastFloodMask        = CPSW_ALE_ALL_PORTS_MASK;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.forceUntaggedEgressMask  = 0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.noLearnMask              = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vidIngressCheck          = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.limitIPNxtHdr            = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.disallowIPFragmentation  = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanIdInfo.outerVlanFlag = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanIdInfo.vlanId        = ETH_TEST_MAC_PORT_VLAN_ID_BASE + CPSW_NORMALIZE_MACPORT(i);
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanMemberList           = CPSW_ALE_ALL_PORTS_MASK;
    }

    pCpswCfg->hostPortConfig.vlanCfg.portPri = 7;
    pCpswCfg->hostPortConfig.vlanCfg.portCfi = 0;
    pCpswCfg->hostPortConfig.vlanCfg.portVID = ETH_TEST_HOST_PORT_VLAN_ID;
    pCpswCfg->vlanConfig.vlanAware           = TRUE;
    
    /* Update the resource usage so that priority OR results in direct mapping
     * to rxFlowIndex allocated starting with default flow.
     * For this the default rx flow index should be 8.
     * This requires the cores MCU_2_0 rxFlowIndex start at 7 as the first
     * rxFlowId is allocated to the reserved flow
     */
    CpswAppUtils_assert(pCpswCfg->resourceConfig.selfCoreId == IPC_MCU2_0);
    pCpswCfg->resourceConfig.resourcePartitionInfo.numCores = 2;

    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[0].numRxFlows = 7;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[0].coreId     = IPC_MPU1_0;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[0].numMacAddress = 1;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[0].numTxCh       = 4;

    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[1].numRxFlows = 5;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[1].coreId     = IPC_MCU2_0;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[1].numMacAddress = 1;
    pCpswCfg->resourceConfig.resourcePartitionInfo.coreDmaResourceInfo[1].numTxCh       = 4;
}

void CpswTestDefaultPriority_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                               Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback              = FALSE;
    pLinkArgs->macConfig.passPriorityTaggedUnchanged = TRUE;

    pLinkArgs->macConfig.vlanConfig.portPri = 7; // CPSW_NORMALIZE_MACPORT(portNum);
    pLinkArgs->macConfig.vlanConfig.portCfi = 0;
    pLinkArgs->macConfig.vlanConfig.portVID = ETH_TEST_MAC_PORT_VLAN_ID_BASE + CPSW_NORMALIZE_MACPORT(portNum);
}

static int32_t CpswTestDefaultPriority_setPolicerAleEntry(CpswTestTaskObj *taskObj,
                                                          uint32_t rxFlowId)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;

    CpswTestStateObj *stateObj = &taskObj->stateObj;

    setPolicerInArgs.policerMatch.policerMatchEnableMask = CPSW_ALE_POLICER_MATCH_PRIORITY;
    setPolicerInArgs.policerMatch.priority               = rxFlowId;
    setPolicerInArgs.threadIdEnable                      = TRUE;
    setPolicerInArgs.threadId                            = taskObj->stateObj.rxFlowObj[rxFlowId].rxFlowIdx;
    setPolicerInArgs.peakRateInBitsPerSec                = 0;
    setPolicerInArgs.commitRateInBitsPerSec              = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    setPolicerInArgs.policerMatch.policerMatchEnableMask = CPSW_ALE_POLICER_MATCH_PRIORITY;
    setPolicerInArgs.policerMatch.priority               = rxFlowId + 4;
    setPolicerInArgs.threadIdEnable                      = TRUE;
    setPolicerInArgs.threadId                            = taskObj->stateObj.rxFlowObj[rxFlowId].rxFlowIdx;
    setPolicerInArgs.peakRateInBitsPerSec                = 0;
    setPolicerInArgs.commitRateInBitsPerSec              = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    return status;
}

static int32_t CpswTestDefaultPriority_setRxDscpPriority(CpswTestTaskObj *taskObj)
{
    CpswPort_DscpPriority setHostInArgs;
    CpswPort_DscpPriority getHostOutArgs;
    Cpsw_IoctlPrms hostPrms;
    CpswMacPort_SetRxDscpPriorityInArgs setMacInArgs;
    CpswMacPort_GenericInArgs getMacInArgs;
    CpswPort_DscpPriority getMacOutArgs;
    Cpsw_IoctlPrms macPrms;
    uint32_t i;
    int32_t status;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t portIter;
    CpswTestMacPortList_t enabledPorts;

    setHostInArgs.enableDscpIPv4 = TRUE;
    setHostInArgs.enableDscpIPv6 = TRUE;
    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setHostInArgs.tosMap); i++)
    {
        setHostInArgs.tosMap[i] = (i % 8);
    }

    CPSW_IOCTL_SET_IN_ARGS(&hostPrms, &setHostInArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY,
                        &hostPrms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY: %d\n",
                           status);
    }

    CPSW_IOCTL_SET_OUT_ARGS(&hostPrms, &getHostOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY,
                        &hostPrms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY: %d\n",
                           status);
    }
    else
    {
        CpswAppUtils_assert(setHostInArgs.enableDscpIPv4 ==
                            getHostOutArgs.enableDscpIPv4);
        CpswAppUtils_assert(setHostInArgs.enableDscpIPv6 ==
                            getHostOutArgs.enableDscpIPv6);
        for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setHostInArgs.tosMap); i++)
        {
            CpswAppUtils_assert(setHostInArgs.tosMap[i] ==
                                getHostOutArgs.tosMap[i]);
        }
    }

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    /* Set Mac Port DSCP priority mapping for all mac ports */
    for (portIter = 0; portIter < enabledPorts.numMacPorts; portIter++)
    {
        setMacInArgs.portNum                     = enabledPorts.macPortList[portIter];
        setMacInArgs.dscpPriority.enableDscpIPv4 = TRUE;
        setMacInArgs.dscpPriority.enableDscpIPv6 = TRUE;
        for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setMacInArgs.dscpPriority.tosMap); i++)
        {
            setMacInArgs.dscpPriority.tosMap[i] = (i % taskObj->taskCfg->numRxFlow);
        }

        CPSW_IOCTL_SET_IN_ARGS(&macPrms, &setMacInArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY,
                            &macPrms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY: %d\n",
                               status);
        }

        getMacInArgs.portNum = enabledPorts.macPortList[portIter];
        CPSW_IOCTL_SET_INOUT_ARGS(&macPrms, &getMacInArgs, &getMacOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY,
                            &macPrms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY: %d\n",
                               status);
        }
        else
        {
            CpswAppUtils_assert(setMacInArgs.dscpPriority.enableDscpIPv4 ==
                                getMacOutArgs.enableDscpIPv4);
            CpswAppUtils_assert(setMacInArgs.dscpPriority.enableDscpIPv6 ==
                                getMacOutArgs.enableDscpIPv6);
            for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setMacInArgs.dscpPriority.tosMap); i++)
            {
                CpswAppUtils_assert(setMacInArgs.dscpPriority.tosMap[i] ==
                                    getMacOutArgs.tosMap[i]);
            }
        }
    }

    return status;
}

static int32_t CpswTestDefaultPriority_setRxPriority(CpswTestTaskObj *taskObj)
{
    CpswPort_RxPriority setHostInArgs;
    Cpsw_IoctlPrms hostPrms;
    CpswMacPort_SetRxPriorityInArgs setMacInArgs;
    Cpsw_IoctlPrms macPrms;
    uint32_t i;
    int32_t status;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t portIter;
    CpswTestMacPortList_t enabledPorts;

    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setHostInArgs.priorityMap); i++)
    {
        setHostInArgs.priorityMap[i] = i;
    }

    CPSW_IOCTL_SET_IN_ARGS(&hostPrms, &setHostInArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY,
                        &hostPrms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY: %d\n",
                           status);
    }

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    /* Set Mac Port DSCP priority mapping for all mac ports */
    for (portIter = 0; portIter < enabledPorts.numMacPorts; portIter++)
    {
        setMacInArgs.portNum = enabledPorts.macPortList[portIter];
        for (i = 0; i < CPSW_UTILS_ARRAYSIZE(setMacInArgs.rxPriority.priorityMap); i++)
        {
            setMacInArgs.rxPriority.priorityMap[i] = i;
        }

        CPSW_IOCTL_SET_IN_ARGS(&macPrms, &setMacInArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_MACPORT_IOCTL_SET_RX_PRIORITY,
                            &macPrms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_MACPORT_IOCTL_SET_RX_PRIORITY: %d\n",
                               status);
        }
    }

    return status;
}

int32_t CpswTestDefaultPriority_processRxPkt(CpswTestTaskObj *taskObj,
                                             uint32_t rxFlowCfgId,
                                             uint32_t pktNum,
                                             CpswDma_PktInfo *pktInfo,
                                             bool *testComplete)
{
    int32_t status = CPSW_SOK;
    uint32_t rxFlowId;
    EthVlanFrame *frame;
    uint32_t pcp;

    /* Consume the packet by just printing its content */
    frame = (EthVlanFrame *)pktInfo->bufPtr;

    pcp = (((htons(frame->hdr.tci)) >> 13) & 0x7);

    rxFlowId = taskObj->stateObj.rxFlowObj[rxFlowCfgId].rxFlowIdx;
    if ((rxFlowId & (CPSW_PRIORITY_NUM - 1)) == pcp)
    {
        status = CPSW_SOK;
    }
    else
    {
        status = CPSW_EFAIL;
    }

    *testComplete = FALSE;
    return status;
}

const uint8_t gCpswTestDefaultPriorityIPPkt[] =
{
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe8, 0x84, 0x4c, 0xeb, 0x64, 0x93, 0x08, 0x00, 0x45, 0x04,
    0x00, 0x28, 0x12, 0x34, 0x40, 0x00, 0xff, 0x06, 0x94, 0x97, 0xac, 0x18, 0xbe, 0x5e, 0xac, 0x18,
    0xbe, 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x64, 0x50, 0x00,
    0x0f, 0xa0, 0xca, 0x7c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
#define CPSW_TEST_DEFAULTPRIORITY_DSCP_OFFSET  (15)
#define CPSW_TEST_DEFAULTPRIORITY_TOS2DSCP(x)  ((x) << 2)
#define CPSW_TEST_DEFAULTPRIORITY_MAC_OFFSET   (11)
int32_t CpswTestDefaultPriority_setTxPktInfo(CpswTestTaskObj *taskObj,
                                             uint32_t txChIndex,
                                             uint32_t pktNum,
                                             CpswDma_PktInfo *pktInfo,
                                             bool *testComplete)
{
    memcpy(pktInfo->bufPtr, gCpswTestDefaultPriorityIPPkt, sizeof(gCpswTestDefaultPriorityIPPkt));
    pktInfo->bufPtr[CPSW_TEST_DEFAULTPRIORITY_DSCP_OFFSET] = CPSW_TEST_DEFAULTPRIORITY_TOS2DSCP(txChIndex);
    pktInfo->bufPtr[CPSW_TEST_DEFAULTPRIORITY_MAC_OFFSET]  = gCpswTestDefaultPriorityIPPkt[CPSW_TEST_DEFAULTPRIORITY_MAC_OFFSET] + txChIndex;

    pktInfo->userBufLen = sizeof(gCpswTestDefaultPriorityIPPkt);
    *testComplete       = FALSE;
    return CPSW_SOK;
}
