/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     main_tirtos.c
 *
 * \brief    Main file for TI-RTOS build.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Semaphore.h>

#include <ti/drv/sciclient/sciclient.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* Test application stack size */
#define APP_TSK_STACK_MAIN                    (10U * 1024U)
/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* Task function */
// static Void taskFxn(UArg a0, UArg a1);

static CpswApp_ClkHandle CpswUtApp_createClock(CpswTestTaskObj *taskObj);

static void CpswUtApp_delete(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Test application stack */
static uint8_t gAppTskStackTick[APP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
volatile uint32_t exitFlag = 0U;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static void CpswUt_testCreatePhyEvent(CpswTestTaskObj *taskObj)
{
    Event_Params evtPrms;

    Event_Params_init(&evtPrms);
    Event_construct(&taskObj->stateObj.phyEventObj, &evtPrms);
    taskObj->stateObj.phyEvent = Event_handle(&taskObj->stateObj.phyEventObj);
}

static void CpswUt_testDeletePhyEvent(CpswTestTaskObj *taskObj)
{
    Event_destruct(&taskObj->stateObj.phyEventObj);
    taskObj->stateObj.phyEvent = NULL;
}

static void CpswUt_testCreatePortEvent(CpswTestTaskObj *taskObj)
{
    Event_Params evtPrms;

    Event_Params_init(&evtPrms);
    Event_construct(&taskObj->stateObj.portEventObj, &evtPrms);
    taskObj->stateObj.portEvent = Event_handle(&taskObj->stateObj.portEventObj);
}

static void CpswUt_testDeletePortEvent(CpswTestTaskObj *taskObj)
{
    Event_destruct(&taskObj->stateObj.portEventObj);
    taskObj->stateObj.portEvent = NULL;
}

static void  CpswUt_initStateObj(CpswTestStateObj *stateObj)
{
    uint32_t i;

    memset(stateObj->rxFlowObj, 0, sizeof(stateObj->rxFlowObj));
    memset(stateObj->txChObj, 0, sizeof(stateObj->txChObj));
    stateObj->phyEvent          = NULL;
    stateObj->hCpsw             = NULL;
    stateObj->task_periodicTick = NULL;
    stateObj->hTimer            = NULL;
    stateObj->timerSem          = NULL;
    stateObj->coreId            = CpswAppSoc_getCoreId();
    stateObj->coreKey           = CPSW_RM_INVALIDCORE;
    stateObj->noPhyPortMask     = 0U;

    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(stateObj->rxFlowObj); i++)
    {
        CpswUtils_initQ(&stateObj->rxFlowObj[i].rxFreeQ);
        CpswUtils_initQ(&stateObj->rxFlowObj[i].rxReadyQ);
    }

    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(stateObj->txChObj); i++)
    {
        CpswUtils_initQ(&stateObj->txChObj[i].txFreePktInfoQ);
    }
}

int32_t CpswUt_testEntry(CpswTestTaskObj *taskObj)
{
    CpswApp_ClkHandle clkhandle;
    int retVal = CPSW_EFAIL;

    CpswUt_initStateObj(&taskObj->stateObj);
    CpswUt_testCreatePhyEvent(taskObj);
    CpswUt_testCreatePortEvent(taskObj);

    clkhandle = CpswUtApp_createClock(taskObj);
    if (clkhandle != NULL)
    {
        retVal = CpswTestCommon_Init(taskObj);
        if (CPSW_SOK == retVal)
        {
            int32_t deInitRetVal;

            /* Start timer */
            CpswUtApp_startClock(clkhandle);
            retVal = taskObj->taskCfg->runTest(taskObj);
            /* Stop timer */
            CpswUtApp_stopClock(clkhandle);
            if (NULL != taskObj->taskCfg->deInitTest)
            {
                deInitRetVal = taskObj->taskCfg->deInitTest(taskObj);
            }
            else
            {
                deInitRetVal = CpswTestCommon_DeInit(taskObj);
            }

            if (CPSW_SOK == retVal)
            {
                /* If test run was successful use the deinit retval as return status for test
                 * If test run failed , use the run fail status as return status for test
                 */
                retVal = deInitRetVal;
            }
        }
    }

    CpswUt_testDeletePhyEvent(taskObj);
    CpswUt_testDeletePortEvent(taskObj);
    CpswUtApp_delete(taskObj);

    return(retVal);
}

static Void CpswApp_timerCallback(UArg arg)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)arg;

    /* Tick! */
    Semaphore_post(timerSem);
}

static void CpswUtApp_periodicTick(UArg a0,
                                   UArg a1)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)a0;
    CpswTestTaskObj *taskObj  = (CpswTestTaskObj *)a1;

    while (!exitFlag)
    {
        Semaphore_pend(timerSem, BIOS_WAIT_FOREVER);
        /* Cpsw_periodicTick should be called from non-ISR context */
        Cpsw_periodicTick(CpswTestCommon_getCpswHandle(taskObj));
    }
}

static CpswApp_ClkHandle CpswUtApp_createClock(CpswTestTaskObj *taskObj)
{
    Task_Params params;
    Semaphore_Params semParams;
    Clock_Params clkParams;
    UInt period = 100U; /* msecs */
    Error_Block eb;

    Semaphore_Params_init(&semParams);
    semParams.mode             = Semaphore_Mode_COUNTING;
    taskObj->stateObj.timerSem = Semaphore_create(0, &semParams, NULL);

    /* Initialize the taskperiodicTick params. Set the task priority higher than
     * the
     * default priority (1) */
    Error_init(&eb);
    Task_Params_init(&params);
    params.priority       = 7U;
    params.stack          = gAppTskStackTick;
    params.stackSize      = sizeof(gAppTskStackTick);
    params.arg0           = (UArg)taskObj->stateObj.timerSem;
    params.arg1           = (UArg)taskObj;
    params.instance->name = "UT Periodic Tick";

    taskObj->stateObj.task_periodicTick =
        Task_create(CpswUtApp_periodicTick, &params,
                    &eb);
    CpswAppUtils_assert(Error_check(&eb) == FALSE);
    if (taskObj->stateObj.task_periodicTick == NULL)
    {
        BIOS_exit(0);
    }

    Clock_Params_init(&clkParams);
    clkParams.startFlag = FALSE;
    clkParams.period    = period;
    clkParams.arg       = (UArg)taskObj->stateObj.timerSem;

    /* Creating timer and setting timer callback function*/
    taskObj->stateObj.hTimer = Clock_create((Clock_FuncPtr) & CpswApp_timerCallback,
                                            period,
                                            &clkParams,
                                            NULL);
    if (taskObj->stateObj.hTimer != NULL)
    {
        /* Set timer expiry time in OS ticks */
        Clock_setTimeout(taskObj->stateObj.hTimer, period);
        Clock_setPeriod(taskObj->stateObj.hTimer, period);
    }
    else
    {
        CpswAppUtils_print("CpswApp_createClock() failed to create clock: %d\n");
    }

    return (CpswApp_ClkHandle)taskObj->stateObj.hTimer;
}

void CpswUtApp_startClock(CpswApp_ClkHandle handle)
{
    Clock_Handle clkHandle = (Clock_Handle)handle;

    /* Start timer */
    Clock_start(clkHandle);
}

void CpswUtApp_stopClock(CpswApp_ClkHandle handle)
{
    Clock_Handle clkHandle = (Clock_Handle)handle;

    /* Stop and delete the tick timer */
    Clock_stop(clkHandle);
}

void CpswUtApp_wait(uint32_t waitTime)
{
    Task_sleep(waitTime);
}

static void CpswUtApp_delete(CpswTestTaskObj *taskObj)
{
    if (taskObj->stateObj.hTimer != NULL)
    {
        Clock_delete(&taskObj->stateObj.hTimer);
    }

    if (taskObj->stateObj.timerSem != NULL)
    {
        Semaphore_delete(&taskObj->stateObj.timerSem);
    }

    if (taskObj->stateObj.task_periodicTick != NULL)
    {
        Task_delete(&taskObj->stateObj.task_periodicTick);
    }
}
