/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_switching_logic.c
 *
 * \brief    This file contains the cpsw_switching_logic test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_fifostats.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define CPSW_TEST_FIFO_STATS_NUM_STATS_ENTRY                             (512)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */
typedef struct CpswTestFifoStats_s
{
    uint32_t packetCount;
    CpswHostPort_FifoStats hostportFifoStats;
    CpswMacPort_FifoStats macportFifoStats;
} CpswTestFifoStats_t;

typedef struct CpswTestFifoStatsTable_s
{
    uint32_t tableIndex;
    CpswTestFifoStats_t CpswTestFifoStats[CPSW_TEST_FIFO_STATS_NUM_STATS_ENTRY];
} CpswTestFifoStatsTable_t;
/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswTestFifoStatsTable_t CpswTestFifoStatsTable;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
static void CpswTestFifoStats_getFifoStats(CpswTestTaskObj *taskObj,
                                           CpswTestFifoStatsTable_t *pfifoStatsTbl,
                                           uint32_t packetCount)
{
    int32_t status;
    Cpsw_IoctlPrms hostprms;
    Cpsw_IoctlPrms macprms;
    CpswMacPort_GenericInArgs macPortInArgs;

    if (pfifoStatsTbl->tableIndex < CPSW_UTILS_ARRAYSIZE(pfifoStatsTbl->CpswTestFifoStats))
    {
        CPSW_IOCTL_SET_OUT_ARGS(&hostprms, &pfifoStatsTbl->CpswTestFifoStats[pfifoStatsTbl->tableIndex].hostportFifoStats);

        status =
            Cpsw_ioctl(taskObj->stateObj.hCpsw, taskObj->stateObj.coreId, CPSW_HOSTPORT_IOCTL_GET_FIFO_STATS,
                       &hostprms);

        if (status == CPSW_SOK)
        {
        }

        macPortInArgs.portNum = CPSW_MAC_PORT_0;
        CPSW_IOCTL_SET_INOUT_ARGS(&macprms, &macPortInArgs, &pfifoStatsTbl->CpswTestFifoStats[pfifoStatsTbl->tableIndex].macportFifoStats);
        status =
            Cpsw_ioctl(taskObj->stateObj.hCpsw, taskObj->stateObj.coreId, CPSW_MACPORT_IOCTL_GET_FIFO_STATS,
                       &macprms);
        pfifoStatsTbl->CpswTestFifoStats[pfifoStatsTbl->tableIndex].packetCount = packetCount;
        pfifoStatsTbl->tableIndex++;
    }
}

static void CpswTestFifoStats_initStatsTable(CpswTestFifoStatsTable_t *pfifoStatsTbl)
{
    pfifoStatsTbl->tableIndex = 0;
}

void CpswTestFifoStats_printFifoStats(CpswTestFifoStatsTable_t *pfifoStatsTbl)
{
    uint32_t i;

    for (i = 0; i < pfifoStatsTbl->tableIndex; i++)
    {
        CpswAppUtils_print(" --------FIFO Stats at Packet #:%u------\n", pfifoStatsTbl->CpswTestFifoStats[i].packetCount);
        CpswAppUtils_print(" --------Host Port------\n");
        CpswAppUtils_print(" RxThroughputRate:%d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.rxThroughputRate);
        CpswAppUtils_print("FIFO Active status priority 0: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[0]);
        CpswAppUtils_print("FIFO Active status priority 1: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[1]);
        CpswAppUtils_print("FIFO Active status priority 2: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[2]);
        CpswAppUtils_print("FIFO Active status priority 3: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[3]);
        CpswAppUtils_print("FIFO Active status priority 4: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[4]);
        CpswAppUtils_print("FIFO Active status priority 5: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[5]);
        CpswAppUtils_print("FIFO Active status priority 6: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[6]);
        CpswAppUtils_print("FIFO Active status priority 7: %d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].hostportFifoStats.txActiveFifo[7]);

        CpswAppUtils_print(" \n\n--------Port %d------\n", CPSW_MAC_PORT_0 + 1);
        CpswAppUtils_print(" RxThroughputRate:%d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.rxThroughputRate);
        CpswAppUtils_print(" RxPreemptBlockCount:%d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.rxPreemptBlockCount);
        CpswAppUtils_print(" RxPreemptBlockCount:%d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.rxPreemptBlockCount);
        CpswAppUtils_print(" RxExpressBlockCount:%d \n",
                           pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.rxExpressBlockCount);
        CpswAppUtils_print(" TxBlockCount:%d \n", pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.txBlockCount);
        CpswAppUtils_print(" RxMaxBlocks:%d \n", pfifoStatsTbl->CpswTestFifoStats[i].macportFifoStats.rxMaxBlocks);
    }
}

int32_t CpswTestFifoStats_Run(CpswTestTaskObj *taskObj)
{
    int32_t status;

    CpswTestFifoStats_initStatsTable(&CpswTestFifoStatsTable);
    CpswAppUtils_print("\n\n----------------FIFO stats before RX TX---------------- \n");
    CpswTestFifoStats_getFifoStats(taskObj, &CpswTestFifoStatsTable, 0);
    CpswTestFifoStats_printFifoStats(&CpswTestFifoStatsTable);

    status = CpswTestCommon_Run(taskObj);
    if (status == CPSW_SOK)
    {
        CpswTestFifoStats_printFifoStats(&CpswTestFifoStatsTable);
    }

    return status;
}

int32_t CpswTestFifoStats_processRxPkt(CpswTestTaskObj *taskObj,
                                       uint32_t rxFlowId,
                                       uint32_t pktNum,
                                       CpswDma_PktInfo *pktInfo,
                                       bool *testComplete)
{
#ifdef ENABLE_PRINTFRAME
    EthFrame *frame;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;
    CpswAppUtils_printFrame
        (frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif

    if ((pktNum % 100) == 0)
    {
        CpswTestFifoStats_getFifoStats(taskObj, &CpswTestFifoStatsTable, pktNum);
    }

    *testComplete = FALSE;
    return CPSW_SOK;
}

void CpswTestFifoStats_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                         Cpsw_MacPort portNum)
{
    /* This test also exercises a different code path PHY present with mac loopback, hence we don'tableIndex
     * set phy config to NO PHY mode */
    pLinkArgs->interface.variantType = CPSW_MAC_VARIANT_FORCED;
    pLinkArgs->macConfig.enableLoopback = TRUE;
}
