/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_hostport_rxfilter.c
 *
 * \brief    This file contains the Host port rx filter test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_hostport_rxfilter.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static void CpswTestRxFilter_validateRxFilter(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Multicast address */
static uint8_t mcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x01U, 0xffU, 0xffU, 0xffU, 0xffU, 0x01U
};

/* nonRegisteredMcastAddr address */
static uint8_t nonRegisteredMcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x01U, 0xffU, 0x00U, 0x00U, 0xffU, 0x02U
};

/* nonRegisteredUcastAddr address */
static uint8_t nonRegisteredUcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x40U, 0xcdU, 0xefU, 0xfeU, 0xdcU, 0xbaU
};

/* Broadcast address */
static uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU
};

typedef struct CpswTestRxFilterPktCounter_s
{
    uint32_t directCnt;
    uint32_t bcastCnt;
    uint32_t mcastCnt;
    uint32_t mcastAllCnt;
    uint32_t promiscousCnt;
} CpswTestRxFilterPktCounter_t;

CpswTestRxFilterPktCounter_t CpswTestRxFilterPktCounter;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestRxFilter_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_IoctlPrms prms;
    CpswAle_HostRxFilterParams setRxFilterInArgs;
    uint32_t rxReadyCnt, timeoutVal, elapsedTIme;

    /* For this test overwrite board host port mac address with standard UT
     * default mac addr so that test packet generation does not change 
     * based on board
     */
    CpswTestCommon_getDefaultHostMacAddr(&stateObj->hostMacAddr[0U]);
    CpswTestRxFilter_validateRxFilter(taskObj);

    status = CpswTestCommon_setAleMulticastEntry(taskObj,
                                                 &mcastAddr[0],
                                                 0,  /* Vlan Id */
                                                 2,
                                                 (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK));

    CpswAppUtils_assert(status == CPSW_SOK);
    setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_NOTHING;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswTestCommon_waitForPortLink(taskObj);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        /* Discard all queued packets before entering timeout loop
         * There maybe packets queued during the validateRxFilter
         * Rx filter transition
         */

        CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);
        CpswTestCommon_discardRxPkts(taskObj, 0);

        timeoutVal  = 5000;
        elapsedTIme = 0;

        do
        {
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, 0);
            if (rxReadyCnt != 0)
            {
                status = CPSW_EFAIL;
                break;
            }
            else
            {
                CpswUtApp_wait(10);
                elapsedTIme += 10;
            }
        }
        while (elapsedTIme < timeoutVal);
    }

    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("Rx Filter direct test\n");
        setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_DIRECT;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                            &prms);
        if (status == CPSW_SOK)
        {
            memset(&CpswTestRxFilterPktCounter,0,sizeof(CpswTestRxFilterPktCounter));
            status = CpswTestCommon_pktTxRx(taskObj);
        }
    }

    return status;
}

static void CpswTestRxFilter_validateRxFilter(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_HostRxFilterParams setRxFilterInArgs;
    CpswAle_HostRxFilterParams getRxFilterOutArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    int32_t i;
    CPSW_ALE_RXFILTER_E rxFilterType[] = {CPSW_ALE_RXFILTER_NOTHING,
                                          CPSW_ALE_RXFILTER_DIRECT,
                                          CPSW_ALE_RXFILTER_BROADCAST,
                                          CPSW_ALE_RXFILTER_MULTICAST,
                                          CPSW_ALE_RXFILTER_ALLMULTICAST,
                                          CPSW_ALE_RXFILTER_ALL};

    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(rxFilterType); i++)
    {
        setRxFilterInArgs.rxFilter = rxFilterType[i];

        CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getRxFilterOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_GET_RX_FILTER,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        CpswAppUtils_assert(setRxFilterInArgs.rxFilter ==

                            getRxFilterOutArgs.rxFilter);
    }
}

void CpswTestRxFilter_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                        Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestRxFilter_setOpenPrms(CpswTestTaskObj *taskObj,
                                  Cpsw_Config *pCpswCfg,
                                  CpswOsal_Prms *pOsalPrms,
                                  CpswUtils_Prms *pUtilsPrms)
{
    // VLAN configurations
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode               = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanNoLearn             = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownForceUntaggedEgressMask = 0U;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask       = (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK);
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask     = (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK);
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask      = (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK);
}

int32_t CpswTestRxFilter_processRxPkt(CpswTestTaskObj *taskObj,
                                      uint32_t rxFlowId,
                                      uint32_t pktNum,
                                      CpswDma_PktInfo *pktInfo,
                                      bool *testComplete)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_IoctlPrms prms;
    CpswAle_HostRxFilterParams setRxFilterInArgs;
    CpswAle_HostRxFilterParams getRxFilterOutArgs;
    uint32_t numIgnBits = 2;
    uint32_t portMask   = (CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK);
    uint32_t vlanId     = 0;
    EthFrame *frame;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;

#ifdef ENABLE_PRINTFRAME
    CpswAppUtils_printFrame
        (frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif

    if (CpswTestRxFilterPktCounter.directCnt >= 1000)
    {
        // CpswAppUtils_print("Rx Filter Broadcast test\n");
        setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_BROADCAST;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                            &prms);
        if (status != CPSW_SOK)
        {
            status = CPSW_EFAIL;
        }

        CpswTestRxFilterPktCounter.directCnt = 0;
    }

    if (CpswTestRxFilterPktCounter.bcastCnt >= 1000)
    {
        // CpswAppUtils_print("Rx Filter registered mcast test\n");
        status = CpswTestCommon_setAleMulticastEntry(taskObj,
                                                     &mcastAddr[0],
                                                     vlanId,  /* Vlan Id */
                                                     numIgnBits,
                                                     portMask);
        if (CPSW_SOK == status)
        {
            setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_MULTICAST;
            CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);
            status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                                &prms);
            if (status != CPSW_SOK)
            {
                status = CPSW_EFAIL;
            }
        }
        else
        {
            status = CPSW_EFAIL;
        }

        CpswTestRxFilterPktCounter.bcastCnt = 0;
    }

    if (CpswTestRxFilterPktCounter.mcastCnt >= 1000)
    {
        // CpswAppUtils_print("Rx Filter all mcast test\n");
        setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_ALLMULTICAST;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                            &prms);
        if (status != CPSW_SOK)
        {
            status = CPSW_EFAIL;
        }

        CpswTestRxFilterPktCounter.mcastCnt = 0;
    }

    if (CpswTestRxFilterPktCounter.mcastAllCnt >= 1000)
    {
        // CpswAppUtils_print("Rx Filter promiscous mode test\n");
        setRxFilterInArgs.rxFilter = CPSW_ALE_RXFILTER_ALL;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setRxFilterInArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_RX_FILTER,
                            &prms);
        if (status != CPSW_SOK)
        {
            status = CPSW_EFAIL;
        }

        CpswTestRxFilterPktCounter.mcastAllCnt = 0;
    }

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &getRxFilterOutArgs);
    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_GET_RX_FILTER,
                        &prms);

    if (status == CPSW_SOK)
    {
        switch (getRxFilterOutArgs.rxFilter)
        {
            case CPSW_ALE_RXFILTER_DIRECT:
                if (memcmp(&frame->hdr.dstMac[0], &stateObj->hostMacAddr[0U], sizeof(frame->hdr.dstMac)) == 0)
                {
                    status = CPSW_SOK;
                    CpswTestRxFilterPktCounter.directCnt++;
                }
                else
                {
                    status = CPSW_EFAIL;
                }

                break;

            case CPSW_ALE_RXFILTER_BROADCAST:
                if ((memcmp(&frame->hdr.dstMac[0], &stateObj->hostMacAddr[0U], sizeof(frame->hdr.dstMac)) == 0) ||
                    (memcmp(&frame->hdr.dstMac[0], &bcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0))
                {
                    status = CPSW_SOK;
                    if (memcmp(&frame->hdr.dstMac[0], &bcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0)
                    {
                        CpswTestRxFilterPktCounter.bcastCnt++;
                    }
                }
                else
                {
                    status = CPSW_EFAIL;
                }

                break;

            case CPSW_ALE_RXFILTER_MULTICAST:
                if ((memcmp(&frame->hdr.dstMac[0], &stateObj->hostMacAddr[0U], sizeof(frame->hdr.dstMac)) == 0) ||
                    (memcmp(&frame->hdr.dstMac[0], &bcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0) ||
                    (memcmp(&frame->hdr.dstMac[0], &mcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0))
                {
                    status = CPSW_SOK;
                    if (memcmp(&frame->hdr.dstMac[0], &mcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0)
                    {
                        CpswTestRxFilterPktCounter.mcastCnt++;
                    }
                }
                else
                {
                    status = CPSW_EFAIL;
                }

                break;

            case CPSW_ALE_RXFILTER_ALLMULTICAST:
                if ((memcmp(&frame->hdr.dstMac[0], &stateObj->hostMacAddr[0U], sizeof(frame->hdr.dstMac)) == 0) ||
                    ((frame->hdr.dstMac[0] & 0x1) == 0x1))
                {
                    if (memcmp(&frame->hdr.dstMac[0], &nonRegisteredMcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0)
                    {
                        CpswTestRxFilterPktCounter.mcastAllCnt++;
                    }

                    status = CPSW_SOK;
                }
                else
                {
                    status = CPSW_EFAIL;
                }

                break;

            case CPSW_ALE_RXFILTER_ALL:
                /* For Rx filter ALL it is in promiscous mode and will receive all packets */
                status = CPSW_SOK;
                if (memcmp(&frame->hdr.dstMac[0], &nonRegisteredUcastAddr[0U], sizeof(frame->hdr.dstMac)) == 0)
                {
                    CpswTestRxFilterPktCounter.promiscousCnt++;
                }

                break;

            default:
                status = CPSW_EFAIL;
                break;
        }
    }

    if (CpswTestRxFilterPktCounter.promiscousCnt >= 1000)
    {
        *testComplete = TRUE;
    }
    else
    {
        *testComplete = FALSE;
    }

    return status;
}
