/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_policer_nomatch.c
 *
 * \brief    This file contains the cpsw_policer nomatch test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_intervlan.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM (CPSW_MAC_PORT_3)
#define CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM  (CPSW_MAC_PORT_2)

#define CPSW_TEST_INTERVLAN_INGRESS_VLANID    (100)
#define CPSW_TEST_INTERVLAN_EGRESS_VLANID     (200)
#define CPSW_TEST_INTERVLAN_HOSTPORT_PVID     (300)
#define CPSW_TEST_INTERVLAN_MACPORT_PVID_BASE (400)
#define CPSW_TEST_INTERVLAN_INTRA_VLANID      (1000)
#define CPSW_TEST_INTERVLAN_INTRA_VLAN_EGRESS_STRIP_ID      (1200)

#define CPSW_TEST_INTERVLAN_IPV6_HOP_LIMIT_OFFSET (7)
#define CPSW_TEST_INTERVLAN_IPV4_TTL_OFFSET       (8)
#define CPSW_TEST_INTERVLAN_IPV6_ETHERTYPE (0x86DD)
#define CPSW_TEST_INTERVLAN_IPV4_ETHERTYPE (0x0800)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswTestInterVlan_setInterVlanMultiEgress(CpswTestTaskObj *taskObj,
                                                         CpswMacPort_InterVLANRouteId expectedAllocRouteId,
                                                         uint32_t *pNumRoutesUsed);

static int32_t CpswTestInterVlan_setInterVlanUniEgress(CpswTestTaskObj *taskObj,
                                                       CpswMacPort_InterVLANRouteId expectedAllocRouteId,
                                                       uint32_t *pNumRoutesUsed);

static int32_t CpswTestInterVlan_setShortIPG(CpswTestTaskObj *taskObj);

static int32_t CpswTestInterVlan_setBlacklistHostportClassifier(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
#define CPSW_TEST_IPV6_OCTET2ARRAY(x)  0x00, x
#define CPSW_TEST_IPV6_HEXTET2ARRAY(x) (x & 0xFF00) >> 8, (x & 0xFF)

uint8_t testDstIpv6Addr[16] = {CPSW_TEST_IPV6_HEXTET2ARRAY(0x2000),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x2),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                               CPSW_TEST_IPV6_OCTET2ARRAY(0x2)};
uint8_t testDstIpv6Addr2[16] = {CPSW_TEST_IPV6_HEXTET2ARRAY(0x2000),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x1),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x4),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                CPSW_TEST_IPV6_OCTET2ARRAY(0x2)};
uint8_t testDstIpv6AddrMcast[] = {CPSW_TEST_IPV6_HEXTET2ARRAY(0x2FFF),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x1),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x4),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                  CPSW_TEST_IPV6_OCTET2ARRAY(0x2)};

static uint8_t testDstMacAddr[] = {0x00, 0x11, 0x02, 0x00, 0x00, 0x01};
static uint8_t testDstMcastMacAddr[] = {0x01, 0x00, 0x00, 0x01, 0x02, 0x08};

static uint8_t testHostMacAddr[] = {0x02, 0x00, 0x00, 0x00, 0x00, 0x02};

static uint8_t testSrcIpv6Addr[16] = {CPSW_TEST_IPV6_HEXTET2ARRAY(0x2000),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x1),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x0),
                                      CPSW_TEST_IPV6_OCTET2ARRAY(0x2)};
static uint8_t testSrcIpv4Addr[4] = {172,
                                     24,
                                     106,
                                     128};

static uint8_t testDstIpv4Addr[4] = {172,
                                     24,
                                     108,
                                     128};

static uint8_t testSrcMacAddr[] = {0x00, 0x11, 0x01, 0x00, 0x00, 0x01};
static uint8_t testBlackListSrcMacAddr[] = {0x00, 0x11, 0x01, 0x00, 0x00, 0x03};

static Cpsw_MacPort testMultiEgressPortList[] = {CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM, CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestInterVlan_Run(CpswTestTaskObj *taskObj)
{
    int32_t status              = CPSW_SOK;
    uint32_t numRoutesAllocated = 0;
    CpswTestStateObj *stateObj  = &taskObj->stateObj;

    status = CpswTestInterVlan_setShortIPG(taskObj);

    if (CPSW_SOK == status)
    {
        status = CpswTestInterVlan_setInterVlanUniEgress(taskObj, CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST, &numRoutesAllocated);
    }

    if (CPSW_SOK == status)
    {
        /* CpswTestInterVlan_setInterVlanUniEgress will use two routes */
        status = CpswTestInterVlan_setInterVlanMultiEgress(taskObj, (CpswMacPort_InterVLANRouteId)(CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST + numRoutesAllocated), &numRoutesAllocated);
    }

    if (CPSW_SOK == status)
    {
        status = CpswTestInterVlan_setBlacklistHostportClassifier(taskObj);
    }

    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES,
                            &prms);

        CpswAppUtils_assert(status == CPSW_SOK);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_PRINT_REGISTERS,
                            &prms);

        CpswAppUtils_assert(status == CPSW_SOK);

        status = CpswTestCommon_Run(taskObj);
    }

    return status;
}

void CpswTestInterVlan_setOpenPrms(CpswTestTaskObj *taskObj,
                                   Cpsw_Config *pCpswCfg,
                                   CpswOsal_Prms *pOsalPrms,
                                   CpswUtils_Prms *pUtilsPrms)
{
    Cpsw_MacPort i;

    /* pCpswCfg->aleConfig.policerGlobalConfig.policingEnable SHOULD BE TRUE for interVLan.
     * Set to FALSE to exercise driver internal logic to auto enable policer when interVLan API
     * is invoked
     */
    pCpswCfg->aleConfig.policerGlobalConfig.policingEnable = FALSE;

    pCpswCfg->aleConfig.policerGlobalConfig.redDropEnable      = FALSE;
    pCpswCfg->aleConfig.policerGlobalConfig.yellowDropEnable   = FALSE;
    pCpswCfg->aleConfig.policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_GREEN;
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode            = TRUE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode           = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.enableVid0Mode                = TRUE;

    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].learningCfg.noLearn              = FALSE;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].vlanCfg.dropUntagged             = FALSE;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.unregMcastFloodMask      = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.regMcastFloodMask        = CPSW_ALE_ALL_PORTS_MASK;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.forceUntaggedEgressMask  = 0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.noLearnMask              = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vidIngressCheck          = 0x0;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.limitIPNxtHdr            = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.disallowIPFragmentation  = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanIdInfo.outerVlanFlag = false;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_HOSTPORT_PVID;
    pCpswCfg->aleConfig.portCfg[CPSW_ALE_HOST_PORT_NUM].pvidCfg.vlanMemberList           = CPSW_ALE_ALL_PORTS_MASK;

    for (i = CPSW_MAC_PORT_0; i <= CPSW_MAC_PORT_LAST; i++)
    {
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].learningCfg.noLearn              = FALSE;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].vlanCfg.dropUntagged             = FALSE;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.unregMcastFloodMask      = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.regMcastFloodMask        = CPSW_ALE_ALL_PORTS_MASK;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.forceUntaggedEgressMask  = 0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.noLearnMask              = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vidIngressCheck          = 0x0;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.limitIPNxtHdr            = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.disallowIPFragmentation  = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanIdInfo.outerVlanFlag = false;
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_MACPORT_PVID_BASE + CPSW_NORMALIZE_MACPORT(i);
        pCpswCfg->aleConfig.portCfg[CPSW_ALE_MACPORT_TO_ALEPORT(i)].pvidCfg.vlanMemberList           = CPSW_ALE_ALL_PORTS_MASK;
    }

    pCpswCfg->hostPortConfig.vlanCfg.portPri = 7;
    pCpswCfg->hostPortConfig.vlanCfg.portCfi = 0;
    pCpswCfg->hostPortConfig.vlanCfg.portVID = CPSW_TEST_INTERVLAN_HOSTPORT_PVID;
    pCpswCfg->vlanConfig.vlanAware           = TRUE;
}

void CpswTestInterVlan_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                         Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;

    pLinkArgs->macConfig.vlanConfig.portPri = CPSW_NORMALIZE_MACPORT(portNum);
    pLinkArgs->macConfig.vlanConfig.portCfi = 0;
    pLinkArgs->macConfig.vlanConfig.portVID = CPSW_TEST_INTERVLAN_MACPORT_PVID_BASE + CPSW_NORMALIZE_MACPORT(portNum);
}

static uint32_t CpswTestInterVlan_getIngressVlanMembershipMask(void)
{
    uint32_t memberShipMask;

    memberShipMask =
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
    memberShipMask |= CPSW_ALE_HOST_PORT_MASK;
    return memberShipMask;
}

static uint32_t CpswTestInterVlan_getEgressVlanMembershipMask(void)
{
    uint32_t memberShipMask;

    memberShipMask =
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM));
    memberShipMask |= CPSW_ALE_HOST_PORT_MASK;
    return memberShipMask;
}

static uint32_t CpswTestInterVlan_getIntraVlanMembershipMask(void)
{
    uint32_t memberShipMask;

    memberShipMask =
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
    memberShipMask |=
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM));
    memberShipMask |= CPSW_ALE_HOST_PORT_MASK;
    return memberShipMask;
}

static int32_t CpswTestInterVlan_addUniEgressAleTableEntries(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setUcastInArgs.addr.addr[0U], testHostMacAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0;
    setUcastInArgs.info.portNum = CPSW_ALE_HOST_PORT_NUM;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("%s() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           __func__, status);
    }

    if (status == CPSW_SOK)
    {
        CpswAle_VlanEntryInfo inArgs;
        CpswAle_AddEntryOutArgs outArgs;

        inArgs.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_INGRESS_VLANID;
        inArgs.vlanIdInfo.outerVlanFlag = false;
        inArgs.vlanMemberList           = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.unregMcastFloodMask      = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.regMcastFloodMask        = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.forceUntaggedEgressMask  = 0U;
        inArgs.noLearnMask              = 0U;
        inArgs.vidIngressCheck          = false;
        inArgs.limitIPNxtHdr            = false;
        inArgs.disallowIPFragmentation  = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                               __func__, status);
        }
    }

    if (status == CPSW_SOK)
    {
        CpswAle_VlanEntryInfo inArgs;
        CpswAle_AddEntryOutArgs outArgs;

        inArgs.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_EGRESS_VLANID;
        inArgs.vlanIdInfo.outerVlanFlag = false;
        inArgs.vlanMemberList           = CpswTestInterVlan_getEgressVlanMembershipMask();
        inArgs.unregMcastFloodMask      = CpswTestInterVlan_getEgressVlanMembershipMask();
        inArgs.regMcastFloodMask        = CpswTestInterVlan_getEgressVlanMembershipMask();
        inArgs.forceUntaggedEgressMask  = 0U;
        inArgs.noLearnMask              = 0U;
        inArgs.vidIngressCheck          = false;
        inArgs.limitIPNxtHdr            = false;
        inArgs.disallowIPFragmentation  = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                               __func__, status);
        }
    }

    return status;
}

static int32_t CpswTestInterVlan_setInterVlanUniEgress(CpswTestTaskObj *taskObj,
                                                       CpswMacPort_InterVLANRouteId expectedAllocRouteId,
                                                       uint32_t *pNumRoutesUsed)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    Cpsw_SetInterVlanRouteUniEgressInArgs inArgs;
    Cpsw_SetInterVlanRouteUniEgressOutArgs outArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    *pNumRoutesUsed = 0;
    status          = CpswTestInterVlan_addUniEgressAleTableEntries(taskObj);

    if (CPSW_SOK == status)
    {
        /* Set to invalid id and confirm outArgs populated correctly after IOCTL
         * called
         */
        outArgs.egressPortRouteId = CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;

        inArgs.inPktMatchCfg.packetMatchEnableMask = 0;

        inArgs.inPktMatchCfg.ingressPort            = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT;

        inArgs.inPktMatchCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.dstIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr, testDstIpv6Addr, sizeof(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST;

        memcpy(inArgs.inPktMatchCfg.dstMacAddr.addr.addr, testHostMacAddr, sizeof(inArgs.inPktMatchCfg.dstMacAddr.addr.addr));
        inArgs.inPktMatchCfg.dstMacAddr.addr.vlanId   = 0;
        inArgs.inPktMatchCfg.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM);

        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST;

        inArgs.inPktMatchCfg.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.srcIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr, testSrcIpv6Addr, sizeof(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC;

        memcpy(inArgs.inPktMatchCfg.srcMacAddr.addr.addr, testSrcMacAddr, sizeof(inArgs.inPktMatchCfg.srcMacAddr.addr.addr));
        inArgs.inPktMatchCfg.srcMacAddr.addr.vlanId    = 0;
        inArgs.inPktMatchCfg.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM);
        inArgs.inPktMatchCfg.packetMatchEnableMask    |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC;

        inArgs.inPktMatchCfg.etherType.etherType    = CPSW_TEST_INTERVLAN_IPV6_ETHERTYPE;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE;

        inArgs.inPktMatchCfg.vlanId         = CPSW_TEST_INTERVLAN_INGRESS_VLANID;
        inArgs.inPktMatchCfg.enableTTLCheck = TRUE;

        inArgs.egressCfg.egressPort                       = CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM;
        inArgs.egressCfg.outPktModCfg.decrementTTL        = TRUE;
        inArgs.egressCfg.outPktModCfg.forceUntaggedEgress = FALSE;
        inArgs.egressCfg.outPktModCfg.replaceDASA         = TRUE;
        memcpy(inArgs.egressCfg.outPktModCfg.srcAddr, testHostMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.srcAddr));
        memcpy(inArgs.egressCfg.outPktModCfg.dstAddr, testDstMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.dstAddr));
        inArgs.egressCfg.outPktModCfg.vlanId = CPSW_TEST_INTERVLAN_EGRESS_VLANID;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestInterVlan_setInterVlanUniEgress() failed CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        *pNumRoutesUsed += 1;
        CpswAppUtils_assert(outArgs.egressPortRouteId == expectedAllocRouteId);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.port == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.portIsTrunk == FALSE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpEnabled == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpcode == (1 + (outArgs.egressPortRouteId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.enableTTLCheck == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.destPortMask == (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.policerMatchEnableMask == (CPSW_ALE_POLICER_MATCH_PORT |
                                                                                           CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_MACDST |
                                                                                           CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                                           CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                                                                           CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_IPDST));
    }

    if (status == CPSW_SOK)
    {
        /* Add another route to send packet out on ingress port.
         * Only destIP changes. Rest of params remain same
         */
        inArgs.inPktMatchCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.dstIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr, testDstIpv6Addr2, sizeof(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST;

        inArgs.egressCfg.egressPort = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS,
                            &prms);

        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestInterVlan_setInterVlanUniEgress() failed CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        /* Do not increment numRoutesUsed as it is a different egress port */
        CpswAppUtils_assert(outArgs.egressPortRouteId == expectedAllocRouteId);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.port == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.portIsTrunk == FALSE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpEnabled == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpcode == (1 + (outArgs.egressPortRouteId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.enableTTLCheck == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.destPortMask == (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.policerMatchEnableMask == (CPSW_ALE_POLICER_MATCH_PORT |
                                                                                           CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_MACDST |
                                                                                           CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                                           CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                                                                           CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_IPDST));
    }

    /* Add a route in opposite direction to enable bidirectional intervlan switching */
    if (CPSW_SOK == status)
    {
        /* Set to invalid id and confirm outArgs populated correctly after IOCTL
         * called
         */
        outArgs.egressPortRouteId = CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;

        inArgs.inPktMatchCfg.packetMatchEnableMask = 0;

        inArgs.inPktMatchCfg.ingressPort            = CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT;

        inArgs.inPktMatchCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.dstIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr, testSrcIpv6Addr, sizeof(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST;

        memcpy(inArgs.inPktMatchCfg.dstMacAddr.addr.addr, testHostMacAddr, sizeof(inArgs.inPktMatchCfg.dstMacAddr.addr.addr));
        inArgs.inPktMatchCfg.dstMacAddr.addr.vlanId   = 0;
        inArgs.inPktMatchCfg.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM);

        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST;

        inArgs.inPktMatchCfg.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.srcIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr, testDstIpv6Addr, sizeof(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC;

        memcpy(inArgs.inPktMatchCfg.srcMacAddr.addr.addr, testDstMacAddr, sizeof(inArgs.inPktMatchCfg.srcMacAddr.addr.addr));
        inArgs.inPktMatchCfg.srcMacAddr.addr.vlanId    = 0;
        inArgs.inPktMatchCfg.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM);
        inArgs.inPktMatchCfg.packetMatchEnableMask    |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC;

        inArgs.inPktMatchCfg.etherType.etherType    = 0x86DD;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE;

        inArgs.inPktMatchCfg.vlanId         = CPSW_TEST_INTERVLAN_EGRESS_VLANID;
        inArgs.inPktMatchCfg.enableTTLCheck = TRUE;

        inArgs.egressCfg.egressPort                       = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;
        inArgs.egressCfg.outPktModCfg.decrementTTL        = TRUE;
        inArgs.egressCfg.outPktModCfg.forceUntaggedEgress = FALSE;
        inArgs.egressCfg.outPktModCfg.replaceDASA         = TRUE;
        memcpy(inArgs.egressCfg.outPktModCfg.srcAddr, testHostMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.srcAddr));
        memcpy(inArgs.egressCfg.outPktModCfg.dstAddr, testSrcMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.dstAddr));
        inArgs.egressCfg.outPktModCfg.vlanId = CPSW_TEST_INTERVLAN_INGRESS_VLANID;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestInterVlan_setInterVlanUniEgress() failed CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        CpswAppUtils_assert(outArgs.egressPortRouteId == (expectedAllocRouteId + 1));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.port == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.portIsTrunk == FALSE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpEnabled == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpcode == (1 + (outArgs.egressPortRouteId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.enableTTLCheck == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.destPortMask == (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.policerMatchEnableMask == (CPSW_ALE_POLICER_MATCH_PORT |
                                                                                           CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_MACDST |
                                                                                           CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                                           CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                                                                           CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_IPDST));
    }

    /* Add IPv4 intervlan switching route */
    if (CPSW_SOK == status)
    {
        /* Set to invalid id and confirm outArgs populated correctly after IOCTL
         * called
         */
        outArgs.egressPortRouteId = CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;

        inArgs.inPktMatchCfg.packetMatchEnableMask = 0;

        inArgs.inPktMatchCfg.ingressPort            = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT;

        inArgs.inPktMatchCfg.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
        inArgs.inPktMatchCfg.srcIp.ipv4.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.srcIp.ipv4.ipv4Addr, testSrcIpv4Addr, sizeof(inArgs.inPktMatchCfg.srcIp.ipv4.ipv4Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC;

        inArgs.inPktMatchCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
        inArgs.inPktMatchCfg.dstIp.ipv4.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.dstIp.ipv4.ipv4Addr, testDstIpv4Addr, sizeof(inArgs.inPktMatchCfg.dstIp.ipv4.ipv4Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST;

        memcpy(inArgs.inPktMatchCfg.dstMacAddr.addr.addr, testHostMacAddr, sizeof(inArgs.inPktMatchCfg.dstMacAddr.addr.addr));
        inArgs.inPktMatchCfg.dstMacAddr.addr.vlanId   = 0;
        inArgs.inPktMatchCfg.dstMacAddr.egressPortNum = CPSW_ALE_HOST_PORT_NUM;

        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST;

        memcpy(inArgs.inPktMatchCfg.srcMacAddr.addr.addr, testSrcMacAddr, sizeof(inArgs.inPktMatchCfg.srcMacAddr.addr.addr));
        inArgs.inPktMatchCfg.srcMacAddr.addr.vlanId    = 0;
        inArgs.inPktMatchCfg.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM);
        inArgs.inPktMatchCfg.packetMatchEnableMask    |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC;

        inArgs.inPktMatchCfg.vlanId         = CPSW_TEST_INTERVLAN_INGRESS_VLANID;
        inArgs.inPktMatchCfg.enableTTLCheck = TRUE;

        inArgs.egressCfg.egressPort                       = CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM;
        inArgs.egressCfg.outPktModCfg.decrementTTL        = TRUE;
        inArgs.egressCfg.outPktModCfg.forceUntaggedEgress = TRUE;
        inArgs.egressCfg.outPktModCfg.replaceDASA         = TRUE;
        memcpy(inArgs.egressCfg.outPktModCfg.srcAddr, testHostMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.srcAddr));
        memcpy(inArgs.egressCfg.outPktModCfg.dstAddr, testDstMacAddr, sizeof(inArgs.egressCfg.outPktModCfg.dstAddr));
        inArgs.egressCfg.outPktModCfg.vlanId = 0;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestInterVlan_setInterVlanUniEgress() failed CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        *pNumRoutesUsed += 1;
        CpswAppUtils_assert(outArgs.egressPortRouteId == (expectedAllocRouteId + 1));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.port == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.portIsTrunk == FALSE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpEnabled == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpcode == (1 + (outArgs.egressPortRouteId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.enableTTLCheck == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.destPortMask == (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.policerMatchEnableMask == (CPSW_ALE_POLICER_MATCH_PORT |
                                                                                           CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_MACDST |
                                                                                           CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                                           CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_IPDST));
    }

    return status;
}

static uint32_t CpswTestInterVlan_getMultiEgressDestPortMask(void)
{
    uint32_t i;
    uint32_t destPortMask;

    destPortMask = 0;
    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(testMultiEgressPortList); i++)
    {
        destPortMask |= (1 << CPSW_ALE_MACPORT_TO_ALEPORT(testMultiEgressPortList[i]));
    }

    return destPortMask;
}

static int32_t CpswTestInterVlan_addMultiEgressAleTableEntries(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswAle_VlanEntryInfo inArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], testHostMacAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0;
    setUcastInArgs.info.portNum = CPSW_ALE_HOST_PORT_NUM;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("%s() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           __func__, status);
    }

    if (status == CPSW_SOK)
    {
        CpswAle_AddEntryOutArgs outArgs;

        inArgs.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_INGRESS_VLANID;
        inArgs.vlanIdInfo.outerVlanFlag = false;
        inArgs.vlanMemberList           = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.unregMcastFloodMask      = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.regMcastFloodMask        = CpswTestInterVlan_getIngressVlanMembershipMask();
        inArgs.forceUntaggedEgressMask  = 0U;
        inArgs.noLearnMask              = 0U;
        inArgs.vidIngressCheck          = false;
        inArgs.limitIPNxtHdr            = false;
        inArgs.disallowIPFragmentation  = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                               __func__, status);
        }
    }

    if (CPSW_SOK == status)
    {
        status = CpswTestCommon_setAleMulticastEntry(taskObj,
                                                     testDstMcastMacAddr,
                                                     0,
                                                     0,
                                                     CpswTestInterVlan_getIngressVlanMembershipMask());
    }

    return status;
}

static int32_t CpswTestInterVlan_setInterVlanMultiEgress(CpswTestTaskObj *taskObj,
                                                         CpswMacPort_InterVLANRouteId expectedAllocRouteId,
                                                         uint32_t *pNumRoutesUsed)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    Cpsw_SetInterVlanRouteMultiEgressInArgs inArgs;
    Cpsw_SetInterVlanRouteMultiEgressOutArgs outArgs;
    uint32_t i;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    *pNumRoutesUsed = 0;
    status          = CpswTestInterVlan_addMultiEgressAleTableEntries(taskObj);

    if (CPSW_SOK == status)
    {
        inArgs.inPktMatchCfg.packetMatchEnableMask = 0;

        inArgs.inPktMatchCfg.ingressPort            = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT;

        inArgs.inPktMatchCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.dstIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr, testDstIpv6AddrMcast, sizeof(inArgs.inPktMatchCfg.dstIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST;

        memcpy(inArgs.inPktMatchCfg.dstMacAddr.addr.addr, testDstMcastMacAddr, sizeof(inArgs.inPktMatchCfg.dstMacAddr.addr.addr));
        inArgs.inPktMatchCfg.dstMacAddr.addr.vlanId   = 0;
        inArgs.inPktMatchCfg.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM);
        inArgs.inPktMatchCfg.packetMatchEnableMask   |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST;

        inArgs.inPktMatchCfg.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV6;
        inArgs.inPktMatchCfg.srcIp.ipv6.numLSBIgnoreBits = 8;
        memcpy(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr, testSrcIpv6Addr, sizeof(inArgs.inPktMatchCfg.srcIp.ipv6.ipv6Addr));
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC;

        memcpy(inArgs.inPktMatchCfg.srcMacAddr.addr.addr, testSrcMacAddr, sizeof(inArgs.inPktMatchCfg.srcMacAddr.addr.addr));
        inArgs.inPktMatchCfg.srcMacAddr.addr.vlanId    = 0;
        inArgs.inPktMatchCfg.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM);

        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC;

        inArgs.inPktMatchCfg.etherType.etherType    = 0x86DD;
        inArgs.inPktMatchCfg.packetMatchEnableMask |= CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE;

        inArgs.inPktMatchCfg.vlanId         = CPSW_TEST_INTERVLAN_INGRESS_VLANID;
        inArgs.inPktMatchCfg.enableTTLCheck = TRUE;

        inArgs.numEgressPorts = CPSW_UTILS_ARRAYSIZE(testMultiEgressPortList);

        for (i = 0; i < inArgs.numEgressPorts; i++)
        {
            inArgs.egressCfg[i].egressPort                       = testMultiEgressPortList[i];
            inArgs.egressCfg[i].outPktModCfg.decrementTTL        = TRUE;
            inArgs.egressCfg[i].outPktModCfg.forceUntaggedEgress = FALSE;
            inArgs.egressCfg[i].outPktModCfg.replaceDASA         = TRUE;
            memcpy(inArgs.egressCfg[i].outPktModCfg.srcAddr, testHostMacAddr, sizeof(inArgs.egressCfg[i].outPktModCfg.srcAddr));
            memcpy(inArgs.egressCfg[i].outPktModCfg.dstAddr, testDstMcastMacAddr, sizeof(inArgs.egressCfg[i].outPktModCfg.dstAddr));
            inArgs.egressCfg[i].outPktModCfg.vlanId = CPSW_TEST_INTERVLAN_EGRESS_VLANID + 1 + i;
        }

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_INTERVLAN_ROUTE_MULTI_EGRESS,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswTestInterVlan_setInterVlanUniEgress() failed CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        *pNumRoutesUsed += 1;
        CpswAppUtils_assert(outArgs.egressPortRouteId == expectedAllocRouteId);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.port == CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.portIsTrunk == FALSE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpEnabled == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.egressOpcode == (1 + (outArgs.egressPortRouteId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST)));
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.enableTTLCheck == TRUE);
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.destPortMask == CpswTestInterVlan_getMultiEgressDestPortMask());
        CpswAppUtils_assert(outArgs.ingressPacketClassifierInfo.policerMatchEnableMask == (CPSW_ALE_POLICER_MATCH_PORT |
                                                                                           CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_MACDST |
                                                                                           CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                                           CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                                                                           CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                                           CPSW_ALE_POLICER_MATCH_IPDST));
    }

    return status;
}

volatile static uint32_t gCpswTestDumpPolicer = false;
uint32_t gCpswTestInterVlanTTL[1000];
static uint32_t gCpswTestInterVlanIndex = 0;

int32_t CpswTestInterVlan_processRxPkt(CpswTestTaskObj *taskObj,
                                       uint32_t rxFlowId,
                                       uint32_t pktNum,
                                       CpswDma_PktInfo *pktInfo,
                                       bool *testComplete)
{
    int32_t status = CPSW_SOK;
    EthVlanFrame *frame;

    frame = (EthVlanFrame *)pktInfo->bufPtr;
    /* Check for blacklist classifier not being received on host port */
    if (memcmp(&frame->hdr.dstMac[0], testHostMacAddr, sizeof(frame->hdr.dstMac)) == 0)
    {
        if ((ntohs(frame->hdr.tci) & 0xFFFU) == CPSW_TEST_INTERVLAN_INTRA_VLANID)
        {
            /* If IPv4 packet */
            if (ntohs(frame->hdr.etherType) == CPSW_TEST_INTERVLAN_IPV4_ETHERTYPE)
            {
                if (memcmp(&frame->hdr.srcMac[0], testBlackListSrcMacAddr, sizeof(frame->hdr.srcMac)) == 0)
                {
                    /* We should not get blacklisted classifier on host port */
                    status = CPSW_EFAIL;
                }
            }
        }
    }

    if (status == CPSW_SOK)
    {
        if ((memcmp(&frame->hdr.dstMac[0], testHostMacAddr, sizeof(frame->hdr.dstMac)) == 0)
            ||
            (memcmp(&frame->hdr.dstMac[0], testDstMcastMacAddr, sizeof(frame->hdr.dstMac)) == 0))
        {
            if (((ntohs(frame->hdr.tci) & 0xFFFU) == CPSW_TEST_INTERVLAN_INGRESS_VLANID)
                ||
                ((ntohs(frame->hdr.tci) & 0xFFFU) == CPSW_TEST_INTERVLAN_EGRESS_VLANID))
            {
                if (ntohs(frame->hdr.etherType) == CPSW_TEST_INTERVLAN_IPV6_ETHERTYPE)
                {
                    if (frame->payload[CPSW_TEST_INTERVLAN_IPV6_HOP_LIMIT_OFFSET] <= 1)
                    {
                        status = CPSW_SOK;
                    }
                    else
                    {
                        gCpswTestInterVlanIndex                        = (gCpswTestInterVlanIndex + 1) % CPSW_UTILS_ARRAYSIZE(gCpswTestInterVlanTTL);
                        gCpswTestInterVlanTTL[gCpswTestInterVlanIndex] = frame->payload[CPSW_TEST_INTERVLAN_IPV6_HOP_LIMIT_OFFSET];
                        status                                         = CPSW_EFAIL;
                    }
                }
                else
                {
                    if (ntohs(frame->hdr.etherType) == CPSW_TEST_INTERVLAN_IPV4_ETHERTYPE)
                    {
                        if (frame->payload[CPSW_TEST_INTERVLAN_IPV4_TTL_OFFSET] <= 1)
                        {
                            status = CPSW_SOK;
                        }
                        else
                        {
                            gCpswTestInterVlanIndex                        = (gCpswTestInterVlanIndex + 1) % CPSW_UTILS_ARRAYSIZE(gCpswTestInterVlanTTL);
                            gCpswTestInterVlanTTL[gCpswTestInterVlanIndex] = frame->payload[CPSW_TEST_INTERVLAN_IPV4_TTL_OFFSET];
                            status                                         = CPSW_EFAIL;
                        }
                    }
                }
            }
        }
    }

    if (gCpswTestDumpPolicer)
    {
        Cpsw_IoctlPrms prms;
        CpswTestStateObj *stateObj = &taskObj->stateObj;
        int32_t dumpStatus;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        dumpStatus = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                                &prms);
        CpswAppUtils_assert(dumpStatus == CPSW_SOK);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        dumpStatus = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES,
                                &prms);

        CpswAppUtils_assert(dumpStatus == CPSW_SOK);
    }

    *testComplete = FALSE;

    return status;
}

#define CPSW_TEST_INTERVLAN_DEFAULT_SHORTIPG_THRESHOLD                    (11)

static int32_t CpswTestInterVlan_setShortIPG(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_SetShortTxIPGConfigInArgs setShortIPGInArgs;
    int32_t status;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &setShortIPGInArgs);
    setShortIPGInArgs.configureGapThreshold                           = TRUE;
    setShortIPGInArgs.ipgTriggerThresholdBlkCount                     = 0;
    setShortIPGInArgs.numMacPorts                                     = 2;
    setShortIPGInArgs.portIPGCfg[0].portNum                           = CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM;
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGapLimit = false;

    setShortIPGInArgs.portIPGCfg[1].portNum                           = CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM;
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGapLimit = false;

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_SHORT_IPG_CONFIG,
                        &prms);
    if (CPSW_SOK == status)
    {
        Cpsw_GetShortTxIPGConfigOutArgs getShortIPGOutArgs;

        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getShortIPGOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_GET_SHORT_IPG_CONFIG,
                            &prms);
        if (CPSW_SOK == status)
        {
            uint32_t i;

            CpswAppUtils_assert(getShortIPGOutArgs.ipgTriggerThresholdBlkCount == 0);
            for (i = 0; i < getShortIPGOutArgs.numMacPorts; i++)
            {
                if ((getShortIPGOutArgs.portIPGCfg[i].portNum == CPSW_TEST_INTERVLAN_EGRESS_PORT_NUM)
                    ||
                    (getShortIPGOutArgs.portIPGCfg[i].portNum == CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM))
                {
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGap == true);
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGapLimit == false);
                }
            }
        }
    }

    return status;
}

static int32_t CpswTestInterVlan_addBlackListHostPortAleTableEntries(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswAle_VlanEntryInfo inArgs;
    CpswAle_AddEntryOutArgs outArgs;

    memset(&inArgs, 0, sizeof(inArgs));
    inArgs.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_INTRA_VLANID;
    inArgs.vlanIdInfo.outerVlanFlag = false;
    inArgs.vlanMemberList           = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.unregMcastFloodMask      = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.regMcastFloodMask        = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.forceUntaggedEgressMask  = 0U;
    inArgs.noLearnMask              = 0U;
    inArgs.vidIngressCheck          = false;
    inArgs.limitIPNxtHdr            = false;
    inArgs.disallowIPFragmentation  = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                           __func__, status);
    }

    memset(&inArgs, 0, sizeof(inArgs));
    inArgs.vlanIdInfo.vlanId        = CPSW_TEST_INTERVLAN_INTRA_VLAN_EGRESS_STRIP_ID;
    inArgs.vlanIdInfo.outerVlanFlag = false;
    inArgs.vlanMemberList           = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.unregMcastFloodMask      = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.regMcastFloodMask        = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.forceUntaggedEgressMask  = CpswTestInterVlan_getIntraVlanMembershipMask();
    inArgs.noLearnMask              = 0U;
    inArgs.vidIngressCheck          = false;
    inArgs.limitIPNxtHdr            = false;
    inArgs.disallowIPFragmentation  = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                           __func__, status);
    }

    return status;
}

static int32_t CpswTestInterVlan_setBlacklistHostportClassifier(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    int32_t status;
    CpswAle_PolicerMatchParams blackListTrafficCfg;
    CpswAle_BlockClassifierHostPortOutArgs outArgs;

    status = CpswTestInterVlan_addBlackListHostPortAleTableEntries(taskObj);
    if (status == CPSW_SOK)
    {
        memset(&blackListTrafficCfg, 0, sizeof(blackListTrafficCfg));

        /* Clear classifier mask to 0 */
        blackListTrafficCfg.policerMatchEnableMask = 0;

        /* Blacklist IPv4 Src address */
        blackListTrafficCfg.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
        blackListTrafficCfg.srcIp.ipv4.numLSBIgnoreBits = 8;
        memcpy(blackListTrafficCfg.srcIp.ipv4.ipv4Addr, testSrcIpv4Addr, sizeof(blackListTrafficCfg.srcIp.ipv4.ipv4Addr));
        blackListTrafficCfg.policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IPSRC;

        /* Blacklist IPv4 Dst address */
        blackListTrafficCfg.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
        blackListTrafficCfg.dstIp.ipv4.numLSBIgnoreBits = 8;
        memcpy(blackListTrafficCfg.dstIp.ipv4.ipv4Addr, testDstIpv4Addr, sizeof(blackListTrafficCfg.dstIp.ipv4.ipv4Addr));
        blackListTrafficCfg.policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IPDST;

        /* Blacklist Src Mac address */
        blackListTrafficCfg.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_INTERVLAN_INGRESS_PORT_NUM);
        CPSW_UTILS_ARRAY_COPY(blackListTrafficCfg.srcMacAddr.addr.addr, testBlackListSrcMacAddr);
        blackListTrafficCfg.policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_MACSRC;

        /* Blacklist IVLAN id */
        blackListTrafficCfg.ivlanId                 = CPSW_TEST_INTERVLAN_INTRA_VLANID;
        blackListTrafficCfg.policerMatchEnableMask |= CPSW_ALE_POLICER_MATCH_IVLAN;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &blackListTrafficCfg, &outArgs);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId,
                            CPSW_ALE_IOCTL_BLOCK_CLASSIFIER_HOSTPORT,
                            &prms);
        if (CPSW_SOK == status)
        {
            /* Assert main configuration is correct,
             * Egress op should be enabled with destPortMask of 0
             * Egress Op code should be non-zero
             */
            CpswAppUtils_assert((outArgs.egressOpEnabled == true)
                                &&
                                (outArgs.destPortMask == 0)
                                &&
                                (outArgs.egressOpcode == 1));

            /* Assert other policer configuration is correct
             */
            CpswAppUtils_assert((outArgs.egressTrunkIndex == 0)
                                &&
                                (outArgs.enableTTLCheck == false)
                                &&
                                (outArgs.threadIdEnable == false)
                                &&
                                (outArgs.commitRateInBitsPerSec == 0)
                                &&
                                (outArgs.peakRateInBitsPerSec == 0));
        }
    }

    return status;
}
