/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_network_security.c
 *
 * \brief    This file contains the cpsw_test_network_security test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_network_security.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define IP_V4_NXT_HDR_TCP           6U
#define IP_V4_NXT_HDR_UDP           17U
#define IP_V4_NXT_HDR_ICMP          1U
#define IP_V4_NXT_HDR_IGMP          2U

#define ETH_TEST_VLAN_ID           100U

#define CPSW_TEST_SECURITY_INGRESS_PORT  (CPSW_MAC_PORT_3)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswTestSecurity_setVLANentry(CpswTestTaskObj *taskObj);

static int32_t CpswTestSecurity_setOUIEntry(CpswTestTaskObj *taskObj);

static int32_t CpswTestSecurity_SetAleEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

static uint8_t testSrcValidOUIAddr[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x00, 0x00, 0x00, 0x00, 0x02
};

static uint8_t testSrcValidOUIAddr2[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x00, 0x00, 0x00, 0x00, 0x04
};

static uint8_t testSrcInvalidOUIAddr[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x02, 0x00, 0x00, 0x00, 0x00
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestSecurity_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    if (CPSW_SOK == status)
    {
        status = CpswTestSecurity_SetAleEntry(taskObj);
    }

    if (CPSW_SOK == status)
    {
        status = CpswTestSecurity_setVLANentry(taskObj);
    }

    if (CPSW_SOK == status)
    {
        status = CpswTestSecurity_setOUIEntry(taskObj);
    }

    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);

        CpswAppUtils_assert(status == CPSW_SOK);
        
        /* Use default mac address instead of board specific mac address so that IXIA test does not change per board */
        CpswTestCommon_getDefaultHostMacAddr(&stateObj->hostMacAddr[0U]);
        status = CpswTestCommon_Run(taskObj);
    }

    return status;
}

void CpswTestSecurity_setOpenPrms(CpswTestTaskObj *taskObj,
                                  Cpsw_Config *pCpswCfg,
                                  CpswOsal_Prms *pOsalPrms,
                                  CpswUtils_Prms *pUtilsPrms)
{
    Cpsw_MacPort i;

    // VLAN configurations
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode               = TRUE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode              = FALSE;
    pCpswCfg->aleConfig.vlanConfig.autoLearnWithVLAN              = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanNoLearn             = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownForceUntaggedEgressMask = 0U;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask       = 0U;
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask     = 0U;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask      = CPSW_ALE_ALL_PORTS_MASK;

    // Network security configurations
    pCpswCfg->aleConfig.nwSecCfg.hostOuiNoMatchDeny                    = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.enableVid0Mode                        = FALSE;
    pCpswCfg->aleConfig.nwSecCfg.malformedPktCfg.disableSrcMcastDrop   = FALSE;
    pCpswCfg->aleConfig.nwSecCfg.malformedPktCfg.enableBadLenPktDrop   = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.enableDefaultNoFragment      = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.enableDefaultNxtHdrWhitelist = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.ipNxtHdrWhitelistCount       = 4U;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.ipNxtHdrWhitelist[0]         = IP_V4_NXT_HDR_TCP;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.ipNxtHdrWhitelist[1]         = IP_V4_NXT_HDR_UDP;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.ipNxtHdrWhitelist[2]         = IP_V4_NXT_HDR_ICMP;
    pCpswCfg->aleConfig.nwSecCfg.ipPktCfg.ipNxtHdrWhitelist[3]         = IP_V4_NXT_HDR_IGMP;
    pCpswCfg->aleConfig.nwSecCfg.macAuthCfg.enableAuthMode             = TRUE;
    pCpswCfg->aleConfig.nwSecCfg.macAuthCfg.macAuthDisableMask         = 0U;

    for (i = CPSW_MAC_PORT_FIRST; i <= CPSW_MAC_PORT_LAST; i++)
    {
        // Port configurations
        pCpswCfg->aleConfig.portCfg[i].learningCfg.noLearn          = TRUE;
        pCpswCfg->aleConfig.portCfg[i].learningCfg.noSaUpdateEnable = TRUE;
        pCpswCfg->aleConfig.portCfg[i].vlanCfg.vidIngressCheck      = TRUE;
        pCpswCfg->aleConfig.portCfg[i].vlanCfg.dropUntagged         = TRUE;
        pCpswCfg->aleConfig.portCfg[i].vlanCfg.dropDualVlan         = TRUE;
        pCpswCfg->aleConfig.portCfg[i].vlanCfg.dropDoubleVlan       = TRUE;
    }
}

void CpswTestSecurity_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                        Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

static int32_t CpswTestSecurity_setVLANentry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_VlanEntryInfo vlanEntry;
    CpswAle_AddEntryOutArgs vlanOutInfo;
    CpswAle_GetVlanEntryOutArgs getVlanOutArgs;
    CpswAle_VlanIdInfo getVlanInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&vlanEntry, 0, sizeof(vlanEntry));
    vlanEntry.disallowIPFragmentation  = TRUE;
    vlanEntry.forceUntaggedEgressMask  = 0x0;
    vlanEntry.limitIPNxtHdr            = TRUE;
    vlanEntry.noLearnMask              = 0x0;
    vlanEntry.regMcastFloodMask        = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    vlanEntry.unregMcastFloodMask      = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    vlanEntry.vidIngressCheck          = TRUE;
    vlanEntry.vlanMemberList           = CpswTestCommon_getMacPortAleMask(taskObj) | CPSW_ALE_HOST_PORT_MASK;
    vlanEntry.vlanIdInfo.vlanId        = ETH_TEST_VLAN_ID;
    vlanEntry.vlanIdInfo.outerVlanFlag = FALSE;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &vlanEntry, &vlanOutInfo);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setAleEntry() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    memset(&getVlanInArgs, 0, sizeof(getVlanInArgs));
    getVlanInArgs.vlanId        = ETH_TEST_VLAN_ID;
    getVlanInArgs.outerVlanFlag = FALSE;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getVlanInArgs, &getVlanOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_LOOKUP_VLAN,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestAutoLearnVlan_setVLANentry() failed CPSW_ALE_IOCTL_LOOKUP_VLAN: %d\n",
                           status);
    }

    return status;
}

static int32_t CpswTestSecurity_setOUIEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setOuiOutArgs;
    CpswAle_OuiEntryInfo setOuiInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setOuiInArgs.ouiAddr[0], testSrcValidOUIAddr, sizeof(setOuiInArgs.ouiAddr));

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setOuiInArgs, &setOuiOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_OUI,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setOUIEntry() failed : %d\n", status);
    }

    return status;
}

static int32_t CpswTestSecurity_SetAleEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], testSrcValidOUIAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = ETH_TEST_VLAN_ID;
    setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_SECURITY_INGRESS_PORT);
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
        memcpy(&setUcastInArgs.addr.addr[0U], testSrcValidOUIAddr2,
               sizeof(setUcastInArgs.addr.addr));
        setUcastInArgs.addr.vlanId  = 0;
        setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_SECURITY_INGRESS_PORT);
        setUcastInArgs.info.blocked = false;
        setUcastInArgs.info.secure  = false;
        setUcastInArgs.info.super   = 0U;
        setUcastInArgs.info.ageable = false;
        setUcastInArgs.info.trunk   = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                               status);
        }
    }

    if (status == CPSW_SOK)
    {
        memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
        memcpy(&setUcastInArgs.addr.addr[0U], testSrcInvalidOUIAddr,
               sizeof(setUcastInArgs.addr.addr));
        setUcastInArgs.addr.vlanId  = 0;
        setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_SECURITY_INGRESS_PORT);
        setUcastInArgs.info.blocked = false;
        setUcastInArgs.info.secure  = false;
        setUcastInArgs.info.super   = 0U;
        setUcastInArgs.info.ageable = false;
        setUcastInArgs.info.trunk   = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                            &prms);

        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                               status);
        }
    }

    return status;
}

int32_t CpswTestSecurity_processRxPkt(CpswTestTaskObj *taskObj,
                                      uint32_t rxFlowId,
                                      uint32_t pktNum,
                                      CpswDma_PktInfo *pktInfo,
                                      bool *testComplete)
{
    EthFrame *frame;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    int32_t status             = CPSW_SOK;
    EthVlanFrame *vlanFrame;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;

    if (memcmp(&frame->hdr.dstMac[0], &stateObj->hostMacAddr[0U], sizeof(frame->hdr.dstMac)) != 0)
    {
        status = CPSW_EFAIL;
    }

    if (memcmp(&frame->hdr.srcMac[0], &testSrcValidOUIAddr, CPSW_ALE_OUIADDR_NUM_OCTETS) != 0)
    {
        status = CPSW_EFAIL;
    }

    vlanFrame = (EthVlanFrame *)frame;
    if ((ntohs(vlanFrame->hdr.tci) & 0xFFFU) != ETH_TEST_VLAN_ID)
    {
        // status = CPSW_EFAIL;
    }

    *testComplete = FALSE;
    return status;
}
