/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_nway.c
 *
 * \brief    This file contains the CPSW auto-negotiation test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>
#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/osal/osal.h>
#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_nway.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

static void CpswTestNway_printLinkConfig(CpswMacPort_LinkConfig *linkCfg);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestNway_Run(CpswTestTaskObj *taskObj)
{
    CpswTestMacPortList_t enabledPorts;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswPhy_GenericInArgs phyInArgs;
    CpswMacPort_LinkConfig phyOutArgs;
    CpswMacPort_GenericInArgs macInArgs;
    CpswMacPort_LinkConfig macOutArgs;
    Cpsw_GenericPortLinkInArgs cpswInArgs;
    CpswMacPort_LinkConfig cpswOutArgs;
    Cpsw_IoctlPrms prms;
    uint32_t portNum;
    int32_t phyStatus;
    int32_t macStatus;
    int32_t cpswStatus;
    int32_t status = CPSW_SOK;

    CpswTestCommon_waitForPortLink(taskObj);
    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    /* Single port test, ignore others */
    portNum = enabledPorts.macPortList[0];
    CpswAppUtils_print("Port %u\n", portNum);

    /* Get port link state (speed & duplexity) from PHY */
    phyInArgs.portNum = (Cpsw_MacPort)portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyInArgs, &phyOutArgs);

    phyStatus = Cpsw_ioctl(stateObj->hCpsw,
                           stateObj->coreId,
                           CPSW_PHY_IOCTL_PHY_GET_LINK_MODE,
                           &prms);
    if (phyStatus != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to get link mode from PHY driver: %d\n", phyStatus);
        status = phyStatus;
    }

    /* Get port link state (speed & duplexity) from MAC port */
    macInArgs.portNum = (Cpsw_MacPort)portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &macInArgs, &macOutArgs);

    macStatus = Cpsw_ioctl(stateObj->hCpsw,
                           stateObj->coreId,
                           CPSW_MACPORT_IOCTL_GET_LINK_CONFIG,
                           &prms);
    if (macStatus != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to get link mode from MAC port: %d\n", macStatus);
        status = macStatus;
    }

    /* Get port link state (speed & duplexity) from CPSW top level */
    cpswInArgs.portNum = (Cpsw_MacPort)portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &cpswInArgs, &cpswOutArgs);

    cpswStatus = Cpsw_ioctl(stateObj->hCpsw,
                            stateObj->coreId,
                            CPSW_IOCTL_GET_PORT_LINK_CONFIG,
                            &prms);
    if (cpswStatus != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to get link mode from CPSW layer: %d\n", cpswStatus);
        status = cpswStatus;
    }

    /* Print all link configs read from different layers */
    if (phyStatus == CPSW_SOK)
    {
        CpswAppUtils_print("PHY link config: ");
        CpswTestNway_printLinkConfig(&phyOutArgs);
    }

    if (macStatus == CPSW_SOK)
    {
        CpswAppUtils_print("MAC link config: ");
        CpswTestNway_printLinkConfig(&macOutArgs);
    }

    if (cpswStatus == CPSW_SOK)
    {
        CpswAppUtils_print("CPSW link config: ");
        CpswTestNway_printLinkConfig(&cpswOutArgs);
    }

    return status;
}

void CpswTestNway_updatePortLinkCfg1GFd(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                        Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
    pLinkArgs->linkConfig.speed         = CPSW_SPEED_AUTO;
    pLinkArgs->linkConfig.duplexity     = CPSW_DUPLEX_AUTO;
    pLinkArgs->phyConfig.nwayCaps       = CPSW_LINK_CAP_FD1000;
}

void CpswTestNway_updatePortLinkCfg100MFd(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                          Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
    pLinkArgs->linkConfig.speed         = CPSW_SPEED_AUTO;
    pLinkArgs->linkConfig.duplexity     = CPSW_DUPLEX_AUTO;
    pLinkArgs->phyConfig.nwayCaps       = CPSW_LINK_CAP_FD100;
}

void CpswTestNway_updatePortLinkCfg100MHd(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                          Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
    pLinkArgs->linkConfig.speed         = CPSW_SPEED_AUTO;
    pLinkArgs->linkConfig.duplexity     = CPSW_DUPLEX_AUTO;
    pLinkArgs->phyConfig.nwayCaps       = CPSW_LINK_CAP_HD100;
}

static void CpswTestNway_printLinkConfig(CpswMacPort_LinkConfig *linkCfg)
{
    const char *speedStr;
    const char *duplexStr;
    bool invalid = false;

    switch (linkCfg->speed)
    {
        case CPSW_SPEED_10MBIT:
            speedStr = "10-Mbps";
            break;

        case CPSW_SPEED_100MBIT:
            speedStr = "100-Mbps";
            break;

        case CPSW_SPEED_1GBIT:
            speedStr = "1-Gbps";
            break;

        default:
            speedStr = "invalid speed";
            invalid  = true;
            break;
    }

    switch (linkCfg->duplexity)
    {
        case CPSW_DUPLEX_HALF:
            duplexStr = "half-duplex";
            break;

        case CPSW_DUPLEX_FULL:
            duplexStr = "full-duplex";
            break;

        default:
            duplexStr = "invalid duplexity";
            invalid   = true;
            break;
    }

    CpswAppUtils_print("%s%s %s\n", speedStr, invalid ? ", " : "", duplexStr);
}
