/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_outer_vlan.c
 *
 * \brief    This file contains the cpsw outer vlan test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethpatterns.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_outer_vlan.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define ETH_TEST_VLAN_ID                      (100U)
#define ETHERTYPE_OUTER_VLAN_TAG              (0x88a8U)
#define ETH_TEST_HOST_PORT_VLAN_ID            (1000U)
#define ETH_TEST_MAC_PORT_VLAN_ID_BASE        (500U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */
/* Test ALE Entry Multicast Address */
static uint8_t testMCastAddr[CPSW_MAC_ADDR_LEN] =
{
    0x01, 0x80, 0xC2, 0x00, 0x00, 0x02
};

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int CpswTestOuterVlan_addAleVlanEntry(CpswTestTaskObj *taskObj);

static int CpswTestOuterVlan_setAleMulticastEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestOuterVlan_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Add ALE entry to force untagged packets on ETH_TEST_VLAN_ID */
    status = CpswTestOuterVlan_addAleVlanEntry(taskObj);
    if (CPSW_SOK == status)
    {
        status = CpswTestOuterVlan_setAleMulticastEntry(taskObj);
    }

    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        status = CpswTestCommon_Run(taskObj);
    }

    return status;
}

void CpswTestOuterVlan_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                         Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestOuterVlan_setOpenPrms(CpswTestTaskObj *taskObj,
                                   Cpsw_Config *pCpswCfg,
                                   CpswOsal_Prms *pOsalPrms,
                                   CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode               = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask      = 0;
    pCpswCfg->aleConfig.vlanConfig.unknownForceUntaggedEgressMask = 0;
    pCpswCfg->aleConfig.vlanConfig.unknownRegMcastFloodMask       = 0;
    pCpswCfg->aleConfig.vlanConfig.unknownUnregMcastFloodMask     = 0;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanNoLearn             = 0;

    pCpswCfg->vlanConfig.vlanSwitch = CPSW_VLAN_SWITCH_SERVICE;
}

static int CpswTestOuterVlan_addAleVlanEntry(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj   = &taskObj->stateObj;
    CpswAle_VlanEntryInfo inArgs =
    {
        .vlanIdInfo.vlanId        = ETH_TEST_VLAN_ID,
        .vlanMemberList           = CPSW_ALE_ALL_PORTS_MASK,
        .unregMcastFloodMask      = 0U,
        .regMcastFloodMask        = CPSW_ALE_ALL_PORTS_MASK,
        .forceUntaggedEgressMask  = 0U,
        .noLearnMask              = 0U,
        .vidIngressCheck          = true,
        .limitIPNxtHdr            = false,
        .disallowIPFragmentation  = false,
        .vlanIdInfo.outerVlanFlag = true,
    };
    CpswAle_AddEntryOutArgs outArgs;
    int32_t status;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestOuterVlan_addAleVlanEntry() ADD_VLAN ioctl failed: %d\n",
                           status);
    }

    return status;
}

static int CpswTestOuterVlan_setAleMulticastEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setMcastOutArgs;
    CpswAle_SetMcastEntryInArgs setMcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memcpy(&setMcastInArgs.addr.addr[0], testMCastAddr,
           sizeof(setMcastInArgs.addr.addr));
    setMcastInArgs.addr.vlanId = 0;

    setMcastInArgs.info.superFlag  = false;
    setMcastInArgs.info.fwdState   = CPSW_ALE_FWDSTLVL_FWD_LRN;
    setMcastInArgs.info.portMask   = CPSW_ALE_ALL_PORTS_MASK;
    setMcastInArgs.info.numIgnBits = 0;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setMcastInArgs, &setMcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_MULTICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswTestOuterVlan_setAleMulticastEntry() failed CPSW_ALE_IOCTL_ADD_MULTICAST: %d\n",
                           status);
    }

    return status;
}

int32_t CpswTestOuterVlan_processRxPkt(CpswTestTaskObj *taskObj,
                                       uint32_t rxFlowId,
                                       uint32_t pktNum,
                                       CpswDma_PktInfo *pktInfo,
                                       bool *testComplete)
{
    int32_t status = CPSW_SOK;

    EthVlanFrame *frame;
    uint32_t vid;
    uint16_t etherType;
    uint32_t tpid;

    /* Consume the packet by just printing its content */
    frame = (EthVlanFrame *)pktInfo->bufPtr;

    tpid      = (ntohs(frame->hdr.tpid) & 0xFFFFU);
    vid       = (ntohs(frame->hdr.tci) & 0xFFFU);
    etherType = ntohs(frame->hdr.etherType);

    if ((vid == ETH_TEST_VLAN_ID) && (etherType == ETHERTYPE_VLAN_TAG) && (tpid == ETHERTYPE_OUTER_VLAN_TAG))
    {
        status = CPSW_SOK;
    }
    else
    {
        status = CPSW_EFAIL;
    }

    *testComplete = FALSE;
    return status;
}
