/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_policer.c
 *
 * \brief    This file contains the cpsw_policer test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_policer.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define ETH_TEST_IVLAN_ID 100U
#define ETH_TEST_OVLAN_ID 150U

#define CPSW_TEST_POLICER_MBPS(x)                   ((x) * 1000000U)

#define RECV_BIT_RATE_PIR_TEST1 CPSW_TEST_POLICER_MBPS(500U)
#define RECV_BIT_RATE_CIR_TEST1 CPSW_TEST_POLICER_MBPS(300U)

#define RECV_BIT_RATE_PIR_TEST2 CPSW_TEST_POLICER_MBPS(700U)
#define RECV_BIT_RATE_CIR_TEST2 CPSW_TEST_POLICER_MBPS(500U)

#define RECV_BIT_RATE_PIR_TEST3 CPSW_TEST_POLICER_MBPS(300U)
#define RECV_BIT_RATE_CIR_TEST3 CPSW_TEST_POLICER_MBPS(150U)

#define RECV_BIT_RATE_RANGE_BASE_PIR CPSW_TEST_POLICER_MBPS(950U)
#define RECV_BIT_RATE_RANGE_BASE_CIR CPSW_TEST_POLICER_MBPS(900U)
#define RECV_BIT_RATE_RANGE_INTERVAL CPSW_TEST_POLICER_MBPS(50U)
#define RECV_BIT_RATE_RANGE_MIN_PIR  CPSW_TEST_POLICER_MBPS(50U)


#define CPSW_TEST_POLICER_INGRESS_PORT  (CPSW_MAC_PORT_3)
#define CPSW_TEST_POLICER_EGRESS_PORT   (CPSW_MAC_PORT_2)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static void CpswTestPolicer_setPolicer(CpswTestTaskObj *taskObj);

static int32_t CpswTestPolicer_setPolicerAleEntryType1(CpswTestTaskObj *taskObj);

static int32_t CpswTestPolicer_setPolicerAleEntryType2(CpswTestTaskObj *taskObj);

static int32_t CpswTestPolicer_setPolicerAleEntryType3(CpswTestTaskObj *taskObj);

static int32_t CpswTestPolicer_Test(CpswTestTaskObj *taskObj,
                                    uint32_t rxFlowCfgId,
                                    uint32_t recvbitRatePir,
                                    uint32_t recvbitRateCir);
int32_t CpswTestPolicer_setPolicerRange(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
static uint8_t testSrcAddr[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x00, 0x00, 0x00, 0x00, 0x08
};

static uint8_t testHostPortDestAddr_Policer1[CPSW_MAC_ADDR_LEN] =
{
    0x04, 0x00, 0x00, 0x00, 0x00, 0x00
};

static uint8_t testEgressPortDestAddr_Policer1[CPSW_MAC_ADDR_LEN] =
{
    0x04, 0x00, 0x00, 0x00, 0x00, 0x02
};

static uint8_t testHostPortDestAddr_Policer2[CPSW_MAC_ADDR_LEN] =
{
    0x08, 0x00, 0x00, 0x00, 0x00, 0x00
};

static uint8_t testEgressPortDestAddr_Policer2[CPSW_MAC_ADDR_LEN] =
{
    0x08, 0x00, 0x00, 0x00, 0x00, 0x02
};

static uint8_t testOuiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
{
    0x02, 0x00, 0x00
};

static uint8_t testSrcIpv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
{
    172, 24, 190, 113
};

static uint8_t testDestIpv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
{
    172, 24, 190, 112
};


static uint8_t testSrcIpv4AddrRange[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
{
    172, 24, 180, 1
};

static uint8_t testSrcAddrRange[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x00, 0x00, 0x00, 0x11, 0x11
};

static uint8_t testEgressPortDestAddrRange[CPSW_MAC_ADDR_LEN] =
{
    0x04, 0x00, 0x00, 0x00, 0x11, 0x11
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestPolicer_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswTestPolicer_setPolicer(taskObj);
    status = CpswTestPolicer_setPolicerAleEntryType1(taskObj);
    if (CPSW_SOK == status)
    {
        status = CpswTestPolicer_setPolicerAleEntryType2(taskObj);
    }

    if (CPSW_SOK == status)
    {
        status = CpswTestPolicer_setPolicerAleEntryType3(taskObj);
    }

    if (CPSW_SOK == status)
    {
        //status = CpswTestPolicer_setPolicerRange(taskObj);
    }
    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw,
                            stateObj->coreId,
                            CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw,
                            stateObj->coreId,
                            CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        status = CpswTestCommon_Run(taskObj);
    }

    if (status == CPSW_SOK)
    {
        status = CpswTestPolicer_Test(taskObj, CPSWTESTPOLICER_TYPE1_RXFLOWCFGID, RECV_BIT_RATE_PIR_TEST1, RECV_BIT_RATE_CIR_TEST1);
    }

    if (status == CPSW_SOK)
    {
        status = CpswTestPolicer_Test(taskObj, CPSWTESTPOLICER_TYPE2_RXFLOWCFGID, RECV_BIT_RATE_PIR_TEST2, RECV_BIT_RATE_CIR_TEST2);
    }

    if (status == CPSW_SOK)
    {
        status = CpswTestPolicer_Test(taskObj, CPSWTESTPOLICER_TYPE3_RXFLOWCFGID, RECV_BIT_RATE_PIR_TEST3, RECV_BIT_RATE_CIR_TEST3);
    }

    return status;
}

void CpswTestPolicer_setOpenPrms(CpswTestTaskObj *taskObj,
                                 Cpsw_Config *pCpswCfg,
                                 CpswOsal_Prms *pOsalPrms,
                                 CpswUtils_Prms *pUtilsPrms)
{
    CpswAle_Config *aleCfg;

    aleCfg = &pCpswCfg->aleConfig;

    aleCfg->policerGlobalConfig.policingEnable     = TRUE;
    aleCfg->policerGlobalConfig.yellowDropEnable   = FALSE;
    aleCfg->policerGlobalConfig.redDropEnable      = TRUE;
    aleCfg->policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_GREEN;
    aleCfg->vlanConfig.aleVlanAwareMode            = TRUE;
    aleCfg->nwSecCfg.hostOuiNoMatchDeny            = TRUE;
}

static void CpswTestPolicer_setPolicer(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswAle_PolicerMatchParams getPolicerInArgs;
    CpswAle_GetPolicerEntryOutArgs getPolicerOutArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
    setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                                            CPSW_ALE_POLICER_MATCH_PRIORITY |
                                                            CPSW_ALE_POLICER_MATCH_OUI |
                                                            CPSW_ALE_POLICER_MATCH_MACSRC |
                                                            CPSW_ALE_POLICER_MATCH_MACDST |
                                                            CPSW_ALE_POLICER_MATCH_IVLAN |
                                                            CPSW_ALE_POLICER_MATCH_OVLAN |
                                                            CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                                            CPSW_ALE_POLICER_MATCH_IPSRC |
                                                            CPSW_ALE_POLICER_MATCH_IPDST);
    setPolicerInArgs.policerMatch.portNum     = 1U;
    setPolicerInArgs.policerMatch.portIsTrunk = FALSE;
    setPolicerInArgs.policerMatch.priority    = 2U;
    memcpy(&setPolicerInArgs.policerMatch.oui.ouiAddr[0U], testOuiAddr,
           sizeof(setPolicerInArgs.policerMatch.oui.ouiAddr));
    memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], testSrcAddr,
           sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
    setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
    memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], testHostPortDestAddr_Policer1,
           sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
    setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = 0;
    setPolicerInArgs.policerMatch.ivlanId                  = ETH_TEST_IVLAN_ID;
    setPolicerInArgs.policerMatch.ovlanId                  = ETH_TEST_OVLAN_ID;
    setPolicerInArgs.policerMatch.etherType.etherType      = ETHERTYPE_EXPERIMENTAL1;
    memcpy(&setPolicerInArgs.policerMatch.srcIp.ipv4.ipv4Addr[0U], testSrcIpv4Addr,
           sizeof(setPolicerInArgs.policerMatch.srcIp.ipv4.ipv4Addr));
    setPolicerInArgs.policerMatch.srcIp.ipv4.numLSBIgnoreBits = 0;
    setPolicerInArgs.policerMatch.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
    memcpy(&setPolicerInArgs.policerMatch.dstIp.ipv4.ipv4Addr[0U], testDestIpv4Addr,
           sizeof(setPolicerInArgs.policerMatch.dstIp.ipv4.ipv4Addr));
    setPolicerInArgs.policerMatch.dstIp.ipv4.numLSBIgnoreBits = 0;
    setPolicerInArgs.policerMatch.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;

    setPolicerInArgs.threadIdEnable         = TRUE;
    setPolicerInArgs.threadId               = 10U;
    setPolicerInArgs.peakRateInBitsPerSec   = RECV_BIT_RATE_PIR_TEST1;
    setPolicerInArgs.commitRateInBitsPerSec = RECV_BIT_RATE_CIR_TEST1;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setPolicyEntry() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    memset(&getPolicerInArgs, 0, sizeof(getPolicerInArgs));
    getPolicerInArgs.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                               CPSW_ALE_POLICER_MATCH_PRIORITY |
                                               CPSW_ALE_POLICER_MATCH_OUI |
                                               CPSW_ALE_POLICER_MATCH_MACSRC |
                                               CPSW_ALE_POLICER_MATCH_MACDST |
                                               CPSW_ALE_POLICER_MATCH_IVLAN |
                                               CPSW_ALE_POLICER_MATCH_OVLAN |
                                               CPSW_ALE_POLICER_MATCH_ETHERTYPE |
                                               CPSW_ALE_POLICER_MATCH_IPSRC |
                                               CPSW_ALE_POLICER_MATCH_IPDST);
    getPolicerInArgs.portNum     = 1U;
    getPolicerInArgs.portIsTrunk = FALSE;
    getPolicerInArgs.priority    = 2U;
    memcpy(&getPolicerInArgs.oui.ouiAddr[0U], testOuiAddr,
           sizeof(getPolicerInArgs.oui.ouiAddr));
    memcpy(&getPolicerInArgs.srcMacAddr.addr.addr[0U], testSrcAddr,
           sizeof(getPolicerInArgs.srcMacAddr.addr.addr));
    getPolicerInArgs.srcMacAddr.addr.vlanId    = 0;
    getPolicerInArgs.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
    memcpy(&getPolicerInArgs.dstMacAddr.addr.addr[0U], testHostPortDestAddr_Policer1,
           sizeof(getPolicerInArgs.dstMacAddr.addr.addr));
    getPolicerInArgs.dstMacAddr.addr.vlanId   = 0;
    getPolicerInArgs.dstMacAddr.egressPortNum = 0;
    getPolicerInArgs.ivlanId                  = ETH_TEST_IVLAN_ID;
    getPolicerInArgs.ovlanId                  = ETH_TEST_OVLAN_ID;
    getPolicerInArgs.etherType.etherType      = ETHERTYPE_EXPERIMENTAL1;
    memcpy(&getPolicerInArgs.srcIp.ipv4.ipv4Addr[0U], testSrcIpv4Addr,
           sizeof(getPolicerInArgs.srcIp.ipv4.ipv4Addr));
    getPolicerInArgs.srcIp.ipv4.numLSBIgnoreBits = 0;
    getPolicerInArgs.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;
    memcpy(&getPolicerInArgs.dstIp.ipv4.ipv4Addr[0U], testDestIpv4Addr,
           sizeof(getPolicerInArgs.dstIp.ipv4.ipv4Addr));
    getPolicerInArgs.dstIp.ipv4.numLSBIgnoreBits = 0;
    getPolicerInArgs.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getPolicerInArgs, &getPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_GET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_getPolicyEntry() failed CPSW_ALE_IOCTL_GET_POLICER: %d\n",
                           status);
    }
    else
    {
        CpswAppUtils_assert(setPolicerOutArgs.policerEntryIndex ==
                            getPolicerOutArgs.policerEntryIndex);
        CpswAppUtils_assert(setPolicerInArgs.policerMatch.policerMatchEnableMask ==
                            getPolicerOutArgs.policerMatchEnableMask);
        CpswAppUtils_assert(setPolicerInArgs.policerMatch.portNum ==
                            getPolicerOutArgs.port);
        CpswAppUtils_assert(setPolicerInArgs.policerMatch.portIsTrunk ==
                            getPolicerOutArgs.portIsTrunk);
        CpswAppUtils_assert(setPolicerInArgs.policerMatch.priority ==
                            getPolicerOutArgs.priority);
        CpswAppUtils_assert(setPolicerOutArgs.ouiAleEntryIndex ==
                            getPolicerOutArgs.ouiAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.srcMacAleEntryIndex ==
                            getPolicerOutArgs.srcMacAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.dstMacAleEntryIndex ==
                            getPolicerOutArgs.dstMacAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.ivlanAleEntryIndex ==
                            getPolicerOutArgs.ivlanAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.ovlanAleEntryIndex ==
                            getPolicerOutArgs.ovlanAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.etherTypeAleEntryIndex ==
                            getPolicerOutArgs.etherTypeAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.srcIpAleEntryIndex ==
                            getPolicerOutArgs.srcIpAleEntryIndex);
        CpswAppUtils_assert(setPolicerOutArgs.dstIpAleEntryIndex ==
                            getPolicerOutArgs.dstIpAleEntryIndex);
        CpswAppUtils_assert(setPolicerInArgs.threadIdEnable ==
                            getPolicerOutArgs.threadIdEnable);
        CpswAppUtils_assert(setPolicerInArgs.threadId ==
                            getPolicerOutArgs.threadId);
        if (setPolicerInArgs.peakRateInBitsPerSec != 0)
        {
            CpswAppUtils_assert((setPolicerInArgs.peakRateInBitsPerSec / getPolicerOutArgs.peakRateInBitsPerSec) == 1);
        }
        else
        {
            CpswAppUtils_assert(getPolicerOutArgs.peakRateInBitsPerSec == 0);
        }

        if (setPolicerInArgs.commitRateInBitsPerSec != 0)
        {
            CpswAppUtils_assert((setPolicerInArgs.commitRateInBitsPerSec / getPolicerOutArgs.commitRateInBitsPerSec) == 1);
        }
        else
        {
            CpswAppUtils_assert(getPolicerOutArgs.commitRateInBitsPerSec == 0);
        }
    }
}

static int32_t CpswTestPolicer_setPolicerAleEntryType1(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
    setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                                            CPSW_ALE_POLICER_MATCH_MACSRC |
                                                            CPSW_ALE_POLICER_MATCH_MACDST);
    setPolicerInArgs.policerMatch.portNum     = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
    setPolicerInArgs.policerMatch.portIsTrunk = FALSE;
    memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], testSrcAddr,
           sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
    setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
    memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], testHostPortDestAddr_Policer1,
           sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
    setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_HOST_PORT_NUM;
    setPolicerInArgs.threadIdEnable                        = TRUE;
    setPolicerInArgs.threadId                              = taskObj->stateObj.rxFlowObj[CPSWTESTPOLICER_TYPE1_RXFLOWCFGID].rxFlowIdx;
    setPolicerInArgs.peakRateInBitsPerSec                  = RECV_BIT_RATE_PIR_TEST1;
    setPolicerInArgs.commitRateInBitsPerSec                = RECV_BIT_RATE_CIR_TEST1;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    if (CPSW_SOK == status)
    {
        memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
        setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                                                CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                CPSW_ALE_POLICER_MATCH_MACDST);
        setPolicerInArgs.policerMatch.portNum     = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
        setPolicerInArgs.policerMatch.portIsTrunk = FALSE;
        memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], testSrcAddr,
               sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
        setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);
        memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], testEgressPortDestAddr_Policer1,
               sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
        setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_EGRESS_PORT);
        setPolicerInArgs.threadIdEnable                        = FALSE;
        setPolicerInArgs.peakRateInBitsPerSec                  = RECV_BIT_RATE_PIR_TEST1;
        setPolicerInArgs.commitRateInBitsPerSec                = RECV_BIT_RATE_CIR_TEST1;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                               status);
        }
    }

    return status;
}

static int32_t CpswTestPolicer_setPolicerAleEntryType2(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], testHostPortDestAddr_Policer2,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = ETH_TEST_IVLAN_ID;
    setUcastInArgs.info.portNum = CPSW_ALE_HOST_PORT_NUM;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], testEgressPortDestAddr_Policer2,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = ETH_TEST_IVLAN_ID;
    setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_EGRESS_PORT);
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           status);
    }

    if (CPSW_SOK == status)
    {
        memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
        setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_IVLAN |
                                                                CPSW_ALE_POLICER_MATCH_ETHERTYPE);
        setPolicerInArgs.policerMatch.ivlanId             = ETH_TEST_IVLAN_ID;
        setPolicerInArgs.policerMatch.etherType.etherType = ETHERTYPE_EXPERIMENTAL1;
        setPolicerInArgs.threadIdEnable                   = TRUE;
        setPolicerInArgs.threadId                         = taskObj->stateObj.rxFlowObj[CPSWTESTPOLICER_TYPE2_RXFLOWCFGID].rxFlowIdx;
        setPolicerInArgs.peakRateInBitsPerSec             = RECV_BIT_RATE_PIR_TEST2;
        setPolicerInArgs.commitRateInBitsPerSec           = RECV_BIT_RATE_CIR_TEST2;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                               status);
        }
    }

    return status;
}

static int32_t CpswTestPolicer_setPolicerAleEntryType3(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Set Policer entry */
    memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
    setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_OUI |
                                                            CPSW_ALE_POLICER_MATCH_IPDST);
    memcpy(&setPolicerInArgs.policerMatch.oui.ouiAddr[0U], testOuiAddr,
           sizeof(setPolicerInArgs.policerMatch.oui.ouiAddr));
    memcpy(&setPolicerInArgs.policerMatch.dstIp.ipv4.ipv4Addr[0U], testDestIpv4Addr,
           sizeof(setPolicerInArgs.policerMatch.dstIp.ipv4.ipv4Addr));
    setPolicerInArgs.policerMatch.dstIp.ipv4.numLSBIgnoreBits = 0;
    setPolicerInArgs.policerMatch.dstIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;

    setPolicerInArgs.threadIdEnable         = TRUE;
    setPolicerInArgs.threadId               = taskObj->stateObj.rxFlowObj[CPSWTESTPOLICER_TYPE3_RXFLOWCFGID].rxFlowIdx;
    setPolicerInArgs.peakRateInBitsPerSec   = RECV_BIT_RATE_PIR_TEST3;
    setPolicerInArgs.commitRateInBitsPerSec = RECV_BIT_RATE_CIR_TEST3;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_setPolicyEntry() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    return status;
}

static int32_t CpswTestPolicer_Test(CpswTestTaskObj *taskObj,
                                    uint32_t rxFlowCfgId,
                                    uint32_t recvbitRatePir,
                                    uint32_t recvbitRateCir)
{
    CpswTestRxFlowObj *rxFlowObj;

    rxFlowObj = &taskObj->stateObj.rxFlowObj[rxFlowCfgId];
    if ((rxFlowObj->receivedBitRate >= recvbitRateCir)
        &&
        (rxFlowObj->receivedBitRate <= (recvbitRatePir * 2)))
    {
        return CPSW_SOK;
    }
    else
    {
        /* The bitrate calculation is not correct when done from host so
         * disable returning failure for now
         */
        CpswAppUtils_print("CpswTestPolicer_Test: Rate Limit check fail."
                           "FlowCfgId:%d, PIR:%d, CIR:%d, Measured:%d \n",
                           rxFlowCfgId, recvbitRatePir, recvbitRateCir, rxFlowObj->receivedBitRate);
        return CPSW_SOK;
    }
}

void CpswTestPolicer_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                       Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

int32_t CpswTestPolicer_setRxOpenParams(CpswDma_OpenRxFlowPrms *pRxFlowPrms,
                                        CpswTestTaskObj *taskObj,
                                        uint32_t rxFlowCfgIdx)
{
    pRxFlowPrms->useProxy = TRUE;

    return CPSW_SOK;
}

int32_t CpswTestPolicer_setTxOpenParams(CpswDma_OpenTxChPrms *pTxChPrms,
                                        CpswTestTaskObj *taskObj,
                                        uint32_t txPSILThreadId,
                                        uint32_t txChCfgIdx)
{
    pTxChPrms->useProxy = TRUE;

    return CPSW_SOK;
}

int32_t CpswTestPolicer_setPolicerRange(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswAle_PolicerMatchParams getPolicerInArgs;
    CpswAle_GetPolicerEntryOutArgs getPolicerOutArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t rangeId = 0;

    status = CPSW_SOK;
    while ((RECV_BIT_RATE_RANGE_BASE_CIR > (rangeId * RECV_BIT_RATE_RANGE_INTERVAL))
           &&
           (CPSW_SOK == status))
    {
        memset(&setPolicerInArgs, 0, sizeof(setPolicerInArgs));
        memset(&setPolicerOutArgs, 0, sizeof(setPolicerOutArgs));
        setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_IPSRC |
                                                                CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                CPSW_ALE_POLICER_MATCH_MACDST);

        memcpy(&setPolicerInArgs.policerMatch.srcIp.ipv4.ipv4Addr[0U], testSrcIpv4AddrRange,
               sizeof(setPolicerInArgs.policerMatch.srcIp.ipv4.ipv4Addr));
        setPolicerInArgs.policerMatch.srcIp.ipv4.ipv4Addr[3U] += rangeId;
        setPolicerInArgs.policerMatch.srcIp.ipv4.numLSBIgnoreBits = 0;
        setPolicerInArgs.policerMatch.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;

        memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr , testSrcAddrRange,
               sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId = 0;
        setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);

        memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr , testEgressPortDestAddrRange,
               sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId = 0;
        setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_EGRESS_PORT);

        setPolicerInArgs.peakRateInBitsPerSec   = (RECV_BIT_RATE_RANGE_BASE_PIR - (rangeId * RECV_BIT_RATE_RANGE_INTERVAL));
        setPolicerInArgs.commitRateInBitsPerSec = (RECV_BIT_RATE_RANGE_BASE_CIR - (rangeId * RECV_BIT_RATE_RANGE_INTERVAL));

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_setPolicyEntry() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                               status);
        }
        if (CPSW_SOK == status)
        {
            memset(&getPolicerInArgs, 0, sizeof(getPolicerInArgs));
            memset(&getPolicerOutArgs, 0, sizeof(getPolicerOutArgs));
            getPolicerInArgs.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_IPSRC |
                                                       CPSW_ALE_POLICER_MATCH_MACSRC |
                                                       CPSW_ALE_POLICER_MATCH_MACDST);
            memcpy(&getPolicerInArgs.srcIp.ipv4.ipv4Addr[0U], testSrcIpv4AddrRange,
                   sizeof(getPolicerInArgs.srcIp.ipv4.ipv4Addr));
            getPolicerInArgs.srcIp.ipv4.ipv4Addr[3U] += rangeId;
            getPolicerInArgs.srcIp.ipv4.numLSBIgnoreBits = 0;
            getPolicerInArgs.srcIp.ipAddrtype            = CPSW_ALE_IPADDR_CLASSIFIER_IPV4;

            memcpy(&getPolicerInArgs.srcMacAddr.addr.addr , testSrcAddrRange,
                   sizeof(getPolicerInArgs.srcMacAddr.addr.addr));
            getPolicerInArgs.srcMacAddr.addr.vlanId = 0;
            getPolicerInArgs.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_INGRESS_PORT);

            memcpy(&getPolicerInArgs.dstMacAddr.addr.addr , testEgressPortDestAddrRange,
                   sizeof(getPolicerInArgs.dstMacAddr.addr.addr));
            getPolicerInArgs.dstMacAddr.addr.vlanId = 0;
            getPolicerInArgs.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_POLICER_EGRESS_PORT);

            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getPolicerInArgs, &getPolicerOutArgs);

            status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_GET_POLICER,
                                &prms);
            if (status != CPSW_SOK)
            {
                CpswAppUtils_print("CpswApp_getPolicyEntry() failed CPSW_ALE_IOCTL_GET_POLICER: %d\n",
                                   status);
            }
            else
            {
                CpswAppUtils_assert(setPolicerOutArgs.policerEntryIndex ==
                                    getPolicerOutArgs.policerEntryIndex);
                CpswAppUtils_assert(setPolicerInArgs.policerMatch.policerMatchEnableMask ==
                                    getPolicerOutArgs.policerMatchEnableMask);
                CpswAppUtils_assert(setPolicerOutArgs.srcIpAleEntryIndex ==
                                    getPolicerOutArgs.srcIpAleEntryIndex);
                CpswAppUtils_assert(setPolicerOutArgs.srcMacAleEntryIndex ==
                                    getPolicerOutArgs.srcMacAleEntryIndex);
                CpswAppUtils_assert(setPolicerOutArgs.dstMacAleEntryIndex ==
                                    getPolicerOutArgs.dstMacAleEntryIndex);

                if (setPolicerInArgs.peakRateInBitsPerSec != 0)
                {
                    CpswAppUtils_assert((setPolicerInArgs.peakRateInBitsPerSec / getPolicerOutArgs.peakRateInBitsPerSec) == 1);
                }
                else
                {
                    CpswAppUtils_assert(getPolicerOutArgs.peakRateInBitsPerSec == 0);
                }

                if (setPolicerInArgs.commitRateInBitsPerSec != 0)
                {
                    CpswAppUtils_assert((setPolicerInArgs.commitRateInBitsPerSec / getPolicerOutArgs.commitRateInBitsPerSec) == 1);
                }
                else
                {
                    CpswAppUtils_assert(getPolicerOutArgs.commitRateInBitsPerSec == 0);
                }
            }
        }
        rangeId++;
    }
    return status;
}

