/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_policer_nomatch.c
 *
 * \brief    This file contains the cpsw_policer nomatch test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_policer_nomatch.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
#define NOMATCHRED_TC_ID                           17U
#define NOMATCHYELLOW_TC_ID                        18U
#define NOMATCHPOLICER_TC_ID                       19U

#define CPSW_TEST_POLICER_MBPS(x)                   ((x) * 1000000U)
#define RECV_BIT_RATE_DEFAULT_PIR                   CPSW_TEST_POLICER_MBPS(200)
#define RECV_BIT_RATE_DEFAULT_CIR                   CPSW_TEST_POLICER_MBPS(100)

#define RECV_BIT_RATE_FLOWTYPE1_PIR                 CPSW_TEST_POLICER_MBPS(500)
#define RECV_BIT_RATE_FLOWTYPE1_CIR                 CPSW_TEST_POLICER_MBPS(250)

#define CPSW_TEST_NOMATCHPOLICER_INGRESS_PORT       (CPSW_MAC_PORT_3)
#define CPSW_TEST_NOMATCHPOLICER_EGRESS_PORT        (CPSW_MAC_PORT_2)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswTestPolicerNomatch_Test(CpswTestTaskObj *taskObj,
                                           uint32_t rxFlowCfgId,
                                           uint32_t recvbitRatePir,
                                           uint32_t recvbitRateCir);

static int32_t CpswTestPolicerNomatch_setPolicerAleEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
static uint8_t testSrcAddr[CPSW_MAC_ADDR_LEN] =
{
    0x02, 0x00, 0x00, 0x00, 0x00, 0x00
};

static uint8_t testHostPortDestAddr[CPSW_MAC_ADDR_LEN] =
{
    0x04, 0x00, 0x00, 0x00, 0x00, 0x00
};

static uint8_t testEgressPortDestAddr[CPSW_MAC_ADDR_LEN] =
{
    0x04, 0x00, 0x00, 0x00, 0x00, 0x02
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestPolicerNomatch_Run(CpswTestTaskObj *taskObj)
{
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t rxReadyCnt, timeoutVal, elapsedTIme;

    status = CpswTestPolicerNomatch_setPolicerAleEntry(taskObj);

    if (CPSW_SOK == status)
    {
        Cpsw_IoctlPrms prms;

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswAppUtils_assert(status == CPSW_SOK);

        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES,
                            &prms);

        CpswAppUtils_assert(status == CPSW_SOK);
    }

    if ((taskObj->testParams->tcId == NOMATCHRED_TC_ID) ||
        (taskObj->testParams->tcId == NOMATCHYELLOW_TC_ID))
    {
        CpswTestCommon_waitForPortLink(taskObj);

        /* Discard all queued packets before entering timeout loop 
         * There maybe packets queued during the validateRxFilter 
         * Rx filter transition
         */
        CpswTestCommon_discardRxPkts(taskObj, CPSWTESTPOLICER_NOMATCH_DEFAULT_RXFLOWCFGID);

        timeoutVal  = 1 * 2 * 60 * 1000;
        elapsedTIme = 0;

        do
        {
            rxReadyCnt = CpswTestCommon_receivePkts(taskObj, CPSWTESTPOLICER_NOMATCH_DEFAULT_RXFLOWCFGID);
            if (rxReadyCnt != 0)
            {
                status = CPSW_EFAIL;
                break;
            }
            else
            {
                CpswUtApp_wait(10);
                elapsedTIme += 10;
            }
        }
        while (elapsedTIme < timeoutVal);
    }
    else if (taskObj->testParams->tcId == NOMATCHPOLICER_TC_ID)
    {
        if (status == CPSW_SOK)
        {
            status = CpswTestCommon_Run(taskObj);
        }

        if (status == CPSW_SOK)
        {
            status = CpswTestPolicerNomatch_Test(taskObj, CPSWTESTPOLICER_NOMATCH_DEFAULT_RXFLOWCFGID, RECV_BIT_RATE_DEFAULT_PIR, RECV_BIT_RATE_DEFAULT_CIR);

            if (status == CPSW_SOK)
            {
                status = CpswTestPolicerNomatch_Test(taskObj, CPSWTESTPOLICER_NOMATCH_TYPE1_RXFLOWCFGID, RECV_BIT_RATE_FLOWTYPE1_PIR, RECV_BIT_RATE_FLOWTYPE1_CIR);
            }
        }
    }

    return status;
}

void CpswTestPolicerNomatch_setOpenPrms(CpswTestTaskObj *taskObj,
                                        Cpsw_Config *pCpswCfg,
                                        CpswOsal_Prms *pOsalPrms,
                                        CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->aleConfig.policerGlobalConfig.policingEnable   = TRUE;
    pCpswCfg->aleConfig.policerGlobalConfig.redDropEnable    = TRUE;
    pCpswCfg->aleConfig.policerGlobalConfig.yellowDropEnable = FALSE;

    switch (taskObj->testParams->tcId)
    {
        case NOMATCHRED_TC_ID:
            pCpswCfg->aleConfig.policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_RED;
            break;

        case NOMATCHYELLOW_TC_ID:
            pCpswCfg->aleConfig.policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_YELLOW;
            pCpswCfg->aleConfig.policerGlobalConfig.yellowThresh       = CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_100;
            pCpswCfg->aleConfig.policerGlobalConfig.yellowDropEnable   = TRUE;

            break;

        case NOMATCHPOLICER_TC_ID:
            pCpswCfg->aleConfig.policerGlobalConfig.policerNoMatchMode                    = CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER;
            pCpswCfg->aleConfig.policerGlobalConfig.noMatchPolicer.peakRateInBitsPerSec   = RECV_BIT_RATE_DEFAULT_PIR;
            pCpswCfg->aleConfig.policerGlobalConfig.noMatchPolicer.commitRateInBitsPerSec = RECV_BIT_RATE_DEFAULT_CIR;
            break;
    }
}

void CpswTestPolicerNomatch_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                              Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

static int32_t CpswTestPolicerNomatch_setPolicerAleEntry(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPolicerEntryOutArgs setPolicerOutArgs;
    CpswAle_SetPolicerEntryInArgs setPolicerInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                                            CPSW_ALE_POLICER_MATCH_MACSRC |
                                                            CPSW_ALE_POLICER_MATCH_MACDST);
    setPolicerInArgs.policerMatch.portNum     = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_NOMATCHPOLICER_INGRESS_PORT);
    setPolicerInArgs.policerMatch.portIsTrunk = FALSE;
    memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], testSrcAddr,
           sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
    setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_NOMATCHPOLICER_INGRESS_PORT);
    memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], testHostPortDestAddr,
           sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
    setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
    setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_HOST_PORT_NUM;
    setPolicerInArgs.threadIdEnable                        = TRUE;
    setPolicerInArgs.threadId                              = taskObj->stateObj.rxFlowObj[CPSWTESTPOLICER_NOMATCH_TYPE1_RXFLOWCFGID].rxFlowIdx;
    setPolicerInArgs.peakRateInBitsPerSec                  = RECV_BIT_RATE_FLOWTYPE1_PIR;
    setPolicerInArgs.commitRateInBitsPerSec                = RECV_BIT_RATE_FLOWTYPE1_CIR;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        setPolicerInArgs.policerMatch.policerMatchEnableMask = (CPSW_ALE_POLICER_MATCH_PORT |
                                                                CPSW_ALE_POLICER_MATCH_MACSRC |
                                                                CPSW_ALE_POLICER_MATCH_MACDST);
        setPolicerInArgs.policerMatch.portNum     = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_NOMATCHPOLICER_INGRESS_PORT);
        setPolicerInArgs.policerMatch.portIsTrunk = FALSE;
        memcpy(&setPolicerInArgs.policerMatch.srcMacAddr.addr.addr[0U], testSrcAddr,
               sizeof(setPolicerInArgs.policerMatch.srcMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.srcMacAddr.addr.vlanId    = 0;
        setPolicerInArgs.policerMatch.srcMacAddr.ingressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_NOMATCHPOLICER_INGRESS_PORT);
        memcpy(&setPolicerInArgs.policerMatch.dstMacAddr.addr.addr[0U], testEgressPortDestAddr,
               sizeof(setPolicerInArgs.policerMatch.dstMacAddr.addr.addr));
        setPolicerInArgs.policerMatch.dstMacAddr.addr.vlanId   = 0;
        setPolicerInArgs.policerMatch.dstMacAddr.egressPortNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSW_TEST_NOMATCHPOLICER_EGRESS_PORT);
        setPolicerInArgs.threadIdEnable                        = FALSE;
        setPolicerInArgs.peakRateInBitsPerSec                  = RECV_BIT_RATE_FLOWTYPE1_PIR;
        setPolicerInArgs.commitRateInBitsPerSec                = RECV_BIT_RATE_FLOWTYPE1_CIR;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setPolicerInArgs, &setPolicerOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_SET_POLICER,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswUtApp_openCpsw() failed CPSW_ALE_IOCTL_SET_POLICER: %d\n",
                               status);
        }
    }

    return status;
}

int32_t CpswTestPolicerNomatch_processRxPkt(CpswTestTaskObj *taskObj,
                                            uint32_t rxFlowId,
                                            uint32_t pktNum,
                                            CpswDma_PktInfo *pktInfo,
                                            bool *testComplete)
{
    int32_t status = CPSW_SOK;
    EthFrame *frame;

    /* Consume the packet by just printing its content */
    frame = (EthFrame *)pktInfo->bufPtr;

#ifdef ENABLE_PRINTFRAME
    CpswAppUtils_printFrame
        (frame, (pktInfo->userBufLen - sizeof(EthFrameHeader)));
#endif

    if (rxFlowId == CPSWTESTPOLICER_NOMATCH_TYPE1_RXFLOWCFGID)
    {
        if ((memcmp(&frame->hdr.dstMac[0], &testHostPortDestAddr, sizeof(frame->hdr.dstMac)) == 0) &&
            (memcmp(&frame->hdr.srcMac[0], &testSrcAddr, sizeof(frame->hdr.dstMac)) == 0))
        {
            status = CPSW_SOK;
        }
        else
        {
            status = CPSW_EFAIL;
        }
    }
    else
    {
        status = CPSW_SOK;
    }

    *testComplete = FALSE;
    return status;
}

static int32_t CpswTestPolicerNomatch_Test(CpswTestTaskObj *taskObj,
                                           uint32_t rxFlowCfgId,
                                           uint32_t recvbitRatePir,
                                           uint32_t recvbitRateCir)
{
    CpswTestRxFlowObj *rxFlowObj;

    rxFlowObj = &taskObj->stateObj.rxFlowObj[rxFlowCfgId];
    if ((rxFlowObj->receivedBitRate >= recvbitRateCir)
        &&
        (rxFlowObj->receivedBitRate <= (recvbitRatePir * 2)))
    {
        return CPSW_SOK;
    }
    else
    {
        /* The bitrate calculation is not correct when done from host so
         * disable returning failure for now
         */
        CpswAppUtils_print("CpswTestPolicer_Test: Rate Limit check fail."
                           "FlowCfgId:%d, PIR:%d, CIR:%d, Measured:%d \n",
                           rxFlowCfgId, recvbitRatePir, recvbitRateCir, rxFlowObj->receivedBitRate);
        return CPSW_SOK;
    }
}
