/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_priority_regeneration.c
 *
 * \brief    This file contains the cpsw_test_priority_regeneration test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <xdc/std.h>
#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethpatterns.h>
#include <ti/osal/osal.h>
#include <ti/board/board.h>

#include "cpsw_test_base.h"
#include "cpsw_test_entry.h"
#include "cpsw_test_priority_regeneration.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define ETH_TEST_MAC_PORT_VLAN_ID_BASE            (100U)
#define ETH_TEST_HOST_PORT_VLAN_ID                (200U)
#define FRAME_HEADER_NB_TO_SHIFT                  (13U)
#define FRAME_HEADER_PRIORITY                     (0x7U)

#define CPSWTEST_PRIORITYREGENERATION_RX_RECV_COUNT  (10000)

#define CPSWTEST_PRIORITYREGENERATION_EGRESS_PORT     (CPSW_MAC_PORT_2)
#define CPSWTEST_PRIORITYREGENERATION_INGRESS_PORT    (CPSW_MAC_PORT_3)
#define CPSWTEST_PRIORITYREGENERATION_VLANID          (1200U)
#define CPSWTEST_PRIORITYREGENERATION_DEFAULT_SHORTIPG_THRESHOLD    (11)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

static void CpswTestPriorityRegeneration_validateTosMap(const uint32_t *tosMap, uint32_t alePortNum);

static void CpswTestPriorityRegeneration_validateRxPri2HdrPktPri(const uint32_t *priorityMap, uint32_t alePortNum);

static void CpswTestPriorityRegeneration_validateHdrPktPri2SwitchPri(const uint32_t *priorityMap, uint32_t alePortNum);

static int32_t CpswTestPriorityRegeneration_getHostRxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_getMacRxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_getHostRxDscpPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_getMacRxDscpPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_getHostTxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_getMacTxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapHostRxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapMacRxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapHostRxDscpPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapMacRxDscpPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapHostTxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_remapMacTxPriority(CpswTestTaskObj *taskObj);

static int32_t CpswTestPriorityRegeneration_xmitTest(CpswTestTaskObj *taskObj,
                                                     uint32_t txCfgIndex,
                                                     uint32_t num,
                                                     const uint8_t *dstAddr,
                                                     const uint8_t *srcAddr,
                                                     uint16_t pcp,
                                                     uint16_t vid,
                                                     uint16_t etherType,
                                                     uint16_t length);

static int32_t CpswTestPriorityRegeneration_transmitPkts(CpswTestTaskObj *taskObj,
                                                         CpswDma_PktInfoQ *pTxSubmitQ,
                                                         uint32_t txCfgIndex);

static void CpswTestPriorityRegeneration_transmitTest(CpswTestTaskObj *taskObj,
                                                      uint32_t txCfgIndex,
                                                      uint16_t pcp);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Broadcast address */
static uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0xFFU, 0xFFU, 0xFFU, 0xFFU, 0xFFU, 0xFFU
};

/* "RX Packet Priority" to "Header Packet Priority" mapping */
static uint32_t mapRxPktPri2HdrPktPri[CPSW_PACKET_PRIORITY_NUM] =
{
    CPSW_HEADER_PACKET_PRIORITY_7, /* 0 -> 7 */
    CPSW_HEADER_PACKET_PRIORITY_6, /* 1 -> 6 */
    CPSW_HEADER_PACKET_PRIORITY_5, /* 2 -> 5 */
    CPSW_HEADER_PACKET_PRIORITY_4, /* 3 -> 4 */
    CPSW_HEADER_PACKET_PRIORITY_3, /* 4 -> 3 */
    CPSW_HEADER_PACKET_PRIORITY_2, /* 5 -> 2 */
    CPSW_HEADER_PACKET_PRIORITY_1, /* 6 -> 1 */
    CPSW_HEADER_PACKET_PRIORITY_0, /* 7 -> 0 */
};

/* "Header Packet Priority" to "Switch Priority" mapping for host port */
static uint32_t mapHostHdrPktPri2SwitchPri[CPSW_HEADER_PACKET_PRIORITY_NUM] =
{
    CPSW_SWITCH_PRIORITY_1, /* 0 -> 1 */
    CPSW_SWITCH_PRIORITY_1,
    CPSW_SWITCH_PRIORITY_3, /* 2 -> 3 */
    CPSW_SWITCH_PRIORITY_3,
    CPSW_SWITCH_PRIORITY_5, /* 4 -> 5 */
    CPSW_SWITCH_PRIORITY_5,
    CPSW_SWITCH_PRIORITY_7, /* 6 -> 7 */
    CPSW_SWITCH_PRIORITY_7,
};

/* "Header Packet Priority" to "Switch Priority" mapping for MAC ports */
static uint32_t mapHdrPktPri2SwitchPri[CPSW_MAC_PORT_NUM][CPSW_HEADER_PACKET_PRIORITY_NUM] =
{
    {                           /* CPSW_MAC_PORT_0 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_0, /* 1 -> 0 */
        CPSW_SWITCH_PRIORITY_0, /* 2 -> 0 */
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_7,
    },
    {   /* CPSW_MAC_PORT_1 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_1, /* 2 -> 1 */
        CPSW_SWITCH_PRIORITY_1, /* 3 -> 1 */
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_7,
    },
    {   /* CPSW_MAC_PORT_2 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_2, /* 3 -> 2 */
        CPSW_SWITCH_PRIORITY_2, /* 4 -> 2 */
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_7,
    },
    {   /* CPSW_MAC_PORT_3 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_3, /* 4 -> 3 */
        CPSW_SWITCH_PRIORITY_3, /* 5 -> 3 */
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_7,
    },
    {   /* CPSW_MAC_PORT_4 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_4, /* 5 -> 4 */
        CPSW_SWITCH_PRIORITY_4, /* 6 -> 4 */
        CPSW_SWITCH_PRIORITY_7,
    },
    {   /* CPSW_MAC_PORT_5 */
        CPSW_SWITCH_PRIORITY_0,
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_5, /* 6 -> 5 */
        CPSW_SWITCH_PRIORITY_5, /* 7 -> 5 */
    },
    {                           /* CPSW_MAC_PORT_6 */
        CPSW_SWITCH_PRIORITY_6, /* 0 -> 6 */
        CPSW_SWITCH_PRIORITY_1,
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_6, /* 7 -> 6 */
    },
    {                           /* CPSW_MAC_PORT_7 */
        CPSW_SWITCH_PRIORITY_0, /* 0 -> 7 */
        CPSW_SWITCH_PRIORITY_1, /* 1 -> 7 */
        CPSW_SWITCH_PRIORITY_2,
        CPSW_SWITCH_PRIORITY_3,
        CPSW_SWITCH_PRIORITY_4,
        CPSW_SWITCH_PRIORITY_5,
        CPSW_SWITCH_PRIORITY_6,
        CPSW_SWITCH_PRIORITY_7,
    },
};

static uint8_t testIngressPortMacAddr[] = {0x00, 0x11, 0x01, 0x00, 0x00, 0x01};
static uint8_t testEgressPortMacAddr[] = {0x00, 0x11, 0x02, 0x00, 0x00, 0x01};
/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
static uint32_t CpswTestPriorityRegeneration_getVlanMembershipMask(void)
{
    uint32_t memberShipMask;

    memberShipMask =
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSWTEST_PRIORITYREGENERATION_INGRESS_PORT));
    memberShipMask |=
        (1 << CPSW_ALE_MACPORT_TO_ALEPORT(CPSWTEST_PRIORITYREGENERATION_EGRESS_PORT));
    memberShipMask |= CPSW_ALE_HOST_PORT_MASK;
    return memberShipMask;
}


static int32_t CpswTestPriorityRegeneration_setShortIPG(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    Cpsw_SetShortTxIPGConfigInArgs setShortIPGInArgs;
    int32_t status;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &setShortIPGInArgs);
    setShortIPGInArgs.configureGapThreshold                           = TRUE;
    setShortIPGInArgs.ipgTriggerThresholdBlkCount                     = 0;
    setShortIPGInArgs.numMacPorts                                     = 2;
    setShortIPGInArgs.portIPGCfg[0].portNum                           = CPSWTEST_PRIORITYREGENERATION_INGRESS_PORT;
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[0].shortGapCfg.enableTxShortGapLimit = false;

    setShortIPGInArgs.portIPGCfg[1].portNum                           = CPSWTEST_PRIORITYREGENERATION_EGRESS_PORT;
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGap      = true;
    setShortIPGInArgs.portIPGCfg[1].shortGapCfg.enableTxShortGapLimit = false;

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_SET_SHORT_IPG_CONFIG,
                        &prms);
    if (CPSW_SOK == status)
    {
        Cpsw_GetShortTxIPGConfigOutArgs getShortIPGOutArgs;

        CPSW_IOCTL_SET_OUT_ARGS(&prms, &getShortIPGOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_IOCTL_GET_SHORT_IPG_CONFIG,
                            &prms);
        if (CPSW_SOK == status)
        {
            uint32_t i;

            CpswAppUtils_assert(getShortIPGOutArgs.ipgTriggerThresholdBlkCount == 0);
            for (i = 0; i < getShortIPGOutArgs.numMacPorts; i++)
            {
                if ((getShortIPGOutArgs.portIPGCfg[i].portNum == CPSWTEST_PRIORITYREGENERATION_INGRESS_PORT)
                    ||
                    (getShortIPGOutArgs.portIPGCfg[i].portNum == CPSWTEST_PRIORITYREGENERATION_EGRESS_PORT))
                {
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGap == true);
                    CpswAppUtils_assert(getShortIPGOutArgs.portIPGCfg[i].shortGapCfg.enableTxShortGapLimit == false);
                }
            }
        }
    }

    return status;
}

static int32_t CpswTestPriorityRegeneration_addAleEntries(CpswTestTaskObj *taskObj)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    int32_t dumpStatus;

    memset(&setUcastInArgs,0,sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], testEgressPortMacAddr,
           sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0;
    setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSWTEST_PRIORITYREGENERATION_EGRESS_PORT);
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("%s() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                           __func__, status);
    }

    if (status == CPSW_SOK)
    {
        memset(&setUcastInArgs,0,sizeof(setUcastInArgs));
        memcpy(&setUcastInArgs.addr.addr[0U], testIngressPortMacAddr,
               sizeof(setUcastInArgs.addr.addr));
        setUcastInArgs.addr.vlanId  = 0;
        setUcastInArgs.info.portNum = CPSW_ALE_MACPORT_TO_ALEPORT(CPSWTEST_PRIORITYREGENERATION_INGRESS_PORT);
        setUcastInArgs.info.blocked = false;
        setUcastInArgs.info.secure  = false;
        setUcastInArgs.info.super   = 0U;
        setUcastInArgs.info.ageable = false;
        setUcastInArgs.info.trunk   = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("%s() failed CPSW_ALE_IOCTL_ADD_UNICAST: %d\n",
                               __func__, status);
        }
    }
    if (status == CPSW_SOK)
    {
        CpswAle_VlanEntryInfo inArgs;
        CpswAle_AddEntryOutArgs outArgs;

        memset(&inArgs,0,sizeof(inArgs));
        inArgs.vlanIdInfo.vlanId        = CPSWTEST_PRIORITYREGENERATION_VLANID;
        inArgs.vlanIdInfo.outerVlanFlag = false;
        inArgs.vlanMemberList           = CpswTestPriorityRegeneration_getVlanMembershipMask();
        inArgs.unregMcastFloodMask      = CpswTestPriorityRegeneration_getVlanMembershipMask();
        inArgs.regMcastFloodMask        = CpswTestPriorityRegeneration_getVlanMembershipMask();
        inArgs.forceUntaggedEgressMask  = 0U;
        inArgs.noLearnMask              = 0U;
        inArgs.vidIngressCheck          = false;
        inArgs.limitIPNxtHdr            = false;
        inArgs.disallowIPFragmentation  = false;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

        status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("%s() failed ADD_VLAN ioctl failed: %d\n",
                               __func__, status);
        }
    }

    CPSW_IOCTL_SET_NO_ARGS(&prms);
    dumpStatus = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
    CpswAppUtils_assert(dumpStatus == CPSW_SOK);

    return status;
}



int32_t CpswTestPriorityRegeneration_run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;
    uint32_t rxReadyCnt = 0;

    status = CpswTestPriorityRegeneration_setShortIPG(taskObj);
    CpswAppUtils_assert(status == CPSW_SOK);

    status = CpswTestPriorityRegeneration_addAleEntries(taskObj);
    CpswAppUtils_assert(status == CPSW_SOK);

    CpswTestCommon_waitForPortLink(taskObj);

    /* Remap Host port Rx packet priority */
    status = CpswTestPriorityRegeneration_remapHostRxPriority(taskObj);
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_getHostRxPriority(taskObj);
    }

    /* Remap Mac port Rx packet priority */
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_remapMacRxPriority(taskObj);
        if (status == CPSW_SOK)
        {
            status = CpswTestPriorityRegeneration_getMacRxPriority(taskObj);
        }
    }

    /* Remap Host port IPDSCP Rx packet priority */
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_remapHostRxDscpPriority(taskObj);
        if (status == CPSW_SOK)
        {
            status = CpswTestPriorityRegeneration_getHostRxDscpPriority(taskObj);
        }
    }

    /* Remap MAC port IPDSCP Rx packet priority */
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_remapMacRxDscpPriority(taskObj);
        if (status == CPSW_SOK)
        {
            status = CpswTestPriorityRegeneration_getMacRxDscpPriority(taskObj);
        }
    }

    /* Remap Host port Tx packet priority */
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_remapHostTxPriority(taskObj);
        if (status == CPSW_SOK)
        {
            status = CpswTestPriorityRegeneration_getHostTxPriority(taskObj);
        }
    }

    /* Remap Mac port Tx packet priority */
    if (status == CPSW_SOK)
    {
        status = CpswTestPriorityRegeneration_remapMacTxPriority(taskObj);
        if (status == CPSW_SOK)
        {
            status = CpswTestPriorityRegeneration_getMacTxPriority(taskObj);
        }
    }

    CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);
    do
    {
        rxReadyCnt += CpswTestCommon_receivePkts(taskObj, 0);
        CpswTestCommon_discardRxPkts(taskObj, 0);
        CpswUtApp_wait(10);
    } while (rxReadyCnt < CPSWTEST_PRIORITYREGENERATION_RX_RECV_COUNT);

    /* Transmit packets with all PCP values.Packets should be captured on egress port and analyzed for correct PCP values */
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 0U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 1U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 2U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 3U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 4U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 5U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 6U);
    CpswTestPriorityRegeneration_transmitTest(taskObj, 0U, 7U);

    return status;
}

void CpswTestPriorityRegeneration_setOpenPrms(CpswTestTaskObj *taskObj,
                                              Cpsw_Config *pCpswCfg,
                                              CpswOsal_Prms *pOsalPrms,
                                              CpswUtils_Prms *pUtilsPrms)
{
    pCpswCfg->vlanConfig.vlanAware = true;

    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode          = TRUE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode         = TRUE;
    pCpswCfg->aleConfig.vlanConfig.unknownVlanMemberListMask = CPSW_ALE_ALL_PORTS_MASK;

    pCpswCfg->hostPortConfig.vlanCfg.portPri       = 3;
    pCpswCfg->hostPortConfig.vlanCfg.portCfi       = 0;
    pCpswCfg->hostPortConfig.vlanCfg.portVID       = ETH_TEST_HOST_PORT_VLAN_ID;
    pCpswCfg->hostPortConfig.enableRxVlanRemap     = true;
    pCpswCfg->hostPortConfig.enableRxDscpIPv4Remap = true;
    pCpswCfg->hostPortConfig.enableRxDscpIPv6Remap = true;
}

void CpswTestPriorityRegeneration_setMacConfig(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                               Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback     = FALSE;
    pLinkArgs->macConfig.vlanConfig.portPri = 3;
    pLinkArgs->macConfig.vlanConfig.portCfi = 0;
    pLinkArgs->macConfig.vlanConfig.portVID = ETH_TEST_MAC_PORT_VLAN_ID_BASE +
                                              CPSW_NORMALIZE_MACPORT(portNum);
}

static void CpswTestPriorityRegeneration_validateTosMap(const uint32_t *tosMap, uint32_t alePortNum)
{
    uint32_t i;

    for (i = 0U; i < CPSW_TOS_PRIORITY_NUM; i++)
    {
        CpswAppUtils_assert(tosMap[i] == (i % 8));
    }

}

static void CpswTestPriorityRegeneration_validateRxPri2HdrPktPri(const uint32_t *priorityMap, uint32_t alePortNum)
{
    uint32_t i;

    for (i = 0U; i < CPSW_PACKET_PRIORITY_NUM; i++)
    {
        CpswAppUtils_assert(priorityMap[i] == mapRxPktPri2HdrPktPri[i]);
    }
}

static void CpswTestPriorityRegeneration_validateHdrPktPri2SwitchPri(const uint32_t *priorityMap, uint32_t alePortNum)
{
    uint32_t i;

    for (i = 0U; i < CPSW_HEADER_PACKET_PRIORITY_NUM; i++)
    {
        if (alePortNum == CPSW_ALE_HOST_PORT_NUM)
        {
            CpswAppUtils_assert(priorityMap[i] == mapHostHdrPktPri2SwitchPri[i]);
        }
        else
        {
           CpswAppUtils_assert(alePortNum > CPSW_ALE_MACPORT_BASE);
           CpswAppUtils_assert((alePortNum - CPSW_ALE_MACPORT_BASE) < CPSW_UTILS_ARRAYSIZE(mapHdrPktPri2SwitchPri));
           CpswAppUtils_assert(priorityMap[i] == mapHdrPktPri2SwitchPri[alePortNum - CPSW_ALE_MACPORT_BASE][i]);
        }
    }

}

/* Get Host port Rx Packet priority */
static int32_t CpswTestPriorityRegeneration_getHostRxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_RxPriority getHostOutArgs;
    Cpsw_IoctlPrms prms;
    int32_t status;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &getHostOutArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_GET_RX_PRIORITY,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswTestPriorityRegeneration_validateRxPri2HdrPktPri(getHostOutArgs.priorityMap, CPSW_ALE_HOST_PORT_NUM);
    }
    else
    {
        CpswAppUtils_print("getHostRxPriority() CPSW_HOSTPORT_IOCTL_GET_RX_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Packet priority to header packet priority map */
static int32_t CpswTestPriorityRegeneration_remapHostRxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_RxPriority setHostInArgs;
    Cpsw_IoctlPrms prms;
    int32_t status;

    /* Copy the test "RX Packet Priority" to "Header Packet Priority" mapping */
    memcpy(setHostInArgs.priorityMap,
           mapRxPktPri2HdrPktPri,
           sizeof(mapRxPktPri2HdrPktPri));
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setHostInArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("remapHostRxPriority() CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Get MAC port Rx Packet priority */
static int32_t CpswTestPriorityRegeneration_getMacRxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_GenericInArgs getMacInArgs;
    CpswPort_RxPriority getMacOutArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        getMacInArgs.portNum = enabledPorts.macPortList[i];
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getMacInArgs, &getMacOutArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_GET_RX_PRIORITY,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswTestPriorityRegeneration_validateRxPri2HdrPktPri(getMacOutArgs.priorityMap, CPSW_ALE_MACPORT_TO_ALEPORT(enabledPorts.macPortList[i]));
        }
        else
        {
            CpswAppUtils_print("getMacRxPriority() CPSW_MACPORT_IOCTL_GET_RX_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

/* Packet priority to header packet priority map */
static int32_t CpswTestPriorityRegeneration_remapMacRxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_SetRxPriorityInArgs setMacInArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    /* Copy the test "RX Packet Priority" to "Header Packet Priority" mapping */
    memcpy(setMacInArgs.rxPriority.priorityMap,
           mapRxPktPri2HdrPktPri,
           sizeof(mapRxPktPri2HdrPktPri));

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    /* Set "RX Packet Priority" to "Header Packet Priority" mapping for all MAC ports */
    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        setMacInArgs.portNum = enabledPorts.macPortList[i];
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setMacInArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_SET_RX_PRIORITY,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("remapMacRxPriority() CPSW_MACPORT_IOCTL_SET_RX_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

/* Get Host port IPDSCP Rx Packet priority */
static int32_t CpswTestPriorityRegeneration_getHostRxDscpPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_DscpPriority getHostOutArgs;
    Cpsw_IoctlPrms prms;
    int32_t status;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &getHostOutArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswTestPriorityRegeneration_validateTosMap(getHostOutArgs.tosMap, CPSW_ALE_HOST_PORT_NUM);
    }
    else
    {
        CpswAppUtils_print("getHostRxDscpPriority() CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Packet priority to header packet priority map */
static int32_t CpswTestPriorityRegeneration_remapHostRxDscpPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_DscpPriority setHostInArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status;

    setHostInArgs.enableDscpIPv4 = TRUE;
    setHostInArgs.enableDscpIPv6 = TRUE;

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(setHostInArgs.tosMap); i++)
    {
        setHostInArgs.tosMap[i] = (i % 8);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &setHostInArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("remapHostRxDscpPriority() CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Get MAC port IPDSCP Rx Packet priority */
static int32_t CpswTestPriorityRegeneration_getMacRxDscpPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_GenericInArgs getMacInArgs;
    CpswPort_DscpPriority getMacOutArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);
    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        getMacInArgs.portNum = enabledPorts.macPortList[i];
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getMacInArgs, &getMacOutArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswTestPriorityRegeneration_validateTosMap(getMacOutArgs.tosMap, CPSW_ALE_MACPORT_TO_ALEPORT(enabledPorts.macPortList[i]));
        }
        else
        {
            CpswAppUtils_print("getMacRxDscpPriorit() CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

/* Packet priority to header packet priority map */
static int32_t CpswTestPriorityRegeneration_remapMacRxDscpPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_SetRxDscpPriorityInArgs setMacInArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    uint32_t j;
    int32_t status = CPSW_SOK;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    /* Set Mac Port DSCP priority mapping for all mac ports */
    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        setMacInArgs.portNum                     = enabledPorts.macPortList[i];
        setMacInArgs.dscpPriority.enableDscpIPv4 = TRUE;
        setMacInArgs.dscpPriority.enableDscpIPv6 = TRUE;

        for (j = 0U; j < CPSW_UTILS_ARRAYSIZE(setMacInArgs.dscpPriority.tosMap); j++)
        {
            setMacInArgs.dscpPriority.tosMap[j] = (j % CPSW_PRIORITY_NUM);
        }

        CPSW_IOCTL_SET_IN_ARGS(&prms, &setMacInArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("remapMacRxDscpPriority() CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

/* Get Host port Header packet priority */
static int32_t CpswTestPriorityRegeneration_getHostTxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_TxPriority getHostOutArgs;
    Cpsw_IoctlPrms prms;
    int32_t status;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &getHostOutArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_GET_TX_PRIORITY,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswTestPriorityRegeneration_validateHdrPktPri2SwitchPri(getHostOutArgs.priorityMap, CPSW_ALE_HOST_PORT_NUM);
    }
    else
    {
        CpswAppUtils_print("getHostTxPriority() CPSW_HOSTPORT_IOCTL_GET_TX_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Header packet priority to switch priority map */
static int32_t CpswTestPriorityRegeneration_remapHostTxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswPort_TxPriority setHostInArgs;
    Cpsw_IoctlPrms prms;
    int32_t status;

    /* Copy the test "Header Packet Priority" to "Switch Priority" mapping */
    memcpy(setHostInArgs.priorityMap,
           mapHostHdrPktPri2SwitchPri,
           sizeof(mapHostHdrPktPri2SwitchPri));
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setHostInArgs);

    status = Cpsw_ioctl(testStateObj->hCpsw,
                        testStateObj->coreId,
                        CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("remapHostTxPriority() CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY failed: %d\n",
                           status);
    }

    return status;
}

/* Get MAC port Header packet priority */
static int32_t CpswTestPriorityRegeneration_getMacTxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_GenericInArgs getMacInArgs;
    CpswPort_TxPriority getMacOutArgs;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        getMacInArgs.portNum = enabledPorts.macPortList[i];
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &getMacInArgs, &getMacOutArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_GET_TX_PRIORITY,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswTestPriorityRegeneration_validateHdrPktPri2SwitchPri(getMacOutArgs.priorityMap, CPSW_ALE_MACPORT_TO_ALEPORT(enabledPorts.macPortList[i]));
        }
        else
        {
            CpswAppUtils_print("getMacTxPriority() CPSW_MACPORT_IOCTL_GET_TX_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

/* Header packet priority to switch priority map */
static int32_t CpswTestPriorityRegeneration_remapMacTxPriority(CpswTestTaskObj *taskObj)
{
    CpswTestStateObj *testStateObj = &taskObj->stateObj;
    CpswTestMacPortList_t enabledPorts;
    CpswMacPort_SetTxPriorityInArgs setMacInArgs;
    Cpsw_MacPort portNum;
    Cpsw_IoctlPrms prms;
    uint32_t i;
    int32_t status = CPSW_SOK;

    CpswTestCommon_getMacPortList(taskObj->taskCfg->macPortMask, &enabledPorts);

    /* Set "Header Packet Priority" to "Switch Priority" mapping for all MAC ports */
    for (i = 0U; i < enabledPorts.numMacPorts; i++)
    {
        portNum = enabledPorts.macPortList[i];

        /* Copy the test "Header Packet Priority" to "Switch Priority" mapping */
        memcpy(setMacInArgs.txPriority.priorityMap,
               mapHdrPktPri2SwitchPri[portNum],
               sizeof(mapHdrPktPri2SwitchPri[portNum]));
        setMacInArgs.portNum = portNum;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setMacInArgs);

        status = Cpsw_ioctl(testStateObj->hCpsw,
                            testStateObj->coreId,
                            CPSW_MACPORT_IOCTL_SET_TX_PRIORITY,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("remapMacTxPriority() CPSW_MACPORT_IOCTL_SET_TX_PRIORITY failed: %d\n",
                               status);
        }
    }

    return status;
}

static int32_t CpswTestPriorityRegeneration_transmitPkts(CpswTestTaskObj *taskObj,
                                                         CpswDma_PktInfoQ *pTxSubmitQ,
                                                         uint32_t txCfgIndex)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t txRetrievePktCnt;
    uint32_t txCnt;
    int32_t status = CPSW_SOK;

    while (0U != CpswUtils_getQCount(pTxSubmitQ))
    {
        txCnt  = CpswUtils_getQCount(pTxSubmitQ);
        status = CpswAppUtils_submitTxPackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                              pTxSubmitQ);

        /* Retrieve TX free packets */
        if (status == CPSW_SOK)
        {
            txCnt            = txCnt - CpswUtils_getQCount(pTxSubmitQ);
            txRetrievePktCnt = 0U;
            while (txRetrievePktCnt != txCnt)
            {
                /* TODO this is not failure as HW is busy sending packets, we need
                 * to wait and again call retrieve packets */
                CpswUtApp_wait(1);
                txRetrievePktCnt += CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
#if DEBUG
                CpswAppUtils_print("failed to retrieve consumed transmit packets: %d\n", status);
#endif
            }
        }
        else
        {
            break;
        }
    }

    return status;
}

static int32_t CpswTestPriorityRegeneration_xmitTest(CpswTestTaskObj *taskObj,
                                                     uint32_t txCfgIndex,
                                                     uint32_t num,
                                                     const uint8_t *dstAddr,
                                                     const uint8_t *srcAddr,
                                                     uint16_t pcp,
                                                     uint16_t vid,
                                                     uint16_t etherType,
                                                     uint16_t length)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint16_t len               = length - ETH_TEST_DATA_HDR_LEN;
    uint32_t pktCnt            = 0U;
    EthVlanFrame *frame;
    DataFramePayload *payload;
    CpswDma_PktInfo *pktInfo;
    CpswDma_PktInfoQ txSubmitQ;
    CpswDma_PktInfoQ *txFreePktInfoQ = &stateObj->txChObj[txCfgIndex].txFreePktInfoQ;
    int32_t status                   = CPSW_SOK;

    while (pktCnt < num)
    {
        CpswUtils_initQ(&txSubmitQ);

        /* Dequeue one free TX Eth packet */
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(txFreePktInfoQ);

        while (NULL != pktInfo)
        {
            /* Fill the TX Eth frame with test content */
            frame = (EthVlanFrame *)pktInfo->bufPtr;
            memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
            memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
            frame->hdr.tpid      = htons(ETHERTYPE_VLAN_TAG);
            frame->hdr.tci       = htons((pcp << 13) | vid);
            frame->hdr.etherType = htons(etherType);

            payload       = (DataFramePayload *)frame->payload;
            payload->type = 0U; /* DataPattern1 */
            payload->len  = htons(len);
            memcpy(payload->data, Cpsw_AppDataPattern1, len);

            pktInfo->userBufLen = length + sizeof(EthVlanFrameHeader);
            pktInfo->appPriv    = &stateObj;

            /* Enqueue the packet for later transmission */
            CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

            pktCnt++;
            if (pktCnt >= num)
            {
                break;
            }

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(txFreePktInfoQ);
        }

        status = CpswTestPriorityRegeneration_transmitPkts(taskObj, &txSubmitQ, txCfgIndex);
    }

    return status;
}

static void CpswTestPriorityRegeneration_transmitTest(CpswTestTaskObj *taskObj,
                                                      uint32_t txCfgIndex,
                                                      uint16_t pcp)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    uint32_t count             = 5;
    uint16_t len               = 500U;
    uint32_t status;

    CpswUtApp_wait(10);

    /* Transmit frames */
    CpswAppUtils_print("transmitTest: transmitting %d packets with PCP %u\n", count, pcp);

    status = CpswTestPriorityRegeneration_xmitTest(taskObj,
                                                   txCfgIndex,
                                                   count,
                                                   bcastAddr,
                                                   &stateObj->hostMacAddr[0U],
                                                   pcp,
                                                   ETH_TEST_VLAN_VID,
                                                   ETHERTYPE_EXPERIMENTAL1,
                                                   len);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("transmitTest: failed to transmit %d packets\n", count);
    }
}
