/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_sanity.c
 *
 * \brief    This file contains the cpsw sanity test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethpatterns.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_sanity.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
uint32_t CpswUtApp_recvTest(CpswTestTaskObj *taskObj,
                            uint32_t rxCfgIndex);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/* Broadcast address */
static uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static int32_t CpswTestSanityMacSpeedCommon_Run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;

    /*fix for gcc unused variable warning */
    Cpsw_AppDataPattern2[0U] = Cpsw_AppDataPattern2[0U];
    Cpsw_AppDataPattern3[0U] = Cpsw_AppDataPattern3[0U];
    Cpsw_AppDataPattern4[0U] = Cpsw_AppDataPattern4[0U];

    CpswAppUtils_assert(taskObj->taskCfg->numTxCh == 1);
    CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);

    /* Send command indicating that we're ready to start the test stage */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_DUT_READY, 0);

    /* test_0001: Receive test packets from host app */
    CpswUtApp_test_0001(taskObj, 0, 0);

    /* test_0002: Transmit test packets to host app */
    CpswUtApp_test_0002(taskObj, 0, 0);

    /* test_0100: Transmit throughput */
    CpswUtApp_test_0100(taskObj, 0, 0);

    CpswUtApp_discardRxPkts(taskObj, 0);

    CpswUtApp_freeRxBuffer(taskObj, 0);

    return status;
}

void CpswTestSanity_updatePortLinkCfgAutoNeg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                             Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestSanity_updatePortLinkCfg100Mbps(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                             Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
    pLinkArgs->linkConfig.speed         = CPSW_SPEED_100MBIT;
    pLinkArgs->linkConfig.duplexity     = CPSW_DUPLEX_FULL;
}

int32_t CpswUtApp_transmitPkts(CpswTestTaskObj *taskObj,
                               CpswDma_PktInfoQ *pTxSubmitQ,
                               uint32_t txCfgIndex)
{
    uint32_t txRetrievePktCnt, txCnt;
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    while (0U != CpswUtils_getQCount(pTxSubmitQ))
    {
        txCnt  = CpswUtils_getQCount(pTxSubmitQ);
        status = CpswAppUtils_submitTxPackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                              pTxSubmitQ);
        /* Retrieve TX free packets */
        if (status == CPSW_SOK)
        {
            txCnt            = txCnt - CpswUtils_getQCount(pTxSubmitQ);
            txRetrievePktCnt = 0U;
            while (txRetrievePktCnt != txCnt)
            {
                /* TODO this is not failure as HW is busy sending packets, we need
                 * to wait and again call retrieve packets */
                CpswUtApp_wait(1);
                txRetrievePktCnt += CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
#if DEBUG
                CpswAppUtils_print("Failed to retrieve consumed transmit packets: %d\n",
                                   status);
#endif
            }
        }
        else
        {
            break;
        }
    }

    return status;
}

int32_t CpswTestMacSpeed_Run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;

    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswPhy_GenericInArgs inArgs;

    CpswTestCommon_waitForPortLink(taskObj);

    CpswAppUtils_print("Port 0");
    inArgs.portNum = CPSW_MAC_PORT_0;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs);
    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_PHY_IOCTL_PRINT_REGS,
                        &prms);

    CpswTestSanityMacSpeedCommon_Run(taskObj);

    return status;
}

int32_t CpswUtApp_xmitTest(CpswTestTaskObj *taskObj,
                           uint32_t txCfgIndex,
                           uint32_t num,
                           const uint8_t *dstAddr,
                           const uint8_t *srcAddr,
                           uint16_t etherType,
                           uint16_t length)
{
    int32_t retVal = CPSW_SOK;
    CpswDma_PktInfo *pktInfo;
    uint16_t len    = length - ETH_TEST_DATA_HDR_LEN;
    uint32_t pktCnt = 0U;
    DataFramePayload *payload;
    CpswDma_PktInfoQ txSubmitQ;
    EthFrame *frame;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    while (pktCnt < num)
    {
        CpswUtils_initQ(&txSubmitQ);

        /* Dequeue one free TX Eth packet */
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);

        while (NULL != pktInfo)
        {
            pktCnt++;
            /* Fill the TX Eth frame with test content */
            frame   = (EthFrame *)pktInfo->bufPtr;
            payload = (DataFramePayload *)frame->payload;
            memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
            memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
            frame->hdr.etherType = htons(etherType);

            payload->type = 0U; /* DataPattern1 */
            payload->len  = htons(len);
            memcpy(payload->data, Cpsw_AppDataPattern1, len);

            pktInfo->userBufLen = length + sizeof(EthFrameHeader);
            pktInfo->appPriv    = &stateObj;

            /* Enqueue the packet for later transmission */
            CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

            if (pktCnt >= num)
            {
                break;
            }

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);
        }

        retVal = CpswUtApp_transmitPkts(taskObj, &txSubmitQ, txCfgIndex);
    }

    return retVal;
}

uint32_t CpswUtApp_recvTest(CpswTestTaskObj *taskObj,
                            uint32_t rxCfgIndex)
{
    CpswDma_PktInfo *pktInfo;
    EthFrame *frame;
    uint32_t num = 0;
    uint8_t cmd;
    bool runTest = true;
    uint32_t rxReadyCnt;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_print("CpswUtApp_recvTest() receiving packets until STOP cmd\n");

    /* Wait for packet reception */
    while (runTest)
    {
        /* Get the packets received so far */
        rxReadyCnt = CpswTestCommon_receivePkts(taskObj, rxCfgIndex);
        if (rxReadyCnt > 0U)
        {
            /* Consume the received packets and release them */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
            while (NULL != pktInfo)
            {
                /* Consume the packet by just printing its content */
                frame = (EthFrame *)pktInfo->bufPtr;
                /* Release the received packet */
                if (ntohs(frame->hdr.etherType) == ETHERTYPE_EXP_CONTROL)
                {
                    cmd = frame->payload[0];
                    if (cmd == CTRL_FRAME_CMD_STOP)
                    {
                        CpswAppUtils_print("CpswUtApp_recvTest() STOP cmd has been received\n");
                        runTest = false;
                    }

                    if (cmd == CTRL_FRAME_CMD_PC_READY)
                    {
                        CpswAppUtils_print("CpswUtApp_recvTest() PC Ready cmd has been received\n");
                        runTest = false;
                    }
                }
                else
                {
                    /* Consume the packet by just printing its content */
#ifdef ENABLE_PRINTFRAME
                    CpswAppUtils_printFrame(frame, pktInfo->userBufLen -
                                            sizeof(EthFrameHeader));
#endif
                    num++;
                }

                CpswUtils_enQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ, &pktInfo->node);
                if (runTest)
                {
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
                }
                else
                {
                    break;
                }
            }

            /*Submit now processed buffers */
            {
                CpswAppUtils_submitRxPackets(stateObj->rxFlowObj[rxCfgIndex].hRxFlow,
                                             &stateObj->rxFlowObj[rxCfgIndex].rxFreeQ);
            }
        }

        CpswUtApp_wait(10U);
    }

    CpswAppUtils_print("CpswUtApp_recvTest() %u test packets received\n", num);

    return num;
}

void CpswUtApp_discardRxPkts(CpswTestTaskObj *taskObj,
                             uint32_t rxCfgIndex)
{
    CpswDma_PktInfo *pkt;
    uint32_t count;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Discard any packets sitting in rxReadyQ before the test starts */
    count = CpswUtils_getQCount(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
    if (count != 0U)
    {
        CpswAppUtils_print("CpswUtApp_discardRxPkts() rxReadyQ has %d packets, discarding all\n",
                           count);
        pkt = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
        while (pkt != NULL)
        {
            /* Release the received packet */
            CpswUtils_enQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ, &pkt->node);
            pkt = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
        }
    }
}

void CpswUtApp_sendCmd(CpswTestTaskObj *taskObj,
                       uint8_t cmd,
                       uint32_t txCfgIndex)
{
    CpswDma_PktInfo *pktInfo;
    EthFrame *frame;
    CtrlFramePayload *payload;
    int32_t retVal;
    CpswDma_PktInfoQ txSubmitQ;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Transmit a single packet */
    CpswUtils_initQ(&txSubmitQ);

    /* Dequeue one free TX Eth packet */
    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);
    frame   = (EthFrame *)pktInfo->bufPtr;
    payload = (CtrlFramePayload *)frame->payload;

    memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
    memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
    frame->hdr.etherType = htons(ETHERTYPE_EXP_CONTROL);

    memset(payload, 0, sizeof(CtrlFramePayload));
    payload->cmd = cmd;

    pktInfo->userBufLen = sizeof(CtrlFramePayload) + sizeof(EthFrameHeader);
    pktInfo->appPriv    = &stateObj;

    /* Enqueue the packet for later transmission */
    CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

    CpswUtApp_wait(1000);

    retVal = CpswUtApp_transmitPkts(taskObj, &txSubmitQ, txCfgIndex);

    if (retVal != CPSW_SOK)
    {
        CpswAppUtils_print("CpswUtApp_sendCmd() failed to send cmd %d: %d\n", cmd, retVal);
    }
}

void CpswUtApp_test_0001(CpswTestTaskObj *taskObj,
                         uint32_t txCfgIndex,
                         uint32_t rxCfgIndex)
{
    uint32_t num;

    CpswAppUtils_print("Test_0001: START\n");

    CpswUtApp_discardRxPkts(taskObj, rxCfgIndex);

    /* Send START cmd */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_START, txCfgIndex);

    /* Receive frames until STOP cmd is detected */
    num = CpswUtApp_recvTest(taskObj, rxCfgIndex);

    if (num != ETH_TEST_ITER_M_COUNT)
    {
        CpswAppUtils_print("Test_0001: received %d of %d frames\n", num, ETH_TEST_ITER_M_COUNT);
    }

    CpswAppUtils_print("Test_0001: END\n\n");
}

void CpswUtApp_test_0002(CpswTestTaskObj *taskObj,
                         uint32_t txCfgIndex,
                         uint32_t rxCfgIndex)
{
    uint32_t iterations = ETH_TEST_ITER_M_COUNT;
    uint16_t len        = 1500U;
    uint32_t retVal;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_print("Test_0002: START\n");

    CpswUtApp_discardRxPkts(taskObj, rxCfgIndex);

    /* Send START cmd */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_START, txCfgIndex);

    CpswUtApp_wait(10);

    /* Transmit frames */
    retVal = CpswUtApp_xmitTest(taskObj,
                                txCfgIndex,
                                iterations,
                                bcastAddr,
                                &stateObj->hostMacAddr[0U],
                                ETHERTYPE_EXPERIMENTAL1,
                                len);
    if (retVal != CPSW_SOK)
    {
        CpswAppUtils_print("Test_0002: failed to transmit %d packets\n", iterations);
    }

    /* Send STOP cmd */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_STOP, txCfgIndex);

    CpswAppUtils_print("Test_0002: END\n\n");
}

void CpswUtApp_test_0100(CpswTestTaskObj *taskObj,
                         uint32_t txCfgIndex,
                         uint32_t rxCfgIndex)
{
    CpswDma_PktInfoQ tempQ;
    CpswDma_PktInfo *pkt;
    EthFrame *frame;
    uint32_t iterations = ETH_TEST_ITER_L_COUNT;
    uint16_t len        = 1500U;
    uint32_t retVal;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswAppUtils_print("Test_0100: START\n");

    CpswUtApp_discardRxPkts(taskObj, rxCfgIndex);

    /* Send START cmd */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_START, txCfgIndex);

    CpswUtApp_wait(10);

    /* Transmit frames (only fill L2 header) */
    while (iterations > 0U)
    {
        CpswUtils_initQ(&tempQ);

        while (iterations > 0U)
        {
            pkt = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);
            if (pkt == NULL)
            {
                break;
            }

            frame = (EthFrame *)pkt->bufPtr;
            memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
            memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
            frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);
            pkt->userBufLen      = len + sizeof(EthFrameHeader);

            CpswUtils_enQ(&tempQ, &pkt->node);
            iterations--;
        }

        retVal = CpswUtApp_transmitPkts(taskObj, &tempQ, txCfgIndex);
        if (retVal != CPSW_SOK)
        {
            CpswAppUtils_print("Test_0100() failed to transmit: %d\n", retVal);
        }
    }

    /* Send STOP cmd */
    CpswUtApp_sendCmd(taskObj, CTRL_FRAME_CMD_STOP, txCfgIndex);

    CpswAppUtils_print("Test_0100: END\n\n");
}

void CpswUtApp_freeRxBuffer(CpswTestTaskObj *taskObj,
                            uint32_t rxCfgIndex)
{
    CpswDma_PktInfo *pktInfo;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ);
    while (pktInfo != NULL)
    {
        CpswAppMemUtils_freeEthPktFxn(pktInfo);
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ);
    }
}

int32_t CpswTestSanity_Run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;

    CpswTestCommon_waitForPortLink(taskObj);

    status = CpswTestSanityMacSpeedCommon_Run(taskObj);

    return status;
}

/* end of file */
