/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_test_vlan.c
 *
 * \brief    This file contains the cpsw vlan test implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>

#include <ti/drv/uart/UART_stdio.h>
#include <ti/csl/csl_cpswitch.h>

#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethpatterns.h>

#include <ti/osal/osal.h>

#include <ti/board/board.h>

#include "cpsw_test_entry.h"
#include "cpsw_test_base.h"
#include "cpsw_test_vlan.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define ETH_TEST_VLAN_ID                (100U)
#define ETH_TEST_VLAN_PRI               (3U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */
static int32_t CpswUtApp_xmitTest(CpswTestTaskObj *taskObj,
                                  uint32_t txCfgIndex,
                                  uint32_t num,
                                  const uint8_t *dstAddr,
                                  const uint8_t *srcAddr,
                                  uint16_t etherType,
                                  uint16_t length);

static int32_t CpswUtApp_transmitPkts(CpswTestTaskObj *taskObj,
                                      CpswDma_PktInfoQ *pTxSubmitQ,
                                      uint32_t txCfgIndex);

static void CpswApp_transmitTest(CpswTestTaskObj *taskObj,
                                 uint32_t txCfgIndex,
                                 uint32_t rxCfgIndex);

static void CpswApp_addAleVlanEntry(CpswTestTaskObj *taskObj);

static void CpswApp_removeAleVlanEntry(CpswTestTaskObj *taskObj);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/* Broadcast address */
static uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
{
    0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int32_t CpswTestVlan_Run(CpswTestTaskObj *taskObj)
{
    int32_t status = CPSW_SOK;

    CpswTestCommon_waitForPortLink(taskObj);

    /*fix for gcc unused variable warning */
    Cpsw_AppDataPattern2[0U] = Cpsw_AppDataPattern2[0U];
    Cpsw_AppDataPattern3[0U] = Cpsw_AppDataPattern3[0U];
    Cpsw_AppDataPattern4[0U] = Cpsw_AppDataPattern4[0U];

    /* Transmit untagged packets
     * VLAN tag will be inserted from port 0 default settings */
    CpswAppUtils_print("Transmit untagged packets\n");
    CpswAppUtils_print("  VLAN will be inserted from port 0 settings:\n");
    CpswAppUtils_print("  - VLAN ID  = %d\n", ETH_TEST_VLAN_ID);
    CpswAppUtils_print("  - VLAN Pri = %d\n", ETH_TEST_VLAN_PRI);
    CpswAppUtils_assert(taskObj->taskCfg->numTxCh == 1);
    CpswAppUtils_assert(taskObj->taskCfg->numRxFlow == 1);

    CpswApp_transmitTest(taskObj, 0, 0);

    /* Add ALE entry to force untagged packets on ETH_TEST_VLAN_ID */
    CpswAppUtils_print("Add ALE entry to force untagged packets on VLAN ID %d\n",
                       ETH_TEST_VLAN_ID);
    CpswApp_addAleVlanEntry(taskObj);

    /* Transmit untagged packets
     * VLAN tag inserted from port 0 default settings to be dropped by ALE's
     *force untagged settings */
    CpswAppUtils_print("Transmit untagged packets\n");
    CpswAppUtils_print("  VLAN will be inserted from port 0 settings, but shall be forced untagged by ALE\n");
    CpswApp_transmitTest(taskObj, 0, 0);

    /* Remove ALE entry for ETH_TEST_VLAN_ID */
    CpswAppUtils_print("Remove ALE entry for VLAN ID %d\n", ETH_TEST_VLAN_ID);
    CpswApp_removeAleVlanEntry(taskObj);

    /* Transmit untagged packets
     * VLAN tag will be inserted from port 0 default settings */
    CpswAppUtils_print("Transmit untagged packets\n");
    CpswAppUtils_print("  VLAN will be inserted from port 0 settings:\n");
    CpswAppUtils_print("  - VLAN ID  = %d\n", ETH_TEST_VLAN_ID);
    CpswAppUtils_print("  - VLAN Pri = %d\n", ETH_TEST_VLAN_PRI);
    CpswApp_transmitTest(taskObj, 0, 0);

    return status;
}

void CpswTestVlan_updatePortLinkCfg(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                    Cpsw_MacPort portNum)
{
    pLinkArgs->macConfig.enableLoopback = FALSE;
}

void CpswTestVlan_setOpenPrms(CpswTestTaskObj *taskObj,
                              Cpsw_Config *pCpswCfg,
                              CpswOsal_Prms *pOsalPrms,
                              CpswUtils_Prms *pUtilsPrms)
{
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    pCpswCfg->aleConfig.vlanConfig.aleVlanAwareMode  = TRUE;
    pCpswCfg->aleConfig.vlanConfig.cpswVlanAwareMode = TRUE;

    pCpswCfg->hostPortConfig.vlanCfg.portPri = ETH_TEST_VLAN_PRI;
    pCpswCfg->hostPortConfig.vlanCfg.portCfi = 0U;
    pCpswCfg->hostPortConfig.vlanCfg.portVID = ETH_TEST_VLAN_ID;

    pCpswCfg->vlanConfig.vlanAware = TRUE;

    if ((stateObj->curIteration % 2) != 0U)
    {
        pCpswCfg->vlanConfig.vlanSwitch = CPSW_VLAN_SWITCH_SERVICE;
    }
    else
    {
        pCpswCfg->vlanConfig.vlanSwitch = CPSW_VLAN_SWITCH_CUSTOMER;
    }
}

static void CpswApp_addAleVlanEntry(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj   = &taskObj->stateObj;
    CpswAle_VlanEntryInfo inArgs =
    {
        .vlanIdInfo.vlanId       = ETH_TEST_VLAN_ID,
        .vlanMemberList          = CPSW_ALE_ALL_PORTS_MASK,
        .unregMcastFloodMask     = 0U,
        .regMcastFloodMask       = 0U,
        .forceUntaggedEgressMask = CPSW_ALE_ALL_PORTS_MASK,
        .noLearnMask             = 0U,
        .vidIngressCheck         = false,
        .limitIPNxtHdr           = false,
        .disallowIPFragmentation = false,
    };
    CpswAle_AddEntryOutArgs outArgs;
    int32_t status;

    if ((stateObj->curIteration % 2) != 0U)
    {
        inArgs.vlanIdInfo.outerVlanFlag = true;
    }
    else
    {
        inArgs.vlanIdInfo.outerVlanFlag = false;
    }

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &outArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_ADD_VLAN, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_addAleVlanEntry() ADD_VLAN ioctl failed: %d\n",
                           status);
    }
}

static void CpswApp_removeAleVlanEntry(CpswTestTaskObj *taskObj)
{
    Cpsw_IoctlPrms prms;
    CpswTestStateObj *stateObj = &taskObj->stateObj;
    CpswAle_VlanIdInfo inArgs  =
    {
        .vlanId = ETH_TEST_VLAN_ID,
    };
    int32_t status;

    if ((stateObj->curIteration % 2) != 0U)
    {
        inArgs.outerVlanFlag = true;
    }
    else
    {
        inArgs.outerVlanFlag = false;
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs);

    status = Cpsw_ioctl(stateObj->hCpsw, stateObj->coreId, CPSW_ALE_IOCTL_REMOVE_VLAN, &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_removeAleVlanEntry() REMOVE_VLAN ioctl failed: %d\n",
                           status);
    }
}

static int32_t CpswUtApp_transmitPkts(CpswTestTaskObj *taskObj,
                                      CpswDma_PktInfoQ *pTxSubmitQ,
                                      uint32_t txCfgIndex)
{
    uint32_t txRetrievePktCnt, txCnt;
    int32_t status             = CPSW_SOK;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    while (0U != CpswUtils_getQCount(pTxSubmitQ))
    {
        txCnt  = CpswUtils_getQCount(pTxSubmitQ);
        status = CpswAppUtils_submitTxPackets(stateObj->txChObj[txCfgIndex].hTxCh,
                                              pTxSubmitQ);
        /* Retrieve TX free packets */
        if (status == CPSW_SOK)
        {
            txCnt            = txCnt - CpswUtils_getQCount(pTxSubmitQ);
            txRetrievePktCnt = 0U;
            while (txRetrievePktCnt != txCnt)
            {
                /* TODO this is not failure as HW is busy sending packets, we need
                 * to wait and again call retrieve packets */
                CpswUtApp_wait(1);
                txRetrievePktCnt += CpswTestCommon_retrieveFreeTxPkts(taskObj, txCfgIndex);
#if DEBUG
                CpswAppUtils_print("Failed to retrieve consumed transmit packets: %d\n",
                                   status);
#endif
            }
        }
        else
        {
            break;
        }
    }

    return status;
}

static int32_t CpswUtApp_xmitTest(CpswTestTaskObj *taskObj,
                                  uint32_t txCfgIndex,
                                  uint32_t num,
                                  const uint8_t *dstAddr,
                                  const uint8_t *srcAddr,
                                  uint16_t etherType,
                                  uint16_t length)
{
    int32_t retVal = CPSW_SOK;
    CpswDma_PktInfo *pktInfo;
    uint16_t len    = length - ETH_TEST_DATA_HDR_LEN;
    uint32_t pktCnt = 0U;
    DataFramePayload *payload;
    CpswDma_PktInfoQ txSubmitQ;
    EthFrame *frame;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    while (pktCnt < num)
    {
        CpswUtils_initQ(&txSubmitQ);

        /* Dequeue one free TX Eth packet */
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);

        while (NULL != pktInfo)
        {
            pktCnt++;
            /* Fill the TX Eth frame with test content */
            frame   = (EthFrame *)pktInfo->bufPtr;
            payload = (DataFramePayload *)frame->payload;
            memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
            memcpy(frame->hdr.srcMac, &stateObj->hostMacAddr[0U],
                   CPSW_MAC_ADDR_LEN);
            frame->hdr.etherType = htons(etherType);

            payload->type = 0U; /* DataPattern1 */
            payload->len  = htons(len);
            memcpy(payload->data, Cpsw_AppDataPattern1, len);

            pktInfo->userBufLen = length + sizeof(EthFrameHeader);
            pktInfo->appPriv    = &stateObj;

            /* Enqueue the packet for later transmission */
            CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

            if (pktCnt >= num)
            {
                break;
            }

            /* Dequeue one free TX Eth packet */
            pktInfo =
                (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->txChObj[txCfgIndex].txFreePktInfoQ);
        }

        retVal = CpswUtApp_transmitPkts(taskObj, &txSubmitQ, txCfgIndex);
    }

    return retVal;
}

static void CpswUtApp_discardRxPkts(CpswTestTaskObj *taskObj,
                                    uint32_t rxCfgIndex)
{
    CpswDma_PktInfo *pkt;
    uint32_t count;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    /* Discard any packets sitting in rxReadyQ before the test starts */
    count = CpswUtils_getQCount(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
    if (count != 0U)
    {
        CpswAppUtils_print("CpswApp_discardRxPkts() rxReadyQ has %d packets, discarding all\n",
                           count);
        pkt = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
        while (pkt != NULL)
        {
            /* Release the received packet */
            CpswUtils_enQ(&stateObj->rxFlowObj[rxCfgIndex].rxFreeQ,
                          &pkt->node);
            pkt = (CpswDma_PktInfo *)CpswUtils_deQ(&stateObj->rxFlowObj[rxCfgIndex].rxReadyQ);
        }
    }
}

static void CpswApp_transmitTest(CpswTestTaskObj *taskObj,
                                 uint32_t txCfgIndex,
                                 uint32_t rxCfgIndex)
{
    uint32_t iterations = ETH_TEST_ITER_S_COUNT;
    uint16_t len        = 500U;
    uint32_t retVal;
    CpswTestStateObj *stateObj = &taskObj->stateObj;

    CpswUtApp_discardRxPkts(taskObj, rxCfgIndex);

    CpswUtApp_wait(10);

    /* Transmit frames */
    CpswAppUtils_print("Test_transmit: transmitting %d packets\n", iterations);
    retVal = CpswUtApp_xmitTest(taskObj,
                                txCfgIndex,
                                iterations,
                                bcastAddr,
                                &stateObj->hostMacAddr[0U],
                                ETHERTYPE_EXPERIMENTAL1,
                                len);
    if (retVal != CPSW_SOK)
    {
        CpswAppUtils_print("Test_transmit: failed to transmit %d packets\n",
                           iterations);
    }
}
