/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file cpsw_test.h
 *
 *  \brief This file contains all the structures, macros, enums
 *  used by the CPSW test application.
 *
 */

#ifndef CPSW_TEST_H_
#define CPSW_TEST_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include <xdc/std.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Mailbox.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Event.h>

#include <ti/csl/tistdtypes.h>
#include <ti/csl/csl_types.h>
#include <ti/csl/soc.h>
#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/uart/UART.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>

#include <ti/osal/osal.h>
#include <ti/osal/SemaphoreP.h>
#include <ti/osal/TaskP.h>

#include "cpsw_test_utils_mem.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <ti/board/board.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */
/* UART read timeout in msec                        */
#define CPSW_TEST_UART_TIMEOUT_MSEC                (10000U)

#define CPSW_TEST_PRINT_BUFSIZE                    (0x4000U)

/* Application name string used in print statements */
#define APP_NAME                        "CPSW_TEST"
#define CPSW_TEST_MAX_TASKS             (1U)
#define CPSW_INSTANCE_ID_MAX            (1U)
#define CPSW_TEST_MAX_INST              (CPSW_INSTANCE_ID_MAX)
#define CPSW_TEST_CFG_MAX_NUM_RX_FLOWS  (4U)
#define CPSW_TEST_CFG_MAX_NUM_TX_CH     (4U)
/* Default values */
#define CPSW_TEST_USE_DEF               (0xFAFAU)
#define CPSW_TEST_DEF_ITERATION_CNT          (1U)

/* Defined the following so that it is easy to understand a particular config */
#define DEF_HEAP_ID                     (UTILS_MEM_HEAP_ID_DDR)
#define TEST_ENABLE                     (TRUE)
#define TEST_DISABLE                    (FALSE)
#define PRF_ENABLE                      (TRUE)
#define PRF_DISABLE                     (FALSE)
#define PRINT_ENABLE                    (TRUE)
#define PRINT_DISABLE                   (FALSE)
#define TIMEST_ENABLE                   (TRUE)
#define TIMEST_DISABLE                  (FALSE)
#define DATA_CHECK_ENABLE               (TRUE)
#define DATA_CHECK_DISABLE              (FALSE)
#define PACING_NONE                     (0U)
#define DEF_PACING                      (20U)

#define CPSW_TEST_RF_SOC_AM65XX         ((uint64_t)0x00000001U)
#define CPSW_TEST_RF_SOC_J721E          ((uint64_t)0x00000002U)
#define CPSW_TEST_RF_SOC_ALL            ((uint64_t)0xFFFFFFFFU)

#define CPSW_TEST_RF_CORE_MPU1_0  ((uint64_t)(((uint64_t)0x0001U) << 32U))
#define CPSW_TEST_RF_CORE_MPU2_0  ((uint64_t)(((uint64_t)0x0002U) << 32U))
#define CPSW_TEST_RF_CORE_MCU1_0  ((uint64_t)(((uint64_t)0x0004U) << 32U))
#define CPSW_TEST_RF_CORE_MCU1_1  ((uint64_t)(((uint64_t)0x0008U) << 32U))
#define CPSW_TEST_RF_CORE_MCU2_0  ((uint64_t)(((uint64_t)0x0010U) << 32U))
#define CPSW_TEST_RF_CORE_MCU2_1  ((uint64_t)(((uint64_t)0x0020U) << 32U))
#define CPSW_TEST_RF_CORE_MCU3_0  ((uint64_t)(((uint64_t)0x0040U) << 32U))
#define CPSW_TEST_RF_CORE_MCU3_1  ((uint64_t)(((uint64_t)0x0080U) << 32U))
#define CPSW_TEST_RF_CORE_C7X_1   ((uint64_t)(((uint64_t)0x0100U) << 32U))
#define CPSW_TEST_RF_CORE_C66X_1  ((uint64_t)(((uint64_t)0x0200U) << 32U))
#define CPSW_TEST_RF_CORE_C66X_2  ((uint64_t)(((uint64_t)0x0400U) << 32U))
#define CPSW_TEST_RF_CORE_ALL     ((uint64_t)(((uint64_t)0xFFFFU) << 32U))
#define CPSW_TEST_RF_CORE_MCU_ALL       (CPSW_TEST_RF_CORE_MCU1_0 | \
                                         CPSW_TEST_RF_CORE_MCU1_1 | \
                                         CPSW_TEST_RF_CORE_MCU2_0 | \
                                         CPSW_TEST_RF_CORE_MCU2_1 | \
                                         CPSW_TEST_RF_CORE_MCU3_0 | \
                                         CPSW_TEST_RF_CORE_MCU3_1)

/* For future when we have dynamic coverage testcases */
#define CPSW_TEST_RF_CFG_ALL      ((uint64_t)(((uint64_t)0xFFFFU) << 48U))

#define CPSW_TEST_VINTR_MAX             (16U)

/**
 *  \brief Testcase types.
 */
typedef enum
{
    /* Category */
    CPSW_TCT_SANITY      = 0x01U,
    CPSW_TCT_REGRESSION  = 0x02U,
    CPSW_TCT_FULL        = 0x04U,
    /* Adequacy */
    CPSW_TCT_FUNCTIONAL  = 0x08U,
    CPSW_TCT_STRESS      = 0x10U,
    CPSW_TCT_NEGATIVE    = 0x20U,
    CPSW_TCT_PERFORMANCE = 0x40U,
    CPSW_TCT_MISC        = 0x80U,
    CPSW_TCT_API         = 0x100U,
    /* Used for Test parser dont use in test case */
    CPSW_TCT_ALL         = 0x1FFU
} CpswTestCaseType;

/**
 *  \brief Test types.
 */
typedef enum
{
    CPSW_TEST_TYPE_J721E    = 0U,

    /**< Test type: Test only J721E EVM test cases
     *   from which it can be viewed off line */
    CPSW_TEST_TYPE_MAXWELL = 1U,

    /**< Test type: Test only MAXWELL EVM test cases
     *   DSS */
} CpswTestTypes;

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* Forward declaration. */
typedef struct CpswTestTxChCfgInfo_t CpswTestTxChCfgInfo;
typedef struct CpswTestRxFlowCfgInfo_t CpswTestRxFlowCfgInfo;
typedef struct CpswTestTxChObj_t CpswTestTxChObj;
typedef struct CpswTestRxFlowObj_t CpswTestRxFlowObj;
typedef struct CpswTestStateObj_t CpswTestStateObj;
typedef struct CpswTestTaskObj_t CpswTestTaskObj;
typedef struct CpswTestTaskCfg_t CpswTestTaskCfg;
typedef struct CpswTestParams_t CpswTestParams;

/** \brief Typedef for test case type function pointer */
typedef void (*CpswTest_OpenPrmsUpdateFxn)(CpswTestTaskObj *taskObj,
                                           Cpsw_Config *pCpswCfg,
                                           CpswOsal_Prms *pOsalPrms,
                                           CpswUtils_Prms *pUtilsPrms);
typedef int32_t (*CpswTest_SetTxChPrmsFxn)(CpswDma_OpenTxChPrms *pTxChPrms,
                                           CpswTestTaskObj *taskObj,
                                           uint32_t txPSILThreadId,
                                           uint32_t txChCfgIndex);
typedef int32_t (*CpswTest_SetRxFlowPrmsFxn)(CpswDma_OpenRxFlowPrms *pRxFlowPrms,
                                             CpswTestTaskObj *taskObj,
                                             uint32_t rxFlowCfgIdx);
typedef void (*CpswTest_SetPortLinkPrmsFxn)(Cpsw_OpenPortLinkInArgs *pLinkArgs,
                                            Cpsw_MacPort portNum);
typedef int32_t (*CpswTest_RunFxn)(CpswTestTaskObj *taskObj);
typedef int32_t (*CpswTest_DeInitFxn)(CpswTestTaskObj *taskObj);
typedef int32_t (*CpswTest_SetTxPktInfoFxn)(CpswTestTaskObj *taskObj,
                                            uint32_t txChIndex,
                                            uint32_t pktNum, CpswDma_PktInfo *pktInfo,
                                            bool *testComplete);
typedef int32_t (*CpswTest_PreTxSendFxn)(CpswTestTaskObj *taskObj, uint32_t txChIndex);
typedef int32_t (*CpswTest_PostTxSendFxn)(CpswTestTaskObj *taskObj,
                                          uint32_t txChIndex);

typedef int32_t (*CpswTest_ProcessRxPktFxn)(CpswTestTaskObj *taskObj, uint32_t rxFlowId,
                                            uint32_t pktNum, CpswDma_PktInfo *pktInfo,
                                            bool *testComplete);
typedef int32_t (*CpswTest_PreProcessRxPktFxn)(CpswTestTaskObj *taskObj,
                                               uint32_t rxCfgIndex);
typedef int32_t (*CpswTest_PostProcessRxPktFxn)(CpswTestTaskObj *taskObj,
                                                uint32_t rxCfgIndex);

struct CpswTestRxFlowObj_t
{
    CpswDma_RxFlowHandle hRxFlow;
    CpswDma_PktInfoQ rxFreeQ;
    CpswDma_PktInfoQ rxReadyQ;
    uint32_t rxFlowStartIdx;
    uint32_t rxFlowIdx;
    Task_Handle hRxTask;
    Semaphore_Handle hRxCQSem;
    Mailbox_Handle hRxTestStatusMbx;
    int32_t rxTaskStatus;
    uint32_t receivedBitRate;
    uint32_t processCount;
    uint32_t stackUsed;
};

struct CpswTestTxChObj_t
{
    CpswDma_TxChHandle hTxCh;
    CpswDma_PktInfoQ txFreePktInfoQ;
    Task_Handle hTxTask;
    Semaphore_Handle hTxCQSem;
    Mailbox_Handle hTxTestStatusMbx;
    int32_t txTaskStatus;
    uint32_t transmitBitRate;
    uint32_t processCount;
    uint32_t retrieveCnt;
    uint32_t stackUsed;
    uint32_t txPSILThreadId;
};

struct CpswTestStateObj_t
{
    /* CPSW driver handle */
    Cpsw_Handle hCpsw;
    Udma_DrvHandle hUdmaDrv;
    CpswTestTxChObj txChObj[CPSW_TEST_CFG_MAX_NUM_TX_CH];
    CpswTestRxFlowObj rxFlowObj[CPSW_TEST_CFG_MAX_NUM_RX_FLOWS];
    uint8_t hostMacAddr[CPSW_MAC_ADDR_LEN];
    void *testCtx;
    CpswTestTxChCfgInfo *txChCfgInfo[CPSW_TEST_CFG_MAX_NUM_TX_CH];
    CpswTestRxFlowCfgInfo *rxFlowCfgInfo[CPSW_TEST_CFG_MAX_NUM_RX_FLOWS];
    uint32_t curIteration;
    Clock_Handle hTimer;
    Task_Handle task_periodicTick;
    Semaphore_Handle timerSem;
    Event_Struct phyEventObj;
    Event_Handle phyEvent;
    Event_Struct portEventObj;
    Event_Handle portEvent;
    uint32_t coreId;
    uint32_t coreKey;
    uint32_t noPhyPortMask;
};

struct CpswTestTxChCfgInfo_t
{
    const uint32_t txChCfgId;

    const CpswTest_SetTxChPrmsFxn updateTxChPrms;
    /**< Tx DMA channel to use */
    const uint32_t maxTxPktLen;
    /**< Max Tx Packet Length */
    const CpswTest_SetTxPktInfoFxn setTxPktInfo;
    const CpswTest_PreTxSendFxn preTxSend;
    const CpswTest_PostTxSendFxn postTxSend;
    const uint32_t pktSendCount;
    const uint32_t pktSendLoopCount;
};

struct CpswTestRxFlowCfgInfo_t
{
    const uint32_t rxFlowCfgId;
    const CpswTest_SetRxFlowPrmsFxn updateRxFlowPrms;
    const uint32_t rxFlowMtu;
    const uint32_t pktRecvCount;
    const CpswTest_ProcessRxPktFxn processRxPkt;
    const CpswTest_PreProcessRxPktFxn preRxProcess;
    const CpswTest_PostProcessRxPktFxn postRxProcess;
    const bool useDefaultFlow;
};

/**
 *  \brief Structure used for CPSW UT task configuration parameters.
 */
struct CpswTestTaskCfg_t
{
    const uint32_t taskCfgId;
    /**< Id of the test task obj */
    const uint32_t txChCfgId[CPSW_TEST_CFG_MAX_NUM_TX_CH];
    /**< Type of test case to run. */
    const uint32_t rxFlowCfgId[CPSW_TEST_CFG_MAX_NUM_RX_FLOWS];
    /**< Type of test case to run. */

    const CpswTest_OpenPrmsUpdateFxn openPrmsUpdate;
    const CpswTest_SetPortLinkPrmsFxn portLinkPrmsUpdate;
    /**< Type of test case to run. */
    const CpswTest_RunFxn runTest;
    /**< Test case deinit hook function - can be used to do additional test specific cleanup . */
    const CpswTest_DeInitFxn deInitTest;
    /**< Type of test case to run. */
    const uint32_t taskId;
    /**< task Id */
    const Cpsw_Type cpswType;
    /**< CPSW Type for the test. */
    const uint32_t macPortMask;
    /**< Bitmask of Mac port number enabled for this test */
    uint32_t numTxCh;
    uint32_t numRxFlow;
};

/**
 *  \brief Structure used for CPSW UT task object parameters.
 */
struct CpswTestTaskObj_t
{
    CpswTestTaskCfg *taskCfg;
    /**< Pointer to CPSW DRV instance configuration. */
    CpswTestStateObj stateObj;
    /**< CPSW Test instance object. Holds all state info for the test */
    int32_t testResult;
    /**< test result. */
    uint32_t durationMs;
    /**< Time taken in ms */
    TaskP_Handle taskHandle;
    /**< task handle */
    SemaphoreP_Handle syncSem;
    /**< Semaphore to synch task start */
    CpswTestParams *testParams;
    /**< Reference to corresponding test parameters */
};

/**
 *  \brief Test case parameter structure.
 */
struct CpswTestParams_t
{
    const bool enableTest;
    /**< Whether test case should be executed or not. */
    const uint32_t tcId;
    /**< Test case ID. */
    const char *tcName;
    /**< Test case name. */
    const char *tcRunInfo;
    /**< Test case running instructions. */
    const char *reqMapping;
    /**< Requirement mapping. */
    const char *disableInfo;
    /**< Reason string for disabling a test case. */
    const bool printEnable;
    /**< Enable/disable print statements, used for stress testing. */
    const bool profilingEnable;
    /**< Enable Profiling prints. */
    const uint32_t tcType;
    /**< Type of testcase  - like BFT, stress etc... */
    const uint32_t dcEnable;
    /**< Enable/disable data check - used for performance. */
    const uint32_t iterationCnt;
    /**< Count for iteration, how many times this test should be run. */
    const uint32_t numTasks;
    /**< Number of tasks to test */
    const uint32_t testType[CPSW_TEST_MAX_TASKS];
    /**< Type of test, Refer enum #CpswTestTypes */
    const uint32_t taskCfgId[CPSW_TEST_MAX_TASKS];
    /**< Id of the test task cfg */
    SemaphoreP_Handle taskCompleteSem;
    /**< Semaphore to indicate task completion */
    CpswTestTaskObj taskObj[CPSW_TEST_MAX_TASKS];

    /**< CPSW Test Task Object. This object is filled runtime.
     *   This is not a input parameter */

    /*
     * Below variables are initialized in code and not in table!!
     */
    bool isRun;
    /**< Flag to indicate whether the test case is run or not. */
    int32_t testResult;
    /**< Test result. */
};
/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*
 * Parser functions
 */
int32_t cpswTestParser(void);

void CpswUT_testPrint(const char *str,
                      ...);

void udmaDrvPrint(const char *str);

/*
 * Parser functions
 */
void cpswTestLogTestResult(const CpswTestParams *testParams,
                           int32_t testResult,
                           uint32_t testCaseId,
                           char *testcaseInfo);

void cpswTestPrintTestResult(const CpswTestParams *testParams,
                             uint32_t skipCount,
                             uint32_t disableCount);

void cpswTestResetTestResult(void);

/*
 * CPSW common functions
 */

void CpswUT_testResetTestResult(void);

void CpswUT_testCalcPerformance(CpswTestTaskObj *taskObj,
                                uint32_t durationMs);

int32_t AppUtils_getCharTimeout(char *ch,
                                uint32_t msec);

uint32_t AppUtils_getCurTimeInMsec(void);

uint32_t AppUtils_getElapsedTimeInMsec(uint32_t startTime);

/* ========================================================================== */
/*      Internal Function Declarations (Needed for other static inlines)      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_TEST_H_ */
