/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *   \file  cpsw_ale.c
 *
 *   \brief
 *          ALE module implementation
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* Project dependency headers */
#include <ti/csl/cslr.h>
#include <ti/csl/csl_cpswitch.h>
#include <ti/osal/osal.h>

/* This module's private header */
#include "cpsw_priv.h"
#include "cpsw_ale_priv.h"
#include "cpsw_trace_priv.h"

/* This module's header */
#include "cpsw_ale.h"
#include "cpsw_ale_config.h"
#include "cpsw_priv.h"
#include "cpsw_types.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define NOT_ZERO(x)                            ((uint32_t)0U != (uint32_t)(x))
#define IS_NULL(x)                             ((typeof(x))NULL == (x))

/*! ALE magic number (assigned randomly) */
#define CPSW_ALE_MAGIC                        (0x73CA4108U)

/*! ALE IOCTL Cmd Minor number */
#define CPSW_ALE_MINOR(name)          (CPSW_IOCTL_GET_MIN(CPSW_ALE_IOCTL_ ## name))

#define CSL_ALE_THREADMAPVAL_RESETVAL                            (0x00000000U)

#define CPSW_ALE_VLAN_ID_MASK                                    (CPSW_VLAN_ID_MAX)
#define CPSW_ALE_AGING_TIMER_DIV_FACTOR                          (1000000)
#define CPSW_ALE_IDL_INC_VAL_DIV_FACTOR                          (32768)
#define CPSW_ALE_NOMATCH_ENTRY0_POLICER_INDEX                    (0)
#define CPSW_ALE_BCAST_ADDR_INIT_LIST                            {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}
#define CPSW_ALE_BLOCKCLASSIFIER_PORTMASK                        (0)
/* Route index maps to egressOp and should be non-zero */
#define CPSW_ALE_BLOCKCLASSIFIER_ROUTEINDEX                      (1)

#define CPSW_ALE_MAC_ADDR_INIT_LIST(macAddr)    {macAddr[0], macAddr[1], \
                                                 macAddr[2], macAddr[3], \
                                                 macAddr[4], macAddr[5]}
#define CPSW_ALE_OUI_ADDR_INIT_LIST(ouiAddr)    {ouiAddr[0], ouiAddr[1], \
                                                 ouiAddr[2]}
#define CPSW_ALE_IPV4_ADDR_INIT_LIST(ipv4Addr)  {ipv4Addr[0], ipv4Addr[1], \
                                                 ipv4Addr[2], ipv4Addr[3]}
#define CPSW_ALE_IPV6_ADDR_INIT_LIST(ipv6Addr)  {ipv6Addr[0], ipv6Addr[1],   \
                                                 ipv6Addr[2], ipv6Addr[3],   \
                                                 ipv6Addr[4], ipv6Addr[5],   \
                                                 ipv6Addr[6], ipv6Addr[7],   \
                                                 ipv6Addr[8], ipv6Addr[9],   \
                                                 ipv6Addr[10], ipv6Addr[11], \
                                                 ipv6Addr[12], ipv6Addr[13], \
                                                 ipv6Addr[14], ipv6Addr[15]}

/* The first ALE Mask Mux is a readonly register set to ALL_PORTS mask */
#define CPSW_ALE_VLAN_MASK_MUX_COUNT(cpswAleRegBase) \
    (CPSW_UTILS_ARRAYSIZE(cpswAleRegBase->I1_ALE_MSK_MUX1) + 1)
#define CPSW_ALE_VLANMASKMUX2REG(index) \
    ((index) - 1U)

#define CPSW_ALE_2G_MAX_ALE_ENTRIES     (64)
#define CPSW_ALE_2G_MAX_POLICERS        (8)

#define CPSW_ALE_5G_MAX_ALE_ENTRIES     (512)
#define CPSW_ALE_5G_MAX_POLICERS        (64)

#define CPSW_ALE_9G_MAX_ALE_ENTRIES     (1024)
#define CPSW_ALE_9G_MAX_POLICERS        (96)

#define CPSW_ALE_SAFE_ASSIGN(ptr, val)   do { \
        if (!IS_NULL(ptr))                    \
        {                                     \
            *ptr = val;                       \
        }                                     \
} while (0)

typedef const uint8_t (*const CpswAle_MacAddrType_t)[CPSW_ALE_MACADDR_NUM_OCTETS];
typedef const uint8_t (*const CpswAle_OuiAddrType_t)[CPSW_ALE_OUIADDR_NUM_OCTETS];
typedef const uint8_t (*const CpswAle_Ipv4AddrType_t)[CPSW_ALE_IPV4ADDR_NUM_OCTETS];
typedef const uint8_t (*const CpswAle_Ipv6AddrType_t)[CPSW_ALE_IPV6ADDR_NUM_OCTETS];

typedef uint8_t (*const CpswAle_MacAddrWriteType_t)[CPSW_ALE_MACADDR_NUM_OCTETS];

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct CpswAle_IoctlValidateEntry_s
{
    CPSW_ALE_IOCTL_E cmd;
    uint32_t inArgsSize;
    uint32_t outArgsSize;
} CpswAle_IoctlValidateEntry;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static void CpswAle_initcpswAleRegBase(CSL_AleRegs **cpswAleRegBasePtr,
                                       Cpsw_Type cpswType);

static uint32_t CpswAle_getNumPorts(Cpsw_Type cpswType);

static int32_t CpswAle_validateIoctl(const CpswAle_IoctlValidateEntry *entry,
                                     CPSW_ALE_IOCTL_E cmd,
                                     const Cpsw_IoctlPrms *prms);

static int32_t CpswAle_lookupUcastAddr(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t *portNum,
                                       bool *blocked,
                                       bool *secure,
                                       bool *super,
                                       bool *ageable,
                                       bool *trunk,
                                       bool *touched,
                                       uint32_t *aleEntryIndex);

static int32_t CpswAle_lookupMcastAddr(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t numIgnBits,
                                       uint32_t *portMask,
                                       CPSW_ALE_FWDSTLVL_E *fwdState,
                                       bool *super,
                                       uint32_t *pNumIgnBits,
                                       uint32_t *aleEntryIndex);

static void CpswAle_initAleRegs(CSL_AleRegs *cpswAleRegBase);

static void CpswAle_setAleConfig(CpswAle_Obj *pAle,
                                 const CpswAle_Config *pALEConfig,
                                 uint32_t aleFreqHz,
                                 Cpsw_Type cpswType);

int32_t CpswAle_disablePortMirror(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setReceiveFilter(CpswAle_Obj *pALE,
                                        CPSW_ALE_RXFILTER_E rxFilter);

static int32_t CpswAle_getReceiveFilter(const CpswAle_Obj *pAle,
                                        CPSW_ALE_RXFILTER_E *pRxFilter);

static int32_t CpswAle_getFreeEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t *entry);

static void CpswAle_clearTableEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t entry);

static int32_t CpswAle_disableBcastMcastLimit(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_findAddr(CSL_AleRegs *cpswAleRegBase,
                                CpswAle_MacAddrType_t addr,
                                uint32_t vlanId,
                                uint32_t *entry);

static int32_t CpswAle_findVlan(CSL_AleRegs *cpswAleRegBase,
                                uint32_t vlanId,
                                bool outerVlanFlag,
                                uint32_t *vlanMemberList,
                                uint32_t *unregMcastFloodMask,
                                uint32_t *regMcastFloodMask,
                                uint32_t *forceUntaggedEgress,
                                uint32_t *noLearnMask,
                                bool *vidIngressCheck,
                                bool *limitIpNxtHdr,
                                bool *disallowFragmentation,
                                uint32_t *entry);

static int32_t CpswAle_setInterVlanConfig(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          const CpswAle_SetInterVLANConfigInArgs *inArgs,
                                          CpswAle_SetInterVLANConfigOutArgs *outArgs);

static void CpswAle_ageOut(CSL_AleRegs *cpswAleRegBase,
                           bool aleTimerActive,
                           uint32_t aleTickTimeOutCount,
                           uint32_t *aleTickCount);

static int32_t CpswAle_delPortUcastEntries(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t portNum);

static int32_t CpswAle_addUcastAddr(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t portNum,
                                    CpswAle_MacAddrType_t addr,
                                    uint32_t vlanId,
                                    uint32_t blocked,
                                    uint32_t secure,
                                    uint32_t ageable,
                                    uint32_t trunk,
                                    uint32_t *entryIndex);

static int32_t CpswAle_addOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                  CpswAle_OuiAddrType_t ouiAddr,
                                  uint32_t *pAleIndex,
                                  bool *pAleEntryAlloced);

static int32_t CpswAle_addMcastAddr(CSL_AleRegs *cpswAleRegBase,
                                    const CpswAle_MacAddrType_t addr,
                                    uint32_t vlanId,
                                    uint32_t portMask,
                                    uint32_t super,
                                    uint32_t fwdState,
                                    uint32_t numIgnBits,
                                    uint32_t *entryIndex);

static int32_t CpswAle_addVlan(CSL_AleRegs *cpswAleRegBase,
                               uint32_t vlanId,
                               bool outerVlanFlag,
                               uint32_t vlanMemberList,
                               uint32_t unregMcastFloodMask,
                               uint32_t regMcastFloodMask,
                               uint32_t forceUntaggedEgress,
                               uint32_t noLearnMask,
                               bool vidIngressCheck,
                               bool limitIPNxtHdr,
                               bool disallowIPFragmentation,
                               uint32_t *aleEntryIndex);

static int32_t CpswAle_delVlan(CSL_AleRegs *cpswAleRegBase,
                               uint32_t vlanId,
                               bool outerVlanFlag);

static int32_t CpswAle_delAddr(CSL_AleRegs *cpswAleRegBase,
                               CpswAle_MacAddrType_t addr,
                               uint32_t vlanId);

static int32_t CpswAle_delOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_OuiAddrType_t ouiAddr);

static void CpswAle_dumpTable(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setAlePortState(CSL_AleRegs *cpswAleRegBase,
                                       uint32_t port_num,
                                       CSL_CPSW_ALE_PORTSTATE portState);

static bool CpswAle_isMcastAddr(CpswAle_MacAddrType_t macAddr);

static void CpswAle_setAleAging(CSL_AleRegs *cpswAleRegBase,
                                const CpswAle_AgingConfig *agingConfig,
                                uint32_t aleFreq,
                                bool *aleSoftTimerActive,
                                uint32_t *aleTickTimeOutCount,
                                uint32_t *aleSoftTickCount);

static void CpswAle_validateIPVersion(CSL_AleRegs *cpswAleRegBase);

void CpswAle_clearGlobalPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                     bool clearAllHit,
                                     bool clearRedHit,
                                     bool clearYellowHit);

static int32_t CpswAle_findOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                    CpswAle_OuiAddrType_t addr,
                                    uint32_t *entry);

static uint32_t CpswAle_getMaxAleEntries(CSL_AleRegs *cpswAleRegBase);

static uint32_t CpswAle_getMaxPolicers(CSL_AleRegs *cpswAleRegBase);

static uint32_t CpswAle_mapNumPktsPerSec2RateLimit(CSL_AleRegs *cpswAleRegBase,
                                                   uint32_t aleFreq,
                                                   uint32_t numPktsPerSec);

static uint32_t CpswAle_mapRateLimit2NumPktsPerSec(CSL_AleRegs *cpswAleRegBase,
                                                   uint32_t aleFreq,
                                                   uint32_t rateLimit);

static int32_t CpswAle_setPolicerControl(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t aleFreqHz,
                                         bool policingEnable,
                                         bool yellowDropEnable,
                                         bool redDropEnable,
                                         CPSW_ALE_POLICER_YELLOWTHRESH_E yellowThresh,
                                         CPSW_ALE_POLICER_NOMATCH_MODE_E policerMatchMode,
                                         const CpswAle_UnregulatedTrafficPolicer *noMatchPolicer);

static int32_t CpswAle_setPolicerDefaultThreadConfig(CSL_AleRegs *cpswAleRegBase,
                                                     bool threadEnable,
                                                     uint32_t threadVal,
                                                     bool enablePriorityOR,
                                                     bool disableMacPortDefaultThread);

static int32_t CpswAle_setPolicerRateLimit(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t peakRateInBitsPs,
                                           uint32_t commitRateInBitsPs,
                                           uint32_t aleFreq);

static uint32_t CpswAle_mapAleAgingIntervalMsToAgingTimer(uint32_t aleFreq,
                                                          uint32_t aleAgingIntervalInMs);

static int32_t CpswAle_findIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                    const CpswAle_Ipv6AddrType_t ipv6Addr,
                                    const uint32_t numLSBIgnore,
                                    uint32_t *entry);

static void CpswAle_clearSelectedPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerIndex);

static int32_t CpswAle_findIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                    CpswAle_Ipv4AddrType_t ipv4Addr,
                                    uint32_t numLSBIgnore,
                                    uint32_t *entry);

static int32_t CpswAle_setPortTrunkConfig(CSL_AleRegs *cpswAleRegBase,
                                          const CpswAle_SetTrunkConfigInArgs *trunkCfg);

static int32_t CpswAle_setPortMirrorConfig(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t maxNumPorts,
                                           const CpswAle_SetPortMirroringConfigInArgs *portMirrorCfg);

static void CpswAle_getPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t policerIndex,
                                    bool *pPolHit,
                                    bool *pPolRedHit,
                                    bool *pPolYellowHit,
                                    bool clearStats);

static int32_t CpswAle_setOuiPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t policerEntryIdx,
                                          CpswAle_OuiAddrType_t ouiAddr,
                                          uint32_t peakRateInBps,
                                          uint32_t commitRateInBps,
                                          uint32_t aleFreq,
                                          uint32_t *aleEntryIndex,
                                           bool *pAleEntryAlloced);

static int32_t CpswAle_setIPv4PolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           CpswAle_Ipv4AddrType_t ipv4Addr,
                                           CPSW_ALE_ADDR_TYPE addrType,
                                           uint32_t numLSBIgnore,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIndex,
                                           bool *pEntryAlloced);

static int32_t CpswAle_setIPv6PolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           CpswAle_Ipv6AddrType_t ipv6Addr,
                                           CPSW_ALE_ADDR_TYPE addrtype,
                                           uint32_t numLSBIgnore,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIndex,
                                           bool *pEntryAlloced);

static int32_t CpswAle_setEtherTypePolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                                uint32_t policerEntryIdx,
                                                uint32_t etherType,
                                                uint32_t peakRateInBps,
                                                uint32_t commitRateInBps,
                                                uint32_t aleFreq,
                                                uint32_t *aleEntryIndex,
                                                bool *pEntryAlloced);

static int32_t CpswAle_setMacAddrPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerEntryIdx,
                                              CpswAle_MacAddrType_t addr,
                                              uint32_t vlanId,
                                              uint32_t portNum,
                                              CPSW_ALE_ADDR_TYPE addrType,
                                              uint32_t peakRateInBps,
                                              uint32_t commitRateInBps,
                                              uint32_t aleFreq,
                                              uint32_t *aleEntryIdx,
                                              bool *pEntryAlloced);

static int32_t CpswAle_setThreadPolicerConfig(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerEntryIdx,
                                              uint32_t threadId);

static int32_t CpswAle_setVlanPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t vlanId,
                                           bool outVlanFlag,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIdx,
                                           bool *pEntryAlloced);

static int32_t CpswAle_setPortPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t port_num,
                                           bool isTrunkPort,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq);

static int32_t CpswAle_setPriorityPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t policerEntryIdx,
                                               uint32_t priority,
                                               uint32_t peakRateInBps,
                                               uint32_t commitRateInBps,
                                               uint32_t aleFreq);

static int32_t CpswAle_allocPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t *freePolicerEntry);

static int32_t CpswAle_delPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                       uint32_t policerIndex,
                                       uint32_t delAleEntryMask);

static int32_t CpswAle_delAllEntries(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_delAllPolicerEntries(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setPortBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t aleFreqHz,
                                              uint32_t portNum,
                                              bool enableBRateLimitForPort,
                                              uint32_t bcastLimitNumPktsPerSec,
                                              bool enableMRateLimitForPort,
                                              uint32_t mcastLimitNumPktsPerSec);

static int32_t CpswAle_getPortBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t aleFreqHz,
                                              uint32_t portNum,
                                              bool *pBRateLimitEnableForPort,
                                              uint32_t *pBcastLimitNumPktsPerSec,
                                              bool *pMRateLimitEnableForPort,
                                              uint32_t *pMcastLimitNumPktsPerSec);

static void CpswAle_dumpPolicerEntries(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_addPortDefaultVlanId(CpswAle_Obj *pAle);

static int32_t CpswAle_addDefaultAleEntries(CpswAle_Obj *pAle);

static int32_t CpswAle_delIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv4AddrType_t ipv4Addr,
                                   uint32_t numLSBIgnore);

static int32_t CpswAle_delEtherType(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t etherType);

static int32_t CpswAle_setIPSrcDstPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t policerEntryIdx,
                                               const CpswAle_IpAddrClassifierInfo *ipClassifier,
                                               CPSW_ALE_ADDR_TYPE addrType,
                                               uint32_t peakRateInBitsPerSec,
                                               uint32_t commitRateInBitsPerSec,
                                               uint32_t aleFreqHz,
                                               uint32_t *aleEntryIndex,
                                               bool *pEntryAlloced);

static int32_t CpswAle_setOAMLpbk(CSL_AleRegs *cpswAleRegBase,
                                  uint32_t enableLpbkPortMask);

static int32_t CpswAle_addEtherType(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t etherType,
                                    uint32_t *pAleIndex,
                                    bool *pEntryAlloced);

static int32_t CpswAle_addIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv4AddrType_t ipv4Addr,
                                   const uint32_t numLSBIgnore,
                                   uint32_t *pAleIndex,
                                   bool *pEntryAlloced);

static int32_t CpswAle_addIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv6AddrType_t ipv6Addr,
                                   const uint32_t numLSBIgnore,
                                   uint32_t *pAleIndex,
                                   bool *pEntryAlloced);

static void CpswAle_setAleModeFlags(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t aleModeFlags);

static int32_t CpswAle_getPortMacAddr(CSL_AleRegs *cpswAleRegBase,
                                      uint32_t portNum,
                                      CpswAle_MacAddrInfo *macAddrBuffer,
                                      uint32_t maxNumAddr,
                                      uint32_t *portTotalMacAddr);

static int32_t CpswAle_configThreadPolicer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t aleFreqHz,
                                           const CpswAle_PolicerMatchParams *policerMatch,
                                           bool threadIdEnable,
                                           uint32_t threadId,
                                           CpswAle_SetPolicerThreadConfigOutArgs *outArgs);

static int32_t CpswAle_setBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          const CpswAle_SetBcastMcastRateLimitInArgs *inArgs);

static int32_t CpswAle_getBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          CpswAle_GetBcastMcastRateLimitOutArgs *outArgs);

static int32_t CpswAle_getPolicerStatsHandler(CSL_AleRegs *cpswAleRegBase,
                                              const CpswAle_PolicerMatchParams *policerMatchParams,
                                              bool *pPolHit,
                                              bool *pPolRedHit,
                                              bool *pPolYellowHit,
                                              bool clearStats);

static int32_t CpswAle_delPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                  uint32_t delAleEntryMask);

static int32_t CpswAle_getPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                  CpswAle_GetPolicerEntryOutArgs *outArgs,
                                  uint32_t aleFreqHz);

static int32_t CpswAle_setPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                  uint32_t peakRateInBitsPerSec,
                                  uint32_t commitRateInBitsPerSec,
                                  uint32_t aleFreqHz,
                                  bool threadIdEnable,
                                  uint32_t threadId,
                                  CpswAle_SetPolicerEntryOutArgs *outArgs);

static int32_t CpswAle_delIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv6AddrType_t ipv6Addr,
                                   uint32_t numLSBIgnore);

static int32_t CpswAle_setRxFilterNone(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setRxFilterDirect(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setRxFilterBcast(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setRxFilterMcast(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setRxFilterAllMcast(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_setRxFilterAll(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_addVlanEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t vlanId,
                                    bool outVlanFlag,
                                    uint32_t *aleEntryIdx);

static int32_t CpswAle_addMacAddrEntry(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t portNum,
                                       CPSW_ALE_ADDR_TYPE addrType,
                                       uint32_t *aleEntryIdx);

static uint32_t CpswAle_mapBw2IdlIncVal(uint32_t aleFreq,
                                        uint32_t rateInBps);

static int32_t CpswAle_getPolicerIndex(CSL_AleRegs *cpswAleRegBase,
                                       const CpswAle_PolicerMatchParams *policerMatchParams,
                                       uint32_t *policerIndex);

static void CpswAle_getVlanMcastPortMask(CSL_AleRegs *cpswAleRegBase,
                                         const CSL_CPSW_ALE_VLAN_ENTRY *vlanEntry,
                                         uint32_t *regMcastFloodMask,
                                         uint32_t *unregMcastFloodMask);

static int32_t CpswAle_setVlanMcastPortMask(CSL_AleRegs *cpswAleRegBase,
                                            CSL_CPSW_ALE_VLAN_ENTRY *vlanEntry,
                                            uint32_t aleIndex,
                                            bool outerVlanFlag,
                                            uint32_t regMcastFloodMask,
                                            uint32_t unregMcastFloodMask);

static void CpswAle_validateRegisterResetVal(volatile uint32_t *regAddr,
                                             uint32_t resetValExpected);

static int32_t CpswAle_addVlanMask(CSL_AleRegs *cpswAleRegBase,
                                   uint32_t mask,
                                   uint32_t *entry);

static int32_t CpswAle_getPortState(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t portNum,
                                    CPSW_ALE_PORTSTATE_E *pPortState);

static void CpswAle_mapMcastMaskBits2RegCfg(uint8_t *addr,
                                            uint32_t numIgnBits,
                                            uint8_t *maskedAddr,
                                            uint32_t *ignMBits,
                                            CSL_CPSW_ALETABLE_TYPE aleType);

static bool CpswAle_isVlanMaskMuxRegSupported(CSL_AleRegs *cpswAleRegBase);

static uintptr_t CpswAle_getBaseAddr(Cpsw_Type cpswType);

static Cpsw_Type CpswAle_getCpswType(uintptr_t cpswAleRegBase);

static bool CpswAle_isIPHdrWhiteListSupported(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_getPolicerDefaultThreadConfig(CSL_AleRegs *cpswAleRegBase,
                                                     CpswAle_GetDefaultThreadConfigOutArgs *defaultThreadConfig);

static bool CpswAle_isFlowPrioritySupported(CSL_AleRegs *cpswAleRegBase);

static void CpswAle_printRegs(CSL_AleRegs *cpswAleRegBase);

static int32_t CpswAle_getPolicerControl(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t aleFreqHz,
                                         CpswAle_GetPolicerGlobalConfigOutArgs *policerCfg);

static int32_t CpswAle_getNoMatchEntry0Policer(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t aleFreq,
                                               uint32_t *pPeakRateInBitsPs,
                                               uint32_t *pCommitRateInBitsPs);

static int32_t CpswAle_delAllPolicerThreadId(CSL_AleRegs *cpswAleRegBase,
                                             uint32_t threadId);

static int32_t CpswAle_blockClassifierToHostPort(CSL_AleRegs *cpswAleRegBase,
                                                 uint32_t aleFreqHz,
                                                 const CpswAle_PolicerMatchParams *policerMatch,
                                                 CpswAle_BlockClassifierHostPortOutArgs *outArgs);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Data to keep track of open references to the ALE */
static CpswAle_Obj *CpswAle_aleRef[CPSW_COUNT] = {NULL, NULL};

/*! \brief CPSW ALE object */
static CpswAle_Obj CpswAle_aleObj[CPSW_COUNT];

const CpswAle_IoctlValidateEntry CpswAleIoctlInfo[] =
{
    [CPSW_ALE_MINOR(ADD_UNICAST)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_UNICAST,
        .inArgsSize  = sizeof(CpswAle_SetUcastEntryInArgs),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_MULTICAST)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_MULTICAST,
        .inArgsSize  = sizeof(CpswAle_SetMcastEntryInArgs),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_VLAN)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_VLAN,
        .inArgsSize  = sizeof(CpswAle_VlanEntryInfo),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_OUI)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_OUI,
        .inArgsSize  = sizeof(CpswAle_OuiEntryInfo),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_IPV4ADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_IPV4ADDR,
        .inArgsSize  = sizeof(CpswAle_Ipv4EntryInfo),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_IPV6ADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_IPV6ADDR,
        .inArgsSize  = sizeof(CpswAle_Ipv6EntryInfo),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(ADD_ETHERTYPE)] =
    {
        .cmd         = CPSW_ALE_IOCTL_ADD_ETHERTYPE,
        .inArgsSize  = sizeof(CpswAle_EtherTypeEntryInfo),
        .outArgsSize = sizeof(CpswAle_AddEntryOutArgs)
    },
    [CPSW_ALE_MINOR(LOOKUP_UNICAST)] =
    {
        .cmd         = CPSW_ALE_IOCTL_LOOKUP_UNICAST,
        .inArgsSize  = sizeof(CpswAle_MacAddrInfo),
        .outArgsSize = sizeof(CpswAle_GetUcastEntryOutArgs)
    },
    [CPSW_ALE_MINOR(LOOKUP_MULTICAST)] =
    {
        .cmd         = CPSW_ALE_IOCTL_LOOKUP_MULTICAST,
        .inArgsSize  = sizeof(CpswAle_GetMcastEntryInArgs),
        .outArgsSize = sizeof(CpswAle_GetMcastEntryOutArgs)
    },
    [CPSW_ALE_MINOR(LOOKUP_VLAN)] =
    {
        .cmd         = CPSW_ALE_IOCTL_LOOKUP_VLAN,
        .inArgsSize  = sizeof(CpswAle_VlanIdInfo),
        .outArgsSize = sizeof(CpswAle_GetVlanEntryOutArgs)
    },
    [CPSW_ALE_MINOR(REMOVE_ADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_ADDR,
        .inArgsSize  = sizeof(CpswAle_MacAddrInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(REMOVE_VLAN)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_VLAN,
        .inArgsSize  = sizeof(CpswAle_VlanIdInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(REMOVE_OUI)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_OUI,
        .inArgsSize  = sizeof(CpswAle_OuiEntryInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(REMOVE_IPV4ADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_IPV4ADDR,
        .inArgsSize  = sizeof(CpswAle_Ipv4EntryInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(REMOVE_IPV6ADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_IPV6ADDR,
        .inArgsSize  = sizeof(CpswAle_Ipv6EntryInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(REMOVE_ETHERTYPE)] =
    {
        .cmd         = CPSW_ALE_IOCTL_REMOVE_ETHERTYPE,
        .inArgsSize  = sizeof(CpswAle_EtherTypeEntryInfo),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_POLICER)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_POLICER,
        .inArgsSize  = sizeof(CpswAle_SetPolicerEntryInArgs),
        .outArgsSize = sizeof(CpswAle_SetPolicerEntryOutArgs)
    },
    [CPSW_ALE_MINOR(GET_POLICER)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_POLICER,
        .inArgsSize  = sizeof(CpswAle_PolicerMatchParams),
        .outArgsSize = sizeof(CpswAle_GetPolicerEntryOutArgs)
    },
    [CPSW_ALE_MINOR(DEL_POLICER)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DEL_POLICER,
        .inArgsSize  = sizeof(CpswAle_DelPolicerEntryInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(DUMP_TABLE)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DUMP_TABLE,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(DUMP_POLICER_ENTRIES)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(GET_POLICER_STATS)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_POLICER_STATS,
        .inArgsSize  = sizeof(CpswAle_GetPolicerStatsInArgs),
        .outArgsSize = sizeof(CpswAle_GetPolicerStatsOutArgs)
    },
    [CPSW_ALE_MINOR(SET_PORT_MIRROR_CFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_PORT_MIRROR_CFG,
        .inArgsSize  = sizeof(CpswAle_SetPortMirroringConfigInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_TRUNK_CFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_TRUNK_CFG,
        .inArgsSize  = sizeof(CpswAle_SetTrunkConfigInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_OAMLPBK_CFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_OAMLPBK_CFG,
        .inArgsSize  = sizeof(CpswAle_SetOAMLoopbackInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(DEL_ALL_ENTRIES)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DEL_ALL_ENTRIES,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_RX_FILTER)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_RX_FILTER,
        .inArgsSize  = sizeof(CpswAle_HostRxFilterParams),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(GET_RX_FILTER)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_RX_FILTER,
        .inArgsSize  = 0,
        .outArgsSize = sizeof(CpswAle_HostRxFilterParams)
    },
    [CPSW_ALE_MINOR(SET_PORT_STATE)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_PORT_STATE,
        .inArgsSize  = sizeof(CpswAle_SetPortStateInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_BCAST_MCAST_LIMIT)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT,
        .inArgsSize  = sizeof(CpswAle_SetBcastMcastRateLimitInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_DEFAULT_THREADCFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG,
        .inArgsSize  = sizeof(CpswAle_SetDefaultThreadConfigInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(AGE_ALE_ENTRIES)] =
    {
        .cmd         = CPSW_ALE_IOCTL_AGE_ALE_ENTRIES,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_INTERVLAN_CFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_INTERVLAN_CFG,
        .inArgsSize  = sizeof(CpswAle_SetInterVLANConfigInArgs),
        .outArgsSize = sizeof(CpswAle_SetInterVLANConfigOutArgs)
    },
    [CPSW_ALE_MINOR(DEL_LEARNT_ENTRIES)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DEL_LEARNT_ENTRIES,
        .inArgsSize  = sizeof(CpswAle_DelLearntEntriesInArgs),
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_POLICER_THREADCFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_POLICER_THREADCFG,
        .inArgsSize  = sizeof(CpswAle_SetPolicerThreadConfigInArgs),
        .outArgsSize = sizeof(CpswAle_SetPolicerThreadConfigOutArgs)
    },
    [CPSW_ALE_MINOR(GET_PORT_MACADDR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_PORT_MACADDR,
        .inArgsSize  = sizeof(CpswAle_GetPortMacAddrInArgs),
        .outArgsSize = sizeof(CpswAle_GetPortMacAddrOutArgs)
    },
    [CPSW_ALE_MINOR(DISABLE_PORT_MIRROR)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DISABLE_PORT_MIRROR,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(DISABLE_BCAST_MCAST_LIMIT)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DISABLE_BCAST_MCAST_LIMIT,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(SET_POLICER_GLOBAL_CONFIG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_SET_POLICER_GLOBAL_CONFIG,
        .inArgsSize  = 0,
        .outArgsSize = 0
    },
    [CPSW_ALE_MINOR(GET_BCAST_MCAST_LIMIT)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT,
        .inArgsSize  = 0,
        .outArgsSize = sizeof(CpswAle_GetBcastMcastRateLimitOutArgs)
    },
    [CPSW_ALE_MINOR(GET_PORT_STATE)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_PORT_STATE,
        .inArgsSize  = sizeof(CpswAle_GetPortStateInArgs),
        .outArgsSize = sizeof(CpswAle_GetPortStateOutArgs)
    },
    [CPSW_ALE_MINOR(GET_DEFAULT_THREADCFG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG,
        .inArgsSize  = 0,
        .outArgsSize = sizeof(CpswAle_GetDefaultThreadConfigOutArgs),
    },
    [CPSW_ALE_MINOR(PRINT_REGS)] =
    {
        .cmd         = CPSW_ALE_IOCTL_PRINT_REGS,
        .inArgsSize  = 0,
        .outArgsSize = 0,
    },
    [CPSW_ALE_MINOR(GET_POLICER_GLOBAL_CONFIG)] =
    {
        .cmd         = CPSW_ALE_IOCTL_GET_POLICER_GLOBAL_CONFIG,
        .inArgsSize  = 0,
        .outArgsSize = sizeof(CpswAle_GetPolicerGlobalConfigOutArgs),
    },
    [CPSW_ALE_MINOR(DEL_ALL_POLICER_THREADID)] =
    {
        .cmd         = CPSW_ALE_IOCTL_DEL_ALL_POLICER_THREADID,
        .inArgsSize  = sizeof(uint32_t),
        .outArgsSize = 0,
    },
    [CPSW_ALE_MINOR(BLOCK_CLASSIFIER_HOSTPORT)] =
    {
        .cmd         = CPSW_ALE_IOCTL_BLOCK_CLASSIFIER_HOSTPORT,
        .inArgsSize  = sizeof(CpswAle_PolicerMatchParams),
        .outArgsSize = sizeof(CpswAle_BlockClassifierHostPortOutArgs),
    },
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

CpswAle_Handle CpswAle_open(Cpsw_Type cpswType,
                            const CpswAle_Config *pALEConfig)
{
    CpswAle_Handle hAle = NULL;
    const CpswSoc_Config *pSocCfg;
    uint32_t numAleEntries;
    uint32_t numPolicers;

    if (IS_NULL(CpswAle_aleRef[cpswType]))
    {
        /* Check open is not invoked previously. */
        /* CpswAle_aleRef will be NULL if open not done previously */
        CpswAle_Obj *pAle = &CpswAle_aleObj[cpswType];
        pSocCfg = CpswSoc_getConfig(cpswType);

        memset(pAle, 0, sizeof(*pAle));
        pAle->magic = CPSW_ALE_MAGIC;
        CpswAle_initcpswAleRegBase(&pAle->cpswAleRegBase, cpswType);

        /* Assert at compile time that offsets match for first and last
         *register in ALE register overlay */
        CPSW_UTILS_COMPILETIME_ASSERT(offsetof(CSL_AleRegs, MOD_VER) == CSL_ALE_MOD_VER);
        CPSW_UTILS_COMPILETIME_ASSERT(offsetof(CSL_AleRegs, THREADMAPVAL) == CSL_ALE_THREADMAPVAL);

        /* Assert flows num greater than CPSW instance max. supported */
        Cpsw_assert(pSocCfg->rxFlowCount <= (CSL_ALE_THREADMAPVAL_THREADVAL_MAX + 1));
        CpswAle_validateIPVersion(pAle->cpswAleRegBase);
        CpswAle_validateRegisterResetVal(&pAle->cpswAleRegBase->THREADMAPVAL,
                                         CSL_ALE_THREADMAPVAL_RESETVAL);
        CpswAle_setAleConfig(pAle,
                             pALEConfig,
                             pSocCfg->cppiClkFreqHz,
                             cpswType);

        /* Add default entries */
        CpswAle_addDefaultAleEntries(pAle);

        numAleEntries = CpswAle_getMaxAleEntries(pAle->cpswAleRegBase);
        numPolicers = CpswAle_getMaxPolicers(pAle->cpswAleRegBase);

        switch (cpswType)
        {
            case CPSW_9G:
                Cpsw_assert(CPSW_ALE_9G_MAX_ALE_ENTRIES == numAleEntries && CPSW_ALE_9G_MAX_POLICERS == numPolicers);
                break;

            case CPSW_5G:
                Cpsw_assert(CPSW_ALE_5G_MAX_ALE_ENTRIES == numAleEntries && CPSW_ALE_5G_MAX_POLICERS == numPolicers);
                break;

            case CPSW_2G:
                Cpsw_assert(CPSW_ALE_2G_MAX_ALE_ENTRIES == numAleEntries && CPSW_ALE_2G_MAX_POLICERS == numPolicers);
                break;

            default:
                Cpsw_assert(FALSE);
        }

        /* Save a pointer to the state structure to be used to provide a
         *reference for future open calls */
        CpswAle_aleRef[cpswType] = pAle;
        hAle = pAle;
    }
    else
    {
        /*
         * If reference count at entry was non-zero, ALE was already opened,
         * so return error as open can be called only once
         */
        hAle = NULL;
    }

    return hAle;
}

int32_t CpswAle_close(CpswAle_Handle hAle)
{
    uint32_t i;
    CpswAle_Obj *pAle = (CpswAle_Obj *)hAle;
    int32_t retVal = CPSW_SOK;

    if (!IS_NULL(pAle) && !IS_NULL(CpswAle_aleRef[pAle->cpswType]))
    {
        /* De-activate the ALE ageout timer here */
        pAle->aleSoftTimerActive = 0;

        CpswAle_delAllPolicerEntries(pAle->cpswAleRegBase);

        CpswAle_delAllEntries(pAle->cpswAleRegBase);

        for (i = 0; i < CpswAle_getNumPorts(pAle->cpswType); i++)
        {
            CpswAle_setAlePortState(pAle->cpswAleRegBase, i,
                                    CSL_ALE_PORTSTATE_DISABLED);
        }

        CpswAle_aleRef[pAle->cpswType] = NULL;

        memset(pAle, 0, sizeof(*pAle));
    }
    else
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    return retVal;
}

int32_t CpswAle_ioctl(CpswAle_Handle hAle,
                      CPSW_ALE_IOCTL_E cmd,
                      Cpsw_IoctlPrms *prms)
{
    CpswAle_Obj *pAle = (CpswAle_Obj *)hAle;
    int32_t retVal = CPSW_EFAIL;

    /* Decode the command and act on it */
    switch (cmd)
    {
        case CPSW_ALE_IOCTL_ADD_UNICAST:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_UNICAST)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetUcastEntryInArgs *inArgs = (const CpswAle_SetUcastEntryInArgs *)prms->inArgs;
                CpswAle_AddEntryOutArgs *outArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;
                const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                    CPSW_ALE_MAC_ADDR_INIT_LIST(inArgs->addr.addr);
                bool blocked = inArgs->info.blocked;
                bool secure = inArgs->info.secure;

                if (inArgs->info.super == TRUE)
                {
                    blocked = TRUE;
                    secure = TRUE;
                }

                retVal = CpswAle_addUcastAddr(pAle->cpswAleRegBase,
                                              inArgs->info.portNum,
                                              &addr,
                                              inArgs->addr.vlanId,
                                              blocked,
                                              secure,
                                              inArgs->info.ageable,
                                              inArgs->info.trunk,
                                              &outArgs->aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_MULTICAST:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_MULTICAST)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetMcastEntryInArgs *inArgs = (const CpswAle_SetMcastEntryInArgs *)prms->inArgs;
                CpswAle_AddEntryOutArgs *outArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;
                const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                    CPSW_ALE_MAC_ADDR_INIT_LIST(inArgs->addr.addr);

                retVal = CpswAle_addMcastAddr(pAle->cpswAleRegBase,
                                              &addr,
                                              inArgs->addr.vlanId,
                                              inArgs->info.portMask,
                                              inArgs->info.superFlag,
                                              inArgs->info.fwdState,
                                              inArgs->info.numIgnBits,
                                              &outArgs->aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_VLAN:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_VLAN)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_VlanEntryInfo *vlanEntryInArgs = (const CpswAle_VlanEntryInfo *)prms->inArgs;
                CpswAle_AddEntryOutArgs *vlanEntryOutArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;

                retVal = CpswAle_addVlan(pAle->cpswAleRegBase,
                                         vlanEntryInArgs->vlanIdInfo.vlanId,
                                         vlanEntryInArgs->vlanIdInfo.outerVlanFlag,
                                         vlanEntryInArgs->vlanMemberList,
                                         vlanEntryInArgs->unregMcastFloodMask,
                                         vlanEntryInArgs->regMcastFloodMask,
                                         vlanEntryInArgs->forceUntaggedEgressMask,
                                         vlanEntryInArgs->noLearnMask,
                                         vlanEntryInArgs->vidIngressCheck,
                                         vlanEntryInArgs->limitIPNxtHdr,
                                         vlanEntryInArgs->disallowIPFragmentation,
                                         &vlanEntryOutArgs->aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_OUI:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_OUI)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_OuiEntryInfo *ouiInArgs = (const CpswAle_OuiEntryInfo *)prms->inArgs;
                CpswAle_AddEntryOutArgs *ouiOutArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;
                const uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
                    CPSW_ALE_OUI_ADDR_INIT_LIST(ouiInArgs->ouiAddr);

                retVal = CpswAle_addOuiAddr(pAle->cpswAleRegBase,
                                            &ouiAddr,
                                            &ouiOutArgs->aleEntryIndex,
                                            NULL);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_IPV4ADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_IPV4ADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_Ipv4EntryInfo *inArgs = (const CpswAle_Ipv4EntryInfo *)prms->inArgs;
                CpswAle_AddEntryOutArgs *outArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;
                const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                    CPSW_ALE_IPV4_ADDR_INIT_LIST(inArgs->ipv4Addr);

                retVal = CpswAle_addIPv4Addr(pAle->cpswAleRegBase,
                                             &ipv4Addr,
                                             inArgs->numLSBIgnoreBits,
                                             &outArgs->aleEntryIndex,
                                             NULL);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_IPV6ADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_IPV6ADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_Ipv6EntryInfo *inArgs = (const CpswAle_Ipv6EntryInfo *)prms->inArgs;
                CpswAle_AddEntryOutArgs *outArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;
                const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                    CPSW_ALE_IPV6_ADDR_INIT_LIST(inArgs->ipv6Addr);

                retVal = CpswAle_addIPv6Addr(pAle->cpswAleRegBase,
                                             &ipv6Addr,
                                             inArgs->numLSBIgnoreBits,
                                             &outArgs->aleEntryIndex,
                                             NULL);
            }

            break;
        }

        case CPSW_ALE_IOCTL_ADD_ETHERTYPE:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(ADD_ETHERTYPE)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_EtherTypeEntryInfo *inArgs = (const CpswAle_EtherTypeEntryInfo *)prms->inArgs;
                CpswAle_AddEntryOutArgs *outArgs = (CpswAle_AddEntryOutArgs *)prms->outArgs;

                retVal = CpswAle_addEtherType(pAle->cpswAleRegBase,
                                              inArgs->etherType,
                                              &outArgs->aleEntryIndex,
                                              NULL);
            }

            break;
        }

        case CPSW_ALE_IOCTL_LOOKUP_UNICAST:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(LOOKUP_UNICAST)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_MacAddrInfo *inArgs = (const CpswAle_MacAddrInfo *)prms->inArgs;
                CpswAle_GetUcastEntryOutArgs *outArgs = (CpswAle_GetUcastEntryOutArgs *)prms->outArgs;
                const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                    CPSW_ALE_MAC_ADDR_INIT_LIST(inArgs->addr);

                retVal = CpswAle_lookupUcastAddr(pAle->cpswAleRegBase,
                                                 &addr,
                                                 inArgs->vlanId,
                                                 &outArgs->unicastInfo.portNum,
                                                 &outArgs->unicastInfo.blocked,
                                                 &outArgs->unicastInfo.secure,
                                                 &outArgs->unicastInfo.super,
                                                 &outArgs->unicastInfo.ageable,
                                                 &outArgs->unicastInfo.trunk,
                                                 &outArgs->touched,
                                                 &outArgs->aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_LOOKUP_MULTICAST:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(LOOKUP_MULTICAST)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_GetMcastEntryInArgs *inArgs = (const CpswAle_GetMcastEntryInArgs *)prms->inArgs;
                CpswAle_GetMcastEntryOutArgs *outArgs = (CpswAle_GetMcastEntryOutArgs *)prms->outArgs;
                const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                    CPSW_ALE_MAC_ADDR_INIT_LIST(inArgs->addr.addr);

                retVal = CpswAle_lookupMcastAddr(pAle->cpswAleRegBase,
                                                 &addr,
                                                 inArgs->addr.vlanId,
                                                 inArgs->numIgnBits,
                                                 &outArgs->outInfo.portMask,
                                                 &outArgs->outInfo.fwdState,
                                                 &outArgs->outInfo.superFlag,
                                                 &outArgs->outInfo.numIgnBits,
                                                 &outArgs->aleIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_LOOKUP_VLAN:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(LOOKUP_VLAN)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_VlanIdInfo *inArgs = (const CpswAle_VlanIdInfo *)prms->inArgs;
                CpswAle_GetVlanEntryOutArgs *outArgs = (CpswAle_GetVlanEntryOutArgs *)prms->outArgs;

                retVal = CpswAle_findVlan(pAle->cpswAleRegBase,
                                          inArgs->vlanId,
                                          inArgs->outerVlanFlag,
                                          &outArgs->vlanMemberList,
                                          &outArgs->unregMcastFloodMask,
                                          &outArgs->regMcastFloodMask,
                                          &outArgs->forceUntaggedEgressMask,
                                          &outArgs->noLearnMask,
                                          &outArgs->vidIngressCheck,
                                          &outArgs->limitIPNxtHdr,
                                          &outArgs->disallowIPFragmentation,
                                          &outArgs->aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_ADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_ADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_MacAddrInfo *inArgs = (const CpswAle_MacAddrInfo *)prms->inArgs;
                const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                    CPSW_ALE_MAC_ADDR_INIT_LIST(inArgs->addr);

                retVal = CpswAle_delAddr(pAle->cpswAleRegBase, &addr,
                                         inArgs->vlanId);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_VLAN:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_VLAN)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_VlanIdInfo *inArgs = (CpswAle_VlanIdInfo *)prms->inArgs;

                retVal = CpswAle_delVlan(pAle->cpswAleRegBase, inArgs->vlanId,
                                         inArgs->outerVlanFlag);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_OUI:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_OUI)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_OuiEntryInfo *inArgs = (const CpswAle_OuiEntryInfo *)prms->inArgs;
                const uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
                    CPSW_ALE_OUI_ADDR_INIT_LIST(inArgs->ouiAddr);

                retVal = CpswAle_delOuiAddr(pAle->cpswAleRegBase, &ouiAddr);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_IPV4ADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_IPV4ADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_Ipv4EntryInfo *inArgs = (const CpswAle_Ipv4EntryInfo *)prms->inArgs;
                const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                    CPSW_ALE_IPV4_ADDR_INIT_LIST(inArgs->ipv4Addr);

                retVal = CpswAle_delIPv4Addr(pAle->cpswAleRegBase,
                                             &ipv4Addr,
                                             inArgs->numLSBIgnoreBits);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_IPV6ADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_IPV6ADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_Ipv6EntryInfo *inArgs = (const CpswAle_Ipv6EntryInfo *)prms->inArgs;
                const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                    CPSW_ALE_IPV6_ADDR_INIT_LIST(inArgs->ipv6Addr);

                retVal = CpswAle_delIPv6Addr(pAle->cpswAleRegBase,
                                             &ipv6Addr,
                                             inArgs->numLSBIgnoreBits);
            }

            break;
        }

        case CPSW_ALE_IOCTL_REMOVE_ETHERTYPE:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(REMOVE_ETHERTYPE)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_EtherTypeEntryInfo *inArgs = (const CpswAle_EtherTypeEntryInfo *)prms->inArgs;

                retVal = CpswAle_delEtherType(pAle->cpswAleRegBase,
                                              inArgs->etherType);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_POLICER:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_POLICER)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetPolicerEntryInArgs *inArgs = (const CpswAle_SetPolicerEntryInArgs *)prms->inArgs;
                CpswAle_SetPolicerEntryOutArgs *outArgs = (CpswAle_SetPolicerEntryOutArgs *)prms->outArgs;

                retVal = CpswAle_setPolicer(pAle->cpswAleRegBase,
                                            &inArgs->policerMatch,
                                            inArgs->peakRateInBitsPerSec,
                                            inArgs->commitRateInBitsPerSec,
                                            pAle->aleFreqHz,
                                            inArgs->threadIdEnable,
                                            inArgs->threadId,
                                            outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_POLICER:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_POLICER)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_PolicerMatchParams *inArgs = (const CpswAle_PolicerMatchParams *)prms->inArgs;
                CpswAle_GetPolicerEntryOutArgs *outArgs = (CpswAle_GetPolicerEntryOutArgs *)prms->outArgs;

                retVal = CpswAle_getPolicer(pAle->cpswAleRegBase,
                                            inArgs,
                                            outArgs,
                                            pAle->aleFreqHz);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DEL_POLICER:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DEL_POLICER)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_DelPolicerEntryInArgs *inArgs = (const CpswAle_DelPolicerEntryInArgs *)prms->inArgs;

                retVal = CpswAle_delPolicer(pAle->cpswAleRegBase,
                                            &inArgs->policerMatch,
                                            inArgs->delAleEntryMask);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DUMP_TABLE:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DUMP_TABLE)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_dumpTable(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DUMP_POLICER_ENTRIES:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DUMP_POLICER_ENTRIES)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_dumpPolicerEntries(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_POLICER_STATS:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_POLICER_STATS)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_GetPolicerStatsInArgs *inArgs = (const CpswAle_GetPolicerStatsInArgs *)prms->inArgs;
                CpswAle_GetPolicerStatsOutArgs *outArgs = (CpswAle_GetPolicerStatsOutArgs *)prms->outArgs;

                retVal = CpswAle_getPolicerStatsHandler(pAle->cpswAleRegBase,
                                                        &inArgs->policerInfo,
                                                        &outArgs->policerHit,
                                                        &outArgs->policerRedHit,
                                                        &outArgs->policerYellowHit,
                                                        inArgs->clearStats);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_PORT_MIRROR_CFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_PORT_MIRROR_CFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetPortMirroringConfigInArgs *inArgs =
                    (const CpswAle_SetPortMirroringConfigInArgs *)prms->inArgs;
                uint32_t maxNumPorts = CpswAle_getNumPorts(pAle->cpswType);

                retVal = CpswAle_setPortMirrorConfig(pAle->cpswAleRegBase,
                                                     maxNumPorts,
                                                     inArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_TRUNK_CFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_TRUNK_CFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetTrunkConfigInArgs *inArgs = (const CpswAle_SetTrunkConfigInArgs *)prms->inArgs;

                retVal = CpswAle_setPortTrunkConfig(pAle->cpswAleRegBase, inArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_OAMLPBK_CFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_OAMLPBK_CFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetOAMLoopbackInArgs *inArgs = (const CpswAle_SetOAMLoopbackInArgs *)prms->inArgs;

                retVal = CpswAle_setOAMLpbk(pAle->cpswAleRegBase,
                                            inArgs->enableLpbkPortMask);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DEL_ALL_ENTRIES:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DEL_ALL_ENTRIES)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                retVal = CpswAle_delAllEntries(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_RX_FILTER:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_RX_FILTER)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_HostRxFilterParams *inArgs = (const CpswAle_HostRxFilterParams *)prms->inArgs;

                retVal = CpswAle_setReceiveFilter(pAle, inArgs->rxFilter);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_RX_FILTER:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_RX_FILTER)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_HostRxFilterParams *outArgs = (CpswAle_HostRxFilterParams *)prms->outArgs;

                retVal = CpswAle_getReceiveFilter(pAle, &outArgs->rxFilter);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_PORT_STATE:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_PORT_STATE)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetPortStateInArgs *inArgs = (const CpswAle_SetPortStateInArgs *)prms->inArgs;

                retVal = CpswAle_setAlePortState(pAle->cpswAleRegBase,
                                                 inArgs->portNum,
                                                 (CSL_CPSW_ALE_PORTSTATE)inArgs->portState);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_BCAST_MCAST_LIMIT:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_BCAST_MCAST_LIMIT)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetBcastMcastRateLimitInArgs *inArgs =
                    (const CpswAle_SetBcastMcastRateLimitInArgs *)prms->inArgs;

                retVal = CpswAle_setBcastMcastLimit(pAle->cpswAleRegBase,
                                                    pAle->aleFreqHz,
                                                    inArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_BCAST_MCAST_LIMIT:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_BCAST_MCAST_LIMIT)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_GetBcastMcastRateLimitOutArgs *outArgs = (CpswAle_GetBcastMcastRateLimitOutArgs *)prms->outArgs;

                retVal = CpswAle_getBcastMcastLimit(pAle->cpswAleRegBase,
                                                    pAle->aleFreqHz,
                                                    outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_DEFAULT_THREADCFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_DEFAULT_THREADCFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetDefaultThreadConfigInArgs *inArgs =
                    (const CpswAle_SetDefaultThreadConfigInArgs *)prms->inArgs;

                retVal = CpswAle_setPolicerDefaultThreadConfig(pAle->cpswAleRegBase,
                                                               inArgs->defaultThreadEnable,
                                                               inArgs->threadId,
                                                               inArgs->enablePriorityOR,
                                                               inArgs->disableMacPortDefaultThread);
            }

            break;
        }

        case CPSW_ALE_IOCTL_AGE_ALE_ENTRIES:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(AGE_ALE_ENTRIES)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_ageOut(pAle->cpswAleRegBase,
                               pAle->aleSoftTimerActive,
                               pAle->aleTickTimeOutCount,
                               &pAle->aleSoftTickCount);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_INTERVLAN_CFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_INTERVLAN_CFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetInterVLANConfigInArgs *inArgs = (const CpswAle_SetInterVLANConfigInArgs *)prms->inArgs;
                CpswAle_SetInterVLANConfigOutArgs *outArgs = (CpswAle_SetInterVLANConfigOutArgs *)prms->outArgs;

                retVal = CpswAle_setInterVlanConfig(pAle->cpswAleRegBase, pAle->aleFreqHz,
                                                    inArgs,
                                                    outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DEL_LEARNT_ENTRIES:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DEL_LEARNT_ENTRIES)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_DelLearntEntriesInArgs *inArgs = (const CpswAle_DelLearntEntriesInArgs *)prms->inArgs;

                retVal = CpswAle_delPortUcastEntries(pAle->cpswAleRegBase,
                                                     inArgs->portNum);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_POLICER_THREADCFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_POLICER_THREADCFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_SetPolicerThreadConfigInArgs *inArgs =
                    (const CpswAle_SetPolicerThreadConfigInArgs *)prms->inArgs;
                CpswAle_SetPolicerThreadConfigOutArgs *outArgs =
                    (CpswAle_SetPolicerThreadConfigOutArgs *)prms->outArgs;

                retVal = CpswAle_configThreadPolicer(pAle->cpswAleRegBase,
                                                     pAle->aleFreqHz,
                                                     &inArgs->policerMatch,
                                                     inArgs->threadIdEnable,
                                                     inArgs->threadId,
                                                     outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_PORT_MACADDR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_PORT_MACADDR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_GetPortMacAddrInArgs *inArgs = (const CpswAle_GetPortMacAddrInArgs *)prms->inArgs;
                CpswAle_GetPortMacAddrOutArgs *outArgs = (CpswAle_GetPortMacAddrOutArgs *)prms->outArgs;

                retVal = CpswAle_getPortMacAddr(pAle->cpswAleRegBase,
                                                inArgs->portNum,
                                                inArgs->macAddr,
                                                inArgs->maxNumAddr,
                                                &outArgs->totalPortMacAddrCount);
                if (CPSW_SOK == retVal)
                {
                    outArgs->macAddr = inArgs->macAddr;
                    if (outArgs->totalPortMacAddrCount < inArgs->maxNumAddr)
                    {
                        outArgs->addrCount = outArgs->totalPortMacAddrCount;
                    }
                    else
                    {
                        outArgs->addrCount = inArgs->maxNumAddr;
                    }
                }
            }

            break;
        }

        case CPSW_ALE_IOCTL_DISABLE_PORT_MIRROR:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DISABLE_PORT_MIRROR)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                retVal = CpswAle_disablePortMirror(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DISABLE_BCAST_MCAST_LIMIT:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DISABLE_BCAST_MCAST_LIMIT)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                retVal = CpswAle_disableBcastMcastLimit(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_SET_POLICER_GLOBAL_CONFIG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(SET_POLICER_GLOBAL_CONFIG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_PolicerGlobalConfig *inArgs = (const CpswAle_PolicerGlobalConfig *)prms->inArgs;

                retVal = CpswAle_setPolicerControl(pAle->cpswAleRegBase,
                                                   pAle->aleFreqHz,
                                                   inArgs->policingEnable,
                                                   inArgs->yellowDropEnable,
                                                   inArgs->redDropEnable,
                                                   inArgs->yellowThresh,
                                                   inArgs->policerNoMatchMode,
                                                   &inArgs->noMatchPolicer);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_PORT_STATE:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_PORT_STATE)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_GetPortStateInArgs *inArgs = (const CpswAle_GetPortStateInArgs *)prms->inArgs;
                CpswAle_GetPortStateOutArgs *outArgs = (CpswAle_GetPortStateOutArgs *)prms->outArgs;

                retVal = CpswAle_getPortState(pAle->cpswAleRegBase,
                                              inArgs->portNum,
                                              &outArgs->portState);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_DEFAULT_THREADCFG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_DEFAULT_THREADCFG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_GetDefaultThreadConfigOutArgs *outArgs =
                    (CpswAle_GetDefaultThreadConfigOutArgs *)prms->outArgs;

                retVal = CpswAle_getPolicerDefaultThreadConfig(pAle->cpswAleRegBase,
                                                               outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_PRINT_REGS:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(PRINT_REGS)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_printRegs(pAle->cpswAleRegBase);
            }

            break;
        }

        case CPSW_ALE_IOCTL_GET_POLICER_GLOBAL_CONFIG:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(GET_POLICER_GLOBAL_CONFIG)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                CpswAle_GetPolicerGlobalConfigOutArgs *outArgs = (CpswAle_GetPolicerGlobalConfigOutArgs *)prms->outArgs;

                retVal = CpswAle_getPolicerControl(pAle->cpswAleRegBase,
                                                   pAle->aleFreqHz,
                                                   outArgs);
            }

            break;
        }

        case CPSW_ALE_IOCTL_DEL_ALL_POLICER_THREADID:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(DEL_ALL_POLICER_THREADID)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const uint32_t *threadId = (const uint32_t *)prms->inArgs;

                retVal = CpswAle_delAllPolicerThreadId(pAle->cpswAleRegBase,
                                                       *threadId);
            }

            break;
        }

        case CPSW_ALE_IOCTL_BLOCK_CLASSIFIER_HOSTPORT:
        {
            const CpswAle_IoctlValidateEntry *entry =
                &CpswAleIoctlInfo[CPSW_ALE_MINOR(BLOCK_CLASSIFIER_HOSTPORT)];

            retVal = CpswAle_validateIoctl(entry, cmd, prms);
            if (CPSW_SOK == retVal)
            {
                const CpswAle_PolicerMatchParams *inArgs = (const CpswAle_PolicerMatchParams *)prms->inArgs;
                CpswAle_BlockClassifierHostPortOutArgs *outArgs = (CpswAle_BlockClassifierHostPortOutArgs *)prms->outArgs;

                retVal = CpswAle_blockClassifierToHostPort(pAle->cpswAleRegBase,
                                                           pAle->aleFreqHz,
                                                           inArgs,
                                                           outArgs);
            }

            break;
        }

        default:
        {
            retVal = CPSW_EUNKNOWN_IOCTL;
            break;
        }
    }

    return retVal;
}

static void CpswAle_initcpswAleRegBase(CSL_AleRegs **cpswAleRegBasePtr,
                                       Cpsw_Type cpswType)
{
    *cpswAleRegBasePtr = (CSL_AleRegs *)CpswAle_getBaseAddr(cpswType);
}

static uint32_t CpswAle_getNumPorts(Cpsw_Type cpswType)
{
    const uint32_t numPorts[CPSW_COUNT] =
    {
        [CPSW_2G] = 2,
        [CPSW_5G] = 5,
        [CPSW_9G] = 9,
    };

    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_COUNT == 3U);
    Cpsw_assert(cpswType < CPSW_UTILS_ARRAYSIZE(numPorts));
    return numPorts[cpswType];
}

static uint32_t CpswAle_getNumPortsFromCpswAleRegBase(uintptr_t cpswAleRegBase)
{
    Cpsw_Type cpswType;

    cpswType = CpswAle_getCpswType(cpswAleRegBase);

    return(CpswAle_getNumPorts(cpswType));
}

static CSL_CPSW_ALETABLE_TYPE CpswAle_getAleTableType(uintptr_t cpswAleRegBase)
{
    const CSL_CPSW_ALETABLE_TYPE aleTableType[CPSW_COUNT] =
    {
        [CPSW_2G] = CSL_CPSW_ALETABLE_TYPE_4PORT,
        [CPSW_5G] = CSL_CPSW_ALETABLE_TYPE_9PORT,
        [CPSW_9G] = CSL_CPSW_ALETABLE_TYPE_9PORT,
    };
    Cpsw_Type cpswType;

    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_COUNT == 3U);
    cpswType = CpswAle_getCpswType(cpswAleRegBase);
    /* Unknown CPSW base address */
    Cpsw_assert(cpswType < CPSW_UTILS_ARRAYSIZE(aleTableType));
    return aleTableType[cpswType];
}

static const CpswSoc_CpswIpInfo *CpswAle_getCpswIpInfoFromBaseAddr(uintptr_t cpswAleRegBase)
{
    uint32_t i;
    const CpswSoc_CpswIpInfoTbl *ipInfoTbl = CpswSoc_getIPInfoTbl();

    for (i = 0; i < ipInfoTbl->numCpswIPs; i++)
    {
#ifdef QNX_OS
        if (ipInfoTbl->info[i].aleInfo.regBaseAddr == CpswUtils_virtToPhyFxn((const void *)cpswAleRegBase, NULL))
#else
        if (ipInfoTbl->info[i].aleInfo.regBaseAddr == cpswAleRegBase)
#endif
        {
            break;
        }
    }

    Cpsw_assert(i < ipInfoTbl->numCpswIPs);
    return &ipInfoTbl->info[i];
}

static Cpsw_Type CpswAle_getCpswType(uintptr_t cpswAleRegBase)
{
    const CpswSoc_CpswIpInfo *ipInfo = CpswAle_getCpswIpInfoFromBaseAddr(cpswAleRegBase);

#ifdef QNX_OS
    if (ipInfo->aleInfo.regBaseAddr != CpswUtils_virtToPhyFxn((const void *)cpswAleRegBase, NULL)) {
        Cpsw_assert(ipInfo->aleInfo.regBaseAddr == CpswUtils_virtToPhyFxn((const void *)cpswAleRegBase, NULL));
    }
#else
    Cpsw_assert(ipInfo->aleInfo.regBaseAddr == cpswAleRegBase);
#endif
    return ipInfo->cpswType;
}

static uintptr_t CpswAle_getBaseAddr(Cpsw_Type cpswType)
{
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
#ifdef QNX_OS
#define CPSW_ALE_SIZE 0x1000
    uintptr_t cpswAleRegBase = NULL;
    cpswAleRegBase = (uintptr_t) mmap_device_memory(0,
        CPSW_ALE_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)ipInfo->aleInfo.regBaseAddr);
    return cpswAleRegBase;
#else
    return ipInfo->aleInfo.regBaseAddr;
#endif
}

static int32_t CpswAle_validateIoctl(const CpswAle_IoctlValidateEntry *entry,
                                     CPSW_ALE_IOCTL_E cmd,
                                     const Cpsw_IoctlPrms *prms)
{
    int32_t retVal = CPSW_SOK;

    if (IS_NULL(prms))
    {
        retVal = CPSW_EMALFORMED_IOCTL;
    }

    if (CPSW_SOK == retVal)
    {
        if (entry->cmd != cmd)
        {
            retVal = CPSW_EMALFORMED_IOCTL;
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (entry->inArgsSize != 0)
        {
            if ((entry->inArgsSize != prms->inArgsSize) || IS_NULL(prms->inArgs))
            {
                retVal = CPSW_EMALFORMED_IOCTL;
            }
        }

        if (entry->outArgsSize != 0)
        {
            if ((entry->outArgsSize != prms->outArgsSize) || IS_NULL(prms->outArgs))
            {
                retVal = CPSW_EMALFORMED_IOCTL;
            }
        }
    }

    if (retVal != CPSW_SOK)
    {
        CPSWTRACE_ERROR("CPSWALE:Malformed IOCTL:%d", cmd);
    }

    return retVal;
}

static int32_t CpswAle_lookupUcastAddr(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t *portNum,
                                       bool *blocked,
                                       bool *secure,
                                       bool *super,
                                       bool *ageable,
                                       bool *trunk,
                                       bool *touched,
                                       uint32_t *aleEntryIndex)
{
    int32_t retVal;
    uint32_t value;

    retVal = CpswAle_findAddr(cpswAleRegBase,
                              addr,
                              vlanId,
                              aleEntryIndex);
    if (CPSW_SOK == retVal)
    {
        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, *aleEntryIndex,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
        {
            CSL_CPSW_ALE_UNICASTADDR_ENTRY ucastAddrEntry;
            CSL_CPSW_getAleUnicastAddrEntry(cpswAleRegBase, *aleEntryIndex,
                                            &ucastAddrEntry,
                                            CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            *portNum = ucastAddrEntry.portNumber;
            *blocked = ucastAddrEntry.blockEnable;
            *secure  = ucastAddrEntry.secureEnable;
            *ageable = ucastAddrEntry.ageable;
            *trunk   = ucastAddrEntry.trunkFlag;
            *touched = ucastAddrEntry.touched;
            if ((ucastAddrEntry.blockEnable) && (ucastAddrEntry.secureEnable))
            {
                *super = TRUE;
                *blocked = FALSE;
                *secure = FALSE;
            }
        }
        else
        {
            CSL_CPSW_ALE_VLANUNICASTADDR_ENTRY vlanUcastAddrEntry;

            Cpsw_assert((value == CSL_ALE_ENTRYTYPE_VLANADDRESS));
            CSL_CPSW_getAleVlanUnicastAddrEntry(cpswAleRegBase, *aleEntryIndex,
                                                &vlanUcastAddrEntry,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            *portNum = vlanUcastAddrEntry.portNumber;
            *blocked = vlanUcastAddrEntry.blockEnable;
            *secure  = vlanUcastAddrEntry.secureEnable;
            *trunk   = vlanUcastAddrEntry.trunkFlag;
            *ageable = vlanUcastAddrEntry.ageable;
            *touched = vlanUcastAddrEntry.touched;
            if ((vlanUcastAddrEntry.blockEnable) &&
                (vlanUcastAddrEntry.secureEnable))
            {
                *super = TRUE;
                *blocked = FALSE;
                *secure = FALSE;
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_lookupMcastAddr(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t numIgnBits,
                                       uint32_t *portMask,
                                       CPSW_ALE_FWDSTLVL_E *fwdState,
                                       bool *super,
                                       uint32_t *pNumIgnBits,
                                       uint32_t *aleEntryIndex)
{
    int32_t retVal;
    uint32_t value;
    uint8_t mcastMaskedAddr[6];
    uint32_t ignMBitsFlag;

    CpswAle_mapMcastMaskBits2RegCfg((uint8_t *)addr,
                                    numIgnBits,
                                    mcastMaskedAddr,
                                    &ignMBitsFlag,
                                    CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

    retVal = CpswAle_findAddr(cpswAleRegBase,
                              (CpswAle_MacAddrType_t)(&mcastMaskedAddr),
                              vlanId,
                              aleEntryIndex);
    if (CPSW_SOK == retVal)
    {
        value = CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                         *aleEntryIndex,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
        {
            CSL_CPSW_ALE_MCASTADDR_ENTRY mcastAddrEntry;
            CSL_CPSW_getAleMcastAddrEntry(cpswAleRegBase,
                                          *aleEntryIndex,
                                          &mcastAddrEntry,
                                          CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

            *portMask    = mcastAddrEntry.portMask;
            *fwdState    = (CPSW_ALE_FWDSTLVL_E)mcastAddrEntry.mcastFwdState;
            *super       = mcastAddrEntry.superEnable;
            *pNumIgnBits = mcastAddrEntry.ignMBits;
            CPSW_UTILS_ARRAY_COPY(mcastMaskedAddr, mcastAddrEntry.macAddress);
        }
        else
        {
            CSL_CPSW_ALE_VLANMCASTADDR_ENTRY vlanMcastAddrEntry;

            Cpsw_assert(value == CSL_ALE_ENTRYTYPE_VLANADDRESS);
            CSL_CPSW_getAleVlanMcastAddrEntry(cpswAleRegBase,
                                              *aleEntryIndex,
                                              &vlanMcastAddrEntry,
                                              CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            *portMask    = vlanMcastAddrEntry.portMask;
            *fwdState    = (CPSW_ALE_FWDSTLVL_E)vlanMcastAddrEntry.mcastFwdState;
            *super       = vlanMcastAddrEntry.superEnable;
            *pNumIgnBits = vlanMcastAddrEntry.ignMBits;
            CPSW_UTILS_ARRAY_COPY(mcastMaskedAddr, vlanMcastAddrEntry.macAddress);
        }

        if (*pNumIgnBits)
        {
            uint32_t bitPos;
            uint32_t word0, word1;

            CSL_CPSW_mapMacAddr2TableWord(&word0, &word1,
                                          mcastMaskedAddr,
                                          CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

            bitPos = 0;

            /* We will check only word0. Ensure max ign bits cannot be greater
             * than 32
             */
            CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_MCAST_IGN_BITS_MAX < 32);
            while (bitPos < CPSW_ALE_MCAST_IGN_BITS_MAX)
            {
                if ((word0 & (1 << bitPos)) == 0)
                {
                    break;
                }

                bitPos++;
            }

            *pNumIgnBits += bitPos;
        }
    }

    return retVal;
}

static void CpswAle_ageOut(CSL_AleRegs *cpswAleRegBase,
                           bool aleTimerActive,
                           uint32_t aleTickTimeOutCount,
                           uint32_t *aleTickCount)
{
    if ((FALSE != aleTimerActive) && ((uint32_t)0U != aleTickTimeOutCount))
    {
        /* aleTickCount can become greater if timeout period is changed in
         *between */
        *aleTickCount += 1;
        if (*aleTickCount >= aleTickTimeOutCount)
        {
            CSL_CPSW_startAleAgeOutNow(cpswAleRegBase);

            while (CSL_CPSW_isAleAgeOutDone(cpswAleRegBase) != TRUE)
            {
            };

            *aleTickCount = 0;
        }
    }
}

static void CpswAle_setUnknownVlanConfig(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t unknownVlanMemberList,
                                         uint32_t unknownVlanForceUntaggedEgress,
                                         uint32_t unknownVlanRegMcastFloodMask,
                                         uint32_t unknownVlanUnRegMcastFloodMask)
{
    uint32_t aleUnknwnVlanVal = 0;

    aleUnknwnVlanVal = CSL_FMK(ALE_ALE_UVLAN_MEMBER_UVLAN_MEMBER_LIST,
                               unknownVlanMemberList);
    CSL_CPSW_setAleUnknwnVlanMemberReg(cpswAleRegBase, aleUnknwnVlanVal);

    aleUnknwnVlanVal = CSL_FMK(ALE_ALE_UVLAN_UNTAG_UVLAN_FORCE_UNTAGGED_EGRESS,
                               unknownVlanForceUntaggedEgress);
    CSL_CPSW_setAleUnknwnVlanUntagReg(cpswAleRegBase, aleUnknwnVlanVal);

    aleUnknwnVlanVal = CSL_FMK(ALE_ALE_UVLAN_RMCAST_UVLAN_REG_MCAST_FLOOD_MASK,
                               unknownVlanRegMcastFloodMask);
    CSL_CPSW_setAleUnknwnVlanRegMcastReg(cpswAleRegBase, aleUnknwnVlanVal);

    aleUnknwnVlanVal = CSL_FMK(ALE_ALE_UVLAN_URCAST_UVLAN_UNREG_MCAST_FLOOD_MASK,
                               unknownVlanUnRegMcastFloodMask);
    CSL_CPSW_setAleUnknwnVlanUnregMcastReg(cpswAleRegBase, aleUnknwnVlanVal);
}

static void CpswAle_initVlanConfig(CSL_AleRegs *cpswAleRegBase,
                                   const CpswAle_InitVlanConfig *initVlanConfig)
{
    uint32_t regVal;

    if (initVlanConfig->aleVlanAwareMode)
    {
        CSL_CPSW_enableAleVlanAware(cpswAleRegBase);
    }
    else
    {
        CSL_CPSW_disableAleVlanAware(cpswAleRegBase);
    }

    regVal = CSL_CPSW_getAleControlReg(cpswAleRegBase);

    CSL_FINS(regVal, ALE_ALE_CONTROL_UVLAN_NO_LEARN, initVlanConfig->unknownVlanNoLearn);
    if (initVlanConfig->autoLearnWithVLAN)
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_LEARN_NO_VLANID, 0);
    }
    else
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_LEARN_NO_VLANID, 1);
    }

    CSL_CPSW_setAleControlReg(cpswAleRegBase, regVal);

    CpswAle_setUnknownVlanConfig(cpswAleRegBase,
                                 initVlanConfig->unknownVlanMemberListMask,
                                 initVlanConfig->unknownForceUntaggedEgressMask,
                                 initVlanConfig->unknownRegMcastFloodMask,
                                 initVlanConfig->unknownUnregMcastFloodMask);
}

static void CpswAle_setIPFilterConfig(CSL_AleRegs *cpswAleRegBase,
                                      const CpswAle_IPPktSecurityConfig *ipCfg)
{
    CSL_CPSW_ALE_CTRL2_IPPKTFLT_CONFIG ipPkt;
    uint32_t i;
    uint8_t ipNxtHdr[CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR];

    ipPkt.ipPktFltEnableDefNoFrag = ipCfg->enableDefaultNoFragment;
    ipPkt.ipPktFltEnableDefNxtHdrLimit = ipCfg->enableDefaultNxtHdrWhitelist;

    CSL_CPSW_setAleCtrl2IPPktFilterConfig(cpswAleRegBase, &ipPkt);

    if (ipCfg->ipNxtHdrWhitelistCount > 0)
    {
        /* If configuring less than 4 IP NXT HDRs, the NXT hdrs should be
         * replicated to ensure valid values present in NXT_HDR_0/1/2/3
         * So first replicate application passed NXT_HDR_0 and then set the
         * app passed NXT hdr values
         */
        for (i = 0; i < CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR; i++)
        {
            ipNxtHdr[i] = ipCfg->ipNxtHdrWhitelist[0];
        }

        for (i = 0; ((i < ipCfg->ipNxtHdrWhitelistCount) &&
                     (i < CPSW_ALE_MAX_WHITELIST_IP_NXT_HDR));
             i++)
        {
            ipNxtHdr[i] = ipCfg->ipNxtHdrWhitelist[i];
        }

        CSL_CPSW_setAleIPNxtHdrWhitelist(cpswAleRegBase, ipNxtHdr[0],
                                         ipNxtHdr[1], ipNxtHdr[2], ipNxtHdr[3]);
    }
}

static void CpswAle_setNetSecConfig(CSL_AleRegs *cpswAleRegBase,
                                    const CpswAle_NetworkSecurityConfig *netsecCfg)
{
    uint32_t regVal;
    CSL_CPSW_ALE_CTRL2_MALFORMEDFRAME_CONFIG badPkt;

    regVal = CSL_CPSW_getAleControlReg(cpswAleRegBase);
    CSL_FINS(regVal, ALE_ALE_CONTROL_ENABLE_VID0_MODE, netsecCfg->enableVid0Mode);
    CSL_FINS(regVal, ALE_ALE_CONTROL_ENABLE_OUI_DENY,  netsecCfg->hostOuiNoMatchDeny);
    CSL_FINS(regVal, ALE_ALE_CONTROL_ENABLE_AUTH_MODE, netsecCfg->macAuthCfg.enableAuthMode);
    CSL_CPSW_setAleControlReg(cpswAleRegBase, regVal);

    badPkt.dropBadLen = netsecCfg->malformedPktCfg.enableBadLenPktDrop;
    badPkt.noDropSrcMcast = netsecCfg->malformedPktCfg.disableSrcMcastDrop;
    CSL_CPSW_setAleCtrl2MalformedFrameConfig(cpswAleRegBase, &badPkt);

    CpswAle_setIPFilterConfig(cpswAleRegBase, &netsecCfg->ipPktCfg);
}

static void CpswAle_setAlePortConfig(CSL_AleRegs *cpswAleRegBase,
                                     uint32_t portNum,
                                     const CpswAle_PortConfig *portConfig,
                                     bool disableMacAuthForPort)
{
    CSL_CPSW_ALE_PORTCONTROL alePortControl;

    CSL_CPSW_getAlePortControlReg(cpswAleRegBase, portNum, &alePortControl);

    alePortControl.macAuthDisable        = disableMacAuthForPort;
    alePortControl.dropDualVlan          = portConfig->vlanCfg.dropDualVlan;
    alePortControl.dropDoubleVlan        = portConfig->vlanCfg.dropDoubleVlan;
    alePortControl.vidIngressCheckEnable = portConfig->vlanCfg.vidIngressCheck;
    alePortControl.dropUntaggedEnable    = portConfig->vlanCfg.dropUntagged;

    alePortControl.noSaUpdateEnable      = portConfig->learningCfg.noSaUpdateEnable;
    alePortControl.noLearnModeEnable     = portConfig->learningCfg.noLearn;

    alePortControl.macOnlyEnable         = portConfig->macModeCfg.macOnlyEnable;
    alePortControl.macOnlyCafEnable      = portConfig->macModeCfg.macOnlyCafEnable;

    /* At time of open , set ALE port state in disabled mode.
     * ALE port must be enabled only after MAC config and
     * PHY link up */
    alePortControl.portState = CSL_ALE_PORTSTATE_DISABLED;

    CSL_CPSW_setAlePortControlReg(cpswAleRegBase, portNum, &alePortControl);
}

static void CpswAle_setAleConfig(CpswAle_Obj *pAle,
                                 const CpswAle_Config *pALEConfig,
                                 uint32_t aleFreqHz,
                                 Cpsw_Type cpswType)
{
    uint32_t i = 0;

    /* Save some config info in the state structure*/
    pAle->cpswType = cpswType;
    pAle->aleFreqHz = aleFreqHz;

    CpswAle_initAleRegs(pAle->cpswAleRegBase);
    CpswAle_setAleModeFlags(pAle->cpswAleRegBase, pALEConfig->modeFlags);
    CpswAle_setAleAging(pAle->cpswAleRegBase,
                        &pALEConfig->agingConfig,
                        pAle->aleFreqHz,
                        &pAle->aleSoftTimerActive,
                        &pAle->aleTickTimeOutCount,
                        &pAle->aleSoftTickCount);

    CpswAle_setUnknownVlanConfig(pAle->cpswAleRegBase,
                                 pALEConfig->vlanConfig.unknownVlanMemberListMask,
                                 pALEConfig->vlanConfig.unknownForceUntaggedEgressMask,
                                 pALEConfig->vlanConfig.unknownRegMcastFloodMask,
                                 pALEConfig->vlanConfig.unknownUnregMcastFloodMask);

    CpswAle_setPolicerControl(pAle->cpswAleRegBase,
                              pAle->aleFreqHz,
                              pALEConfig->policerGlobalConfig.policingEnable,
                              pALEConfig->policerGlobalConfig.yellowDropEnable,
                              pALEConfig->policerGlobalConfig.redDropEnable,
                              pALEConfig->policerGlobalConfig.yellowThresh,
                              pALEConfig->policerGlobalConfig.policerNoMatchMode,
                              &pALEConfig->policerGlobalConfig.noMatchPolicer);

    /* Clear all policers on init */
    CpswAle_clearGlobalPolicerStats(pAle->cpswAleRegBase, TRUE, TRUE, TRUE);

    /* By default disable default thread enable */
    CpswAle_setPolicerDefaultThreadConfig(pAle->cpswAleRegBase, FALSE, 0, FALSE, FALSE);

    /* Configure ALE Port Control Register */
    for (i = 0; i < CpswAle_getNumPorts(pAle->cpswType); i++)
    {
        bool macAuthDisable;

        macAuthDisable =
            ((pALEConfig->nwSecCfg.macAuthCfg.macAuthDisableMask & (1 << i)) == 1);
        CpswAle_setAlePortConfig(pAle->cpswAleRegBase,
                                 i,
                                 &pALEConfig->portCfg[i],
                                 macAuthDisable);
        pAle->pvid[i] = pALEConfig->portCfg[i].pvidCfg;
    }

    CpswAle_setNetSecConfig(pAle->cpswAleRegBase, &pALEConfig->nwSecCfg);
    CpswAle_initVlanConfig(pAle->cpswAleRegBase, &pALEConfig->vlanConfig);
    pAle->pvidEnabled = pALEConfig->vlanConfig.cpswVlanAwareMode;
    pAle->rxFilter = CPSW_ALE_RXFILTER_DIRECT;
}

static void CpswAle_initAleRegs(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t aleCtlVal = 0;

    /* On ALE open init ALE control register to known state */
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_CLEAR_TABLE, 1U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_AGE_OUT_NOW, 1U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_BCAST_MCAST_CTL, 0U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_ENABLE_RATE_LIMIT, 0U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_UPD_STATIC, 0U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_UPD_BW_CTRL, 0U);
    aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_UPD_STATIC, 0U);

    /* Set ALE control register */
    CSL_CPSW_setAleControlReg(cpswAleRegBase, aleCtlVal);
    CSL_CPSW_setAlePrescaleReg(cpswAleRegBase, CPSW_ALE_CFG_BCAST_MCAST_PRESCALE);
}

static void CpswAle_setAleModeFlags(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t aleModeFlags)
{
    uint32_t aleCtlVal;

    aleCtlVal = CSL_CPSW_getAleControlReg(cpswAleRegBase);
    if (NOT_ZERO(aleModeFlags & CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE))
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_ENABLE_ALE, 1U);
    }
    else
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_ENABLE_ALE, 0U);
    }

    if (NOT_ZERO(aleModeFlags & CPSW_ALE_CONFIG_MASK_ALE_BYPASS_ENABLE))
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_ENABLE_BYPASS, 1U);
    }
    else
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_ENABLE_BYPASS, 0U);
    }

    if (NOT_ZERO(aleModeFlags & CPSW_ALE_CONFIG_MASK_UNKNOWN_UNICAST_FLOOD2HOST))
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_EN_HOST_UNI_FLOOD, 1U);
    }
    else
    {
        aleCtlVal |= CSL_FMK(ALE_ALE_CONTROL_EN_HOST_UNI_FLOOD, 0U);
    }

    /* Update ALE control register */
    CSL_CPSW_setAleControlReg(cpswAleRegBase, aleCtlVal);
}

/*
 *  Finds index to the next free entry in ALE table.
 */
static int32_t CpswAle_getFreeEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_SOK;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        if (CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                     i,
                                     CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) == CSL_ALE_ENTRYTYPE_FREE)
        {
            break;
        }
    }

    if (aleTableDepth == i)
    {
        ret = CPSW_EALE_TABLE_FULL;
    }
    else
    {
        *entry = i;
    }

    return ret;
}

/*
 * This function clears the given entry from the ALE table.
 */
static void CpswAle_clearTableEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t entry)
{
    Cpsw_assert(entry <= CpswAle_getMaxAleEntries(cpswAleRegBase));

    CSL_CPSW_clearAleEntry(cpswAleRegBase, entry);

    return;
}

static int32_t CpswAle_setReceiveFilter(CpswAle_Obj *pAle,
                                        CPSW_ALE_RXFILTER_E rxFilter)
{
    int32_t retVal;

    switch (rxFilter)
    {
        case CPSW_ALE_RXFILTER_ALL:
            retVal = CpswAle_setRxFilterAll(pAle->cpswAleRegBase);
            break;

        case CPSW_ALE_RXFILTER_ALLMULTICAST:
            retVal = CpswAle_setRxFilterAllMcast(pAle->cpswAleRegBase);
            break;

        case CPSW_ALE_RXFILTER_MULTICAST:
            retVal = CpswAle_setRxFilterMcast(pAle->cpswAleRegBase);
            break;

        case CPSW_ALE_RXFILTER_BROADCAST:
            retVal = CpswAle_setRxFilterBcast(pAle->cpswAleRegBase);
            break;

        case CPSW_ALE_RXFILTER_DIRECT:
            retVal = CpswAle_setRxFilterDirect(pAle->cpswAleRegBase);
            break;

        case CPSW_ALE_RXFILTER_NOTHING:
            retVal = CpswAle_setRxFilterNone(pAle->cpswAleRegBase);
            break;

        default:
            retVal = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == retVal)
    {
        pAle->rxFilter = rxFilter;
    }

    return retVal;
}

static int32_t CpswAle_getReceiveFilter(const CpswAle_Obj *pAle,
                                        CPSW_ALE_RXFILTER_E *pRxFilter)
{
    int32_t retVal;

    CPSW_ALE_SAFE_ASSIGN(pRxFilter, pAle->rxFilter);

    /* Validate our handle */
    if (IS_NULL(pRxFilter))
    {
        retVal = CPSW_EINVALID_PARAMS;
    }
    else
    {
        retVal = CPSW_SOK;
    }

    return retVal;
}

static int32_t CpswAle_findAddr(CSL_AleRegs *cpswAleRegBase,
                                CpswAle_MacAddrType_t addr,
                                uint32_t vlanId,
                                uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint8_t aleEntryMacAddr[6];
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;
        uint32_t reg0, reg1, reg2;

        /* read from the ALE table */
        CSL_CPSW_getAleTableEntry(cpswAleRegBase, i, &reg0, &reg1, &reg2);

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if ((value == CSL_ALE_ENTRYTYPE_ADDRESS) ||
            (value == CSL_ALE_ENTRYTYPE_VLANADDRESS))
        {
            CSL_CPSW_mapTableWord2MacAddr(reg0, reg1, &aleEntryMacAddr[0], CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            if (memcmp(aleEntryMacAddr, (*addr), CPSW_ALE_MACADDR_NUM_OCTETS) == 0)
            {
                uint32_t vid = CSL_CPSW_extractVid(reg1, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                if ((vlanId > CPSW_VLAN_ID_MAX) || (vid == vlanId) ||
                    (vlanId == 0))
                {
                    break;
                }
            }
        }
    }

    if (i != aleTableDepth)
    {
        *entry = i;
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_findVlan(CSL_AleRegs *cpswAleRegBase,
                                uint32_t vlanId,
                                bool outerVlanFlag,
                                uint32_t *vlanMemberList,
                                uint32_t *unregMcastFloodMask,
                                uint32_t *regMcastFloodMask,
                                uint32_t *forceUntaggedEgress,
                                uint32_t *noLearnMask,
                                bool *vidIngressCheck,
                                bool *limitIpNxtHdr,
                                bool *disallowFragmentation,
                                uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_VLAN_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);
    uint32_t matchVlanMemberList = 0U;
    uint32_t matchUnregMcastFloodMask = 0U;
    uint32_t matchRegMcastFloodMask = 0U;
    uint32_t matchForceUntaggedEgress = 0U;
    uint32_t matchNoLearnMask = 0U;
    bool matchIngressCheck = false;
    bool matchLimitIpNxtHdr = false;
    bool matchDisallowFragmentation = false;

    bool matchFound;

    matchFound = FALSE;
    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_VLAN)
        {
            if (outerVlanFlag)
            {
                if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                    i,
                                                    CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                    CSL_ALE_POLICER_ENTRYTYPE_OVLAN)
                {
                    CSL_CPSW_ALE_OUTER_VLAN_ENTRY oVlanEntry;

                    CSL_CPSW_getAleOutVlanEntry(cpswAleRegBase, i, &oVlanEntry,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (oVlanEntry.vlanId == vlanId)
                    {
                        matchFound                 = TRUE;
                        matchVlanMemberList        = oVlanEntry.vlanMemList;
                        matchForceUntaggedEgress   = oVlanEntry.forceUntaggedEgress;
                        matchNoLearnMask           = oVlanEntry.noLearnMask;
                        matchIngressCheck          = oVlanEntry.ingressCheckFlag;
                        matchLimitIpNxtHdr         = oVlanEntry.limitIPNxtHdr;
                        matchDisallowFragmentation = oVlanEntry.disallowIPFragmentation;
                        CpswAle_getVlanMcastPortMask(cpswAleRegBase, &oVlanEntry,
                                                     &matchRegMcastFloodMask,
                                                     &matchUnregMcastFloodMask);
                        break;
                    }
                }
            }
            else
            {
                if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                    i,
                                                    CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                    CSL_ALE_POLICER_ENTRYTYPE_VLAN)
                {
                    CSL_CPSW_ALE_VLAN_ENTRY vlanEntry;

                    CSL_CPSW_getAleVlanEntry(cpswAleRegBase, i, &vlanEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (vlanEntry.vlanId == vlanId)
                    {
                        matchFound               = TRUE;
                        matchVlanMemberList      = vlanEntry.vlanMemList;
                        matchForceUntaggedEgress = vlanEntry.forceUntaggedEgress;
                        matchNoLearnMask         = vlanEntry.noLearnMask;
                        matchIngressCheck        = vlanEntry.ingressCheckFlag;
                        if (!CpswAle_isIPHdrWhiteListSupported(cpswAleRegBase))
                        {
                            matchLimitIpNxtHdr = 0;
                            matchDisallowFragmentation = 0;
                        }
                        else
                        {
                            matchLimitIpNxtHdr = vlanEntry.limitIPNxtHdr;
                            matchDisallowFragmentation = vlanEntry.disallowIPFragmentation;
                        }

                        CpswAle_getVlanMcastPortMask(cpswAleRegBase,
                                                     &vlanEntry,
                                                     &matchRegMcastFloodMask,
                                                     &matchUnregMcastFloodMask);

                        break;
                    }
                }
            }
        }
    } /* end of for i loop */

    if (matchFound)
    {
        CPSW_ALE_SAFE_ASSIGN(vlanMemberList, matchVlanMemberList);
        CPSW_ALE_SAFE_ASSIGN(unregMcastFloodMask, matchUnregMcastFloodMask);
        CPSW_ALE_SAFE_ASSIGN(regMcastFloodMask, matchRegMcastFloodMask);
        CPSW_ALE_SAFE_ASSIGN(forceUntaggedEgress, matchForceUntaggedEgress);
        CPSW_ALE_SAFE_ASSIGN(noLearnMask, matchNoLearnMask);
        CPSW_ALE_SAFE_ASSIGN(vidIngressCheck, matchIngressCheck);
        CPSW_ALE_SAFE_ASSIGN(limitIpNxtHdr, matchLimitIpNxtHdr);
        CPSW_ALE_SAFE_ASSIGN(disallowFragmentation, matchDisallowFragmentation);
        CPSW_ALE_SAFE_ASSIGN(entry, i);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_addUcastAddr(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t portNum,
                                    CpswAle_MacAddrType_t addr,
                                    uint32_t vlanId,
                                    uint32_t blocked,
                                    uint32_t secure,
                                    uint32_t ageable,
                                    uint32_t trunk,
                                    uint32_t *entryIndex)
{
    uint32_t freeEntry, vlanAware;
    int32_t ret;
    uint32_t addrEntry = 0;

    /* Check the current setting for vlanAware */
    vlanAware = CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase);

    /* ignore input vlanId if not in VLAN aware mode */
    if (vlanAware == 0)
    {
        vlanId = 0;
    }

    /* Check for space in the ALE table */
    ret = CpswAle_findAddr(cpswAleRegBase, addr, vlanId, &addrEntry);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
    {
        ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
    }
    else
    {
        /* update the existing entry itself */
        freeEntry = (uint32_t)addrEntry;
    }

    if (ret != CPSW_EALE_TABLE_FULL)
    {
        if ((vlanId != 0) && (vlanId <= CPSW_VLAN_ID_MAX))
        {
            CSL_CPSW_ALE_VLANUNICASTADDR_ENTRY vlanUcastAddrEntry;

            vlanUcastAddrEntry.blockEnable  = blocked;
            CPSW_UTILS_ARRAY_COPY(vlanUcastAddrEntry.macAddress, (*addr));
            vlanUcastAddrEntry.portNumber   = portNum;
            vlanUcastAddrEntry.secureEnable = secure;
            vlanUcastAddrEntry.vlanId       = vlanId;
            vlanUcastAddrEntry.ageable      = ageable;
            vlanUcastAddrEntry.trunkFlag    = trunk;
            CSL_CPSW_setAleVlanUnicastAddrEntry(cpswAleRegBase,
                                                freeEntry,
                                                &vlanUcastAddrEntry,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }
        else
        {
            CSL_CPSW_ALE_UNICASTADDR_ENTRY ucastAddrEntry;

            /* Add new entry to ALE table */
            ucastAddrEntry.blockEnable  = blocked;
            CPSW_UTILS_ARRAY_COPY(ucastAddrEntry.macAddress, (*addr));
            ucastAddrEntry.portNumber   = portNum;
            ucastAddrEntry.secureEnable = secure;
            ucastAddrEntry.ageable      = ageable;
            ucastAddrEntry.trunkFlag    = trunk;

            CSL_CPSW_setAleUnicastAddrEntry(cpswAleRegBase,
                                            freeEntry,
                                            &ucastAddrEntry,
                                            CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }

        CPSW_ALE_SAFE_ASSIGN(entryIndex, freeEntry);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_addOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                  CpswAle_OuiAddrType_t ouiAddr,
                                  uint32_t *pAleIndex,
                                  bool *pAleEntryAlloced)
{
    uint32_t freeEntry;
    int32_t ret = CPSW_SOK;
    uint32_t addrEntry;

    /* Check for space in the ALE table */
    ret = CpswAle_findOuiAddr(cpswAleRegBase, ouiAddr, &addrEntry);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
    {
        ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
        if (ret != CPSW_EALE_TABLE_FULL)
        {
            CPSW_ALE_SAFE_ASSIGN(pAleEntryAlloced, TRUE);
        }
    }
    else
    {
        freeEntry = addrEntry;
        CPSW_ALE_SAFE_ASSIGN(pAleEntryAlloced, FALSE);
    }

    if (ret != CPSW_EALE_TABLE_FULL)
    {
        CSL_CPSW_ALE_OUIADDR_ENTRY ouiAddrEntry;

        CPSW_UTILS_ARRAY_COPY(ouiAddrEntry.ouiAddress, (*ouiAddr));
        CSL_CPSW_setAleOUIAddrEntry(cpswAleRegBase,
                                    freeEntry,
                                    &ouiAddrEntry,
                                    CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        CPSW_ALE_SAFE_ASSIGN(pAleIndex, freeEntry);
        ret = CPSW_SOK;
    }

    return ret;
}

static void CpswAle_mapMcastMaskBits2RegCfg(uint8_t *addr,
                                            uint32_t numIgnBits,
                                            uint8_t *maskedAddr,
                                            uint32_t *ignMBits,
                                            CSL_CPSW_ALETABLE_TYPE aleType)
{
    if (numIgnBits)
    {
        uint32_t ignClrMsk, ignSetMsk;
        uint32_t word0, word1;

        ignClrMsk = CPSW_UTILS_MK_ONES((numIgnBits - 1), 0);
        ignSetMsk = ignClrMsk >> 1;
        CSL_CPSW_mapMacAddr2TableWord(&word0, &word1, addr, aleType);
        word0 &= ~ignClrMsk;
        word0 |= ignSetMsk;
        CSL_CPSW_mapTableWord2MacAddr(word0, word1, maskedAddr, aleType);
        *ignMBits = 1;
    }
    else
    {
        /* Cannot use CPSW_UTILS_ARRAY_COPY here as maskedAddr is a pointer
         * and not an array
         */
        memcpy(maskedAddr, addr, CPSW_ALE_MACADDR_NUM_OCTETS);
        *ignMBits = 0;
    }
}

static int32_t CpswAle_validateAddMcastAddr(CpswAle_MacAddrType_t addr,
                                            uint32_t numIgnBits)
{
    int32_t retVal = CPSW_SOK;

    if (numIgnBits > CPSW_ALE_MCAST_IGN_BITS_MAX)
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == retVal)
    {
        if (CpswAle_isMcastAddr(addr) != TRUE)
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    if (retVal != CPSW_SOK)
    {
        CPSWTRACE_ERROR("CPSWALE:Bad MCast Addr:%x:%x:%x:%x:%x:%x/%d",
                        addr[0], addr[1], addr[2], addr[3], addr[4], addr[5],
                        numIgnBits);
    }

    return retVal;
}

static int32_t CpswAle_addMcastAddr(CSL_AleRegs *cpswAleRegBase,
                                    const CpswAle_MacAddrType_t addr,
                                    uint32_t vlanId,
                                    uint32_t portMask,
                                    uint32_t super,
                                    uint32_t fwdState,
                                    uint32_t numIgnBits,
                                    uint32_t *entryIndex)
{
    uint32_t freeEntry, vlanAware;
    int32_t ret = CPSW_SOK;
    uint32_t addrEntry = 0;
    uint8_t mcastMaskedAddr[6];
    uint32_t ignMBitsFlag;

    ret = CpswAle_validateAddMcastAddr(addr, numIgnBits);

    if (CPSW_SOK == ret)
    {
        /* Check the current setting for vlanAware */

        vlanAware = CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase);

        /* ignore input vlanId if not in VLAN aware mode */
        if (vlanAware == 0)
        {
            vlanId = 0;
        }

        CpswAle_mapMcastMaskBits2RegCfg((uint8_t *)addr,
                                        numIgnBits,
                                        mcastMaskedAddr,
                                        &ignMBitsFlag,
                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

        /* Check for space in the ALE table */
        ret = CpswAle_findAddr(cpswAleRegBase,
                               (CpswAle_MacAddrType_t)(&mcastMaskedAddr),
                               vlanId,
                               &addrEntry);
        if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
        {
            ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
        }
        else
        {
            /* update the existing entry */
            freeEntry = addrEntry;
        }

        if (ret == CPSW_EALE_TABLE_FULL)
        {
            ret = CPSW_EALE_TABLE_FULL;
        }
        else
        {
            if ((uint32_t)0U != vlanId)
            {
                CSL_CPSW_ALE_VLANMCASTADDR_ENTRY vlanMcastAddrCfg;

                vlanMcastAddrCfg.mcastFwdState = fwdState;
                vlanMcastAddrCfg.portMask = portMask;
                vlanMcastAddrCfg.superEnable = super;
                vlanMcastAddrCfg.vlanId = vlanId;
                vlanMcastAddrCfg.ignMBits = ignMBitsFlag;
                CPSW_UTILS_ARRAY_COPY(vlanMcastAddrCfg.macAddress, mcastMaskedAddr);

                CSL_CPSW_setAleVlanMcastAddrEntry(cpswAleRegBase,
                                                  freeEntry,
                                                  &vlanMcastAddrCfg,
                                                  CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            }
            else
            {
                CSL_CPSW_ALE_MCASTADDR_ENTRY mcastAddrCfg;

                mcastAddrCfg.mcastFwdState = fwdState;
                mcastAddrCfg.portMask = portMask;
                mcastAddrCfg.superEnable = super;
                mcastAddrCfg.ignMBits = ignMBitsFlag;
                CPSW_UTILS_ARRAY_COPY(mcastAddrCfg.macAddress, mcastMaskedAddr);
                /* Add new entry to the ALE table */
                CSL_CPSW_setAleMcastAddrEntry(cpswAleRegBase,
                                              freeEntry,
                                              &mcastAddrCfg,
                                              CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            }

            CPSW_ALE_SAFE_ASSIGN(entryIndex, freeEntry);
            ret = CPSW_SOK;
        }
    }

    return ret;
}

static int32_t CpswAle_addVlan(CSL_AleRegs *cpswAleRegBase,
                               uint32_t vlanId,
                               bool outerVlanFlag,
                               uint32_t vlanMemberList,
                               uint32_t unregMcastFloodMask,
                               uint32_t regMcastFloodMask,
                               uint32_t forceUntaggedEgress,
                               uint32_t noLearnMask,
                               bool vidIngressCheck,
                               bool limitIPNxtHdr,
                               bool disallowIPFragmentation,
                               uint32_t *aleEntryIndex)
{
    uint32_t vlanAware = 0, freeEntry = 0;
    int32_t ret = CPSW_SOK;
    uint32_t vlanEntry = 0;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    /* Check the current setting for vlanAware */
    vlanAware = CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase);

    if (vlanAware == 0)
    {
        ret = CPSW_EALE_VLAN_UNAWARE_MODE;
    }
    else
    {
        vlanId              &= (uint32_t)CPSW_ALE_VLAN_ID_MASK;
        vlanMemberList      &= (uint32_t)CPSW_ALE_ALL_PORTS_MASK;
        unregMcastFloodMask &= (uint32_t)CPSW_ALE_ALL_PORTS_MASK;
        regMcastFloodMask   &= (uint32_t)CPSW_ALE_ALL_PORTS_MASK;
        forceUntaggedEgress &= (uint32_t)CPSW_ALE_ALL_PORTS_MASK;
        noLearnMask         &= (uint32_t)CPSW_ALE_ALL_PORTS_MASK;

        ret = CpswAle_findVlan(cpswAleRegBase,
                               vlanId,
                               outerVlanFlag,
                               NULL,
                               NULL,
                               NULL,
                               NULL,
                               NULL,
                               NULL,
                               NULL,
                               NULL,
                               &vlanEntry);
        if ((CPSW_EALE_VLAN_NOT_FOUND != ret) && (vlanEntry < aleTableDepth))
        {
            /* update existing entry */
            freeEntry = vlanEntry;
        }
        else
        {
            /* Update VLAN Info */
            ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
        }

        if (ret != CPSW_EALE_TABLE_FULL)
        {
            if (CPSW_SOK == ret)
            {
                if (!outerVlanFlag)
                {
                    CSL_CPSW_ALE_VLAN_ENTRY vlanCfg;

                    vlanCfg.forceUntaggedEgress     = forceUntaggedEgress;
                    vlanCfg.vlanMemList             = vlanMemberList;
                    vlanCfg.vlanId                  = vlanId;
                    vlanCfg.noLearnMask             = noLearnMask;
                    vlanCfg.ingressCheckFlag        = vidIngressCheck;
                    vlanCfg.limitIPNxtHdr           = limitIPNxtHdr;
                    vlanCfg.disallowIPFragmentation = disallowIPFragmentation;

                    ret = CpswAle_setVlanMcastPortMask(cpswAleRegBase,
                                                       &vlanCfg,
                                                       freeEntry,
                                                       FALSE,
                                                       regMcastFloodMask,
                                                       unregMcastFloodMask);
                }
                else
                {
                    CSL_CPSW_ALE_OUTER_VLAN_ENTRY oVlanCfg;

                    oVlanCfg.forceUntaggedEgress     = forceUntaggedEgress;
                    oVlanCfg.vlanMemList             = vlanMemberList;
                    oVlanCfg.vlanId                  = vlanId;
                    oVlanCfg.noLearnMask             = noLearnMask;
                    oVlanCfg.ingressCheckFlag        = vidIngressCheck;
                    oVlanCfg.limitIPNxtHdr           = limitIPNxtHdr;
                    oVlanCfg.disallowIPFragmentation = disallowIPFragmentation;

                    ret = CpswAle_setVlanMcastPortMask(cpswAleRegBase,
                                                       &oVlanCfg,
                                                       freeEntry,
                                                       TRUE,
                                                       regMcastFloodMask,
                                                       unregMcastFloodMask);
                }

                CPSW_ALE_SAFE_ASSIGN(aleEntryIndex, freeEntry);
            }
        }
    }

    return ret;
}

static int32_t CpswAle_delVlan(CSL_AleRegs *cpswAleRegBase,
                               uint32_t vlanId,
                               bool outerVlanFlag)
{
    uint32_t aleIndex;
    int32_t ret = CPSW_EALE_VLAN_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    ret = CpswAle_findVlan(cpswAleRegBase,
                           vlanId,
                           outerVlanFlag,
                           NULL,
                           NULL,
                           NULL,
                           NULL,
                           NULL,
                           NULL,
                           NULL,
                           NULL,
                           &aleIndex);
    if ((CPSW_EALE_VLAN_NOT_FOUND != ret) && (aleIndex < aleTableDepth))
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_delAddr(CSL_AleRegs *cpswAleRegBase,
                               CpswAle_MacAddrType_t addr,
                               uint32_t vlanId)
{
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint32_t aleIndex;

    ret = CpswAle_findAddr(cpswAleRegBase, addr, vlanId, &aleIndex);
    if (CPSW_SOK == ret)
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
    }

    return ret;
}

static bool CpswAle_isIpv6HighEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t aleIndex,
                                    uint32_t numPending)
{
    bool retVal = FALSE;
    uint32_t highOffset;
    uint32_t checkAleIndex;

    if (numPending > 0)
    {
        highOffset = CSL_CPSW_getAleIPv6HighEntryOffset(cpswAleRegBase);

        checkAleIndex = aleIndex - highOffset;
        if ((CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                      checkAleIndex,
                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
             CSL_ALE_ENTRYTYPE_POLICER)
            &&
            (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                             checkAleIndex,
                                             CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
             CSL_ALE_POLICER_ENTRYTYPE_IPV6))
        {
            retVal = TRUE;
        }
    }

    return retVal;
}

static void CpswAle_dumpTable(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t numPending = 0;
    uint32_t i, freeEntries = 0;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    CpswUtils_printf("\n");

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t info0, info1, info2;
        bool skipPrintEntryRaw = FALSE;

        CSL_CPSW_getAleTableEntry(cpswAleRegBase, i, &info0, &info1, &info2);

        if ((info0 == 0xFFFFFFFEU) && ((info1 & 0x0FFFFU) == 0x0FFFFU))
        {
        }
        else
        {
            skipPrintEntryRaw = FALSE;
            switch (CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)))
            {
                case CSL_ALE_ENTRYTYPE_FREE:
                {
                    freeEntries++;
                    skipPrintEntryRaw = TRUE;
                    break;
                }

                case CSL_ALE_ENTRYTYPE_ADDRESS: /* MAC address */
                {
                    CSL_CPSW_ALE_ADDRTYPE addrType =
                        CSL_CPSW_getALEAddressType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

                    if (CSL_ALE_ADDRTYPE_MCAST == addrType)
                    {
                        CSL_CPSW_ALE_MCASTADDR_ENTRY mcastAddrEntry;

                        CSL_CPSW_getAleMcastAddrEntry(cpswAleRegBase,
                                                      i,
                                                      &mcastAddrEntry,
                                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                        CpswUtils_printf("  %4d: Address: %02x%02x%02x%02x%02x%02x, Member:%03x Su=%x FWDSTLVL=%d IGNMBITS=%d",
                                         i,
                                         mcastAddrEntry.macAddress[0],
                                         mcastAddrEntry.macAddress[1],
                                         mcastAddrEntry.macAddress[2],
                                         mcastAddrEntry.macAddress[3],
                                         mcastAddrEntry.macAddress[4],
                                         mcastAddrEntry.macAddress[5],
                                         mcastAddrEntry.portMask,
                                         mcastAddrEntry.superEnable,
                                         mcastAddrEntry.mcastFwdState,
                                         mcastAddrEntry.ignMBits);
                    }
                    else
                    {
                        CSL_CPSW_ALE_UNICASTADDR_ENTRY ucastAddrEntry;

                        CSL_CPSW_getAleUnicastAddrEntry(cpswAleRegBase,
                                                        i,
                                                        &ucastAddrEntry,
                                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

                        CpswUtils_printf("  %4d: Address: %02x%02x%02x%02x%02x%02x, Port: %03x Se=%x Bl=%d TOUCH=%d AGE=%d TRUNK=%d",
                                         i,
                                         ucastAddrEntry.macAddress[0],
                                         ucastAddrEntry.macAddress[1],
                                         ucastAddrEntry.macAddress[2],
                                         ucastAddrEntry.macAddress[3],
                                         ucastAddrEntry.macAddress[4],
                                         ucastAddrEntry.macAddress[5],
                                         ucastAddrEntry.portNumber,
                                         ucastAddrEntry.secureEnable,
                                         ucastAddrEntry.blockEnable,
                                         ucastAddrEntry.touched,
                                         ucastAddrEntry.ageable,
                                         ucastAddrEntry.trunkFlag);
                    }

                    break;
                }

                case CSL_ALE_ENTRYTYPE_POLICER: /*VLAN*/
                {
                    CSL_CPSW_ALE_POLICER_ENTRYTYPE policerType;

                    policerType = CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                                  i,
                                                                  CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    switch (policerType)
                    {
                        case CSL_ALE_POLICER_ENTRYTYPE_OVLAN:
                        {
                            CSL_CPSW_ALE_OUTER_VLAN_ENTRY ovlan;
                            uint32_t regMcastFloodMask;
                            uint32_t unRegMcastFloodMask;

                            CSL_CPSW_getAleOutVlanEntry(cpswAleRegBase,
                                                        i,
                                                        &ovlan,
                                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                            CpswAle_getVlanMcastPortMask(cpswAleRegBase,
                                                         &ovlan,
                                                         &regMcastFloodMask,
                                                         &unRegMcastFloodMask);
                            CpswUtils_printf("  %4d: Outer Vlanid: %04x, UTagged: %x, Mult: %x, UMult: %x, Member: %x",
                                             i,
                                             ovlan.vlanId,
                                             ovlan.forceUntaggedEgress,
                                             regMcastFloodMask,
                                             unRegMcastFloodMask,
                                             ovlan.vlanMemList);
                            break;
                        }

                        case CSL_ALE_POLICER_ENTRYTYPE_VLAN:
                        {
                            CSL_CPSW_ALE_VLAN_ENTRY vlanEntry;
                            uint32_t regMcastFloodMask;
                            uint32_t unRegMcastFloodMask;

                            CSL_CPSW_getAleVlanEntry(cpswAleRegBase,
                                                     i,
                                                     &vlanEntry,
                                                     CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                            CpswAle_getVlanMcastPortMask(cpswAleRegBase,
                                                         &vlanEntry,
                                                         &regMcastFloodMask,
                                                         &unRegMcastFloodMask);
                            CpswUtils_printf("  %4d: Vlanid: %04x, UTagged: %x, Mult: %x, UMult: %x, Member: %x",
                                             i,
                                             vlanEntry.vlanId,
                                             vlanEntry.forceUntaggedEgress,
                                             regMcastFloodMask,
                                             unRegMcastFloodMask,
                                             vlanEntry.vlanMemList);
                            break;
                        }

                        case CSL_ALE_POLICER_ENTRYTYPE_ETHERTYPE:
                        {
                            CSL_CPSW_ALE_ETHERTYPE_ENTRY etherType;

                            CSL_CPSW_getAleEthertypeEntry(cpswAleRegBase,
                                                          i,
                                                          &etherType,
                                                          CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                            CpswUtils_printf("  %4d: EtherType: %04x",
                                             i, etherType.ethertype);
                            break;
                        }

                        case CSL_ALE_POLICER_ENTRYTYPE_IPV4:
                        {
                            CSL_CPSW_ALE_IPv4_ENTRY ipv4Type;

                            CSL_CPSW_getAleIPv4Entry(cpswAleRegBase,
                                                     i,
                                                     &ipv4Type,
                                                     CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                            CpswUtils_printf("  %4d: IPv4: Address: %02d:%02d:%02d:%02d, IGNBITS:%d",
                                             i,
                                             ipv4Type.address[0],
                                             ipv4Type.address[1],
                                             ipv4Type.address[2],
                                             ipv4Type.address[3],
                                             ipv4Type.numLSBIgnore);
                            break;
                        }

                        case CSL_ALE_POLICER_ENTRYTYPE_IPV6:
                        {
                            CSL_CPSW_ALE_IPv6_ENTRY ipv6Type;
                            bool isHighEntry;

                            isHighEntry = CpswAle_isIpv6HighEntry(cpswAleRegBase, i, numPending);
                            if (isHighEntry)
                            {
                                skipPrintEntryRaw = TRUE;
                                numPending--;
                            }
                            else
                            {
                                numPending++;
                                CSL_CPSW_getAleIPv6Entry(cpswAleRegBase,
                                                         i,
                                                         &ipv6Type,
                                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                                CpswUtils_printf("  %4d: IPv6:     ,"
                                                 " Address: %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:"
                                                 "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x,"
                                                 " IGNBITS:%d",
                                                 i,
                                                 ipv6Type.address[0],
                                                 ipv6Type.address[1],
                                                 ipv6Type.address[2],
                                                 ipv6Type.address[3],
                                                 ipv6Type.address[4],
                                                 ipv6Type.address[5],
                                                 ipv6Type.address[6],
                                                 ipv6Type.address[7],
                                                 ipv6Type.address[8],
                                                 ipv6Type.address[9],
                                                 ipv6Type.address[10],
                                                 ipv6Type.address[11],
                                                 ipv6Type.address[12],
                                                 ipv6Type.address[13],
                                                 ipv6Type.address[14],
                                                 ipv6Type.address[15],
                                                 ipv6Type.numLSBIgnore);
                            }

                            break;
                        }
                    }

                    break;
                }

                case CSL_ALE_ENTRYTYPE_VLANADDRESS: /* VLAN MAC address */
                {
                    CSL_CPSW_ALE_ADDRTYPE addrType =
                        CSL_CPSW_getALEAddressType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

                    if (CSL_ALE_ADDRTYPE_MCAST == addrType)
                    {
                        CSL_CPSW_ALE_VLANMCASTADDR_ENTRY mcastAddrEntry;

                        CSL_CPSW_getAleVlanMcastAddrEntry(cpswAleRegBase,
                                                          i,
                                                          &mcastAddrEntry,
                                                          CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                        CpswUtils_printf("  %4d: VlanId:%03x, Address: %02x%02x%02x%02x%02x%02x, Member:%03x Su=%x FWDSTLVL=%d IGNMBITS=%d",
                                         i,
                                         mcastAddrEntry.vlanId,
                                         mcastAddrEntry.macAddress[0],
                                         mcastAddrEntry.macAddress[1],
                                         mcastAddrEntry.macAddress[2],
                                         mcastAddrEntry.macAddress[3],
                                         mcastAddrEntry.macAddress[4],
                                         mcastAddrEntry.macAddress[5],
                                         mcastAddrEntry.portMask,
                                         mcastAddrEntry.superEnable,
                                         mcastAddrEntry.mcastFwdState,
                                         mcastAddrEntry.ignMBits);
                    }
                    else
                    {
                        CSL_CPSW_ALE_VLANUNICASTADDR_ENTRY ucastAddrEntry;

                        CSL_CPSW_getAleVlanUnicastAddrEntry(cpswAleRegBase,
                                                            i,
                                                            &ucastAddrEntry,
                                                            CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                        CpswUtils_printf("  %4d: VlanId:%03x, Address: %02x%02x%02x%02x%02x%02x, Port: %03x Se=%x Bl=%d TOUCH=%d AGE=%d TRUNK=%d",
                                         i,
                                         ucastAddrEntry.vlanId,
                                         ucastAddrEntry.macAddress[0],
                                         ucastAddrEntry.macAddress[1],
                                         ucastAddrEntry.macAddress[2],
                                         ucastAddrEntry.macAddress[3],
                                         ucastAddrEntry.macAddress[4],
                                         ucastAddrEntry.macAddress[5],
                                         ucastAddrEntry.portNumber,
                                         ucastAddrEntry.secureEnable,
                                         ucastAddrEntry.blockEnable,
                                         ucastAddrEntry.touched,
                                         ucastAddrEntry.ageable,
                                         ucastAddrEntry.trunkFlag);
                    }

                    break;
                }

                default:
                {
                    break;
                }
            } /* end of switch construct */

            /*Raw Dump,followed by parsed output*/
            if (!skipPrintEntryRaw)
            {
                /* print Info for entried that are not free */
                CpswUtils_printf(" RAW:[%01x %02x%02x%02x",
                                 info2 & 0xFU,
                                 (info1 >> 24U) & 0xFFU,
                                 (info1 >> 16U) & 0xFFU,
                                 (info1 >> 8U) & 0xFFU);
                CpswUtils_printf("%02x %02x%02x%02x%02x] \n",
                                 info1 & 0xFFU,
                                 (info0 >> 24U) & 0xFFU,
                                 (info0 >> 16U) & 0xFFU,
                                 (info0 >> 8U) & 0xFFU, info0 & 0xFFU);
            }
        } /* end of else conditional block */
    }     /* end of for loop over ALE entries */

    CpswUtils_printf("\n\r");
    CpswUtils_printf("%d Free Entries \r\n", freeEntries);
}

static bool  CpswAle_isMcastAddr(CpswAle_MacAddrType_t addr)
{
    if (((*addr)[0] & 0x1) == 1)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

static void CpswAle_setAleAging(CSL_AleRegs *cpswAleRegBase,
                                const CpswAle_AgingConfig *agingConfig,
                                uint32_t aleFreq,
                                bool *aleSoftTimerActive,
                                uint32_t *aleTickTimeOutCount,
                                uint32_t *aleSoftTickCount)
{
    if (agingConfig->enableAutoAging)
    {
        uint32_t agingPeriod = CpswAle_mapAleAgingIntervalMsToAgingTimer(aleFreq,
                                                                         agingConfig->agingPeriodInMs);

        CSL_CPSW_setAleAgingTimerReg(cpswAleRegBase,
                                     CSL_ALE_AGT_PRESACLE_1M,
                                     agingPeriod);
        *aleSoftTimerActive = FALSE;
    }
    else
    {
        *aleSoftTimerActive = TRUE;
        *aleTickTimeOutCount = agingConfig->agingPeriodInMs;
        *aleSoftTickCount = 0;
    }
}

static uint32_t CpswAle_getMaxPolicers(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t numPolicers;

    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicers);

    return numPolicers;
}

static uint32_t CpswAle_getMaxAleEntries(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t numAleEntries;

    CSL_CPSW_getAleStatusNumAleEntries(cpswAleRegBase, &numAleEntries);

    return numAleEntries;
}

static int32_t  CpswAle_validatePolicerIndex(CSL_AleRegs *cpswAleRegBase,
                                             uint32_t policerEntryIdx)
{
    int32_t retVal = CPSW_SOK;

    if (policerEntryIdx >= CpswAle_getMaxPolicers(cpswAleRegBase))
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    return retVal;
}

static int32_t CpswAle_validateAleEntryIndex(CSL_AleRegs *cpswAleRegBase,
                                             uint32_t aleEntryIdx)
{
    int32_t retVal = CPSW_SOK;

    if (aleEntryIdx >= CpswAle_getMaxAleEntries(cpswAleRegBase))
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    return retVal;
}

static int32_t CpswAle_validateSetPolicerCommon(CSL_AleRegs *cpswAleRegBase,
                                                uint32_t policerEntryIdx,
                                                uint32_t aleEntryIdx)
{
    int32_t retVal = CPSW_SOK;

    retVal = CpswAle_validatePolicerIndex(cpswAleRegBase, policerEntryIdx);

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateAleEntryIndex(cpswAleRegBase, aleEntryIdx);
    }

    return retVal;
}

static int32_t CpswAle_validateOuiPolicer(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t policerEntryIdx,
                                          uint32_t aleEntryIdx,
                                          CpswAle_OuiAddrType_t ouiAddr)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase,
                                              policerEntryIdx,
                                              aleEntryIdx);

    if (CPSW_SOK == retVal)
    {
        uint32_t matchEntry;

        retVal = CpswAle_findOuiAddr(cpswAleRegBase, ouiAddr, &matchEntry);
        if (CPSW_SOK == retVal)
        {
            if (matchEntry != aleEntryIdx)
            {
                retVal = CPSW_EINVALID_PARAMS;
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_findOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_OuiAddrType_t addr,
                                   uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        CSL_CPSW_ALE_ENTRYTYPE value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                         i,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

        /* We only need to check the entry type and address fields */
        if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
        {
            if (CSL_CPSW_getALEAddressType(cpswAleRegBase,
                                           i,
                                           CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) == CSL_ALE_ADDRTYPE_OUI)
            {
                CSL_CPSW_ALE_OUIADDR_ENTRY ouiEntry;

                CSL_CPSW_getAleOUIAddrEntry(cpswAleRegBase, i, &ouiEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));

                if (memcmp(ouiEntry.ouiAddress, (*addr), CPSW_UTILS_ARRAYSIZE(ouiEntry.ouiAddress)) == 0)
                {
                    break;
                }
            }
        }
    }

    if (i != aleTableDepth)
    {
        *entry = i;
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_setOuiPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t policerEntryIdx,
                                          CpswAle_OuiAddrType_t ouiAddr,
                                          uint32_t peakRateInBps,
                                          uint32_t commitRateInBps,
                                          uint32_t aleFreq,
                                          uint32_t *aleEntryIndex,
                                          bool *pAleEntryAlloced)
{
    int32_t retVal;

    retVal = CpswAle_addOuiAddr(cpswAleRegBase, ouiAddr, aleEntryIndex, pAleEntryAlloced);

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateOuiPolicer(cpswAleRegBase,
                                            policerEntryIdx,
                                            *aleEntryIndex,
                                            ouiAddr);
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_OUI_VALID;
        polCfg.ouiIdx = *aleEntryIndex;
        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase, policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validateIPv4Policer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t aleEntryIdx,
                                           CpswAle_Ipv4AddrType_t ipv4Addr,
                                           uint32_t numLSBIgnore)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase,
                                              policerEntryIdx,
                                              aleEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (numLSBIgnore > CSL_CPSW_getIpv4IgnBitsMax(CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)))
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t matchEntry;

        retVal = CpswAle_findIPv4Addr(cpswAleRegBase, ipv4Addr, numLSBIgnore, &matchEntry);
        if (CPSW_SOK == retVal)
        {
            if (matchEntry != aleEntryIdx)
            {
                retVal = CPSW_EINVALID_PARAMS;
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_setIPPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t policerEntryIdx,
                                         uint32_t aleEntryIdx,
                                         CPSW_ALE_ADDR_TYPE addrType)
{
    CSL_CPSW_ALE_POLICER_ENTRY polCfg;

    CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);

    if (addrType == CPSW_ALE_ADDR_TYPE_SRC)
    {
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_SRC_IP_VALID;
        polCfg.srcIpIdx = aleEntryIdx;
    }
    else
    {
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_DST_IP_VALID;
        polCfg.dstIpIdx = aleEntryIdx;
    }

    CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    return CPSW_SOK;
}

static uint32_t CpswAle_mapIPv4AddrOctet2RegWord(CpswAle_Ipv4AddrType_t ipv4Addr)
{
    uint32_t regWord;

    regWord = ((*ipv4Addr)[0] << 24) |
              ((*ipv4Addr)[1] << 16) |
              ((*ipv4Addr)[2] << 8) |
              ((*ipv4Addr)[3] << 0);
    return regWord;
}

static uint32_t CpswAle_maskIPv4AddrRegWord(const uint32_t ipv4AddrRegWord,
                                            const uint32_t numLSBIgnore)
{
    uint32_t maskedAddr;

    maskedAddr = ipv4AddrRegWord & (uint32_t) ~((1U << numLSBIgnore) - 1U);
    return maskedAddr;
}

static int32_t CpswAle_findIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                    CpswAle_Ipv4AddrType_t ipv4Addr,
                                    uint32_t numLSBIgnore,
                                    uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);
    uint32_t ipv4MatchAddr = CpswAle_mapIPv4AddrOctet2RegWord(ipv4Addr);
    uint32_t ipv4MatchAddrMasked = CpswAle_maskIPv4AddrRegWord(ipv4MatchAddr, numLSBIgnore);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                         i,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_POLICER)
        {
            if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                i,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                CSL_ALE_POLICER_ENTRYTYPE_IPV4)
            {
                CSL_CPSW_ALE_IPv4_ENTRY ipv4Entry;
                uint32_t entryIpv4AddrReg;
                uint32_t entryIpv4AddrMaskedReg;

                CSL_CPSW_getAleIPv4Entry(cpswAleRegBase,
                                         i,
                                         &ipv4Entry,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                {
                    const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                        CPSW_ALE_IPV4_ADDR_INIT_LIST(ipv4Entry.address);
                    entryIpv4AddrReg = CpswAle_mapIPv4AddrOctet2RegWord(&ipv4Addr);
                }
                entryIpv4AddrMaskedReg = CpswAle_maskIPv4AddrRegWord(entryIpv4AddrReg,
                                                                     numLSBIgnore);
                if (entryIpv4AddrMaskedReg == ipv4MatchAddrMasked)
                {
                    break;
                }
            }
        }
    }

    if (i != aleTableDepth)
    {
        *entry = i;
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_addIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv4AddrType_t ipv4Addr,
                                   const uint32_t numLSBIgnore,
                                   uint32_t *pAleIndex,
                                   bool *pEntryAlloced)
{
    uint32_t freeEntry;
    int32_t ret = CPSW_SOK;
    uint32_t addrEntry;

    /* Check for space in the ALE table */
    ret = CpswAle_findIPv4Addr(cpswAleRegBase, ipv4Addr, numLSBIgnore,
                               &addrEntry);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
    {
        ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
        if (ret != CPSW_EALE_TABLE_FULL)
        {
            CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, TRUE);
        }
    }
    else
    {
        freeEntry = addrEntry;
        CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, FALSE);
    }

    if (ret != CPSW_EALE_TABLE_FULL)
    {
        CSL_CPSW_ALE_IPv4_ENTRY ipv4Cfg;

        memcpy(ipv4Cfg.address, (*ipv4Addr), sizeof(ipv4Cfg.address));
        ipv4Cfg.numLSBIgnore = numLSBIgnore;

        CSL_CPSW_setAleIPv4Entry(cpswAleRegBase, freeEntry, &ipv4Cfg, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        CPSW_ALE_SAFE_ASSIGN(pAleIndex, freeEntry);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_setIPv4PolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           CpswAle_Ipv4AddrType_t ipv4Addr,
                                           CPSW_ALE_ADDR_TYPE addrType,
                                           uint32_t numLSBIgnore,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIndex,
                                           bool *pEntryAlloced)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_addIPv4Addr(cpswAleRegBase,
                                     ipv4Addr,
                                     numLSBIgnore,
                                     aleEntryIndex,
                                     pEntryAlloced);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateIPv4Policer(cpswAleRegBase,
                                             policerEntryIdx,
                                             *aleEntryIndex,
                                             ipv4Addr,
                                             numLSBIgnore);
    }

    if (CPSW_SOK == retVal)
    {
        CpswAle_setIPPolicerEntry(cpswAleRegBase,
                                  policerEntryIdx,
                                  *aleEntryIndex,
                                  addrType);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase,
                                             policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validateIPv6Policer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t aleEntryIdx,
                                           CpswAle_Ipv6AddrType_t ipv6Addr,
                                           uint32_t numLSBIgnore)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase,
                                              policerEntryIdx,
                                              aleEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (numLSBIgnore >
            CSL_CPSW_getIpv6IgnBitsMax(CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)))
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t matchEntry;

        retVal = CpswAle_findIPv6Addr(cpswAleRegBase,
                                      ipv6Addr,
                                      numLSBIgnore,
                                      &matchEntry);
        if (CPSW_SOK == retVal)
        {
            if (matchEntry != aleEntryIdx)
            {
                retVal = CPSW_EINVALID_PARAMS;
            }
        }
    }

    return retVal;
}

static void  CpswAle_mapIPv6AddrOctet2RegWord(CpswAle_Ipv6AddrType_t ipv6Addr,
                                              uint32_t(*ipv6RegWord)[4])
{
    uint32_t i;

    /* Setup the IPv6 configuration */
    for (i = 0; i < 4; i++)
    {
        (*ipv6RegWord)[i] = ((*ipv6Addr)[4 * i] << 24) |
                            ((*ipv6Addr)[4 * i + 1] << 16) |
                            ((*ipv6Addr)[4 * i + 2] << 8) |
                            ((*ipv6Addr)[4 * i + 3] << 0);
    }
}

static void  CpswAle_maskIPv6AddrRegWord(uint32_t ipv6RegWord[4],
                                         const uint32_t numLSBIgnore)
{
    Uint32 addrMask[4];
    UInt32 addrMaskBitPos;
    UInt32 addrMaskBitRemain;
    uint32_t i;

    /* Setup the IPv6 address mask configuration */
    addrMaskBitRemain = numLSBIgnore;
    addrMaskBitPos = 96;
    for (i = 0; i < 4; i++)
    {
        if (addrMaskBitRemain > addrMaskBitPos)
        {
            addrMask[i] =
                ~((1U << (addrMaskBitRemain - addrMaskBitPos)) - 1U);
            addrMaskBitRemain -= addrMaskBitPos;
        }
        else
        {
            addrMask[i] = ~0U;
        }

        addrMaskBitPos -= 32;
    }

    /* Setup the IPv6 configuration */
    for (i = 0; i < 4; i++)
    {
        ipv6RegWord[i] &= addrMask[i];
    }
}

static bool  CpswAle_matchIPv6AddrRegWord(const uint32_t ipv6RegWord1[4],
                                          const uint32_t ipv6RegWord2[4])
{
    bool match;
    uint32_t i;

    /* Setup the IPv6 configuration */
    for (i = 0; i < 4; i++)
    {
        if (ipv6RegWord1[i] != ipv6RegWord2[i])
        {
            break;
        }
    }

    if (i < 4)
    {
        match = FALSE;
    }
    else
    {
        match = TRUE;
    }

    return match;
}

static int32_t CpswAle_findIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                    const CpswAle_Ipv6AddrType_t ipv6Addr,
                                    const uint32_t numLSBIgnore,
                                    uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);
    uint32_t ipv6MatchAddr[4];

    CpswAle_mapIPv6AddrOctet2RegWord(ipv6Addr, &ipv6MatchAddr);
    CpswAle_maskIPv6AddrRegWord(ipv6MatchAddr, numLSBIgnore);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_POLICER)
        {
            if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                i,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                CSL_ALE_POLICER_ENTRYTYPE_IPV6)
            {
                CSL_CPSW_ALE_IPv6_ENTRY ipv6Entry;
                uint32_t entryIpv6AddrReg[4];

                CSL_CPSW_getAleIPv6Entry(cpswAleRegBase, i, &ipv6Entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                {
                    const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                        CPSW_ALE_IPV6_ADDR_INIT_LIST(ipv6Entry.address);
                    CpswAle_mapIPv6AddrOctet2RegWord(&ipv6Addr, &entryIpv6AddrReg);
                }
                CpswAle_maskIPv6AddrRegWord(entryIpv6AddrReg, numLSBIgnore);

                if (CpswAle_matchIPv6AddrRegWord(entryIpv6AddrReg, ipv6MatchAddr))
                {
                    break;
                }
            }
        }
    }

    if (i != aleTableDepth)
    {
        *entry = i;
        ret = CPSW_SOK;
    }

    return ret;
}

/*
 *  Finds index to the next free entry in ALE table.
 */
static int32_t CpswAle_getFreeIPv6Entry(CSL_AleRegs *cpswAleRegBase,
                                        uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_SOK;
    uint32_t aleTableDepth;
    uint32_t ramDepth;

    aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    ramDepth = CSL_CPSW_getAleStatusRamDepth(cpswAleRegBase);
    switch (ramDepth)
    {
        case CSL_ALE_RAMDEPTH_32:  /** RAMDEPTH is 32  */
            aleTableDepth -= 32;
            break;

        case CSL_ALE_RAMDEPTH_64:  /** RAMDEPTH is 64  */
            aleTableDepth -= 64;
            break;

        case CSL_ALE_RAMDEPTH_128: /** RAMDEPTH is 128  */
            aleTableDepth -= 128;
            break;
    }

    for (i = 0; i < aleTableDepth; i++)
    {
        if ((CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                      CSL_CPSW_getAleIPv6LowEntryIndex(cpswAleRegBase, i),
                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
             CSL_ALE_ENTRYTYPE_FREE)
            &&
            (CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                      CSL_CPSW_getAleIPv6HighEntryIndex(cpswAleRegBase, i),
                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
             CSL_ALE_ENTRYTYPE_FREE))
        {
            break;
        }
    }

    if (aleTableDepth == i)
    {
        ret = CPSW_EALE_TABLE_FULL;
    }
    else
    {
        *entry = i;
    }

    return ret;
}

static int32_t CpswAle_addIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv6AddrType_t ipv6Addr,
                                   const uint32_t numLSBIgnore,
                                   uint32_t *pAleIndex,
                                   bool *pEntryAlloced)
{
    uint32_t freeEntry;
    int32_t ret = CPSW_SOK;
    uint32_t addrEntry;

    /* Check for space in the ALE table */
    ret = CpswAle_findIPv6Addr(cpswAleRegBase,
                               ipv6Addr,
                               numLSBIgnore,
                               &addrEntry);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
    {
        ret = CpswAle_getFreeIPv6Entry(cpswAleRegBase, &freeEntry);
        if (ret != CPSW_EALE_TABLE_FULL)
        {
            CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, TRUE);
        }
    }
    else
    {
        freeEntry = addrEntry;
        CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, FALSE);
    }

    if (ret != CPSW_EALE_TABLE_FULL)
    {
        CSL_CPSW_ALE_IPv6_ENTRY ipv6Cfg;

        CPSW_UTILS_ARRAY_COPY(ipv6Cfg.address, (*ipv6Addr));
        ipv6Cfg.numLSBIgnore = numLSBIgnore;

        CSL_CPSW_setAleIPv6Entry(cpswAleRegBase,
                                 freeEntry,
                                 &ipv6Cfg,
                                 CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        CPSW_ALE_SAFE_ASSIGN(pAleIndex, freeEntry);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_setIPv6PolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           CpswAle_Ipv6AddrType_t ipv6Addr,
                                           CPSW_ALE_ADDR_TYPE addrtype,
                                           uint32_t numLSBIgnore,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIndex,
                                           bool *pEntryAlloced)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_addIPv6Addr(cpswAleRegBase,
                                     ipv6Addr,
                                     numLSBIgnore,
                                     aleEntryIndex,
                                     pEntryAlloced);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateIPv6Policer(cpswAleRegBase,
                                             policerEntryIdx,
                                             *aleEntryIndex,
                                             ipv6Addr,
                                             numLSBIgnore);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setIPPolicerEntry(cpswAleRegBase,
                                           policerEntryIdx,
                                           *aleEntryIndex,
                                           addrtype);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase, policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_findEtherType(CSL_AleRegs *cpswAleRegBase,
                                     uint32_t etherType,
                                     uint32_t *entry)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                         i,
                                         CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_POLICER)
        {
            if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                i,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                CSL_ALE_POLICER_ENTRYTYPE_ETHERTYPE)
            {
                CSL_CPSW_ALE_ETHERTYPE_ENTRY etherTypeEntry;

                CSL_CPSW_getAleEthertypeEntry(cpswAleRegBase,
                                              i,
                                              &etherTypeEntry,
                                              CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                if (etherTypeEntry.ethertype == etherType)
                {
                    break;
                }
            }
        }
    }

    if (i != aleTableDepth)
    {
        *entry = i;
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_addEtherType(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t etherType,
                                    uint32_t *pAleIndex,
                                    bool *pEntryAlloced)
{
    uint32_t freeEntry;
    int32_t ret = CPSW_SOK;
    uint32_t addrEntry;

    /* Check for space in the ALE table */
    ret = CpswAle_findEtherType(cpswAleRegBase, etherType, &addrEntry);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == ret)
    {
        ret = CpswAle_getFreeEntry(cpswAleRegBase, &freeEntry);
        if (ret != CPSW_EALE_TABLE_FULL)
        {
            CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, TRUE);
        }
    }
    else
    {
        freeEntry = addrEntry;
        CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, FALSE);
    }

    if (ret != CPSW_EALE_TABLE_FULL)
    {
        CSL_CPSW_ALE_ETHERTYPE_ENTRY etherTypeEntry;

        etherTypeEntry.ethertype = etherType;
        CSL_CPSW_setAleEthertypeEntry(cpswAleRegBase,
                                      freeEntry,
                                      &etherTypeEntry,
                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        CPSW_ALE_SAFE_ASSIGN(pAleIndex, freeEntry);
        ret = CPSW_SOK;
    }

    return ret;
}

static int32_t CpswAle_validateEtherTypePolicer(CSL_AleRegs *cpswAleRegBase,
                                                uint32_t etherType,
                                                uint32_t policerEntryIdx,
                                                uint32_t aleEntryIdx)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase,
                                              policerEntryIdx,
                                              aleEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (etherType >
            CSL_CPSW_getEthertypeMax(CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)))
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t matchEntry;

        retVal = CpswAle_findEtherType(cpswAleRegBase, etherType, &matchEntry);
        if (CPSW_SOK == retVal)
        {
            if (matchEntry != aleEntryIdx)
            {
                retVal = CPSW_EINVALID_PARAMS;
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_setEtherTypePolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                                uint32_t policerEntryIdx,
                                                uint32_t etherType,
                                                uint32_t peakRateInBps,
                                                uint32_t commitRateInBps,
                                                uint32_t aleFreq,
                                                uint32_t *aleEntryIndex,
                                                bool *pEntryAlloced)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_addEtherType(cpswAleRegBase,
                                      etherType,
                                      aleEntryIndex,
                                      pEntryAlloced);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateEtherTypePolicer(cpswAleRegBase,
                                                  etherType,
                                                  policerEntryIdx,
                                                  *aleEntryIndex);
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase,
                                    policerEntryIdx,
                                    &polCfg);

        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_ETHERTYPE_VALID;
        polCfg.ethertypeIdx = *aleEntryIndex;

        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase,
                                    policerEntryIdx,
                                    &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase,
                                             policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validateMacAddrPolicer(CSL_AleRegs *cpswAleRegBase,
                                              CpswAle_MacAddrType_t addr,
                                              uint32_t policerEntryIdx,
                                              uint32_t aleEntryIdx,
                                              CPSW_ALE_ADDR_TYPE addrType)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase,
                                              policerEntryIdx,
                                              aleEntryIdx);

    if (CPSW_SOK == retVal)
    {
        uint32_t word0, word1;
        bool mcastAddr;

        CSL_CPSW_mapMacAddr2TableWord(&word0,
                                      &word1,
                                      (uint8_t *)addr,
                                      CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        mcastAddr = CpswAle_isMcastAddr(addr);
        if ((mcastAddr) && (addrType == CPSW_ALE_ADDR_TYPE_SRC))
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    return retVal;
}

static int32_t CpswAle_setMacAddrPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerEntryIdx,
                                              CpswAle_MacAddrType_t addr,
                                              uint32_t vlanId,
                                              uint32_t portNum,
                                              CPSW_ALE_ADDR_TYPE addrType,
                                              uint32_t peakRateInBps,
                                              uint32_t commitRateInBps,
                                              uint32_t aleFreq,
                                              uint32_t *aleEntryIdx,
                                              bool *pEntryAlloced)
{
    int32_t retVal;

    retVal = CpswAle_findAddr(cpswAleRegBase, addr, vlanId, aleEntryIdx);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == retVal)
    {
        /* If entry is not found add the entry with default settings */
        retVal = CpswAle_addMacAddrEntry(cpswAleRegBase,
                                         addr,
                                         vlanId,
                                         portNum,
                                         addrType,
                                         aleEntryIdx);
        if (CPSW_SOK == retVal)
        {
            CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, TRUE);
        }
    }
    else
    {
        CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, FALSE);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateMacAddrPolicer(cpswAleRegBase,
                                                addr,
                                                policerEntryIdx,
                                                *aleEntryIdx,
                                                addrType);
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        if (addrType == CPSW_ALE_ADDR_TYPE_SRC)
        {
            polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_SRC_MAC_VALID;
            polCfg.srcMacIdx = *aleEntryIdx;
        }
        else
        {
            polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_DST_MAC_VALID;
            polCfg.dstMacIdx = *aleEntryIdx;
        }

        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase,
                                             policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validateThreadPolicer(CSL_AleRegs *cpswAleRegBase,
                                             uint32_t policerEntryIdx,
                                             uint32_t threadId)
{
    int32_t retVal;

    retVal = CpswAle_validatePolicerIndex(cpswAleRegBase, policerEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (threadId > CSL_ALE_THREADMAPVAL_THREADVAL_MAX)
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY policerEntry;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &policerEntry);
        if (policerEntry.validBitmap == 0)
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    return retVal;
}

static int32_t CpswAle_setThreadPolicerConfig(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerEntryIdx,
                                              uint32_t threadId)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateThreadPolicer(cpswAleRegBase, policerEntryIdx, threadId);
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_THREAD_VALID;
        polCfg.thread = threadId;
        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    return retVal;
}

static int32_t CpswAle_disableThreadPolicerConfig(CSL_AleRegs *cpswAleRegBase,
                                                  uint32_t policerEntryIdx)
{
    CSL_CPSW_disableAlePolicerThread(cpswAleRegBase, policerEntryIdx);
    return CPSW_SOK;
}

static int32_t CpswAle_validateVlanPolicer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t aleEntryIdx)
{
    int32_t retVal;

    retVal = CpswAle_validateSetPolicerCommon(cpswAleRegBase, policerEntryIdx, aleEntryIdx);

    return retVal;
}

static int32_t CpswAle_setVlanPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t vlanId,
                                           bool outVlanFlag,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq,
                                           uint32_t *aleEntryIdx,
                                           bool *pEntryAlloced)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_findVlan(cpswAleRegBase, vlanId, outVlanFlag,
                                  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                                  aleEntryIdx);
        if (CPSW_EALE_VLAN_NOT_FOUND == retVal)
        {
            /* If entry is not found add the entry with default settings */
            retVal = CpswAle_addVlanEntry(cpswAleRegBase, vlanId, outVlanFlag, aleEntryIdx);
            if (retVal == CPSW_SOK)
            {
                CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, TRUE);
            }
        }
        else
        {
            CPSW_ALE_SAFE_ASSIGN(pEntryAlloced, FALSE);
        }
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_validateVlanPolicer(cpswAleRegBase, policerEntryIdx, *aleEntryIdx);
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        if (outVlanFlag)
        {
            polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_OVLAN_VALID;
            polCfg.ovlanIdx = *aleEntryIdx;
        }
        else
        {
            polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_VLAN_VALID;
            polCfg.vlanIdx = *aleEntryIdx;
        }

        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase,
                                             policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validatePortPolicer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t port_num)
{
    int32_t retVal = CPSW_SOK;

    retVal = CpswAle_validatePolicerIndex(cpswAleRegBase, policerEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (port_num >=
            CpswAle_getNumPortsFromCpswAleRegBase((uintptr_t)cpswAleRegBase))
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    return retVal;
}

static int32_t CpswAle_setPortPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t port_num,
                                           bool isTrunkPort,
                                           uint32_t peakRateInBps,
                                           uint32_t commitRateInBps,
                                           uint32_t aleFreq)
{
    int32_t retVal = CPSW_SOK;

    retVal = CpswAle_validatePortPolicer(cpswAleRegBase, policerEntryIdx,
                                         port_num);

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_PORT_VALID;
        polCfg.port = port_num;
        if (isTrunkPort)
        {
            polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_PORT_TRUNK_VALID;
        }

        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase,
                                             policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_validatePriorityPolicer(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t policerEntryIdx,
                                               uint32_t priority)
{
    int32_t retVal = CPSW_SOK;

    retVal = CpswAle_validatePolicerIndex(cpswAleRegBase, policerEntryIdx);

    if (CPSW_SOK == retVal)
    {
        if (priority > CSL_ALE_POLICECFG0_PRI_VAL_MAX)
        {
            retVal = CPSW_EINVALID_PARAMS;
        }
    }

    return retVal;
}

static int32_t CpswAle_setPriorityPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t policerEntryIdx,
                                               uint32_t priority,
                                               uint32_t peakRateInBps,
                                               uint32_t commitRateInBps,
                                               uint32_t aleFreq)
{
    int32_t retVal = CPSW_SOK;

    retVal = CpswAle_validatePriorityPolicer(cpswAleRegBase, policerEntryIdx,
                                             priority);

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_PRI_VALID;
        polCfg.pri = priority;
        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_setPolicerRateLimit(cpswAleRegBase, policerEntryIdx,
                                             peakRateInBps,
                                             commitRateInBps,
                                             aleFreq);
    }

    return retVal;
}

static int32_t CpswAle_setNoMatchEntry0Policer(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t aleFreq,
                                               uint32_t peakRateInBitsPs,
                                               uint32_t commitRateInBitsPs)
{
    CSL_CPSW_ALE_POLICER_ENTRY polCfg;
    uint32_t pir_idl_inc_val = CpswAle_mapBw2IdlIncVal(aleFreq, peakRateInBitsPs);
    uint32_t cir_idl_inc_val = CpswAle_mapBw2IdlIncVal(aleFreq, commitRateInBitsPs);

    CSL_CPSW_getAlePolicerEntry(cpswAleRegBase,
                                CPSW_ALE_NOMATCH_ENTRY0_POLICER_INDEX,
                                &polCfg);

    /* Confirm entry 0 is free. If entry 0 is not free then , ALE open
     * sequence must be modified to ensure entry 0 is free to allow
     * configuration of NO_MATCH policer
     */
    Cpsw_assert(polCfg.validBitmap == 0);
    polCfg.validBitmap = (CSL_CPSW_ALE_POLICER_PIR_VALID |
                          CSL_CPSW_ALE_POLICER_CIR_VALID);
    polCfg.pirIdleIncVal = pir_idl_inc_val;
    polCfg.cirIdleIncVal = cir_idl_inc_val;

    CSL_CPSW_setAlePolicerEntry(cpswAleRegBase,
                                CPSW_ALE_NOMATCH_ENTRY0_POLICER_INDEX,
                                &polCfg);
    CpswAle_clearSelectedPolicerStats(cpswAleRegBase,
                                      CPSW_ALE_NOMATCH_ENTRY0_POLICER_INDEX);
    return CPSW_SOK;
}

static int32_t CpswAle_setPolicerControl(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t aleFreqHz,
                                         bool policingEnable,
                                         bool yellowDropEnable,
                                         bool redDropEnable,
                                         CPSW_ALE_POLICER_YELLOWTHRESH_E yellowThresh,
                                         CPSW_ALE_POLICER_NOMATCH_MODE_E policerMatchMode,
                                         const CpswAle_UnregulatedTrafficPolicer *noMatchPolicer)
{
    CSL_CPSW_ALE_POLICER_CONTROL policerControl;

    if (CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER == policerMatchMode)
    {
        CpswAle_setNoMatchEntry0Policer(cpswAleRegBase,
                                        aleFreqHz,
                                        noMatchPolicer->peakRateInBitsPerSec,
                                        noMatchPolicer->commitRateInBitsPerSec);
    }

    CSL_CPSW_getAlePolicerControlReg(cpswAleRegBase, &policerControl);

    policerControl.policingEnable   = policingEnable;
    policerControl.redDropEnable    = redDropEnable;
    policerControl.yellowDropEnable = yellowDropEnable;
    policerControl.yellowDropThresh = (CSL_CPSW_ALE_POLICER_CONTROL_YELLOWTHRESH)yellowThresh;
    policerControl.policeMatchMode  = (CSL_CPSW_ALE_POLICER_CONTROL_POLICING_MATCH_MODE)policerMatchMode;

    CSL_CPSW_setAlePolicerControlReg(cpswAleRegBase, &policerControl);

    return CPSW_SOK;
}

static void  CpswAle_validateIPVersion(CSL_AleRegs *cpswAleRegBase)
{
    CSL_CPSW_ALE_VERSION versionInfo;
    const CpswSoc_CpswIpInfo *ipInfo = CpswAle_getCpswIpInfoFromBaseAddr((uintptr_t)cpswAleRegBase);

    CSL_CPSW_getAleVersionInfo(cpswAleRegBase, &versionInfo);
    Cpsw_assert(versionInfo.id == ipInfo->aleInfo.versionInfo.moduleId);
    Cpsw_assert(versionInfo.majorVer == ipInfo->aleInfo.versionInfo.majorVer);
    Cpsw_assert(versionInfo.minorVer == ipInfo->aleInfo.versionInfo.minorVer);
    Cpsw_assert(versionInfo.rtlVer == ipInfo->aleInfo.versionInfo.rtlVer);
}

static void CpswAle_validateRegisterResetVal(volatile uint32_t *regAddr,
                                             uint32_t resetValExpected)
{
    Cpsw_assert(CSL_REG_RD(regAddr) == resetValExpected);
}

static int32_t CpswAle_setAlePortState(CSL_AleRegs *cpswAleRegBase,
                                       uint32_t port_num,
                                       CSL_CPSW_ALE_PORTSTATE portState)
{
    CSL_CPSW_setAlePortState(cpswAleRegBase, port_num, portState);
    return CPSW_SOK;
}

static uint32_t CpswAle_mapNumPktsPerSec2RateLimit(CSL_AleRegs *cpswAleRegBase,
                                                   uint32_t aleFreq,
                                                   uint32_t numPktsPerSec)
{
    uint32_t alePrescale = CSL_CPSW_getAlePrescaleReg(cpswAleRegBase);
    uint32_t preScalerPulseFreq;

    if (alePrescale == 0)
    {
        alePrescale = 1;
    }

    preScalerPulseFreq = aleFreq / alePrescale;
    return(numPktsPerSec / preScalerPulseFreq);
}

static uint32_t CpswAle_mapRateLimit2NumPktsPerSec(CSL_AleRegs *cpswAleRegBase,
                                                   uint32_t aleFreq,
                                                   uint32_t rateLimit)
{
    uint32_t alePrescale = CSL_CPSW_getAlePrescaleReg(cpswAleRegBase);
    uint32_t preScalerPulseFreq;

    if (alePrescale == 0)
    {
        alePrescale = 1;
    }

    preScalerPulseFreq = aleFreq / alePrescale;
    return(rateLimit * preScalerPulseFreq);
}

static uint32_t CpswAle_mapAleAgingIntervalMsToAgingTimer(uint32_t aleFreq,
                                                          uint32_t
                                                          aleAgingIntervalInMs)
{
    uint32_t agingTimer =
        (aleFreq / CPSW_ALE_AGING_TIMER_DIV_FACTOR) * aleAgingIntervalInMs;

    agingTimer /= 1000;

    /* Round up the aging timer register value if
     * above division operation truncated the timer
     * value
     */
    if ((agingTimer % 1000) != 0)
    {
        agingTimer++;
    }

    return(agingTimer);
}

static int32_t CpswAle_setPolicerDefaultThreadConfig(CSL_AleRegs *cpswAleRegBase,
                                                     bool threadEnable,
                                                     uint32_t threadVal,
                                                     bool enablePriorityOR,
                                                     bool disableMacPortDefaultThread)
{
    CSL_CPSW_ALE_POLICER_GLOB_CONFIG defThreadConfig;
    CSL_CPSW_ALE_POLICER_CONTROL policerControl;

    defThreadConfig.defThreadEnable = threadEnable;
    defThreadConfig.defThread = threadVal;

    CSL_CPSW_setAlePolicerGlobConfig(cpswAleRegBase, &defThreadConfig);

    if (CpswAle_isFlowPrioritySupported(cpswAleRegBase))
    {
        /* disableMacPort and enablePriorityOR are part of policer control register */
        CSL_CPSW_getAlePolicerControlReg(cpswAleRegBase, &policerControl);

        policerControl.disableMacPortDefaultThread = disableMacPortDefaultThread;
        policerControl.enablePriorityOR = enablePriorityOR;

        CSL_CPSW_setAlePolicerControlReg(cpswAleRegBase, &policerControl);
    }

    return CPSW_SOK;
}

static int32_t CpswAle_getPolicerDefaultThreadConfig(CSL_AleRegs *cpswAleRegBase,
                                                     CpswAle_GetDefaultThreadConfigOutArgs *defaultThreadConfig)
{
    CSL_CPSW_ALE_POLICER_GLOB_CONFIG defThreadConfig;
    CSL_CPSW_ALE_POLICER_CONTROL policerControl;

    CSL_CPSW_getAlePolicerGlobConfig(cpswAleRegBase, &defThreadConfig);

    if (CpswAle_isFlowPrioritySupported(cpswAleRegBase))
    {
        /* disableMacPort and enablePriorityOR are part of policer control register
        **/
        CSL_CPSW_getAlePolicerControlReg(cpswAleRegBase, &policerControl);

        defaultThreadConfig->disableMacPortDefaultThread = policerControl.disableMacPortDefaultThread;
        defaultThreadConfig->enablePriorityOR = policerControl.enablePriorityOR;
    }
    else
    {
        defaultThreadConfig->disableMacPortDefaultThread = FALSE;
        defaultThreadConfig->enablePriorityOR = FALSE;
    }

    defaultThreadConfig->defaultThreadEnable = defThreadConfig.defThreadEnable;
    defaultThreadConfig->threadId = defThreadConfig.defThread;

    return CPSW_SOK;
}

static uint32_t CpswAle_mapBw2IdlIncVal(uint32_t aleFreq,
                                        uint32_t rateInBps)
{
    uint64_t idl_inc_val;

    /*
     * CIR policing rate in Mbit/s = ((ALE frequency in Mhz) * cir_idle_inc_val)
     * / 32768
     * PIR policing rate in Mbit/s = ((ALE frequency in Mhz) * pir_idle_inc_val)
     * / 32768
     */
    idl_inc_val = ((uint64_t)rateInBps * CPSW_ALE_IDL_INC_VAL_DIV_FACTOR);
    idl_inc_val /= aleFreq;

    return (uint32_t)idl_inc_val;
}

static uint32_t CpswAle_mapIdlIncVal2Bw(uint32_t aleFreq,
                                        uint32_t idl_inc_val)
{
    uint64_t rateInBps;

    /*
     * CIR policing rate in Mbit/s = ((ALE frequency in Mhz) * cir_idle_inc_val)
     * / 32768
     * PIR policing rate in Mbit/s = ((ALE frequency in Mhz) * pir_idle_inc_val)
     * / 32768
     */
    rateInBps = ((uint64_t)idl_inc_val * aleFreq);
    rateInBps /= CPSW_ALE_IDL_INC_VAL_DIV_FACTOR;

    return (uint32_t)rateInBps;
}

static int32_t CpswAle_setPolicerRateLimit(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t policerEntryIdx,
                                           uint32_t peakRateInBitsPs,
                                           uint32_t commitRateInBitsPs,
                                           uint32_t aleFreq)
{
    CSL_CPSW_ALE_POLICER_ENTRY polCfg;
    uint32_t pir_idl_inc_val = CpswAle_mapBw2IdlIncVal(aleFreq, peakRateInBitsPs);
    uint32_t cir_idl_inc_val = CpswAle_mapBw2IdlIncVal(aleFreq, commitRateInBitsPs);

    CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);

    polCfg.validBitmap |= (CSL_CPSW_ALE_POLICER_PIR_VALID |
                           CSL_CPSW_ALE_POLICER_CIR_VALID);
    polCfg.pirIdleIncVal = pir_idl_inc_val;
    polCfg.cirIdleIncVal = cir_idl_inc_val;

    CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerEntryIdx, &polCfg);
    CpswAle_clearSelectedPolicerStats(cpswAleRegBase, policerEntryIdx);
    return CPSW_SOK;
}

void CpswAle_clearGlobalPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                     bool clearAllHit,
                                     bool clearRedHit,
                                     bool clearYellowHit)
{
    CSL_CPSW_ALE_POLICER_TEST_CONTROL policerTstCtrl;

    policerTstCtrl.polClrallHit       = clearAllHit;
    policerTstCtrl.polClrallRedhit    = clearRedHit;
    policerTstCtrl.polClrallYellowhit = clearYellowHit;
    policerTstCtrl.polClrselAll       = 0;
    policerTstCtrl.polTestIdx         = 0; /* Dont care param */
    CSL_CPSW_setAlePolicerTestControlReg(cpswAleRegBase, &policerTstCtrl);
}

static void CpswAle_clearSelectedPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t policerIndex)
{
    CSL_CPSW_ALE_POLICER_TEST_CONTROL policerTstCtrl;

    policerTstCtrl.polClrallHit       = 0;
    policerTstCtrl.polClrallRedhit    = 0;
    policerTstCtrl.polClrallYellowhit = 0;
    policerTstCtrl.polClrselAll       = 1;
    policerTstCtrl.polTestIdx         = policerIndex; /* Dont care param */
    CSL_CPSW_setAlePolicerTestControlReg(cpswAleRegBase, &policerTstCtrl);
}

static void CpswAle_getPolicerStats(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t policerIndex,
                                    bool *pPolHit,
                                    bool *pPolRedHit,
                                    bool *pPolYellowHit,
                                    bool clearStats)
{
    CSL_CPSW_ALE_POLICER_TEST_CONTROL policerTstCtrl;
    CSL_CPSW_ALE_POLICER_HSTAT policerHStatCfg;

    policerTstCtrl.polClrallHit       = 0;
    policerTstCtrl.polClrallRedhit    = 0;
    policerTstCtrl.polClrallYellowhit = 0;
    policerTstCtrl.polClrselAll       = 0;
    policerTstCtrl.polTestIdx         = policerIndex; /* Dont care param */
    CSL_CPSW_setAlePolicerTestControlReg(cpswAleRegBase, &policerTstCtrl);

    CSL_CPSW_getAlePolicerHstatReg(cpswAleRegBase, &policerHStatCfg);
    *pPolHit = policerHStatCfg.polHit;
    *pPolRedHit = policerHStatCfg.polRedhit;
    *pPolYellowHit = policerHStatCfg.polYellowhit;

    if (clearStats)
    {
        policerTstCtrl.polClrselAll = 1;
        policerTstCtrl.polTestIdx = policerIndex; /* Dont care param */
        CSL_CPSW_setAlePolicerTestControlReg(cpswAleRegBase, &policerTstCtrl);
    }
}

static int32_t CpswAle_getMirrorMatchIndex(CSL_AleRegs *cpswAleRegBase,
                                           const CpswAle_MirrorMatchParams *mirrorMatchParams,
                                           uint32_t *aleEntryIndex)
{
    int32_t retVal = CPSW_SOK;

    switch (mirrorMatchParams->entryType)
    {
        case CPSW_ALE_TBLENTRYTYPE_ADDR:
        {
            const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                CPSW_ALE_MAC_ADDR_INIT_LIST(mirrorMatchParams->macAddrInfo.addr.addr);

            retVal = CpswAle_findAddr(cpswAleRegBase,
                                      &addr,
                                      mirrorMatchParams->macAddrInfo.addr.vlanId,
                                      aleEntryIndex);
            if (CPSW_EALE_ADDRESS_NOT_FOUND == retVal)
            {
                /* As per spec enabling Mirror Match Entry enables the match
                 * mirror option. When this bit is set any traffic whose
                 * destination, source, VLAN or OUI matches the
                 * mirror_midx entry index will have that traffic also sent to
                 * the mirror_top port.
                 * So address type to dest to allow MCast address to be
                 * mirrored
                 */
                retVal = CpswAle_addMacAddrEntry(cpswAleRegBase,
                                                 &addr,
                                                 mirrorMatchParams->macAddrInfo.addr.vlanId,
                                                 mirrorMatchParams->macAddrInfo.egressPortNum,
                                                 CPSW_ALE_ADDR_TYPE_DST,
                                                 aleEntryIndex);
            }

            break;
        }

        case CPSW_ALE_TBLENTRYTYPE_VLAN:
            retVal = CpswAle_findVlan(cpswAleRegBase,
                                      mirrorMatchParams->vlanIdInfo.vlanId,
                                      mirrorMatchParams->vlanIdInfo.outerVlanFlag,
                                      NULL,
                                      NULL,
                                      NULL,
                                      NULL,
                                      NULL,
                                      NULL,
                                      NULL,
                                      NULL,
                                      aleEntryIndex);
            if (CPSW_EALE_VLAN_NOT_FOUND == retVal)
            {
                retVal = CpswAle_addVlanEntry(cpswAleRegBase,
                                              mirrorMatchParams->vlanIdInfo.vlanId,
                                              mirrorMatchParams->vlanIdInfo.outerVlanFlag,
                                              aleEntryIndex);
            }

            break;

        case CPSW_ALE_TBLENTRYTYPE_ETHERTYPE:
            retVal = CpswAle_addEtherType(cpswAleRegBase,
                                          mirrorMatchParams->etherTypeInfo.etherType,
                                          aleEntryIndex,
                                          NULL);
            break;

        case CPSW_ALE_TBLENTRYTYPE_OUI:
        {
            const uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
                CPSW_ALE_OUI_ADDR_INIT_LIST(mirrorMatchParams->ouiInfo.ouiAddr);

            retVal = CpswAle_addOuiAddr(cpswAleRegBase,
                                        &ouiAddr,
                                        aleEntryIndex,
                                        NULL);
            break;
        }

        case CPSW_ALE_TBLENTRYTYPE_IPV4:
        {
            const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV4_ADDR_INIT_LIST(mirrorMatchParams->ipv4Info.ipv4Addr);

            retVal = CpswAle_addIPv4Addr(cpswAleRegBase,
                                         &ipv4Addr,
                                         mirrorMatchParams->ipv4Info.numLSBIgnoreBits,
                                         aleEntryIndex,
                                         NULL);
            break;
        }

        case CPSW_ALE_TBLENTRYTYPE_IPV6:
        {
            const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV6_ADDR_INIT_LIST(mirrorMatchParams->ipv6Info.ipv6Addr);

            retVal = CpswAle_addIPv6Addr(cpswAleRegBase,
                                         &ipv6Addr,
                                         mirrorMatchParams->ipv6Info.numLSBIgnoreBits,
                                         aleEntryIndex,
                                         NULL);
            break;
        }

        default:
            retVal = CPSW_EINVALID_PARAMS;
    }

    return retVal;
}

static int32_t CpswAle_setPortMirrorConfig(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t maxNumPorts,
                                           const CpswAle_SetPortMirroringConfigInArgs *portMirrorCfg)
{
    int32_t retVal = CPSW_SOK;
    uint32_t regVal;
    uint32_t i;

    regVal = CSL_CPSW_getAleControlReg(cpswAleRegBase);
    if (portMirrorCfg->enableSrc)
    {
        bool enableSrcPortMirror;
        for (i = 0; i < maxNumPorts; i++)
        {
            enableSrcPortMirror = NOT_ZERO((portMirrorCfg->srcPortNumMask >> i) & 0x1);
            CSL_CPSW_setAlePortMirrorSouce(cpswAleRegBase, i, enableSrcPortMirror);
        }

        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_SEN, TRUE);
    }
    else
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_SEN, FALSE);
    }

    if (portMirrorCfg->enableDst)
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_DP, portMirrorCfg->destPortNum);
        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_DEN, TRUE);
    }
    else
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_DEN, FALSE);
    }

    if (portMirrorCfg->enableMatch)
    {
        uint32_t mirrorMatchIndex;

        retVal = CpswAle_getMirrorMatchIndex(cpswAleRegBase,
                                             &portMirrorCfg->matchParams,
                                             &mirrorMatchIndex);
        if (CPSW_SOK == retVal)
        {
            CSL_CPSW_setAleCtrl2MirrorMatchIndex(cpswAleRegBase,
                                                 mirrorMatchIndex);
            CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_MEN, TRUE);
        }
    }
    else
    {
        CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_MEN, FALSE);
    }

    CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_TOP, portMirrorCfg->toPortNum);
    CSL_CPSW_setAleControlReg(cpswAleRegBase, regVal);

    return retVal;
}

static int32_t CpswAle_setPortTrunkConfig(CSL_AleRegs *cpswAleRegBase,
                                          const CpswAle_SetTrunkConfigInArgs *trunkCfg)
{
    uint32_t i;
    CSL_CPSW_ALE_CTRL2_TRUNK_CONFIG cslTrunkCfg;

    for (i = 0; i < trunkCfg->numPorts; i++)
    {
        CSL_CPSW_setAlePortControlTrunk(cpswAleRegBase,
                                        trunkCfg->trunkPortIndex[i],
                                        TRUE,
                                        trunkCfg->trunkId);
    }

    cslTrunkCfg.trunkBase            = trunkCfg->trunkHashBase;
    cslTrunkCfg.trunkEnableDestIP    = trunkCfg->enableDstIP;
    cslTrunkCfg.trunkEnableSrcIP     = trunkCfg->enableSrcIP;
    cslTrunkCfg.trunkEnableInnerVLAN = trunkCfg->enableInnerVLAN;
    cslTrunkCfg.trunkEnablePri       = trunkCfg->enablePri;
    cslTrunkCfg.trunkEnableSrc       = trunkCfg->enableSrc;
    cslTrunkCfg.trunkEnableDst       = trunkCfg->enableDst;
    CSL_CPSW_setAleCtrl2TrunkParams(cpswAleRegBase, &cslTrunkCfg);

    return CPSW_SOK;
}

void CpswAle_initParams(CpswAle_Config *pALEConfig)
{
    uint32_t i;

    memset(pALEConfig, 0, sizeof(*pALEConfig));
    pALEConfig->modeFlags = CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE;
    pALEConfig->policerGlobalConfig.policingEnable   = TRUE;
    pALEConfig->policerGlobalConfig.redDropEnable    = TRUE;
    pALEConfig->policerGlobalConfig.yellowDropEnable = FALSE;

    pALEConfig->agingConfig.enableAutoAging  = TRUE;
    pALEConfig->agingConfig.agingPeriodInMs  = CPSW_ALE_CFG_DEFAULT_AGING_PERIOD_MS;
    pALEConfig->vlanConfig.aleVlanAwareMode  = TRUE;
    pALEConfig->vlanConfig.cpswVlanAwareMode = FALSE;
    pALEConfig->vlanConfig.unknownForceUntaggedEgressMask = 0;
    pALEConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    pALEConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    pALEConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
    pALEConfig->nwSecCfg.enableVid0Mode                        = TRUE;
    pALEConfig->nwSecCfg.hostOuiNoMatchDeny                    = FALSE;
    pALEConfig->nwSecCfg.ipPktCfg.enableDefaultNoFragment      = FALSE;
    pALEConfig->nwSecCfg.ipPktCfg.enableDefaultNxtHdrWhitelist = FALSE;
    pALEConfig->nwSecCfg.macAuthCfg.enableAuthMode             = FALSE;
    pALEConfig->nwSecCfg.malformedPktCfg.enableBadLenPktDrop   = FALSE;
    pALEConfig->nwSecCfg.malformedPktCfg.disableSrcMcastDrop  = FALSE;

    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(pALEConfig->portCfg); i++)
    {
        pALEConfig->portCfg[i].learningCfg.noLearn             = FALSE;
        pALEConfig->portCfg[i].learningCfg.noSaUpdateEnable    = FALSE;
        pALEConfig->portCfg[i].macModeCfg.macOnlyEnable        = FALSE;
        pALEConfig->portCfg[i].pvidCfg.disallowIPFragmentation = FALSE;
        pALEConfig->portCfg[i].pvidCfg.forceUntaggedEgressMask = 0;
        pALEConfig->portCfg[i].pvidCfg.limitIPNxtHdr       = FALSE;
        pALEConfig->portCfg[i].pvidCfg.noLearnMask         = 0;
        pALEConfig->portCfg[i].pvidCfg.regMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
        pALEConfig->portCfg[i].pvidCfg.unregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
        pALEConfig->portCfg[i].pvidCfg.vidIngressCheck     = 0;
        pALEConfig->portCfg[i].pvidCfg.vlanIdInfo.outerVlanFlag = FALSE;
        pALEConfig->portCfg[i].pvidCfg.vlanIdInfo.vlanId   = 0;
        pALEConfig->portCfg[i].pvidCfg.vlanMemberList      = CPSW_ALE_ALL_PORTS_MASK;
    }
}

static int32_t CpswAle_allocPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t *freePolicerEntry)
{
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;
    uint32_t numPolicerEntrys;
    int32_t retVal;
    uint32_t i;

    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);
    for (i = 0; i < numPolicerEntrys; i++)
    {
        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, i, &policerEntry);
        if (policerEntry.validBitmap == 0)
        {
            break;
        }
    }

    if (i < numPolicerEntrys)
    {
        *freePolicerEntry = i;
        retVal = CPSW_SOK;
    }
    else
    {
        retVal = CPSW_EALE_POLICER_TABLE_FULL;
    }

    return retVal;
}

static bool CpswAle_delIPv6HighEntry(CSL_AleRegs *cpswAleRegBase,
                                     uint32_t aleEntryIndex)
{
    bool ipv6EntryDeleted = FALSE;

    if (CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                 aleEntryIndex,
                                 CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) == CSL_ALE_ENTRYTYPE_POLICER)
    {
        if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                            aleEntryIndex,
                                            CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
            CSL_ALE_POLICER_ENTRYTYPE_IPV6)
        {
            CpswAle_clearTableEntry(cpswAleRegBase,
                                    CSL_CPSW_getAleIPv6HighEntryIndex(cpswAleRegBase, aleEntryIndex));
            CpswAle_clearTableEntry(cpswAleRegBase,
                                    CSL_CPSW_getAleIPv6LowEntryIndex(cpswAleRegBase, aleEntryIndex));
            ipv6EntryDeleted = TRUE;
        }
    }

    return ipv6EntryDeleted;
}

static int32_t CpswAle_delPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                       uint32_t policerIndex,
                                       uint32_t delAleEntryMask)
{
    uint32_t numPolicerEntrys;
    int32_t retVal = CPSW_SOK;
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;

    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);

    if (policerIndex >= numPolicerEntrys)
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t aleDelEntryIndex;

        CpswAle_clearSelectedPolicerStats(cpswAleRegBase, policerIndex);

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerIndex, &policerEntry);

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_OUI_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_OUI))
        {
            aleDelEntryIndex = policerEntry.ouiIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_DST_MAC_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST))
        {
            aleDelEntryIndex = policerEntry.dstMacIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_SRC_MAC_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACSRC))
        {
            aleDelEntryIndex = policerEntry.srcMacIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_OVLAN_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_OVLAN))
        {
            aleDelEntryIndex = policerEntry.ovlanIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_VLAN_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_IVLAN))
        {
            aleDelEntryIndex = policerEntry.vlanIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_ETHERTYPE_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_ETHERTYPE))
        {
            aleDelEntryIndex = policerEntry.ethertypeIdx;
            CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_SRC_IP_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPSRC))
        {
            aleDelEntryIndex = policerEntry.srcIpIdx;

            if (CpswAle_delIPv6HighEntry(cpswAleRegBase, aleDelEntryIndex) == FALSE)
            {
                CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
            }
        }

        if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_DST_IP_VALID)
            &&
            (delAleEntryMask & CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPDST))
        {
            aleDelEntryIndex = policerEntry.dstIpIdx;

            if (CpswAle_delIPv6HighEntry(cpswAleRegBase, aleDelEntryIndex) == FALSE)
            {
                CpswAle_clearTableEntry(cpswAleRegBase, aleDelEntryIndex);
            }
        }

        policerEntry.validBitmap = 0;
        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerIndex, &policerEntry);
        CSL_CPSW_disableAlePolicerThread(cpswAleRegBase, policerIndex);
    }

    return retVal;
}

static int32_t CpswAle_delAllPolicerEntries(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t policerIndex;
    uint32_t numPolicerEntrys;
    int32_t retVal = CPSW_SOK;

    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);

    for (policerIndex = 0; policerIndex < numPolicerEntrys; policerIndex++)
    {
        retVal = CpswAle_delPolicerEntry(cpswAleRegBase,
                                         policerIndex,
                                         CPSW_ALE_POLICER_TABLEENTRY_DELETE_ALL);
        if (retVal != CPSW_SOK)
        {
            break;
        }
    }

    return retVal;
}

static int32_t CpswAle_delAllEntries(CSL_AleRegs *cpswAleRegBase)
{
    int32_t retVal = CPSW_SOK;
    uint32_t aleEntryIndex;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (aleEntryIndex = 0; aleEntryIndex < aleTableDepth; aleEntryIndex++)
    {
        if (CSL_CPSW_getALEEntryType(cpswAleRegBase,
                                     aleEntryIndex,
                                     CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) != CSL_ALE_ENTRYTYPE_FREE)
        {
            CpswAle_clearTableEntry(cpswAleRegBase, aleEntryIndex);
        }
    }

    return retVal;
}

static int32_t CpswAle_setPortBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t aleFreqHz,
                                              uint32_t portNum,
                                              bool enableBRateLimitForPort,
                                              uint32_t bcastLimitNumPktsPerSec,
                                              bool enableMRateLimitForPort,
                                              uint32_t mcastLimitNumPktsPerSec)
{
    int32_t retVal = CPSW_SOK;
    CSL_CPSW_ALE_PORTCONTROL alePortControl;

    CSL_CPSW_getAlePortControlReg(cpswAleRegBase, portNum, &alePortControl);
    if (enableBRateLimitForPort)
    {
        alePortControl.bcastLimit = CpswAle_mapNumPktsPerSec2RateLimit(cpswAleRegBase,
                                                                       aleFreqHz,
                                                                       bcastLimitNumPktsPerSec);
    }
    else
    {
        alePortControl.bcastLimit = 0;
    }

    if (enableMRateLimitForPort)
    {
        alePortControl.mcastLimit = CpswAle_mapNumPktsPerSec2RateLimit(cpswAleRegBase,
                                                                       aleFreqHz,
                                                                       mcastLimitNumPktsPerSec);
    }
    else
    {
        alePortControl.mcastLimit = 0;
    }

    CSL_CPSW_setAlePortControlReg(cpswAleRegBase, portNum, &alePortControl);

    return retVal;
}

static int32_t CpswAle_getPortBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                              uint32_t aleFreqHz,
                                              uint32_t portNum,
                                              bool *pBRateLimitEnableForPort,
                                              uint32_t *pBcastLimitNumPktsPerSec,
                                              bool *pMRateLimitEnableForPort,
                                              uint32_t *pMcastLimitNumPktsPerSec)
{
    int32_t retVal = CPSW_SOK;
    CSL_CPSW_ALE_PORTCONTROL alePortControl;

    CSL_CPSW_getAlePortControlReg(cpswAleRegBase, portNum, &alePortControl);
    if (alePortControl.bcastLimit != 0)
    {
        *pBcastLimitNumPktsPerSec = CpswAle_mapRateLimit2NumPktsPerSec(cpswAleRegBase,
                                                                       aleFreqHz,
                                                                       alePortControl.bcastLimit);
        *pBRateLimitEnableForPort = TRUE;
    }
    else
    {
        *pBcastLimitNumPktsPerSec = 0;
        *pBRateLimitEnableForPort = FALSE;
    }

    if (alePortControl.mcastLimit != 0)
    {
        *pMcastLimitNumPktsPerSec = CpswAle_mapRateLimit2NumPktsPerSec(cpswAleRegBase,
                                                                       aleFreqHz,
                                                                       alePortControl.mcastLimit);
        *pMRateLimitEnableForPort = TRUE;
    }
    else
    {
        *pMcastLimitNumPktsPerSec = 0;
        *pMRateLimitEnableForPort = FALSE;
    }

    return retVal;
}

static void CpswAle_dumpPolicerEntries(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t i, freeEntries;
    uint32_t numPolicerEntrys;
    bool policerHit, policerRedHit, policerYellowHit;
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;
    bool trunkIndicator;

    CpswUtils_printf("\n");

    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);

    freeEntries = 0;
    for (i = 0; i < numPolicerEntrys; i++)
    {
        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, i, &policerEntry);
        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_PORT_VALID)
        {
            if (policerEntry.validBitmap &
                CSL_CPSW_ALE_POLICER_PORT_TRUNK_VALID)
            {
                trunkIndicator = TRUE;
            }
            else
            {
                trunkIndicator = FALSE;
            }

            CpswUtils_printf(" %4d: POLICER_PORT, PORT_NUM: %d ISTRUNK:%d\n",
                             i, policerEntry.port, trunkIndicator);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_PRI_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_PRI, PRI:%d",
                             i, policerEntry.pri);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_OUI_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_OUI, ALE Index: %d\n",
                             i, policerEntry.ouiIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_DST_MAC_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_DST_MAC,ALE Index: %d\n",
                             i, policerEntry.dstMacIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_SRC_MAC_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_SRC_MAC,ALE Index: %d\n",
                             i, policerEntry.srcMacIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_OVLAN_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_OVLAN,ALE Index: %d\n",
                             i, policerEntry.ovlanIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_VLAN_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_VLAN,ALE Index: %d\n",
                             i, policerEntry.vlanIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_ETHERTYPE_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_ETHERTYPE,ALE Index: %d\n",
                             i, policerEntry.ethertypeIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_SRC_IP_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_SRC_IP,ALE Index: %d\n",
                             i, policerEntry.srcIpIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_DST_IP_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_DST_IP,ALE Index: %d\n",
                             i, policerEntry.dstIpIdx);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_THREAD_VALID)
        {
            CpswUtils_printf(" %4d: POLICER_THREAD,THREAD ID:%d\n",
                             i, policerEntry.thread);
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_PIR_VALID)
        {
            if (policerEntry.pirIdleIncVal)
            {
                CpswUtils_printf(" %4d: POLICER_PIR_IDL_INC_VAL: %d\n",
                                 i, policerEntry.pirIdleIncVal);
            }
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_CIR_VALID)
        {
            if (policerEntry.cirIdleIncVal)
            {
                CpswUtils_printf(" %4d: POLICER_CIR_IDL_INC_VAL: %d\n",
                                 i, policerEntry.cirIdleIncVal);
            }
        }

        if (policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_EGRESSOP_VALID)
        {
            CpswUtils_printf(" %4d: "
                             "EGRESS_OP: %x \t"
                             "TRUNK_IDX: %d \t"
                             "TTL_CHECK: %d \t"
                             "DEST_PORT_MASK: %x\n",
                             i,
                             policerEntry.egressOp,
                             policerEntry.egressTrunkIndex,
                             policerEntry.enableTTLCheck,
                             policerEntry.destPortMask);
        }

        if (policerEntry.validBitmap == 0)
        {
            freeEntries++;
        }
        else
        {
            CpswAle_getPolicerStats(cpswAleRegBase, i, &policerHit,
                                    &policerRedHit, &policerYellowHit,
                                    FALSE);
            CpswUtils_printf(" %4d: POLICER_STATS: Hit: %d, RedHit: %d, YellowHit: %d\n",
                             i, policerHit, policerRedHit, policerYellowHit);
        }
    }

    CpswUtils_printf("\n\r");
    CpswUtils_printf("%d Free Entries \r\n", freeEntries);
}

static int32_t CpswAle_addPortDefaultVlanId(CpswAle_Obj *pAle)
{
    uint32_t i;
    int32_t ret = CPSW_SOK;
    uint32_t aleEntryIndex;

    for (i = 0; i < CpswAle_getNumPorts(pAle->cpswType); i++)
    {
        /* Also Program VLAN entry settings for Default VID */
        ret = CpswAle_addVlan(pAle->cpswAleRegBase,
                              pAle->pvid[i].vlanIdInfo.vlanId,
                              pAle->pvid[i].vlanIdInfo.outerVlanFlag,
                              pAle->pvid[i].vlanMemberList,
                              pAle->pvid[i].unregMcastFloodMask,
                              pAle->pvid[i].regMcastFloodMask,
                              pAle->pvid[i].forceUntaggedEgressMask,
                              pAle->pvid[i].noLearnMask,
                              pAle->pvid[i].vidIngressCheck,
                              pAle->pvid[i].limitIPNxtHdr,
                              pAle->pvid[i].disallowIPFragmentation,
                              &aleEntryIndex);

        if (ret != CPSW_SOK)
        {
            break;
        }
    }

    return ret;
}

static int32_t CpswAle_addDefaultAleEntries(CpswAle_Obj *pAle)
{
    int32_t ret = CPSW_SOK;

    if (CSL_CPSW_isAleVlanAwareEnabled(pAle->cpswAleRegBase) &&
        pAle->pvidEnabled)
    {
        ret = CpswAle_addPortDefaultVlanId(pAle);
    }

    return ret;
}

static int32_t CpswAle_delOuiAddr(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_OuiAddrType_t ouiAddr)
{
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint32_t aleIndex;

    ret = CpswAle_findOuiAddr(cpswAleRegBase, ouiAddr, &aleIndex);
    if (CPSW_SOK == ret)
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
    }

    return ret;
}

static int32_t CpswAle_delIPv4Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv4AddrType_t ipv4Addr,
                                   uint32_t numLSBIgnore)
{
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint32_t aleIndex;

    ret = CpswAle_findIPv4Addr(cpswAleRegBase, ipv4Addr, numLSBIgnore, &aleIndex);
    if (CPSW_SOK == ret)
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
    }

    return ret;
}

static int32_t CpswAle_delIPv6Addr(CSL_AleRegs *cpswAleRegBase,
                                   CpswAle_Ipv6AddrType_t ipv6Addr,
                                   uint32_t numLSBIgnore)
{
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint32_t aleIndex;

    ret = CpswAle_findIPv6Addr(cpswAleRegBase, ipv6Addr, numLSBIgnore, &aleIndex);
    if (CPSW_SOK == ret)
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
    }

    return ret;
}

static int32_t CpswAle_delEtherType(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t etherType)
{
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint32_t aleIndex;

    ret = CpswAle_findEtherType(cpswAleRegBase, etherType, &aleIndex);
    if (CPSW_SOK == ret)
    {
        CpswAle_clearTableEntry(cpswAleRegBase, aleIndex);
    }

    return ret;
}

static int32_t CpswAle_setPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                   uint32_t peakRateInBitsPerSec,
                                   uint32_t commitRateInBitsPerSec,
                                   uint32_t aleFreqHz,
                                   bool threadIdEnable,
                                   uint32_t threadId,
                                   CpswAle_SetPolicerEntryOutArgs *outArgs)
{
    uint32_t policerEntryIndex = (uint32_t)CPSW_EALE_POLICER_TABLE_FULL;
    int32_t retVal = CPSW_SOK;
    bool policerAlloced;
    uint32_t aleFreeEntryMask;

    policerAlloced = FALSE;
    aleFreeEntryMask = 0;
    if (NOT_ZERO(policerMatchParams->policerMatchEnableMask))
    {
        retVal = CpswAle_allocPolicerEntry(cpswAleRegBase, &policerEntryIndex);
        if (CPSW_SOK == retVal)
        {
            outArgs->policerEntryIndex = policerEntryIndex;
            policerAlloced = TRUE;
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_OUI))
        {
            bool newEntry;
            const CpswAle_OuiEntryInfo *ouiEntryInfo = &policerMatchParams->oui;
            const uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
                CPSW_ALE_OUI_ADDR_INIT_LIST(ouiEntryInfo->ouiAddr);

            retVal = CpswAle_setOuiPolicerEntry(cpswAleRegBase,
                                                policerEntryIndex,
                                                &ouiAddr,
                                                peakRateInBitsPerSec,
                                                commitRateInBitsPerSec,
                                                aleFreqHz,
                                                &outArgs->ouiAleEntryIndex,
                                                &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_OUI;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_IPSRC))
        {
            bool newEntry;

            retVal = CpswAle_setIPSrcDstPolicerEntry(cpswAleRegBase,
                                                     policerEntryIndex,
                                                     &policerMatchParams->srcIp,
                                                     CPSW_ALE_ADDR_TYPE_SRC,
                                                     peakRateInBitsPerSec,
                                                     commitRateInBitsPerSec,
                                                     aleFreqHz,
                                                     &outArgs->srcIpAleEntryIndex,
                                                     &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPSRC;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_IPDST))
        {
            bool newEntry;

            retVal = CpswAle_setIPSrcDstPolicerEntry(cpswAleRegBase,
                                                     policerEntryIndex,
                                                     &policerMatchParams->dstIp,
                                                     CPSW_ALE_ADDR_TYPE_DST,
                                                     peakRateInBitsPerSec,
                                                     commitRateInBitsPerSec,
                                                     aleFreqHz,
                                                     &outArgs->dstIpAleEntryIndex,
                                                     &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPDST;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_ETHERTYPE))
        {
            bool newEntry;

            retVal = CpswAle_setEtherTypePolicerEntry(cpswAleRegBase,
                                                      policerEntryIndex,
                                                      policerMatchParams->etherType.etherType,
                                                      peakRateInBitsPerSec,
                                                      commitRateInBitsPerSec,
                                                      aleFreqHz,
                                                      &outArgs->etherTypeAleEntryIndex,
                                                      &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_ETHERTYPE;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_MACSRC))
        {
            bool newEntry;
            const uint8_t macAddr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                CPSW_ALE_MAC_ADDR_INIT_LIST(policerMatchParams->srcMacAddr.addr.addr);

            retVal = CpswAle_setMacAddrPolicerEntry(cpswAleRegBase,
                                                    policerEntryIndex,
                                                    &macAddr,
                                                    policerMatchParams->srcMacAddr.addr.vlanId,
                                                    policerMatchParams->srcMacAddr.ingressPortNum,
                                                    CPSW_ALE_ADDR_TYPE_SRC,
                                                    peakRateInBitsPerSec,
                                                    commitRateInBitsPerSec,
                                                    aleFreqHz,
                                                    &outArgs->srcMacAleEntryIndex,
                                                    &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACSRC;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_MACDST))
        {
            bool newEntry;
            const uint8_t macAddr[CPSW_ALE_MACADDR_NUM_OCTETS] =
                CPSW_ALE_MAC_ADDR_INIT_LIST(policerMatchParams->dstMacAddr.addr.addr);

            retVal = CpswAle_setMacAddrPolicerEntry(cpswAleRegBase,
                                                    policerEntryIndex,
                                                    &macAddr,
                                                    policerMatchParams->dstMacAddr.addr.vlanId,
                                                    policerMatchParams->dstMacAddr.egressPortNum,
                                                    CPSW_ALE_ADDR_TYPE_DST,
                                                    peakRateInBitsPerSec,
                                                    commitRateInBitsPerSec,
                                                    aleFreqHz,
                                                    &outArgs->dstMacAleEntryIndex,
                                                    &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_IVLAN))
        {
            bool newEntry;

            retVal = CpswAle_setVlanPolicerEntry(cpswAleRegBase,
                                                 policerEntryIndex,
                                                 policerMatchParams->ivlanId,
                                                 FALSE,
                                                 peakRateInBitsPerSec,
                                                 commitRateInBitsPerSec,
                                                 aleFreqHz,
                                                 &outArgs->ivlanAleEntryIndex,
                                                 &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_IVLAN;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_OVLAN))
        {
            bool newEntry;

            retVal = CpswAle_setVlanPolicerEntry(cpswAleRegBase,
                                                 policerEntryIndex,
                                                 policerMatchParams->ovlanId,
                                                 TRUE,
                                                 peakRateInBitsPerSec,
                                                 commitRateInBitsPerSec,
                                                 aleFreqHz,
                                                 &outArgs->ovlanAleEntryIndex,
                                                 &newEntry);
            if (newEntry)
            {
                aleFreeEntryMask |= CPSW_ALE_POLICER_TABLEENTRY_DELETE_OVLAN;
            }
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_PORT))
        {
            retVal = CpswAle_setPortPolicerEntry(cpswAleRegBase,
                                                 policerEntryIndex,
                                                 policerMatchParams->portNum,
                                                 policerMatchParams->portIsTrunk,
                                                 peakRateInBitsPerSec,
                                                 commitRateInBitsPerSec,
                                                 aleFreqHz);
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (NOT_ZERO(policerMatchParams->policerMatchEnableMask &
                     CPSW_ALE_POLICER_MATCH_PRIORITY))
        {
            retVal = CpswAle_setPriorityPolicerEntry(cpswAleRegBase,
                                                     policerEntryIndex,
                                                     policerMatchParams->priority,
                                                     peakRateInBitsPerSec,
                                                     commitRateInBitsPerSec,
                                                     aleFreqHz);
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (threadIdEnable)
        {
            retVal = CpswAle_setThreadPolicerConfig(cpswAleRegBase,
                                                    policerEntryIndex,
                                                    threadId);
        }
    }

    if ((retVal != CPSW_SOK) && (policerAlloced))
    {
        CpswAle_delPolicerEntry(cpswAleRegBase, policerEntryIndex, aleFreeEntryMask);
    }

    return retVal;
}

static int32_t CpswAle_setIPSrcDstPolicerEntry(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t policerEntryIdx,
                                               const CpswAle_IpAddrClassifierInfo *ipClassifier,
                                               CPSW_ALE_ADDR_TYPE addrType,
                                               uint32_t peakRateInBitsPerSec,
                                               uint32_t commitRateInBitsPerSec,
                                               uint32_t aleFreqHz,
                                               uint32_t *aleEntryIndex,
                                               bool *pEntryAlloced)
{
    int32_t retVal = CPSW_SOK;

    if (CPSW_ALE_IPADDR_CLASSIFIER_IPV4 == ipClassifier->ipAddrtype)
    {
        const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
            CPSW_ALE_IPV4_ADDR_INIT_LIST(ipClassifier->ipv4.ipv4Addr);

        retVal = CpswAle_setIPv4PolicerEntry(cpswAleRegBase,
                                             policerEntryIdx,
                                             &ipv4Addr,
                                             addrType,
                                             ipClassifier->ipv4.numLSBIgnoreBits,
                                             peakRateInBitsPerSec,
                                             commitRateInBitsPerSec,
                                             aleFreqHz,
                                             aleEntryIndex,
                                             pEntryAlloced);
    }
    else
    {
        const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
            CPSW_ALE_IPV6_ADDR_INIT_LIST(ipClassifier->ipv6.ipv6Addr);

        retVal = CpswAle_setIPv6PolicerEntry(cpswAleRegBase,
                                             policerEntryIdx,
                                             &ipv6Addr,
                                             addrType,
                                             ipClassifier->ipv6.numLSBIgnoreBits,
                                             peakRateInBitsPerSec,
                                             commitRateInBitsPerSec,
                                             aleFreqHz,
                                             aleEntryIndex,
                                             pEntryAlloced);
    }

    return retVal;
}

static int32_t CpswAle_addMacAddrEntry(CSL_AleRegs *cpswAleRegBase,
                                       CpswAle_MacAddrType_t addr,
                                       uint32_t vlanId,
                                       uint32_t portNum,
                                       CPSW_ALE_ADDR_TYPE addrType,
                                       uint32_t *aleEntryIdx)
{
    int32_t retVal;

    if (CpswAle_isMcastAddr(addr) != TRUE)
    {
        retVal = CpswAle_addUcastAddr(cpswAleRegBase,
                                      portNum,
                                      addr,
                                      vlanId,
                                      CPSW_ALE_UNIADDR_ENTRY_DEFAULT_BLOCKEDFLAG,
                                      CPSW_ALE_UNIADDR_ENTRY_DEFAULT_SECUREFLAG,
                                      CPSW_ALE_UNIADDR_ENTRY_DEFAULT_AGEABLEFLAG,
                                      CPSW_ALE_UNIADDR_ENTRY_DEFAULT_TRUNKFLAG,
                                      aleEntryIdx);
    }
    else
    {
        if (CPSW_ALE_ADDR_TYPE_SRC == addrType)
        {
            /* Multicast cannot be source address */
            retVal = CPSW_EINVALID_PARAMS;
        }
        else
        {
            retVal = CpswAle_addMcastAddr(cpswAleRegBase,
                                          addr,
                                          vlanId,
                                          CPSW_SET_BIT(portNum),
                                          CPSW_ALE_MCASTADDR_ENTRY_DEFAULT_SUPERFLAG,
                                          CPSW_ALE_MCASTADDR_ENTRY_DEFAULT_FWDSTATE,
                                          CPSW_ALE_MCASTADDR_ENTRY_DEFAULT_IGNBITS,
                                          aleEntryIdx);
        }
    }

    return retVal;
}

static int32_t CpswAle_addVlanEntry(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t vlanId,
                                    bool outVlanFlag,
                                    uint32_t *aleEntryIdx)
{
    int32_t retVal;
    uint32_t unknownVlanMemberList, unknownForceUntaggedEgress;
    uint32_t unknownRegMcastFloodMask, unknownMcastFloodMask;

    /* For adding vlan entry without info from application regarding
     * vlan port membership, use unknwon vlan membership mask values
     */
    CSL_CPSW_getAleUnkownVlanReg(cpswAleRegBase,
                                 &unknownVlanMemberList,
                                 &unknownMcastFloodMask,
                                 &unknownRegMcastFloodMask,
                                 &unknownForceUntaggedEgress);

    retVal = CpswAle_addVlan(cpswAleRegBase,
                             vlanId,
                             outVlanFlag,
                             unknownVlanMemberList,
                             unknownMcastFloodMask,
                             unknownRegMcastFloodMask,
                             unknownForceUntaggedEgress,
                             CPSW_ALE_VLAN_ENTRY_DEFAULT_NOLEARNMASK,
                             CPSW_ALE_VLAN_ENTRY_DEFAULT_VIDINGRESSCHK,
                             CPSW_ALE_VLAN_ENTRY_DEFAULT_LIMITIPNXTHDR,
                             CPSW_ALE_VLAN_ENTRY_DEFAULT_DISALLOWFRAGMENTATION,
                             aleEntryIdx);
    return retVal;
}

static bool CpswAle_checkPolicerMatch(CSL_AleRegs *cpswAleRegBase,
                                      CSL_CPSW_ALE_POLICER_ENTRY *policerEntry,
                                      const CpswAle_PolicerMatchParams *policerMatchParams)
{
    bool matchFound = TRUE;
    uint32_t aleEntryIndex;
    uint32_t retVal;

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_PORT)
    {
        if (policerEntry->port != policerMatchParams->portNum)
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_PRIORITY)
    {
        if (policerEntry->pri != policerMatchParams->priority)
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_OUI)
    {
        const uint8_t ouiAddr[CPSW_ALE_OUIADDR_NUM_OCTETS] =
            CPSW_ALE_OUI_ADDR_INIT_LIST(policerMatchParams->oui.ouiAddr);

        retVal = CpswAle_findOuiAddr(cpswAleRegBase,
                                     &ouiAddr,
                                     &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->ouiIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_MACSRC)
    {
        const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
            CPSW_ALE_MAC_ADDR_INIT_LIST(policerMatchParams->srcMacAddr.addr.addr);

        retVal = CpswAle_findAddr(cpswAleRegBase,
                                  &addr,
                                  policerMatchParams->srcMacAddr.addr.vlanId,
                                  &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->srcMacIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_MACDST)
    {
        const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
            CPSW_ALE_MAC_ADDR_INIT_LIST(policerMatchParams->dstMacAddr.addr.addr);

        retVal = CpswAle_findAddr(cpswAleRegBase,
                                  &addr,
                                  policerMatchParams->dstMacAddr.addr.vlanId,
                                  &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->dstMacIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IVLAN)
    {
        retVal = CpswAle_findVlan(cpswAleRegBase,
                                  policerMatchParams->ivlanId,
                                  FALSE,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->vlanIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_OVLAN)
    {
        retVal = CpswAle_findVlan(cpswAleRegBase,
                                  policerMatchParams->ovlanId,
                                  TRUE,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->ovlanIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_ETHERTYPE)
    {
        retVal = CpswAle_findEtherType(cpswAleRegBase,
                                       policerMatchParams->etherType.etherType,
                                       &aleEntryIndex);
        if ((retVal != CPSW_SOK) ||
            (policerEntry->ethertypeIdx != aleEntryIndex))
        {
            matchFound = FALSE;
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IPSRC)
    {
        if (policerMatchParams->srcIp.ipAddrtype ==
            CPSW_ALE_IPADDR_CLASSIFIER_IPV4)
        {
            const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV4_ADDR_INIT_LIST(policerMatchParams->srcIp.ipv4.ipv4Addr);

            retVal = CpswAle_findIPv4Addr(cpswAleRegBase,
                                          &ipv4Addr,
                                          policerMatchParams->srcIp.ipv4.numLSBIgnoreBits,
                                          &aleEntryIndex);
            if ((retVal != CPSW_SOK) ||
                (policerEntry->srcIpIdx != aleEntryIndex))
            {
                matchFound = FALSE;
            }
        }
        else
        {
            const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV6_ADDR_INIT_LIST(policerMatchParams->srcIp.ipv6.ipv6Addr);

            retVal = CpswAle_findIPv6Addr(cpswAleRegBase,
                                          &ipv6Addr,
                                          policerMatchParams->srcIp.ipv6.numLSBIgnoreBits,
                                          &aleEntryIndex);
            if ((retVal != CPSW_SOK) ||
                (policerEntry->srcIpIdx != aleEntryIndex))
            {
                matchFound = FALSE;
            }
        }
    }

    if (policerMatchParams->policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IPDST)
    {
        if (policerMatchParams->dstIp.ipAddrtype ==
            CPSW_ALE_IPADDR_CLASSIFIER_IPV4)
        {
            const uint8_t ipv4Addr[CPSW_ALE_IPV4ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV4_ADDR_INIT_LIST(policerMatchParams->dstIp.ipv4.ipv4Addr);

            retVal = CpswAle_findIPv4Addr(cpswAleRegBase,
                                          &ipv4Addr,
                                          policerMatchParams->dstIp.ipv4.numLSBIgnoreBits,
                                          &aleEntryIndex);
            if ((retVal != CPSW_SOK) ||
                (policerEntry->dstIpIdx != aleEntryIndex))
            {
                matchFound = FALSE;
            }
        }
        else
        {
            const uint8_t ipv6Addr[CPSW_ALE_IPV6ADDR_NUM_OCTETS] =
                CPSW_ALE_IPV6_ADDR_INIT_LIST(policerMatchParams->dstIp.ipv6.ipv6Addr);

            retVal = CpswAle_findIPv6Addr(cpswAleRegBase,
                                          &ipv6Addr,
                                          policerMatchParams->dstIp.ipv6.numLSBIgnoreBits,
                                          &aleEntryIndex);
            if ((retVal != CPSW_SOK) ||
                (policerEntry->dstIpIdx != aleEntryIndex))
            {
                matchFound = FALSE;
            }
        }
    }

    return matchFound;
}

static uint32_t CpswAle_mapPolicer2CslMask(uint32_t policerMatchEnableMask)
{
    uint32_t matchBitmap;

    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_PORT == CSL_CPSW_ALE_POLICER_PORT_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_PRIORITY == CSL_CPSW_ALE_POLICER_PRI_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_OUI == CSL_CPSW_ALE_POLICER_OUI_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_MACSRC == CSL_CPSW_ALE_POLICER_SRC_MAC_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_MACDST == CSL_CPSW_ALE_POLICER_DST_MAC_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_IVLAN == CSL_CPSW_ALE_POLICER_VLAN_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_OVLAN == CSL_CPSW_ALE_POLICER_OVLAN_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_ETHERTYPE == CSL_CPSW_ALE_POLICER_ETHERTYPE_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_IPSRC == CSL_CPSW_ALE_POLICER_SRC_IP_VALID);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_MATCH_IPDST == CSL_CPSW_ALE_POLICER_DST_IP_VALID);

    matchBitmap = 0;

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_PORT))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_PORT_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_PRIORITY))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_PRI_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_OUI))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_OUI_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_MACSRC))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_SRC_MAC_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_MACDST))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_DST_MAC_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IVLAN))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_VLAN_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_OVLAN))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_OVLAN_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_ETHERTYPE))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_ETHERTYPE_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IPSRC))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_SRC_IP_VALID;
    }

    if (NOT_ZERO(policerMatchEnableMask & CPSW_ALE_POLICER_MATCH_IPDST))
    {
        matchBitmap |= CSL_CPSW_ALE_POLICER_DST_IP_VALID;
    }

    return matchBitmap;
}

static int32_t CpswAle_getPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                  CpswAle_GetPolicerEntryOutArgs *outArgs,
                                  uint32_t aleFreqHz)
{
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;
    int32_t retVal;
    uint32_t policerIndex;

    retVal = CpswAle_getPolicerIndex(cpswAleRegBase, policerMatchParams, &policerIndex);

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerIndex, &policerEntry);

        /* Do not set valid bitmask to CSL valid bitmask as it is not understood
         * by application */
        outArgs->policerMatchEnableMask = policerMatchParams->policerMatchEnableMask;
        outArgs->policerEntryIndex      = policerIndex;
        outArgs->port                   = policerEntry.port;
        outArgs->portIsTrunk            = NOT_ZERO(policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_PORT_TRUNK_VALID);
        outArgs->priority               = policerEntry.pri;
        outArgs->ouiAleEntryIndex       = policerEntry.ouiIdx;
        outArgs->etherTypeAleEntryIndex = policerEntry.ethertypeIdx;
        outArgs->ivlanAleEntryIndex     = policerEntry.vlanIdx;
        outArgs->ovlanAleEntryIndex     = policerEntry.ovlanIdx;
        outArgs->srcMacAleEntryIndex    = policerEntry.srcMacIdx;
        outArgs->dstMacAleEntryIndex    = policerEntry.dstMacIdx;
        outArgs->srcIpAleEntryIndex     = policerEntry.srcIpIdx;
        outArgs->dstIpAleEntryIndex     = policerEntry.dstIpIdx;
        outArgs->threadId               = policerEntry.thread;
        outArgs->threadIdEnable         = NOT_ZERO(policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_THREAD_VALID);
        outArgs->commitRateInBitsPerSec = CpswAle_mapIdlIncVal2Bw(aleFreqHz, policerEntry.cirIdleIncVal);
        outArgs->peakRateInBitsPerSec   = CpswAle_mapIdlIncVal2Bw(aleFreqHz, policerEntry.pirIdleIncVal);
        outArgs->egressOpEnabled        = (policerEntry.egressOp != 0);
        outArgs->egressOpcode           = policerEntry.egressOp;
        outArgs->egressTrunkIndex       = policerEntry.egressTrunkIndex;
        outArgs->enableTTLCheck         = policerEntry.enableTTLCheck;
        outArgs->destPortMask           = policerEntry.destPortMask;
    }

    return retVal;
}

static int32_t CpswAle_getPolicerIndex(CSL_AleRegs *cpswAleRegBase,
                                       const CpswAle_PolicerMatchParams *policerMatchParams,
                                       uint32_t *policerIndex)
{
    uint32_t numPolicerEntrys;
    int32_t retVal;
    uint32_t i;
    uint32_t matchBitmap;
    bool matchFound;
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;

    matchBitmap = CpswAle_mapPolicer2CslMask(policerMatchParams->policerMatchEnableMask);
    CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);
    for (i = 0; i < numPolicerEntrys; i++)
    {
        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, i, &policerEntry);
        if ((policerEntry.validBitmap & matchBitmap) == matchBitmap)
        {
            matchFound = CpswAle_checkPolicerMatch(cpswAleRegBase,
                                                   &policerEntry,
                                                   policerMatchParams);
            if (matchFound)
            {
                break;
            }
        }
    }

    if (i < numPolicerEntrys)
    {
        *policerIndex = i;
        retVal = CPSW_SOK;
    }
    else
    {
        retVal = CPSW_EFAIL;
    }

    return retVal;
}

static int32_t CpswAle_delPolicer(CSL_AleRegs *cpswAleRegBase,
                                  const CpswAle_PolicerMatchParams *policerMatchParams,
                                  uint32_t delAleEntryMask)
{
    uint32_t policerEntryIndex;
    int32_t retVal;

    retVal = CpswAle_getPolicerIndex(cpswAleRegBase,
                                     policerMatchParams,
                                     &policerEntryIndex);

    if (CPSW_SOK == retVal)
    {
        CpswAle_delPolicerEntry(cpswAleRegBase, policerEntryIndex, delAleEntryMask);
    }

    return retVal;
}

static int32_t CpswAle_delAllPolicerThreadId(CSL_AleRegs *cpswAleRegBase,
                                             uint32_t threadId)
{
    uint32_t numPolicerEntrys;
    uint32_t i;
    CSL_CPSW_ALE_POLICER_ENTRY policerEntry;
    int32_t retVal = CPSW_SOK;

    if (threadId > CSL_ALE_THREADMAPVAL_THREADVAL_MAX)
    {
        retVal = CPSW_EINVALID_PARAMS;
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_getAleStatusNumPolicers(cpswAleRegBase, &numPolicerEntrys);
        for (i = 0; i < numPolicerEntrys; i++)
        {
            CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, i, &policerEntry);
            if ((policerEntry.validBitmap & CSL_CPSW_ALE_POLICER_THREAD_VALID)
                &&
                (policerEntry.thread == threadId))
            {
                CpswAle_delPolicerEntry(cpswAleRegBase, i, CPSW_ALE_POLICER_TABLEENTRY_DELETE_ALL);
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_getPolicerStatsHandler(CSL_AleRegs *cpswAleRegBase,
                                              const CpswAle_PolicerMatchParams *policerMatchParams,
                                              bool *pPolHit,
                                              bool *pPolRedHit,
                                              bool *pPolYellowHit,
                                              bool clearStats)
{
    uint32_t policerIndex;
    int32_t retVal;

    retVal = CpswAle_getPolicerIndex(cpswAleRegBase,
                                     policerMatchParams,
                                     &policerIndex);
    if (CPSW_SOK == retVal)
    {
        CpswAle_getPolicerStats(cpswAleRegBase,
                                policerIndex,
                                pPolHit,
                                pPolRedHit,
                                pPolYellowHit,
                                clearStats);
    }

    return retVal;
}

static int32_t CpswAle_setOAMLpbk(CSL_AleRegs *cpswAleRegBase,
                                  uint32_t enableLpbkPortMask)
{
    CSL_CPSW_setAleOAMLpbkControl(cpswAleRegBase,
                                  enableLpbkPortMask);
    return CPSW_SOK;
}

static void CpswAle_updateMcastHostPortMask(CSL_AleRegs *cpswAleRegBase,
                                            bool addHostPort)
{
    uint32_t i;
    uint8_t aleEntryMacAddr[6];
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;
        uint32_t reg0, reg1, reg2;

        /* read from the ALE table */
        CSL_CPSW_getAleTableEntry(cpswAleRegBase, i, &reg0, &reg1, &reg2);

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if ((value == CSL_ALE_ENTRYTYPE_ADDRESS) ||
            (value == CSL_ALE_ENTRYTYPE_VLANADDRESS))
        {
            CSL_CPSW_mapTableWord2MacAddr(reg0, reg1, aleEntryMacAddr, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            if (CpswAle_isMcastAddr((CpswAle_MacAddrType_t)&aleEntryMacAddr))
            {
                if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
                {
                    CSL_CPSW_ALE_MCASTADDR_ENTRY entry;

                    CSL_CPSW_getAleMcastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (addHostPort)
                    {
                        entry.portMask |= CPSW_ALE_HOST_PORT_MASK;
                    }
                    else
                    {
                        entry.portMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                    }

                    CSL_CPSW_setAleMcastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                }
                else
                {
                    CSL_CPSW_ALE_VLANMCASTADDR_ENTRY entry;

                    CSL_CPSW_getAleVlanMcastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (addHostPort)
                    {
                        entry.portMask |= CPSW_ALE_HOST_PORT_MASK;
                    }
                    else
                    {
                        entry.portMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                    }

                    CSL_CPSW_setAleVlanMcastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                }
            }
        }
    }
}

static void CpswAle_getVlanMcastPortMask(CSL_AleRegs *cpswAleRegBase,
                                         const CSL_CPSW_ALE_VLAN_ENTRY *vlanEntry,
                                         uint32_t *pRegMcastFloodMask,
                                         uint32_t *pUnregMcastFloodMask)
{
    if (CpswAle_isVlanMaskMuxRegSupported(cpswAleRegBase))
    {
        CSL_CPSW_getAleVlanMaskMuxEntryReg(cpswAleRegBase,
                                           vlanEntry->unRegMcastFloodIndex,
                                           pUnregMcastFloodMask);
        CSL_CPSW_getAleVlanMaskMuxEntryReg(cpswAleRegBase,
                                           vlanEntry->regMcastFloodIndex,
                                           pRegMcastFloodMask);
    }
    else
    {
        *pUnregMcastFloodMask = vlanEntry->unRegMcastFloodMask;
        *pRegMcastFloodMask = vlanEntry->regMcastFloodMask;
    }
}

static int32_t CpswAle_setVlanMcastPortMask(CSL_AleRegs *cpswAleRegBase,
                                            CSL_CPSW_ALE_VLAN_ENTRY *vlanEntry,
                                            uint32_t aleIndex,
                                            bool outerVlanFlag,
                                            uint32_t regMcastFloodMask,
                                            uint32_t unregMcastFloodMask)
{
    int32_t retVal = CPSW_SOK;

    if (CpswAle_isVlanMaskMuxRegSupported(cpswAleRegBase))
    {
        retVal = CpswAle_addVlanMask(cpswAleRegBase,
                                     regMcastFloodMask,
                                     &vlanEntry->regMcastFloodIndex);
        if (CPSW_SOK == retVal)
        {
            retVal = CpswAle_addVlanMask(cpswAleRegBase,
                                         unregMcastFloodMask,
                                         &vlanEntry->unRegMcastFloodIndex);
        }
    }
    else
    {
        vlanEntry->unRegMcastFloodMask = unregMcastFloodMask;
        vlanEntry->regMcastFloodMask = regMcastFloodMask;
    }

    if (CPSW_SOK == retVal)
    {
        if (outerVlanFlag)
        {
            CSL_CPSW_setAleOutVlanEntry(cpswAleRegBase,
                                        aleIndex,
                                        vlanEntry,
                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }
        else
        {
            CSL_CPSW_setAleVlanEntry(cpswAleRegBase,
                                     aleIndex,
                                     vlanEntry,
                                     CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }
    }

    return retVal;
}

static int32_t CpswAle_updateVlanMcastHostPortMask(CSL_AleRegs *cpswAleRegBase,
                                                   bool updateRegMcast,
                                                   bool updateUnregMcast,
                                                   bool addHostPort)
{
    uint32_t i;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);
    uint32_t regMcastFloodMask, unregMcastFloodMask;
    int32_t retVal = CPSW_SOK;

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_VLAN)
        {
            if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                i,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                CSL_ALE_POLICER_ENTRYTYPE_OVLAN)
            {
                CSL_CPSW_ALE_OUTER_VLAN_ENTRY oVlanEntry;

                CSL_CPSW_getAleOutVlanEntry(cpswAleRegBase, i, &oVlanEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                if (NOT_ZERO(oVlanEntry.vlanMemList & CPSW_ALE_HOST_PORT_MASK))
                {
                    CpswAle_getVlanMcastPortMask(cpswAleRegBase, &oVlanEntry,
                                                 &regMcastFloodMask,
                                                 &unregMcastFloodMask);
                    if (updateUnregMcast)
                    {
                        if (addHostPort)
                        {
                            unregMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
                        }
                        else
                        {
                            unregMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                        }
                    }

                    if (updateRegMcast)
                    {
                        if (addHostPort)
                        {
                            regMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
                        }
                        else
                        {
                            regMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                        }
                    }

                    retVal = CpswAle_setVlanMcastPortMask(cpswAleRegBase,
                                                          &oVlanEntry,
                                                          i,
                                                          TRUE,
                                                          regMcastFloodMask,
                                                          unregMcastFloodMask);
                    if (retVal != CPSW_SOK)
                    {
                        break;
                    }
                }
            }

            if (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase,
                                                i,
                                                CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)) ==
                CSL_ALE_POLICER_ENTRYTYPE_VLAN)
            {
                CSL_CPSW_ALE_VLAN_ENTRY vlanEntry;

                CSL_CPSW_getAleVlanEntry(cpswAleRegBase, i, &vlanEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                if (NOT_ZERO(vlanEntry.vlanMemList & CPSW_ALE_HOST_PORT_MASK))
                {
                    CpswAle_getVlanMcastPortMask(cpswAleRegBase,
                                                 &vlanEntry,
                                                 &regMcastFloodMask,
                                                 &unregMcastFloodMask);
                    if (updateUnregMcast)
                    {
                        if (addHostPort)
                        {
                            unregMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
                        }
                        else
                        {
                            unregMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                        }
                    }

                    if (updateRegMcast)
                    {
                        if (addHostPort)
                        {
                            regMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
                        }
                        else
                        {
                            regMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
                        }
                    }

                    retVal = CpswAle_setVlanMcastPortMask(cpswAleRegBase,
                                                          &vlanEntry,
                                                          i,
                                                          FALSE,
                                                          regMcastFloodMask,
                                                          unregMcastFloodMask);
                    if (retVal != CPSW_SOK)
                    {
                        break;
                    }
                }
            }
        }
    } /* end of for i loop */

    return retVal;
}

static int32_t CpswAle_delPortUcastEntries(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t portNum)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint8_t aleEntryMacAddr[6];
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;
        uint32_t reg0, reg1, reg2;

        /* read from the ALE table */
        CSL_CPSW_getAleTableEntry(cpswAleRegBase, i, &reg0, &reg1, &reg2);

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if ((value == CSL_ALE_ENTRYTYPE_ADDRESS) ||
            (value == CSL_ALE_ENTRYTYPE_VLANADDRESS))
        {
            CSL_CPSW_mapTableWord2MacAddr(reg0, reg1, aleEntryMacAddr, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            if (!CpswAle_isMcastAddr((CpswAle_MacAddrType_t)&aleEntryMacAddr))
            {
                if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
                {
                    CSL_CPSW_ALE_UNICASTADDR_ENTRY entry;

                    CSL_CPSW_getAleUnicastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (portNum == entry.portNumber)
                    {
                        /* Clear ALE table unicast entry for the given port
                         * only if following conditions are met
                         * (ageable == TRUE)  : Dont delete static entries
                         * (secure  == FALSE) : Dont delete secured address
                         * (block  == FALSE) : Dont delete blocked address
                         * (trunk == FALSE)  : Entry is not part of trunk
                         */
                        if ((entry.ageable == TRUE) &&
                            (entry.secureEnable == FALSE) &&
                            (entry.blockEnable == FALSE) &&
                            (entry.trunkFlag == FALSE))
                        {
                            CpswAle_clearTableEntry(cpswAleRegBase, i);
                        }
                    }

                    CSL_CPSW_setAleUnicastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                }
                else
                {
                    CSL_CPSW_ALE_VLANUNICASTADDR_ENTRY entry;

                    CSL_CPSW_getAleVlanUnicastAddrEntry(cpswAleRegBase, i,
                                                        &entry,
                                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (portNum == entry.portNumber)
                    {
                        if ((entry.ageable == TRUE) &&
                            (entry.secureEnable == FALSE) &&
                            (entry.blockEnable == FALSE) &&
                            (entry.trunkFlag == FALSE))
                        {
                            CpswAle_clearTableEntry(cpswAleRegBase, i);
                        }
                    }

                    CSL_CPSW_setAleVlanUnicastAddrEntry(cpswAleRegBase, i,
                                                        &entry,
                                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                }
            }
        }
    }

    return ret;
}

static void CpswAle_updateUnknownVlanMcastHostPortMask(CSL_AleRegs *cpswAleRegBase,
                                                       bool updateRegMcast,
                                                       bool updateUnregMcast,
                                                       bool addHostPort)
{
    uint32_t unknownVlanMemList, unknownForceUntaggedEgress;
    uint32_t unknownVlanUnRegMcastFloodMask, unknownVlanRegMcastFloodMask;

    CSL_CPSW_getAleUnkownVlanReg(cpswAleRegBase,
                                 &unknownVlanMemList,
                                 &unknownVlanUnRegMcastFloodMask,
                                 &unknownVlanRegMcastFloodMask,
                                 &unknownForceUntaggedEgress);
    if (updateRegMcast)
    {
        if (addHostPort)
        {
            unknownVlanRegMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
            CSL_CPSW_setAleUnknwnVlanRegMcastReg(cpswAleRegBase,
                                                 unknownVlanRegMcastFloodMask);
        }
        else
        {
            unknownVlanRegMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
            CSL_CPSW_setAleUnknwnVlanRegMcastReg(cpswAleRegBase,
                                                 unknownVlanRegMcastFloodMask);
        }
    }

    if (updateUnregMcast)
    {
        if (addHostPort)
        {
            unknownVlanUnRegMcastFloodMask |= CPSW_ALE_HOST_PORT_MASK;
            CSL_CPSW_setAleUnknwnVlanUnregMcastReg(cpswAleRegBase,
                                                   unknownVlanUnRegMcastFloodMask);
        }
        else
        {
            unknownVlanUnRegMcastFloodMask &= CPSW_ALE_ALL_MACPORTS_MASK;
            CSL_CPSW_setAleUnknwnVlanUnregMcastReg(cpswAleRegBase,
                                                   unknownVlanUnRegMcastFloodMask);
        }
    }
}

static int32_t CpswAle_setRxFilterAll(CSL_AleRegs *cpswAleRegBase)
{
    if (NOT_ZERO(CSL_CPSW_isAleBypassEnabled(cpswAleRegBase)) == FALSE)
    {
        /* Set ALE in bypass */
        CSL_CPSW_enableAleBypass(cpswAleRegBase);
    }

    /* Set all ports in MAC only mode */

    /* Set all ports in MAC CAF enable */

    /* Disable learning on all ports */

    /* Clear all ALE table entries */

    /* Set ALE port state to forward for host port */
    CpswAle_setAlePortState(cpswAleRegBase,
                            CPSW_ALE_HOST_PORT_NUM,
                            CSL_ALE_PORTSTATE_FORWARD);

    return CPSW_SOK;
}

static int32_t CpswAle_setRxFilterAllMcast(CSL_AleRegs *cpswAleRegBase)
{
    if (CSL_CPSW_isAleBypassEnabled(cpswAleRegBase))
    {
        /* Clear ALE bypass */
        CSL_CPSW_disableAleBypass(cpswAleRegBase);
    }

    if (CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase))
    {
        /* Enable Host port for Unknown VLAN,reg and unreg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, TRUE);
        /* Enable Host port for Known VLAN,reg and unreg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, TRUE);
    }

    CpswAle_updateMcastHostPortMask(cpswAleRegBase, TRUE);
    CpswAle_setAlePortState(cpswAleRegBase,
                            CPSW_ALE_HOST_PORT_NUM,
                            CSL_ALE_PORTSTATE_FORWARD);

    return CPSW_SOK;
}

static int32_t CpswAle_setRxFilterMcast(CSL_AleRegs *cpswAleRegBase)
{
    if (CSL_CPSW_isAleBypassEnabled(cpswAleRegBase))
    {
        /* Clear ALE bypass */
        CSL_CPSW_disableAleBypass(cpswAleRegBase);
    }

    if (CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase))
    {
        /* Enable Host port for Unknown VLAN reg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, TRUE, FALSE, TRUE);
        /* Disable Host port for Unknown VLAN unreg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, FALSE, TRUE, FALSE);
        /* Enable Host port for Known VLAN,reg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, TRUE, FALSE, TRUE);
        /* Disable Host port for Known VLAN,unreg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, FALSE, TRUE, FALSE);
    }

    CpswAle_updateMcastHostPortMask(cpswAleRegBase, TRUE);
    CpswAle_setAlePortState(cpswAleRegBase,
                            CPSW_ALE_HOST_PORT_NUM,
                            CSL_ALE_PORTSTATE_FORWARD);

    return CPSW_SOK;
}

static int32_t CpswAle_setRxFilterBcast(CSL_AleRegs *cpswAleRegBase)
{
    const uint8_t addr[CPSW_ALE_MACADDR_NUM_OCTETS] =
        CPSW_ALE_BCAST_ADDR_INIT_LIST;
    uint32_t aleEntryIndex;
    int32_t retVal = CPSW_SOK;

    if (CSL_CPSW_isAleBypassEnabled(cpswAleRegBase))
    {
        /* Clear ALE bypass */
        CSL_CPSW_disableAleBypass(cpswAleRegBase);
    }

    if (CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase))
    {
        /* Enable Host port for Unknown VLAN reg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, TRUE, FALSE, TRUE);
        /* Disable Host port for Unknown VLAN unreg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, FALSE, TRUE, FALSE);
        /* Enable Host port for Known VLAN,reg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, TRUE, FALSE, TRUE);
        /* Disable Host port for Known VLAN,unreg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, FALSE, TRUE, FALSE);
    }

    /* Disable host port mask as we want to receive only broadcast */
    CpswAle_updateMcastHostPortMask(cpswAleRegBase, FALSE);

    retVal = CpswAle_findAddr(cpswAleRegBase,
                              &addr,
                              0,
                              &aleEntryIndex);
    if (CPSW_EALE_ADDRESS_NOT_FOUND == retVal)
    {
        retVal = CpswAle_addMcastAddr(cpswAleRegBase,
                                      &addr,
                                      0,
                                      CPSW_ALE_HOST_PORT_MASK,
                                      0,
                                      CPSW_ALE_FWDSTLVL_FORWARDING,
                                      0,
                                      &aleEntryIndex);
    }

    if (CPSW_SOK == retVal)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, aleEntryIndex, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
        {
            CSL_CPSW_ALE_MCASTADDR_ENTRY entry;

            CSL_CPSW_getAleMcastAddrEntry(cpswAleRegBase, aleEntryIndex, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            entry.portMask |= CPSW_ALE_HOST_PORT_MASK;
            CSL_CPSW_setAleMcastAddrEntry(cpswAleRegBase, aleEntryIndex, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }
        else
        {
            CSL_CPSW_ALE_VLANMCASTADDR_ENTRY entry;

            CSL_CPSW_getAleVlanMcastAddrEntry(cpswAleRegBase, aleEntryIndex,
                                              &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            entry.portMask |= CPSW_ALE_HOST_PORT_MASK;
            CSL_CPSW_setAleVlanMcastAddrEntry(cpswAleRegBase, aleEntryIndex,
                                              &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        }
    }

    if (CPSW_SOK == retVal)
    {
        CpswAle_setAlePortState(cpswAleRegBase,
                                CPSW_ALE_HOST_PORT_NUM,
                                CSL_ALE_PORTSTATE_FORWARD);
    }

    return retVal;
}

static int32_t CpswAle_setRxFilterDirect(CSL_AleRegs *cpswAleRegBase)
{
    if (CSL_CPSW_isAleBypassEnabled(cpswAleRegBase))
    {
        /* Clear ALE bypass */
        CSL_CPSW_disableAleBypass(cpswAleRegBase);
    }

    if (CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase))
    {
        /* Disable Host port for Unknown VLAN unreg/reg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, FALSE);
        /* Disable Host port for Known VLAN,reg,unreg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, FALSE);
    }

    /* Disable host port mask for all multicast addresses */
    CpswAle_updateMcastHostPortMask(cpswAleRegBase, FALSE);

    CpswAle_setAlePortState(cpswAleRegBase,
                            CPSW_ALE_HOST_PORT_NUM,
                            CSL_ALE_PORTSTATE_FORWARD);

    return CPSW_SOK;
}

static int32_t CpswAle_setRxFilterNone(CSL_AleRegs *cpswAleRegBase)
{
    if (CSL_CPSW_isAleBypassEnabled(cpswAleRegBase))
    {
        /* Clear ALE bypass */
        CSL_CPSW_disableAleBypass(cpswAleRegBase);
    }

    if (CSL_CPSW_isAleVlanAwareEnabled(cpswAleRegBase))
    {
        /* Disable Host port for Unknown VLAN unreg/reg mcast */
        CpswAle_updateUnknownVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, FALSE);
        /* Disable Host port for Known VLAN,reg,unreg mcast */
        CpswAle_updateVlanMcastHostPortMask(cpswAleRegBase, TRUE, TRUE, FALSE);
    }

    /* Disable host port mask for all multicast addresses */
    CpswAle_updateMcastHostPortMask(cpswAleRegBase, FALSE);
    /* Delete all unicast addresses with host port as port Number */
    CpswAle_delPortUcastEntries(cpswAleRegBase, CPSW_ALE_HOST_PORT_NUM);

    /* When Rx filter is none , set ALE host port state to disabled
     * to block all traffic to host port
     */
    CpswAle_setAlePortState(cpswAleRegBase,
                            CPSW_ALE_HOST_PORT_NUM,
                            CSL_ALE_PORTSTATE_DISABLED);
    return CPSW_SOK;
}

static int32_t CpswAle_setBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          const CpswAle_SetBcastMcastRateLimitInArgs *inArgs)
{
    uint32_t i;
    int32_t retVal = CPSW_SOK;

    if (inArgs->rateLimitAtTxPort)
    {
        CSL_CPSW_enableAleTxRateLimit(cpswAleRegBase);
    }
    else
    {
        CSL_CPSW_disableAleTxRateLimit(cpswAleRegBase);
    }

    for (i = 0; i < inArgs->numPorts; i++)
    {
        retVal = CpswAle_setPortBcastMcastLimit(cpswAleRegBase,
                                                aleFreqHz,
                                                inArgs->portPrms[i].portNum,
                                                inArgs->portPrms[i].enableBRateLimitForPort,
                                                inArgs->portPrms[i].bcastLimitNumPktsPerSec,
                                                inArgs->portPrms[i].enableMRateLimitForPort,
                                                inArgs->portPrms[i].mcastLimitNumPktsPerSec);
        if (retVal != CPSW_SOK)
        {
            break;
        }
    }

    CSL_CPSW_enableAleRateLimit(cpswAleRegBase);

    return retVal;
}

static int32_t CpswAle_getBcastMcastLimit(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          CpswAle_GetBcastMcastRateLimitOutArgs *outArgs)
{
    uint32_t i;
    int32_t retVal = CPSW_SOK;
    uint32_t numPorts = CpswAle_getNumPortsFromCpswAleRegBase((uintptr_t)cpswAleRegBase);

    outArgs->rateLimitEnabled = CSL_CPSW_isAleRateLimitEnabled(cpswAleRegBase);
    outArgs->rateLimitAtTxPort = CSL_CPSW_isAleTxRateLimitEnabled(cpswAleRegBase);
    outArgs->numPorts = 0;

    if (outArgs->rateLimitEnabled)
    {
        for (i = 0; i < numPorts; i++)
        {
            bool enableBRateLimitForPort, enableMRateLimitForPort;
            uint32_t bcastLimitNumPktsPerSec, mcastLimitNumPktsPerSec;

            retVal = CpswAle_getPortBcastMcastLimit(cpswAleRegBase,
                                                    aleFreqHz,
                                                    i,
                                                    &enableBRateLimitForPort,
                                                    &bcastLimitNumPktsPerSec,
                                                    &enableMRateLimitForPort,
                                                    &mcastLimitNumPktsPerSec);
            if (retVal != CPSW_SOK)
            {
                break;
            }
            else
            {
                if ((enableBRateLimitForPort == true) || (enableMRateLimitForPort == true))
                {
                    Cpsw_assert(outArgs->numPorts < CPSW_UTILS_ARRAYSIZE(outArgs->portPrms));
                    outArgs->portPrms[outArgs->numPorts].portNum = i;
                    outArgs->portPrms[outArgs->numPorts].enableBRateLimitForPort = enableBRateLimitForPort;
                    outArgs->portPrms[outArgs->numPorts].enableMRateLimitForPort = enableMRateLimitForPort;
                    outArgs->portPrms[outArgs->numPorts].bcastLimitNumPktsPerSec = bcastLimitNumPktsPerSec;
                    outArgs->portPrms[outArgs->numPorts].mcastLimitNumPktsPerSec = mcastLimitNumPktsPerSec;
                    outArgs->numPorts++;
                }
            }
        }
    }

    return retVal;
}

static int32_t CpswAle_setInterVlanConfig(CSL_AleRegs *cpswAleRegBase,
                                          uint32_t aleFreqHz,
                                          const CpswAle_SetInterVLANConfigInArgs *inArgs,
                                          CpswAle_SetInterVLANConfigOutArgs *outArgs)
{
    int32_t retVal;
    uint32_t policerIndex;

    retVal = CpswAle_getPolicerIndex(cpswAleRegBase,
                                     &inArgs->policerMatch,
                                     &policerIndex);

    if (retVal != CPSW_SOK)
    {
        CpswAle_SetPolicerEntryOutArgs policerSetOutArgs;

        retVal = CpswAle_setPolicer(cpswAleRegBase,
                                    &inArgs->policerMatch,
                                    CPSW_ALE_PEAKBITRATE_DISABLE,
                                    CPSW_ALE_COMMITBITRATE_DISABLE,
                                    aleFreqHz,
                                    FALSE,
                                    CPSW_ALE_THREADID_INVALID,
                                    &policerSetOutArgs);
        if (CPSW_SOK == retVal)
        {
            policerIndex = policerSetOutArgs.policerEntryIndex;
        }
    }

    if (CPSW_SOK == retVal)
    {
        CSL_CPSW_ALE_POLICER_ENTRY polCfg;

        CSL_CPSW_getAlePolicerEntry(cpswAleRegBase, policerIndex, &polCfg);
        polCfg.validBitmap |= CSL_CPSW_ALE_POLICER_EGRESSOP_VALID;
        polCfg.egressOp = inArgs->routeIndex;
        polCfg.egressTrunkIndex = inArgs->egressTrunkIndex;
        polCfg.enableTTLCheck = inArgs->enableTTLCheck;
        polCfg.destPortMask = inArgs->destPortMask;
        CSL_CPSW_setAlePolicerEntry(cpswAleRegBase, policerIndex, &polCfg);
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_getPolicer(cpswAleRegBase,
                                    &inArgs->policerMatch,
                                    outArgs,
                                    aleFreqHz);
    }

    return retVal;
}

static int32_t CpswAle_configThreadPolicer(CSL_AleRegs *cpswAleRegBase,
                                           uint32_t aleFreqHz,
                                           const CpswAle_PolicerMatchParams *policerMatch,
                                           bool threadIdEnable,
                                           uint32_t threadId,
                                           CpswAle_SetPolicerThreadConfigOutArgs *outArgs)
{
    int32_t retVal;
    uint32_t policerIndex;

    retVal = CpswAle_getPolicerIndex(cpswAleRegBase,
                                     policerMatch,
                                     &policerIndex);

    if (retVal != CPSW_SOK)
    {
        CpswAle_SetPolicerEntryOutArgs policerSetOutArgs;

        retVal = CpswAle_setPolicer(cpswAleRegBase,
                                    policerMatch,
                                    CPSW_ALE_PEAKBITRATE_DISABLE,
                                    CPSW_ALE_COMMITBITRATE_DISABLE,
                                    aleFreqHz,
                                    FALSE,
                                    CPSW_ALE_THREADID_INVALID,
                                    &policerSetOutArgs);
        if (CPSW_SOK == retVal)
        {
            policerIndex = policerSetOutArgs.policerEntryIndex;
        }
    }

    if (CPSW_SOK == retVal)
    {
        if (threadIdEnable)
        {
            retVal = CpswAle_setThreadPolicerConfig(cpswAleRegBase, policerIndex, threadId);
        }
        else
        {
            retVal = CpswAle_disableThreadPolicerConfig(cpswAleRegBase, policerIndex);
        }
    }

    if (CPSW_SOK == retVal)
    {
        retVal = CpswAle_getPolicer(cpswAleRegBase, policerMatch, outArgs, aleFreqHz);
    }

    return retVal;
}

static int32_t CpswAle_getPortMacAddr(CSL_AleRegs *cpswAleRegBase,
                                      uint32_t portNum,
                                      CpswAle_MacAddrInfo *macAddrBuffer,
                                      uint32_t maxNumAddr,
                                      uint32_t *portTotalMacAddr)
{
    uint32_t i;
    int32_t ret = CPSW_EALE_ADDRESS_NOT_FOUND;
    uint8_t aleEntryMacAddr[6];
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);
    uint32_t macAddrBufferIndex;

    *portTotalMacAddr = 0;
    macAddrBufferIndex = 0;
    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;
        uint32_t reg0, reg1, reg2;

        /* read from the ALE table */
        CSL_CPSW_getAleTableEntry(cpswAleRegBase, i, &reg0, &reg1, &reg2);

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        /* We only need to check the entry type and address fields */

        if ((value == CSL_ALE_ENTRYTYPE_ADDRESS) ||
            (value == CSL_ALE_ENTRYTYPE_VLANADDRESS))
        {
            CSL_CPSW_mapTableWord2MacAddr(reg0, reg1, aleEntryMacAddr, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
            if (!CpswAle_isMcastAddr((CpswAle_MacAddrType_t)&aleEntryMacAddr))
            {
                if (value == CSL_ALE_ENTRYTYPE_ADDRESS)
                {
                    CSL_CPSW_ALE_UNICASTADDR_ENTRY entry;

                    CSL_CPSW_getAleUnicastAddrEntry(cpswAleRegBase, i, &entry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (portNum == entry.portNumber)
                    {
                        *portTotalMacAddr = *portTotalMacAddr + 1;
                        if (macAddrBufferIndex < maxNumAddr)
                        {
                            CpswAle_MacAddrInfo *destEntry =
                                &macAddrBuffer[macAddrBufferIndex];

                            CPSW_UTILS_ARRAY_COPY(destEntry->addr, entry.macAddress);
                            destEntry->vlanId = 0;
                            macAddrBufferIndex++;
                        }
                    }
                }
                else
                {
                    CSL_CPSW_ALE_VLANUNICASTADDR_ENTRY entry;

                    CSL_CPSW_getAleVlanUnicastAddrEntry(cpswAleRegBase, i,
                                                        &entry,
                                                        CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (portNum == entry.portNumber)
                    {
                        *portTotalMacAddr = *portTotalMacAddr + 1;
                        if (macAddrBufferIndex < maxNumAddr)
                        {
                            CpswAle_MacAddrInfo *destEntry =
                                &macAddrBuffer[macAddrBufferIndex];

                            CPSW_UTILS_ARRAY_COPY(destEntry->addr, entry.macAddress);
                            destEntry->vlanId = entry.vlanId;
                            macAddrBufferIndex++;
                        }
                    }
                }
            }
        }
    }

    return ret;
}

int32_t CpswAle_disablePortMirror(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t regVal;

    regVal = CSL_CPSW_getAleControlReg(cpswAleRegBase);

    CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_SEN, FALSE);
    CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_DEN, FALSE);
    CSL_FINS(regVal, ALE_ALE_CONTROL_MIRROR_MEN, FALSE);

    CSL_CPSW_setAleControlReg(cpswAleRegBase, regVal);

    return CPSW_SOK;
}

static int32_t CpswAle_disableBcastMcastLimit(CSL_AleRegs *cpswAleRegBase)
{
    CSL_CPSW_disableAleRateLimit(cpswAleRegBase);
    return CPSW_SOK;
}

static int32_t CpswAle_findVlanMaskMuxEntry(CSL_AleRegs *cpswAleRegBase,
                                            uint32_t matchMask,
                                            uint32_t *entry)
{
    uint32_t i;
    uint32_t entryMask;
    int32_t retVal;

    /* The first ALE Mask Mux is a readonly register set to ALL_PORTS mask */
    for (i = 0; i < CPSW_ALE_VLAN_MASK_MUX_COUNT(cpswAleRegBase); i++)
    {
        CSL_CPSW_getAleVlanMaskMuxEntryReg(cpswAleRegBase, i, &entryMask);
        if (matchMask == entryMask)
        {
            break;
        }
    }

    if (i < CPSW_ALE_VLAN_MASK_MUX_COUNT(cpswAleRegBase))
    {
        *entry = i;
        retVal = CPSW_SOK;
    }
    else
    {
        retVal = CPSW_EALE_VLANMASK_NOT_FOUND;
    }

    return retVal;
}

static int32_t CpswAle_getVlanMaskMuxFreeEntry(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t *entry)
{
    uint32_t vlanMaskUsedList[7];
    uint32_t i;
    int32_t retVal;
    const uint32_t aleTableDepth = CpswAle_getMaxAleEntries(cpswAleRegBase);

    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_UTILS_ARRAYSIZE(vlanMaskUsedList) ==
                                  CPSW_UTILS_ARRAYSIZE(cpswAleRegBase->I1_ALE_MSK_MUX1));

    memset(vlanMaskUsedList, 0, sizeof(vlanMaskUsedList));

    for (i = 0; i < aleTableDepth; i++)
    {
        uint32_t value;

        value = CSL_CPSW_getALEEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
        if (value == CSL_ALE_ENTRYTYPE_POLICER)
        {
            switch (CSL_CPSW_getALEPolicerEntryType(cpswAleRegBase, i, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase)))
            {
                case CSL_ALE_POLICER_ENTRYTYPE_VLAN:
                {
                    CSL_CPSW_ALE_VLAN_ENTRY vlanEntry;

                    CSL_CPSW_getAleVlanEntry(cpswAleRegBase, i, &vlanEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    if (vlanEntry.unRegMcastFloodIndex > 0)
                    {
                        vlanMaskUsedList[CPSW_ALE_VLANMASKMUX2REG(vlanEntry.unRegMcastFloodIndex)] = 1;
                    }

                    if (vlanEntry.regMcastFloodIndex > 0)
                    {
                        vlanMaskUsedList[CPSW_ALE_VLANMASKMUX2REG(vlanEntry.regMcastFloodIndex)] = 1;
                    }

                    break;
                }

                case CSL_ALE_POLICER_ENTRYTYPE_OVLAN:
                {
                    CSL_CPSW_ALE_OUTER_VLAN_ENTRY vlanEntry;

                    CSL_CPSW_getAleOutVlanEntry(cpswAleRegBase, i, &vlanEntry, CpswAle_getAleTableType((uintptr_t)cpswAleRegBase));
                    vlanMaskUsedList[CPSW_ALE_VLANMASKMUX2REG(vlanEntry.unRegMcastFloodIndex)] = 1;
                    vlanMaskUsedList[CPSW_ALE_VLANMASKMUX2REG(vlanEntry.regMcastFloodIndex)] = 1;
                    break;
                }

                default:
                    ; /* Do nothing for other policer types */
            }
        }
    }

    /* Entry 0 is readonly and cannot be used as free entry */
    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(vlanMaskUsedList); i++)
    {
        if (vlanMaskUsedList[i] != 1)
        {
            break;
        }
    }

    if (i < CPSW_UTILS_ARRAYSIZE(vlanMaskUsedList))
    {
        *entry = i;
        retVal = CPSW_SOK;
    }
    else
    {
        retVal = CPSW_EALE_TABLE_FULL;
    }

    return retVal;
}

static int32_t CpswAle_addVlanMask(CSL_AleRegs *cpswAleRegBase,
                                   uint32_t mask,
                                   uint32_t *entry)
{
    int32_t retVal;
    uint32_t maskEntryIdx;

    retVal = CpswAle_findVlanMaskMuxEntry(cpswAleRegBase, mask, &maskEntryIdx);

    if (retVal != CPSW_SOK)
    {
        retVal = CpswAle_getVlanMaskMuxFreeEntry(cpswAleRegBase, &maskEntryIdx);
        if (retVal == CPSW_SOK)
        {
            CSL_CPSW_setAleVlanMaskMuxEntryReg(cpswAleRegBase, maskEntryIdx, mask);
        }
    }

    if (retVal == CPSW_SOK)
    {
        *entry = maskEntryIdx;
    }

    return retVal;
}

static int32_t CpswAle_getPortState(CSL_AleRegs *cpswAleRegBase,
                                    uint32_t portNum,
                                    CPSW_ALE_PORTSTATE_E *pPortState)
{
    CSL_CPSW_getAlePortState(cpswAleRegBase, portNum, (CSL_CPSW_ALE_PORTSTATE *)pPortState);
    return CPSW_SOK;
}

static bool CpswAle_isVlanMaskMuxRegSupported(CSL_AleRegs *cpswAleRegBase)
{
    bool vlanMsk08, vlanMsk12;

    CSL_CPSW_getAleStatusVlanMask(cpswAleRegBase, &vlanMsk08, &vlanMsk12);

    if ((vlanMsk08 == FALSE) && (vlanMsk12 == FALSE))
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

static bool CpswAle_isIPHdrWhiteListSupported(CSL_AleRegs *cpswAleRegBase)
{
    const CpswSoc_CpswIpInfo *ipInfo = CpswAle_getCpswIpInfoFromBaseAddr((uintptr_t)cpswAleRegBase);
    bool isIpWhiteListSupported;

    if (ipInfo->featureBitmap & CPSW_FEATURE_IPHDRWHITELIST)
    {
        isIpWhiteListSupported = TRUE;
    }
    else
    {
        isIpWhiteListSupported = FALSE;
    }

    return isIpWhiteListSupported;
}

static bool CpswAle_isFlowPrioritySupported(CSL_AleRegs *cpswAleRegBase)
{
    const CpswSoc_CpswIpInfo *ipInfo = CpswAle_getCpswIpInfoFromBaseAddr((uintptr_t)cpswAleRegBase);
    bool isFlowPrioritySupported;

    if (ipInfo->featureBitmap & CPSW_FEATURE_FLOWPRIORITY)
    {
        isFlowPrioritySupported = TRUE;
    }
    else
    {
        isFlowPrioritySupported = FALSE;
    }

    return isFlowPrioritySupported;
}

static void CpswAle_printRegs(CSL_AleRegs *cpswAleRegBase)
{
    uint32_t *regAddr = (uint32_t *)cpswAleRegBase;
    uint32_t regIdx = 0;

    while (((uintptr_t)regAddr) < ((uintptr_t)cpswAleRegBase + sizeof(*cpswAleRegBase)))
    {
        if (*regAddr)
        {
            CpswUtils_printf("ALE: %u: 0x%08x\n", regIdx, *regAddr);
        }

        regAddr++;
        regIdx++;
    }
}

static int32_t CpswAle_getPolicerControl(CSL_AleRegs *cpswAleRegBase,
                                         uint32_t aleFreqHz,
                                         CpswAle_GetPolicerGlobalConfigOutArgs *policerCfg)
{
    CSL_CPSW_ALE_POLICER_CONTROL policerControl;

    CSL_CPSW_getAlePolicerControlReg(cpswAleRegBase, &policerControl);
    policerCfg->policingEnable = policerControl.policingEnable;
    policerCfg->redDropEnable = policerControl.redDropEnable;
    policerCfg->yellowDropEnable = policerControl.yellowDropEnable;
#ifndef __aarch64__
    /* Direct assignment of CSL enums to CPSW enums. Ensure by compile time check that the enum values are equivalent */

    /* This check is disabled for GCC compiler due to warning -Werror=enum-compare so disabled for GCC compiler.
     * However the check must be there so that any change in CPSW or CSL enum values are flagged as compile time error
     */
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_100 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_100);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_50 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_50);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_33 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_33);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_25 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_25);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_20 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_20);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_17 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_17);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_14 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_14);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_YELLOWTHRESH_DROP_PERCENT_13 == CSL_ALE_POLICER_CONTROL_YELLOWTHRESH_DROP_PERCENT_13);
#endif
    policerCfg->yellowThresh = (CPSW_ALE_POLICER_YELLOWTHRESH_E)policerControl.yellowDropThresh;
    /* Direct assignment of CSL enums to CPSW enums. Ensure by compile time check that the enum values are equivalent */
#ifndef __aarch64__
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_NOMATCH_MODE_GREEN == CSL_ALE_POLICER_CONTROL_POLICING_MATCH_MODE_NOMATCH_GREEN);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_NOMATCH_MODE_YELLOW == CSL_ALE_POLICER_CONTROL_POLICING_MATCH_MODE_NOMATCH_YELLOW);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_NOMATCH_MODE_RED == CSL_ALE_POLICER_CONTROL_POLICING_MATCH_MODE_NOMATCH_RED);
    CPSW_UTILS_COMPILETIME_ASSERT(CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER ==
                                  CSL_ALE_POLICER_CONTROL_POLICING_MATCH_MODE_NOMATCH_ENTRY0STATE);
#endif
    policerCfg->policerNoMatchMode = (CPSW_ALE_POLICER_NOMATCH_MODE_E)policerControl.policeMatchMode;

    if (CPSW_ALE_POLICER_NOMATCH_MODE_UNREGULATED_TRAFFIC_POLICER ==
        policerCfg->policerNoMatchMode)
    {
        uint32_t retVal;

        retVal = CpswAle_getNoMatchEntry0Policer(cpswAleRegBase,
                                                 aleFreqHz,
                                                 &policerCfg->noMatchPolicer.peakRateInBitsPerSec,
                                                 &policerCfg->noMatchPolicer.commitRateInBitsPerSec);
        Cpsw_assert(retVal == CPSW_SOK);
    }

    return CPSW_SOK;
}

static int32_t CpswAle_getNoMatchEntry0Policer(CSL_AleRegs *cpswAleRegBase,
                                               uint32_t aleFreq,
                                               uint32_t *pPeakRateInBitsPs,
                                               uint32_t *pCommitRateInBitsPs)
{
    CSL_CPSW_ALE_POLICER_ENTRY polCfg;

    CSL_CPSW_getAlePolicerEntry(cpswAleRegBase,
                                CPSW_ALE_NOMATCH_ENTRY0_POLICER_INDEX,
                                &polCfg);
    if (polCfg.validBitmap & CSL_CPSW_ALE_POLICER_PIR_VALID)
    {
        *pPeakRateInBitsPs = CpswAle_mapIdlIncVal2Bw(aleFreq, polCfg.pirIdleIncVal);
    }
    else
    {
        *pPeakRateInBitsPs = 0;
    }

    if (polCfg.validBitmap & CSL_CPSW_ALE_POLICER_CIR_VALID)
    {
        *pCommitRateInBitsPs = CpswAle_mapIdlIncVal2Bw(aleFreq, polCfg.cirIdleIncVal);
    }
    else
    {
        *pCommitRateInBitsPs = 0;
    }

    return CPSW_SOK;
}

static int32_t CpswAle_blockClassifierToHostPort(CSL_AleRegs *cpswAleRegBase,
                                                 uint32_t aleFreqHz,
                                                 const CpswAle_PolicerMatchParams *policerMatch,
                                                 CpswAle_BlockClassifierHostPortOutArgs *outArgs)
{
    CpswAle_SetInterVLANConfigInArgs interVlanCfg;
    int32_t retVal;

    memset(&interVlanCfg, 0, sizeof(interVlanCfg));

    interVlanCfg.policerMatch = *policerMatch;
    interVlanCfg.destPortMask = CPSW_ALE_BLOCKCLASSIFIER_PORTMASK;
    interVlanCfg.egressTrunkIndex = 0;
    interVlanCfg.enableTTLCheck = false;
    interVlanCfg.routeIndex = CPSW_ALE_BLOCKCLASSIFIER_ROUTEINDEX;
    retVal = CpswAle_setInterVlanConfig(cpswAleRegBase, aleFreqHz, &interVlanCfg,
                                        outArgs);
    return retVal;
}
